function [r,s,p] = rates(err, hvec,plt,nvec,sym,nmin)

% RATES computes convergence rates of finite difference schemes
%
%    [R,S,P] = RATES(ERR,HVEC,C,NVEC,SYM) computes the convergence rate
%    from a series of error norms computed N different grids. ERR is
%    assumed to be an Nx3 vector, with the first column corresonding to the
%    one-norm error, the second column corresponding to the 2-norm error,
%    and the third column corresponding to the inf-norm error.  The rows of
%    ERR correspond to errors solutions on different grids.  HVEC is a
%    vector of decreasing mesh widths, C is the a number (1,2,or 3)
%    indicating the column of ERR to plot, NVEC is a vector of grid
%    resolutions (usually LEN/HVEC, where LEN is the domain length).  NVEC
%    is used to label the x-axis.  Finally, SYM is the symbol to use on the
%    LOGLOG plot of the errors.
%
%    [R,S,P] = RATES(ERR,HVEC) only prints the convergence rates.
%
%    The return arguments are R, the array of convergence rates, S, the
%    estimate of the convergence rate, using a best-fit line through the
%    error data, and P, a handle to the line object plotted.
%


[len,c] = size(err);

if (nargin < 6)
  nmin = 1;
  if (nargin < 5)
    sym = 'b*';
    if (nargin < 3)
      plt = 0;
      if (nargin < 2)
	hvec = 0.5.^[0:len-1]';
      end;
    end;
  end;
end;

d1 = hvec(1:len-1)./hvec(2:len);
divvec = kron(d1,ones(1,c));

% Vector of convergence rates returned.
r = log(err(1:len-1,:)./err(2:len,:))./log(divvec);

plt_str = strvcat('one-norm','two-norm','inf-norm');

if (plt > 0)
  p(1) = loglog(hvec,err(:,plt),['',sym]);
  hold on;
  A = [log(hvec(nmin:end)) ones(len-nmin+1,1)];
  b = log(err(nmin:end,plt));
  pc = inv(A'*A)*A'*b;
  hold on;
  p(2) = loglog(hvec(nmin:end),exp(pc(2))*hvec(nmin:end).^pc(1),'r--');

  str = [plt_str(plt,:),sprintf(';  Rate = %6.2f',pc(1))];
  title(str);
  set(gca,'XTick',flipud(hvec));
  set(gca,'XTickLabel',flipud(nvec));
  set(gca,'XLim',[hvec(end)/2, hvec(1)*2]);
  set(gca,'YLim',[1e-4 1]);
  set(gca,'XScale','log');
  set(gca,'YScale','log');
  s = pc(1);
end;
