/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMR_LEVEL_GODUNOV_H_
#define _AMR_LEVEL_GODUNOV_H_

#include "FArrayBox.H"
#include "LevelData.H"
#include "AMRLevel.H"
#include "CoarseAverage.H"
#include "FineInterp.H"
#include "LevelFluxRegister.H"

#include "LevelGodunov.H"
#include "PhysIBC.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "DisjointBoxLayout.H"

/// AMR Godunov
/**
 */
class AMRLevelGodunov : public AMRLevel
{
public:
  ///
  AMRLevelGodunov();

  ///
  virtual
  ~AMRLevelGodunov();

  /// define
  virtual void define(AMRLevel*  a_coarser_level_ptr,
                      const Box& a_problem_domain,
                      int        a_level,
                      int        a_ref_ratio);

  /// define
  virtual void define(AMRLevel*  a_coarser_level_ptr,
                      const ProblemDomain& a_problem_domain,
                      int        a_level,
                      int        a_ref_ratio);


  /// advance by one timestep
  virtual Real advance();

  /// things to do after a timestep
  virtual void postTimeStep();

  /// create tags
  virtual void tagCells(IntVectSet& a_tags) ;

  /// create tags at initialization
  virtual void tagCellsInit(IntVectSet& a_tags);

  /// regrid
  virtual void regrid(const Vector<Box>& a_new_grids);

  /// initialize grids
  virtual void initialGrid(const Vector<Box>& a_new_grids);

  /// initialize data
  virtual void initialData();

  /// things to do after initialization
  virtual void postInitialize();

  /// set physical boundary conditions object
  virtual void physIBC(RefCountedPtr<PhysIBC> a_phys_ibc_ptr);

#ifdef HDF5
  /// write checkpoint header
  virtual void writeCheckpointHeader(HDF5Handle& a_handle) const;

  /// write checkpoint data for this level
  virtual void writeCheckpointLevel(HDF5Handle& a_handle) const;

  /// read checkpoint header
  virtual void readCheckpointHeader(HDF5Handle& a_handle);

  /// read checkpoint data for this level
  virtual void readCheckpointLevel(HDF5Handle& a_handle);

  /// write plotfile header
  virtual void writePlotHeader(HDF5Handle& a_handle) const;

  /// write plot file data for this level
  virtual void writePlotLevel(HDF5Handle& a_handle) const;
#endif

  /// compute dt
  virtual Real computeDt();

  /// compute dt with initial data
  virtual Real computeInitialDt();

  /// CFL
  virtual void CFL(Real a_cfl);

  /// refinement threshold
  virtual void refinementThreshold(Real a_refine_thresh);

  /// tag buffer size
  void tagBufferSize(int a_tag_buffer_size);

protected:
  DisjointBoxLayout loadBalance(const Vector<Box>& a_grids);

  Real calculateMass() const;

  void levelSetup(const DisjointBoxLayout& a_level_domain);

protected:
  // state vector at old time
  LevelData<FArrayBox> m_state_old;

  // state vector at new time
  LevelData<FArrayBox> m_state_new;

  // number of components of m_state
  int m_num_comps;

  // names of components
  Vector<string> m_state_names;

  // grid spacing
  Real m_dx;

  // cfl number
  Real m_cfl;

  // new time step
  Real m_dtNew;

  // averaging from coarse to fine level
  CoarseAverage m_coarse_average;

  // interpolation from fine to coarse level
  FineInterp m_fine_interp;

  // level solver
  LevelGodunov m_levelgodunov;

  // flux register
  LevelFluxRegister m_flux_register;

  // boundary conditions
  RefCountedPtr<PhysIBC> m_phys_ibc_ptr;

  // refinement threshold for gradient
  Real m_refine_thresh;

  // tag buffer size
  int m_tag_buffer_size;
};

#endif
