/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LoHiSide.H"

#include "ExplosionIBC.H"
#include "ExplosionF_F.H"

#include "LoHiCenter.H"

ExplosionIBC::ExplosionIBC()
{
    m_isFortranCommonSet = false;
}

ExplosionIBC::ExplosionIBC(const Real&     a_gamma,
                           const Real&     a_ms,
                           const RealVect& a_center,
                           const Real&     a_size,
                           const RealVect& a_velocity)
{
    setFortranCommon(a_gamma,
                     a_ms,
                     a_center,
                     a_size,
                     a_velocity);
}

void ExplosionIBC::setFortranCommon(const Real&     a_gamma,
                                    const Real&     a_ms,
                                    const RealVect& a_center,
                                    const Real&     a_size,
                                    const RealVect& a_velocity)
{
    assert(m_isFortranCommonSet == false);

    FORT_SETEXPLOSION(CHF_CONST_REAL(a_gamma),
                      CHF_CONST_REAL(a_ms),
                      CHF_CONST_REALVECT(a_center),
                      CHF_CONST_REAL(a_size),
                      CHF_CONST_REALVECT(a_velocity));

    m_isFortranCommonSet = true;
}

void ExplosionIBC::fluxBC(FArrayBox&       a_flux,
                          const FArrayBox& a_conState,
                          const FArrayBox&       a_divu,
                          const FArrayBox&       a_qgdnv,
                          const FArrayBox&       a_prim,
                          const int&        a_dir,
                          const Real&      a_time,
                          const Real&      a_dt,
                          const Real&      a_dx,
                          const Box&       a_DProblem
                          ) const
{
  ProblemDomain physdomain(a_DProblem);
  fluxBC(a_flux, a_conState, a_divu, a_qgdnv, a_prim, a_dir, a_time, a_dt, a_dx, physdomain);
}

void ExplosionIBC::fluxBC(FArrayBox&           a_flux,
                          const FArrayBox&     a_conState,
                          const FArrayBox&           a_divu,
                          const FArrayBox&           a_qgdnv,
                          const FArrayBox&           a_prim,
                          const int&            a_dir,
                          const Real&          a_time,
                          const Real&          a_dt,
                          const Real&          a_dx,
                          const ProblemDomain& a_DProblem
                          ) const

{
  assert(m_isFortranCommonSet == true);

  // don't do anything if periodic
  if (!a_DProblem.isPeriodic(a_dir))
    {
      for (int sign = -1; sign < 2; sign += 2)
        {
          Box fluxBox = a_flux.box();
          Box tmp = fluxBox;
          
          tmp.shiftHalf(a_dir,sign);
          if (!a_DProblem.contains(tmp))
            {
              tmp &= a_DProblem;
              
              Box dBoundary;
              
              if (sign == -1) 
                {
                  dBoundary = bdryLo(tmp,a_dir);
                }
              else
                {
                  dBoundary = bdryHi(tmp,a_dir);
                }
              dBoundary.shiftHalf(a_dir,1);
              FArrayBox& qgdnv = (FArrayBox&)a_qgdnv;
              FArrayBox& divu  = (FArrayBox&)a_divu;

              a_flux.shiftHalf(a_dir,1);
              qgdnv.shiftHalf(a_dir,1);
              divu.shiftHalf(a_dir,1);
              
              FORT_SOLIDBC(CHF_BOX(dBoundary),
                           CHF_FRA(a_flux),
                           CHF_CONST_FRA(a_conState),
                           CHF_CONST_FRA(qgdnv),
                           CHF_CONST_FRA1(divu,0),
                           CHF_CONST_REAL(a_dx),
                           CHF_CONST_INT(a_dir),
                           CHF_INT(sign));

              a_flux.shiftHalf(a_dir,-1);
              qgdnv.shiftHalf(a_dir,-1);
              divu.shiftHalf(a_dir,-1);
            }
        }
    }
}

void ExplosionIBC::setBdrySlopes(FArrayBox&       a_dq,
                                 const Box&       a_DProblem,
                                 const Real&      a_dx,
                                 const int&       a_dir,
                                 const FArrayBox& a_prim
                                 ) const
{
  ProblemDomain physdomain(a_DProblem);
  setBdrySlopes(a_dq, physdomain, a_dx, a_dir, a_prim);
}

void ExplosionIBC::setBdrySlopes(FArrayBox&           a_dq,
                                 const ProblemDomain& a_DProblem,
                                 const Real&          a_dx,
                                 const int&           a_dir,
                                 const FArrayBox&     a_prim
                                 ) const
{
  assert(m_isFortranCommonSet == true);

  // in periodic case, this doesn't do anything
  if (!a_DProblem.isPeriodic(a_dir))
    {
      Box loBox,hiBox,centerBox,domain;
      int haslo,hashi;
      Box slopeBox = a_dq.box()&a_DProblem;
      loHiCenter(slopeBox,a_DProblem,a_dir,
           loBox,hiBox,centerBox,domain,haslo,hashi);
      if ((haslo != 0) || (hashi != 0))
        {
          FArrayBox& Prim = (FArrayBox& ) a_prim;
          
          FORT_SLOPEBCS(CHF_BOX(loBox),
                        CHF_BOX(hiBox),
                        CHF_CONST_INT(a_dir),
                        CHF_CONST_INT(haslo),
                        CHF_CONST_INT(hashi),
                        CHF_CONST_REAL(a_dx),
                        CHF_FRA(Prim),
                        CHF_FRA(a_dq));
        }
    }
}

void ExplosionIBC::initialize(LevelData<FArrayBox>& a_conState,
                              const Box&                   a_DProblem,
                              const Real&                  a_dx) const
{
  ProblemDomain physdomain(a_DProblem);
  initialize(a_conState, a_DProblem, a_dx);
}

void ExplosionIBC::initialize(LevelData<FArrayBox>& a_conState,
                              const ProblemDomain&        a_DProblem,
                              const Real&                  a_dx) const
{
  assert(m_isFortranCommonSet == true);

  DataIterator dit = a_conState.boxLayout().dataIterator();
  
  for (dit.begin(); dit.ok(); ++dit)
  {
    FArrayBox& state = a_conState[dit()];
    Box dCalc = state.box();
    dCalc &= a_DProblem;
    
    FORT_EXPLOSIONINIT(CHF_BOX(dCalc),
                       CHF_CONST_FRA(state),
                       CHF_CONST_REAL(a_dx));
  }
}
