/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LevelGodunov.H"
#include "DisjointBoxLayout.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "LevelFluxRegister.H"
#include "PiecewiseLinearFillPatch.H"
#include "ExtrapFillPatch.H"

#include "PhysIBC.H"
#include "LGintegrator.H"
#include "GodunovF_F.H"
#include "LoHiSide.H"

#include "LoHiCenter.H"

LevelGodunov::LevelGodunov()
{
  m_defined = false;
  m_bc = NULL;
  m_dx = 0.0;
  m_refine_coarse = 0;
}

LevelGodunov::~LevelGodunov()
{
}

void LevelGodunov::define(const DisjointBoxLayout& a_thisDisjointBoxLayout,
                          const DisjointBoxLayout& a_coarserDisjointBoxLayout,
                          const Box&               a_DProblem,
                          const int&               a_numGhost,
                          const int&               a_nRefine,
                          const Real&              a_dx,
                          const RefCountedPtr<PhysIBC>&   a_bc)
{
  ProblemDomain physdomain(a_DProblem);

  define(a_thisDisjointBoxLayout,a_coarserDisjointBoxLayout,physdomain,
         a_numGhost,a_nRefine,a_dx,a_bc);
}

void LevelGodunov::define(const DisjointBoxLayout& a_thisDisjointBoxLayout,
                          const DisjointBoxLayout& a_coarserDisjointBoxLayout,
                          const ProblemDomain&     a_DProblem,
                          const int&                a_numGhost,
                          const int&                a_nRefine,
                          const Real&              a_dx,
                          const RefCountedPtr<PhysIBC>&   a_bc)
{
  (DisjointBoxLayout&) m_ba = a_thisDisjointBoxLayout;
  m_bc = a_bc;

  m_dx = a_dx;
  m_refine_coarse = a_nRefine;

  m_numGhost = a_numGhost;

  const int innerRadius = 1;
  const int outerRadius = m_numGhost;

  m_numCons        = CNUM;
  m_numPrim        = QNUM;
  m_numSlopes      = QSLOPE;
  m_numFluxes      = CNUM;
  m_DProblem       = a_DProblem;
  m_DProblemNodal  = a_DProblem.domainBox();
  m_DProblemNodal.surroundingNodes();

  m_patcher.define(a_thisDisjointBoxLayout,
                   a_coarserDisjointBoxLayout,
                   m_numCons,
                   coarsen(a_DProblem,a_nRefine),
                   a_nRefine,
                   innerRadius);

  m_extrap_filler.define(a_thisDisjointBoxLayout,
                         a_DProblem,
                         Interval(innerRadius+1,outerRadius));

  m_leveque.define(a_thisDisjointBoxLayout,
                   a_DProblem,
                   a_dx,
                   m_numCons);

  m_mask.define(a_thisDisjointBoxLayout,
                a_DProblem,
                innerRadius);

  for (int dir = 0; dir < SpaceDim; ++dir)
    {
      m_edgeType[dir] = IndexType::TheCellType();
      m_edgeType[dir].setType(dir,IndexType::NODE);
    }

  m_defined = true;
}

// The function advances the solution by "a_dt" by using a operator splitting
// method.  It uses half "a_dt" to sweep in one ordering of directions and
// half "a_dt" to sweep in the reverse ordering (Strang alternation).  As a
// result the new (max) time step it computes is twice the maximum time step
// for a single sweep.
Real LevelGodunov::step(LevelData<FArrayBox>&       a_U,
                        LevelFluxRegister&          a_myFluxRegister,
                        LevelFluxRegister&          a_fineFluxRegister,
                        const LevelData<FArrayBox>& a_UCoarseOld,
                        const LevelData<FArrayBox>& a_UCoarseNew,
                        const Real&                  a_time,
                        const Real&                  a_TCold,
                        const Real&                  a_TCnew,
                        const Real&                  a_dt
                        )
{
  if (!m_defined){
    MayDay::Error("LevelGodunov::m_defined == false");
  }

  // clear boundary registers
  if (a_myFluxRegister.isDefined())
  {
    a_myFluxRegister.setToZero();
  }

  // Half the time step and do two sweeps.
  Real dt = 0.5*a_dt;
  bool evenParity = true;

  for (int i = 0; i < 2; i++)
  {
    // Step one, declare and fill with 'zero' the diffusions coefficient
    // for artificial viscosity
    int nGhost = m_numGhost;
    LayoutData<FArrayBox> divu[SpaceDim];
    for (int dir = 0; dir < SpaceDim; ++dir)
    {
        divu[dir].define(m_ba);
        DataIterator ditv = m_ba.dataIterator();

        for (ditv.begin(); ditv.ok(); ++ditv)
        {
            Box divbox = surroundingNodes(m_ba[ditv()],dir);
            divu[dir][ditv()].resize(divbox,1);
            divu[dir][ditv()].setVal(0.0);
        }
    }

    IntVect ivGhost = IntVect::Unit;
    ivGhost *= nGhost;

    LevelData<FArrayBox> UTemp(m_ba,a_U.nComp(),ivGhost);
    Interval consInterval = UTemp.interval();
    a_U.copyTo(consInterval,UTemp,consInterval);

    // Begin Loop over directions

    for (int dir = 0; dir < SpaceDim; dir++)
    {
      // Use Strang alternation of ordering of fractional steps to preserve
      // second-order accuracy in time.

      int sweepDir;

      if (evenParity)
      {
        sweepDir = dir;
      }
      else
      {
        sweepDir = SpaceDim - dir - 1;
      }

      // PC: We need a convention of what to do about coarse data being
      // defined.

      // Declare LevelData<FArrayBox> for temporary storage in this sweep

      if (dir == 0)
      {
        // Fill UTemp using fillInterp, and copyTo.

        if (a_UCoarseOld.isDefined())
        {
          Real alpha = (a_time - a_TCold) / (a_TCnew - a_TCold);

          // truncate to the range [0,1] to remove floating-point subtraction
          // roundoff effects
          Real eps = 0.04 * dt / m_refine_coarse;

          if (Abs(alpha) < eps)
          {
            alpha = 0.0;
          }

          if (Abs(1.0-alpha) < eps)
          {
            alpha = 1.0;
          }

          if (alpha < 0.0)
          {
            MayDay::Error( "LevelGodunov::step: alpha < 0.0");
          }

          if (alpha > 1.0)
          {
            MayDay::Error( "LevelGodunov::step: alpha > 1.0");
          }

          m_patcher.fillInterp(UTemp,
                               a_UCoarseOld,
                               a_UCoarseNew,
                               alpha,
                               0,0,m_numCons);
        }
      }

      m_extrap_filler.fillExtrap(UTemp,
                                 sweepDir,
                                 0,m_numCons);

      UTemp.exchange(consInterval);

      // Beginning of loop through grids.

      DataIterator dit = m_ba.dataIterator();

      for (dit.begin(); dit.ok(); ++dit)
      {
        // int i = dit().intCode();
        FArrayBox& conState = UTemp[dit()];

        // Current box
        Box curBox = m_ba[dit()];

        // Compute primitive variables.
        Box primBox = curBox;
        primBox.grow(sweepDir,nGhost-1).grow(1);
        primBox &= m_DProblem;

        // primitive variables
        FArrayBox primState(primBox,m_numPrim);

        // calculate primState
        consToPrim(conState,primState);

        // If this is the first time through the direction loop, form
        // Artificial viscosity coefficients for all three directions.

        if (dir == 0)
        {
          for (int dirVisc = 0; dirVisc < SpaceDim; ++dirVisc)
          {
            divU(primState,curBox,dirVisc,divu[dirVisc][dit()]);
          }
        }

        // Compute slopes.
        Box slopeBox = curBox;
        slopeBox.grow(sweepDir,1);
        slopeBox &= m_DProblem;
        FArrayBox dq(slopeBox,m_numSlopes);

        slope(primState,dit(),sweepDir,dq);

        // Compute predictor step to obtain extrapolated edge values.
        FArrayBox qlo;
        FArrayBox qhi;
        normalPred(primState,dq,dt,sweepDir,qlo,qhi);

        Box edgeBox = surroundingNodes(curBox,sweepDir);
        FArrayBox qGdnv(edgeBox,m_numPrim);
        riemann(qhi,qlo,sweepDir,qGdnv);

        // Compute fluxes and then apply artificial viscosity
        FArrayBox flux(edgeBox,m_numFluxes);
        fluxCalc(qGdnv,sweepDir,flux);
        artificialViscosity(qGdnv,conState,divu[sweepDir][dit()],sweepDir,flux);

        // No more changes are performed on 'flux' now in the conservative
        // scheme.  Here we invoke the user supplied PhysBC object to
        // apply the boundary conditions
        //
        m_bc->fluxBC(flux, conState,divu[sweepDir][dit()],
                     qGdnv,primState,sweepDir,a_time,dt,
                     m_dx,m_DProblem);

        // Perform conservative update, incrementing LevequeCorrection object.
        updateState(flux,sweepDir,dt,curBox,UTemp[dit()]);
        m_leveque.correctBoundary(conState,flux,dt,m_dx,sweepDir,dit());

        // increment flux registers, as appropriate. It seems that FluxRegister
        // will handle the case where is does nothing when there are no points,
        // but I'll leave the check here anyways for now.
        if (a_myFluxRegister.isDefined())
        {
          a_myFluxRegister.incrementCoarse(flux,dt,dit(),
                                           consInterval,consInterval,sweepDir);
        }

        if (a_fineFluxRegister.isDefined())
        {
          a_fineFluxRegister.incrementFine(flux,dt,dit(),consInterval,
                                           consInterval,sweepDir,Side::Lo);
          a_fineFluxRegister.incrementFine(flux,dt,dit(),consInterval,
                                           consInterval,sweepDir,Side::Hi);
        }
      } // end of dit loop over grids.
    } // end of dir loop over sweep directions.

    // Now that we have completed the updates of all
    // the patches in this sweep,
    // we copy the contents of UTemp into U.
    UTemp.copyTo(consInterval,a_U,consInterval);

    evenParity = !evenParity;
  }

  // Compute the new (max) dt - doubled because of the two sweeps being done
  // at one half this dt (see above).
  Real vmax = getMaxWaveSpeed(a_U);
  Real dtNew = 2.0 * m_dx / vmax;

  // gather dtNew
  Vector<Real> all_dt;
  gather(all_dt,dtNew,uniqueProc(SerialTask::compute));
  if (procID() == uniqueProc(SerialTask::compute))
  {
    dtNew = all_dt[0];
    for (int i = 1; i < all_dt.size(); ++i)
    {
      dtNew = Min(dtNew,all_dt[i]);
    }
  }
  broadcast(dtNew,uniqueProc(SerialTask::compute));

  return dtNew;
  // should pass back 'dtout', but I need a proper algorithm for this number
}

Real LevelGodunov::getMaxWaveSpeed(const LevelData<FArrayBox>& a_state)
{
  const DisjointBoxLayout& dba = a_state.disjointBoxLayout();
  Real speed = 0.0;

  DataIterator dit = dba.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
  {
    Real speedBox = 0.0;
    FArrayBox& stateFab = (FArrayBox &)a_state[dit()];
    Box bx = dba[dit()];

    FORT_MAXWAVESPEED(CHF_BOX(bx),
                      CHF_FRA(stateFab),
                      CHF_REAL(speedBox));

    speed = Max(speed,speedBox);
  }

  // gather speed
  Vector<Real> all_speeds;
  gather(all_speeds,speed,uniqueProc(SerialTask::compute));
  if (procID() == uniqueProc(SerialTask::compute))
  {
    speed = all_speeds[0];
    for (int i = 1; i < all_speeds.size (); ++i)
    {
      speed = Max(speed,all_speeds[i]);
    }
  }
  broadcast(speed,uniqueProc(SerialTask::compute));

  return speed;
}

void LevelGodunov::getFlattening(const FArrayBox& a_q,
                                 const DataIndex  a_di,
                                 const int        a_dir,
                                 FArrayBox&       a_flatCoef) const
{
  // const Real STRONG_SHOCK = 0.33;

  Box gradBox,loBox,hiBox,centerBox;
  int haslo,hashi;

  loHiCenter(a_q.box(),m_DProblem,a_dir,
             loBox,hiBox,centerBox,gradBox,haslo,hashi);

  FArrayBox dp(gradBox,1);
  FArrayBox du(gradBox,1);

  FORT_GETGRAD(CHF_BOX(centerBox),
               CHF_BOX(loBox),
               CHF_BOX(hiBox),
               CHF_CONST_INT(a_dir),
               CHF_CONST_INT(haslo),
               CHF_CONST_INT(hashi),
               CHF_CONST_FRA1(a_q,QPRES),
               CHF_FRA1(dp,0));

  FORT_GETGRAD(CHF_BOX(centerBox),
               CHF_BOX(loBox),
               CHF_BOX(hiBox),
               CHF_CONST_INT(a_dir),
               CHF_CONST_INT(haslo),
               CHF_CONST_INT(hashi),
               CHF_CONST_FRA1(a_q,QVELX+a_dir),
               CHF_FRA1(du,0));

  FArrayBox pmin(gradBox,1);

  FORT_MIN3PTS(CHF_BOX(centerBox),
               CHF_BOX(loBox),
               CHF_BOX(hiBox),
               CHF_CONST_INT(a_dir),
               CHF_CONST_INT(haslo),
               CHF_CONST_INT(hashi),
               CHF_CONST_FRA1(a_q,QPRES),
               CHF_FRA1(pmin,0));

  for (RadiusMaskIterator center(m_mask,a_di,a_dir); center.ok(); center.next())
  {
    // IntVect iv = center();
    int sign = center.sign();
    const IntVect& inside = center.inside();
    dp(center()) = sign * (a_q(center(),QPRES) - a_q(inside,QPRES));
    du(center()) = sign * (a_q(center(),QVELX+a_dir)
                         - a_q(inside  ,QVELX+a_dir));
    pmin(center()) = Min(a_q(center(),QPRES),a_q(inside,QPRES));
  }

  Box grad2Box;
  loHiCenter(gradBox,m_DProblem,a_dir,
             loBox,hiBox,centerBox,grad2Box,haslo,hashi);

  FArrayBox d2p(grad2Box,1);

  FORT_GETDPTWO(CHF_BOX(centerBox),
                CHF_BOX(loBox),
                CHF_BOX(hiBox),
                CHF_CONST_INT(a_dir),
                CHF_CONST_INT(haslo),
                CHF_CONST_INT(hashi),
                CHF_FRA1(dp,0),
                CHF_FRA1(d2p,0));

  for (RadiusMaskIterator center(m_mask,a_di,a_dir); center.ok(); center.next())
  {
    // int sign = center.sign();
    const IntVect& inside = center.inside();
    d2p(center()) = dp(center()) + dp(inside);
  }

  FArrayBox chitwiddle(grad2Box,1);

  FORT_GETFLAT(CHF_BOX(grad2Box),
               CHF_FRA1(pmin,0),
               CHF_FRA1(dp,0),
               CHF_FRA1(d2p,0),
               CHF_FRA1(du,0),
               CHF_FRA1(chitwiddle,0));

  Box flatBox;
  loHiCenter(grad2Box,m_DProblem,a_dir,
             loBox,hiBox,centerBox,flatBox,haslo,hashi);

  a_flatCoef.resize(flatBox,1);

  FORT_MIN3PTS(CHF_BOX(centerBox),
               CHF_BOX(loBox),
               CHF_BOX(hiBox),
               CHF_CONST_INT(a_dir),
               CHF_CONST_INT(haslo),
               CHF_CONST_INT(hashi),
               CHF_FRA1(chitwiddle,0),
               CHF_FRA1(a_flatCoef,0));

  for (RadiusMaskIterator center(m_mask,a_di,a_dir); center.ok(); center.next())
  {
    const IntVect& inside = center.inside();
    a_flatCoef(center()) = Min(chitwiddle(center()),chitwiddle(inside));
  }
}

void LevelGodunov::consToPrim(const FArrayBox& a_conState,
                              FArrayBox&       a_primState) const
{
  FORT_CONS2PRM(CHF_BOX(a_primState.box()),
                CHF_CONST_FRA(a_conState),
                CHF_FRA(a_primState));
}

void LevelGodunov::divU(const FArrayBox& a_q,
                        const Box&       a_box,
                        const int        a_dir,
                        FArrayBox&       a_divU) const
{
  // first, need to calculate the dir derivatives of
  // the tangential components of velocity at the
  // cell-centers.

  FArrayBox dveltan(grow(a_box,1) & m_DProblem,SpaceDim-1);

  for (int i = 0, dir = 0; dir < SpaceDim; ++dir)
  {
    if (dir != a_dir)
    {
    // Velocity is tangential. Build domain on which d(v_dir)/d(x_dir) is
    // to be computed.
      Box dveltanBox = a_box;
      dveltanBox.grow(dir,1).grow(a_dir,1);
      dveltanBox &= m_DProblem;

      Box gradBox,hiBox,loBox,centerBox;
      int haslo,hashi;

      loHiCenter(dveltanBox,m_DProblem,dir,
                 loBox,hiBox,centerBox,gradBox,haslo,hashi);

      // Compute d(v_dir)/d(x_dir).

      FORT_GETGRAD(CHF_BOX(centerBox),
                   CHF_BOX(loBox),
                   CHF_BOX(hiBox),
                   CHF_INT(dir),
                   CHF_CONST_INT(haslo),
                   CHF_CONST_INT(hashi),
                   CHF_CONST_FRA1(a_q,QVELX+dir),
                   CHF_FRA1(dveltan,i));
      ++i;
    }
  }

  // Now, we can calculate the divergence of the normal velocity
  // at the center normal-direction edges. To do this, we determine
  // which edges at which we have sufficient data to compute centered
  // estimates of h*(div(u)). At the remaining edges. i.e. those
  // corresponding to the physical boundaries, we use zeroth-order
  // extrapolation.

  Box divBox,loBox,hiBox,centerBox;
  int haslo,hashi;

  divBox = a_divU.box();
  divBox.shiftHalf(a_dir,1);
  Box tmp = divBox & m_DProblem;
  centerBox = tmp;
  hashi = 0;

  if (tmp != divBox)
  {
    hashi = 1;
    hiBox = bdryHi(tmp,a_dir);
    hiBox.shiftHalf(a_dir,1);
  }

  divBox.shift(a_dir,-1);
  centerBox.shift(a_dir,-1);
  tmp = divBox & m_DProblem;
  centerBox &= m_DProblem;
  centerBox.shift(a_dir,1);
  haslo = 0;
  if (tmp != divBox)
  {
    haslo = 1;
    loBox = bdryLo(tmp,a_dir);
    loBox.shiftHalf(a_dir,1);
  }

  // All of the boxes computed above are shifted so as to be
  // cell-centered, with the index of the cell center being identified
  // with the low edge. We shift a_divU to be compatible with that
  // convention on input, and undo the shift on output.

  a_divU.shiftHalf(a_dir,1);

  FORT_DIVUEDGE(CHF_BOX(centerBox),
                CHF_BOX(loBox),
                CHF_BOX(hiBox),
                CHF_CONST_INT(a_dir),
                CHF_INT(haslo),
                CHF_INT(hashi),
                CHF_CONST_FRA1(a_q,QVELX+a_dir),
                CHF_FRA(dveltan),
                CHF_FRA1(a_divU,0));

  a_divU.shiftHalf(a_dir,-1);
}

void LevelGodunov::slope(const FArrayBox& a_q,
                         const DataIndex  a_di,
                         const int        a_dir,
                         FArrayBox&       a_dq) const
{
  FArrayBox flattening;

  // flattening is sized correctly inside getFlattening.
  getFlattening(a_q,a_di,a_dir,flattening);

  Box dq2Box,dq4Box,loBox,hiBox,centerBox;
  int haslo,hashi;
  Box tmp;

  loHiCenter(a_q.box(),m_DProblem,a_dir,
             loBox,hiBox,centerBox,tmp,haslo,hashi);
  loHiCenter(tmp,m_DProblem,a_dir,
             loBox,hiBox,centerBox,dq2Box,haslo,hashi);
  FArrayBox dq2(dq2Box,m_numSlopes);
  FArrayBox dqlim(dq2Box,m_numSlopes);

  FORT_SECSLOPE(CHF_BOX(centerBox),
                CHF_BOX(loBox),
                CHF_BOX(hiBox),
                CHF_CONST_INT(a_dir),
                CHF_INT(haslo),
                CHF_INT(hashi),
                CHF_CONST_FRA(a_q),
                CHF_FRA(dqlim),
                CHF_FRA(dq2),
                CHF_CONST_INT(m_numSlopes));

  m_bc->setBdrySlopes(dq2, m_DProblem,m_dx,a_dir,a_q);

  // Apply one-sided differencing to the slopes on interpolated boundary
  // cells.
  for (RadiusMaskIterator center(m_mask,a_di,a_dir); center.ok(); center.next())
  {
    int sign = center.sign();
    const IntVect& inside = center.inside();

    for (int ll = 0; ll < m_numSlopes; ll++)
    {
      dq2(center(),ll) = sign * (a_q(center(),ll) - a_q(inside,ll));
    }
  }

  loHiCenter(dq2Box,m_DProblem,a_dir,
             loBox,hiBox,centerBox,dq4Box,haslo,hashi);

  if (haslo == 1)
  {
    loBox &= a_dq.box();
  }

  if (hashi == 1)
  {
    hiBox &= a_dq.box();
  }

  centerBox &= a_dq.box();

  dq4Box = a_dq.box();

  FArrayBox& dq4 = a_dq;

  FORT_FORSLOPE(CHF_BOX(centerBox),
                CHF_BOX(loBox),
                CHF_BOX(hiBox),
                CHF_CONST_INT(a_dir),
                CHF_INT(haslo),
                CHF_INT(hashi),
                CHF_CONST_FRA(a_q),
                CHF_FRA(dq2),
                CHF_FRA(dqlim),
                CHF_FRA1(flattening,0),
                CHF_FRA(dq4),
                CHF_CONST_INT(m_numSlopes));

  // now need to perform the one-sided versions in the
  // interpolated cells
  for (RadiusMaskIterator center(m_mask,a_di,a_dir); center.ok(); center.next())
  {
    int sign = center.sign();
    const IntVect& inside = center.inside();

    for (int ll = 0; ll < m_numSlopes; ll++)
    {
      dq4(center(),ll) = sign * (a_q(center(),ll) - a_q(inside,ll));
    }
  }
}

void LevelGodunov::normalPred(const FArrayBox& a_q,
                              const FArrayBox& a_dq,
                              const Real&      a_dtime,
                              const int        a_dir,
                              FArrayBox&       a_qlo,
                              FArrayBox&       a_qhi) const
{
  Real dtbydx = a_dtime/m_dx;
  const Box& slopeBox = a_dq.box();

  a_qlo.resize(slopeBox,m_numPrim);
  a_qhi.resize(slopeBox,m_numPrim);

  FORT_PRED(CHF_BOX(slopeBox),
            CHF_CONST_FRA(a_q),
            CHF_CONST_FRA(a_dq),
            CHF_FRA(a_qlo),
            CHF_FRA(a_qhi),
            CHF_CONST_INT(a_dir),
            CHF_CONST_REAL(dtbydx));

  a_qlo.shiftHalf(a_dir,-1);
  a_qhi.shiftHalf(a_dir, 1);
}

void LevelGodunov::riemann(const FArrayBox& a_qleft,
                           const FArrayBox& a_qright,
                           const int        a_dir,
                           FArrayBox&       a_qGdnv) const
{
  Box interiorEdges = a_qleft.box() & a_qright.box();

  // Riemann solver computes qgdnv all edges not on the physical boundary.
  FORT_RIEMANNF(CHF_BOX(interiorEdges),
                CHF_CONST_FRA(a_qleft),
                CHF_CONST_FRA(a_qright),
                CHF_FRA(a_qGdnv),
                CHF_CONST_INT(a_dir));

  // FORT_RIEMANN is a point-wise operation.  The special treatment
  // of interpolated cell points in m_mask is not required here. bvs

  // PC: however, we copy the defined values of a_qhi,a_qlo along the
  // physical boundary into the corresponding storage in LevelGodunov, in
  // preparation for using those values in the boundary condition routines.

  Box loBdry = bdryLo(m_DProblem,a_dir);
  if (a_qGdnv.box().intersects(loBdry))
  {
    Box bdryBox = a_qGdnv.box() & loBdry;
    a_qGdnv.copy(a_qright,bdryBox);
  }

  Box hiBdry = bdryHi(m_DProblem,a_dir);
  if (a_qGdnv.box().intersects(hiBdry))
  {
    Box bdryBox = a_qGdnv.box() & hiBdry;
    a_qGdnv.copy(a_qleft,bdryBox);
  }
}

void LevelGodunov::fluxCalc(const FArrayBox& a_qGdnv,
                            const int        a_dir,
                            FArrayBox&       a_flux) const
{
  // Compute fluxes on all interior edges, i.e. not on physical boundary.
  // The physical bounday conditions will be taken care of in the
  // interface class.

  // PC: warning - a_qGdnv should be defined on boundary edges.

  Box edges = a_qGdnv.box();

  FORT_GETFLUX(CHF_BOX(edges),
               CHF_CONST_FRA(a_qGdnv),
               CHF_CONST_INT(a_dir),
               CHF_FRA(a_flux));
}

void LevelGodunov::artificialViscosity(const FArrayBox& a_qGdnv,
                                       const FArrayBox& a_conState,
                                       const FArrayBox& a_divU,
                                       const int        a_dir,
                                       FArrayBox&       a_flux) const
{
  // Apply artificial viscosity to the fluxes

  Box edges = a_qGdnv.box();
  Box visc_edges = grow(surroundingNodes(m_DProblem.domainBox(),a_dir),
                        -BASISV(a_dir)) & edges;
  FORT_ARTVISC(CHF_BOX(visc_edges),
               CHF_CONST_FRA(a_conState),
               CHF_CONST_FRA1(a_divU,0),
               CHF_CONST_INT(a_dir),
               CHF_FRA(a_flux));
}

void LevelGodunov::updateState(const FArrayBox& a_flux,
                               const int        a_dir,
                               const Real       a_dtime,
                               const Box&       a_box,
                               FArrayBox&       a_conState) const
{
  Real dtbydx = a_dtime/m_dx;

  FORT_UPDATE(CHF_BOX(a_box),
              CHF_FRA(a_conState),
              CHF_CONST_FRA(a_flux),
              CHF_CONST_INT(a_dir),
              CHF_CONST_REAL(dtbydx));
}
