/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Implementation details: bvs Tue May 18 , 1999
//
//       +-+-------------+-+
//       | |             | |
//       | |    box      | |
//       | |             | |
//       +-+-------------+-+
//   boxLo                 boxHi    where spanningFAA are defined.
//
//         +--------------+
//         |              |
// --------+   other box  |
//         |              |
//         +--------------+
//         |x
//  box    |x
//         |x
//         |x
// --------+x  x = IntVectSet mask.  Points in boxHi that need correctBoundary

#include "REAL.H"
#include "SPACE.H"
#include "MayDay.H"
#include "DataIndex.H"
#include "DataIterator.H"
#include "LayoutIterator.H"
#include "DisjointBoxLayout.H"
#include "LayoutData.H"
#include "IntVectSet.H"
#include "FArrayBox.H"
#include "Box.H"
#include "IntVect.H"

#include "LevequeCorrector.H"


LevequeCorrector::LevequeCorrector()
{
  m_isDefined = false;
}


LevequeCorrector::~LevequeCorrector()
{
  clearAll();
}

void LevequeCorrector::define(const DisjointBoxLayout& a_ba,
                              const Box&               a_dProblem,
                              const Real&              a_dx,
                              const int                a_numFluxes)
{ 
  ProblemDomain physdomain(a_dProblem);

  define(a_ba,physdomain,a_dx,a_numFluxes);
}

void LevequeCorrector::define(const DisjointBoxLayout& a_ba,
                              const ProblemDomain&     a_dProblem,
                              const Real&              a_dx,
                              const int                a_numFluxes)
{
  clearAll();

  m_numFluxes = a_numFluxes;

  for (int sweepdir = 0; sweepdir < SpaceDim; ++sweepdir)
  {
    LayoutData<IntVectSet>& lo_mask = m_lo_masks[sweepdir];
    LayoutData<IntVectSet>& hi_mask = m_hi_masks[sweepdir];

    lo_mask.define(a_ba);
    hi_mask.define(a_ba);

    DataIterator dit = a_ba.dataIterator();
    for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& box = a_ba[dit()];
      Box loBox = adjCellLo(box,sweepdir);
      Box hiBox = adjCellHi(box,sweepdir);

      IntVectSet& mask_hi = hi_mask[dit()];
      IntVectSet& mask_lo = lo_mask[dit()];

      // lowPAtch and hiPatch are allocated interleaved intentionally
      if (loBox.bigEnd(sweepdir) < a_dProblem.domainBox().smallEnd(sweepdir))
      {
        // boundary passed, leave this zero sized
        loBox = Box();
      }
      else
      {
        mask_lo.define(loBox);
      }

      if (hiBox.smallEnd(sweepdir) > a_dProblem.domainBox().bigEnd(sweepdir))
      {
        // boundary passed, leave this zero sized
        hiBox = Box();
      }
      else
      {
        mask_hi.define(hiBox);
      }

      // OK, now to handle the masking.
      // Brutal search across all other boxes.

      // this loop could be done much more efficiently if we had some
      // conecpt of proximity in our code. Unfortunately, straight graph
      // representations won't cut it....

      LayoutIterator lit = a_ba.layoutIterator();
      for (lit.begin(); lit.ok(); ++lit)
      {
        // can experiment with doing the copy here for cache later...
        const Box& nextBox = a_ba[lit()];

        if (nextBox.bigEnd(0) < box.smallEnd(0) - 1)
        {
          // do nothing, nextBox not near these corrector registers
        }
        else if (nextBox.smallEnd(0) > box.bigEnd(0) + 1)
        {
          lit.end();
        }
        else
        {
          mask_hi -= (nextBox);
          mask_lo -= (nextBox);
        }
      }
    }
  }

  m_isDefined = true;
}

void LevequeCorrector::clearAll()
{
  m_isDefined = false;
}

void LevequeCorrector::correctBoundary(FArrayBox&       a_UPatch,
                                       FArrayBox&       a_flux,
                                       Real             a_dt,
                                       Real             a_dx,
                                       int              a_fluxDir,
                                       const DataIndex& a_patchIndex) const
{
  assert(m_numFluxes == a_UPatch.nComp());
  assert(m_numFluxes == a_flux.nComp());

  if (!m_isDefined)
  {
    MayDay::Error("LevequeCorrector::correctBoundary !m_isDefined");
  }

  for (int dir = 0; dir < SpaceDim; ++dir)
  {
    if (dir != a_fluxDir)
    {
      a_flux.shift(dir,-1);
      IVSIterator loIterator(m_lo_masks[dir][a_patchIndex]);
      IVSIterator hiIterator(m_hi_masks[dir][a_patchIndex]);

      for (loIterator.begin(); loIterator.ok(); ++loIterator)
      {
        IntVect ivLo = loIterator();
        IntVect ivHi = ivLo + BASISV(a_fluxDir);
        for (int comp = 0; comp < m_numFluxes; comp++)
        {
          a_UPatch(ivLo,comp) += a_dt*a_flux(ivLo,comp)/a_dx;
          a_UPatch(ivLo,comp) -= a_dt*a_flux(ivHi,comp)/a_dx;
        }
      }

      a_flux.shift(dir,2);

      for (hiIterator.begin(); hiIterator.ok(); ++hiIterator)
      {
        IntVect ivLo = hiIterator();
        IntVect ivHi = ivLo + BASISV(a_fluxDir);
        for (int comp = 0; comp < m_numFluxes; comp++)
        {
          a_UPatch(ivLo,comp) += a_dt*a_flux(ivLo,comp)/a_dx;
          a_UPatch(ivLo,comp) -= a_dt*a_flux(ivHi,comp)/a_dx;
        }
      }

      a_flux.shift(dir,-1);
    }
  }
}
