/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// RadiusMask.H
// ============
//
// Similar to what other coarse-fine classes handle.
//
// the main method of this helper class is
//
// const IntVectSet& intVectSet(int boxIndex,
//                     int sweepdir,
//                     Side::LoHiSide side)
//
//
// IntVects for various boxes for sweepdir=0, side = Lo
// and radius 2
// given by numbers:
//
//     +---------------+      33|           |/
//   11|               |      33|    box3   |/
//   11|               +--------+-----------+/
//   --+               |                    |/
//     |               |                    |/
// box5|      box1     |          box2      |/
//   --+               |                    |/
//   11|               |                    |/
//     +---------------+----+---------+-----+/
//                        44|         |      /
//                        44| box4    |      /
//                        44|         |      /

// original design        : Phil Colella and Brian Van Straalen
// original implementation: Brian Van Straalen
// implementation date    : 07/02/99
//

#ifndef _RADIUS_MASK_H
#define _RADIUS_MASK_H

#include "Vector.H"
#include "IntVectSet.H"
#include "LoHiSide.H"
#include "BoxLayout.H"
#include "LayoutData.H"
#include "LayoutIterator.H"
#include "DataIterator.H"
#include "DisjointBoxLayout.H"
#include "DataIndex.H"

///
/**
   Similar to what other coarse-fine classes handle.
 */
class RadiusMask
{
public:
  ///
  RadiusMask(){;};

  ///
  virtual ~RadiusMask(){;}; // in case someone wants to safely derive

  // default copy and assign should be fine.

  ///
  RadiusMask(const DisjointBoxLayout& a_ba,
             const Box&               a_DProblem,
             int                      a_radius);

  ///
  RadiusMask(const DisjointBoxLayout& a_ba,
             const ProblemDomain&     a_DProblem,
             int                      a_radius);

  ///
  void define(const DisjointBoxLayout& a_ba,
              const Box&               a_DProblem,
              int                      a_radius);

  ///
  void define(const DisjointBoxLayout& a_ba,
              const ProblemDomain&     a_DProblem,
              int                      a_radius);

  ///
  const IntVectSet& intVectSet(DataIndex      a_di,
                               int            a_sweepdir,
                               Side::LoHiSide a_side) const;

private:
  LayoutData<IntVectSet> m_masks[SpaceDim*2];

  // convention 2*sweepdir   == lowSide
  //            2*sweepdir+1 == hiSide
  int                    m_radius;
};

// This class has a built in flaw currently in that
// it exposes implementation by the return of the
// 'const reference': in C++ the only way to implement
// return by const reference with thread-safe code is
// with data members of that type.
//
//  when we see better where and how this class is used
//  we can see how to build a more encapsulated version.
//
//  At a future date we might experiment with having this
//  class calculate the return object on the fly
//  (explore the cache vs. compute tradeoff).

//=============================

class RadiusMaskIterator
{
public:
  RadiusMaskIterator(const RadiusMask& a_mask,
                     DataIndex         a_di,
                     int               a_sweepdir);

  void define(const RadiusMask& a_mask,
              DataIndex         a_di,
              int               a_sweepdir);

  // default destructor and copy constructor are fine right now.
  void begin();

  inline bool ok();

  void next();

  inline const IntVect& operator() () const;

  /** helper function that returns the IntVect one cell in the oposite
      direction of the current sign of the iteration. */
  inline const IntVect& inside() const;

  inline int sign() const ;

private:
  int         m_setIndex;
  int         m_sweepdir;
  IVSIterator m_set[2];
  IntVect     m_current;
  IntVect     m_inside;
};

//========end of declarations, inline implementation from there down=====

inline const IntVectSet& RadiusMask::intVectSet(DataIndex      a_di,
                                                int            a_sweepdir,
                                                Side::LoHiSide a_side) const
{
  return m_masks[2*a_sweepdir+a_side][a_di];
}

inline RadiusMaskIterator::RadiusMaskIterator(const RadiusMask& a_mask,
                                              DataIndex         a_di,
                                              int               a_sweepdir)
{
  define(a_mask,a_di,a_sweepdir);
}

inline bool RadiusMaskIterator::ok()
{
  return (m_setIndex < 2) && m_set[m_setIndex].ok();
}

inline const IntVect& RadiusMaskIterator::operator() () const
{
  return m_current;
}

inline int RadiusMaskIterator::sign() const
{
  assert(m_setIndex == 1 || m_setIndex == 0);
  return ((m_setIndex == 0) ? -1 : 1);
}

inline const IntVect& RadiusMaskIterator::inside() const
{
  return m_inside;
}

#endif
