/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "RadiusMask.H"
#include "DisjointBoxLayout.H"
#include "Box.H"

RadiusMask::RadiusMask(const DisjointBoxLayout& a_ba,
                       const Box&               a_DProblem,
                       int                      a_radius)
{
  define(a_ba,a_DProblem,a_radius);
}



RadiusMask::RadiusMask(const DisjointBoxLayout& a_ba,
                       const ProblemDomain&     a_DProblem,
                       int                      a_radius)
{
  define(a_ba,a_DProblem,a_radius);
}

void RadiusMask::define(const DisjointBoxLayout& a_ba,
                        const Box&               a_DProblem,
                        int                      a_radius)
{
  ProblemDomain physdomain(a_DProblem);

  define(a_ba,physdomain,a_radius);
}


void RadiusMask::define(const DisjointBoxLayout& a_ba,
                        const ProblemDomain&     a_DProblem,
                        int                      a_radius)
{
  DataIterator di = a_ba.dataIterator();
  m_radius = a_radius;

  for (int dir = 0; dir < SpaceDim; ++dir)
  {
    LayoutData<IntVectSet>& loVector = m_masks[2*dir];
    LayoutData<IntVectSet>& hiVector = m_masks[2*dir+1];

    loVector.define(a_ba);
    hiVector.define(a_ba);

    for (di.begin(); di.ok(); ++di)
    {
      IntVectSet& loSet = loVector[di()];
      IntVectSet& hiSet = hiVector[di()];

      Box loBox(adjCellLo(a_ba[di()],dir,a_radius));
      Box hiBox(adjCellHi(a_ba[di()],dir,a_radius));

      loBox &= a_DProblem;
      hiBox &= a_DProblem;

      loSet.define(loBox);
      hiSet.define(hiBox);

      LayoutIterator li2 = a_ba.layoutIterator();
      for (li2.begin(); li2.ok(); ++li2)
        {
          const Box& nextBox = a_ba[li2()];
          loSet -= nextBox;
          hiSet -= nextBox;
        }
    }
  }
}

//================================================================

void RadiusMaskIterator::define(const RadiusMask& a_mask,
                                DataIndex         a_di,
                                int               a_sweepdir)
{
  m_sweepdir = a_sweepdir;
  m_set[0].define(a_mask.intVectSet(a_di,a_sweepdir,Side::Lo));
  m_set[1].define(a_mask.intVectSet(a_di,a_sweepdir,Side::Hi));

  begin();
}

void RadiusMaskIterator::begin()
{
  m_set[0].begin();
  m_set[1].begin();

  if (m_set[0].ok())
  {
    m_setIndex = 0;
    m_current = m_set[0]();
    m_inside = m_current;
    m_inside[m_sweepdir] += 1;
  }
  else
  {
    m_setIndex = 1;
    m_set[1].begin();

    if (m_set[1].ok())
    {
      m_current = m_set[1]();
      m_inside = m_current;
      m_inside[m_sweepdir] -= 1;
    }
    else
    {
      m_setIndex = 2;
    }
  }
}

void RadiusMaskIterator::next()
{
  assert(m_setIndex == 0 || m_setIndex == 1);

  ++(m_set[m_setIndex]);

  if (!m_set[m_setIndex].ok())
  {
    ++m_setIndex;
  }

  if (m_setIndex < 2)
  {
    m_current = m_set[m_setIndex]();
    m_inside  = m_current;
    m_inside[m_sweepdir] -= sign();
  }
}
