/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LoHiSide.H"

#include "WaveMHDIBC.H"
#include "WaveMHDIBCF_F.H"

#include "LoHiCenter.H"

// Null constructor
WaveMHDIBC::WaveMHDIBC()
{
    m_isParameterSet = false;
}

// Constructor which defines parameters
WaveMHDIBC::WaveMHDIBC(const Real& a_gamma,
		       const int& a_pdir,
		       const int& a_kratio,
		       const int& a_waveNumber,
		       const Real& a_pertAmplitude
		       )
{
  setParameters(a_gamma,a_pdir,a_kratio,a_waveNumber,a_pertAmplitude);
}

// Sets parameters in a common block used by Fortran routines:
//   a_gamma   - Gamma for polytropic, gamma-law gas
void WaveMHDIBC::setParameters(const Real& a_gamma,
			       const int& a_pdir,
			       const int& a_kratio,
			       const int& a_waveNumber,
			       const Real& a_pertAmplitude
		       )
{
  assert(m_isParameterSet == false);

  FORT_SETWAVEMHDF(CHF_CONST_REAL(a_gamma));
  FORT_SETWAVEMHDPARAMETERSF(CHF_CONST_INT(a_pdir),
			    CHF_CONST_INT(a_kratio),
			    CHF_CONST_INT(a_waveNumber),
			    CHF_CONST_REAL(a_pertAmplitude));

  //  pdir=a_pdir;
  //  kratio=a_kratio;
  //  waveNumber=a_waveNumber;
  //  pertAmplitude=a_pertAmplitude;

  m_isParameterSet = true;
}

// Set the flag m_isParameterSet to true so that new IBCs made with
// new_physIBC() will have this flag set without calling setParameters()
// (this is a clumsy design and should be improved).
void WaveMHDIBC::setParameterSet()
{
  m_isParameterSet = true;
}

// Factory method - this object is its own factory:
//   Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
//   its define() must be called before it is used) and m_isParameterSet
//   set to value of m_isParametersset in the current (factory) object.
PhysIBC* WaveMHDIBC::new_physIBC()
{
  WaveMHDIBC* retval = new WaveMHDIBC();

  if (m_isParameterSet == true)
  {
    retval->setParameterSet();
  }

  return static_cast<PhysIBC*>(retval);
}

// Set boundary fluxes
void WaveMHDIBC::fluxBC(FArrayBox&            a_F,
                     const FArrayBox&      a_W,
                     const FArrayBox&      a_Wextrap,
                     const int&            a_dir,
                     const Side::LoHiSide& a_side,
                     const Real&           a_time)
{
  assert(m_isParameterSet == true);
  assert(m_isDefined == true);

  int sign;
  Box FBox = a_F.box();
  Box tmp = FBox;

  // Determine which side and thus shifting directions
  if (a_side == Side::Lo)
  {
    sign = -1;
  }
  else
  {
    sign = 1;
  }

  cout << "tmp0 boundaryBox=" << tmp << endl;
  tmp.shiftHalf(a_dir,sign);

  // Is there a domain boundary next to this grid
  cout << "a_side=" << a_side << endl;
  cout << "tmp1 boundaryBox=" << tmp << endl;
  if (!m_domain.contains(tmp))
  {
    tmp &= m_domain;

    cout << "tmp2 boundaryBox=" << tmp << endl;
    Box boundaryBox;

    // Find the strip of cells next to the domain boundary
    if (a_side == Side::Lo)
    {
      boundaryBox = bdryLo(tmp,a_dir);
      cout << "Lo boundaryBox=" << boundaryBox << endl;
    }
    else
    {
      boundaryBox = bdryHi(tmp,a_dir);
      cout << "Hi boundaryBox=" << boundaryBox << endl;
    }

    // Shift things to all line up correctly
    boundaryBox.shiftHalf(a_dir,-sign);
    a_F.shiftHalf(a_dir,-sign);

    // Set the solid wall boundary fluxes- not for the Rotor problem
    //        FORT_MHDSOLIDBCF(CHF_FRA(a_F),
    //                             CHF_CONST_FRA(a_Wextrap),
    //                             CHF_CONST_REAL(m_dx),
    //                             CHF_CONST_INT(a_dir),
    //                             CHF_BOX(boundaryBox));

    // Set the time varying boundary fluxes
    FORT_MHDWAVEBCF(CHF_FRA(a_F),
               CHF_CONST_FRA(a_Wextrap),
	       //  CHF_CONST_FRA(a_W), RS: changed W to Wextrap
                CHF_CONST_REAL(a_time),
                CHF_CONST_INT(sign),
                CHF_CONST_REAL(m_dx),
                CHF_CONST_INT(a_dir),
                CHF_BOX(boundaryBox));

    // Shift returned fluxes to be face centered
    a_F.shiftHalf(a_dir,sign);
  }
}

// Set up initial conditions
void WaveMHDIBC::initialize(LevelData<FArrayBox>& a_U)
{
  assert(m_isParameterSet == true);
  assert(m_isDefined == true);

  DataIterator dit = a_U.boxLayout().dataIterator();

  // Iterator of all grids in this level
  for (dit.begin(); dit.ok(); ++dit)
  {
    // Storage for current grid
    FArrayBox& U = a_U[dit()];

    // Box of current grid
    Box uBox = U.box();
    uBox &= m_domain;
    
    // Set up initial condition in this grid
    FORT_MHDWAVEINITF(CHF_FRA(U),
		     CHF_CONST_REAL(m_dx),
		     CHF_BOX(uBox));
  }
}

// Set boundary slopes:
//   The boundary slopes in a_dW are already set to one sided difference
//   approximations.  If this function doesn't change them they will be
//   used for the slopes at the boundaries.
void WaveMHDIBC::setBdrySlopes(FArrayBox&       a_dW,
                            const FArrayBox& a_W,
                            const int&       a_dir,
				const Real&      a_time)
{
  assert(m_isParameterSet == true);
  assert(m_isDefined == true);

  Box loBox,hiBox,centerBox,domain;
  int hasLo,hasHi;
  Box slopeBox = a_dW.box()&m_domain;

  // Generate the domain boundary boxes, loBox and hiBox, if there are
  // domain boundarys there
  loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
             slopeBox,m_domain,a_dir);

  // Set the boundary slopes if necessary
  if ((hasLo != 0) | (hasHi != 0))
  {
    FORT_MHDWAVESLOPEBCSF(CHF_FRA(a_dW),
                      CHF_CONST_FRA(a_W),
                      CHF_CONST_REAL(m_dx),
                      CHF_CONST_INT(a_dir),
                      CHF_BOX(loBox),
                      CHF_CONST_INT(hasLo),
                      CHF_BOX(hiBox),
                      CHF_CONST_INT(hasHi));
  }
}
