/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _PATCHPOLYTROPICGAS_H_
#define _PATCHPOLYTROPICGAS_H_

#include "OldPatchGodunov.H"
#include "OldPatchGodunovF_F.H"

/// An example PatchGodunov for polytropic, gamma-law gas dynamics.
/**
 */
class PatchPolytropicGas : public OldPatchGodunov
{
public:
  ///
  /**
   */
  PatchPolytropicGas();

  /// Factory method - this object is its own factory
  /**
     Return a pointer to new OldPatchGodunov object with its initial and boundary
     condtions, slope parameters, and artificial viscosity information defined.
   */
  virtual OldPatchGodunov* new_patchGodunov() const;

  /// Compute the maximum wave speed
  /**
   */
  virtual Real getMaxWaveSpeed(const FArrayBox& a_U,
                               const Box&       a_box);

  /// Number of conserved variables
  /**
     Return the number of conserved variables.
   */
  virtual int numConserved();

  /// Names of the conserved variables
  /**
     Return the names of the conserved variables.
   */
  virtual Vector<string> stateNames();

  /// Number of flux variables
  /**
     Return the  number of flux variables.  This can be greater than the number
     of conserved variables if addition fluxes/face-centered quantities are
     computed.
   */
  virtual int numFluxes();


protected:
  /// Number of primitive variables
  /**
     Return the number of primitive variables.  This may be greater than the
     number of conserved variables if derived/redundant quantities are also
     stored for convenience.
   */
  virtual int numPrimitives();

  ///  Number of primitive variables for which slopes are computed
  /**
     Return the number of primitive variables for which slopes are computed.
     Only slopes corresponding to primitive variables in the interval 0 to
     numSlopes() - 1 are computed and only primitive variables in that interval
     are updated using the slopes.
   */
  virtual int numSlopes();

  /// Compute the primitive variables from the conserved variables within a_box
  /**
   */
  virtual void consToPrim(FArrayBox&       a_W,
                          const FArrayBox& a_U,
                          const Box&       a_box);

  /// Compute the conserved variables from the primitive variables within a_box
  /**
   */
  virtual void primToCons(FArrayBox&       a_U,
                          const FArrayBox& a_W,
                          const Box&       a_box);

  /// Extrapolate the primitive variables to the cell faces
  /**
     Extrapolate the primitive variables, a_W, in the minus and plus direction,
     a_dir, using the slopes a_dW within a_box.  See document for
     details.
   */
  virtual void normalPred(FArrayBox&       a_WMinus,
                          FArrayBox&       a_WPlus,
                          const FArrayBox& a_W,
                          const FArrayBox& a_dW,
                          const Real&      a_scale,
                          const int&       a_dir,
                          const Box&       a_box);

  /// Compute a Riemann problem and generate fluxes at the faces
  /**
     Given input left and right states in a direction, a_dir, compute a
     Riemann problem and generate fluxes at the faces within a_box.
   */
  virtual void riemann(FArrayBox&       a_F,
                       const FArrayBox& a_WLeft,
                       const FArrayBox& a_WRight,
                       const int&       a_dir,
                       const Box&       a_box);

  /// Update the primitive variables using fluxes (of the conserved variables)
  /**
     Given the fluxes, a_F, in a direction, a_dir, and a scaling factor,
     a_scale, update the primitive variables, a_WMinus and a_WPlus, within
     a_box.  Note:  The fluxes are fluxes of conserved variables.
   */
  virtual void updatePrim(FArrayBox&       a_WMinus,
                          FArrayBox&       a_WPlus,
                          const FArrayBox& a_F,
                          const Real&      a_scale,
                          const int&       a_dir,
                          const Box&       a_box);

  /// Update the conserved variable using fluxes and a scaling factor
  /**
     Given the fluxes, a_F, in a direction, a_dir, and a scaling factor,
     a_scale, update the conserved variables, a_U, within a_box:

     a_U = a_U - a_scale * (a_F(a_dir,HiSide) - a_F(a_dir,LoSide))
   */
  virtual void updateCons(FArrayBox&       a_U,
                          const FArrayBox& a_F,
                          const Real&      a_scale,
                          const int&       a_dir,
                          const Box&       a_box);

  /// Perform final update of conserved variable using fluxes
  /**
     Given the fluxes, a_F, in a direction, a_dir, and a scaling factor,
     a_scale, update the conserved variables, a_U, within a_box:

     a_U = a_U - a_scale * (a_F(a_dir,HiSide) - a_F(a_dir,LoSide))
   */
  virtual void finalUpdate(FArrayBox&       a_U,
                           const FArrayBox& a_F,
                           const Real&      a_scale,
                           const int&       a_dir,
                           const Box&       a_box);

  virtual void allTogetherNow(FArrayBox&       a_WMinus,
                              FArrayBox&       a_WPlus,
                              const FArrayBox& a_F,
                              const Real&      a_scale,
                              const int&       a_dir,
                              const Box&       a_box);

  /// Interval within the primitive variables corresponding to the velocities
  /**
     Return the interval of component indices within the primitive variable
     of the velocities.  Used for slope flattening (slope computation) and
     computing the divergence of the velocity (artificial viscosity).
   */
  virtual Interval velocityInterval();

  /// Component index within the primitive variables of the pressure
  /**
     Return the component index withn the primitive variables for the
     pressure.  Used for slope flattening (slope computation).
   */
  virtual int pressureIndex();

  /// Component index within the primitive variables of the bulk modulus
  /**
     Return the component index withn the primitive variables for the
     bulk modulus.  Used for slope flattening (slope computation) used
     as a normalization to measure shock strength.
   */
  virtual int bulkModulusIndex();
};

#endif
