/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "LoHiSide.H"

#include "RampIBC.H"
#include "RampF_F.H"

#include "OldLoHiCenter.H"

#include "Timer.H"
#include "TimerPrototypes2.H"

// Null constructor
RampIBC::RampIBC()
{
    m_isFortranCommonSet = false;
}

// Constructor which defines parameters used by Fortran routines
RampIBC::RampIBC(const Real& a_gamma,
                 const Real& a_alpha,
                 const Real& a_ms,
                 const Real& a_xcorner,
                 const Real& a_artvisc)
{
  setFortranCommon(a_gamma,
                   a_alpha,
                   a_ms,
                   a_xcorner,
                   a_artvisc);
}

// Sets parameters in a common block used by Fortran routines:
//   a_gamma   - Gamma for polytropic, gamma-law gas
//   a_alpha   - The angle from horizontal to the wedge edge
//   a_ms      - Mach shock number
//   a_xcorner - Position of the corner/tip of the wedge
//   a_artvisc - Artificial viscosity coefficient
void RampIBC::setFortranCommon(const Real& a_gamma,
                               const Real& a_alpha,
                               const Real& a_ms,
                               const Real& a_xcorner,
                               const Real& a_artvisc)
{
  assert(m_isFortranCommonSet == false);

  TF_SETRAMP.start();
  FORT_SETRAMPF(CHF_CONST_REAL(a_gamma),
                CHF_CONST_REAL(a_alpha),
                CHF_CONST_REAL(a_ms),
                CHF_CONST_REAL(a_xcorner),
                CHF_CONST_REAL(a_artvisc));
  TF_SETRAMP.stop();

  m_isFortranCommonSet = true;
}

// Factory method - this object is its own factory:
//   Return a pointer to a new OldPhysIBC object with m_isDefined = false (i.e.,
//   its define() must be called before it is used) and m_isFortranCommonSet
//   set to value of m_isFortranCommonset in the current (factory) object.
OldPhysIBC* RampIBC::new_physIBC()
{
  RampIBC* retval = new RampIBC();
  retval->m_isFortranCommonSet = m_isFortranCommonSet;

  return static_cast<OldPhysIBC*>(retval);
}

// Set up initial conditions
void RampIBC::initialize(LevelData<FArrayBox>& a_U)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  DataIterator dit = a_U.boxLayout().dataIterator();

  // Iterator of all grids in this level
  for (dit.begin(); dit.ok(); ++dit)
  {
    // Storage for current grid
    FArrayBox& U = a_U[dit()];

    // Box of current grid
    Box uBox = U.box();
    uBox &= m_domain;

    // Set up initial condition in this grid
    TF_RAMPINIT.start();
    FORT_RAMPINITF(CHF_FRA(U),
                   CHF_CONST_REAL(m_dx),
                   CHF_BOX(uBox));
    TF_RAMPINIT.stop();
  }
}

// Set boundary fluxes
void RampIBC::fluxBC(FArrayBox&            a_F,
                     const FArrayBox&      a_W,
                     const FArrayBox&      a_Wextrap,
                     const int&            a_dir,
                     const Side::LoHiSide& a_side,
                     const Real&           a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  int sign;
  Box FBox = a_F.box();
  Box tmp = FBox;

  // Determine which side and thus shifting directions
  if (a_side == Side::Lo)
  {
    sign = -1;
  }
  else
  {
    sign = 1;
  }

  tmp.shiftHalf(a_dir,sign);

  // Is there a domain boundary next to this grid
  if (!m_domain.contains(tmp))
  {
    tmp &= m_domain;

    Box boundaryBox;

    // Find the strip of cells next to the domain boundary
    if (a_side == Side::Lo)
    {
      boundaryBox = bdryLo(tmp,a_dir);
    }
    else
    {
      boundaryBox = bdryHi(tmp,a_dir);
    }

    // Shift things to all line up correctly
    boundaryBox.shiftHalf(a_dir,-sign);
    a_F.shiftHalf(a_dir,-sign);

    // Set the solid wall boundary fluxes
    TF_RAMPSOLIDBC.start();
    FORT_RAMPSOLIDBCF(CHF_FRA(a_F),
                      CHF_CONST_FRA(a_Wextrap),
                      CHF_CONST_INT(sign),
                      CHF_CONST_REAL(m_dx),
                      CHF_CONST_INT(a_dir),
                      CHF_BOX(boundaryBox));
    TF_RAMPSOLIDBC.stop();

    // Set the time varying boundary fluxes
    TF_RAMPBC.start();
    FORT_RAMPBCF(CHF_FRA(a_F),
                 CHF_CONST_FRA(a_W),
                 CHF_CONST_REAL(a_time),
                 CHF_CONST_INT(sign),
                 CHF_CONST_REAL(m_dx),
                 CHF_CONST_INT(a_dir),
                 CHF_BOX(boundaryBox));
    TF_RAMPBC.stop();

    // Shift returned fluxes to be face centered
    a_F.shiftHalf(a_dir,sign);
  }
}

// Set boundary slopes:
//   The boundary slopes in a_dW are already set to one sided difference
//   approximations.  If this function doesn't change them they will be
//   used for the slopes at the boundaries.
void RampIBC::setBdrySlopes(FArrayBox&       a_dW,
                            const FArrayBox& a_W,
                            const int&       a_dir,
                            const Real&      a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  Box loBox,hiBox,centerBox,domain;
  int hasLo,hasHi;
  Box slopeBox = a_dW.box()&m_domain;

  // Generate the domain boundary boxes, loBox and hiBox, if there are
  // domain boundarys there
  oldLoHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
                slopeBox,m_domain,a_dir);

  // Set the boundary slopes if necessary
  if ((hasLo != 0) || (hasHi != 0))
  {
    TF_RAMPSLOPEBCS.start();
    FORT_RAMPSLOPEBCSF(CHF_FRA(a_dW),
                       CHF_CONST_FRA(a_W),
                       CHF_CONST_REAL(m_dx),
                       CHF_CONST_INT(a_dir),
                       CHF_BOX(loBox),
                       CHF_CONST_INT(hasLo),
                       CHF_BOX(hiBox),
                       CHF_CONST_INT(hasHi));
    TF_RAMPSLOPEBCS.stop();
  }
}

// Adjust boundary fluxes to account for artificail viscosity
void RampIBC::artViscBC(FArrayBox&       a_F,
                        const FArrayBox& a_U,
                        const FArrayBox& a_divVel,
                        const int&       a_dir,
                        const Real&      a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  FArrayBox &divVel = (FArrayBox &)a_divVel;

  Box fluxBox = divVel.box();
  fluxBox.enclosedCells(a_dir);
  fluxBox.grow(a_dir,1);
  fluxBox &= m_domain;

  assert(a_U.box().contains(fluxBox));

  Box loBox,hiBox,centerBox,entireBox;
  int hasLo,hasHi;

  oldLoHiCenterFace(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                    fluxBox,m_domain,a_dir);

  if (hasLo == 1)
  {
    loBox.shiftHalf(a_dir,1);
    divVel.shiftHalf(a_dir,1);
    a_F.shiftHalf(a_dir,1);

    int loHiSide = -1;

    TF_RAMPARTVISC.start();
    FORT_RAMPARTVISCF(CHF_FRA(a_F),
                      CHF_CONST_FRA(a_U),
                      CHF_CONST_FRA1(divVel,0),
                      CHF_CONST_INT(loHiSide),
                      CHF_CONST_REAL(m_dx),
                      CHF_CONST_INT(a_dir),
                      CHF_BOX(loBox));
    TF_RAMPARTVISC.stop();

    loBox.shiftHalf(a_dir,-1);
    divVel.shiftHalf(a_dir,-1);
    a_F.shiftHalf(a_dir,-1);
  }

  if (hasHi == 1)
  {
    hiBox.shiftHalf(a_dir,-1);
    divVel.shiftHalf(a_dir,-1);
    a_F.shiftHalf(a_dir,-1);

    int loHiSide = 1;

    TF_RAMPARTVISC.start();
    FORT_RAMPARTVISCF(CHF_FRA(a_F),
                      CHF_CONST_FRA(a_U),
                      CHF_CONST_FRA1(divVel,0),
                      CHF_CONST_INT(loHiSide),
                      CHF_CONST_REAL(m_dx),
                      CHF_CONST_INT(a_dir),
                      CHF_BOX(hiBox));
    TF_RAMPARTVISC.stop();

    hiBox.shiftHalf(a_dir,1);
    divVel.shiftHalf(a_dir,1);
    a_F.shiftHalf(a_dir,1);
  }
}
