/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial \purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
#include <fstream>
#include <string>
using std::fstream;
using std::string;
#include <cstdio>
#include <cmath>

#include "localFuncs.H"
#include "parstream.H"
#include "BoxIterator.H"
#include "PoissProbF_F.H"
#include "LoadBalance.H"
#include "AMRIO.H"
#include "PoissonBC.H"
#include "BRMeshRefine.H"
#include "CoarseAverage.H"

#include "exampleBallsF.H"

int outputData(const Vector<LevelData<FArrayBox>* >& vectPhi,
               const Vector<LevelData<FArrayBox>* >& vectRhs,
               const Vector<DisjointBoxLayout>& vectGrids,
               const Vector<ProblemDomain>& vectDomain,
               const Vector<int>& vectRatio,
               Real dxCoarsest,
               int numlevels,
               bool verbose)
{
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  int ncomp = vectPhi[0]->nComp();

  string phiNameBase("phi");
  string rhsNameBase("rhs");

  Vector<string> vectName(2*ncomp);
  for (int comp=0; comp<ncomp; comp++) 
    {
      char labelString[80];
      sprintf(labelString, "%s%d", phiNameBase.c_str(), comp);
      string phiName(labelString);
      vectName[comp] = phiName;
      sprintf(labelString, "%s%d", rhsNameBase.c_str(), comp);
      string rhsName(labelString);
      vectName[ncomp+comp] = rhsName;
    }
  vectName.push_back("procID");
  Box domain = vectDomain[0].domainBox();
  //XXX -- not used
  //Real dx = 1.;
#ifdef HDF5
  Real dt = 1.;
  Real time = 1.;
#endif

  // problem is that invalid regions of phi contain whatever garbage
  // is left over from the solve; this can make viewing output difficult
  // since the garbage may affect the dynamic range in the plotfile.
  // fix this by doing an average down of finer-level solution(if it
  // exists) onto covered regions
  for (int ilev = numlevels-2; ilev >= 0; ilev--)
    {      
      const DisjointBoxLayout& fineGrids = vectPhi[ilev+1]->getBoxes();
      CoarseAverage avgObject(fineGrids, 
                              ncomp,
                              vectRatio[ilev]);
      
      avgObject.averageToCoarse(*vectPhi[ilev],
                                *vectPhi[ilev+1]);
    }
  

  Vector<LevelData<FArrayBox>* > vectPhiAndRHS(numlevels, NULL);
  for(int ilev = 0; ilev < numlevels; ilev++)
    {
      vectPhiAndRHS[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], 2*ncomp+1);
      
      Interval phiInterval(0,ncomp-1);
      Interval rhsInterval(ncomp, 2*ncomp-1);

      vectPhi[ilev]->copyTo(vectPhi[ilev]->interval(), 
                            *vectPhiAndRHS[ilev],
                            phiInterval);
      vectRhs[ilev]->copyTo(vectRhs[ilev]->interval(), 
                            *vectPhiAndRHS[ilev],
                            rhsInterval);
      DataIterator dit = vectPhiAndRHS[ilev]->dataIterator();
      for(; dit.ok(); ++dit){
        vectPhiAndRHS[ilev]->operator[](dit()).setVal(procID(), 2*ncomp);
      }
    }
#ifdef HDF5
  string filename("poissonOut.hdf5");
  WriteAMRHierarchyHDF5(filename,                             
                        vectGrids, 
                        vectPhiAndRHS, 
                        vectName,                 
                        domain,                               
                        dxCoarsest, dt, time,                                         
                        vectRatio,                           
                        numlevels);
#endif

  for(int ilev = 0; ilev < numlevels; ilev++)
    delete vectPhiAndRHS[ilev];

  return 0;
}

/*
  Set RHS on hierarchy from input file
 */
int setRHS(Vector<LevelData<FArrayBox>* >& vectRhs,
           const Vector<Real>&             vectDx, 
           const Vector<ProblemDomain>&    vectDomain, 
           const int numlevels, 
           const bool verbose)
{

#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  Real rhono, rno;

  ParmParse ppLocal("main");
  bool useConstRHS = false;
  Real constRHS = -1.0;
  if (ppLocal.contains("constRhs"))
    {
      useConstRHS = true;
      ppLocal.get("constRhs", constRHS);
    }

  // ?  i dunno ?
  // problem 1 is better for dirichlet problem, while 
  // problem 2 is for the periodic case
  int iprob;   // defaults...  
  if (vectDomain[0].isPeriodic())
    {
      iprob = 2;
    } else {
      iprob = 1;
    }

  // if iprob is in input -- then override here...
  if (ppLocal.contains("iprob"))
    {
      ppLocal.get("iprob", iprob);
    }

  pout() << " iprob = " << iprob << endl;

  // apparently these aren't actually inputs...
  rhono = 0.75;
  rno = 0.5;

  if(verbose) {
    pout() 
      << " rhono  = " << rhono
      << " rno  = " << rno
      << " iprob  = " << iprob
      << endl;
  }

  for(int ilev = 0; ilev < numlevels; ilev++)
    {
      if(verbose)
        pout() 
          << " ilev = " << ilev 
          << " dom = " << vectDomain[ilev]
          << " dx  = " << vectDx[ilev]
          << endl;

      Real dxlev = vectDx[ilev];
      Box domlev = vectDomain[ilev].domainBox();
      LevelData<FArrayBox>& rhsLD = *vectRhs[ilev];
      DataIterator dit =  rhsLD.dataIterator();
      for(dit.reset(); dit.ok(); ++dit)
        {
          FArrayBox& rhsFab = rhsLD[dit()];

          if (iprob == 3) {


            int ndom = 512;
            int nsplit = 8;
            int refToCoarse = 16;

            //int ndom = 64;
            //int nsplit = 4;
            //int refToCoarse = 4;

            int nballs = 1;

            Vector<RealVect> ballCenters(nballs);
            Real* ballRadii = new Real[nballs];
            ballCenters[0] = 0.5 * RealVect::Unit; // center of spherical charge
            ballRadii[0] = 0.45;
            
            // need to send amid1, amid2, amid3
            Real* amidx = new Real[nballs];
            Real* amidy = new Real[nballs];
            Real* amidz = new Real[nballs];
            for (int iball = 0; iball < nballs; iball++)
              {
                RealVect ballCenterThis = ballCenters[iball];
                amidx[iball] = ballCenterThis[0];
                amidy[iball] = ballCenterThis[1];
                amidz[iball] = ballCenterThis[2];
              }


            int s = 0; // size of buffer for rhs or phi
            const Box& bx = rhsFab.box();
            const int *bxLo = bx.loVect();
            const int *bxHi = bx.hiVect();
            int iverbose = 0;
            //if (verbose) iverbose = 1;
            
            FORT_INITIALIZERHS(rhsFab.dataPtr(),
                               &bxLo[0], &bxLo[1], &bxLo[2],
                               &bxHi[0], &bxHi[1], &bxHi[2],
                               &ndom, &s, &dxlev, &nballs,
                               amidx, amidy, amidz, ballRadii,
                               &iverbose);

            delete amidx;
            delete amidy;
            delete amidz;
            delete ballRadii;

          } else {

            //kluge to get things the same in 0.2
            rhsFab.setVal(constRHS);
            /**/
              if (!useConstRHS)
                {
                  FORT_GETRHSPOIS(CHF_FRA(rhsFab),
                                  CHF_BOX(rhsFab.box()),
                                  CHF_BOX(domlev),
                                  CHF_CONST_REAL(dxlev),
                                  CHF_CONST_REAL(rhono),
                                  CHF_CONST_REAL(rno),
                                  CHF_CONST_INT(iprob));          
                }
          }
        }
#ifdef MPI
      MPI_Barrier(Chombo_MPI::comm);
#endif
    }
  return 0;
}

/*
  Set grid hierarchy from input file
 */
int setGrids(Vector<DisjointBoxLayout>& vectGrids,
             Vector<ProblemDomain>&     vectDomain, 
             Vector<Real>&              vectDx, 
             Vector<int>&               vectRefRatio, 
             int& numlevels, bool verbose)
{
  barrier();
  int max_level;
  Real fillRat = 0.77;
  std::vector<int> ancells(SpaceDim), refs;
  std::vector<Real>  prob_loa(SpaceDim);
  std::vector<Real>  prob_hia(SpaceDim);
  // set default to non-periodic
  std::vector<int> is_periodica(SpaceDim,0);
  bool is_periodic[SpaceDim];
  int temp;
  bool predefinedGrids = false;
  bool gridsFromFile = false;
  string gridFile;
  int maxboxsize = 32;
  Real refine_thresh = 0.5;
 

  ParmParse pp("main");
  pp.get("max_level", max_level);
  numlevels = max_level + 1;
  //kluge to get things the same in 0.2
  //max_level = 1;
  pp.query("fill_ratio", fillRat);
  pp.getarr("n_cell", ancells, 0, SpaceDim);
  pp.getarr("ref_ratio", refs, 0, numlevels);
  vectRefRatio = refs;
  pp.getarr("prob_lo",prob_loa,0,SpaceDim);
  pp.getarr("prob_hi",prob_hia,0,SpaceDim);
  pp.queryarr("is_periodic", is_periodica,0, SpaceDim);
  pp.query("maxboxsize",maxboxsize);

  numlevels = max_level + 1;

  temp = predefinedGrids;
  pp.query("presetGrids", temp);
  predefinedGrids = (temp == 1);

  if (pp.contains("grids_file"))
    {
      gridsFromFile = true;
      pp.get("grids_file", gridFile);
    }

  bool useConstantRHS = false;
  Real constRHS = -1.0;
  if (pp.contains("constRHS"))
    {
      useConstantRHS = true;
      pp.get("constRHS", constRHS);
    }

  pp.query("refinement_threshold", refine_thresh);

  // grid generation parameters
  int maxrat = 2;
  for(int irat = 0; irat < vectRefRatio.size(); irat++)
    maxrat = Max(vectRefRatio[irat], maxrat);
  int blockFactor = maxrat*2;
  if (maxboxsize/blockFactor < 1) blockFactor = maxboxsize;
  int nesting_radius = 1;


  IntVect ivlo = IntVect::TheZeroVector();
  IntVect ivhi;
  for(int idir = 0; idir < SpaceDim; idir++)
    ivhi[idir] = ancells[idir] - 1;

  Box basedom = Box(ivlo, ivhi);

  vectGrids.resize(numlevels);
  vectDomain.resize(numlevels);
  vectDx.resize(numlevels);


  vectDx.resize(numlevels,0.0);
  vectDx[0] = (prob_hia[0]-prob_loa[0])/ancells[0];
  for(int ilev = 1; ilev < numlevels; ilev++)
    {
      assert(vectRefRatio[ilev-1] > 0);
      vectDx[ilev] = vectDx[ilev-1]/vectRefRatio[ilev-1];
    }


  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++) 
    {
      is_periodic[dim] = (is_periodica[dim] == 1);
      if (is_periodic[dim] && verbose && procID() == 0)
        pout() << "Using Periodic BCs in direction: " << dim << endl;
    }

  vectDomain[0] = ProblemDomain(basedom,is_periodic);
  for(int ilev = 1;ilev < numlevels; ilev++)
    {
      vectDomain[ilev] = refine(vectDomain[ilev-1],vectRefRatio[ilev-1]);
    }

  int maxLevel = numlevels-1;
  Vector<Vector<Box> > newBoxes(numlevels);
  Vector<Vector<Box> > oldBoxes(numlevels);

  // use predefined grid configuration designed to test quadCFInterp configs
  if (predefinedGrids) 
    {
      int nc = ancells[0];
      int nmi = nc/2;//16
      int nqu = nc/4;//8
      int ntf = (nc*3)/4;  //24
      int nte = (nc*3)/8; //12
      int nfe = (nc*5)/8; //20
#if(CH_SPACEDIM ==2)
      Box boxf1(IntVect(0, nqu), IntVect(nmi-1,ntf-1));
      Box boxf2(IntVect(nmi,nte), IntVect(ntf-1,nfe-1));
      Box boxf3(IntVect(nqu,0  ), IntVect(nfe-1,nqu-1));
      Box boxf4(IntVect(nfe,nqu), IntVect(nc -1,nte-1));
#else
      Box boxf1(IntVect(0, nqu,nqu), IntVect(nmi-1,ntf-1,ntf-1));
      Box boxf2(IntVect(nmi,nte,nte), IntVect(ntf-1,nfe-1,nfe-1));
      Box boxf3(IntVect(nqu,0,0  ), IntVect(nfe-1,nqu-1,nqu-1));
      Box boxf4(IntVect(nfe,nqu,nqu), IntVect(nc -1,nte-1,nte-1));
#endif
      //comment out for kluge 
      /**/
      IntVectSet tags;
      tags |= boxf1;
      tags |= boxf2;
      tags |= boxf3;
      tags |= boxf4;
      
      for(int ilev = 0; ilev <numlevels; ilev++)
        {
          oldBoxes[ilev].push_back(vectDomain[ilev].domainBox());
        }
      int baseLevel = 0;
      int topLevel  = numlevels - 2;
      int eekflag = 0;
      if(topLevel >= 0) 
        {
          int newFinestLev;
          BRMeshRefine meshrefine(vectDomain[0],vectRefRatio, fillRat,
                                  blockFactor, nesting_radius, maxboxsize);
          newFinestLev = meshrefine.regrid(newBoxes, tags, baseLevel, 
                                           topLevel, oldBoxes);
        } else {
          newBoxes = oldBoxes;
        }
      
      Vector< Vector<int> > procAssign;
      Real effRatio = 0.75;
      Vector< Vector<long> > loads(numlevels);
      for(int ilev = 0; ilev <numlevels; ilev++)
        {
          loads[ilev].resize(newBoxes[ilev].size());
          for(int ibox = 0; ibox < newBoxes[ilev].size() ; ibox++)
            {
              loads[ilev][ibox] = newBoxes[ilev][ibox].numPts();
            }
        }
      LoadBalance(procAssign, effRatio, newBoxes, loads, vectRefRatio);
      
      if(eekflag != 0)
        {
          cerr << "setGrids: loadBalance returned error code " 
               << eekflag << endl;
          return(1);
        }
      for(int ilev = 0; ilev <numlevels; ilev++)
        {
          vectGrids[ilev].define(newBoxes[ilev], procAssign[ilev],
                                 vectDomain[ilev]);
          vectGrids[ilev].close();
        }
    } 
  else if (gridsFromFile)
    {
      Vector<Vector<Box> > amrGrids(numlevels);
      Vector<Vector<int> > amrProcAssign(numlevels);

#ifdef MPI
      MPI_Barrier(Chombo_MPI::comm);
      if (procID() == 0) 
        {
#endif
      // read grids (and proc assignments, if MPI) from file
      ifstream is(gridFile.c_str(), ios::in);
      if (is.fail()) 
        {
          pout() << "Cannot open grids file " << gridFile << endl;
        }
#ifdef MPI
      if (verbose) 
        {
          pout() << "procID: " << procID() << "  opening gridfile \n"
                 << endl;          
        }
#endif
      // format of file -- number of levels, then for each level starting with 
      // level 1, number of grids on level, list of boxes (along with proc
      // assignments if mpi)
      int in_numLevels;
      while (is.get() != '=');
      is >> in_numLevels;
      if (verbose) 
        {
          pout() << "NumLevels in grids file = " << in_numLevels << endl;
        }

      numlevels = min(in_numLevels, numlevels);
      while (is.get() != '\n');
      amrGrids.resize(numlevels);
      int ngrid;
      for (int lev=0; lev<numlevels; lev++) 
        {
          while (is.get() != ':');
          is >> ngrid;
          if (verbose) 
            {
              pout() << "level " << lev << " numGrids = " << ngrid << endl;
              //pout() << "grids = " ;
            }
          while (is.get() != '\n');
          amrGrids[lev].resize(ngrid);
          amrProcAssign[lev].resize(ngrid);
          Box bx;
          int this_proc=0;

          for (int i=0; i<ngrid; i++) 
            {
              is >> bx;
#ifdef MPI
              while (is.get() != '[');
              is >> this_proc;
#endif
              // advance to next box
              while (char ch = is.get())
                {
                  if (ch == '#') break;
                  if (ch == '\n') break;
                }
              
              // assume that if we're defining the grids, we know what
              // we're doing (so comment out this test)
              /*              
              // quick check on box size
              Box bxRef(bx);
              if (bxRef.longside() > maxboxsize) {
                pout() << "Grid " << bx << " too large" << endl;
                MayDay::Error();
              }
              */ 

              if (verbose) {
                //pout() << bx << endl;
              }
              amrGrids[lev][i] = bx;
              amrProcAssign[lev][i] = this_proc;
            } // end loop over boxes on this level
        } // end loop over levels
#ifdef MPI
        } // end if procID = 0
      broadcast(amrGrids,0);
      broadcast(amrProcAssign,0);      
#endif

      //LoadBalance(procAssign, effRatio, newBoxes, loads, vectRefRatio);

      //pout() << " number of boxes " << amrGrids.size() << endl;
      //pout() << " amrProcAssign " << amrProcAssign << endl;

      // now define disjointBoxLayouts
      for (int lev=0; lev<amrGrids.size(); lev++) 
        {
          const DisjointBoxLayout newDBL(amrGrids[lev], 
                                         amrProcAssign[lev], 
                                         vectDomain[lev]);
          vectGrids[lev] = newDBL;          
        }
    

    } else {
      // determine grids dynamically, based on grad(RHS)
      // will need temp storage for RHS
      Vector<LevelData<FArrayBox>* > vectRHS(maxLevel+1,NULL);
      int ncomps = 1;

      // define base level first
      Vector< Vector<int> > procAssign(maxLevel+1);
      domainSplit(vectDomain[0], oldBoxes[0], maxboxsize, blockFactor);
      procAssign[0].resize(oldBoxes[0].size());
      LoadBalance(procAssign[0],oldBoxes[0]);
      
      vectGrids[0].define(oldBoxes[0],procAssign[0],vectDomain[0]);
      
      vectRHS[0] = new LevelData<FArrayBox>(vectGrids[0], ncomps, 
                                            IntVect::TheZeroVector());

      int topLevel = 0;

      bool moreLevels = (maxLevel > 0);
      
      // create grid generation object
      BRMeshRefine meshrefine(vectDomain[0], vectRefRatio, fillRat,
                              blockFactor, nesting_radius, maxboxsize);

      while (moreLevels) 
        {
          // default is moreLevels = false
          // (only repeat loop in the case where a new level 
          // is generated which is still less than maxLevel)
          moreLevels = false;

          int baseLevel = 0;
          int oldTopLevel = topLevel;

          // now initialize RHS for this existing hierarchy
          setRHS(vectRHS, vectDx, vectDomain, topLevel+1, verbose);

          Vector<IntVectSet> tagVect(topLevel+1);
          int tags_grow = 1;
          tagCells(vectRHS, tagVect, vectDx, vectDomain, refine_thresh,
                   tags_grow, baseLevel, topLevel+1);

//           if (procID() == uniqueProc(SerialTask::compute))
//             {
//               if (topLevel >= 0 && ! tagVect[topLevel].isEmpty())
//                 {
                  
                  int new_finest = meshrefine.regrid(newBoxes, tagVect, 
                                                     baseLevel, 
                                                     topLevel, oldBoxes);
                  
                  if (new_finest > topLevel)
                    topLevel++;
                  
       //          } // end if there was a new level generated
//             } // end if proc is serial node
//           broadcast(topLevel, uniqueProc(SerialTask::compute) );
//           broadcast(newBoxes, uniqueProc(SerialTask::compute) );

          oldBoxes = newBoxes;

          //  no need to do this for the base level (already done)
          for (int lev=1; lev<= topLevel; lev++) 
            {
              // do load balancing
              procAssign[lev].resize(newBoxes[lev].size());
              LoadBalance(procAssign[lev], newBoxes[lev]);
              const DisjointBoxLayout newDBL(newBoxes[lev], procAssign[lev],
                                             vectDomain[lev]);
              vectGrids[lev] = newDBL;
              delete vectRHS[lev];
              vectRHS[lev] = new LevelData<FArrayBox>(vectGrids[lev], ncomps,
                                                      IntVect::TheZeroVector());
            } // end loop over levels for initialization
          
          // figure out whether we need another pass through grid generation
          if ((topLevel<maxLevel) && (topLevel > oldTopLevel))
            moreLevels = true;

        } // end while moreLevels loop
      // clean up temp storage
      for (int ilev=0; ilev <vectRHS.size(); ilev++)
        {
          if (vectRHS[ilev] != NULL) 
            {
              delete vectRHS[ilev];
              vectRHS[ilev] = NULL;
            }
        }

    } // end if grids generated dynamically
  
  return 0;
}

/*
  tag cells for refinement based on magnitude(RHS)
*/
void
tagCells(Vector<LevelData<FArrayBox>* >& vectRHS,
         Vector<IntVectSet>& tagVect,
         Vector<Real>& vectDx,
         Vector<ProblemDomain>& vectDomain,
         const Real refine_thresh, 
         const int tags_grow,
         const int baseLevel, 
         int numLevels) 
{
  for (int lev=baseLevel; lev!= numLevels; lev++) 
    {
      IntVectSet local_tags;
      LevelData<FArrayBox> & levelRhs = *vectRHS[lev];
      DisjointBoxLayout level_domain = levelRhs.getBoxes();
      DataIterator dit = levelRhs.dataIterator();

      Real maxRHS = 0;

      maxRHS = norm(levelRhs, levelRhs.interval(), 0);

      Real tagVal = maxRHS * refine_thresh;
      
      // now loop through grids and tag cells where RHS > tagVal
      for (dit.reset(); dit.ok(); ++dit) 
        {
          const Box thisBox = level_domain.get(dit());
          const FArrayBox& thisRhs = levelRhs[dit()];
          BoxIterator bit(thisBox);
          for (bit.begin(); bit.ok(); ++bit) 
            {
              const IntVect& iv = bit();
              if (abs(thisRhs(iv)) >= tagVal)
                local_tags |= iv;
            }
        } // end loop over grids on this level
      
      local_tags.grow(tags_grow);
      const Box& domainBox = vectDomain[lev].domainBox();
      local_tags &= domainBox;
      //  changed by bvs Fri Jan 16 14:40:45 PST 2004
      //
      //  somehow I had managed to not update the relationship between MeshRefine::regrid
      //  change and the AMRPoisson example code.
      tagVect[lev] = local_tags;
 //      int numTags = local_tags.numPts();
//       Vector<IntVectSet> all_tags;
//       const int dest_proc = uniqueProc(SerialTask::compute);
//       gather (all_tags, local_tags, dest_proc);
//       if (procID() == uniqueProc (SerialTask::compute)) {
//         for (int i=0; i< all_tags.size(); ++i) 
//           {
//             tagVect[lev] |= all_tags[i];
//           }
//       }
//       numTags = tagVect[lev].numPts();
    } // end loop over levels
}  


/*
  set domain boundary conditions from input file
  note that periodic BC's will override whichever physical
  boundary conditions are set here.
 */
int
setDomainBC(DomainGhostBC& domghostbc, bool verbose, int ncomp)
{
  barrier();
  std::vector<int> ibclo;
  std::vector<int> ibchi;
  ParmParse pp("main");
  pp.getarr("bc_lo", ibclo, 0, SpaceDim);
  pp.getarr("bc_hi", ibchi, 0, SpaceDim);

  Interval comps(0,ncomp-1);

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //lo bc
      {
        NeumannBC neumbc(idir, Side::Lo,comps);
        DirichletBC dircbc(idir, Side::Lo,comps);
        if(ibclo[idir] == 0) 
          {
            if(verbose&& procID()==0)
              pout() << "Dirichlet bcs in low direction for side " << idir << endl;
            domghostbc.setBoxGhostBC(dircbc);
          }
        else if(ibclo[idir] == 1)
          {
            if(verbose&& procID()==0)
              pout() << "neumann bcs in low direction for side " << idir << endl;
            domghostbc.setBoxGhostBC(neumbc);
          }
        else
          {
            if(verbose)
              cerr << "setDomainBC:: bogus input bc_lo flag" << endl;
            return(1);
          }
      }
      //hi bc
      {
        NeumannBC neumbc(idir, Side::Hi,comps);
        DirichletBC dircbc(idir, Side::Hi,comps);
        if(ibchi[idir] == 0) 
          {
            if(verbose&& procID()==0)
              pout() << "Dirichlet bcs in high direction for side " << idir << endl;
            domghostbc.setBoxGhostBC(dircbc);
          }
        else if(ibchi[idir] == 1)
          {
            if(verbose&& procID()==0)
              pout() << "neumann bcs in high direction for side " << idir << endl;
            domghostbc.setBoxGhostBC(neumbc);
          }
        else
          {
            if(verbose)
              cerr << "setDomainBC:: bogus input bc_hi flag" << endl;
            return(2);
          }
      }
    }
  return(0);
}
