/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include  <iostream>
#include "LevelOp.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "PoissonBC.H"
#include "ParmParse.H"
#include "AMRSolver.H"
#include "PoissonOp.H"
#include "Vector.H"
#include "AMRIO.H"
#include "SPMD.H"
#include "BRMeshRefine.H"
//#include "LoadBalance.H"
#include "localFuncs.H"
#include "computeSum.H"
#include "PoissProbF_F.H"
#include "ProblemDomain.H"

//#define HALEM_PROC_SPEED
#ifdef HALEM_PROC_SPEED
#include <cstdio>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>
#endif

#include "NoelsPrettyPrint.H"
#include "whatisdefined.H"
#include "memusage.H"
#include "Timer.H"
Timer Everything        ("Everything", 0);
Timer TimeReadInput     ("read input",    Everything);
Timer TimeProblemSetup  ("problem setup", Everything);
Timer TimeInitializePhi ("init phi",      Everything);
Timer TimeSolveAMR      ("solveAMR",      Everything);
Timer TimeOutputData    ("output data",   Everything);

#ifdef TIMER
extern void writeOutNoelsSooperIterationCounterStuff(void);
#endif

int global_granularity=1; //temp hack for now (noel)
int write_bottomsmoother_file=0; //temp hack for now (noel)
int num_levelGSRBs_inPC=2; //temp hack for now (noel)
int global_relaxation_method=1;  // default GSRB    GS 0   GSRB 1
int global_mgRelax_numBottomGSRB=4;
int global_mgRelax_numSmoothUp=4;
int global_mgRelax_numSmoothDown=4;
int global_BS_algorithm=0; // default bicgstab    bicgstab 0  bisimple 1  cg 2  cgsq 3

double global_BS_tolerance = 1.0e-10;
double global_GSRB_relaxFactor = 1.0; // only matters if turned on in FORT_GSRB

// define main loop here because that's the way amrgodunov does it!
void mainLoop();

extern void dumpmemoryatexit(void);

//#define TRAP_FPE  //(should be off by default)
#ifdef TRAP_FPE
// funproto
static void __attribute__ ((constructor)) trapfpe(void);
#endif

int main(int argc, char* argv[])
{

#ifdef MPI
  MPI_Init(&argc, &argv);
  //registerDebugger();
#ifdef CH_AIX  
  H5dont_atexit();
#endif
  setChomboMPIErrorHandler();
  barrier();
#endif
  //scoping trick
  {
  int rank, number_procs;
#ifdef MPI
  MPI_Comm_rank(Chombo_MPI::comm, &rank);
  MPI_Comm_size(Chombo_MPI::comm, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif

  // init silly global variables
  IterationFlavor = 0;
  for (int i=0; i<MAX_ITERATION_COUNTERS; ++i) {
    NoelsSooperIterationCounter[i].Count = 0;
    NoelsSooperIterationCounter[i].Value = 0;
    NoelsSooperIterationCounter[i].NmaxIteration = 0;
    NoelsSooperIterationCounter[i].wc = 0.0;
  }

  if (rank==0) {
    pout() << " number_procs= " << number_procs << endl;
    dumpAllThoseCrazyDefines();
  }

  Timer::TimerInit(rank);

#ifdef HALEM_PROC_SPEED
  pout() << " Halem Processor Speed: " << HalemProcessorSpeed() << endl;
#endif

#ifdef TRAP_FPE
  trapfpe();
#endif
  

  Everything.start();
  
  // read inputs
  TimeReadInput.start();
  char* in_file = argv[1];
  NoelsDumpInputFile(in_file);
  ParmParse pp(argc-2, argv+2,NULL, in_file);
  
  barrier();
  if (argc < 2) 
    {
      cerr<< " usage " << argv[0] << " <input_file_name> " << endl;
      exit(0);
    }
  TimeReadInput.stop();

  TimeProblemSetup.start();
  print_memory_line("after reading input");
  
  int lbase;
  int iverbose;
  ParmParse pp2("main");
  pp2.get("lbase", lbase);
  pp2.get("verbose", iverbose);
  bool verbose = (iverbose <= 1);
  int ncomp = 2;
  pp2.query("ncomp", ncomp);
      

  DomainGhostBC domghostbc;
  int eekflag = setDomainBC(domghostbc, verbose,ncomp);
  if(eekflag != 0)
    {
      cerr << "error: setDomainBC returned error code  " 
           << eekflag << endl;
      exit(1);
    }
  PoissonOp levelop;
  levelop.setDomainGhostBC(domghostbc);
  print_memory_line("after PoissonOp levelop");

  // check to see if we can set bottom smoother here
  //NoOpSmoother testBottomSmoother;
  //levelop.setBottomSmoother(testBottomSmoother);

  int numlevels;
  Vector<DisjointBoxLayout> vectGrids;
  Vector<ProblemDomain> vectDomain;
  Vector<int> vectRefRatio;
  Vector<Real> vectDx;
  eekflag = setGrids(vectGrids, vectDomain, vectDx, 
                     vectRefRatio, numlevels, verbose);
  if(eekflag != 0)
    {
      cerr << "error: setGrids returned error code " 
           << eekflag << endl;
      exit(2);
    }
  
  Vector<LevelData<FArrayBox>* > phi(numlevels, NULL);
  Vector<LevelData<FArrayBox>* > rhs(numlevels, NULL);
  print_memory_line("after Vector<LevelData<FArrayBox>* > phi and rhs");  

  bool constPhi = false;
  Real initialPhi = 0.0;
  if (pp2.contains("initialGuess"))
    {
      constPhi = true;
      pp2.get("initialGuess", initialPhi);
    }

  char tempstring1[120];
  long long NpointsAllLevels=0;
  for(int ilev = 0; ilev < numlevels; ilev++)
    {
      phi[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], ncomp, IntVect::TheUnitVector());
      rhs[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], ncomp, IntVect::TheZeroVector());
      LevelData<FArrayBox>& phifabs= *phi[ilev];

      int Nboxes=0, Npoints=0, NpointsFromDataIterator=0;
      NoelsLDFABgetUsefulInfo(phifabs, Nboxes, Npoints, NpointsFromDataIterator);
      NpointsAllLevels += Npoints;
      sprintf(tempstring1, "number of points updated at level %2d = %16ld  boxes=%10ld NpointsFromDataIterator=%12ld\n",
              ilev, Npoints, Nboxes, NpointsFromDataIterator);
      pout() << tempstring1;
      print_memory_line(" after new on this level");

      Real rval = Real(3*ilev);
      for(DataIterator dit = phifabs.dataIterator(); dit.ok(); ++dit)
      {

        rval += 1.0;
        if (constPhi)
          {
            // option to set phi to constant-value initial guess 
            phifabs[dit()].setVal(initialPhi);
          }
        else
          {          
            phifabs[dit()].setVal(rval);
          }
      }
    }
  sprintf(tempstring1, "total number of points updated       = %16ld",
          NpointsAllLevels);
  pout() << tempstring1;
  sprintf(tempstring1, "   %20.5e\n", (Real)NpointsAllLevels);
  pout() << tempstring1;

  print_memory_line("before setRHS");

  eekflag = setRHS(rhs, vectDx, vectDomain, numlevels, verbose);
  if(eekflag != 0)
    {
      cerr << "error: setRHS returned error code " 
           << eekflag << endl;
      exit(3);
    }
  
  // now compute sum(RHS) as a diagnostic
  // (should be 0 for all-Neumann or periodic BC's for solvability).
  for (int comp=0; comp<ncomp; comp++) 
    {
      Interval comps(comp,comp);
      Real sumRHS = computeSum(rhs,vectRefRatio,
                               vectDx[0], comps, 0);
      
      if (verbose && procID()==0)
        pout() << "Component " << comp << ",  Sum(RHS) = " << sumRHS << endl;
      
    } // end loop over components
  
  print_memory_line("before amrSolver creation");
  
  AMRSolver amrSolver(vectGrids,  vectDomain,
                      vectDx,     vectRefRatio, 
                      numlevels, lbase, &levelop,
                      ncomp);
  print_memory_line("after amrSolver creation");

  amrSolver.setVerbose(verbose);
  int maxiter;
  pp2.get("max_iterations", maxiter);
  amrSolver.setMaxIter(maxiter);
  int numvbot;
  pp2.get("num_v_cycles_bottom", numvbot);
  amrSolver.setNumVCyclesBottom(numvbot);
  if (pp2.contains("solverTol"))
    {
      Real solverTol;
      pp2.get("solverTol", solverTol);
      amrSolver.setTolerance(solverTol);
    }
  
  if (pp2.contains("numSmoothUp"))
    {
      int numSmoothUp;
      pp2.get("numSmoothUp", numSmoothUp);
      amrSolver.setNumSmoothUp(numSmoothUp);
      pout() << "numSmoothUp = " << numSmoothUp << endl;
    }
  
  
  if (pp2.contains("numSmoothDown"))
    {
      int numSmoothDown;
      pp2.get("numSmoothDown", numSmoothDown);
      amrSolver.setNumSmoothDown(numSmoothDown);
      pout() << "numSmoothDown = " << numSmoothDown << endl;
    }
  TimeProblemSetup.stop();
  
  // Add the ability to loop over the solve (the same solve) for benchmarking 
  //  purposes only.  (ndk)  note that it will default to one time thru...
  int solve_iterations = 1;
  pp2.get("iterations", solve_iterations);
  for(int i=0; i<solve_iterations; i++) {
    
    TimeInitializePhi.start();
#define INITIALIZE_PHI_EACH_ITERATION
#ifdef INITIALIZE_PHI_EACH_ITERATION
    for(int ilev = 0; ilev < numlevels; ilev++)
      {
        LevelData<FArrayBox>& phifabs= *phi[ilev];        
        Real rval = Real(3*ilev);
        
        for(DataIterator dit = phifabs.dataIterator(); dit.ok(); ++dit)
          {
            rval += 1.0;
            if (constPhi)
              {
                // option to set phi to constant-value initial guess 
                phifabs[dit()].setVal(initialPhi);
              }
            else
              {          
                phifabs[dit()].setVal(rval);
              }
          }
      }
    eekflag = setRHS(rhs, vectDx, vectDomain, numlevels, verbose);
    print_memory_line("after setRHS");
    if(eekflag != 0)
      {
        cerr << "error: setRHS returned error code " 
             << eekflag << endl;
        exit(3);
      }
#endif
    TimeInitializePhi.stop();

    Timer onesolve;
    onesolve.start();
    TimeSolveAMR.start();
    IterationFlavor = 1; //+ m_level;
    amrSolver.solveAMR(phi, rhs, false);
    TimeSolveAMR.stop();
    onesolve.stop();
    sprintf(tempstring1, " time to solve = %14.3f\n", onesolve.wc_time());
    pout() << tempstring1;
    print_memory_line("after solveAMR");
  }

  TimeOutputData.start();
  //#define WRITE_OUTPUT
#ifdef WRITE_OUTPUT
  eekflag = outputData(phi, rhs, vectGrids, vectDomain,
                     vectRefRatio, vectDx[0], numlevels, verbose);
  if(eekflag != 0)
  {
    cerr << "error: outputData returned error code " 
         << eekflag << endl;
    exit(4);
  }
#endif
  for(int ilev = 0; ilev < numlevels; ilev++)
    {
      delete phi[ilev];
      delete rhs[ilev];
    }
  
  writeOutNoelsSooperIterationCounterStuff();
  
  Real end_memory = get_memory_usage_from_OS();
  
#ifdef MPI
  // Gather peak memory from procs and write a single line to screen.  that's all.
  Real avg_memory, min_memory, max_memory;
  gather_memory_from_procs(end_memory, avg_memory, min_memory, max_memory);
#endif

  TimeOutputData.stop();

  Everything.stop();
  
  print_memory_line("after all");

  pout() << "Everything done.  mem= " << end_memory 
         << "MB wall-clock time: "    << setiosflags(ios::fixed)
         << Everything.wc_time()      << " seconds" << endl;
    
  Timer::TimerSummary();



#ifdef COUNT_PEX  
  FILE *PEXOUT;
  char filename[80];
  sprintf(filename, "pex%03d.dat", rank);
  PEXOUT = fopen(filename, "w");
  if(PEXOUT == NULL) {
    printf("problem opening pex out\n");  
    exit(0); 
  }
  if (rank==0) printf("rank%d NPEX = %d\n", rank, NPEX);
  for(int i=1; i<NPEX; ++i) { 
    fprintf(PEXOUT, "%6d %5d %10d %10d %10d %10.2e\n", i, pex_Nsends[i], 
            pex_Nlocal[i], pex_Nto[i], pex_Nfrom[i], 1000*pex_delta_time[i]);
    if (pex_Nlocal[i] == 0 && pex_delta_time[i] < 1.0e-16) break;
  }
  fclose(PEXOUT);
#endif

  }//end scoping trick
#ifdef MPI
#ifndef CH_AIX
  //dumpmemoryatexit(); // was hanging at end...
#endif
  MPI_Finalize();
#endif

  return(0);
}
