/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdio>
#include <string>
using std::string;

#include "LoHiSide.H"

#include "GodunovUtilities.H"
#include "LoHiCenter.H"
#include "GodunovPhysics.H"
#include "GodunovUtilitiesF_F.H"

#include "Timer.H"
#include "TimerPrototypes.H"

// Flag everything as not defined or set
GodunovUtilities::GodunovUtilities()
{
  m_isDefined = false;
}

GodunovUtilities::~GodunovUtilities()
{
}

// Define this object and the boundary condition object
void GodunovUtilities::define(ProblemDomain& a_domain,
                              const Real&    a_dx)
{
  // Store the domain and grid spacing
  m_domain = a_domain;
  m_dx = a_dx;
  m_isDefined = true;
}

// Compute the flattening coefficients from the primitive variables
void GodunovUtilities::computeFlattening(FArrayBox&       a_flattening,
                                         const FArrayBox& a_W,
                                         const Interval&  a_velInt,
                                         const int&       a_presInd,
                                         const int&       a_bulkModulusInd,
                                         const Box&       a_box)
{
  assert(m_isDefined);
  assert(a_W.box().contains(a_box));

  // The current direction
  int idir;

  // The directional flattening coefficients
  FArrayBox zetaDir(a_box,SpaceDim);

  // The divergence of the velocity
  FArrayBox dVel(a_box,SpaceDim);

  // The interval of the primitive variables corresponding to the velocity
  Interval velInterval= a_velInt;
  int v0index = velInterval.begin();

  // Get the directional flattening coefficients in each direction
  for (idir = 0; idir < SpaceDim; idir++)
    {
      // A box one larger (in direction "idir") than the final result box
      Box box1 = a_box;
      box1.grow(idir,1);
      box1 &= m_domain;

      // A box two larger (in direction "idir") than the final result box
      Box box2 = a_box;
      box2.grow(idir,2);
      box2 &= m_domain;

      // A box three larger (in direction "idir") than the final result box
      Box box3 = a_box;
      box3.grow(idir,3);
      box3 &= m_domain;

      // The primitive variables need to be defined over the largest box
      assert(a_W.box().contains(box3));

      // Compute where centered differences can be used and where one sided
      // differences need to be used.  The data used for the differences is
      // defined on "box3"
      Box loBox,hiBox,centerBox,entireBox;
      int hasLo,hasHi;

      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                 box3,m_domain,idir);

      // Compute the first differences in "pressure"
      FArrayBox delta1p(entireBox,1);

      int pressIndex = a_presInd;

      TF_GETGRAD.start();
      FORT_GETGRAD(CHF_FRA1(delta1p,0),
                   CHF_CONST_FRA1(a_W,pressIndex),
                   CHF_CONST_INT(idir),
                   CHF_BOX(loBox),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(hiBox),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(centerBox));
      TF_GETGRAD.stop();

      // Compute where centered differences can be used and where one sided
      // differences need to be used.  The data used for the differences is
      // defined on "box2"
      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                 box2,m_domain,idir);

      // Compute the second differences in "pressure"
      FArrayBox delta2p(entireBox,1);

      TF_GETDPTWO.start();
      FORT_GETDPTWO(CHF_FRA1(delta2p,0),
                    CHF_CONST_FRA1(delta1p,0),
                    CHF_CONST_INT(idir),
                    CHF_BOX(loBox),
                    CHF_CONST_INT(hasLo),
                    CHF_BOX(hiBox),
                    CHF_CONST_INT(hasHi),
                    CHF_BOX(centerBox));
      TF_GETDPTWO.stop();

      // Compute a 3-way minimum of the "bulk modulus"
      FArrayBox bulkMin(entireBox,1);

      int bulkIndex = a_bulkModulusInd;

      TF_MIN3PTS.start();
      FORT_MIN3PTS(CHF_FRA1(bulkMin,0),
                   CHF_CONST_FRA1(a_W,bulkIndex),
                   CHF_CONST_INT(idir),
                   CHF_BOX(loBox),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(hiBox),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(centerBox));
      TF_MIN3PTS.stop();

      // Use the first and second differences normalized by the 3-way minimum
      // computed above to generate directional flattening coefficients
      FArrayBox zetaTwiddleDir(entireBox,1);

      TF_GETFLAT.start();
      FORT_GETFLAT(CHF_FRA1(zetaTwiddleDir,0),
                   CHF_CONST_FRA1(delta1p,0),
                   CHF_CONST_FRA1(delta2p,0),
                   CHF_CONST_FRA1(bulkMin,0),
                   CHF_BOX(entireBox));
      TF_GETFLAT.stop();

      // Compute where centered differences can be used and where one sided
      // differences need to be used.  The data used for the differences is
      // defined on "box1"
      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                 box1,m_domain,idir);

      // Take a 3-way minimum of the directional flattening coefficients
      TF_MIN3PTS.start();
      FORT_MIN3PTS(CHF_FRA1(zetaDir,idir),
                   CHF_CONST_FRA1(zetaTwiddleDir,0),
                   CHF_CONST_INT(idir),
                   CHF_BOX(loBox),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(hiBox),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(centerBox));
      TF_MIN3PTS.stop();

      // Compute each component of the divergence of the velocity
      TF_GETGRAD.start();
      FORT_GETGRAD(CHF_FRA1(dVel,idir),
                   CHF_CONST_FRA1(a_W,v0index+idir),
                   CHF_CONST_INT(idir),
                   CHF_BOX(loBox),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(hiBox),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(centerBox));
      TF_GETGRAD.stop();
    }

  // At each point, set the flattening coefficient to the minimum of all
  // the directional flattening coefficients if the divergence of the velocity
  // is negative, otherwise set it to 1 (no flattening).
  TF_MINFLAT.start();
  FORT_MINFLAT(CHF_FRA1(a_flattening,0),
               CHF_CONST_FRA(zetaDir),
               CHF_CONST_FRA(dVel),
               CHF_BOX(a_box));
  TF_MINFLAT.stop();
}
void
GodunovUtilities::applyFlattening(
                                        FArrayBox& a_dW,
                                  const FArrayBox& a_flat,
                                  const Box&       a_box)
{
  assert(a_dW.box().contains(a_box));
  assert(a_flat.box().contains(a_box));
  int numSlopes = a_dW.nComp();
  for (int islope = 0;islope < numSlopes;islope++)
  {
    a_dW.mult(a_flat,a_box,0,islope);
  }
}


////////////
void
GodunovUtilities::oneSidedDifferences(FArrayBox&       a_dWMinu,
                                      FArrayBox&       a_dWPlus,
                                      const FArrayBox& a_W,
                                      const int&       a_dir,
                                      const Box&       a_box)
{
  int numSlopes = a_dWMinu.nComp();

  // A box one larger (in direction "a_dir") than the final result box
  Box box1 = a_box;
  box1.grow(a_dir,1);
  box1 &= m_domain;

  // Compute where centered differences can be used and where one sided
  // differences need to be used.
  Box loBox,hiBox,centerBox,entireBox;
  int hasLo,hasHi;

  loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
             box1,m_domain,a_dir);

  // Compute 2nd order slopes - including one sided differences
  FArrayBox deltaWC(entireBox, numSlopes);

  slopes(deltaWC, a_dWMinu, a_dWPlus, a_W, numSlopes, a_dir,
         loBox, hiBox, centerBox, entireBox, hasLo, hasHi);
}

////////
void GodunovUtilities::slopes(FArrayBox&       a_dWCent,
                              FArrayBox&       a_dWMinu,
                              FArrayBox&       a_dWPlus,
                              const FArrayBox& a_W,
                              const int&       a_numSlopes,
                              const int&       a_dir,
                              const Box&       a_loBox,
                              const Box&       a_hiBox,
                              const Box&       a_centerBox,
                              const Box&       a_entireBox,
                              const int&       a_hasLo,
                              const int&       a_hasHi)
{
  assert(a_dWCent.nComp() == a_numSlopes);
  assert(a_dWMinu.nComp() == a_numSlopes);
  assert(a_dWPlus.nComp() == a_numSlopes);
  assert(a_W.nComp() >= a_numSlopes);

  assert(a_dWCent.box().contains(a_entireBox));
  assert(a_dWMinu.box().contains(a_entireBox));
  assert(a_dWPlus.box().contains(a_entireBox));
  assert(a_W.box().contains( a_entireBox));

  assert((a_dir >= 0 )&& (a_dir < SpaceDim));

  TF_SECONDSLOPEDIFFS.start();
  FORT_SECONDSLOPEDIFFS(CHF_FRA(a_dWCent),
                        CHF_FRA(a_dWMinu),
                        CHF_FRA(a_dWPlus),
                        CHF_CONST_FRA(a_W),
                        CHF_CONST_INT(a_numSlopes),
                        CHF_CONST_INT(a_dir),
                        CHF_BOX(a_loBox),
                        CHF_CONST_INT(a_hasLo),
                        CHF_BOX(a_hiBox),
                        CHF_CONST_INT(a_hasHi),
                        CHF_BOX(a_centerBox));
  TF_SECONDSLOPEDIFFS.stop();


}

////////
void GodunovUtilities::vanLeerSlopes(FArrayBox&       a_dW,
                                     const FArrayBox& a_W,
                                     const int&       a_numSlopes,
                                     const int&       a_dir,
                                     const Box&       a_box)
{
  // A box one larger (in direction "a_dir") than the final result box
  Box box1 = a_box;
  box1.grow(a_dir,1);
  box1 &= m_domain;

  // Compute where centered differences can be used and where one sided
  // differences need to be used.
  Box loBox,hiBox,centerBox,entireBox;
  int hasLo,hasHi;

  loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
             box1,m_domain,a_dir);

  // Compute 2nd order slopes - including one sided differences
  FArrayBox dWMinu(entireBox,a_numSlopes);
  FArrayBox dWPlus(entireBox,a_numSlopes);

  slopes(a_dW, dWMinu, dWPlus, a_W, a_numSlopes, a_dir,
         loBox, hiBox, centerBox, entireBox, hasLo, hasHi);


  // Apply the slope limiter.

  slopeLimiter(a_dW,dWMinu,dWPlus,a_numSlopes,a_dir,centerBox);

}
void GodunovUtilities::fourthOrderSlopes(FArrayBox&       a_dW,
                                         const FArrayBox& a_W,
                                         const FArrayBox& a_dWvL,
                                         const int&       a_numSlopes,
                                         const int&       a_dir,
                                         const Box&       a_box)
{
  // Number of slopes to compute
  int numSlope = a_numSlopes;

  assert(a_dW.nComp() == numSlope);
  assert(a_W.nComp() >= numSlope);

  // A box one larger (in direction "a_dir") than the final result box
  Box box1 = a_box;
  box1.grow(a_dir,1);
  box1 &= m_domain;

  // Compute where centered differences can be used and where one sided
  // differences need to be used.
  Box loBox,hiBox,centerBox,entireBox;
  int hasLo,hasHi;

  loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
             box1,m_domain,a_dir);

  assert(a_dW.box().contains(entireBox));

  TF_FOURTHSLOPEDIFFS.start();
  FORT_FOURTHSLOPEDIFFS(CHF_FRA(a_dW),
                        CHF_CONST_FRA(a_W),
                        CHF_CONST_FRA(a_dWvL),
                        CHF_CONST_INT(numSlope),
                        CHF_CONST_INT(a_dir),
                        CHF_BOX(loBox),
                        CHF_CONST_INT(hasLo),
                        CHF_BOX(hiBox),
                        CHF_CONST_INT(hasHi),
                        CHF_BOX(centerBox));
  TF_FOURTHSLOPEDIFFS.stop();
}
// Piecewise linear normal predictor. Computes increments in the
// characteristic amplitudes.

void GodunovUtilities::PLMNormalPredictor(FArrayBox&       a_dWCharLo,
                                          FArrayBox&       a_dWCharHi,
                                          const FArrayBox& a_dWChar,
                                          const FArrayBox& a_Lambda,
                                          const Real&      a_dtbydx,
                                          const Box&       a_box)
{
  int numPrim = a_dWChar.nComp();
  assert(a_dWCharLo.nComp() == numPrim);
  assert(a_dWCharHi.nComp() == numPrim);
  assert(a_Lambda.nComp() == numPrim);
  assert(a_dWCharLo.box().contains(a_box));
  assert(a_dWCharHi.box().contains(a_box));
  assert(a_dWChar.box().contains(a_box));
  assert(a_Lambda.box().contains(a_box));

  TF_PLMPRED.start();
  FORT_PLMPRED(CHF_FRA(a_dWCharLo),
               CHF_FRA(a_dWCharHi),
               CHF_CONST_FRA(a_dWChar),
               CHF_CONST_FRA(a_Lambda),
               CHF_CONST_REAL(a_dtbydx),
               CHF_CONST_INT(numPrim),
               CHF_BOX(a_box));
  TF_PLMPRED.stop();
}

// Increment fluxes with artificial viscosity.

void GodunovUtilities::artificialViscosity(FArrayBox&       a_F,
                                           const FArrayBox& a_U,
                                           const FArrayBox& a_divVel,
                                           const Real&      a_coeff,
                                           const int&       a_dir,
                                           const Box&       a_box)
{
  assert((a_dir >= 0 )&& (a_dir < SpaceDim));
  assert(a_F.box().contains(a_box));
  assert(a_divVel.box().contains(a_box));
  assert(a_coeff >= 0.0);
  //U is centered diff than a_box

  TF_ARTVISC.start();
  FORT_ARTVISC(CHF_FRA(a_F),
               CHF_CONST_FRA(a_U),
               CHF_CONST_FRA1(a_divVel,0),
               CHF_CONST_REAL(a_coeff),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
  TF_ARTVISC.stop();
}

// Apply a van Leer limiter directly to the slopes.

void GodunovUtilities::slopeLimiter(FArrayBox&       a_dW,
                                    const FArrayBox& a_dWLeft,
                                    const FArrayBox& a_dWRigh,
                                    const int&       a_numSlopes,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  assert(m_isDefined);
  assert(a_dW.nComp()      == a_numSlopes);
  assert(a_dWLeft.nComp()  == a_numSlopes);
  assert(a_dWRigh.nComp() == a_numSlopes);
  assert((a_dir >= 0) && a_dir < SpaceDim);
  assert(a_dW.box().contains(a_box));
  assert(a_dWLeft.box().contains(a_box));
  assert(a_dWRigh.box().contains(a_box));

  TF_VANLEERLIMITER.start();
  FORT_VANLEERLIMITER(CHF_FRA(a_dW),
                      CHF_CONST_FRA(a_dWLeft),
                      CHF_CONST_FRA(a_dWRigh),
                      CHF_CONST_INT(a_numSlopes),
                      CHF_CONST_INT(a_dir),
                      CHF_BOX(a_box));
  TF_VANLEERLIMITER.stop();
}

// Compute a face centered divergence of the velocity
void GodunovUtilities::divVel(FArrayBox&       a_divVel,
                              const FArrayBox& a_W,
                              const Interval&  a_velInt,
                              const int&       a_dir,
                              const Box&       a_box)
{
  assert((a_dir >= 0 )&& (a_dir < SpaceDim));
  Box dveltanBox = a_divVel.box();
  dveltanBox.enclosedCells(a_dir);
  dveltanBox.grow(1);
  dveltanBox &= m_domain;

  // First, we need to calculate the directional derivatives of
  // the tangential components of velocity at the cell-centers.
  FArrayBox dveltan(dveltanBox,SpaceDim-1);

  // Get the interval of the primitive variables corresponding to the velocity
  int v0index = a_velInt.begin();

  // Go through the tangential directions
  for (int i = 0, dir = 0; dir < SpaceDim; ++dir)
    {
      if (dir != a_dir)
        {
          // This velocity component is tangential.  Build the box in which
          // d(v[dir])/d(x[dir]) is to be computed.
          Box primBox = a_divVel.box();
          primBox.enclosedCells(a_dir);
          primBox.grow(dir,1).grow(a_dir,1);
          primBox &= m_domain;

          // Compute where centered differences can be used and where one sided
          // differences need to be used.  The data used for the differences is
          // defined on "dveltanBox"
          Box gradBox,hiBox,loBox,centerBox;
          int hasLo,hasHi;

          loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,gradBox,
                     primBox,m_domain,dir);

          // Compute d(v[dir])/d(x[dir]).
          TF_GETGRAD.start();
          FORT_GETGRAD(CHF_FRA1(dveltan,i),
                       CHF_CONST_FRA1(a_W,v0index+dir),
                       CHF_CONST_INT(dir),
                       CHF_BOX(loBox),
                       CHF_CONST_INT(hasLo),
                       CHF_BOX(hiBox),
                       CHF_CONST_INT(hasHi),
                       CHF_BOX(centerBox));
          TF_GETGRAD.stop();
          ++i;
        }
    }

  // Now, we can calculate the divergence of the normal velocity
  // at the center normal-direction edges. To do this, we determine
  // which edges at which we have sufficient data to compute centered
  // estimates of h*(div(u)). At the remaining edges. i.e. those
  // corresponding to the physical boundaries, we use zeroth-order
  // extrapolation.

  Box divBox = a_divVel.box();
  divBox.enclosedCells(a_dir);
  divBox.grow(a_dir,1);
  divBox &= m_domain;

  Box loBox,hiBox,centerBox,entireBox;
  int hasLo,hasHi;

  loHiCenterFace(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                 divBox,m_domain,a_dir);

  // All of the boxes computed above are shifted so as to be cell-centered,
  // with the index of the cell center being identified with the low edge.
  // We then shift a_divVel to be compatible with that convention on input,
  // and undo the shift on output.  Basically, everything is made compatible
  // a_W (which is cell-centered).

  loBox.shiftHalf(a_dir,1);
  centerBox.shiftHalf(a_dir,1);
  hiBox.shiftHalf(a_dir,1);

  a_divVel.shiftHalf(a_dir,1);

  TF_DIVUEDGE.start();
  FORT_DIVUEDGE(CHF_FRA1(a_divVel,0),
                CHF_CONST_FRA1(a_W,v0index+a_dir),
                CHF_CONST_FRA(dveltan),
                CHF_CONST_INT(a_dir),
                CHF_BOX(loBox),
                CHF_CONST_INT(hasLo),
                CHF_BOX(hiBox),
                CHF_CONST_INT(hasHi),
                CHF_BOX(centerBox));
  TF_DIVUEDGE.stop();

  a_divVel.shiftHalf(a_dir,-1);
}
void GodunovUtilities::PPMFaceValues(
                          FArrayBox& a_WFace,
                    const FArrayBox& a_W,
                    const int&       a_numSlopes,
                    const int&       a_dir,
                    const Box&       a_box,
                    const GodunovPhysics* a_physPtr)
{

// Input box is the face-centered box on which a_WFace is computed.

// No limiter needs to be applied to these values, since the van Leer
// limiter already takes care of such issues.

  assert(a_WFace.box().contains(a_box));

  Box interiorBox = a_box;
  interiorBox.enclosedCells();
  interiorBox.grow(a_dir,1);
  interiorBox &= m_domain;
  interiorBox.grow(a_dir,-1);
  interiorBox.surroundingNodes(a_dir);
  Real time = 0;
  Box vlBox = interiorBox;
  vlBox.enclosedCells(a_dir);
  vlBox.grow(a_dir,1);

  assert(a_W.box().contains(vlBox));

  FArrayBox dW(vlBox,a_numSlopes);
  vanLeerSlopes(dW,a_W,a_numSlopes,a_dir,vlBox);
  if (a_physPtr != NULL) a_physPtr->getPhysIBC()->setBdrySlopes(dW,a_W,a_dir,time);

  Box bdryBoxLo = bdryLo(m_domain,a_dir,1);
  bdryBoxLo &= a_box;
  Box bdryBoxHi = bdryHi(m_domain,a_dir,1);
  bdryBoxHi &= a_box;

  int hasLo = 1;
  if (bdryBoxLo.isEmpty()) hasLo = 0;
  int hasHi = 1;
  if (bdryBoxHi.isEmpty()) hasHi = 0;
  TF_PPMFACE.start();
  FORT_PPMFACE(
               CHF_FRA(a_WFace),
               CHF_CONST_FRA(a_W),
               CHF_CONST_FRA(dW),
               CHF_CONST_INT(a_numSlopes),
               CHF_CONST_INT(a_dir),
               CHF_BOX(interiorBox),
               CHF_CONST_INT(hasLo),
               CHF_BOX(bdryBoxLo),
               CHF_CONST_INT(hasHi),
               CHF_BOX(bdryBoxHi));
  TF_PPMFACE.stop();

}

void GodunovUtilities::PPMLimiter(
                          FArrayBox& a_dWLow,
                          FArrayBox& a_dWHigh,
                    const int&       a_numSlopes,
                    const int&       a_dir,
                    const Box&       a_box)
{
  TF_PPMLIMITER.start();
  FORT_PPMLIMITER(
                CHF_FRA(a_dWLow),
                CHF_FRA(a_dWHigh),
                CHF_CONST_INT(a_numSlopes),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
  TF_PPMLIMITER.stop();
}
void GodunovUtilities::PPMNormalPredictor(
                          FArrayBox& a_dWLow,
                          FArrayBox& a_dWHigh,
                    const FArrayBox& a_Lambda,
                    const Real&      a_dtbydx,
                    const int&       a_numSlopes,
                    const Box&       a_box)
{
  TF_PPMPRED.start();
  FORT_PPMPRED(
                CHF_FRA(a_dWLow),
                CHF_FRA(a_dWHigh),
                CHF_CONST_FRA(a_Lambda),
                CHF_CONST_REAL(a_dtbydx),
                CHF_CONST_INT(a_numSlopes),
                CHF_BOX(a_box));
  TF_PPMPRED.stop();
}
