/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _PATCH_GODUNOV_H_
#define _PATCH_GODUNOV_H_

#include "GodunovPhysics.H"
#include "GodunovUtilities.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"

///
/**
   The base class PatchGodunov provides an implementation of a second-order,
   unsplit Godunov method acting on a single grid/patch.  PatchGodunov
   provides an interface to the level integrator, LevelGodunov, which manages
   the entire level and flux corrections (via flux registers) to the coarser
   and finer levels.
 */
class PatchGodunov
{
public:
  ///

  /// Constructor
  /**
   */
  PatchGodunov();

  /// Destructor
  /**
   */
  virtual ~PatchGodunov();

  /// Define the object
  /**
   */
  virtual void define(const ProblemDomain&         a_domain,
                      Real                         a_dx,
                      bool                         a_useArtificialViscosity,
                      Real                         a_artificialViscosity,
                      bool                         a_useFlattening,
                      bool                         a_limitSlopes,
                      bool                         a_useCharLimiting,
                      const GodunovPhysics* const  a_gdnvPhysicsPtr);

  /// Set the current time before calling updateState, computeWHalf, computeUpdate.
  /**
   */
  virtual void setCurrentTime(const Real& a_currentTime);

  /// Set the current box before calling updateState, computeWHalf, computeUpdate.
  /**
   */
  virtual void setCurrentBox(const Box& a_currentBox);

  /// Update the conserved variables and return the fluxes used to do this.
  /**
     Update the conserved variables and return the final fluxes that were used
     for this.  Compute the fluxes using a second-order, unsplit Godunov method
     based on the input conserved variables, a_U, and source terms, a_S.  If
     there are no source terms then a_S should be null constructed.  Also
     return the maximum wave speed.
   */
  virtual void updateState(FArrayBox&       a_U,
                           FArrayBox        a_F[CH_SPACEDIM],
                           Real&            a_maxWaveSpeed,
                           const FArrayBox& a_S,
                           const Real&      a_dt,
                           const Box&       a_box);

  /// Compute the time-centered values of the primitive variables on cell faces.
  /**
     This API is used in cases where some operation over the whole
     level must be performed on the face-centered variables prior to the
     final difference update. Examples include incompressible flow and MHD,
     in which it is necessary to compute the projection of a face-centered
     vector field on its divergence-free part. To complete the differencing,
     it is necessary to call the member function computeUpdate.
   */
  virtual void computeWHalf(FArrayBox        a_WHalf[CH_SPACEDIM],
                            const FArrayBox& a_U,
                            const FArrayBox& a_S,
                            const Real&      a_dt,
                            const Box&       a_box);

  /// Compute the increment in the conserved variables from face variables.
  /**
   Compute dU = dt*dUdt, the change in the conserved variables over
   the time step. The fluxes are returned are suitable for use in refluxing.
   */
  virtual void computeUpdate(FArrayBox&       a_dU,
                             FArrayBox        a_Flux[CH_SPACEDIM],
                             const FArrayBox& a_U,
                             const FArrayBox  a_WHalf[CH_SPACEDIM],
                             const Real&      a_dt,
                             const Box&       a_box);


  void computeFluxes(FArrayBox        a_F[CH_SPACEDIM],
                     const FArrayBox  a_wHalf[CH_SPACEDIM],
                     const Box&       a_box);


  void
  PLMNormalPred(FArrayBox&       a_WMinu,
                FArrayBox&       a_WPlus,
                const Real&      a_dtbydx,
                const FArrayBox& a_W,
                const FArrayBox& a_flat,
                const int&       a_dir,
                const Box&       a_box);

  void
  PPMNormalPred(FArrayBox&       a_WMinu,
                FArrayBox&       a_WPlus,
                const Real&      a_dtbydx,
                const FArrayBox& a_W,
                const FArrayBox& a_flat,
                const int&       a_dir,
                const Box&       a_box);

  /// Is the object completely defined
  /**
     Return true if the object is completely defined.
   */
  virtual bool isDefined() const;

protected:

  bool m_useFlattening;
  bool m_limitSlopes;
  bool m_useCharLimiting;
  bool m_isSlopeSet;
  bool m_useArtificialViscosity;
  bool m_isArtViscSet;
  Real m_artificialViscosity;
  Real m_time;

  // Has define() been called
  bool m_isDefined;

  // Problem domain and grid spacing
  ProblemDomain m_domain;
  Real m_dx;
  GodunovPhysics*  m_gdnvPhysics;
  GodunovUtilities m_util;
  // Current time and has it been set
  Real m_currentTime;
  bool m_isCurrentTimeSet;

  // Current box and has it been set
  Box m_currentBox;
  bool m_isCurrentBoxSet;

private:
  // Disallowed for all the usual reasons
  void operator=(const PatchGodunov& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  PatchGodunov(const PatchGodunov& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
