/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LoHiSide.H"

#include "ExplosionIBC.H"
#include "ExplosionF_F.H"

#include "LoHiCenter.H"

#include "Timer.H"
#include "TimerPrototypes2.H"

// Null constructor
ExplosionIBC::ExplosionIBC()
{
  m_isFortranCommonSet = false;
}
ExplosionIBC::~ExplosionIBC()
{
  m_isFortranCommonSet = false;
}

// Constructor which defines parameters used by Fortran routines
ExplosionIBC::ExplosionIBC(const Real&     a_gamma,
                           const Real&     a_ms,
                           const RealVect& a_center,
                           const Real&     a_size,
                           const Real&     a_artvisc)
{
  setFortranCommon(a_gamma,
                   a_ms,
                   a_center,
                   a_size,
                   a_artvisc);
}

// Sets parameters in a common block used by Fortran routines:
//   a_gamma    - Gamma for polytropic, gamma-law gas
//   a_ms       - Mach shock number of the discontinuity
//   a_center   - Center of the explosion
//   a_size     - Initial radius of the explosion
//   a_velocity - Initial velocity of the gas
//   a_artvisc  - Artificial viscosity coefficient
void 
ExplosionIBC::setFortranCommon(const Real&     a_gamma,
                               const Real&     a_ms,
                               const RealVect& a_center,
                               const Real&     a_size,
                               const Real&     a_artvisc)
{
  assert(m_isFortranCommonSet == false);

  TF_SETEXPLOSION.start();
  FORT_SETEXPLOSIONF(CHF_CONST_REAL(a_gamma),
                     CHF_CONST_REAL(a_ms),
                     CHF_CONST_REALVECT(a_center),
                     CHF_CONST_REAL(a_size),
                     CHF_CONST_REAL(a_artvisc));
  TF_SETEXPLOSION.stop();

  m_isFortranCommonSet = true;
}

// Factory method - this object is its own factory:
//   Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
//   its define() must be called before it is used) and m_isFortranCommonSet
//   set to value of m_isFortranCommonset in the current (factory) object.
PhysIBC* ExplosionIBC::new_physIBC()
{
  ExplosionIBC* retval = new ExplosionIBC();
  retval->m_isFortranCommonSet = m_isFortranCommonSet;

  return static_cast<PhysIBC*>(retval);
}

void ExplosionIBC::
primBC(FArrayBox&            a_WGdnv,
       const FArrayBox&      a_Wextrap,
       const FArrayBox&      a_W,
       const int&            a_dir,
       const Side::LoHiSide& a_side,
       const Real&           a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  int isign = sign(a_side);
  Box gdnvBox = a_WGdnv.box();
  Box bdryBox;
  if (a_side == Side::Lo) bdryBox = bdryLo(m_domain,a_dir,1);
  if (a_side == Side::Hi) bdryBox = bdryHi(m_domain,a_dir,1);
  bdryBox &= gdnvBox;

  // Is there a domain boundary next to this grid
  if (!bdryBox.isEmpty())
    {
      TF_EXPLNBC.start();
      FORT_EXPLNBC(CHF_FRA(a_WGdnv),
                   CHF_CONST_FRA(a_Wextrap),
                   CHF_CONST_FRA(a_W),
                   CHF_CONST_INT(a_dir),
                   CHF_CONST_INT(isign),
                   CHF_BOX(bdryBox));
      TF_EXPLNBC.stop();
    }
}
// Set up initial conditions
void ExplosionIBC::
initialize(LevelData<FArrayBox>& a_U)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  DataIterator dit = a_U.boxLayout().dataIterator();
  
  // Iterator of all grids in this level
  for (dit.begin(); dit.ok(); ++dit)
    {
      // Storage for current grid
      FArrayBox& U = a_U[dit()];

      // Box of current grid
      Box uBox = U.box();
      uBox &= m_domain;
    
      // Set up initial condition in this grid
      TF_EXPLOSIONINIT.start();
      FORT_EXPLOSIONINITF(CHF_CONST_FRA(U),
                          CHF_CONST_REAL(m_dx),
                          CHF_BOX(uBox));
      TF_EXPLOSIONINIT.stop();
    }
}
