#include  <iostream>
#include "LevelOp.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "PoissonBC.H"
#include "ParmParse.H"
#include "AMRSolver.H"
#include "PoissonOp.H"
#include "Vector.H"
#include "AMRIO.H"
#include "SPMD.H"
#include "IntVectSet.H"
#include "LoadBalance.H"
#include "localfunctions.H"
#include "fortranfunctions.H"
#include "DebugOut.H"
//#include "BoxIterator.H"
//#include "LayoutIterator.H"
#include "BRMeshRefine.H"

typedef LevelData<FArrayBox> LDF;
typedef DisjointBoxLayout    DBL;

void compareTwoLDFs(const Vector<LDF*>& grid1, const Vector<LDF*>& grid2,
		    Vector<LDF*>& absDifference);

void coarseFineBoundaryCount(const DisjointBoxLayout& dbl,
			     int& numberCoarseFineBoundaryFaces,
			     int& numberProblemBoundaryFaces);

int main(int argc, char* argv[]) {

  int SIZE = 64;
  if(argc > 1 ) {
    SIZE = atoi(argv[1]);
  }
  
  int rank, number_procs;
#ifdef MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif
  cout << "rank = " << rank << " number_procs= " << number_procs << endl;
  { // parallel scope


  
  // for now, i just want a 2D problem and a 3D problem
  // that will execute at about the same length of time.
  // so i do the following:

#if CH_SPACEDIM == 2
  int maxLevel  = 3;  
  int Ncells[2] = {SIZE, SIZE};
#else
  int maxLevel  = 4; 
  int Ncells[3] = {8,8,8};
#endif

  int lbase               = 0;
  int iverbose            = 1;
  int maxIterations       = 20;
  int numberVcyclesBottom = 1;
  int blockFactor         = 8;
  int maxBoxSize          = 32;
  int bufferSize          = 1;
  int baseLevel           = 0; 

  int ibclo[3]  = {0,0,0};
  int ibchi[3]  = {0,0,0};

  //Real fillRatio          = 0.77;
  //Real refineThreshold    = 0.005;
  Real ProblemLowCord[3]  = {0.0, 0.0, 0.0};
  Real ProblemHighCord[3] = {1.0, 1.0, 1.0};

  bool verbose = (iverbose == 1);

  Vector<int> referenceRatioVector;
  referenceRatioVector.resize(maxLevel+1);
  referenceRatioVector[0] = 2;
  for(int i=1; i<=maxLevel; i++) {
    referenceRatioVector[i] = 2;
  }

  const IntVect unitVector(IntVect::TheUnitVector());

  DomainGhostBC domghostbc;
  setDomainBC(domghostbc, ibclo, ibchi, verbose);
  cout << "DomainGhostBC domghostbc built" << endl;

  PoissonOp poissonOperator;
  poissonOperator.setDomainGhostBC(domghostbc);
  cout << "PoissonOp poissonOperator built" << endl;

  int numLevels=1; // ??

  Vector<LDF*>                rhs(maxLevel+1, NULL);
  Vector<LDF*>           exactlap(maxLevel+1, NULL);
  Vector<DBL>           DBLVector(maxLevel+1);
  Vector<Box>   boundingBoxVector(maxLevel+1);
  Vector<Real>           DxVector(maxLevel+1);
  
  // make one big box with Ncells dimensions
  Box baseDomain = createBox(Ncells, SpaceDim);
  boundingBoxVector[0] = baseDomain;
  cout << " base domain -- boundingBoxVector[0]= " << baseDomain << endl;
  
  for(int ilev = 1;ilev <= maxLevel; ilev++) {
    boundingBoxVector[ilev] = 
      refine(boundingBoxVector[ilev-1],referenceRatioVector[ilev-1]);
    //cout << " boundingBoxVector[" << ilev 
    // << "]= " << boundingBoxVector[ilev] << endl;
  }
  
  DxVector[0] = (ProblemHighCord[0] - ProblemLowCord[0])/Ncells[0];
  for(int ilev = 1; ilev <= maxLevel; ilev++) {
    assert(referenceRatioVector[ilev-1] > 0);
    DxVector[ilev] = DxVector[ilev-1]/referenceRatioVector[ilev-1];
  }

  Vector<Vector<Box> >    amrBoxes(maxLevel+1);
  Vector<Vector<int> >     procIDs(maxLevel+1);
  
  // define base level first
  domainSplit(boundingBoxVector[0], amrBoxes[0], maxBoxSize, blockFactor);

  procIDs[0].resize(amrBoxes[0].size());
  LoadBalance(procIDs[0], amrBoxes[0]);
  
  cout << " amrBoxes[0].size() = " << amrBoxes[0].size() << endl;
  //cout << " procIDs[0]= "             << procIDs[0] << endl;
  cout << " procIDs[0].size() = "     << procIDs[0].size() << endl;
  cout << endl;
  
  DBLVector[0].define(amrBoxes[0], procIDs[0]);

  rhs[0] = new LDF(DBLVector[0], 1, unitVector);

  // now initialize data on this level
  LDF& levelrhs = *rhs[0];
  for (DataIterator dit = levelrhs.dataIterator(); dit.ok(); ++dit) {
    FORT_SETRHS(CHF_FRA( levelrhs[dit()]),
		CHF_BOX( levelrhs[dit()].box()),
		CHF_CONST_REAL(DxVector[0]));    
  }
  //prettyPrintLDF(rhs[0], "rhs");



  AMRSolver amrSolver(DBLVector,  boundingBoxVector,
                      DxVector,   referenceRatioVector, 
                      numLevels, lbase, &poissonOperator);

  amrSolver.setVerbose(verbose);
  amrSolver.setMaxIter(maxIterations);
  amrSolver.setNumVCyclesBottom(numberVcyclesBottom);

  Vector<LDF*> phi(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    phi[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
    //// do this if you want to give phi an initial value for each box
    // LDF& levelphi= *phi[ilev];      
    // Real rval = 3.0*ilev;
    // for(DataIterator dit = levelphi.dataIterator(); dit.ok(); ++dit) {
    //   rval += 1.0;
    //   levelphi[dit()].setVal(rval);
    // }
  }
  //amrSolver.solveAMR(phi, rhs);  
  //prettyPrintLDF(phi[0], "phi");

  // Compute the Laplacian of the RHS
  Vector<LDF*> rhs_lap(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    rhs_lap[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
    amrSolver.applyAMROperator(rhs, *rhs_lap[ilev], ilev);
  }
  //prettyPrintLDF(rhs_lap[0], "rhs lap");


  // instead of solving for phi, set it equal to f
  //for(int ilev = 0; ilev < numLevels; ilev++) { 
  // phi[ilev] = rhs[ilev];
  //}


  // now set the RHS to be it's laplacian
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    rhs[ilev] = rhs_lap[ilev];
  }

  amrSolver.solveAMR(phi, rhs);  



  

  
  //const Real k1 = (DxVector[0]*DxVector[0])/12.0;
  const Real k1 = -(DxVector[0]*DxVector[0])/8.0;
  cout << " k1 = " << k1 << endl;
  
  // create tempLDF and set it equal to RHS
  Vector<LDF*> tempLDF(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    tempLDF[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
    tempLDF[ilev] = rhs[ilev];
  }
  //prettyPrintLDF(tempLDF[0], " tempLDF");

  // compute tempLDF = phi + k1*RHS
  // can't do this easily -- must go thru each box of each level.
  for(int i=0; i<tempLDF.size(); ++i) {
    LDF& leveltempLDF = *(tempLDF[i]);
    for(DataIterator dit = leveltempLDF.dataIterator(); dit.ok(); ++dit) {
      leveltempLDF[dit()] *= k1;
      leveltempLDF[dit()] += (*phi[i])[dit()];
    }
  }
  //prettyPrintLDF(tempLDF[0], " tempLDF");




  // make a coarser grid
  Vector<DBL> DBLVector_coarse(numLevels);
  Vector<Box> boundingBoxVector_coarse(numLevels);  
  Vector<Real> DxVector_coarse(numLevels);
  Vector<LDF*> phi_coarse(numLevels, NULL);

  for (int ilev=0; ilev<numLevels; ilev++) {
    coarsen(DBLVector_coarse[ilev], DBLVector[ilev], 2);
    boundingBoxVector_coarse[ilev] = coarsen(boundingBoxVector[ilev], 2);
    DxVector_coarse[ilev] = 2.0*DxVector[ilev];
    phi_coarse[ilev] = new LDF(DBLVector_coarse[ilev], 1, unitVector);
  }

  // average soln to coarser levels
  for (int lev=numLevels-1; lev>=0; lev--) {
    //printf(" coarsening level %d\n", lev);
    //cout << DBLVector[lev] << endl;
    CoarseAverage avgDown(DBLVector[lev], 1, 2);
    avgDown.averageToCoarse(*phi_coarse[lev], *tempLDF[lev]);
  }
  //prettyPrintLDF(phi_coarse[0], " phi coarse");

  outputHDF5(rhs, DBLVector, boundingBoxVector,
	     referenceRatioVector, numLevels, "rhs.hdf5");

  outputHDF5(phi, DBLVector, boundingBoxVector,
	     referenceRatioVector, numLevels, "phi.hdf5");
  
  outputHDF5(phi_coarse, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "phi_coarse.hdf5");
    

  // so phi_coarse is our 'answer' so far.  let's test it.










  Vector<LDF*> rhs_coarse(numLevels, NULL);
  rhs_coarse[0] = new LDF(DBLVector_coarse[0], 1, unitVector);

  LDF& levelrhs_coarse = *rhs_coarse[0];
  Real dx = 2.0*DxVector[0];
  cout << " dx on coarse grid = " << dx << endl;
  for (DataIterator dit = levelrhs_coarse.dataIterator(); dit.ok(); ++dit) {
    FORT_SETRHS(CHF_FRA(levelrhs_coarse[dit()]), 
		CHF_BOX(levelrhs_coarse[dit()].box()),
		CHF_CONST_REAL(dx));    
  }
  //prettyPrintLDF(rhs_coarse[0], "rhs_coarse");

  Vector<LDF*> diffabs(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    diffabs[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
  }
  compareTwoLDFs(rhs_coarse, phi_coarse, diffabs);
  
  outputHDF5(diffabs, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "diffabs1.hdf5");

  // make new operator and solver on coarse level
  PoissonOp poissonOperator_coarse;
  poissonOperator_coarse.setDomainGhostBC(domghostbc);
  AMRSolver amrSolver_coarse(DBLVector_coarse,
			     boundingBoxVector_coarse,
			     DxVector_coarse,   
			     referenceRatioVector, 
			     numLevels, 
			     lbase, 
			     &poissonOperator_coarse);

  amrSolver_coarse.setVerbose(verbose);
  amrSolver_coarse.setMaxIter(maxIterations);
  amrSolver_coarse.setNumVCyclesBottom(numberVcyclesBottom);

  Vector<LDF*> lap_phi_coarse(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    lap_phi_coarse[ilev] = new LDF(DBLVector_coarse[ilev], 1, unitVector);
  }

  for (int lev=0; lev<numLevels; lev++) {
    amrSolver_coarse.applyAMROperator(phi_coarse, *lap_phi_coarse[lev], lev);
  }

  outputHDF5(lap_phi_coarse, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "coarse_lap.hdf5");

  //prettyPrintLDF(lap_phi_coarse[0], "lap_phi_coarse");
  //  now have the laplacian of phi_coarse

  //  compare with laplacian of phi averaged onto coarse grid


  // Compute the Laplacian of phi
  Vector<LDF*> phi_lap(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    phi_lap[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
  }
  for (int lev=0; lev<numLevels; lev++) {
    amrSolver.applyAMROperator(phi, *phi_lap[lev], lev);
  }
  //prettyPrintLDF(phi_lap[0], " phi_lap");

  // average Lap of phi to coarser levels
  Vector<LDF*> lap_phi_coarse_avg(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    lap_phi_coarse_avg[ilev] = new LDF(DBLVector_coarse[ilev], 1, unitVector);
  }

  for (int lev=numLevels-1; lev>=0; lev--) {
    //cout << DBLVector[lev] << endl;
    CoarseAverage avgDown(DBLVector[lev], 1, 2);
    avgDown.averageToCoarse(*lap_phi_coarse_avg[lev], *phi_lap[lev]);
  }
  //prettyPrintLDF(lap_phi_coarse_avg[0], " lap_phi_coarse_avg");

  outputHDF5(lap_phi_coarse_avg, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "lap_phi_coarse_avg.hdf5");

  compareTwoLDFs(lap_phi_coarse, lap_phi_coarse_avg, diffabs);

  outputHDF5(diffabs, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "diffabs2.hdf5");


  int numberCoarseFineBoundaryFaces;
  int numberProblemBoundaryFaces;
  for(int i=0; i<lap_phi_coarse.size(); ++i) {
    LDF& lev = *(lap_phi_coarse[i]);
    const DBL dbl = lev.disjointBoxLayout();
    coarseFineBoundaryCount(dbl, 
			    numberCoarseFineBoundaryFaces,
			    numberProblemBoundaryFaces);
    cout << " numberCoarseFineBoundaryFaces " 
	 << numberCoarseFineBoundaryFaces << endl;
    cout << " numberProblemBoundaryFaces    " 
	 << numberProblemBoundaryFaces << endl;  
  }






  for(int ilev = 0; ilev < numLevels; ilev++) {
    delete phi[ilev];
    delete rhs[ilev];
  }


  } // end parallel scope
#ifdef MPI
  //dumpmemoryatexit();
  MPI_Finalize();
#endif

  return(0);
}


void compareTwoLDFs(const Vector<LDF*>& grid1, const Vector<LDF*>& grid2,
		    Vector<LDF*>& absDifference) {

  //printf(" grid1 size = %d   grid2 size = %d\n", grid1.size(), grid2.size());
  assert(grid1.size() == grid2.size());
  assert(grid1.size() == absDifference.size());
  const int numLevels = grid1.size();

  // Subtract grid2 from grid1 and store into diff
  Vector<LDF*> diff(numLevels, NULL);
  for(int i=0; i<numLevels; ++i) {
    diff[i] = grid1[i];
  }
  for(int i=0; i<numLevels; ++i) {
    for(DataIterator dit = (*diff[i]).dataIterator(); dit.ok(); ++dit) {
      (*diff[i])[dit()] -= (*grid2[i])[dit()];
    }
  }
  //prettyPrintLDF(diff[0], " zero?");


  // now find the largest absolute value in matrix
  //Vector<LDF*> diffabs(numLevels, NULL);
  for(int ilev=0; ilev<diff.size(); ++ilev) {
    absDifference[ilev] = diff[ilev];
  }

  LDF& levDiff = *(absDifference[0]);
  const DBL dbl = levDiff.disjointBoxLayout();
  Real maxError = 0;
  IntVect indexOfMaxError;

  for(DataIterator dit = levDiff.dataIterator(); dit.ok(); ++dit) {
    levDiff[dit()].abs();

    //printf(" max=%16.7e at (%4d,%4d)  min=%16.7e at (%4d,%4d)  norm=%16.7e n=%ld\n",
    //   levDiff[dit()].max(dbl[dit()]),
    //   (levDiff[dit()].maxIndex(dbl[dit()]))[0],
    //   (levDiff[dit()].maxIndex(dbl[dit()]))[1],
    //   levDiff[dit()].min(dbl[dit()]),
    //   (levDiff[dit()].minIndex(dbl[dit()]))[0],
    //   (levDiff[dit()].minIndex(dbl[dit()]))[1],
    //   levDiff[dit()].norm(dbl[dit()]),
    //   dbl[dit()].numPts());

    if (levDiff[dit()].max(dbl[dit()]) > maxError) {
      maxError = levDiff[dit()].max(dbl[dit()]);
      indexOfMaxError =  levDiff[dit()].maxIndex(dbl[dit()]);
    }

  }

  //printf(" size=%10.0f  h=%16.7e  max_error=%16.7e  at (%d, %d)\n",
  // 1.0/DxVector[0], DxVector[0], maxError,
  // indexOfMaxError[0], indexOfMaxError[1]);
  printf("  max_error=%16.7e  at (%d, %d)\n",
	   maxError,
	 indexOfMaxError[0], indexOfMaxError[1]);

}




void coarseFineBoundaryCount(const DisjointBoxLayout& dbl,
			     int& numberCoarseFineBoundaryFaces,
			     int& numberProblemBoundaryFaces) {

  LayoutIterator it0 = dbl.layoutIterator();
  Box bound = dbl.get(it0()); 
  ++it0;
  for(it0.begin(); it0.ok(); ++it0) {
    bound = minBox(bound, dbl.get(it0()));
  }
  cout << " Bounding Box of dbl = " << bound << endl;


  const int SX = bound.smallEnd()[0];
  const int SY = bound.smallEnd()[1];
  const int BX = bound.bigEnd()[0];
  const int BY = bound.bigEnd()[1];
  //cout << " BOUNDING: smallx = " << SX
  //   << " smally = " << SY
  //   << " bigx = " << BX
  //   << " bigy = " << BY
  //   << endl;

#if (CH_SPACEDIM == 3)
  const int SZ = bound.smallEnd()[2];
  const int BZ = bound.bigEnd()[2];
  //cout << " BOUNDING: smallz = " << SZ
  //   << " BIGZ = " << BZ << endl;
#endif

  numberCoarseFineBoundaryFaces=0;
  numberProblemBoundaryFaces=0;
  int faceCount=0;

  IntVectSet skinSet;

  LayoutIterator it1 = dbl.layoutIterator();
  for (it1.begin(); it1.ok(); ++it1)    {             
    //cout << " di()=" << it1().intCode() << endl;

    Box currentBox = dbl[it1()];
    //cout << " current box=" << currentBox << endl;    

    const int sx = currentBox.smallEnd()[0];
    const int sy = currentBox.smallEnd()[1];
    const int bx = currentBox.bigEnd()[0];
    const int by = currentBox.bigEnd()[1];
#if (CH_SPACEDIM == 3)
    const int sz = currentBox.smallEnd()[2];
    const int bz = currentBox.bigEnd()[2];
    //cout << " smallx = " << sx
    // << " smally = " << sy
    // << " smallz = " << sz
    // << " bigx = " << bx
    // << " bigy = " << by
    // << " bigz = " << bz
    //<< endl;
#else
    const int sz = 0;
    const int bz = 0;
    //cout << " smallx = " << sx
    // << " smally = " << sy
    // << " bigx = " << bx
    // << " bigy = " << by
    // << endl;
#endif
    
    if(sx > SX) { // add left side

      // first make a single layer slab on the side
      IntVect siv(D_DECL(sx-1, sy, sz));
      IntVect biv(D_DECL(sx-1, by, bz));
      Box b(siv, biv);

      // count the number of cells -- and since it's
      // only one layer deep, that's also the number of 
      // outward facing faces.
      faceCount += b.numPts();

      // Then add this box to the IntVectSet collection.
      // Per IntVectSet rules, identical cells will not be 
      // added twice which is what we want here.
      skinSet |= b;
      //cout << " adding left side = " << b << endl;
    } else {
      // else we are on the physical boundary, so add up
      // the faces -- which is the area.  (again one layer)
      numberProblemBoundaryFaces += (by-sy+1)*(bz-sz+1);
    }

    if(bx < BX) { // add right side
      IntVect siv(D_DECL(bx+1, sy, sz));
      IntVect biv(D_DECL(bx+1, by, bz));
      Box b(siv, biv);
      faceCount += b.numPts();
      skinSet |= b;
      //cout << " adding right side = " << b << endl;
    } else {
      numberProblemBoundaryFaces += (by-sy+1)*(bz-sz+1);
    }

    if(sy > SY) { // add bottom side
      IntVect siv(D_DECL(sx, sy-1, sz));
      IntVect biv(D_DECL(bx, sy-1, bz));
      Box b(siv, biv);
      faceCount += b.numPts();
      skinSet |= b;
      //cout << " adding bot side = " << b << endl;
    } else {
      numberProblemBoundaryFaces += (bx-sx+1)*(bz-sz+1);
    }

    if(by < BY) { // add top side
      IntVect siv(D_DECL(sx, by+1, sz));
      IntVect biv(D_DECL(bx, by+1, bz));
      Box b(siv, biv);
      faceCount += b.numPts();
      skinSet |= b;
      //cout << " adding top side = " << b << endl;
    } else {
      numberProblemBoundaryFaces += (bx-sx+1)*(bz-sz+1);
    }

#if (CH_SPACEDIM == 3)
    if(sz > SZ) { // add front side
      IntVect siv(D_DECL(sx, sy, sz-1));
      IntVect biv(D_DECL(bx, by, sz-1));
      Box b(siv, biv);
      faceCount += b.numPts();
      skinSet |= b;
      //cout << " adding front side = " << b << endl;
    } else {
      numberProblemBoundaryFaces += (bx-sx+1)*(by-sy+1);
    }

    if(bz < BZ) { // add back side
      IntVect siv(D_DECL(sx, sy, bz+1));
      IntVect biv(D_DECL(bx, by, bz+1));
      Box b(siv, biv);
      faceCount += b.numPts();
      skinSet |= b;
      //cout << " adding back side = " << b << endl;
    } else {
      numberProblemBoundaryFaces += (bx-sx+1)*(by-sy+1);
    }
#endif

  }

  const int totalSkinSetCount = skinSet.numPts();

  // We now have an IntVectSet of all the cells making up
  //  a one-level skin covering all boxes within the problem
  //  (minus any skin going outside of the problem domain)
  //  Some of this skin includes cells of already defined boxes.
  // So now we can go thru each box in the DBL, 
  //   and subtract those
  //   cells from our skinSet to obtain only the cells
  //   on a coarse-fine boundary.
  for (it1.begin(); it1.ok(); ++it1)    {             
    Box b = dbl[it1()];
    skinSet -= b;
    //cout << "  subtracting = " << b << endl;
  }
      
  const int numberSharedFaces = totalSkinSetCount - skinSet.numPts();
   
  cout << " totalSkinSetCount " << totalSkinSetCount
       << " numberSharedFaces " << numberSharedFaces <<endl;
 
  numberCoarseFineBoundaryFaces = faceCount - numberSharedFaces;

  // And if you want a set of all the cells comprising the
  //  coarse-fine boundary -- you have it
  //for(IVSIterator it(skinSet); it.ok(); ++it) {
  //cout << " skinSet = " << it()  << endl;
  //}
  //cout << " size = " << skinSet.numPts() << endl;

  //cout << " numberCoarseFineBoundaryFaces " << numberCoarseFineBoundaryFaces << endl;
  //cout << " numberProblemBoundaryFaces    " << numberProblemBoundaryFaces << endl;
}


