#include "localfunctions.H"
#include "parstream.H"
#include "fortranfunctions.H"

#include "BoxIterator.H"
#include "LayoutIterator.H"
#include "BRMeshRefine.H"
#include "FineInterp.H"


Box createBox(const int *Ncells, const int dimensions) {
  
  IntVect ivlo = IntVect::TheZeroVector();
  IntVect ivhi;
  for(int idir = 0; idir < dimensions; idir++) {
    ivhi[idir] = Ncells[idir] - 1;
  }    
  return Box(ivlo, ivhi);    
}


//  Init grid hierarchy from input file
int InitializeGrids(Vector<LDFB*>& rhs,
                    Vector<DisjointBoxLayout>&  DBLVector,
                    Vector<Box>&                boundingBoxVector, 
                    Vector<Real>&               DxVector, 
                    const Vector<int>&          referenceRatioVector,
                    const int                   *Ncells,
                    const Real                  *ProblemLowCord,
                    const Real                  *ProblemHighCord,
                    const int                   maxBoxSize,
                    const int                   blockFactor,
                    const int                   bufferSize,
                    const Real                  fillRatio,
                    const Real                  refineThreshold,
                    const int                   maxLevel,
                    int&                        numLevels, 
                    const bool                  verbose) {
  
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif

  for(int i=0; i<SpaceDim; i++) {
    printf(" direction%d Ncells=%d LowCord=%f HighCord=%f\n",
           i, Ncells[i], ProblemLowCord[i], ProblemHighCord[i]);
  }
  printf(" maxBoxSize=%d  blockFactor=%d  bufferSize=%d  maxLevel=%d\n",
         maxBoxSize, blockFactor, bufferSize, maxLevel);
  printf(" fillRatio=%f refineThreshold=%f\n",
         fillRatio, refineThreshold);
  
  // this is level 0 
  numLevels = 0;

  // make one big box with Ncells dimensions
  Box baseDomain = createBox(Ncells, SpaceDim);

  DBLVector.resize(maxLevel+1);
  boundingBoxVector.resize(maxLevel+1);
  DxVector.resize(maxLevel+1);
  rhs.resize(maxLevel+1, NULL);

  DxVector[0] = (ProblemHighCord[0]- ProblemLowCord[0])/Ncells[0];
  for(int ilev = 1; ilev <= maxLevel; ilev++) {
    assert(referenceRatioVector[ilev-1] > 0);
    DxVector[ilev] = DxVector[ilev-1]/referenceRatioVector[ilev-1];
  }
  
  boundingBoxVector[0] = baseDomain;
  cout << " base domain -- boundingBoxVector[0]= " << baseDomain << endl;
  
  for(int ilev = 1;ilev <= maxLevel; ilev++) {
    boundingBoxVector[ilev] = 
      refine(boundingBoxVector[ilev-1], referenceRatioVector[ilev-1]);
    cout << " boundingBoxVector[" << ilev << "]= " << boundingBoxVector[ilev] << endl;
  }

  Vector<Vector<Box> > amrBoxes(maxLevel+1);
  Vector<Vector<Box> > oldAmrBoxes(maxLevel+1);
  Vector<Vector<int> > procIDs(maxLevel+1);
  
  // define base level first
  domainSplit(boundingBoxVector[0], oldAmrBoxes[0], maxBoxSize, blockFactor);

  procIDs[0].resize(oldAmrBoxes[0].size());
  LoadBalance(procIDs[0], oldAmrBoxes[0]);
  
  //cout << " boundingBoxVector[0]= "  << boundingBoxVector[0] << endl;
  //cout << " oldAmrBoxes[0]= " << oldAmrBoxes[0] << endl;
  cout << " oldAmrBoxes[0].size() = " << oldAmrBoxes[0].size() << endl;
  cout << " procIDs[0]= "     << procIDs[0] << endl;
  cout << " procIDs[0].size() = "     << procIDs[0].size() << endl;
  cout << endl;
  
  DBLVector[0].define(oldAmrBoxes[0],procIDs[0]);
  rhs[0] = new LDFB(DBLVector[0], 1, IntVect::TheUnitVector());
  
  // now initialize data on this level
  LDFB& levelRhs = *rhs[0];
  DataIterator dit = levelRhs.dataIterator();
  for (dit.reset(); dit.ok(); ++dit)    {
    FORT_SETRHS(CHF_FRA(levelRhs[dit()]),
                CHF_BOX(levelRhs[dit()].box()),
                CHF_CONST_REAL(DxVector[0]));
  }
  
  numLevels = 1;  
  bool moreLevels = (maxLevel > 0);
  
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif

  BRMeshRefine meshRefine(boundingBoxVector[0], referenceRatioVector,
                          fillRatio, blockFactor, bufferSize, 
                          maxBoxSize);

  while (moreLevels) {
    
    // default is moreLevels = false
    // (only repeat loop in the case where a new level
    // is generated which is still less than maxLevel)
    moreLevels = false;
    
    int baseLevel = 0;
    int topLevel = numLevels -1;
    
    printf(" calling tagCells. numLevels=%d topLevel=%d maxLevel=%d \n", 
           numLevels, topLevel, maxLevel);      
    
    Vector<IntVectSet> tagVect(maxLevel+1);
    tagCells(tagVect, // only output
           rhs, boundingBoxVector, refineThreshold,
           bufferSize, baseLevel, numLevels);
    
    cout << " tagVect.size()= " << tagVect.size() << endl;
    cout << " tagVect[topLevel].isEmpty() = " 
         << tagVect[topLevel].isEmpty() << endl;

    int newFinestLevel;
    if(procID() == uniqueProc(SerialTask::compute) ) {
      if(topLevel >= 0 && !tagVect[topLevel].isEmpty())  {
        cout << " meshRefine... " << endl;

        newFinestLevel = meshRefine.regrid(amrBoxes, // output
                                           tagVect,  // input&output (not used again?)
                                           baseLevel, topLevel, 
                                           oldAmrBoxes);
        
        topLevel++;
        
      } // end if there was a new level generated
    } // end if proc is serial node.
    
    broadcast(topLevel, uniqueProc(SerialTask::compute) );
    broadcast(amrBoxes, uniqueProc(SerialTask::compute) );

    // do this only if a new level was generated
    oldAmrBoxes = amrBoxes;    
    
    printf(" topLevel=%d maxLevel=%d numLevels=%d\n", 
           topLevel, maxLevel, numLevels);
    
    // no need to do this for the base level
    for (int lev=1; lev<=topLevel; lev++) {

      printf("  lev=%d  LoadBalance-ing ...\n", lev);

      // do load balancing
      procIDs[lev].resize(amrBoxes[lev].size());
      LoadBalance(procIDs[lev], amrBoxes[lev]);
      
#ifdef MPI      
      cout << " procIDs[lev]=" << procIDs[lev] << endl;
#endif
      cout << " amrBoxes[lev].size()=" << amrBoxes[lev].size() << endl;
      
      const DisjointBoxLayout newDBL(amrBoxes[lev], procIDs[lev]);
      DBLVector[lev] = newDBL;
      if (rhs[lev] != NULL) delete rhs[lev];
      rhs[lev] = new LDFB(DBLVector[lev], 1, IntVect::TheUnitVector());
      
      // now initialize data on this level
      LDFB& levelRhs = *rhs[lev];
      DataIterator dit = levelRhs.dataIterator();
      for (dit.reset(); dit.ok(); ++dit) {      
        FORT_SETRHS(CHF_FRA(levelRhs[dit()]),
                    CHF_BOX(levelRhs[dit()].box()),
                    CHF_CONST_REAL(DxVector[lev]));       
      }
      
    } // end loop over levels for initialization
    
    int old_numLevels = numLevels;
    numLevels = topLevel+1;
    printf(" old_numLevels=%d  numLevels=%d topLevel=%d\n", 
           old_numLevels, numLevels, topLevel);
    if ((numLevels <maxLevel+1) && (numLevels > old_numLevels) ) {
      moreLevels = true;
    }
    
    
  } // end while more levels loop
  
  return 0;
}
















// if u wanna start from a restart file -- here ya go

//   } else {  // restarting from file:

//     Vector<string> vectnames;
//     Vector<LDFB* > rhsTemp;
//     Real dt;
//     // read in everything from restart file
//     ReadAMRHierarchyHDF5(restart_file,
//                       DBLVector,
//                       rhsTemp,
//                       vectnames,
//                       boundingBoxVector[0],
//                       DxVector[0],
//                       dt,
//                       time,
//                       referenceRatioVector,
//                       numLevels);


//     // now allocate rhs for each level and copy temp data into it
//     for (int lev = 0; lev<numLevels; ++lev) {
//       assert (rhsTemp[lev] != NULL);
//       const DisjointBoxLayout levelDBL = DBLVector[lev];
//       rhs[lev] = new LDFB(levelDBL,1,
//                                             IntVect::TheUnitVector());
      
//       Interval rhsComp(0,0);
//       Interval destComps(0,0);
//       rhsTemp[lev]->copyTo(rhsComp,*rhs[lev],destComps);
      
//       // delete temporary storage
//       delete rhsTemp[lev];
//       rhsTemp[lev] = NULL;

//     }


//   }  // end if restart
    


//BaseFab dump
//  void dump() {
//     cout << " BaseFab box= " << domain << endl;
//     printf(" Number components=%d \n", nvar);
//     printf(" Cached number of points in FAB=%d \n", numpts);
//     printf(" nvar*numpts that was allocated on heap=%d \n", truesize);
//     for(int i=0; i<nvar; i++ ) {
//       for(int j=0; j<numpts; j++ ) {
//      printf("  comp%4d dptr[%4d]=%20.4e\n", i, j, dptr[i*numpts+j]);
//      //cout << "   comp" << i << " j=" << j << " " << dptr[i*numpts+j] << endl;
//       }
//     }
//   }


