#include "localfunctions.H"
#include "parstream.H"
#include "fortranfunctions.H"

#include "BoxIterator.H"
#include "LayoutIterator.H"
#include "BRMeshRefine.H"
#include "FineInterp.H"


int regrid(Vector<LevelData<FArrayBox>*>& vectData,
           Vector<DisjointBoxLayout>&     vectGrids,
           const Vector<Box>&             vectDomain,
           const Vector<int>&             vectRefRatio,
           const int  blockFactor,
           const int  maxBoxSize,
           const int  bufferSize,
           const Real fillRatio,
           const Real refineThreshold,
           const int  maxLevel,    
           const int  base_level,
           int& numlevels) {

  //#ifdef MPI
  //MPI_Barrier(Chombo_MPI::comm);
  //#endif
  
  int top_level = numlevels-1;

  Vector<Vector<Box> > old_grids(top_level+1);
  Vector<Vector<Box> > new_grids;
  Vector<IntVectSet> tags(top_level+1);

  tagCells(tags, // only output of function
           vectData, 
	   vectDomain,
           refineThreshold, 
	   bufferSize,
	   base_level,
           numlevels);
    
  
    
  for (int level = base_level; level <= top_level; ++level)   {  
    // this is clunky, but i can't think of another way
    // to do this at the moment
    // note that we need a LayoutIterator here instead of a 
    // DataIterator, because we want _all_ the boxes, not just
    // the boxes on this processor
    const DisjointBoxLayout& oldDBL = vectGrids[level];
    LayoutIterator lit = oldDBL.layoutIterator();

    //cout << endl;
    //cout << endl;
    //oldDBL.print();

    
    for (lit.reset(); lit.ok(); ++lit) {
      old_grids[level].push_back(oldDBL[lit()]);
    }
  }
  
  int new_finest_level;
  
  if(procID() == uniqueProc(SerialTask::compute) )  {
    

    BRMeshRefine meshrefine(vectDomain[0], vectRefRatio,
                            fillRatio, blockFactor, 
                            bufferSize, maxBoxSize);

    new_finest_level = meshrefine.regrid(new_grids, 
                                         tags,
                                         base_level,
                                         top_level,
                                         old_grids);

    
  }
  
  broadcast(new_finest_level, uniqueProc(SerialTask::compute) );
  broadcast(new_grids,        uniqueProc(SerialTask::compute) );
  
  // can only change the new finest level if we're regridding
  // the entire hierarchy.  in this code, that's not such a big
  // deal, but is important in the case of subcycling...
  if(base_level == 0)    {
    numlevels = Min(new_finest_level, maxLevel)+1;
  }
  
  int nComp = vectData[0]->nComp();
  const IntVect& ghostVect = vectData[0]->ghostVect();
  
  
  // now need to redefine each level
  for (int level = base_level+1; level <numlevels; ++level) {
    int numGridsNew = new_grids[level].size();
    Vector<int> procIDs(numGridsNew);
    LoadBalance(procIDs,new_grids[level]);
    const DisjointBoxLayout newDBL(new_grids[level], procIDs);
    vectGrids[level] = newDBL;
    
    LevelData<FArrayBox>* oldDataPtr = vectData[level];
    LevelData<FArrayBox>* newDataPtr = new LevelData<FArrayBox>(newDBL, 
                                                                nComp,
                                                                ghostVect);
    
    // first fill with interpolated data from coarser level
    FineInterp interpolator(newDBL,nComp,vectRefRatio[level-1],
                            vectDomain[level]);
    
    interpolator.interpToFine(*newDataPtr, *vectData[level-1]);
    
    // now copy old data on this level to new data holder
    if (oldDataPtr != NULL) {
      Interval dataComps = newDataPtr->interval();
      oldDataPtr->copyTo(dataComps, *newDataPtr, dataComps);
      delete oldDataPtr;
    }
    
    // now replace old data holder with new one, and delete old data holder
    vectData[level] = newDataPtr;
    
    
  }  // end loop over currently defined levels
  
  // now ensure that any remaining levels are null pointers
  for (int lev=numlevels; lev <= maxLevel; ++lev) {
    if (vectData[lev] != NULL) {
      delete vectData[lev];
      vectData[lev] = NULL;
    }
    
    DisjointBoxLayout emptyDBL;
    vectGrids[lev] = emptyDBL;
  }

  return 1;
  
}
