#include  <iostream>
#include "LevelOp.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "PoissonBC.H"
#include "ParmParse.H"
#include "AMRSolver.H"
#include "HelmholtzOp.H"
#include "Vector.H"
#include "AMRIO.H"
#include "SPMD.H"
#include "MeshRefine.H"
#include "LoadBalance.H"
#include "localFuncs.H"
#include "tdglF_F.H"
// #include "ArrayView.H"

int main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  //scoping trick
  {
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  if (argc < 2) 
    {
      cerr<< " usage " << argv[0] << " <input_file_name> " << endl;
      exit(0);
    }
  char* in_file = argv[1];
  ParmParse pp(argc-2,argv+2,NULL,in_file);

  int lbase;

  int iverbose;
  ParmParse pp2("main");
  pp2.get("lbase", lbase);
  pp2.get("verbose", iverbose);
  bool verbose = (iverbose == 1);



  DomainGhostBC solverBC;
  SideIterator sit;
  for (int dir=0; dir<SpaceDim; dir++) {
    for (sit.begin(); sit.ok(); ++sit) {
      printf(" Setting BC dir=%d sit=%d\n", dir, sit);
      DircheletBC thisBC(dir, sit());
      solverBC.setBoxGhostBC(thisBC);
    }
  }
  

  Real time = 0;
  int numlevels, maxLevel;
  Vector<DisjointBoxLayout> vectGrids;
  Vector<Box> vectDomain;
  Vector<int> vectRefRatio;
  Vector<Real> vectDx;
  Vector<LevelData<FArrayBox>* > phi;
  int eekflag = setGrids(phi, vectGrids, vectDomain, vectDx, 
                         vectRefRatio, time, maxLevel, numlevels, verbose);
  if(eekflag != 0) {
    cerr << "error: setGrids returned error code " << eekflag << endl;
    exit(2);
  }

  // allocate phihalf and rhs
  Vector<LevelData<FArrayBox>* > rhs(maxLevel+1, NULL);
  Vector<LevelData<FArrayBox>* > phihalfVect(maxLevel+1, NULL);
  Vector<LevelData<FArrayBox>* > lapPhiVect(maxLevel+1, NULL);

  for(int ilev = 0; ilev < numlevels; ilev++)  {
    rhs[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], 1, 
                                         IntVect::TheZeroVector());    
    phihalfVect[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], 1, 
                                                 IntVect::TheZeroVector());
    lapPhiVect[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], 1, 
                                                IntVect::TheZeroVector());
  }


  // will fix this later -- should be read in from inputs file
  Real maxTime = 0.25;
  int step = 0;
  int maxSteps = 10000;
  int plotInterval = 1;
  int regridInterval = 2;

  // output intial conditions
  outputData(phi, phihalfVect, lapPhiVect, rhs, vectGrids, vectDomain, 
             vectRefRatio, vectDx, step, time, numlevels, verbose);

  cout <<"here"<<endl;

  while ((time < maxTime) && (step < maxSteps)) {
    if ((step!=0) && (step%regridInterval == 0) ) {

        // base level for regridding is 0
        regrid(phi, vectGrids, vectDomain, vectDx, vectRefRatio,
               maxLevel, numlevels, 0);

        // may need to redefine rhs, other data
        for (int lev=1; lev<numlevels; lev++) {
          if (phihalfVect[lev] != NULL) { 
            delete phihalfVect[lev];
          }
          if (lapPhiVect[lev] != NULL) { 
            delete lapPhiVect[lev];
          }
          if (rhs[lev] != NULL) {
            delete rhs[lev];
          }

          // now allocate new level's worth of data;
          rhs[lev] = new LevelData<FArrayBox>(vectGrids[lev], 1, 
                                              IntVect::TheZeroVector());
          
          phihalfVect[lev] = 
            new LevelData<FArrayBox>(vectGrids[lev], 1, 
                                     IntVect::TheZeroVector());
          lapPhiVect[lev] = 
            new LevelData<FArrayBox>(vectGrids[lev], 1, 
                                     IntVect::TheZeroVector());
          
        } // end loop over current levels
        
        // make sure that any levels finer than current finest level
        // are deleted
        for (int lev=numlevels; lev<maxLevel+1; lev++) {
          if (phihalfVect[lev] != NULL) { 
            delete phihalfVect[lev];
            phihalfVect[lev] = NULL;
          }
          if (lapPhiVect[lev] != NULL) { 
            delete lapPhiVect[lev];
            lapPhiVect[lev] = NULL;
          }
          if (rhs[lev] != NULL) {
            delete rhs[lev];
            rhs[lev] = NULL;
          }     
        } // end loop over levels finer than current finest
        
    } // end if it's time to regrid
    
    
    Real dt = computeDt(phi, vectDx, vectRefRatio, vectDomain, numlevels);
    if ((time + dt) > maxTime) dt = maxTime-time;
    
    cout << "Step # " << step << ", time = " << time << ", dt = " 
         << dt << endl;
    
    HelmholtzOp levelop;
    
    // CNcoeff = 1 for backwards euler, 0.5 for CN
    Real CNcoeff = 1.0;
    Real helmcoeff = -CNcoeff*dt;
    levelop.setHelmCoeff(helmcoeff);
    levelop.setDomainGhostBC(solverBC);

    AMRSolver bigSolver(vectGrids, vectDomain, vectDx, 
                        vectRefRatio, numlevels, 0, &levelop);

    bigSolver.setVerbose(false);


    // this one is for the RHS computation
    PoissonOp poissOp;
    poissOp.setDomainGhostBC(solverBC);
    AMRSolver poissonSolver(vectGrids, vectDomain, vectDx, 
                            vectRefRatio, numlevels, 0, &poissOp);

    // compute RHS
    for (int lev=0; lev<numlevels; lev++) {
      LevelData<FArrayBox>& levelPhi = *phi[lev];
      LevelData<FArrayBox>& levelRHS = *rhs[lev];
      const DisjointBoxLayout& levelGrids = vectGrids[lev];

      // temporary storage for this level
      LevelData<FArrayBox>& LofPhi = *lapPhiVect[lev];

      poissonSolver.applyAMROperator(phi, LofPhi, lev);


      DataIterator levelDit = levelPhi.dataIterator();
      for (levelDit.reset(); levelDit.ok(); ++levelDit) {

        const Box dataBox = levelPhi[levelDit()].box();
        FArrayBox& phihalf = (*phihalfVect[lev])[levelDit()];
        const Box gridBox = levelGrids.get(levelDit());

        FORT_PREDICTPHI(CHF_FRA(phihalf),
                        CHF_CONST_FRA(levelPhi[levelDit()]),
                        CHF_CONST_FRA(LofPhi[levelDit()]),
                        CHF_BOX(gridBox),
                        CHF_CONST_REAL(vectDx[lev]),
                        CHF_CONST_REAL(dt));

        //ArrayView(&(levelPhi[levelDit()]));
        //ArrayView(&phihalf);

        FORT_SETUPRHS(CHF_FRA(levelRHS[levelDit()]),
                      CHF_CONST_FRA(levelPhi[levelDit()]),
                      CHF_CONST_FRA(phihalf),
                      CHF_CONST_FRA(LofPhi[levelDit()]),
                      CHF_BOX(gridBox),
                      CHF_CONST_REAL(vectDx[lev]),
                      CHF_CONST_REAL(dt));
      } // end loop over grids on this level

    } // end loop over levels


    // do implicit solve
    bigSolver.solveAMR(phi,rhs);

    // average soln to coarser levels
    for (int lev=numlevels-1; lev>0; lev--) {
      CoarseAverage avgDown(vectGrids[lev],1,vectRefRatio[lev-1]);
      avgDown.averageToCoarse(*phi[lev-1], *phi[lev]);
    }

    time = time + dt;
    ++step;

    // output data
    if (step%plotInterval == 0) {
      outputData(phi, phihalfVect, lapPhiVect, rhs, vectGrids, 
                 vectDomain, vectRefRatio, vectDx,
                 step, time, numlevels, verbose);
    }
  
  } // end loop over time                    
    
  
  }//end scoping trick
#ifdef MPI
  MPI_Finalize();
#endif

  return(0);
}
