#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
#include <cmath>
#include <cstdio>
#include <string>

#include "localFuncs.H"
#include "parstream.H"
#include "BoxIterator.H"
#include "LayoutIterator.H"
#include "MeshRefineObject.H"
#include "FineInterp.H"

int outputData(const Vector<LevelData<FArrayBox>* >& vectPhi,
               const Vector<LevelData<FArrayBox>* >& vectPhiHalf,
               const Vector<LevelData<FArrayBox>* >& vectLapPhi,
               const Vector<LevelData<FArrayBox>* >& vectRhs,
               const Vector<DisjointBoxLayout>& vectGrids,
               const Vector<Box>& vectDomain,
               const Vector<int>& vectRatio,
               const Vector<Real>& vectDx,
               const int nstep,
               const Real time,
               int numlevels,
               bool verbose)
{
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  string phiName("phi");
  string phihalfName("phihalf");
  string lapphiname("lapPhi");
  string rhsName("rhs");
  Vector<string> vectName(4);
  vectName[0] = phiName;
  vectName[1] = phihalfName;
  vectName[2] = lapphiname;
  vectName[3] = rhsName;
  Box domain = vectDomain[0];
  Real dx = vectDx[0];
  Real dt = 1.;
  Vector<LevelData<FArrayBox>* > vectPhiAndRHS(numlevels, NULL);
  for(int ilev = 0; ilev < numlevels; ilev++)
    {
      vectPhiAndRHS[ilev] = new LevelData<FArrayBox>(vectGrids[ilev], 4);
      Interval phiInterval(0,0);
      Interval phihalfInterval(1,1);
      Interval lapPhiInterval(2,2);
      Interval rhsInterval(3,3);

      vectPhi[ilev]->copyTo(vectPhi[ilev]->interval(), 
                            *vectPhiAndRHS[ilev],
                            phiInterval);
      vectPhiHalf[ilev]->copyTo(vectPhiHalf[ilev]->interval(), 
                            *vectPhiAndRHS[ilev],
                            phihalfInterval);

      vectLapPhi[ilev]->copyTo(vectLapPhi[ilev]->interval(), 
                               *vectPhiAndRHS[ilev],
                               lapPhiInterval);
      vectRhs[ilev]->copyTo(vectRhs[ilev]->interval(), 
                            *vectPhiAndRHS[ilev],
                            rhsInterval);
    }
#ifdef HDF5
  //string filename("tdglOut.hdf5");

  char iter_str[100];
  sprintf(iter_str, 
          "%s%d.%dd.hdf5", 
          "plt", nstep, SpaceDim );
  string filename(iter_str);

  WriteAMRHierarchyHDF5(filename,                             
                        vectGrids, 
                        vectPhiAndRHS, 
                        vectName,                 
                        domain,                               
                        dx, dt, time,                                         
                        vectRatio,                           
                        numlevels);
#endif

  for(int ilev = 0; ilev < numlevels; ilev++)
    delete vectPhiAndRHS[ilev];

  return 0;
}

/*
  Set grid hierarchy from input file
 */
int setGrids(Vector<LevelData<FArrayBox>*>& vectData,
             Vector<DisjointBoxLayout>& vectGrids,
             Vector<Box>&               vectDomain, 
             Vector<Real>&              vectDx, 
             Vector<int>&               vectRefRatio,
             Real& time,
             int& maxLevel,
             int& numlevels, bool verbose)
{
  
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  Real fillRat = 0.77;
  int blockFactor = 8;
  Real refine_thresh = 0.5;
  Vector<int> ancells(SpaceDim);
  Vector<Real>  prob_loa(SpaceDim);
  Vector<Real>  prob_hia(SpaceDim);
  int maxboxsize = 32;
  int buffersize = 1;
  bool doRestart = false;
  string restart_file;

  if(procID() == 0)
    {
      ParmParse pp("main");
      pp.get("max_level", maxLevel);
      numlevels = 0;
      pp.query("fill_ratio", fillRat);
      pp.query("blockFactor", blockFactor);
      pp.getarr("n_cell", ancells, 0, SpaceDim);
      pp.getarr("ref_ratio", vectRefRatio, 0, maxLevel+1);
      pp.getarr("prob_lo",prob_loa,0,SpaceDim);
      pp.getarr("prob_hi",prob_hia,0,SpaceDim);
      pp.get("maxboxsize",maxboxsize);
      pp.query("refine_thresh", refine_thresh);
      pp.query("tags_grow", buffersize);
      doRestart = pp.contains("restart_file");
      if (doRestart) {
        pp.get("restart_file", restart_file);
      } 
    }

  broadcast(maxLevel, 0);
  broadcast(fillRat, 0);
  broadcast(ancells, 0);
  broadcast(vectRefRatio, 0);
  broadcast(prob_loa, 0);
  broadcast(prob_hia, 0);
  broadcast(maxboxsize, 0);
  numlevels = 0;
  
  IntVect ivlo = IntVect::TheZeroVector();
  IntVect ivhi;
  for(int idir = 0; idir < SpaceDim; idir++)
    ivhi[idir] = ancells[idir] - 1;

  Box basedom = Box(ivlo, ivhi);

  vectGrids.resize(maxLevel+1);
  vectDomain.resize(maxLevel+1);
  vectDx.resize(maxLevel+1);
  vectData.resize(maxLevel+1,NULL);

  vectDx[0] = (prob_hia[0]-prob_loa[0])/ancells[0];
  for(int ilev = 1; ilev <= maxLevel; ilev++)
    {
      assert(vectRefRatio[ilev-1] > 0);
      vectDx[ilev] = vectDx[ilev-1]/vectRefRatio[ilev-1];
    }

  vectDomain[0] = basedom;
  cout << " vectDomain[0]= " << basedom << endl;

  for(int ilev = 1;ilev <= maxLevel; ilev++) {
    vectDomain[ilev] = refine(vectDomain[ilev-1],vectRefRatio[ilev-1]);
    cout << " vectDomain[" << ilev << "]= " << vectDomain[ilev] << endl;
  }
  

  if (!doRestart) {

    Vector<Vector<Box> > amrBoxes(maxLevel+1);
    Vector<Vector<Box> > oldAmrBoxes(maxLevel+1);
    Vector<Vector<int> > procIDs(maxLevel+1);

    // define base level first
    domainSplit(vectDomain[0],oldAmrBoxes[0],maxboxsize,blockFactor);
    procIDs[0].resize(oldAmrBoxes[0].size());
    LoadBalance(procIDs[0],oldAmrBoxes[0]);

    vectGrids[0].define(oldAmrBoxes[0],procIDs[0]);
    vectData[0] = new LevelData<FArrayBox>(vectGrids[0],1,
                                           IntVect::TheUnitVector());

    // now initialize data on this level
    LevelData<FArrayBox>& levelPhi = *vectData[0];
    DataIterator dit = levelPhi.dataIterator();
    for (dit.reset(); dit.ok(); ++dit)
      {
        FORT_SETPHI(CHF_FRA(levelPhi[dit()]),
                    CHF_BOX(levelPhi[dit()].box()),
                    CHF_CONST_REAL(vectDx[0]));
      }
  
    numlevels = 1;
    bool moreLevels = (maxLevel > 0);

    while (moreLevels) {
    
      // default is moreLevels = false
      // (only repeat loop in the case where a new level
      // is generated which is still less than maxLevel)
      moreLevels = false;

      int baseLevel = 0;
      int topLevel = numlevels -1;

      Vector<IntVectSet> tagVect(maxLevel+1);
      tagCells(vectData, tagVect, vectDx,vectDomain, refine_thresh,
               buffersize,baseLevel, numlevels);
    
      int eekflag = 0;
      if(procID() == uniqueProc(SerialTask::compute) ) {
        if(topLevel >= 0 && !tagVect[topLevel].isEmpty()) 
          {
            eekflag = meshRefine(amrBoxes, tagVect, baseLevel, topLevel, 
                                 vectDomain, oldAmrBoxes, vectRefRatio, 
                                 fillRat, blockFactor,
                                 buffersize, maxboxsize);
          
            topLevel++;

            if(eekflag != 0)
              {
                cerr << "setGrids: meshRefine returned error code " << eekflag
                     << endl;
                return(1);
              }
          
            broadcast(topLevel, uniqueProc(SerialTask::compute) );
            broadcast(amrBoxes,      uniqueProc(SerialTask::compute) );
          
          } // end if there was a new level generated
      } // end if proc is serial node.
    
      // do this only if a new level was generated
      oldAmrBoxes = amrBoxes;
    
      // no need to do this for the base level
      for (int lev=1; lev<=topLevel; lev++) {
        // do load balancing
        procIDs[lev].resize(amrBoxes[lev].size());
        LoadBalance(procIDs[lev],amrBoxes[lev]);
        const DisjointBoxLayout newDBL(amrBoxes[lev], procIDs[lev]);
        vectGrids[lev] = newDBL;
        if (vectData[lev] != NULL) delete vectData[lev];
        vectData[lev] = new LevelData<FArrayBox>(vectGrids[lev],1,
                                                 IntVect::TheUnitVector());
      
        //now initialize data on this level
        LevelData<FArrayBox>& levelPhi = *vectData[lev];
        DataIterator dit = levelPhi.dataIterator();
        for (dit.reset(); dit.ok(); ++dit)
          {
            FORT_SETPHI(CHF_FRA(levelPhi[dit()]),
                        CHF_BOX(levelPhi[dit()].box()),
                        CHF_CONST_REAL(vectDx[lev]));
          }
      
      } // end loop over levels for initialization
      int old_numlevels = numlevels;
      numlevels = topLevel+1;
      if ((numlevels <maxLevel+1) && (numlevels > old_numlevels) )
        {
          moreLevels = true;
        }
    
      
    } // end while more levels loop
  } else { 
    Vector<string> vectnames;
    Vector<LevelData<FArrayBox>* > vectDataTemp;
    Real dt;
    // read in everything from restart file
    ReadAMRHierarchyHDF5(restart_file,
                         vectGrids,
                         vectDataTemp,
                         vectnames,
                         vectDomain[0],
                         vectDx[0],
                         dt,
                         time,
                         vectRefRatio,
                         numlevels);


    // now allocate vectData for each level and copy temp data into it
    for (int lev = 0; lev<numlevels; ++lev) {
      assert (vectDataTemp[lev] != NULL);
      const DisjointBoxLayout levelDBL = vectGrids[lev];
      vectData[lev] = new LevelData<FArrayBox>(levelDBL,1,
                                               IntVect::TheUnitVector());

      Interval phiComp(0,0);
      Interval destComps(0,0);
      vectDataTemp[lev]->copyTo(phiComp,*vectData[lev],destComps);

      // delete temporary storage
      delete vectDataTemp[lev];
      vectDataTemp[lev] = NULL;

    }


  }  // end if restart
    
  return 0;
}






Real 
computeDt(Vector<LevelData<FArrayBox>*>& a_phi, 
          const Vector<Real>& a_dxVect, 
          const Vector<int>& a_nRefVect,
          const Vector<Box>& a_domainVect,
          const int& a_numLevels)

{

  Real dt = 1000000;
  for (int lev=0; lev<a_numLevels; lev++) {
    Real dtLev = 1000000;
    LevelData<FArrayBox>& thisPhi = *a_phi[lev];
    DataIterator ditLev = thisPhi.dataIterator();
    for (ditLev.reset(); ditLev.ok(); ++ditLev) {
      Real localDt = a_dxVect[lev];
      dtLev = min(dtLev,localDt);
    }
    dt = min(dt, dtLev);
  }

  Real cfl = 0.25;
  dt = dt*cfl;
  return dt;
}

void
tagCells(Vector<LevelData<FArrayBox>*>& vectPhi,
         Vector<IntVectSet>& tagVect,
         Vector<Real>& vectDx,
         Vector<Box>& vectDomain,
         const Real refine_thresh, const int buffer_size,
         const int baseLevel, int numlevels) {
  
  // loop over levels
  for (int lev=baseLevel; lev<numlevels; lev++) {
   
    IntVectSet local_tags;

    LevelData<FArrayBox>& levelPhi = *vectPhi[lev];
    DisjointBoxLayout level_domain = levelPhi.getBoxes();
    DataIterator dit = levelPhi.dataIterator();

    // loop over boxes in level
    for (dit.reset(); dit.ok(); ++dit) {
      
      const Box& b = level_domain.get(dit());
      FArrayBox grad_fab (b, SpaceDim);
      const FArrayBox& state_fab = levelPhi[dit()];

      // loop over directions (dimensions 2,or3) get grad_fab
      for (int dir = 0; dir < SpaceDim; ++dir) {

        const Box bcenter = b & grow ( vectDomain[lev], -BASISV(dir) );
        const Box blo = b & adjCellLo( bcenter, dir );
        const Box bhi = b & adjCellHi( bcenter, dir );
        const int haslo = ! blo.isEmpty();
        const int hashi = ! bhi.isEmpty();
        FORT_GETGRAD ( CHF_BOX(bcenter),
                       CHF_BOX(blo),
                       CHF_BOX(bhi),
                       CHF_CONST_INT(dir),
                       CHF_CONST_INT(haslo),
                       CHF_CONST_INT(hashi),
                       CHF_CONST_FRA1(state_fab, 0),
                       CHF_FRA1(grad_fab,dir) );
      }

      FArrayBox grad_mag_fab (b, 1);
      FORT_MAGNITUDE ( CHF_BOX(b),
                       CHF_FRA1(grad_mag_fab, 0),
                       CHF_CONST_FRA(grad_fab) );

      // tag where gradient exceeds threshold
      BoxIterator bit (b);

      // loop over every cell in box
      for (bit.begin(); bit.ok(); ++bit)
        {
          const IntVect& iv = bit();
          if ( grad_mag_fab(iv) >= refine_thresh )
            {
              local_tags |= iv;
            }
        }
    } // end loop over grids on this level

    local_tags.grow(buffer_size);
    local_tags &= vectDomain[lev];

    int numTags = local_tags.numPts();
    

    Vector<IntVectSet> all_tags;
    const int dest_proc = uniqueProc (SerialTask::compute);

    gather(all_tags, local_tags, dest_proc);

    if (procID() == uniqueProc (SerialTask::compute) )
      {
        for (int i = 0; i < all_tags.size(); ++i)
          {
            tagVect[lev] |= all_tags[i];
          }
      }

     numTags = tagVect[lev].numPts();
  } // end loop over levels

}
      


int regrid(Vector<LevelData<FArrayBox>*>& a_vectData,
           Vector<DisjointBoxLayout>& a_vectGrids,
           Vector<Box>& a_vectDomain,
           Vector<Real>& a_vectDx,
           Vector<int>& a_vectRefRatio,
           int& a_maxLevel,
           int& a_numlevels, int a_base_level)
{

  int numlevels_old = a_numlevels;
  int top_level = a_numlevels-1;

  Vector<Vector<Box> > old_grids(top_level+1);
  Vector<Vector<Box> > new_grids;
  Vector<IntVectSet> tags(top_level+1);

  int blockFactor = 8;
  Real fillRat = 0.77;;
  int maxboxsize;
  Real refine_thresh;
  int buffer_size = 1;
  string restartFile;

  
  // this is pretty clunky.  probably want to reorganize
  // this so that you don't have to access parmparse every
  // regridding step
  // i think i can do this on every processor, not just
  // the serial node, but if this breaks, then do it on
  // the serial node and broadcast the data...
  ParmParse pp("main");
  pp.query("fill_ratio", fillRat);
  pp.query("blockFactor", blockFactor);
  pp.get("maxboxsize",maxboxsize);
  pp.query("refine_thresh", refine_thresh);
  pp.query("tags_grow", buffer_size);

  tagCells(a_vectData, tags, a_vectDx, a_vectDomain,
           refine_thresh, buffer_size, a_base_level,
           a_numlevels);
    
    
  for (int level = a_base_level; level <= top_level; ++level) 
    {  
      // this is clunky, but i can't think of another way
      // to do this at the moment
      // note that we need a LayoutIterator here instead of a 
      // DataIterator, because we want _all_ the boxes, not just
      // the boxes on this processor
      const DisjointBoxLayout& oldDBL = a_vectGrids[level];
      LayoutIterator lit = oldDBL.layoutIterator();
        
      for (lit.reset(); lit.ok(); ++lit) {
        old_grids[level].push_back(oldDBL[lit()]);
      }
    }
    
  int new_finest_level;
    
  if(procID() == uniqueProc(SerialTask::compute) )
    {
        
      MeshRefineObject meshrefine;
      meshrefine.maxSize(maxboxsize);
      meshrefine.blockFactor(blockFactor);
        
      meshrefine.regrid(new_grids, 
                        new_finest_level,
                        tags,
                        a_base_level,
                        top_level,
                        a_vectDomain,
                        old_grids,
                        a_vectRefRatio);
        
    }
    
  broadcast(new_finest_level, uniqueProc(SerialTask::compute) );
    
    // can only change the new finest level if we're regridding
    // the entire hierarchy.  in this code, that's not such a big
    // deal, but is important in the case of subcycling...
  if(a_base_level == 0)
    {
      a_numlevels = Min(new_finest_level, a_maxLevel)+1;
    }
  broadcast(new_grids,      uniqueProc(SerialTask::compute) );
    
  int nComp = a_vectData[0]->nComp();
  const IntVect& ghostVect = a_vectData[0]->ghostVect();
    
    
    // now need to redefine each level
  for (int level = a_base_level+1; level <a_numlevels; ++level)
    {
      int numGridsNew = new_grids[level].size();
      Vector<int> procIDs(numGridsNew);
      LoadBalance(procIDs,new_grids[level]);
      const DisjointBoxLayout newDBL(new_grids[level], procIDs);
      a_vectGrids[level] = newDBL;
        
      LevelData<FArrayBox>* oldDataPtr = a_vectData[level];
      LevelData<FArrayBox>* newDataPtr = new LevelData<FArrayBox>(newDBL, 
                                                                  nComp,
                                                                  ghostVect);
        
      // first fill with interpolated data from coarser level
      FineInterp interpolator(newDBL,nComp,a_vectRefRatio[level-1],
                              a_vectDomain[level]);
        
      interpolator.interpToFine(*newDataPtr, *a_vectData[level-1]);
        
      // now copy old data on this level to new data holder
      if (oldDataPtr != NULL) 
        {
          Interval dataComps = newDataPtr->interval();
          oldDataPtr->copyTo(dataComps, *newDataPtr, dataComps);
          delete oldDataPtr;
        }
        
      // now replace old data holder with new one, and delete old data holder
      a_vectData[level] = newDataPtr;
        
        
    }  // end loop over currently defined levels
    
  // now ensure that any remaining levels are null pointers
  for (int lev=a_numlevels; lev <= a_maxLevel; ++lev) {
    if (a_vectData[lev] != NULL) 
      {
        delete a_vectData[lev];
        a_vectData[lev] = NULL;
      }
      
    DisjointBoxLayout emptyDBL;
    a_vectGrids[lev] = emptyDBL;
  }
    

  // I do believe that's it!

}

