#include  <iostream>
#include "LevelOp.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "PoissonBC.H"
#include "ParmParse.H"
#include "AMRSolver.H"
#include "PoissonOp.H"
#include "Vector.H"
#include "AMRIO.H"
#include "SPMD.H"
#include "IntVectSet.H"
#include "LoadBalance.H"
#include "localfunctions.H"
#include "fortranfunctions.H"
#include "DebugOut.H"

#include "Timer.H"

// Must initialize the static list defined in Timer.hh
//list<Timer*> Timer::TimerList(0); 

Timer Everything("Everything");
Timer InitGrids ("Init Grids",  Everything);
Timer Regrid    ("Regrid    ",  Everything);
Timer AMRSetup  ("AMR Setup ",  Everything);
Timer AMRSolve  ("AMR Solve ",  Everything);
Timer Output    ("Output    ",  Everything);

// Timer applyopI  ("applyopI  ", Everything, DIAGNOSTIC);
// Timer Reflux    ("Reflux    ", Everything, DIAGNOSTIC);

// Timer iFRCoarse ("iFRCoarse ", Everything, DIAGNOSTIC);
// Timer iFRFine   ("iFRFine   ", Everything, DIAGNOSTIC);
// Timer finereflux("finereflux", Everything, DIAGNOSTIC);

// Timer getflux   ("getflux   ", Everything, DIAGNOSTIC);
// Timer inCoarse  ("inCoarse  ", Everything, DIAGNOSTIC);

// Timer gf1("gf1", Everything, DIAGNOSTIC);
// Timer gf2("gf2", Everything, DIAGNOSTIC);
// Timer gf3("gf3", Everything, DIAGNOSTIC);
// Timer gf4("gf4", Everything, DIAGNOSTIC);

// Timer VcycleMG("AMRVcycleMG",  AMRSolve);
// Timer compRes ("AMRcompRes ",  AMRSolve);

// Timer vcAMRres    ("vcAMRres      ",  VcycleMG);
// Timer vcdownsweep ("vcdownsweep   ",  VcycleMG);
// Timer vclevelSolve("vclevelSolve  ",  VcycleMG);
// Timer vcupsweep   ("vcupsweep     ",  VcycleMG);

// Timer dsSmooth  ("dswp Smooth     ", vcdownsweep);
// Timer dsApplyICF("dswp ApplyOpICF ", vcdownsweep);
// Timer dsAvg2Coar("dswp Avg2Coarse ", vcdownsweep);
// Timer dsAMRRes  ("dswp AMRResidual", vcdownsweep);

// Timer relax1("relax1", Everything, DIAGNOSTIC);
// Timer relax2("relax2", Everything, DIAGNOSTIC);
// Timer relax3("relax3", Everything, DIAGNOSTIC);
// Timer relax4("relax4", Everything, DIAGNOSTIC);
// Timer relax5("relax5", Everything, DIAGNOSTIC);
// Timer relax6("relax6", Everything, DIAGNOSTIC);
// Timer relax7("relax7", Everything, DIAGNOSTIC);
// Timer relax8("relax8", Everything, DIAGNOSTIC);
// Timer relax9("relax9", Everything, DIAGNOSTIC);


typedef LevelData<FArrayBox> LDF;
typedef DisjointBoxLayout    DBL;

void compareTwoLDFs(const Vector<LDF*>& grid1, const Vector<LDF*>& grid2,
		    Vector<LDF*>& absDifference);


int main(int argc, char* argv[]) {

  int SIZE = 64;
  if(argc > 1 ) {
    SIZE = atoi(argv[1]);
  }

  int rank, number_procs;
#ifdef MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif
  cout << "rank = " << rank << " number_procs= " << number_procs << endl;
  { // parallel scope


  
  Everything.start();

  // for now, i just want a 2D problem and a 3D problem
  // that will execute at about the same length of time.
  // so i do the following:

#if CH_SPACEDIM == 2
  int maxLevel  = 3;  
  int Ncells[2] = {SIZE, SIZE};
#else
  int maxLevel  = 4; 
  int Ncells[3] = {8,8,8};
#endif

  int lbase               = 0;
  int iverbose            = 1;
  int maxIterations       = 20;
  int numberVcyclesBottom = 1;
  int blockFactor         = 8;
  int maxBoxSize          = 32;
  int bufferSize          = 1;
  int baseLevel           = 0; 

  int ibclo[3]  = {0,0,0};
  int ibchi[3]  = {0,0,0};

  Real fillRatio          = 0.77;
  Real refineThreshold    = 0.005;
  Real ProblemLowCord[3]  = {0.0, 0.0, 0.0};
  Real ProblemHighCord[3] = {1.0, 1.0, 1.0};

  Vector<int> referenceRatioVector;
  referenceRatioVector.resize(maxLevel+1);
  referenceRatioVector[0] = 2;
  for(int i=1; i<=maxLevel; i++) {
     referenceRatioVector[i] = 2;
  }

  bool verbose = (iverbose == 1);

  const IntVect unitVector(IntVect::TheUnitVector());

  DomainGhostBC domghostbc;
  setDomainBC(domghostbc, ibclo, ibchi, verbose);
  cout << "DomainGhostBC domghostbc built" << endl;

  PoissonOp poissonOperator;
  poissonOperator.setDomainGhostBC(domghostbc);
  cout << "PoissonOp poissonOperator built" << endl;

  Vector<LDF*> rhs;
  Vector<DBL> DBLVector;
  Vector<Box> boundingBoxVector;
  Vector<Real> DxVector;
  int numLevels=0;

  InitGrids.start();
  cout << " Initializing Grid..." << endl;
  InitializeGrids(rhs,                  //output
                  DBLVector,            //output
                  boundingBoxVector,    //output
                  DxVector,             //output
                  referenceRatioVector, 
                  Ncells,
                  ProblemLowCord,
                  ProblemHighCord,
                  maxBoxSize,
                  blockFactor,
                  bufferSize,
                  fillRatio,
                  refineThreshold,
                  maxLevel,               
                  numLevels,             //output
                  verbose);
  cout << " Initializing Grid...  Done." << endl;  
  InitGrids.stop();

  printf(" numLevels=%d\n", numLevels);


  Regrid.start();
  cout << " Regridding..." << endl;  
  regrid(rhs,                  //output
	 DBLVector,            //output
	 boundingBoxVector, 
	 referenceRatioVector,
         blockFactor, 
	 maxBoxSize, 
	 bufferSize, 
	 refineThreshold,   
         fillRatio, 
	 maxLevel, 
	 baseLevel, 
	 numLevels);           //output 
  cout << " Regridding... Done." << endl;  

  printf(" numLevels=%d\n", numLevels);

  AMRSolver amrSolver(DBLVector,  
		      boundingBoxVector,
                      DxVector,   
		      referenceRatioVector, 
                      numLevels, 
		      lbase, 
		      &poissonOperator);

  amrSolver.setVerbose(verbose);
  amrSolver.setMaxIter(maxIterations);
  amrSolver.setNumVCyclesBottom(numberVcyclesBottom);

  Vector<LDF*> phi(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    phi[ilev] = new LDF(DBLVector[ilev], 1, unitVector);
    //// do this is you want to give phi an initial value for each box
    // LDF& levelphi= *phi[ilev];      
    // Real rval = 3.0*ilev;
    // for(DataIterator dit = levelphi.dataIterator(); dit.ok(); ++dit) {
    //   rval += 1.0;
    //   levelphi[dit()].setVal(rval);
    // }
  }
  amrSolver.solveAMR(phi, rhs);  

  // make a coarser grid
  Vector<DBL> DBLVector_coarse(numLevels);
  Vector<Box> boundingBoxVector_coarse(numLevels);  
  Vector<Real> DxVector_coarse(numLevels);
  Vector<LDF*> phi_coarse(numLevels, NULL);

  for (int ilev=0; ilev<numLevels; ilev++) {
    coarsen(DBLVector_coarse[ilev], DBLVector[ilev], 2);
    boundingBoxVector_coarse[ilev] = coarsen(boundingBoxVector[ilev], 2);
    DxVector_coarse[ilev] = 2.0*DxVector[ilev];
    phi_coarse[ilev] = new LDF(DBLVector_coarse[ilev], 1, unitVector);
  }

  // average soln to coarser levels
  for (int lev=numLevels-1; lev>=0; lev--) {
    CoarseAverage avgDown(DBLVector[lev], 1, 2);
    avgDown.averageToCoarse(*phi_coarse[lev], *phi[lev]);
    //prettyPrintLDF(&phi_coarse);
    //prettyPrintLDF(phi[lev]);
  }

  outputHDF5(phi, DBLVector, boundingBoxVector,
	     referenceRatioVector, numLevels, "phi.hdf5");
  
  outputHDF5(phi_coarse, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "coarse.hdf5");
    

  PoissonOp poissonOperator_coarse;
  poissonOperator_coarse.setDomainGhostBC(domghostbc);
  AMRSolver amrSolver_coarse(DBLVector_coarse,
			     boundingBoxVector_coarse,
			     DxVector,   
			     referenceRatioVector, 
			     numLevels, 
			     lbase, 
			     &poissonOperator_coarse);

  amrSolver_coarse.setVerbose(verbose);
  amrSolver_coarse.setMaxIter(maxIterations);
  amrSolver_coarse.setNumVCyclesBottom(numberVcyclesBottom);

  Vector<LDF* > phi_coarse_lap(numLevels, NULL);
  for(int ilev = 0; ilev < numLevels; ilev++) { 
    phi_coarse_lap[ilev] = new LDF(DBLVector_coarse[ilev], 1, unitVector);
  }

  for (int lev=0; lev<numLevels; lev++) {
    amrSolver_coarse.applyAMROperator(phi_coarse, 
				      *phi_coarse_lap[lev], lev);
    //prettyPrintLDF(phi[lev]);
  }

  outputHDF5(phi_coarse_lap, DBLVector_coarse, boundingBoxVector_coarse,
	     referenceRatioVector, numLevels, "coarse_lap.hdf5");

  Output.start();
//   if(rank==0) {
//     cout << " outputData..." << endl;
//     outputData(phi, rhs, DBLVector, boundingBoxVector,
//             referenceRatioVector, numLevels, verbose);
//     cout << " outputData... done" << endl;
//   }
  Output.stop();


  for(int ilev = 0; ilev < numLevels; ilev++) {
    delete phi[ilev];
    delete rhs[ilev];
  }

  Everything.stop();

  Timer::TimerSummary(rank, number_procs);

  } // end parallel scope
#ifdef MPI
  //dumpmemoryatexit();
  MPI_Finalize();
#endif

  return(0);
}
