/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _BICGSTABSMOOTHERSIMPLE_H_
#define _BICGSTABSMOOTHERSIMPLE_H_

#include <iostream>
#include <cmath>
#include <cassert>
#include <cstdlib>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"

#include "BaseBottomSmoother.H"

/// Implements BiCGStab as a bottom smoother for LevelOp-derived classes
/**
    The BiCGStabSmoother class implements the BiCGStab algorithm for
    LevelOp-derived classes
*/
class BiCGStabSmootherSimple: public BaseBottomSmoother
{
public:
  /// Null constructor
  BiCGStabSmootherSimple();

  /// Destructor
  virtual ~BiCGStabSmootherSimple();

  /// "Virtual constructor" workaround.
  /**
      This gets around the "no virtual constructor" rule
   */
  virtual BaseBottomSmoother* new_bottomSmoother() const;

  /// Set maximum number of BiCGStab iterations.
  void setMaxIter(int a_maxIter);

  /// Set solver tolerance
  /**
      Set solver tolerance
   */
  void setSolverTol(Real a_solverTol);

  /// Set whether to enable restart capabillity
  /**
      If this is true, if BiCGStab reaches a hang point, the
      solver will recompute the residual and attempt to restart the
      computation
  */
  void setEnableRestart(bool a_enableRestart);

  /// Set verbosity level
  /**
      Set whether i/o is done when solving (default is false)
  */
  void setVerbose(bool a_verbose);

  /// Do smoothing
  /**
      Performs BiCGStab smoothing of L(a_phi) = a_rhs; L is the operator
      defined in a_levelopPtr, and the preconditioner is
      a_levelopPtr->smooth
  */
  virtual void doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              LevelOp*                    a_levelopPtr);

  /// set metric for judging convergence (if not set, defaults to 1.0)
  virtual void setConvergenceMetric(Real a_metric, int a_comp);

  virtual void clearMemory() {};

protected:
  /**
     Maximum number of BiCGStab iterations
  */
  int m_maxIter;

  /**
      Solver convergence metrics
  */
  Vector<Real> m_convergenceMetrics;

  /**
     Solver tolerance (we declare the problem solved if the residual is
     smaller than solverTol*convergenceMetric
  */
  Real m_solverTol;

  /**
     Small parameter to check for division by zero
  */
  Real m_small;

  /**
     Small parameter to check for size of correction (to prevent hanging)
  */
  Real m_convergeSmall;

  /**
     If true, enable restarting capability
  */
  bool m_enableRestart;

  /**
     Default is false
  */
  bool m_verbose;
};

#endif
