/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"

#include "CGSmoother.H"

#include <cstdio>  // for debugging sprintfs

extern double global_BS_tolerance;
extern double global_GSRB_relaxFactor;

#include "Timer.H"
#include "NoelsPrettyPrint.H"
extern Timer Everything;
Timer TBOTCG_doBottomSmooth("in doBS CG", 109);

Timer TBOTCG_init              ("init",           TBOTCG_doBottomSmooth);
Timer TBOTCG_compInitResidual  ("comp init res",  TBOTCG_doBottomSmooth);
Timer TBOTCG_compNorm          ("comp norm",      TBOTCG_doBottomSmooth);
Timer TBOTCG_preconditioner1   ("precond1",       TBOTCG_doBottomSmooth);
Timer TBOTCG_DotProduct1       ("DotProduct1",    TBOTCG_doBottomSmooth);
Timer TBOTCG_badDotProd_plus   ("badDotProd plus",TBOTCG_doBottomSmooth);
Timer TBOTCG_iter1stuff        ("iter1 stuff",    TBOTCG_doBottomSmooth);
Timer TBOTCG_stuff             ("stuff",          TBOTCG_doBottomSmooth);
Timer TBOTCG_applyOp1          ("applyOp1",       TBOTCG_doBottomSmooth);
Timer TBOTCG_DotProduct2       ("DotProduct2",    TBOTCG_doBottomSmooth);
Timer TBOTCG_increment_residual("inc residual",   TBOTCG_doBottomSmooth);
Timer TBOTCG_check_convergence ("check converg",  TBOTCG_doBottomSmooth);

CGSmoother::CGSmoother()
{
  m_maxIter = 73;
  m_solverTol = global_BS_tolerance;

#ifdef CH_USE_FLOAT
  m_small = 1.0e-30;
#endif
#ifdef CH_USE_DOUBLE
  m_small = 1.0e-60;
#endif
  m_verbose = false;
}

CGSmoother::~CGSmoother(){}

void CGSmoother::setMaxIter(int a_maxIter) {
  m_maxIter = a_maxIter;
}

void CGSmoother::setConvergenceMetric(Real a_metric, int a_comp) {

  // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size()) {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++) {
      m_convergenceMetrics[i] = tempVect[i];
    }
    for (int i=tempVect.size(); i<a_comp; i++) {
      m_convergenceMetrics[i] = 1.0;
    }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
}

void CGSmoother::setSolverTol(Real a_solverTol) {
  m_solverTol = a_solverTol;
}

void CGSmoother::setVerbose(bool a_verbose) {
  m_verbose = a_verbose;
}

BaseBottomSmoother* CGSmoother::new_bottomSmoother() const {
  CGSmoother* newsmoother = new CGSmoother();
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL) {
    MayDay::Error("Out of Memory in CGSmoother::new_bottomSmoother");
  }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

void
CGSmoother::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                           const LevelData<FArrayBox>& a_rhs,
                           LevelOp*                    a_levelopPtr) {

  TBOTCG_doBottomSmooth.start();
  Timer BS_timer;
  BS_timer.start();
  TBOTCG_init.start();
  Real small2 = 1.0e-8;
  int ncomp = a_rhs.nComp(); // 1
  char tempchar[128];

  int Npoints=0, Nboxes=0, Nghosts=0;
  NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);
  Real trelax = global_GSRB_relaxFactor;
  if (Npoints < 150) global_GSRB_relaxFactor = 0.7;
  //if (GlobalLevel == 0) global_GSRB_relaxFactor = 0.7;
  //global_GSRB_relaxFactor = 0.7;

  // in the case where the convergence metrics haven't already been
  // defined, default to norm(rhs) for the bottom smooth
  if (m_convergenceMetrics.size() != ncomp) {
    m_convergenceMetrics = Vector<Real>(ncomp, 0);
    Interval thisInterval(0,0);
    int normType = 0;
    m_convergenceMetrics[0] = norm(a_rhs, thisInterval, normType);
  } // end if we need to resize convergence metrics

  const DisjointBoxLayout grids = a_rhs.getBoxes();

  LevelData<FArrayBox> rmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> pmf(grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> zmf(grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> qmf(grids, ncomp, IntVect::TheZeroVector());

  DataIterator dit = rmf.dataIterator();
  for (dit.begin(); dit.ok(); ++dit) {
    // this is necessary to prevent uninitialized memory reads
    rmf[dit()].setVal(0.0);
    pmf[dit()].setVal(0.0);
    zmf[dit()].setVal(0.0);
    qmf[dit()].setVal(0.0);
  }
  TBOTCG_init.stop();

  TBOTCG_compInitResidual.start();
  // compute initial residual:  residual = b - Ax
  a_levelopPtr->applyOpH(a_phi, rmf);  // residual = Ax

  for (dit.begin(); dit.ok(); ++dit) {
    rmf[dit()] -= a_rhs[dit()];  // residual -= b
    rmf[dit()].negate();         // residual = -residual
  }
  TBOTCG_compInitResidual.stop();

  // compute residual and rhs norms
  TBOTCG_compNorm.start();
  Real rnorm, alpha;
  Interval compInterval(0,0);
  rnorm = norm(rmf, compInterval, 0);

  // if initial residual is already 0, don't do anything here
  bool initialResidualZero = false;
  if (rnorm < m_small) {
    initialResidualZero = true;
  }
  TBOTCG_compNorm.stop();

  //sprintf(tempchar, "       CGSmoother BS  initNorm=%12.4e  zI=%d tol=%10.4e cmet=%10.4e small=%10.4e  maxIter=%d\n",
  //    rnorm, initialResidualZero, m_solverTol, m_convergenceMetrics[0],
  //    m_small, m_maxIter);
  //pout() << tempchar;
  //Timer IterationTimer;
  //Real previous_rnorm = rnorm;
  Real rho, previous_rho;

  bool finished = initialResidualZero;
  bool badDotProd = false;
  int iter;
  for (iter = 1; (iter < m_maxIter) && !finished; iter++) {

    //IterationTimer.clear();
    //IterationTimer.start();

    // solve M(z(i-1)) = r(i-1)
    TBOTCG_preconditioner1.start();
    a_levelopPtr->levelPreconditioner(zmf, rmf);
    TBOTCG_preconditioner1.stop();

    // rho(i) = r(i-1)^T z(i-1)
    TBOTCG_DotProduct1.start();
    previous_rho = rho;
    rho = DotProduct(rmf, zmf, grids, compInterval);

    if (Abs(rho) < m_small) {
      badDotProd = true;
    } else {
      badDotProd = false;
    }
    TBOTCG_DotProduct1.stop();

    //  this sometimes does happen, but is rare.
    // what do i set phi to be?
    if (badDotProd) {
      TBOTCG_badDotProd_plus.start();
      TBOTCG_badDotProd_plus.stop();
      goto done;
    }

    if (iter == 1) {
      // p(1) = z(0)
      TBOTCG_iter1stuff.start();
      for (dit.begin(); dit.ok(); ++dit) {
        pmf[dit()].copy(zmf[dit()]);
      }
      TBOTCG_iter1stuff.stop();
      previous_rho = rho;
    } else {
      // beta(i-1) = rho(i-1)/rho(i-2)
      // p(i) = z(i-1) + beta(i-1)*p(i-1)
      TBOTCG_stuff.start();
      Real beta = rho/previous_rho;
      for (dit.begin(); dit.ok(); ++dit) {
        pmf[dit()] *= beta;
        pmf[dit()] += zmf[dit()];
      }
      TBOTCG_stuff.stop();
    }

    // q(i) = A(p(i))
    TBOTCG_applyOp1.start();
    a_levelopPtr->applyOpH(pmf, qmf);
    TBOTCG_applyOp1.stop();

    // alpha(i) = rho(i-1)/(p(i)^T q(i))
    TBOTCG_DotProduct2.start();
    Real denom = DotProduct(pmf, qmf, grids, compInterval);
    bool badCorrection = false;
    if (Abs(denom) > small2*Abs(rho)) {
      alpha = rho/denom;
    } else {  // this isn't happening, and i don't know what to do if it does... CHECK
      alpha = 0.0;
      badCorrection = true;
      pout() << " CGSmoother badCorrection denom=" << denom
             << " small2*Abs(rho)= " << small2*Abs(rho) << endl;
    }
    TBOTCG_DotProduct2.stop();

    // x(i) = x(i-1) + alpha(i)*p(i)
    // r(i) = r(i-1) - alpha(i)*q(i)
    TBOTCG_increment_residual.start();
    for (dit.begin(); dit.ok(); ++dit) {
      a_phi[dit()].plus(pmf[dit()], alpha, 0, 0, 1);  // update phi here CHECK
      rmf[dit()].plus(qmf[dit()], -alpha, 0, 0, 1);
    }
    TBOTCG_increment_residual.stop();

    // now check convergence by comparing norm of residual with conv criteria
    TBOTCG_check_convergence.start();
    rnorm = norm(rmf, compInterval, 0);

    if ((rnorm < m_solverTol*m_convergenceMetrics[0])) {
      finished = true;
    }

    //IterationTimer.stop();
    //sprintf(tempchar, "       CG BS after %3d iters, rnorm=%11.4e (%8.2f) w=%5.3f\n",
    //    iter, rnorm, previous_rnorm/rnorm,
    //    IterationTimer.wc_time());
    //pout() << tempchar;
    //previous_rnorm = rnorm;

    TBOTCG_check_convergence.stop();

  } // end iterations on iter
 done:
  TBOTCG_doBottomSmooth.stop();

  BS_timer.stop();

  //#if 0
  // debug print statements
  //int Npoints=0, Nboxes=0, Nghosts=0;
  //NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);
  // Nboxes should now be same as grids.size()
  char stuff[180];
  sprintf(stuff, "     end doBScg L%d Npoints=%6d Nboxes=%4d iter=%2d zI=%1d badDP=%1d rnorm=%9.3e tol=%7.1e relax=%5.2f wc=%6.3f\n",
          GlobalLevel, Npoints, Nboxes, iter,
          initialResidualZero, badDotProd, rnorm,
          m_solverTol, global_GSRB_relaxFactor,
          BS_timer.wc_time());
  pout() << stuff;

  global_GSRB_relaxFactor = trelax;

  bool converged=true;
  if (iter >= m_maxIter) converged = false;
  NoelsSooperIterationCounter[BIterationFlavor].Count++;
  NoelsSooperIterationCounter[BIterationFlavor].Value+=iter;
  if (!converged) {
    NoelsSooperIterationCounter[BIterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[BIterationFlavor].wc += BS_timer.wc_time();
  BIterationFlavor = 0;

}
