/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"

#include "CGSquaredSmoother.H"

#include <cstdio>  // for debugging sprintfs

extern double global_BS_tolerance;

#include "Timer.H"
#include "NoelsPrettyPrint.H"
extern Timer Everything;
Timer TBOTCGS_doBottomSmooth("in doBS CGSquard", 110);

Timer TBOTCGS_init              ("init",           TBOTCGS_doBottomSmooth);
Timer TBOTCGS_compInitResidual  ("comp init res",  TBOTCGS_doBottomSmooth);
Timer TBOTCGS_compNorm          ("comp norm",      TBOTCGS_doBottomSmooth);
Timer TBOTCGS_DotProduct1       ("DotProduct1",    TBOTCGS_doBottomSmooth);
Timer TBOTCGS_badDotProd_plus   ("badDotProd plus",TBOTCGS_doBottomSmooth);
Timer TBOTCGS_iter1stuff        ("iter1 stuff",    TBOTCGS_doBottomSmooth);
Timer TBOTCGS_stuff             ("stuff",          TBOTCGS_doBottomSmooth);
Timer TBOTCGS_preconditioner1   ("precond1",       TBOTCGS_doBottomSmooth);
Timer TBOTCGS_applyOp1          ("applyOp1",       TBOTCGS_doBottomSmooth);
Timer TBOTCGS_DotProduct2       ("DotProduct2",    TBOTCGS_doBottomSmooth);
Timer TBOTCGS_stuff2            ("stuff2",         TBOTCGS_doBottomSmooth);
Timer TBOTCGS_preconditioner2   ("precond2",       TBOTCGS_doBottomSmooth);
Timer TBOTCGS_increment_corr    ("inc corr",       TBOTCGS_doBottomSmooth);
Timer TBOTCGS_applyOp2          ("applyOp2",       TBOTCGS_doBottomSmooth);
Timer TBOTCGS_increment_residual("inc residual",   TBOTCGS_doBottomSmooth);
Timer TBOTCGS_check_convergence ("check converg",  TBOTCGS_doBottomSmooth);
Timer TBOTCGS_doublecheck_convergence ("doublecheck", TBOTCGS_doBottomSmooth);

CGSquaredSmoother::CGSquaredSmoother()
{
  m_maxIter = 103;
  m_solverTol = global_BS_tolerance;

#ifdef CH_USE_FLOAT
  m_small = 1.0e-30;
#endif
#ifdef CH_USE_DOUBLE
  m_small = 1.0e-60;
#endif
  m_convergeSmall = 1.0e-2;
  m_enableRestart = true;
  m_verbose = false;
}

CGSquaredSmoother::~CGSquaredSmoother(){}

void CGSquaredSmoother::setMaxIter(int a_maxIter) {
  m_maxIter = a_maxIter;
}

void CGSquaredSmoother::setConvergenceMetric(Real a_metric, int a_comp) {

  // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size()) {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++) {
      m_convergenceMetrics[i] = tempVect[i];
    }
    for (int i=tempVect.size(); i<a_comp; i++) {
      m_convergenceMetrics[i] = 1.0;
    }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
}

void CGSquaredSmoother::setSolverTol(Real a_solverTol) {
  m_solverTol = a_solverTol;
}

void CGSquaredSmoother::setEnableRestart(bool a_enableRestart) {
  m_enableRestart = a_enableRestart;
}

void CGSquaredSmoother::setVerbose(bool a_verbose) {
  m_verbose = a_verbose;
}

BaseBottomSmoother* CGSquaredSmoother::new_bottomSmoother() const {
  CGSquaredSmoother* newsmoother = new CGSquaredSmoother();
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL) {
    MayDay::Error("Out of Memory in CGSquaredSmoother::new_bottomSmoother");
  }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

void
CGSquaredSmoother::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                                  const LevelData<FArrayBox>& a_rhs,
                                  LevelOp*                    a_levelopPtr) {

  TBOTCGS_doBottomSmooth.start();
  Timer BS_timer;
  BS_timer.start();
  TBOTCGS_init.start();
  Real small2 = 1.0e-8;
  int ncomp = a_rhs.nComp(); // 1
  char tempchar[128];

  // in the case where the convergence metrics haven't already been
  // defined, default to norm(rhs) for the bottom smooth
  if (m_convergenceMetrics.size() != ncomp) {
    m_convergenceMetrics = Vector<Real>(ncomp, 0);
    Interval thisInterval(0,0);
    int normType = 0;
    m_convergenceMetrics[0] = norm(a_rhs, thisInterval, normType);
  } // end if we need to resize convergence metrics

  const DisjointBoxLayout grids = a_rhs.getBoxes();

  LevelData<FArrayBox> corremf(grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> phatmf (grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> uhatmf (grids, ncomp, IntVect::TheUnitVector());

  LevelData<FArrayBox> residmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> rtwidmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> pmf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> vhatmf (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> qhatmf (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> umf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> qmf    (grids, ncomp, IntVect::TheZeroVector());

  DataIterator dit = residmf.dataIterator();
  for (dit.begin(); dit.ok(); ++dit) {
    // this is necessary to prevent uninitialized memory reads
    corremf[dit()].setVal(0.0);
    phatmf[dit()] .setVal(0.0);
    uhatmf[dit()] .setVal(0.0);

    residmf[dit()].setVal(0.0);
    rtwidmf[dit()].setVal(0.0);
    pmf[dit()]    .setVal(0.0);
    vhatmf[dit()] .setVal(0.0);
    qhatmf[dit()] .setVal(0.0);
    umf[dit()]    .setVal(0.0);
    qmf[dit()]    .setVal(0.0);
  }
  TBOTCGS_init.stop();

  // compute initial residual  residual = b - Ax
  TBOTCGS_compInitResidual.start();
  a_levelopPtr->applyOpH(a_phi, residmf);  // residual = Ax

  for (dit.begin(); dit.ok(); ++dit) {
    residmf[dit()] -= a_rhs[dit()];  // residual -= b
    residmf[dit()].negate();         // residual = -residual
  }
  TBOTCGS_compInitResidual.stop();

  // compute residual and rhs norms
  TBOTCGS_compNorm.start();
  Real rho, rho_previous, rnorm, alpha;
  Interval compInterval(0,0);
  rnorm = norm(residmf, compInterval, 0);

  // if initial residual is already 0, don't do anything here
  bool initialResidualZero = false;
  if (rnorm < m_small) {
    initialResidualZero = true;
  } else {
    // set r-twiddle to be the initial residual.
    for (dit.begin(); dit.ok(); ++dit) {
      rtwidmf[dit()].copy(residmf[dit()], 0, 0, ncomp);
    }
  }
  TBOTCGS_compNorm.stop();

  //sprintf(tempchar, "      CGSquaredSmoother BS        initNorm=%12.4e  tol=%10.4e small=%10.4e  maxIter=%d\n",
  //      rnorm, m_solverTol, m_small, m_maxIter);
  //pout() << tempchar;
  Timer IterationTimer;
  Real previous_rnorm = rnorm;

  bool finished = initialResidualZero;
  bool badDotProd = false;
  int iter;
  for (iter = 1; (iter < m_maxIter) && !finished; iter++) {

    //IterationTimer.clear();
    //IterationTimer.start();

    TBOTCGS_DotProduct1.start();
    rho = DotProduct(residmf, rtwidmf, grids, compInterval);

    if (Abs(rho) < m_small) {
      badDotProd = true;
    } else {
      badDotProd = false;
    }
    TBOTCGS_DotProduct1.stop();

    if (badDotProd) {
      TBOTCGS_badDotProd_plus.start();
      for (dit.begin(); dit.ok(); ++dit) {
        a_phi[dit()].plus(corremf[dit()], 0, 0, ncomp);
      }
      TBOTCGS_badDotProd_plus.stop();
      goto done;
    }

    if (iter == 1) {
      // u(1) = p(1) = r(0)
      TBOTCGS_iter1stuff.start();
      for (dit.begin(); dit.ok(); ++dit) {
        umf[dit()].copy(residmf[dit()]);
        pmf[dit()].copy(residmf[dit()]);
      }
      TBOTCGS_iter1stuff.stop();
      rho_previous = rho;
    } else {

      // u(i) = r(i-1) + beta(i-1)*q(i-1)
      // p(i) = u(i) + beta(i-1)*(q(i-1) + beta(i-1)*p(i-1))
      TBOTCGS_stuff.start();
      Real beta = rho/rho_previous;
      for (dit.begin(); dit.ok(); ++dit) {
        umf[dit()].setVal(0.0);
        umf[dit()].plus(qmf[dit()], beta, 0, 0, 1);
        umf[dit()] += residmf[dit()];

        qmf[dit()].plus(pmf[dit()], beta, 0, 0, 1);
        pmf[dit()].setVal(0.0);
        pmf[dit()].plus(qmf[dit()], beta, 0, 0, 1);
        pmf[dit()] += umf[dit()];
      } // end loop over grids
      TBOTCGS_stuff.stop();
    } // end if i ==  1 's else

    // solve M(phat) = p(i)
    TBOTCGS_preconditioner1.start();
    a_levelopPtr->levelPreconditioner(phatmf, pmf);
    TBOTCGS_preconditioner1.stop();

    // vhat(i) = A(phat)
    TBOTCGS_applyOp1.start();
    a_levelopPtr->applyOpH(phatmf, vhatmf);
    TBOTCGS_applyOp1.stop();

    // alpha(i) = rho(i-1)/(rtwid^T vhat(i))
    TBOTCGS_DotProduct2.start();
    Real denom = DotProduct(vhatmf, rtwidmf, grids, compInterval);
    bool badCorrection = false;
    if (Abs(denom) > small2*Abs(rho)) {
      alpha = rho/denom;
    } else {
      alpha = 0.0;
      badCorrection = true;
      pout() << " CGSquaredSmoother badCorrection denom=" << denom
             << "  small2*Abs(rho) = "
             << small2*Abs(rho) << endl;
    }
    rho_previous = rho;
    TBOTCGS_DotProduct2.stop();

    // q(i) = u(i) - alpha*vhat(i)
    TBOTCGS_stuff2.start();
    for (dit.begin(); dit.ok(); ++dit) {
      qmf[dit()].copy(umf[dit()]);
      qmf[dit()].plus(vhatmf[dit()], -alpha, 0, 0, 1);
    }

    // actually need q(i)+u(i) below,
    // so store that in u(i) since we don't need u(i) anymore
    // u'(i) = q(i) + u(i)
    for (dit.begin(); dit.ok(); ++dit) {
      umf[dit()] += qmf[dit()];
    }
    TBOTCGS_stuff2.stop();

    // solve M(uhat) = (u(i) + q(i)) = u'(i)
    TBOTCGS_preconditioner2.start();
    a_levelopPtr->levelPreconditioner(uhatmf, umf);
    TBOTCGS_preconditioner2.stop();

    // increment correction: corr(i)  = corr(i-1)  + alpha*uhat
    TBOTCGS_increment_corr.start();
    for (dit.begin(); dit.ok(); ++dit) {
      corremf[dit()].plus(uhatmf[dit()], alpha, 0, 0, 1);
    }
    TBOTCGS_increment_corr.stop();

    // qhat(i) = A(uhat(i))
    TBOTCGS_applyOp2.start();
    a_levelopPtr->applyOpH(uhatmf, qhatmf);
    TBOTCGS_applyOp2.stop();

    // increment residual: resid(i) = resid(i-1) - alpha*qhat
    TBOTCGS_increment_residual.start();
    for (dit.begin(); dit.ok(); ++dit) {
      residmf[dit()].plus(qhatmf[dit()], -alpha, 0, 0, 1);
    }
    TBOTCGS_increment_residual.stop();

    // now check convergence by comparing norm of residual with conv criteria
    TBOTCGS_check_convergence.start();
    rnorm = norm(residmf, compInterval, 0);

    if ((rnorm < m_solverTol*m_convergenceMetrics[0])) {
      finished = true;
    }
    TBOTCGS_check_convergence.stop();

    // if residnorm is small enough (or if correction is
    // small enough), then recompute residual from scratch
    // and check again (to avoid roundoff-drift issues).  if
    // we _still_ think we're done, then exit
    TBOTCGS_doublecheck_convergence.start();
    if (finished) {
      // recompute residual from scratch -- first increment a_phi with correction
      for (dit.begin(); dit.ok(); ++dit) {
        a_phi[dit()].plus(corremf[dit()], 0, 0, ncomp);
      }

      a_levelopPtr->applyOpH(a_phi, residmf);

      for (dit.begin(); dit.ok(); ++dit) {
        residmf[dit()].minus(a_rhs[dit()], 0, 0, 1);
      }

      // now recompute residual norms and check for convergence
      rnorm = norm(residmf, compInterval, 0);

      if (rnorm < m_solverTol*m_convergenceMetrics[0]) {
        // really finished.
      } else {
        pout() << " CGSquaredSmoother: allDone -- but need to try again " << endl;
        // otherwise, go back and try again with new residual

        // also reset \tilde{r} and correction
        for (dit.begin(); dit.ok(); ++dit) {
          rtwidmf[dit()].copy(residmf[dit()],0,0,ncomp);
          corremf[dit()].setVal(0.0);
        }

        finished = false;
      }
    } // end if we're recomputing the residual
    TBOTCGS_doublecheck_convergence.stop();

    //IterationTimer.stop();
    //sprintf(tempchar, "  CGSQ BS after %3d iters, rnorm=%11.4e (%8.2f) w=%5.3f\n",
    //      iter, rnorm, previous_rnorm/rnorm, IterationTimer.wc_time());
    //pout() << tempchar;
    //previous_rnorm = rnorm;

  } // end iterations on iter
 done:
  TBOTCGS_doBottomSmooth.stop();

  BS_timer.stop();

  //#if 0
  // debug print statements
  int Npoints=0, Nboxes=0, Nghosts=0;
  NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);
  // Nboxes should now be same as grids.size()
  char stuff[180];
  sprintf(stuff, "   end doBScgsq L%d Npoints=%6d Nboxes=%4d iter=%2d zI=%1d badDP=%1d rnorm=%9.3e tol=%7.1e met=%9.3e wc=%6.3f\n",
          GlobalLevel, Npoints, Nboxes, iter,
          initialResidualZero, badDotProd, rnorm,
          m_solverTol, m_solverTol*m_convergenceMetrics[0],
          BS_timer.wc_time());
  pout() << stuff;

  bool converged = iter >= m_maxIter;
  NoelsSooperIterationCounter[BIterationFlavor].Count++;
  NoelsSooperIterationCounter[BIterationFlavor].Value+=iter;
  if (!converged) {
    NoelsSooperIterationCounter[BIterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[BIterationFlavor].wc += BS_timer.wc_time();
  BIterationFlavor = 0;

}
