/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// from NoOpSmoother by DFMartin, Sun, May 5, 2002
// petermc, 10 Jan 2003
#ifdef HYPRE_AS_BS

#ifndef _HYPREBOTTOMSMOOTHER_H_
#define _HYPREBOTTOMSMOOTHER_H_

#include <iostream>
#include <cassert>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"

#include "BaseBottomSmoother.H"
#include "FluxBox.H"

/// Implements Hypre as a bottom smoother for LevelOp-derived classes
/**
    The HypreBottomSmoother class implements the HypreBottom algorithm for
    LevelOp-derived classes
*/
class HypreBottomSmoother: public BaseBottomSmoother
{

public:
  enum hypre_solver_type {SMG, PFMG, BICGSTAB, PCG};

  /// constructor
  HypreBottomSmoother();

  HypreBottomSmoother(const Real a_dxLevel,
                      const Real a_dxCoarse,
                      const ProblemDomain& a_domain,
                      const hypre_solver_type a_solverType);

  /// Destructor
  virtual ~HypreBottomSmoother();

  virtual void define(const Real a_dxLevel,
                      const Real a_dxCoarse,
                      const ProblemDomain& a_domain,
                      const hypre_solver_type a_solverType);

  /// "Virtual constructor" workaround.
  /**
      This gets around the "no virtual constructor" rule
   */
  virtual BaseBottomSmoother* new_bottomSmoother() const;

  ///
  //void clearMemory();

  ///
  /** set up solver & matrix for HYPRE
   */
  void setup_solver(const DisjointBoxLayout& a_grids,
                    const ProblemDomain& a_domain,
                    const Real& a_dxLevel,
                    const Real& a_dxCoarse);

  ///
  /** appropriately modify matrix at CF interfaces
   */
  void modifyMatrixAtCFBoundaries(const Box bx,
                                  FArrayBox &matFab,
                                  int &Ncells_modified);

  /// Set maximum number of HypreBottom iterations.
  void setMaxIter(int a_maxIter);

  /// Set solver tolerance
  /**
      Set solver tolerance
   */
  void setSolverTol(Real a_solverTol);

  /// Set verbosity level
  /**
      Set whether i/o is done when solving (default is false)
  */
  void setVerbose(bool a_verbose);

  /// Do smoothing
  /**
      Performs HypreBottom smoothing of L(a_phi) = a_rhs; L is the operator
      defined in a_levelopPtr, and the preconditioner is
      a_levelopPtr->smooth
  */
  virtual void doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              LevelOp*                    a_levelopPtr);

  /// set metric for judging convergence (if not set, defaults to 1.0)
  virtual void setConvergenceMetric(Real a_metric, int a_comp);

  virtual void clearMemory();

  hypre_solver_type solverType() const;

  struct HYPRE_impl;

protected:

  /**
      Solver convergence metrics
  */
  Vector<Real> m_convergenceMetrics;

  // problem domain at this level
  ProblemDomain m_domain;

  //
  HYPRE_impl* m_impl;

  // SMG, PFMG, BICGSTAB
  hypre_solver_type m_solverType;

  /**
     Maximum number of HypreBottom iterations
  */
  int m_maxIter;

  /**
     Solver tolerance (amount by which residual must be reduced before we
     declare the problem "solved")
  */
  Real m_tolerance;

  /**
     Default is false
  */
  bool m_verbose;

  // have you filled in the matrix?
  // bool m_initializedSolver;

  enum boundaryCondition { DIRICHLET, NEUMANN };

  boundaryCondition m_boundaryCondition;

  // do you have m_acoefs and m_bcoefs?
  bool m_initializedHypre;

  LevelData<FArrayBox>* m_acoefs;
  LevelData<FluxBox>* m_bcoefs;

  int m_numPre;
  int m_numPost;

  // grids at this level
  DisjointBoxLayout  m_grids;

  Real m_alpha;
  Real m_beta;
  Real m_dxLevel;
  Real m_dxCoarse;

private:

   ///
  /** Transpose a_fab to a_array.
   */
  void transpose(Real* a_array, const FArrayBox& a_fab) const;

};

#endif

#endif // on HYPRE_AS_BS
