/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifdef HYPRE_AS_BS

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
#include "ProblemDomain.H"
#include "CFIVS.H"

#include "HypreBottomSmoother.H"

#include <cstdio>  // for debugging sprintfs

#include "memusage.H"
#include "Timer.H"
#include "NoelsPrettyPrint.H"
extern Timer Everything;

#include "AMRIO.H"
extern int NbottomSmooth;
extern int write_bottomsmoother_file;  // hack for now (noel)

#include "FORT_PROTO.H"
#include "HABECF_F.H"

#include "HYPRE_struct_ls.h"
#include "struct_mv.h"

Timer TimeHBS_doBottomSmooth("HYPRE botSmooth", 378);
Timer TimeHBS_doBottomSmoothInit ("HBS Init",  TimeHBS_doBottomSmooth);
Timer TimeHBS_doBottomSmoothSolve("HBS Solve", TimeHBS_doBottomSmooth);

Timer TimehypreInit;

#define STENCILSIZE (2*CH_SPACEDIM + 1)

HypreBottomSmoother::HypreBottomSmoother()
 :  m_convergenceMetrics(1,1.0)
{
  // set defaults

  //m_solverType = SMG;
  //m_solverType = PFMG;
  m_solverType = BICGSTAB;
  //m_solverType = PCG;

  m_initializedHypre = false;
  //pout() << " New HypreBottomSmoother HYPRE " << get_memory_usage_from_OS() << endl;
}

HypreBottomSmoother::HypreBottomSmoother(Real  a_dxLevel,
                                         Real  a_dxCoarse,
                                         const ProblemDomain& a_domain,
                                         const hypre_solver_type a_solverType)
 :  m_convergenceMetrics(1,1.0)
{
  define(a_dxLevel, a_dxCoarse, a_domain, a_solverType);
}

void
HypreBottomSmoother::define(Real  a_dxLevel,
                            Real  a_dxCoarse,
                            const ProblemDomain& a_domain,
                            const hypre_solver_type a_solverType)
{
  m_verbose = false;
  m_domain = a_domain;
  m_dxCoarse = a_dxCoarse;
  m_dxLevel = a_dxLevel;
  m_solverType = a_solverType;
  m_initializedHypre = false;
  //pout() << " New HypreBottomSmoother (with arguments) HYPRE " << get_memory_usage_from_OS() << endl;
}

HypreBottomSmoother::~HypreBottomSmoother()
{
}

void HypreBottomSmoother::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
}

void HypreBottomSmoother::setConvergenceMetric(Real a_metric, int a_comp)
{
  // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size())
  {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++)
      {
        m_convergenceMetrics[i] = tempVect[i];
      }
    for (int i=tempVect.size(); i<a_comp; i++)
      {
        m_convergenceMetrics[i] = 1.0;
      }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
}

void HypreBottomSmoother::setSolverTol(Real a_tolerance)
{
  m_tolerance = a_tolerance;
}

void HypreBottomSmoother::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}

BaseBottomSmoother* HypreBottomSmoother::new_bottomSmoother() const
{
  HypreBottomSmoother* newsmoother = new HypreBottomSmoother(m_dxLevel,
                                                             m_dxCoarse,
                                                             m_domain,
                                                             m_solverType);
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL)
    {
      MayDay::Error("Out of Memory in HypreBottomSmoother::new_bottomSmoother");
    }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

struct HypreBottomSmoother::HYPRE_impl
{
  HYPRE_StructGrid    grid;
  //HYPRE_StructStencil stencil;

  HYPRE_StructMatrix  A; // , A0;  don't need A0
  HYPRE_StructVector  b;
  HYPRE_StructVector  x;

  HYPRE_StructSolver  precond;
  HYPRE_StructSolver  solver;
};

namespace
{
  MPI_Comm mpi_comm = MPI_COMM_WORLD;

  inline int ispow2(int i)
  {
    return (i == 1) ? 1 : (((i <= 0) || (i & 1)) ? 0 : ispow2(i / 2));
  }

  inline int*
  loV(const Box& b)
  {
    return const_cast<int*>(b.loVect());
  }

  inline int*
  hiV(const Box& b)
  {
    return const_cast<int*>(b.hiVect());
  }
}

void HypreBottomSmoother::setup_solver(const DisjointBoxLayout& a_grids,
                                       const ProblemDomain& a_domain,
                                       const Real& a_dxLevel,
                                       const Real& a_dxCoarse)
{
  pout() << " HypreBottomSmoother::setup_solver" << endl;

  if (a_dxLevel == 0) {
    MayDay::Error("HypreBottomSmoother::setup_solver a_dxLevel=0");
  }

  if (a_dxCoarse == 0) {
    MayDay::Error("HypreBottomSmoother::setup_solver a_dxCoarse=0");
  }

  m_impl = new HYPRE_impl();

  m_numPre  = 1; // was 1  (was 4, back to 1 [ndk])
  m_numPost = 1; // was 1  (was 4, back to 1 [ndk])
  m_tolerance = 1.0e-7;
  m_maxIter = 60;

  m_grids    = a_grids;
  m_domain   = a_domain;
  m_dxLevel  = a_dxLevel;
  m_dxCoarse = a_dxCoarse;

  m_boundaryCondition = NEUMANN;  // DIRICHLET

  //  Set up structured-grid object, m_impl->grid.

  // first create empty structured-grid object, m_impl->grid,
  // living on mpi_comm.
  HYPRE_StructGridCreate(mpi_comm, CH_SPACEDIM, &m_impl->grid);
  // Remove the periodic stuff because it causes problems.
  // (Must be power of 2, and boxes need to wrap around.)
  /*
  if ( m_domain.isPeriodic() )  // periodic in any direction
    {
      const Box& bx = m_domain.domainBox();
      int periodicity[SpaceDim];
      for ( int idim = 0; idim < SpaceDim; idim++ )
        {
          periodicity[idim] = 0;
          // if ( geom.isPeriodic(i) )
          if ( m_domain.isPeriodic(idim) )
            {
              periodicity[idim] = bx.size(idim);
              // verify that periodicity[idim] is a power of 2, because
              // hypre requires this.

              assert(ispow2(periodicity[idim]));

              assert(bx.smallEnd(0) == 0);
            }
        }
      // Set periodicities of box.
      HYPRE_StructGridSetPeriodic(m_impl->grid, periodicity);
    }
  */

  // DataIterator gets data in this proc only.
  DataIterator dit = m_grids.dataIterator();

  // (In the original, if num_procs != 1, then use distribution
  // mapping to set only on this proc.)
  for (dit.begin(); dit.ok(); ++dit)
    {
      // Add a new box to the structured-grid object.
      const Box& bx = a_grids.get(dit());
      HYPRE_StructGridSetExtents(m_impl->grid, loV(bx), hiV(bx));
    }
  // Finalize assembly of structured-grid object.
  HYPRE_StructGridAssemble(m_impl->grid);

  //  Set up stencil.
#if (CH_SPACEDIM == 2)
  // 5-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1
  //   4
  int offsets[5][2] = {{ 0,  0},
                       {+1,  0},
                       {-1,  0},
                       { 0, +1},
                       { 0, -1}};
#elif (CH_SPACEDIM == 3)
  // 7-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1   (and:  5 above 0, 6 below 0)
  //   4
  int offsets[7][3] = {{ 0,  0,  0},
                       {+1,  0,  0},
                       {-1,  0,  0},
                       { 0, +1,  0},
                       { 0, -1,  0},
                       { 0,  0, +1},
                       { 0,  0, -1}};
#endif

  // number of ghost layers in each direction.  always array of length 6.
#if (CH_SPACEDIM == 2)
  int A_num_ghost[6] = { 1, 1, 1, 1, 0, 0 };
#elif (CH_SPACEDIM == 3)
  int A_num_ghost[6] = { 1, 1, 1, 1, 1, 1 };
#endif

  HYPRE_StructStencil stencil;
  // first create empty CH_SPACEDIM-dimensional, STENCILSIZE-point
  // stencil object, stencil.
  HYPRE_StructStencilCreate(CH_SPACEDIM, STENCILSIZE, &stencil);

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      // assign array numbers for stencil entry icomp.
      HYPRE_StructStencilSetElement(stencil, icomp, offsets[icomp]);
    }

  // Set up matrix.

  // first create empty matrix object, m_impl->A, with m_impl->grid and stencil.
  HYPRE_StructMatrixCreate(mpi_comm, m_impl->grid, stencil, &m_impl->A);
  HYPRE_StructMatrixSetNumGhost(m_impl->A, A_num_ghost);
  // tell hypre that matrix coefficients are now ready to be set.
  HYPRE_StructMatrixInitialize(m_impl->A);

  // Matrix coefficients are set in setup_solver(), not here.

  //  Set up right-hand side vector.

  // create empty vector objects, m_impl->b and m_impl->x, with m_impl->grid.
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->b);
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->x);

  HYPRE_StructStencilDestroy(stencil); // no longer needed
  // tell hypre that vector values are now ready to be set.
  HYPRE_StructVectorInitialize(m_impl->b);
  HYPRE_StructVectorInitialize(m_impl->x);

  //  Other stuff:  define m_acoefs and m_bcoefs.
  const int ncomp=1;
  const int ngrow=0;
  IntVect ghostVect = ngrow * IntVect::Unit;
  m_acoefs = new LevelData<FArrayBox>(a_grids, ncomp, ghostVect);
  m_bcoefs = new LevelData<FluxBox>  (a_grids, ncomp, ghostVect);

  //  Set up hypre solver.

  m_alpha = 0.; // weight of identity
  m_beta  = 1.; // weight of div grad

  for (dit.begin(); dit.ok(); ++dit)
    {
      m_acoefs->operator[](dit()).setVal(0.);
      m_bcoefs->operator[](dit()).setVal(1.);
    }

  ///////////////////////////////////////////////////////////////////////
  // Now setup the matrix

  // stencil size is 5 or 7 (2*dim + 1)
  int stencil_indices[STENCILSIZE];

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      stencil_indices[icomp] = icomp;
    }

  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  char tempp[180];
  sprintf(tempp, "m_alpha=%10.4f m_beta=%10.4f m_tolerance=%10.4e m_dxLevel=%10.4f m_dxCoarse=%10.4f\n",
        m_alpha, m_beta, m_tolerance, m_dxLevel, m_dxCoarse);
  pout() << tempp;

  int Nboxes, Npoints, Nghosts;
  NoelsLDFABgetUsefulInfo(m_acoefs, Nboxes, Npoints, Nghosts);
  //pout() << " Nboxes=" << Nboxes << " Npoints=" << Npoints << endl;

  Nboxes=0;
  int totalNcellsModifiedAtCF = 0;
  for (DataIterator dit = m_acoefs->dataIterator(); dit.ok(); ++dit, ++Nboxes)
    {
      const Box& bx = m_grids.get(dit());
      //pout() << " box N" << Nboxes << " bx=" << bx << endl;

      FArrayBox matFab(bx, STENCILSIZE);

      //build matrix interior
      // Operator on phi(:) is:
      // L(phi)(:) = alpha * a(:) * phi(:) + beta * div (b(:) * grad phi(:))

      // Set coefficients of first term, alpha * a(:) * phi(:).
      // currently, m_alpha is 0 (ndk)
      if (m_alpha == 0.)
        {
          matFab.setVal(0., 0); // set component 0
        }
      else
        {
          FArrayBox& acoefsFab = m_acoefs->operator[](dit());
          matFab.copy(acoefsFab, 0, 0, 1); // srccomp, destcomp, numcomp
          matFab.mult(m_alpha, 0, 1); // startcomp, numcomp
        }
      //        FORT_HACOEF(CHF_FRA(mat),
      //                    CHF_CONST_FRA1(acoefsFab, 0),
      //                    CHF_CONST_REAL(m_alpha));

      // Set coefficients of second term, beta * div (b(:) * grad phi(:)).
      // for now, b (m_bcoefs) are all 1  (ndk)
      for ( int idim = 0; idim < SpaceDim; idim++ )
    {
          const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];
          FORT_HBCOEF(CHF_FRA(matFab),
                      CHF_CONST_FRA1(bfab, 0),
                      CHF_CONST_REAL(m_beta),
                      CHF_CONST_REAL(m_dxLevel),
                      CHF_CONST_INT(idim));
    }

      // add boundary conditions to matrix diagonal, and
      //   zero out offdiag values at domain boundaries.
      SideIterator sit;
      for (int idim = 0; idim < SpaceDim; idim++) {
        // if ( ! m_domain.isPeriodic(idim) ) // hypre has probs with periodic
        for (sit.begin(); sit.ok(); sit.next())
          {
            Box region(bx);
            //cout << " idim=" << idim << " sit=" << sit() << endl;

            int iside;
            bool intersectsBoundary = false;
            switch (sit())
              {
              case Side::Lo:
                {
                  int smallCoordinate = region.smallEnd(idim);
                  if (smallCoordinate == loDomain[idim])
                    {
                      intersectsBoundary = true;
                      region.setBig(idim, smallCoordinate);
                      iside = -1;
                    }
                  break;
                }
              case Side::Hi:
                {
                  int bigCoordinate = region.bigEnd(idim);
                  if (bigCoordinate == hiDomain[idim])
                    {
                      intersectsBoundary = true;
                      region.setSmall(idim, bigCoordinate);
                      iside = +1;
                    }
                  break;
                }
              default:
                cerr << "SideIterator error" << endl;
                abort();
              }

            if (intersectsBoundary)
              {
                //cout << " intersectsBoundary dim=" << idim << " iside=" << iside << endl;
                const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];

                // DIRICHLET 0
                // NEUMANN 1
                const int bc=m_boundaryCondition;
                FORT_HBMAT(CHF_FRA(matFab),
                           CHF_CONST_FRA1(bfab, 0),
                           CHF_BOX(region),
                           CHF_CONST_INT(idim),
                           CHF_CONST_INT(iside),
                           CHF_CONST_REAL(m_beta),
                           CHF_CONST_REAL(m_dxLevel),
                           CHF_CONST_INT(bc));
              }
          } // end iteration over faces

      } // end side loop

      // NoelsPrettyPrint(&matFab);

      // adjust for coarse-fine boundaries
      int Ncells_modified=0;
      modifyMatrixAtCFBoundaries(bx, matFab, Ncells_modified);
      totalNcellsModifiedAtCF += Ncells_modified;

      const int volume = bx.numPts();
      Real* mat = hypre_CTAlloc(Real, STENCILSIZE*volume);
      // original has mat(0:3, regl0:regh0, regl1:regh1, regl2:regh2)
      // note that the component indices are FIRST,
      // but Chombo Fortran assumes they are LAST.
      // Will need to do some complicated choreography to transpose.

      transpose(mat, matFab);

      // initialize matrix

      // fill in the part of the matrix m_impl->A on box bx, with mat.
      HYPRE_StructMatrixSetBoxValues(m_impl->A, loV(bx), hiV(bx),
                                     STENCILSIZE, stencil_indices, mat);

      hypre_TFree(mat);

#if 0
      char tok[200];
      for(BoxIterator bit(bx); bit.ok(); ++bit) {
        sprintf(tok, "box%4d iv0:%3d iv1:%3d matfab: %4.0f %4.0f %4.0f %4.0f %4.0f\n",
                Nboxes, bit().getVect()[0], bit().getVect()[1],
                matFab.get(bit(), 0), matFab.get(bit(), 1),
                matFab.get(bit(), 2), matFab.get(bit(), 3),
                matFab.get(bit(), 4));
        cout << tok;
      }
#endif

    }

  //

  // Finalize assembly of matrix object.
  HYPRE_StructMatrixAssemble(m_impl->A);

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op

  if ( m_solverType == SMG )
    {
      HYPRE_StructSMGCreate         (mpi_comm, &m_impl->solver);
      HYPRE_StructSMGSetMemoryUse   (m_impl->solver, 0);
      HYPRE_StructSMGSetMaxIter     (m_impl->solver, m_maxIter);
      HYPRE_StructSMGSetRelChange   (m_impl->solver, 0);
      HYPRE_StructSMGSetTol         (m_impl->solver, m_tolerance);
      HYPRE_StructSMGSetNumPreRelax (m_impl->solver, m_numPre);
      HYPRE_StructSMGSetNumPostRelax(m_impl->solver, m_numPost);
      HYPRE_StructSMGSetLogging     (m_impl->solver, 1);
      HYPRE_StructSMGSetup          (m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if ( m_solverType == PFMG )
    {
      HYPRE_StructPFMGCreate         (mpi_comm, &m_impl->solver);
      //HYPRE_StructPFMGSetMemoryUse(solver, 0);
      // for isotropic problems, can set to 1.  otherwise, 0.
      HYPRE_StructPFMGSetSkipRelax   (m_impl->solver, 1);
      HYPRE_StructPFMGSetMaxIter     (m_impl->solver, m_maxIter);
      HYPRE_StructPFMGSetRelChange   (m_impl->solver, 0);
      HYPRE_StructPFMGSetTol         (m_impl->solver, m_tolerance);
      HYPRE_StructPFMGSetNumPreRelax (m_impl->solver, m_numPre);
      HYPRE_StructPFMGSetNumPostRelax(m_impl->solver, m_numPost);
      HYPRE_StructPFMGSetLogging     (m_impl->solver, 1);
      HYPRE_StructPFMGSetup          (m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if ( m_solverType == BICGSTAB )
    {
      HYPRE_StructBiCGSTABCreate       (mpi_comm, &m_impl->solver);
      HYPRE_StructBiCGSTABSetMaxIter   (m_impl->solver, m_maxIter);
      HYPRE_StructBiCGSTABSetPrintLevel(m_impl->solver, 0);
      HYPRE_StructBiCGSTABSetTol       (m_impl->solver, m_tolerance);
      HYPRE_StructBiCGSTABSetLogging   (m_impl->solver, 0);

#define USE_PFMG_PRECONDITIONER
#ifdef USE_PFMG_PRECONDITIONER
      // use symmetric PFMG as preconditioner
      HYPRE_StructPFMGCreate         (mpi_comm, &m_impl->precond);
      HYPRE_StructPFMGSetMaxIter     (m_impl->precond, 1);
      HYPRE_StructPFMGSetTol         (m_impl->precond, 0.0);
      HYPRE_StructPFMGSetZeroGuess   (m_impl->precond);

      // coarse grid operator type
      // 0 - Galerkin (default?)
      // 1 - non-Galerkin ParFlow operators
      // 2 - Galerkin, general operators
      HYPRE_StructPFMGSetRAPType     (m_impl->precond, 0);

      // relaxation type
      // 0 - Jacobi
      // 1 - Weighted Jacobi (default?)
      // 2 - R/B Gauss-Seidel
      // 3 - R/B Gauss-Seidel (nonsymmetric)
      HYPRE_StructPFMGSetRelaxType   (m_impl->precond, 1);

      // number of pre and post relaxations
      HYPRE_StructPFMGSetNumPreRelax (m_impl->precond, 1);
      HYPRE_StructPFMGSetNumPostRelax(m_impl->precond, 1);

      // skip levels in PFMG (0 or 1)
      HYPRE_StructPFMGSetSkipRelax   (m_impl->precond, 1);

      //HYPRE_StructPFMGSetDxyz      (m_impl->precond, dxyz);

      HYPRE_StructPFMGSetPrintLevel  (m_impl->precond, 0);
      HYPRE_StructPFMGSetLogging     (m_impl->precond, 0);
      HYPRE_StructBiCGSTABSetPrecond( m_impl->solver,
                                      HYPRE_StructPFMGSolve,
                                      HYPRE_StructPFMGSetup,
                                      m_impl->precond);
#endif

      //#define USE_SMG_PRECONDITIONER
#ifdef USE_SMG_PRECONDITIONER
      // use symmetric SMG as preconditioner
      HYPRE_StructSMGCreate         (mpi_comm, &m_impl->precond);
      HYPRE_StructSMGSetMaxIter     (m_impl->precond, 1);
      HYPRE_StructSMGSetTol         (m_impl->precond, 0.0);
      HYPRE_StructSMGSetZeroGuess   (m_impl->precond);

      // number of pre and post relaxations
      HYPRE_StructSMGSetNumPreRelax (m_impl->precond, 1);
      HYPRE_StructSMGSetNumPostRelax(m_impl->precond, 1);

      HYPRE_StructSMGSetPrintLevel  (m_impl->precond, 0);
      HYPRE_StructSMGSetLogging     (m_impl->precond, 0);
      HYPRE_StructBiCGSTABSetPrecond( m_impl->solver,
                                      HYPRE_StructSMGSolve,
                                      HYPRE_StructSMGSetup,
                                      m_impl->precond);
#endif

      HYPRE_StructBiCGSTABSetup(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if ( m_solverType == PCG )
    {
      HYPRE_StructPCGCreate       (mpi_comm, &m_impl->solver);
      HYPRE_StructPCGSetMaxIter   (m_impl->solver, m_maxIter);
      HYPRE_StructPCGSetPrintLevel(m_impl->solver, 0 );
      HYPRE_StructPCGSetTol       (m_impl->solver, m_tolerance);
      HYPRE_StructPCGSetLogging   (m_impl->solver, 1 );
      //HYPRE_PCGSetRelChange       (m_impl->solver, 0 );
      //HYPRE_PCGSetTwoNorm         (m_impl->solver, 1 );

#define USE_PFMG_PRECONDITIONER
#ifdef USE_PFMG_PRECONDITIONER
      // use symmetric PFMG as preconditioner
      HYPRE_StructPFMGCreate         (mpi_comm, &m_impl->precond);
      HYPRE_StructPFMGSetMaxIter     (m_impl->precond, 1);
      HYPRE_StructPFMGSetTol         (m_impl->precond, 0.0);
      HYPRE_StructPFMGSetZeroGuess   (m_impl->precond);

      // coarse grid operator type
      // 0 - Galerkin (default?)
      // 1 - non-Galerkin ParFlow operators
      // 2 - Galerkin, general operators
      HYPRE_StructPFMGSetRAPType     (m_impl->precond, 0);

      // relaxation type
      // 0 - Jacobi
      // 1 - Weighted Jacobi (default?)
      // 2 - R/B Gauss-Seidel
      // 3 - R/B Gauss-Seidel (nonsymmetric)
      HYPRE_StructPFMGSetRelaxType   (m_impl->precond, 1);

      // number of pre and post relaxations
      HYPRE_StructPFMGSetNumPreRelax (m_impl->precond, 1);
      HYPRE_StructPFMGSetNumPostRelax(m_impl->precond, 1);

      // skip levels in PFMG (0 or 1)
      HYPRE_StructPFMGSetSkipRelax   (m_impl->precond, 1);

      //HYPRE_StructPFMGSetDxyz      (m_impl->precond, dxyz);

      HYPRE_StructPFMGSetPrintLevel  (m_impl->precond, 0);
      HYPRE_StructPFMGSetLogging     (m_impl->precond, 0);
      HYPRE_StructPCGSetPrecond      (m_impl->solver,
                                      HYPRE_StructPFMGSolve,
                                      HYPRE_StructPFMGSetup,
                                      m_impl->precond);
#endif

      HYPRE_StructPCGSetup        (m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else
    {
      MayDay::Error("HypreBottomSmoother: no such solver");
    }

  //cout << " totalNcellsModifiedAtCF=" << totalNcellsModifiedAtCF << endl;
}

void
HypreBottomSmoother::clearMemory()
{
  if (m_initializedHypre)
    {
      //pout() << " Free HYPRE memory before:" << get_memory_usage_from_OS() << endl;
      delete m_acoefs;
      delete m_bcoefs;

      // destroy the solver before destroying m_impl
      if (m_solverType == SMG)
        {
          HYPRE_StructSMGDestroy(m_impl->solver);
        }
      else if (m_solverType == PFMG)
        {
          HYPRE_StructPFMGDestroy(m_impl->solver);
        }
      else if (m_solverType == BICGSTAB)
        {
          HYPRE_StructBiCGSTABDestroy(m_impl->solver);
        }
      else if (m_solverType == PCG)
        {
          HYPRE_StructPCGDestroy(m_impl->solver);
        }

      HYPRE_StructVectorDestroy(m_impl->b);
      HYPRE_StructVectorDestroy(m_impl->x);
      HYPRE_StructMatrixDestroy(m_impl->A);
      HYPRE_StructGridDestroy(m_impl->grid);
      delete m_impl;
      //pout() << " Free HYPRE memory after:" << get_memory_usage_from_OS() << endl;
    } else {
      //pout() << " HYPRE not initialized, so memory not freed" << get_memory_usage_from_OS()  << endl;
    }
  m_initializedHypre = false;
}

void
HypreBottomSmoother::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                                    const LevelData<FArrayBox>& a_rhs,
                                    LevelOp*                    a_levelopPtr)
{
  pout() << " HypreBottomSmoother::doBottomSmooth mem=" << get_memory_usage_from_OS() << endl;

  TimeHBS_doBottomSmooth.start();
  Timer HBS;
  HBS.start();

  if (m_initializedHypre == false) {
    TimeHBS_doBottomSmoothInit.start();
    pout() << " Initialize HYPRE solver  mem=" << get_memory_usage_from_OS() << endl;
    setup_solver(a_phi.getBoxes(), m_domain, m_dxLevel, m_dxCoarse);
    pout() << " DONE Initialize HYPRE solver  mem=" << get_memory_usage_from_OS() << endl;
    TimeHBS_doBottomSmoothInit.stop();
    m_initializedHypre = true;
  }

  NbottomSmooth++;

  //#define WRITE_BOTTOMSMOOTH_FILE
#ifdef WRITE_BOTTOMSMOOTH_FILE
  if (write_bottomsmoother_file == 1) {
    pout() << " (from HYPRE) writing BOTTOMSMOOTH_FILE NbottomSmooth=" << NbottomSmooth << endl;
    //void WriteAMRHierarchyHDF5(    //This is blocking
    //     const string& filename,
    //     const Vector<DisjointBoxLayout>& a_vectGrids, //grids at each level
    //     const Vector<LevelData<FArrayBox>* > & a_vectData, //data at each level
    //     const Vector<string>& a_vectNames,       //names of variables
    //     const Box& a_domain,                   //domain at coarsest level
    //     const Real& a_dx,     //grid spacing at coarsest level
    //     const Real& a_dt,    //time step at coarsest level
    //     const Real& a_time,
    //     const Vector<int>& a_vectRatio,    //refinement ratio at all levels
    //     const int& a_numLevels);//number of levels to output

    if (NbottomSmooth>1 ) {
      int numComp = 2;  // for phi and rhs
      Vector<LevelData<FArrayBox>*> debugData(2, NULL);

      Vector<DisjointBoxLayout> vectGrids(1);
      vectGrids[0] = a_rhs.getBoxes(); // same as a_phi

      debugData[0] = new LevelData<FArrayBox>(a_rhs.getBoxes(), numComp);

      Interval rhsInterval(0,0);
      a_rhs.copyTo(a_rhs.interval(), *debugData[0], rhsInterval);

      Interval phiInterval(1,1);
      a_phi.copyTo(a_phi.interval(), *debugData[0], phiInterval);

      Box domainBox = m_domain.domainBox();

      Vector<string> dataNames(numComp);
      dataNames[0] = "rhs";
      dataNames[1] = "phi";

      Vector<int> vectRefRatio(2,1);  // what's 2,1 for?
      vectRefRatio[0] = (int)(m_dxCoarse/m_dxLevel);

      char filename[45];
      sprintf(filename, "bs%dd-%03d.hdf5", CH_SPACEDIM, NbottomSmooth);

      //char filename2[45];
      //sprintf(filename2, "bsl%03d.hdf5", NbottomSmooth);

      //writeLevelname(&a_rhs, filename2);
      WriteAMRHierarchyHDF5(filename,
                            vectGrids,       // vector of 1
                            debugData,       //
                            dataNames,       //
                            domainBox,
                            m_dxLevel,       // dt, hack it with dxLevel
                            m_dxCoarse,
                            0.0,             // newTime, doesn't matter
                            vectRefRatio,    //
                            1);              // one level

      //cout << " m_dxLevel=" << m_dxLevel << endl;
      //cout << " m_dxCrse=" << m_dxCrse << endl;
      //cout << " domainBox=" << domainBox << endl;

      //NoelsPrettyPrint(&a_rhs, "rhs");
      //NoelsPrettyPrint(&a_phi, "phi");
    }
  }
#endif

  const DisjointBoxLayout& grids = a_phi.getBoxes();
  // assert( grids == m_grids );

  //phi.setVal(0.0);
  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  Real *vec;
  bool has_ghosts = ( a_phi.ghostVect() != IntVect::Zero );
  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      // initialize phi, since we will reuse the space to set up rhs below:

      FArrayBox& phiFab = a_phi[dit()];
      FArrayBox* phiFabPtr = &phiFab;

      if (has_ghosts)
    {
      phiFabPtr = new FArrayBox(bx, 1);
          phiFabPtr->copy(phiFab);
    }

      // sharing space, phi will be overwritten below
      vec = phiFabPtr->dataPtr();

      // set vector coefficients for this box
      HYPRE_StructVectorSetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      const FArrayBox& rhsFab = a_rhs[dit()];
      phiFabPtr->copy(rhsFab); // why do i do this?

      // incorporate boundary conditions into rhs:
      // do nothing, because this is the homogeneous case.

      // initialize rhs

      HYPRE_StructVectorSetBoxValues(m_impl->b, loV(bx), hiV(bx), vec);

      if (has_ghosts) delete phiFabPtr; // contains vec
    }

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op

  int Nboxes, Npoints, Nghosts;
  NoelsLDFABgetUsefulInfo(a_phi, Nboxes, Npoints, Nghosts);
  //if (Npoints < 1000) {
  //HYPRE_StructMatrixPrint("hmatA", m_impl->A, 0);
  //HYPRE_StructVectorPrint("hmatb", m_impl->b, 0);
  //}

  TimeHBS_doBottomSmoothSolve.start();
  Timer hypreSolve;
  hypreSolve.start();

  //tol_hypre = tol_chombo*convergenceMetric/norm(b)
  //Real norminit = norm(residmf, compInt, 0);
  //Real hypreTolerance = m_tolerance*m_convergenceMetric[0]/;
  // need to change the tolerance here for each call to HYPRE
  // based on the new convergence metric stuff
  if (m_solverType == SMG)
    {
      pout() << " call HYPRE_StructSMGSolve" << endl;
      HYPRE_StructSMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if (m_solverType == PFMG)
    {
      pout() << " call HYPRE_StructPFMGSolve" << endl;
      HYPRE_StructPFMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if (m_solverType == BICGSTAB)
    {
      pout() << " call HYPRE_StructBiCGSTABSolve" << endl;
      HYPRE_StructBiCGSTABSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else if (m_solverType == PCG)
    {
      pout() << " call HYPRE_StructPCGSolve" << endl;
      HYPRE_StructPCGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  hypreSolve.stop();
  TimeHBS_doBottomSmoothSolve.stop();
  pout() << " hypre solve done wc=" << hypreSolve.wc_time() << endl;

  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      FArrayBox& phiFab = a_phi[dit()];
      FArrayBox* phiFabPtr = &phiFab;

      if (has_ghosts)
    {
      phiFabPtr = new FArrayBox(bx, 1);
          phiFabPtr->copy(phiFab);
    }

      // sharing space, phi will be overwritten below
      vec = phiFabPtr->dataPtr();

      HYPRE_StructVectorGetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      if (has_ghosts)
    {
      a_phi[dit()].copy(*phiFabPtr);
      delete phiFabPtr;
    }
    }

//   if (m_verbose && procID() == uniqueProc(SerialTask::compute))
//     {
//       int num_iterations;
//       Real res;

//       if (m_solverType == SMG)
//         {
//    HYPRE_StructSMGGetNumIterations(m_impl->solver, &num_iterations);
//    HYPRE_StructSMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
//         }
//       else if (m_solverType == PFMG)
//         {
//    HYPRE_StructPFMGGetNumIterations(m_impl->solver, &num_iterations);
//    HYPRE_StructPFMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
//         }
//       else if (m_solverType == BICGSTAB)
//         {
//    HYPRE_StructBiCGSTABGetNumIterations(m_impl->solver, &num_iterations);
//    HYPRE_StructBiCGSTABGetFinalRelativeResidualNorm(m_impl->solver, &res);
//         }

//       //const int oldprec = pout().precision(20);
//       //pout() << num_iterations << " Hypre Multigrid Iterations, Relative Residual "
//       //     << res << endl;
//       //pout().precision(oldprec);
//     }

  HBS.stop();
#define DEBUG_PRINT
#ifdef DEBUG_PRINT
  int num_iterations=0;
  Real res=0.0;
  if (m_solverType == SMG)
    {
      HYPRE_StructSMGGetNumIterations(m_impl->solver, &num_iterations);
      HYPRE_StructSMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
    }
  else if (m_solverType == PFMG)
    {
      HYPRE_StructPFMGGetNumIterations(m_impl->solver, &num_iterations);
      HYPRE_StructPFMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
    }
  else if (m_solverType == BICGSTAB)
    {
      HYPRE_StructBiCGSTABGetNumIterations(m_impl->solver, &num_iterations);
      HYPRE_StructBiCGSTABGetFinalRelativeResidualNorm(m_impl->solver, &res);
    }
  else if (m_solverType == PCG)
    {
      HYPRE_StructPCGGetNumIterations(m_impl->solver, &num_iterations);
      HYPRE_StructPCGGetFinalRelativeResidualNorm(m_impl->solver, &res);
    }

  // debug print statements
  // Nboxes should now be same as grids.size()
  char stuff[180];
  //sprintf(stuff, " end HYPREdoBottomSmooth Npoints=%7d Nboxes=%7d iterations=%3d res=%10.3e hbs1=%8.3f hbs2=%8.3f hbs3=%8.3f hbs4=%8.3f wc=%10.3f\n",
  //      Npoints, Nboxes, num_iterations, res,
  //      hbs1.wc_time(), hbs2.wc_time(), hbs3.wc_time(), hbs4.wc_time(),
  //      HBS.wc_time());
  sprintf(stuff, " end HYPREdoBottomSmooth L%d Np=%7d Nbox=%5d iters=%3d res=%10.3e m=%5.1f wc=%10.3f\n",
          GlobalLevel, Npoints, Nboxes, num_iterations, res, get_memory_usage_from_OS(),
          HBS.wc_time());
  pout() << stuff;
#endif

  TimeHBS_doBottomSmooth.stop();
}

void
HypreBottomSmoother::transpose(Real* a_array,
                               const FArrayBox& a_fab) const
{
  Box bx = a_fab.box();
  int ncomps = a_fab.nComp();
  BoxIterator bit(bx);
  IntVect bxLo = bx.smallEnd();
  IntVect bxHi = bx.bigEnd();
  IntVect lens = bxHi - bxLo + IntVect::Unit;
  for (bit.begin(); bit.ok(); ++bit)
    {
      IntVect iv = bit();
      IntVect ivOff = iv - bxLo;
      // ptIndex = (ivOff[0] * lens[0] + ivOff[1]) * lens[1] + ivOff[2] ?
      // ptIndex = ivOff[2] + lens[2] * (ivOff[1] + lens[1] * ivOff[0]) ?
      // ptIndex = ivOff[0] + lens[0] * (ivOff[1] + lens[1] * ivOff[2])
      int ptOffIndex = ivOff[SpaceDim-1];
      for (int idim = SpaceDim-2; idim >= 0; idim--)
        {
          ptOffIndex *= lens[idim];
          ptOffIndex += ivOff[idim];
        }
      for (int nv = 0; nv < ncomps; nv++)
        {
          // int ind = nv * volume + ptOffIndex;
          int ind = ptOffIndex * ncomps + nv;
          a_array[ind] = a_fab(iv, nv);
        }
    }
}

void
HypreBottomSmoother::modifyMatrixAtCFBoundaries(const Box bx,
                                                FArrayBox &matFab,
                                                int &Ncells_modified)
{
  // don't need to do this if dxCoarse or dxLevel is negative or 0
  // as that means we are on level 0 and should only have physical boundary interfaces
  // And the CFIVS object should return an empty set anyway.
  if (m_dxCoarse > 0.0 && m_dxLevel > 0.0) {
    const Real rr = m_dxCoarse/m_dxLevel;
    const Real Beta = 1.0 + (rr - 3.0)/(1.0 + rr);
    const Real Alpha = (1.0 - rr)/(3.0 + rr);
    //printf(" dxCoarse=%10.4f dxLevel=%10.4f rr=%10.4f Alpha=%10.4f Beta=%10.4f idxc2=%10.4f\n",
    //     m_dxCoarse, m_dxLevel[0], rr, Alpha, Beta, 1.0/(m_dxCoarse*m_dxCoarse));
    const Real dxLevel = m_dxLevel;
    SideIterator sit;

    // Find all CF interfaces and modify matFab for each.
    for (int idim = 0; idim < SpaceDim; idim++) {
      for (sit.begin(); sit.ok(); sit.next()) {
        Box region(bx);
        // if side is 0, then sidePole is +1.  if side is 1, then sidePole is -1.
        const int sidePole = -1*(sit()+sit()) + 1;
        // if side is 0, then arrow points to the +1 side, otherwise the -1 side
        const IntVect arrow = sidePole * BASISV(idim);
        // convert 0,1,or2 to 1,3,or5
        const int dir135 = 2*idim + 1;
        //cout << " idim=" << idim << " sit=" << sit() << endl;
        CFIVS cfivsObject(m_domain, region, m_grids, idim, sit());
        //cout << cfivsObject.isDefined() << endl;
        //cout << cfivsObject.getFineIVS() << endl;
        for(IVSIterator it(cfivsObject.getFineIVS()); it.ok(); ++it) {
          IntVect iv(it());
          iv.shift(arrow);
          //cout << "   " << it() << " --> " << iv
          //   <<  " " << matFab.get(iv, 0)
          //   <<  " " << matFab.get(iv, 1)
          //   <<  " " << matFab.get(iv, 2)
          //   <<  " " << matFab.get(iv, 3)
          //   <<  " " << matFab.get(iv, 4)
          //   <<  "   sit=" << sit() << " sp" << sidePole << " d" << dir135
          //   << endl;
          //void set(const IntVect& iv, int comp, Real val)
          //Real get(const IntVect& iv, int comp) const

          // this cell is modified
          matFab(iv, 0) += Beta/(dxLevel*dxLevel);

          if (sit() == 0) { // low side
            // one cell higher is modified
            matFab(iv, dir135)   += Alpha/(dxLevel*dxLevel);
            // one cell lower is zeroed
            matFab(iv, dir135+1)  = 0.0;
          } else { // high side
            // one cell higher is zeroed
            matFab(iv, dir135)    = 0.0;
            // one cell lower is modifed
            matFab(iv, dir135+1) += Alpha/(dxLevel*dxLevel);
          }

          Ncells_modified++;

          //cout << "ch " << it() << " --> " << iv
          //   <<  " " << matFab.get(iv, 0)
          //   <<  " " << matFab.get(iv, 1)
          //   <<  " " << matFab.get(iv, 2)
          //   <<  " " << matFab.get(iv, 3)
          //   <<  " " << matFab.get(iv, 4)
          //   << endl;
        }
      }
    }
  }

}

HypreBottomSmoother::hypre_solver_type
HypreBottomSmoother::solverType() const
{
  return m_solverType;
}

#endif // on #ifdef HYPRE_AS_BS
