/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Fri, July 23, 1999

#include "LevelSolver.H"
#include "IntVectSet.H"
#include "parstream.H"
#include "LayoutIterator.H"

using std::cout;
using std::cerr;
using std::endl;

#include "Timer.H"
Timer TLS_levelSolve   ("in levelSolve",  101);
Timer TLS_applyOpI     ("applyOpI",       TLS_levelSolve);
Timer TLS_applyOpH     ("applyOpH",       TLS_levelSolve);
Timer TLS_mgRelax      ("mgRelax",        TLS_levelSolve);
Timer TLS_ditOperations("dit Operations", TLS_levelSolve);
Timer TLS_norm         ("norm",           TLS_levelSolve);

// for Timers in LevelMG.cpp
extern Timer TLMG_mgRelax_all;
extern Timer Everything;
Timer TLevelSolver_define("LS def", Everything, DTIMEDEFINES, true);

Timer TLS_def_all("LS define", Everything, 213, true);
Timer TLS_def1("LS def1", TLS_def_all);
Timer TLS_def2("define levelOp", TLS_def_all);
Timer TLS_def3("LS def3", TLS_def_all);
Timer TLS_def4("LS def4", TLS_def_all);
Timer TLS_def5("define levelMG", TLS_def_all);

//#include "NoelsPrettyPrint.H"

bool LevelSolver::isDefined() const
{
  return m_isDefined;
}

void LevelSolver::setDefaultValues()
{
  m_levelOpPtr = NULL;

  // set this to an invalid value
  m_nRefCrse = -1;
  m_dxLevel = -1.0;

  m_maxIter = 33;
  m_minIter = 4;

  m_tolerance = 1.0e-10;
  m_operatorTolerance = 1.0e-4;
#ifdef CH_USE_FLOAT
  m_tolerance = sqrt(m_tolerance);
#endif

  // default is maxNorm
  m_normType = 0;

  m_verbose = false;

  m_isDefined = false;
}

void LevelSolver::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}

void LevelSolver::setNumSmoothUp(int a_numSmoothUp)
{
  assert(isDefined());

  m_numSmoothUp = a_numSmoothUp;
  m_levelMG.setNumSmoothUp(a_numSmoothUp);
}

void LevelSolver::setNumSmoothDown(int a_numSmoothDown)
{
  assert(isDefined());

  m_numSmoothDown = a_numSmoothDown;
  m_levelMG.setNumSmoothDown(a_numSmoothDown);
}

void LevelSolver::setTolerance(Real a_tolerance)
{
  m_tolerance = a_tolerance;
}

// set tolerance for stopping due to hung convergence
void LevelSolver::setOperatorTolerance(Real a_tolerance)
{
  assert(a_tolerance >= 0.0);

  m_operatorTolerance = a_tolerance;
}

void LevelSolver::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
}

void LevelSolver::setMinIter(int a_minIter)
{
  m_minIter = a_minIter;
}

void
LevelSolver::setConvergenceMetric(Real a_metric, int a_comp)
{
    // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size())
  {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++)
      {
        m_convergenceMetrics[i] = tempVect[i];
      }
    for (int i=tempVect.size(); i<a_comp; i++)
      {
        m_convergenceMetrics[i] = 1.0;
      }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
  // pass this on through to the LevelMG (and from there to the LevelOps)
  m_levelMG.setConvergenceMetric(a_metric, a_comp);
}

void
LevelSolver::resetConvergenceMetrics()
{
  m_convergenceMetrics.clear();
  // don't reset levelMG convergence metrics,
  // since when the're re-set in the LevelSolver, the new
  // metrics will be passed through.
}

void
LevelSolver::setNormType(int a_normType)
{
  m_normType = a_normType;
}

///////////////
// default constructor
LevelSolver::LevelSolver()
{
  setDefaultValues();
}

// destructor
LevelSolver::~LevelSolver()
{
  clearMemory();
}

void LevelSolver::define(const DisjointBoxLayout& a_grids,
                         const DisjointBoxLayout* a_baseGrids,
                         const Box&               a_domain,
                         Real                     a_dxLevel,
                         int                      a_nRefCrse,
                         const LevelOp* const     a_opin,
                         int                      a_ncomp,
                         bool                     a_limitCoarsening)
{
  ProblemDomain physdomain(a_domain);

  define(a_grids, a_baseGrids, a_domain, a_dxLevel, a_nRefCrse,
         a_opin, a_ncomp, a_limitCoarsening);
}

void LevelSolver::define(const DisjointBoxLayout& a_grids,
                         const DisjointBoxLayout* a_baseGrids,
                         const ProblemDomain&     a_domain,
                         Real                     a_dxLevel,
                         int                      a_nRefCrse,
                         const LevelOp* const     a_opin,
                         int                      a_ncomp,
                         bool                     a_limitCoarsening)
{
  TLevelSolver_define.start();
  TLS_def_all.start();

  TLS_def1.start();
  clearMemory();

  m_grids = a_grids;
  m_domain = a_domain;
  m_nRefCrse = a_nRefCrse;
  m_dxLevel = a_dxLevel;

  m_resid.  define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_scratch.define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_corr.   define(m_grids, a_ncomp, IntVect::TheUnitVector());

  m_isDefined = true;
  TLS_def1.stop();

  TLS_def2.start();
  m_levelOpPtr = a_opin->new_levelop();
  m_levelOpPtr->define(m_grids, a_baseGrids, m_dxLevel,
                       m_nRefCrse, m_domain, false, a_ncomp);
  TLS_def2.stop();

  // coarsen boxes once to start, because coarsest level boxes
  // must also be coarsenable (for proper coarse-fine interpolation)
  // however, if grid merely covers entire physical domain, then
  // want to coarsen all the way down
  int nCoarserLevels = -1;

  // if limiting coarsening, only coarsen down to next AMR level
  // only relevant if there's a coarser AMR level defined
  //pout() << " a_limitCoarsening = " << a_limitCoarsening
  //     << " (a_baseGrids != NULL) = " << (a_baseGrids != NULL) << endl;
  if (a_limitCoarsening && (a_baseGrids != NULL) )
    {
      TLS_def3.start();
      nCoarserLevels = 1;
      int temp = m_nRefCrse/2;
      // funny way to do a log base 2
      while (temp > 1)
        {
          nCoarserLevels++;
          temp /= 2;
        } // end log base 2
      TLS_def3.stop();
    }  // end if we're limiting the coarsening
  else
    {
      TLS_def4.start();
      // begin if we're coarsening as far as possible

      // in this case, if there is a coarse-fine interface, we can
      // only coarsen down to grids with a dimension of 2, since we
      // need 2 fine cells to do coarse-fine interpolation.
      // However, if the entire domain is covered, then we can coarsen
      // as much as possible (down to a 1x1 grid), since we don't need
      // to do any coarse-fine interpolation.

      // (DFM 10/21/02) don't bother doing _any_ of this if m_grids
      // is an empty set
      if (m_grids.size() > 0)
        {
          // why do it this way rather than as a dense IVS?
          IntVectSet ivsCheck(m_domain.domainBox());

          LayoutIterator lit = m_grids.layoutIterator();
          for (lit.reset(); lit.ok(); ++lit)
            {
              ivsCheck -= m_grids.get(lit());
            }

          if (ivsCheck.isEmpty())
            {
              nCoarserLevels += 1;
            }

          int refrattot = 2;

          bool getOut = false;
          while (!getOut)
            {
              bool addone = true;
              for (lit.reset(); lit.ok(); ++lit)
                {
                  Box blocal = m_grids.get(lit());

                  if (refine(coarsen(blocal,refrattot),refrattot) != blocal)
                    {
                      addone = false;
                    }

                  if (!addone)
                    {
                      break;
                    }
                }

              if (addone)
                {
                  nCoarserLevels += 1;
                  refrattot *= 2;
                }
              else
                {
                  getOut = true;
                }
            } // end while loop
        } // end if there are actually grids

      TLS_def4.stop();
    } // end if we're coarsening as far as possible

  nCoarserLevels = Max(nCoarserLevels, 0);

  TLS_def5.start();
  m_levelMG.define(m_grids, a_baseGrids, m_dxLevel,
                   m_nRefCrse, m_domain, nCoarserLevels, a_opin, a_ncomp);
  TLS_def5.stop();

  TLS_def_all.stop();
  TLevelSolver_define.stop();
}

// complete constructor
LevelSolver::LevelSolver(const DisjointBoxLayout& a_grids,
                         const DisjointBoxLayout* a_baseGrids,
                         const Box&               a_domain,
                         Real                     a_dxLevel,
                         int                      a_nRefCrse,
                         const LevelOp* const     a_opin,
                         int                      a_ncomp,
                         bool                     a_limitCoarsening)
{
  setDefaultValues();

  ProblemDomain physdomain(a_domain);

  define(a_grids, a_baseGrids, physdomain, a_dxLevel,
         a_nRefCrse, a_opin, a_ncomp, a_limitCoarsening);
}

// complete constructor
LevelSolver::LevelSolver(const DisjointBoxLayout& a_grids,
                         const DisjointBoxLayout* a_baseGrids,
                         const ProblemDomain&     a_domain,
                         Real                     a_dxLevel,
                         int                      a_nRefCrse,
                         const LevelOp* const     a_opin,
                         int                      a_ncomp,
                         bool                     a_limitCoarsening)
{
  setDefaultValues();

  define(a_grids, a_baseGrids, a_domain, a_dxLevel,
         a_nRefCrse, a_opin, a_ncomp, a_limitCoarsening);
}

// returns LevelSolver to a basically undefined state
void LevelSolver::clearMemory()
{
  if (m_levelOpPtr != NULL)
    {
      delete m_levelOpPtr;
    }

  m_levelOpPtr = NULL;
}

// solve on just this level, inhomogeneous bcs
void LevelSolver::levelSolve(LevelData<FArrayBox>&       a_phi,
                             const LevelData<FArrayBox>* a_phic,
                             const LevelData<FArrayBox>& a_rhs,
                             bool                        a_initializePhiToZero)
{
  TLS_levelSolve.start();
  Timer LS_timer; //temp guy
  LS_timer.start();
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /* compute initial residual
     -- can't use residual() fn because
     don't want to zero out higher levels
     -- also note that we use LNf,
     because this is a LEVEL solve
  */

  assert(m_levelOpPtr->isDefined());
  TLS_ditOperations.start();
  DataIterator dit = a_phi.dataIterator();

  if (a_initializePhiToZero)
    {
      //initial guess of phi is zero
      for (dit.reset(); dit.ok(); ++dit)
        {
          a_phi[dit()].setVal(0.0);
        }
    }
  TLS_ditOperations.stop();

  TLS_applyOpI.start();
  m_levelOpPtr->applyOpI(a_phi, a_phic, m_resid);
  TLS_applyOpI.stop();

  TLS_ditOperations.start();
  for (dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();

      m_corr[dit()].setVal(0.0);
    }
  TLS_ditOperations.stop();

  Vector<Real> initRes(m_resid.nComp()); // chombo no have this
  Vector<Real> oldRes(m_resid.nComp());

  bool done = true;

  for (int comp = 0; comp < m_resid.nComp(); comp++)
    {
      Interval thisInterval(comp,comp);
      TLS_norm.start();
      // unscaled norm
      oldRes[comp] = norm(m_resid, thisInterval, m_normType);
      TLS_norm.stop();
      initRes[comp] = oldRes[comp]; // chombo doesn't have an initRes[]  noel

      // do scaling
      if (m_normType != 0)
        {
          Real exponent = SpaceDim;
          exponent /= m_normType;
          Real scale = pow(m_dxLevel, exponent);
          oldRes[comp] *= scale;
        }

      // if initRes = 0, don't bother solving
      if (oldRes[comp] > 0.01*m_tolerance)
        {
          done=false;
        }
    }

  // if this hasn't already been set,
  // compute norm(RHS) for use as convergence metrics
  if (m_convergenceMetrics.size() == 0)
    {
      m_convergenceMetrics = Vector<Real>(a_rhs.nComp(),0);

      for (int comp=0; comp<a_rhs.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);
          // unscaled norm
          m_convergenceMetrics[comp] = norm(a_rhs, thisInterval, m_normType);

          // do scaling if necessary
          if (m_normType != 0)
            {
              Real exponent = SpaceDim;
              exponent /= m_normType;
              Real scale = pow(m_dxLevel, exponent);
              m_convergenceMetrics[comp] *= scale;
            }

          // if norm(rhs) is 0, replace with norm(residual)
          if (m_convergenceMetrics[comp] == 0.0)
            {
              m_convergenceMetrics[comp] = oldRes[comp];
            }
          m_levelMG.setConvergenceMetric(m_convergenceMetrics[comp],
                                         comp);
        }
    } // end if we are computing norm(RHS) as a convergence metric

#if 0
  char temp17[193];
  sprintf(temp17, " LevelSolve init res=%10.4e m_tolerance=%10.4e m_convergenceMetric=%10.4e\n",
          oldRes[0], m_tolerance, m_convergenceMetrics[0]);
  pout() << temp17;
#endif

  //if (done)
  //  {
  //    return;
  //  }

  Real currentRes = 0.0;
  int iter = 0;
  while (!done)
    {
      Timer LSiter;
      LSiter.start();

      TLS_mgRelax.start();
      TLMG_mgRelax_all.start();
      m_levelMG.mgRelax(m_corr, m_resid, true);
      TLMG_mgRelax_all.stop();
      TLS_mgRelax.stop();

      TLS_applyOpH.start();
      m_levelOpPtr->applyOpH(m_corr, m_scratch);
      TLS_applyOpH.stop();

      TLS_ditOperations.start();
      for (dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }
      TLS_ditOperations.stop();

      iter++;

      done = true;
      for (int comp=0; comp<m_resid.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);

          TLS_norm.start();
          // unscaled norm
          currentRes = norm(m_scratch, thisInterval, m_normType);
          TLS_norm.stop();

          // do scaling if necessary
          if (m_normType != 0)
            {
              Real exponent = SpaceDim;
              exponent /= m_normType;
              Real scale = pow(m_dxLevel, exponent);
              currentRes *= scale;
            }

          if (m_verbose)
            {
              //ndk///pout() << "LevelSolve iteration # " << iter
              //ndk///     << " Max(res) = " << currentRes
              //ndk///     << endl;
            }

          // DFM (9/20/02) third test is to catch case where
          // convergence hangs due to machine precision (or
          // solvability) issues.  This follows the approach
          // used in AMRSolver
          //done = done && ((iter > m_maxIter)
          //                || (currentRes <= m_tolerance*initRes[comp] )
          //                || ((iter > m_minIter)
          //                    && (currentRes
          //                        > (1.0-m_operatorTolerance)*oldRes[comp])));

          // DFM (1/16/04) remove hang test (same as in AMRSolver),
          // since it's not clear we really want to do it this way
          done = done && ((iter > m_maxIter)
                          || (currentRes <= m_tolerance*m_convergenceMetrics[comp]));

          // save currentRes into oldRes
          oldRes[comp] = currentRes;
        } // end loop over components

      LSiter.stop();

#if 0
      for (int comp=0; comp<m_resid.nComp(); comp++) {
        char temp7[123];
        sprintf(temp7, " LevelSolve Iter#%2d currentRes=%10.4e goal=%10.4e wc=%6.3f\n",
                iter, currentRes,
                m_tolerance*m_convergenceMetrics[comp], LSiter.wc_time());
        pout() << temp7;
      }
#endif
    }  // end solver iterations [end while (!done)]

  // update phi
  TLS_ditOperations.start();
  for (dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }
  TLS_ditOperations.stop();

  // reset convergence metrics on exit to prevent re-using them
  // if the solver is being re-used
  resetConvergenceMetrics();

  // temp print
  //for (int comp=0; comp<m_scratch.nComp(); comp++)
  //  {
  //  if (oldRes[comp] > m_tolerance*initRes[comp])
  //    pout() << "LevelSolve() Component " << comp << " not converged! Initial Res = "
  //           << initRes[comp] << ", final Res = " << oldRes[comp] << endl;
  //}

  TLS_levelSolve.stop();
  LS_timer.stop();

#if 0
  char topi[190];
  sprintf(topi, " endLS iterations:%2d m_maxIter=%2d currentRes=%10.4e wc=%6.3f\n",
          iter, m_maxIter, currentRes, LS_timer.wc_time());
  pout() << topi;
#endif

  NoelsSooperIterationCounter[IterationFlavor].Count++;
  NoelsSooperIterationCounter[IterationFlavor].Value+=iter;
  if (iter > m_maxIter) {
    NoelsSooperIterationCounter[IterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[IterationFlavor].wc += LS_timer.wc_time();
#ifdef DEBUG_COUNTER
  pout() << " LevelSolver::levelSolve IterationFlavor=" << IterationFlavor
         << " n=" << NoelsSooperIterationCounter[IterationFlavor].Count
         << " wc=" << LS_timer.wc_time()
         << endl;
#endif
  IterationFlavor = 0;
}

// solve on just this level, homogeneous bcs
void LevelSolver::levelSolveH(LevelData<FArrayBox>&       a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              bool                        a_initializePhiToZero)
{
  Timer LS_timer;
  LS_timer.start();
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /** compute initial residual
      -- can't use residual() fn because
      don't want to zero out higher levels
      -- also note that we use LNf,
      because this is a LEVEL solve
  **/

  assert(m_levelOpPtr->isDefined());

  DataIterator dit = a_phi.dataIterator();

  if (a_initializePhiToZero)
    {
      //initial guess of phi is zero
      for (dit.reset(); dit.ok(); ++dit)
        {
          a_phi[dit()].setVal(0.0);
        }
    }

  m_levelOpPtr->applyOpH(a_phi, m_resid);

  for (dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();
      m_corr[dit()].setVal(0.0);
    }

  Vector<Real> initRes(m_resid.nComp()); // chombo no have this
  Vector<Real> oldRes(m_resid.nComp());

  bool done = true;

  for (int comp = 0; comp < m_resid.nComp(); comp++)
    {
      Interval thisInterval(comp,comp);
      // unscaled norm...
      oldRes[comp] = norm(m_resid, thisInterval, m_normType);
      initRes[comp] = oldRes[comp]; // chombo doesn't have an initRes[]  noel

      // do scaling if necessary
      if (m_normType != 0)
        {
          Real exponent = SpaceDim;
          exponent /= m_normType;
          Real scale = pow(m_dxLevel, exponent);
          oldRes[comp] *= scale;
        }

      // if initRes = 0, don't bother solving
      if (oldRes[comp] > 0.01*m_tolerance)
        {
          done = false;
        }
    }

  // if this hasn't already been set,
  // compute norm(RHS) for use as convergence metrics
  if (m_convergenceMetrics.size() == 0)
    {
      m_convergenceMetrics = Vector<Real>(a_rhs.nComp(),0);

      for (int comp=0; comp<a_rhs.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);
          // unscaled norm
          m_convergenceMetrics[comp] = norm(a_rhs, thisInterval, m_normType);

          // do scaling if necessary
          if (m_normType != 0)
            {
              Real exponent = SpaceDim;
              exponent /= m_normType;
              Real scale = pow(m_dxLevel, exponent);
              m_convergenceMetrics[comp] *= scale;
            }

          // if norm(rhs) is 0, replace with norm(residual)
          if (m_convergenceMetrics[comp] == 0.0)
            {
              m_convergenceMetrics[comp] = oldRes[comp];
            }
          m_levelMG.setConvergenceMetric(oldRes[comp], comp);
        }
    } // end if we are computing norm(RHS) as a convergence metric

  //if (done)
  //  {
  //    return;
  //  }

  Real currentRes = 0.0;
  int iter = 0;
  while (!done)
    {
      iter++;

      // compute modified residual LofPhi = res - LNf(corr)
      // use homogeneous form of ApplyOp because want to use
      // homogeneous CF BC's for correction
      TLMG_mgRelax_all.start();
      m_levelMG.mgRelax(m_corr, m_resid, true);
      TLMG_mgRelax_all.stop();

      m_levelOpPtr->applyOpH(m_corr, m_scratch);

      for (dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }

      done = true;
      for (int comp = 0; comp < m_scratch.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);
          // unscaled norm
          currentRes = norm(m_scratch, thisInterval, m_normType);

          // do scaling if necessary
          if (m_normType != 0)
            {
              Real exponent = SpaceDim;
              exponent /= m_normType;
              Real scale = pow(m_dxLevel, exponent);
              currentRes *= scale;
            }

          if (m_verbose)
            {
              pout() << "LevelSolveH iteration # " << iter
                     << ": Max(res) = " << currentRes
                     << endl;
            }

          // DFM (9/20/02) third test is to catch case where
          // convergence hangs due to machine precision (or
          // solvability) issues.  This follows the approach
          // used in AMRSolver
          //done = done && ((iter > m_maxIter)
          //                || (currentRes <= m_tolerance*initRes[comp] )
          //                    || ((iter < m_minIter)
          //                        && (currentRes
          //                           > (1.0-m_operatorTolerance)*oldRes[comp])));

          // DFM (1/16/04) remove hang test because it's not clear that we want
          // to do it this way...
          done = done && ((iter > m_maxIter)
                          || (currentRes <= m_tolerance*m_convergenceMetrics[comp] ));

          oldRes[comp] = currentRes;
        } // end loop over components
    }  // end solver iterations

  // update phi
  for (dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }

#if 0
  // temporary diagnostic prints
  char ttt[100];
  sprintf(ttt, "levelSolveH iterations: %d  max=%d  currentRes=%12.5e\n",
          iter, m_maxIter, currentRes);
  pout() << ttt;

  for (int comp=0; comp<m_scratch.nComp(); comp++)
    {
      if (currentRes > m_tolerance*m_convergenceMetrics[comp])
        pout() << " levelSolveH component " << comp
               << " not converged! Initial Res = " << m_convergenceMetrics[comp]
               << ", final Res = " << currentRes << endl;
    }
#endif

  LS_timer.stop();
  NoelsSooperIterationCounter[IterationFlavor].Count++;
  NoelsSooperIterationCounter[IterationFlavor].Value+=iter;
  if (iter > m_maxIter) {
    NoelsSooperIterationCounter[IterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[IterationFlavor].wc += LS_timer.wc_time();
#ifdef DEBUG_COUNTER
  pout() << " LevelSolver::levelSolveH IterationFlavor=" << IterationFlavor
         << " n=" << NoelsSooperIterationCounter[IterationFlavor].Count
         << " wc=" << LS_timer.wc_time()
         << endl;
#endif
  IterationFlavor = 0;

}
