/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BaseFab.H"
#include "REAL.H"
#include "DataIterator.H"
#include "LevelMGF_F.H"

#include "MGInterp.H"

#include "Timer.H"
extern Timer Everything;
Timer TF_INTERPMG("INTERPMG", Everything, DTIMEFORT, true);

Timer TMPI_MGInterp_interpToFine("interpToF cp", Everything, DTIMECOMM, true);

MGInterp::MGInterp()
  :m_isDefined(false)
{
}

MGInterp::~MGInterp()
{
}

MGInterp::MGInterp(const DisjointBoxLayout& a_fineDomain,
                   int                      a_numcomps,
                   int                      a_refRatio,
                   const Box&               a_problemDomain)
  :m_isDefined(false)
{
  ProblemDomain finePhysdomain(a_problemDomain);

  define(a_fineDomain, a_numcomps, a_refRatio, finePhysdomain);
}

MGInterp::MGInterp(const DisjointBoxLayout& a_fineDomain,
                   int                      a_numcomps,
                   int                      a_refRatio,
                   const ProblemDomain&     a_problemDomain)
  :m_isDefined(false)
{
  define(a_fineDomain, a_numcomps, a_refRatio, a_problemDomain);
}

void MGInterp::define(const DisjointBoxLayout& a_fineDomain,
                      int                      a_numcomps,
                      int                      a_refRatio,
                      const Box&               a_problemDomain)
{
  ProblemDomain finePhysdomain(a_problemDomain);

  define(a_fineDomain, a_numcomps, a_refRatio, finePhysdomain);
}

void MGInterp::define(const DisjointBoxLayout& a_fineDomain,
                      int                      a_numcomps,
                      int                      a_refRatio,
                      const ProblemDomain&     a_problemDomain)
{
  m_refRatio = a_refRatio;
  m_problemDomain = a_problemDomain;
  m_grids = a_fineDomain;

  // create the work array
  DisjointBoxLayout coarsenedFineDomain;

  coarsen(coarsenedFineDomain, a_fineDomain, m_refRatio);

  m_coarsenedFineData.define(coarsenedFineDomain,
                             a_numcomps,
                             IntVect::TheZeroVector());

  m_isDefined = true;
}

// interpolate from coarse level to fine level
void MGInterp::interpToFine(LevelData<FArrayBox>&       a_fineData,
                            const LevelData<FArrayBox>& a_coarseData)
{
  assert(m_isDefined);
  TMPI_MGInterp_interpToFine.start();
  a_coarseData.copyTo(a_coarseData.interval(),
                      m_coarsenedFineData,
                      m_coarsenedFineData.interval());
  TMPI_MGInterp_interpToFine.stop();

  DataIterator dit = m_grids.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& fine = a_fineData[dit()];
      const FArrayBox& coarsenedFine = m_coarsenedFineData[dit()];

      Box crseBox = coarsenedFine.box();
      Box fineBox = m_grids.get(dit());

      fineBox.coarsen(m_refRatio);

      assert(fineBox == crseBox);

      Box nrefbox(IntVect::TheZeroVector(),
                  (m_refRatio-1)*IntVect::TheUnitVector());

      TF_INTERPMG.start();
      FORT_INTERPMG(CHF_FRA(fine),
                    CHF_FRA(coarsenedFine),
                    CHF_BOX(crseBox),
                    CHF_CONST_INT(m_refRatio),
                    CHF_BOX(nrefbox));
      TF_INTERPMG.stop();
    }
}

bool MGInterp::isDefined() const
{
  return m_isDefined;
}
