/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "Box.H"
#include "FArrayBox.H"
#include "REAL.H"
#include "SPACE.H"
#include "Tuple.H"
#include "Vector.H"
#include "LoHiSide.H"

#include "PoissonBC.H"
#include "MayDay.H"

NeumannBC::NeumannBC() : BoxGhostBC(), m_inhomVal(0.0)
{
}

NeumannBC::~NeumannBC()
{
}
void NeumannBC::setInhomVal(Real a_inhomVal)
{
  m_inhomVal = a_inhomVal;
}
void DirichletBC::setInhomVal(Real a_inhomVal)
{
  m_inhomVal = a_inhomVal;
}
void NeumannBC::fillBCValues(FArrayBox& a_neumfac,
                             FArrayBox& a_dircfac,
                             FArrayBox& a_inhmval,
                             Real       a_dx,
                             const Box& a_domain) const
{
  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());

  a_neumfac.setVal(1.0);
  a_dircfac.setVal(0.0);
  a_inhmval.setVal(m_inhomVal);
}

void NeumannBC::fillBCValues(FArrayBox&           a_neumfac,
                             FArrayBox&           a_dircfac,
                             FArrayBox&           a_inhmval,
                             Real                 a_dx,
                             const ProblemDomain& a_domain) const
{
  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());

  a_neumfac.setVal(1.0);
  a_dircfac.setVal(0.0);
  a_inhmval.setVal(m_inhomVal);
}

void DirichletBC::fillBCValues(FArrayBox& a_neumfac,
                               FArrayBox& a_dircfac,
                               FArrayBox& a_inhmval,
                               Real       a_dx,
                               const Box& a_domain) const
{
  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());

  a_neumfac.setVal(0.0);
  a_dircfac.setVal(1.0);
  a_inhmval.setVal(m_inhomVal);
}

void DirichletBC::fillBCValues(FArrayBox&           a_neumfac,
                               FArrayBox&           a_dircfac,
                               FArrayBox&           a_inhmval,
                               Real                 a_dx,
                               const ProblemDomain& a_domain) const
{
  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());

  a_neumfac.setVal(0.0);
  a_dircfac.setVal(1.0);
  a_inhmval.setVal(m_inhomVal);
}

BoxGhostBC* NeumannBC::new_boxghostbc() const
{
  NeumannBC* newop = new NeumannBC();

  if (newop == NULL)
    {
      MayDay::Error("Out of Memory in NeumannBC::new_boxghostbc");
    }
  newop->setInhomVal(m_inhomVal);
  return static_cast<BoxGhostBC*>(newop);
}

BoxGhostBC* DirichletBC::new_boxghostbc() const
{
  DirichletBC* newop = new DirichletBC();

  if (newop == NULL)
    {
      MayDay::Error("Out of Memory in DirichletBC::new_boxghostbc");
    }

  newop->setInhomVal(m_inhomVal);
  return static_cast<BoxGhostBC*>(newop);
}

NeumannBC::NeumannBC(int             a_dir,
                     Side::LoHiSide  a_side,
                     const Interval& a_comps)
  :BoxGhostBC(a_dir,a_side,a_comps),m_inhomVal(0.0)
{
}

NeumannBC::NeumannBC(int a_dir,
                     Side::LoHiSide a_side)
  :BoxGhostBC(a_dir,a_side), m_inhomVal(0.0)
{
}

DirichletBC::DirichletBC()
  :BoxGhostBC(), m_inhomVal(0.0)
{
}

DirichletBC::~DirichletBC()
{
}

DirichletBC::DirichletBC(int dir, Side::LoHiSide side)
  :BoxGhostBC(dir,side), m_inhomVal(0.0)
{
}

DirichletBC::DirichletBC(int             a_dir,
                         Side::LoHiSide  a_side,
                         const Interval& a_comps)
  :BoxGhostBC(a_dir,a_side,a_comps), m_inhomVal(0.0)
{
}
