/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Weds, July 21, 1999

#ifndef _POISSONOP_H_
#define _POISSONOP_H_

#include <iostream>
#include <cmath>
#include <cassert>
#include <cstdlib>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BaseFab.H"
#include "FArrayBox.H"
#include "QuadCFInterp.H"
#include "Copier.H"

#include "GhostBC.H"
#include "LevelOp.H"
#include "CFIVS.H"
#include "BaseBottomSmoother.H"

///  one-sided differencing version of poisson operator
/** this class contains stuff to solve  div (grad phi) = rhs
 */
class PoissonOp: public LevelOp
{
public:
  /// Default constructor
  PoissonOp();

  /// Destructor
  ~PoissonOp();

  /// Virtual constructor workaround
  /**
  */
  virtual LevelOp* new_levelop() const;

  ///
  /**
  */
  void levelGSMC(LevelData<FArrayBox>&       a_phi,
                 const LevelData<FArrayBox>& a_rhs);

  /// Has this object been defined?
  virtual bool isDefined() const;

  /// Full define function.
  /**
      Makes all coarse-fine information and sets internal variables
  */
  void define(const DisjointBoxLayout& a_grids,
              const DisjointBoxLayout* a_baseBAPtr,
              Real                     a_dxLevel,
              int                      a_refRatio,
              const ProblemDomain&     a_domain,
              bool                     a_homogeneousOnly=false,
              int                      a_ncomp = 1);

  // my special define for bottom smoother test (noel)
  void define(const DisjointBoxLayout& a_grids,
              Real                     a_dxLevel,
              Real                     a_dxCoarse,
              const ProblemDomain&     a_domain,
              int                      a_ncomp = 1);

  /// Full define function. Define from finer levelop.
  /**
  */
  void define(const LevelOp* a_opfine,
              int            a_reftoFine);

  /// Full define function (Deprecated -- use ProblemDomain version)
  /**
      Full define function.  makes all coarse-fine information and sets
      internal variables
  */
  void define(const DisjointBoxLayout& a_grids,
              const DisjointBoxLayout* a_baseBAPtr,
              Real                     a_dxLevel,
              int                      a_refRatio,
              const Box&               a_domain,
              bool                     a_homogeneousOnly=false,
              int                      a_ncomp = 1);

  /// Smoother.
  /**
      Assumes that problem has already been put in residual correction form,
      so that C/F boundary conditions are homogeneous.
  */
  void smooth(LevelData<FArrayBox>&       a_phi,
              const LevelData<FArrayBox>& a_rhs);

  /// Apply preconditioning to the solution on this level.
  /**
      Given rhshat, returns phihat from M(phihat) = rhshat
  */
  virtual void levelPreconditioner(LevelData<FArrayBox>&       a_phihat,
                                   const LevelData<FArrayBox>& a_rhshat);

  /// Evaluate Operator, inhomogeneous boundary conditions.
  /**
      Uses inhomogeneous C/F boundary conditions.
      and inhomogeneous phys boundary conditions
  */
  void applyOpI(LevelData<FArrayBox>&       a_phi,
                const LevelData<FArrayBox>* a_phicPtr,
                LevelData<FArrayBox>&       a_lofPhi);

  /// Evaluate Operator, mixed boundary conditions.
  /**
      Uses inhomogeneous form of C/F boundary conditions.
      and homogeneous physical boundary conditions
  */
  void applyOpIcfHphys(LevelData<FArrayBox>&       a_phi,
                       const LevelData<FArrayBox>* a_phicPtr,
                       LevelData<FArrayBox>&       a_lofPhi);

  /// Evaluate Operator using homogeneous boundary conditions.
  /**
      Uses homogeneous form of both coarse-fine and physical  boundary
      conditions.
  */
  void applyOpH(LevelData<FArrayBox>& a_phi,
                LevelData<FArrayBox>& a_lofPhi);

  /// Coarse / Fine interpolation operator.
  /**
  */
  void CFInterp(LevelData<FArrayBox>&       a_phif,
                const LevelData<FArrayBox>& a_phic);

  /// Evaluate Operator
  /**
      Uses homogeneous form of C/F boundary conditions.
      with inhomogeneous physical boundary conditions
  */
  void applyOpHcfIphys(LevelData<FArrayBox>& a_phi,
                       LevelData<FArrayBox>& a_lOfPhi);

  /// Homogeneous coarse/fine interpolation operator
  /**
  */
  void homogeneousCFInterp(LevelData<FArrayBox>& a_phif);

  /// Set physical boundary condition types
  void setDomainGhostBC(const DomainGhostBC& a_dombcin);

  /// Set the bottom smoother to use in solve
  virtual void setBottomSmoother(const BaseBottomSmoother& a_bottomSmoother);

  /// Does homogeneous coarse/fine interpolation
  /**
  */
  void
  homogeneousCFInterp(LevelData<FArrayBox>& a_phif,
                      const DataIndex&      a_datInd,
                      int                   a_idir,
                      Side::LoHiSide        a_hiorlo);

  /// This performs the smoothing at the bottom of a multigrid V-cycle.
  /**
  */
  void bottomSmoother(LevelData<FArrayBox>&       a_phi,
                      const LevelData<FArrayBox>& a_rhs);

  /// Get flux on this level in the direction dir on the given Box
  /**
      The fluxes live on the cell faces with direction a_dir.
      Fluxes are computed for all interior edges of a_data.
      The flux fab is resized inside the routine.
  */
  void getFlux(FArrayBox&       a_fineFlux,
               const FArrayBox& a_data,
               const DataIndex& a_datInd,
               int              a_dir);

  /// pass-through to bottom smoother
  virtual void setConvergenceMetric(Real a_metric, int a_comp);

protected:
  /**
     Boundary conditons
  */
  DomainGhostBC m_domghostbc;

  /**
     Domain of fine grid
  */
  ProblemDomain m_domain;

  /**
     The grids at the current level
  */
  DisjointBoxLayout m_grids;

  /**
     Copier object that knows how to perform exchange operations on
     LevelData objects defined on m_grids
  */
  Copier m_exchangeCopier;

  /**
     DisjointBoxLayout for next coarser level
     this is == NULL if there is no coarser level
  */
  DisjointBoxLayout m_baseBA;

  /**
     Number of components
  */
  int m_ncomp;

  /**
     Refinement ratio between this and the next coarser level
  */
  int m_refRatio;

  /**
     Mesh spacing at this level
  */
  Real m_dxLevel;

  /**
     Mesh spacing at next coarser level
  */
  Real m_dxCrse;

  /*
    Object which performs Coarse-fine interpolation between this level
    and the next coarser level
  */
  QuadCFInterp m_quadCFI;

  /**
     Bottom smoother object
  */
  BaseBottomSmoother* m_bottomSmootherPtr;

  /**
     Has full define function been called?
  */
  bool m_isDefined;

  /**
     Has bc been defined?
  */
  bool m_isBCDefined;

  /**
     Is inhomogeneous CF interpolation possible?
  */
  bool m_ihcfiEnabled;

  /**
     Coarse-fine stencils for homogeneous CFInterp
  */
  LayoutData<CFIVS> m_loCFIVS[SpaceDim];
  /**
     Coarse-fine stencils for homogeneous CFInterp
  */
  LayoutData<CFIVS> m_hiCFIVS[SpaceDim];

private:
  // Internally useful functions
  void levelGSRB(LevelData<FArrayBox>&       a_phi,
                 const LevelData<FArrayBox>& a_rhs);

//   void interpOnIVS(LevelData<FArrayBox>& a_phif,
//                    const FArrayBox&      a_phistar,
//                    const DataIndex&      a_datInd,
//                    const int             a_idir,
//                    const Side::LoHiSide  a_hiorlo,
//                    const IntVectSet&     a_interpIVS);

  void interpOnIVSHomo(LevelData<FArrayBox>& a_phif,
                       const DataIndex&      a_datInd,
                       const int             a_idir,
                       const Side::LoHiSide  a_hiorlo,
                       const IntVectSet&     a_interpIVS);

  void clearMemory();

  void setDefaultValues();
};

#endif
