/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// #define USE_SUPERLU
#ifdef USE_SUPERLU

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
#include "CFIVS.H"

#include "SuperLUSmoother.H"

#include <cstdio>  // for debugging sprintfs

#include "Timer.H"
#include "NoelsPrettyPrint.H"
#include "superlu_ddefs.h"

#include "FORT_PROTO.H"
#include "HABECF_F.H"

#define STENCILSIZE (2*CH_SPACEDIM + 1)

struct point {
  int i;
  int j;
  int k;
};

void findNeighbor(const vector<point>& po,
                  const int findi,
                  const int findj,
                  int& answerdood) {

  answerdood = -1;
  for(int i=0; i<po.size(); i++) {
    if (po[i].i == findi && po[i].j == findj) {
      answerdood = i;
      return;
    }
  }
}

SuperLUSmoother::SuperLUSmoother()
{

  m_verbose = false;
}

SuperLUSmoother::SuperLUSmoother(const Real a_dxLevel,
                                 const Real a_dxCoarse,
                                 const ProblemDomain& a_domain)
{
  m_verbose = false;
  m_domain = a_domain;
  m_dxCoarse = a_dxCoarse;
  m_dxLevel = a_dxLevel;

}

SuperLUSmoother::~SuperLUSmoother()
{

}

void SuperLUSmoother::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
}

void SuperLUSmoother::setTolerance(Real a_tolerance)
{
  m_tolerance = a_tolerance;
}

void SuperLUSmoother::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}

BaseBottomSmoother* SuperLUSmoother::new_bottomSmoother() const
{
  SuperLUSmoother* newsmoother = new SuperLUSmoother();
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL)
    {
      MayDay::Error("Out of Memory in SuperLUSmoother::new_bottomSmoother");
    }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

/***********************/
void SuperLUSmoother::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                                     const LevelData<FArrayBox>& a_rhs,
                                     LevelOp*                    a_levelopPtr)
{
  pout() << " SuperLUSmoother::doBottomSmooth" << endl;

  //if (m_initializedSuperLU == false) {
  //pout() << " initialize superlu solver" << endl;
  //setup_solver(a_phi.getBoxes(), m_domain, m_dxLevel, m_dxCoarse);
  //m_initializedSuperLU = true;
  //}

  m_grids = a_phi.getBoxes();

  // DataIterator gets data in this proc only.
  DataIterator dit = m_grids.dataIterator();

#if (CH_SPACEDIM == 2)
  // 5-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1
  //   4
  //int offsets[5][2] = {{ 0,  0},
  //                   {+1,  0},
  //                   {-1,  0},
  //                   { 0, +1},
  //                   { 0, -1}};
#elif (CH_SPACEDIM == 3)
  // 7-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1   (and:  5 above 0, 6 below 0)
  //   4
  //int offsets[7][3] = {{ 0,  0,  0},
  //                   {+1,  0,  0},
  //                   {-1,  0,  0},
  //                   { 0, +1,  0},
  //                   { 0, -1,  0},
  //                   { 0,  0, +1},
  //                   { 0,  0, -1}};
#endif

  const int ncomp=1;
  const int ngrow=0;
  IntVect ghostVect = ngrow * IntVect::Unit;
  m_acoefs = new LevelData<FArrayBox>(a_rhs.getBoxes(), ncomp, ghostVect);
  m_bcoefs = new LevelData<FluxBox>  (a_rhs.getBoxes(), ncomp, ghostVect);

  m_alpha = 0.; // weight of identity
  m_beta  = 1.; // weight of div grad

  for (dit.begin(); dit.ok(); ++dit)
    {
      m_acoefs->operator[](dit()).setVal(0.);
      m_bcoefs->operator[](dit()).setVal(1.);
    }

  int stencil_indices[STENCILSIZE];

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      stencil_indices[icomp] = icomp;
    }

  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  printf("m_alpha=%10.4f m_beta=%10.4f m_tolerance=%10.4e m_dxLevel=%10.4f m_dxCoarse=%10.4f\n",
         m_alpha, m_beta, m_tolerance, m_dxLevel, m_dxCoarse);

  int Nboxes, Npoints, Nghosts;
  NoelsLDFABgetUsefulInfo(m_acoefs, Nboxes, Npoints, Nghosts);
  cout << " Nboxes=" << Nboxes << " Npoints=" << Npoints << endl;

  // store
  vector<point> pointString(Npoints);
  int Np=0;
  int Nb=0;
  for (DataIterator dit = m_acoefs->dataIterator(); dit.ok(); ++dit, ++Nb) {
    const Box& bx = m_grids.get(dit());
    //dit.get(bit(), 0)
    //cout << " box N" << Nb << " bx=" << bx << endl;
    for(BoxIterator bit(bx); bit.ok(); ++bit, ++Np) {
      //cout << "point" << bit() << Np << endl;
      pointString[Np].i = bit().getVect()[0];
      pointString[Np].j = bit().getVect()[1];
#if (CH_SPACEDIM==3)
      pointString[Np].k = bit().getVect()[2];
#endif
    }
  }

  vector<vector<double> > fullA;
  fullA.resize(Npoints);
  for(int i=0; i<Npoints; i++) fullA[i].resize(Npoints);

  int totalNcellsModifiedAtCF = 0;

  Nboxes=0;
  Np=0;
  for (DataIterator dit = m_acoefs->dataIterator(); dit.ok(); ++dit, ++Nboxes)
    {
      const Box& bx = m_grids.get(dit());
      cout << " box N" << Nboxes << " bx=" << bx << endl;

      FArrayBox matFab(bx, STENCILSIZE);

      //build matrix interior
      // Operator on phi(:) is:
      // L(phi)(:) = alpha * a(:) * phi(:) + beta * div (b(:) * grad phi(:))

      // Set coefficients of first term, alpha * a(:) * phi(:).
      // currently, m_alpha is 0 (ndk)
      if (m_alpha == 0.)
        {
          matFab.setVal(0., 0); // set component 0
        }
      else
        {
          FArrayBox& acoefsFab = m_acoefs->operator[](dit());
          matFab.copy(acoefsFab, 0, 0, 1); // srccomp, destcomp, numcomp
          matFab.mult(m_alpha, 0, 1); // startcomp, numcomp
        }
      //        FORT_HACOEF(CHF_FRA(mat),
      //                    CHF_CONST_FRA1(acoefsFab, 0),
      //                    CHF_CONST_REAL(m_alpha));

      // Set coefficients of second term, beta * div (b(:) * grad phi(:)).
      // for now, b (m_bcoefs) are all 1  (ndk)
      for ( int idim = 0; idim < SpaceDim; idim++ )
    {
          const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];
          FORT_HBCOEF(CHF_FRA(matFab),
                      CHF_CONST_FRA1(bfab, 0),
                      CHF_CONST_REAL(m_beta),
                      CHF_CONST_REAL(m_dxLevel),
                      CHF_CONST_INT(idim));
    }

      // add boundary conditions to matrix diagonal, and
      //   zero out offdiag values at domain boundaries.

      SideIterator sit;
      for (int idim = 0; idim < SpaceDim; idim++) {
        // if ( ! m_domain.isPeriodic(idim) ) // hypre has probs with periodic
        for (sit.begin(); sit.ok(); sit.next())
          {
            Box region(bx);
            //cout << " idim=" << idim << " sit=" << sit() << endl;

            int iside;
            bool intersectsBoundary = false;
            switch (sit())
              {
              case Side::Lo:
                {
                  int smallCoordinate = region.smallEnd(idim);
                  if (smallCoordinate == loDomain[idim])
                    {
                      intersectsBoundary = true;
                      region.setBig(idim, smallCoordinate);
                      iside = -1;
                    }
                  break;
                }
              case Side::Hi:
                {
                  int bigCoordinate = region.bigEnd(idim);
                  if (bigCoordinate == hiDomain[idim])
                    {
                      intersectsBoundary = true;
                      region.setSmall(idim, bigCoordinate);
                      iside = +1;
                    }
                  break;
                }
              default:
                cerr << "SideIterator error" << endl;
                abort();
              }

            if (intersectsBoundary)
              {
                cout << " intersectsBoundary dim=" << idim << " iside=" << iside << endl;
                const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];

                // 0 for dirichlet
                // 1 for neumann
                const int boundaryCondition=1;
                FORT_HBMAT(CHF_FRA(matFab),
                           CHF_CONST_FRA1(bfab, 0),
                           CHF_BOX(region),
                           CHF_CONST_INT(idim),
                           CHF_CONST_INT(iside),
                           CHF_CONST_REAL(m_beta),
                           CHF_CONST_REAL(m_dxLevel),
                           CHF_CONST_INT(boundaryCondition));
              }

          } // end iteration over faces

      } // end side loop

      // NoelsPrettyPrint(&matFab);

      //   3
      // 2 0 1
      //   4
      //int offsets[5][2] = {{ 0,  0},  0
      //                     {+1,  0},  1
      //                     {-1,  0},  2
      //                     { 0, +1},  3
      //                     { 0, -1}}; 4
      //
      //   3
      // 2 0 1   (and:  5 above 0, 6 below 0)
      //   4
      //int offsets[7][3] = {{ 0,  0,  0},  0
      //                     {+1,  0,  0},  1
      //                     {-1,  0,  0},  2
      //                     { 0, +1,  0},  3
      //                     { 0, -1,  0},  4
      //                     { 0,  0, +1},  5
      //                     { 0,  0, -1}}; 6

      // adjust for coarse-fine boundaries
      int Ncells_modified=0;
      modifyMatrixAtCFBoundaries(bx, matFab, Ncells_modified);
      totalNcellsModifiedAtCF += Ncells_modified;

      //NoelsPrettyPrint(&matFab);

#if 0
      char tok[200];
      for(BoxIterator bit(bx); bit.ok(); ++bit) {
        sprintf(tok, "box%4d iv0:%3d iv1:%3d matfab: %4.0f %4.0f %4.0f %4.0f %4.0f\n",
                Nboxes, bit().getVect()[0], bit().getVect()[1],
                matFab.get(bit(), 0), matFab.get(bit(), 1),
                matFab.get(bit(), 2), matFab.get(bit(), 3),
                matFab.get(bit(), 4));
        cout << tok;
      }
#endif

      int leftdood, rightdood, updood, downdood, indood, outdood;
      //const Box& bx = m_grids.get(dit());
      //cout << " bbox N" << Nb << " bx=" << bx << endl;
      for(BoxIterator bit(bx); bit.ok(); ++bit, ++Np) {
        //printf(" finding i) %d %d %d  j)  %d %d %d\n",
        //     pointString[Np].i-1, pointString[Np].i, pointString[Np].i+1,
        //     pointString[Np].j-1, pointString[Np].j, pointString[Np].j+1);

        findNeighbor(pointString, pointString[Np].i-1, pointString[Np].j, leftdood);
        findNeighbor(pointString, pointString[Np].i+1, pointString[Np].j, rightdood);
        findNeighbor(pointString, pointString[Np].i, pointString[Np].j+1, updood);
        findNeighbor(pointString, pointString[Np].i, pointString[Np].j-1, downdood);
#if (CH_SPACEDIM==3)
        //don't work in 3d yet!
        findNeighbor(pointString, pointString[Np].i, pointString[Np].j+1, updood);
        findNeighbor(pointString, pointString[Np].i, pointString[Np].j-1, downdood);
#endif
        //cout << "point" << bit() << Np
        //   << " l" << leftdood
        //   << " r" << rightdood
        //   << " u" << updood
        //   << " d" << downdood
        //   << endl;
        for(int i=0; i<Npoints; i++) {
          fullA[Np][i] = 0.0;
        }
        fullA[Np][Np] = matFab.get(bit(), 0);
        if (rightdood >= 0) fullA[Np][rightdood] = matFab.get(bit(), 1);
        if (leftdood >= 0)  fullA[Np][leftdood]  = matFab.get(bit(), 2);
        if (updood >= 0)    fullA[Np][updood]    = matFab.get(bit(), 3);
        if (downdood >= 0)  fullA[Np][downdood]  = matFab.get(bit(), 4);
      }

    }
  //pout() << " 1 HypreSmoother::setup_matrix()" << m_impl->A << endl;

  int nnz=0;
  printf(" Npoints=%d\n", Npoints);
  for(int row=0; row<Npoints; row++) {
    if (Npoints < 128) printf(" row%2d", row);
    for(int col=0; col<Npoints; col++) {
      if (Npoints < 128) printf("%4d", (int)fullA[row][col]);
      if (fullA[row][col] != 0.0) { nnz++; }
    }
    if (Npoints < 128) printf("\n");
  }

  /////////////////////////////////////
  double* nzval;
  int *rowind, *colptr;

  nzval  = new double[nnz];
  rowind = new int[nnz];
  colptr = new int[Npoints + 1];

  int Nnonzero=0;
  colptr[0] = 0;
  for(int col=0; col<Npoints; col++) {
    for(int row=0; row<Npoints; row++) {
      if (fullA[row][col] != 0.0) {
        nzval[Nnonzero] = fullA[row][col];
        rowind[Nnonzero] = row;
        Nnonzero++;
      }
    }
    colptr[col+1] = Nnonzero;
  }
  printf(" nnz=%d  size of nzval=%d\n", nnz, Nnonzero);

  superlu_options_t options;
  SuperLUStat_t     stat;
  ScalePermstruct_t ScalePermstruct;
  LUstruct_t        LUstruct;
  //SOLVEstruct_t     SOLVEstruct;
  gridinfo_t        grid;

  int iam, info;
  int nprow = 1;  /* Default process rows.      */
  int npcol = 1;  /* Default process columns.   */
  //int nrhs = 1;   /* Number of right-hand side. */

  // INITIALIZE THE SUPERLU PROCESS GRID.
  superlu_gridinit(Chombo_MPI::comm, nprow, npcol, &grid);
  //superlu_gridinit(MPI_COMM_WORLD, nprow, npcol, &grid);

  /* Bail out if I do not belong in the grid. */
  iam = grid.iam;
  if ( iam >= nprow * npcol ) {
    exit(0);
    //goto out;
  }
  if ( !iam ) printf("\tProcess grid\t%d X %d\n", grid.nprow, grid.npcol);

  int Ncolumns = Npoints;
  int Nrows = Npoints;
  int Nrhs = 1;

  SuperMatrix superMatrixA;

  dCreate_CompCol_Matrix_dist(&superMatrixA, Nrows, Ncolumns, nnz, nzval,
                              rowind, colptr,
                              SLU_NC, SLU_D, SLU_GE);
  dPrint_CompCol_Matrix_dist(&superMatrixA);

  int leadingDim_b = Nrows;
  int leadingDim_x = Ncolumns;

  double *xtrue_global;
  xtrue_global = new double[Ncolumns * Nrhs];
  dGenXtrue_dist(Ncolumns, Nrhs, xtrue_global, leadingDim_x);

  double *b_global;
  b_global = new double[Nrows * Nrhs];
  char     trans[1];
  *trans = 'N';
  dFillRHS_dist(trans, Nrhs, xtrue_global, Ncolumns, &superMatrixA, b_global, leadingDim_b);

  double *berr;
  berr = new double[Nrhs];

  set_default_options(&options);

  /* Initialize ScalePermstruct and LUstruct. */
  ScalePermstructInit(Nrows, Ncolumns, &ScalePermstruct);
  LUstructInit       (Nrows, Ncolumns, &LUstruct);

  /* Initialize the statistics variables. */
  PStatInit(&stat);

  double *rhs;
  rhs = new double[Nrows];
  for (int ij=0; ij< Nrows; ij++) {
    rhs[ij] = 1.0;
  }

  //SuperMatrix myB;
  //dCreate_Dense_Matrix_dist(&myB, Nrows, Nrhs, rhs, Nrows, SLU_DN, SLU_D, SLU_GE);
  //dPrint_Dense_Matrix_dist(&myB);

  double *linear_rhs;
  linear_rhs = new double[Npoints];
  int pindex=0;
  for (DataIterator dit = a_rhs.dataIterator(); dit.ok(); ++dit) {
    const Box& bx = a_rhs.getBoxes().get(dit());
    for(BoxIterator bit(bx); bit.ok(); ++bit) {
      linear_rhs[pindex++] = (a_rhs)[dit()](bit());
      //cout << " (a_rhs)[dit()](bit()) = " << (a_rhs)[dit()](bit()) << endl;
    }
  }

  /* Call the linear equation solver. */
  cout << " Call the linear equation solver. " << endl;
  //pdgssvx(&options, &myA, &ScalePermstruct, rhs, myleadingDim_b, myNrhs, &grid,
  //      &LUstruct, &SOLVEstruct, berr, &stat, &info);
  pdgssvx_ABglobal(&options, &superMatrixA, &ScalePermstruct, linear_rhs, leadingDim_b,
                   Nrhs, &grid, &LUstruct, berr, &stat, &info);
  cout << " pdgssvx_ABglobal done. " << endl;

  //  const int ncomp=1;
  //const int ngrow=0;
  IntVect ghostVect2 = ngrow * IntVect::Unit;
  LevelData<FArrayBox> *Z_su;
  Z_su = new LevelData<FArrayBox>(m_grids, ncomp, ghostVect2);

  // copy back into a LevelData<FAB> so we can work with it...
  int jindex=0;
  for (DataIterator dit = Z_su->dataIterator(); dit.ok(); ++dit) {
    const Box& bx = m_grids.get(dit());
    for(BoxIterator bit(bx); bit.ok(); ++bit) {
      (*Z_su)[dit()](bit()) = linear_rhs[jindex++];
    }
  }

  if (Npoints < 256) {
    NoelsPrettyPrint(Z_su, "Z_su");
  }

  //for (int ij=0; ij< Nrows; ij++) {
  //printf("SU %d %10.4e\n", ij, rhs[ij]);
  //}

  /* Check the accuracy of the solution. */
  //if (!iam) dinf_norm_error_dist(Ncolumns, Nrhs, rhs, leadingDim_b,
  //                             xtrue_global, leadingDim_x, &grid);

  //PStatPrint(&options, &stat, &grid);        /* Print the statistics. */

  //DEALLOCATE STORAGE.
  //PStatFree(&stat);

  //   Destroy_CompCol_Matrix_dist(&superMatrixA);
  //ScalePermstructFree(&ScalePermstruct);
  //Destroy_LU(Nrows, &grid, &LUstruct);
  //LUstructFree(&LUstruct);

  //   if ( options.SolveInitialized ) {
  //     dSolveFinalize(&options, &SOLVEstruct);
  //   }

  delete[] rhs;
  delete[] linear_rhs;
  delete[] nzval;
  delete[] rowind;
  delete[] colptr;
  delete[] b_global;
  delete[] xtrue_global;
  delete[] berr;

  //RELEASE THE SUPERLU PROCESS GRID.
  superlu_gridexit(&grid);

}

void
SuperLUSmoother::modifyMatrixAtCFBoundaries(const Box bx,
                                            FArrayBox &matFab,
                                            int &Ncells_modified)
{
  // don't need to do this if dxCoarse or dxLevel is negative or 0
  // as that means we are on level 0 and should only have physical boundary interfaces
  // And the CFIVS object should return an empty set anyway.
  if (m_dxCoarse > 0.0 && m_dxLevel > 0.0) {
    const Real rr = m_dxCoarse/m_dxLevel;
    const Real Beta = 1.0 + (rr - 3.0)/(1.0 + rr);
    const Real Alpha = (1.0 - rr)/(3.0 + rr);
    //printf(" dxCoarse=%10.4f dxLevel=%10.4f rr=%10.4f Alpha=%10.4f Beta=%10.4f idxc2=%10.4f\n",
    //     m_dxCoarse, m_dxLevel[0], rr, Alpha, Beta, 1.0/(m_dxCoarse*m_dxCoarse));
    const Real dxLevel = m_dxLevel;
    SideIterator sit;

    // Find all CF interfaces and modify matFab for each.
    for (int idim = 0; idim < SpaceDim; idim++) {
      for (sit.begin(); sit.ok(); sit.next()) {
        Box region(bx);
        // if side is 0, then sidePole is +1.  if side is 1, then sidePole is -1.
        const int sidePole = -1*(sit()+sit()) + 1;
        // if side is 0, then arrow points to the +1 side, otherwise the -1 side
        const IntVect arrow = sidePole * BASISV(idim);
        // convert 0,1,or2 to 1,3,or5
        const int dir135 = 2*idim + 1;
        //cout << " idim=" << idim << " sit=" << sit() << endl;
        CFIVS cfivsObject(m_domain, region, m_grids, idim, sit());
        //cout << cfivsObject.isDefined() << endl;
        //cout << cfivsObject.getFineIVS() << endl;
        for(IVSIterator it(cfivsObject.getFineIVS()); it.ok(); ++it) {
          IntVect iv(it());
          iv.shift(arrow);
          //cout << "   " << it() << " --> " << iv
          //   <<  " " << matFab.get(iv, 0)
          //   <<  " " << matFab.get(iv, 1)
          //   <<  " " << matFab.get(iv, 2)
          //   <<  " " << matFab.get(iv, 3)
          //   <<  " " << matFab.get(iv, 4)
          //   <<  "   sit=" << sit() << " sp" << sidePole << " d" << dir135
          //   << endl;
          //void set(const IntVect& iv, int comp, Real val)
          //Real get(const IntVect& iv, int comp) const

          // this cell is modified
          matFab(iv, 0) += Beta/(dxLevel*dxLevel);

          if (sit() == 0) { // low side
            // one cell higher is modified
            matFab(iv, dir135)   += Alpha/(dxLevel*dxLevel);
            // one cell lower is zeroed
            matFab(iv, dir135+1)  = 0.0;
          } else { // high side
            // one cell higher is zeroed
            matFab(iv, dir135)    = 0.0;
            // one cell lower is modifed
            matFab(iv, dir135+1) += Alpha/(dxLevel*dxLevel);
          }

          Ncells_modified++;

          //cout << "ch " << it() << " --> " << iv
          //   <<  " " << matFab.get(iv, 0)
          //   <<  " " << matFab.get(iv, 1)
          //   <<  " " << matFab.get(iv, 2)
          //   <<  " " << matFab.get(iv, 3)
          //   <<  " " << matFab.get(iv, 4)
          //   << endl;
        }
      }
    }
  }

}

#endif  // on USE_SUPERLU
