/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "LoHiSide.H"

#include "PatchGodunov.H"
#include "PatchGodunovF_F.H"
#include "LoHiCenter.H"

#include "PatchGodunovF_F.H"

#include "Timer.H"
#include "TimerPrototypes.H"

// Flag everything as not defined or set
PatchGodunov::PatchGodunov()
{
  m_gdnvPhysics      = NULL;
  m_isDefined        = false;
  m_isCurrentTimeSet = false;
  m_isCurrentBoxSet  = false;
}

PatchGodunov::~PatchGodunov()
{
  // Delete the initial/boundary condition object - if it exists
  if (m_gdnvPhysics != NULL)
  {
    delete m_gdnvPhysics;
    m_gdnvPhysics  = NULL;
  }
}

// Define this object and the boundary condition object
void PatchGodunov::define(const ProblemDomain&           a_domain,
                          const Real&                    a_dx,
                          const GodunovPhysics* const    a_physPtr,
                          const int&                     a_normalPredOrder,
                          const bool&                    a_useFourthOrderSlopes,
                          const bool&                    a_usePrimLimiting,
                          const bool&                    a_useCharLimiting,
                          const bool&                    a_useFlattening,
                          const bool&                    a_useArtificialViscosity,
                          const Real&                    a_artificialViscosity)
{
  // Store the domain and grid spacing
  m_domain = a_domain;
  m_dx     = a_dx;

  if (m_gdnvPhysics != NULL)
  {
    delete m_gdnvPhysics;
    m_gdnvPhysics  = NULL;
  }

  m_gdnvPhysics = a_physPtr->new_godunovPhysics();
  m_gdnvPhysics->define(m_domain, m_dx);

  m_util.define(m_domain, m_dx);

  m_normalPredOrder = a_normalPredOrder;

  // Slope flattening is only valid when 4th order slopes are being computed
  // and limited in some way
  assert(!a_useFlattening || (a_useFourthOrderSlopes &&
                              (a_usePrimLimiting || a_useCharLimiting)));

  m_useFourthOrderSlopes = a_useFourthOrderSlopes;
  m_usePrimLimiting      = a_usePrimLimiting;
  m_useCharLimiting      = a_useCharLimiting;
  m_useFlattening        = a_useFlattening;

  // Artificial viscosity coefficient must be greater than zero
  assert(!a_useArtificialViscosity || (a_artificialViscosity >= 0.0));

  // Store the artificial viscosity flag and coefficient
  m_useArtificialViscosity = a_useArtificialViscosity;
  m_artificialViscosity    = a_artificialViscosity;

  m_isDefined = true;
}

// Set the current box.
void PatchGodunov::setCurrentTime(const Real& a_currentTime)
{
  m_currentTime      = a_currentTime;
  m_isCurrentTimeSet = true;
}

// Set the current box.
void PatchGodunov::setCurrentBox(const Box& a_currentBox)
{
  m_currentBox      = a_currentBox;
  m_isCurrentBoxSet = true;
}

void PatchGodunov::updateState(FArrayBox&       a_U,
                               FArrayBox        a_F[CH_SPACEDIM],
                               Real&            a_maxWaveSpeed,
                               const FArrayBox& a_S,
                               const Real&      a_dt,
                               const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  TimeUpstateInit.start();
  int nprim = m_gdnvPhysics->numPrimitives();
  int nflux = m_gdnvPhysics->numFluxes();

  FArrayBox whalf[CH_SPACEDIM];
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    Box faceBox = surroundingNodes(a_box, idir);

    whalf[idir].resize(faceBox, nprim);
    a_F[idir]  .resize(faceBox, nflux);

    whalf[idir].setVal(0.0);
    a_F  [idir].setVal(0.0);
  }
  TimeUpstateInit.stop();

  TimeUpstateComputeWHalf.start();
  computeWHalf(whalf, a_U, a_S, a_dt, a_box);
  TimeUpstateComputeWHalf.stop();

  FArrayBox dU(a_U.box(),a_U.nComp());

  TimeUpstateComputeUpdate.start();
  computeUpdate(dU, a_F, a_U, whalf, a_dt, a_box);
  TimeUpstateComputeUpdate.stop();

  a_U += dU;

  // Get and return the maximum wave speed on this patch/grid
  TimeUpstateMaxWaveSpeed.start();
  a_maxWaveSpeed = m_gdnvPhysics->getMaxWaveSpeed(a_U, m_currentBox);
  TimeUpstateMaxWaveSpeed.stop();
}

// Compute the time-centered values of the primitive variable on the
// interior cell faces of the cell-centered box a_box.
void PatchGodunov::computeWHalf(FArrayBox        a_WHalf[CH_SPACEDIM],
                                const FArrayBox& a_U,
                                const FArrayBox& a_S,
                                const Real&      a_dt,
                                const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  TimeCWH_boxSetup.start();
  // Get the number of various variables
  int numPrim = m_gdnvPhysics->numPrimitives();
  int numFlux = m_gdnvPhysics->numFluxes();

  // The current box
  Box curBox = m_currentBox;

  // Boxes for face centered state - used for the riemann() and
  // artificialViscosity() calls
  Box faceBox[SpaceDim];

  // Boxes for face centered fluxes
  Box fluxBox[SpaceDim];

  // Boxes for cell centered state - used for the updatePrim() calls
  Box ccBox[SpaceDim];

  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
  {
    // This box is face centered in direction "dir1", is one bigger than the
    // input box in all directions except "dir1" and stays one cell away from
    // the domain boundary in "dir1"
    faceBox[dir1] = curBox;
    faceBox[dir1].grow(1);
    faceBox[dir1] &= m_domain;
    faceBox[dir1].grow(dir1,-1);
    faceBox[dir1].surroundingNodes(dir1);

    // This box is face centered in direction "dir1", is one bigger than the
    // input box in all directions except "dir1"
    fluxBox[dir1] = curBox;
    fluxBox[dir1].grow(1);
    fluxBox[dir1].grow(dir1,-1);
    fluxBox[dir1] &= m_domain;
    fluxBox[dir1].surroundingNodes(dir1);

    // This is the cell centered analog of "fluxBox"
    ccBox[dir1] = curBox;
    ccBox[dir1].grow(1);
    ccBox[dir1].grow(dir1,-1);
    ccBox[dir1] &= m_domain;
  }

  // Primitive variables
  Box WBox = a_U.box();
  WBox &= m_domain;

  FArrayBox W(WBox,numPrim);
  TimeCWH_boxSetup.stop();

  // Create a temp storage for a_U - RS 06/10/02
  TimeCWH_aUcopy.start();
  FArrayBox Uold(a_U.box(),a_U.nComp());
  Uold.copy(a_U);
  TimeCWH_aUcopy.stop();

  // Calculate the primitive variables from the conserved variables
  TimeCWH_consToPrim.start();
  m_gdnvPhysics->consToPrim(W,a_U,WBox);
  TimeCWH_consToPrim.stop();

  // Define the box where slopes will be needed (one larger than the final
  // update box)
  TimeCWH_flattening.start();
  Box slopeBox = curBox;
  slopeBox.grow(1);
  slopeBox &= m_domain;

  // Compute flattening once for all slopes if needed
  FArrayBox flattening(slopeBox,1);

  if (m_useFlattening)
  {
    Interval velInt    = m_gdnvPhysics->velocityInterval();
    int pressureIndex  = m_gdnvPhysics->pressureIndex();
    Real smallPressure = m_gdnvPhysics->smallPressure();
    int bulkIndex      = m_gdnvPhysics->bulkModulusIndex();

    m_util.computeFlattening(flattening,
                             W,
                             velInt,
                             pressureIndex,
                             smallPressure,
                             bulkIndex,
                             slopeBox);
  }

  // Intermediate, extrapolated primitive variables
  FArrayBox WMinus[SpaceDim];
  FArrayBox WPlus [SpaceDim];

  // Initial fluxes
  FArrayBox WHalf1[SpaceDim];
  FArrayBox Flux1[SpaceDim];
  TimeCWH_flattening.stop();

  FArrayBox localSource;

  // If the source term is valid, make a local copy, increment it, and scale
  // it by 1/2 the timestep
  if (!a_S.box().isEmpty())
  {
    TimeCWH_incSource.start();
    localSource.define(a_S.box(), a_S.nComp());
    localSource.copy(a_S);

    m_gdnvPhysics->incrementSource(localSource,W,slopeBox);

    localSource *= 0.5 * a_dt;
    TimeCWH_incSource.stop();
  }

  // Compute initial fluxes
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
  {
    // Size the intermediate, extrapolated primitive variables
    TimeCWH_PPMNormalPred.start();
    WMinus[dir1].resize(slopeBox,numPrim);
    WPlus [dir1].resize(slopeBox,numPrim);

    Real dtbydx = a_dt/m_dx;
    // Compute predictor step to obtain extrapolated primitive variables
    if (m_normalPredOrder == 1)
    {
      PLMNormalPred(WMinus[dir1],WPlus[dir1],dtbydx,W,
                    flattening,dir1,slopeBox);
    }
    else if (m_normalPredOrder == 2)
    {
      PPMNormalPred(WMinus[dir1],WPlus[dir1],dtbydx,W,
                    flattening,dir1,slopeBox);
    }
    else
    {
      MayDay::Error("PatchGodunov::computeWHalf: Normal predictor order must be 1 (PLM) or 2 (PPM)");
    }
    TimeCWH_PPMNormalPred.stop();

    // If the source term is valid add it to the primitive quantities
    TimeCWH_addSourceTerm.start();
    if (!localSource.box().isEmpty())
    {
      WMinus[dir1] += localSource;
      WPlus [dir1] += localSource;
    }
    TimeCWH_addSourceTerm.stop();

    // Solve the Riemann problem and get fluxes
    TimeCWH_solveRiemann.start();
    WHalf1[dir1].resize(fluxBox[dir1],numFlux);
    Flux1[dir1].resize(fluxBox[dir1],numFlux);
    m_gdnvPhysics->riemann(WHalf1[dir1],WPlus[dir1],WMinus[dir1],W,
                           m_currentTime,dir1,faceBox[dir1]);
    TimeCWH_solveRiemann.stop();
    TimeCWH_getFlux.start();
    m_gdnvPhysics->getFlux(Flux1[dir1],WHalf1[dir1],dir1,Flux1[dir1].box());
    TimeCWH_getFlux.stop();
  }

#if (CH_SPACEDIM == 3)
  TimeCWH_3DintermediateFluxes.start();
  // In 3D, compute some additional intermediate fluxes
  //
  // NOTE:  The diagonal entries of this array of fluxes are not
  // used and will not be defined.
  FArrayBox WHalf2[SpaceDim][SpaceDim];

  // Compute the intermediate, corrected fluxes in each direction
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
  {
    // Correct fluxes using fluxes from a different direction
    for (int dir2 = 0; dir2 < SpaceDim; dir2++)
    {
      // A different direction has been found
      if (dir2 != dir1)
      {
        // Temporary primitive variables
        FArrayBox WTempMinus(WMinus[dir1].box(),numPrim);
        FArrayBox WTempPlus (WPlus [dir1].box(),numPrim);
        FArrayBox AdWdx(WPlus[dir1].box(),numPrim);

        // preventing uninitialized memory reads which cause
        // FPE's on some machines
        // AdWdx.setVal(666.666);

        // Copy data for in place modification
        WTempMinus.copy(WMinus[dir1]);
        WTempPlus .copy(WPlus [dir1]);

        // Update the current, extrapolated primitive variable using a flux
        // in a different direction

        m_gdnvPhysics->quasilinearUpdate(AdWdx,
            WHalf1[dir2],WTempMinus,-(1.0/3.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WTempMinus += AdWdx;

        m_gdnvPhysics->quasilinearUpdate(AdWdx,
            WHalf1[dir2],WTempPlus ,-(1.0/3.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WTempPlus  += AdWdx;

        // Solve the Riemann problem.

        WHalf2[dir1][dir2].resize(fluxBox[dir1],numPrim);
        m_gdnvPhysics->riemann(WHalf2[dir1][dir2],WTempPlus,WTempMinus,W,
                               m_currentTime,dir1,faceBox[dir1]);
      }
    }
  }
  TimeCWH_3DintermediateFluxes.stop();
#endif

  // faceBox and fluxBox are now a bit smaller for the final corrections
  TimeCWH_boxSurroundingNodes.start();
  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
  {
    faceBox[dir1] = curBox;
    faceBox[dir1].grow(dir1,1);
    faceBox[dir1] &= m_domain;
    faceBox[dir1].grow(dir1,-1);
    faceBox[dir1].surroundingNodes(dir1);

    fluxBox[dir1] = curBox;
    fluxBox[dir1].surroundingNodes(dir1);
  }
  TimeCWH_boxSurroundingNodes.stop();

  // Do the final corrections to the fluxes
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
  {
    // Correct the flux using fluxes in the remaining direction(s)
    for (int dir2 = 0; dir2 < SpaceDim; dir2++)
    {
      // A different direction has been found
      if (dir2 != dir1)
      {
#if (CH_SPACEDIM == 2)
        // In 2D, the current primitive state is updated by a flux in
        // the other direction
        TimeCWH_2DfinalFluxCorrection.start();
        FArrayBox AdWdx(WPlus[dir1].box(),numPrim);

        m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf1[dir2],WMinus[dir1],
                                         -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WMinus[dir1] += AdWdx;

        m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf1[dir2],WPlus [dir1],
                                         -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WPlus[dir1] += AdWdx;
        TimeCWH_2DfinalFluxCorrection.stop();
#elif (CH_SPACEDIM == 3)
        // In 3D, find a direction different from the two above
        TimeCWH_3DfinalFluxCorrection.start();
        int dir3 = 3 - dir1 - dir2;

        // Update the conservative state using both corrected fluxes in
        // the other two directions
        FArrayBox AdWdx(WPlus[dir1].box(),numPrim);

        m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf2[dir2][dir3],WMinus[dir1],
                                         -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WMinus[dir1].plus(AdWdx,0,0,numPrim);

        m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf2[dir2][dir3],WPlus [dir1],
                                         -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
        WPlus [dir1].plus(AdWdx,0,0,numPrim);
        TimeCWH_3DfinalFluxCorrection.stop();
#else
        // Only 2D and 3D should be possible
        MayDay::Error("PatchGodunov::computeWHalf: CH_SPACEDIM not 2 or 3!");
#endif
      }
    }

    // Solve the Riemann problem to obtain time-centered face values.
    // be returned
    TimeCWH_finalRiemann.start();
    FArrayBox& WHalf = a_WHalf[dir1];
    m_gdnvPhysics->riemann(WHalf,WPlus[dir1],WMinus[dir1],W,
                           m_currentTime,dir1,faceBox[dir1]);
    TimeCWH_finalRiemann.stop();
  }
}

void PatchGodunov::computeUpdate(FArrayBox&       a_dU,
                                 FArrayBox        a_F[CH_SPACEDIM],
                                 const FArrayBox& a_U,
                                 const FArrayBox  a_wHalf[CH_SPACEDIM],
                                 const Real&      a_dt,
                                 const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  TimeCUadUzero.start();
  a_dU.setVal(0.0);
  TimeCUadUzero.stop();

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    // Get flux from wHal
    TimeCUgetFlux.start();
    m_gdnvPhysics->getFlux(a_F[idir],a_wHalf[idir],idir,a_F[idir].box());
    TimeCUgetFlux.stop();

    if (m_useArtificialViscosity)
    {
      TimeCUconsToPrim.start();
      FArrayBox divu(a_F[idir].box(),1);
      int numPrim = m_gdnvPhysics->numPrimitives();
      FArrayBox W(a_U.box(),numPrim);

      m_gdnvPhysics->consToPrim(W,a_U,a_U.box());
      TimeCUconsToPrim.stop();

      TimeCUdivVel.start();
      Interval velInt = m_gdnvPhysics->velocityInterval();
      m_util.divVel(divu,W,velInt,idir,a_F[idir].box());
      TimeCUdivVel.stop();

      TimeCUfaceBoxCalc.start();
      Box faceBox = enclosedCells(a_F[idir].box());
      faceBox.grow(idir,1);
      faceBox &= m_domain;
      faceBox.grow(idir,-1);
      faceBox.surroundingNodes(idir);
      TimeCUfaceBoxCalc.stop();

      // CLEANUP:: faceBox should be related to a_box.
      TimeCUartificialViscosity.start();
      m_util.artificialViscosity(a_F[idir],a_U,
                                 divu,m_artificialViscosity,idir,faceBox);
      TimeCUartificialViscosity.stop();

      TimeCUartViscBC.start();
      m_gdnvPhysics->getPhysIBC()->artViscBC(a_F[idir],a_U,divu,
                                             idir,m_currentTime);
      TimeCUartViscBC.stop();
    }

    // Compute flux difference fHi-fLo
    TimeCUfluxDiff.start();
    FArrayBox diff(a_dU.box(), a_dU.nComp());
    diff.setVal(0.0);

    TF_FLUXDIFF.start();
    FORT_FLUXDIFFF(CHF_FRA(diff),
                   CHF_CONST_FRA(a_F[idir]),
                   CHF_CONST_INT(idir),
                   CHF_BOX(a_box));
    TF_FLUXDIFF.stop();

    // Add flux difference to dU
    a_dU += diff;
    TimeCUfluxDiff.stop();
  }

  // Multiply dU by dt/dx because that is what the output expects
  TimeCUmultiplyBydtdx.start();
  Real dtbydx = -a_dt/m_dx;
  a_dU *= dtbydx;
  TimeCUmultiplyBydtdx.stop();
}

void PatchGodunov::PLMNormalPred(FArrayBox&       a_WMinus,
                                 FArrayBox&       a_WPlus,
                                 const Real&      a_dtbydx,
                                 const FArrayBox& a_W,
                                 const FArrayBox& a_flat,
                                 const int&       a_dir,
                                 const Box&       a_box)
{
  int numprim = m_gdnvPhysics->numPrimitives();

  // This will hold 2nd or 4th order slopes
  FArrayBox dW(a_box,numprim);

  if (m_useFourthOrderSlopes)
  {
    // 2nd order slopes need to be computed over a larger box to accommodate
    // the 4th order slope computation
    Box boxVL = a_box;
    boxVL.grow(a_dir,1);
    boxVL &= m_domain;

    // Compute 2nd order (van Leer) slopes
    FArrayBox dWvL(boxVL, numprim);
    m_util.vanLeerSlopes(dWvL,a_W,numprim,
                         m_useCharLimiting || m_usePrimLimiting,
                         a_dir,boxVL);
    m_gdnvPhysics->getPhysIBC()->setBdrySlopes(dWvL,a_W,a_dir,m_currentTime);

    // Compute 4th order slopes, without limiting.
    m_util.fourthOrderSlopes(dW,a_W,dWvL, numprim, a_dir,a_box);
  }
  else
  {
    // Compute 2nd order (van Leer) slopes
    m_util.vanLeerSlopes(dW,a_W,numprim,
                         m_useCharLimiting || m_usePrimLimiting,
                         a_dir,a_box);
    m_gdnvPhysics->getPhysIBC()->setBdrySlopes(dW,a_W,a_dir,m_currentTime);
  }

  // To save on storage, we use the input values as temporaries for the
  // delta's
  a_WMinus.setVal(0.0);
  a_WPlus .setVal(0.0);

  // Compute one-sided differences as inputs for limiting.
  m_util.oneSidedDifferences(a_WMinus,a_WPlus,a_W,a_dir,a_box);

  FArrayBox lambda(a_box, numprim);
  m_gdnvPhysics->charValues(lambda, a_W, a_dir,a_box);

  if (m_useCharLimiting && m_usePrimLimiting)
  {
    MayDay::Error("PatchGodunov::PLMNormalPred: Attempt to limit slopes in primitive AND characteristic coordinates - not implemented");
  }

  // Apply limiter on characteristic or primitive variables. Either
  // way, must end up with characteristic variables to pass to to the
  // normal predictor utility.  Currently, cannot do both.

  // If doing characteristic limiting then transform before limiting
  if (m_useCharLimiting)
  {
    // Transform from primitive to characteristic variables
    m_gdnvPhysics->charAnalysis(a_WMinus,a_W,a_dir,a_box);
    m_gdnvPhysics->charAnalysis(a_WPlus ,a_W,a_dir,a_box);
    m_gdnvPhysics->charAnalysis(dW      ,a_W,a_dir,a_box);
  }

  if (m_useCharLimiting || m_usePrimLimiting)
  {
    // Do slope limiting
    m_util.slopeLimiter(dW,a_WMinus,a_WPlus,numprim,a_box);

    // Do slope flattening
    if (m_useFlattening)
    {
      m_util.applyFlattening(dW,a_flat,a_box);
    }
  }

  // If not doing characteristic limiting then transform after any limiting
  if (!m_useCharLimiting)
  {
    // Transform from primitive to characteristic variables
    m_gdnvPhysics->charAnalysis(dW,a_W,a_dir,a_box);
  }

  // To the normal prediction in characteristic variables
  m_util.PLMNormalPred(a_WMinus,a_WPlus,dW,lambda,a_dtbydx,a_box);

  // Construct the increments to the primitive variables
  m_gdnvPhysics->charSynthesis(a_WMinus,a_W,a_dir,a_box);
  m_gdnvPhysics->charSynthesis(a_WPlus ,a_W,a_dir,a_box);

  // Apply a physics-dependent post-normal predictor step:
  // For example:
  //   - adjust/bound delta's so constraints on extrapolated primitive
  //     quantities are enforced (density and pressure > 0).
  //   - compute source terms that depend on the spatially varying
  //     coefficients.
  m_gdnvPhysics->postNormalPred(a_WMinus,a_WPlus,a_W,a_dir,a_box);

  // Compute the state from the increments
  a_WMinus += a_W;
  a_WPlus  += a_W;
}

void PatchGodunov::PPMNormalPred(FArrayBox&       a_WMinus,
                                 FArrayBox&       a_WPlus,
                                 const Real&      a_dtbydx,
                                 const FArrayBox& a_W,
                                 const FArrayBox& a_flat,
                                 const int&       a_dir,
                                 const Box&       a_box)
{
  int numprim = m_gdnvPhysics->numPrimitives();

  Box faceBox = a_box;
  faceBox.surroundingNodes(a_dir);
  FArrayBox WFace(faceBox,numprim);

  m_util.PPMFaceValues(WFace,a_W,numprim,
                       m_useCharLimiting || m_usePrimLimiting,
                       a_dir,faceBox,m_currentTime,m_gdnvPhysics);

  // To save on storage, we use the input values as temporaries for the
  // delta's
  a_WMinus.setVal(0.0);
  a_WPlus .setVal(0.0);

  a_WMinus -= a_W;
  a_WPlus  -= a_W;

  WFace.shiftHalf(a_dir,1);
  a_WMinus += WFace;

  WFace.shift(a_dir,-1);
  a_WPlus  += WFace;

  FArrayBox lambda(a_box, numprim);
  m_gdnvPhysics->charValues(lambda, a_W, a_dir,a_box);

  if (m_useCharLimiting && m_usePrimLimiting)
  {
    MayDay::Error("PatchGodunov::PPMNormalPred: Attempt to limit slopes in primitive AND characteristic coordinates - not implemented");
  }

  // Apply limiter on characteristic or primitive variables. Either
  // way, must end up with characteristic variables to pass to to the
  // normal predictor utility.  Currently, cannot do both.

  // If doing characteristic limiting then transform before limiting
  if (m_useCharLimiting)
  {
    // Transform from primitive to characteristic variables
    m_gdnvPhysics->charAnalysis(a_WMinus,a_W,a_dir,a_box);
    m_gdnvPhysics->charAnalysis(a_WPlus ,a_W,a_dir,a_box);
  }

  if (m_useCharLimiting || m_usePrimLimiting)
  {
    // Do slope limiting
    m_util.PPMLimiter(a_WMinus,a_WPlus,numprim,a_box);

    // Do slope flattening
    if (m_useFlattening)
    {
      m_util.applyFlattening(a_WMinus,a_flat,a_box);
      m_util.applyFlattening(a_WPlus ,a_flat,a_box);
    }
  }

  // If not doing characteristic limiting then transform after any limiting
  if (!m_useCharLimiting)
  {
    // Transform from primitive to characteristic variables
    m_gdnvPhysics->charAnalysis(a_WMinus,a_W,a_dir,a_box);
    m_gdnvPhysics->charAnalysis(a_WPlus ,a_W,a_dir,a_box);
  }

  // To the normal prediction in characteristic variables
  m_util.PPMNormalPred(a_WMinus,a_WPlus,lambda,a_dtbydx,numprim,a_box);

  // Construct the increments to the primitive variables
  m_gdnvPhysics->charSynthesis(a_WMinus,a_W,a_dir,a_box);
  m_gdnvPhysics->charSynthesis(a_WPlus ,a_W,a_dir,a_box);

  // Apply a physics-dependent post-normal predictor step:
  // For example:
  //   - adjust/bound delta's so constraints on extrapolated primitive
  //     quantities are enforced (density and pressure > 0).
  //   - compute source terms that depend on the spatially varying
  //     coefficients.
  m_gdnvPhysics->postNormalPred(a_WMinus,a_WPlus,a_W,a_dir,a_box);

  // Compute the state from the increments
  a_WMinus += a_W;
  a_WPlus  += a_W;
}

// Return true if everything is defined and setup
bool PatchGodunov::isDefined() const
{
  return m_isDefined        &&
         m_isCurrentTimeSet &&
         m_isCurrentBoxSet  ;
}
