/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

//  -*- Mode: C++; Modified: "Tue 02 Feb 1999 14:29:43 by graves"; -*-
#include "LayoutIterator.H"
#include "CFStencil.H"
#include "LayoutIterator.H"
#include "DataIterator.H"
#include "ProblemDomain.H"

#include "Timer.H"
extern Timer Everything;
Timer cfs2("cfs2", Everything, 671, true);

bool
CFStencil::isDefined() const
{
    return m_isDefined;
}

/**
   are there any interpolation points?  \\
   returns false if so.*/
bool
CFStencil::isEmpty() const
{
    assert(m_isDefined);
    return m_fineIVS.isEmpty();
}

/** get fine intvects which need to be interpolated  \\
    This will be empty if isEmpty() returns true*/
const IntVectSet&
CFStencil::getFineIVS() const
{
    assert(m_isDefined);
    return m_fineIVS;
}

/** get coarse intvects that underly fiIVS.
    This will be empty if isEmpty() returns true*/
const IntVectSet&
CFStencil::getCoarIVS() const
{
    assert(m_isDefined);
    return m_coarIVS;
}

void
CFStencil::clear()
{
    m_fineIVS.define();
    m_coarIVS.define();
    setDefaultValues();
}
void
CFStencil::setDefaultValues()
{
    m_isDefined = false;
    m_direction = -777;
}
CFStencil::CFStencil()
{
    setDefaultValues();
}
CFStencil::CFStencil(const Box& a_b, int a_nComp)
{
  setDefaultValues();
  define(a_b, a_nComp);

}
void
CFStencil::define(const Box& a_b, int a_nComp)
{
  clear();
}

CFStencil::~CFStencil()
{
    clear();
}

CFStencil::CFStencil(
    const Box& a_fineDomain,
    const Box& a_grid,
    const DisjointBoxLayout& a_fineBoxes,
    const DisjointBoxLayout& a_coarBoxes,
    int a_refRatio,
    int a_direction,
    Side::LoHiSide a_hiorlo)
{
  ProblemDomain physDomain(a_fineDomain);
  setDefaultValues();
  define(physDomain,  a_grid, a_fineBoxes, a_coarBoxes,
         a_refRatio, a_direction, a_hiorlo);
}

CFStencil::CFStencil(
                     const ProblemDomain& a_fineDomain,
                     const Box& a_grid,
                     const DisjointBoxLayout& a_fineBoxes,
                     const DisjointBoxLayout& a_coarBoxes,
                     int a_refRatio,
                     int a_direction,
                     Side::LoHiSide a_hiorlo)
{
  setDefaultValues();
  define(a_fineDomain, a_grid, a_fineBoxes, a_coarBoxes,
         a_refRatio, a_direction, a_hiorlo);
}

void
CFStencil::define(
                  const Box& a_fineDomain,
                  const Box& a_grid,
                  const DisjointBoxLayout& a_fineBoxes,
                  const DisjointBoxLayout& a_coarBoxes,
                  int a_refRatio,
                  int a_direction,
                  Side::LoHiSide a_hiorlo)
{
  ProblemDomain physDomain(a_fineDomain);
  define(physDomain, a_grid, a_fineBoxes, a_coarBoxes, a_refRatio,
         a_direction, a_hiorlo);
}

void
CFStencil::define(
                  const ProblemDomain& a_fineDomain,
                  const Box& a_grid,
                  const DisjointBoxLayout& a_fineBoxes,
                  const DisjointBoxLayout& a_coarBoxes,
                  int a_refRatio,
                  int a_direction,
                  Side::LoHiSide a_hiorlo)
{
  m_isDefined = true;
  assert(a_refRatio >= 1);
  assert(a_direction >= 0);
  assert(a_direction < SpaceDim);
  assert((a_hiorlo == Side::Lo) ||
         (a_hiorlo == Side::Hi));

  assert(!a_fineDomain.isEmpty());

  //set internal vars.  most of these are kept around
  //just to keep the class from having an identity crisis.
  m_direction = a_direction;
  m_hiorlo =  a_hiorlo;

  Box finebox = a_grid;
  //this things are guaranteed by proper nesting.
  ProblemDomain coardom= coarsen(a_fineDomain, a_refRatio);

  //compute intvectset of all points on fine grid that
  //need to be interpolated

  //shift direction
  int hilo = sign(a_hiorlo);

  //create fine stencil
  Box edgebox;
  assert((hilo ==1) || (hilo == -1));
  if(hilo == -1)
    {
      edgebox = adjCellLo(finebox,m_direction,1);
    }
  else
    {
      edgebox = adjCellHi(finebox,m_direction,1);
    }

  //pout() << " finebox=" << finebox
  //     << " coardom=" << coardom
  //     << " a_fineDomain=" << a_fineDomain
  //     << endl;

  //pout() << " edgebox=" << edgebox << endl;

  edgebox = a_fineDomain & edgebox;
  //pout() << "m_fineIVS=" << m_fineIVS << endl;

  if(!edgebox.isEmpty())
    {
      Box periodicTestBox(a_fineDomain.domainBox());
      if (a_fineDomain.isPeriodic())
        {
          for (int idir=0; idir<SpaceDim; idir++)
            {
              if (a_fineDomain.isPeriodic(idir))
                {
                  periodicTestBox.grow(idir,-1);
                }
            }
        }

      m_fineIVS.define(edgebox);
      LayoutIterator lit = a_fineBoxes.layoutIterator();

      //pout() << "=====================================" << endl;
      //pout() << " edgebox=" << edgebox
      //<< " edgebox.numPts()=" << edgebox.numPts()
      //   << " m_fineIVS.numPts()=" << m_fineIVS.numPts()
      //   << endl;

      //int save = m_fineIVS.numPts();
      // appears this loop is the only thing taking up time in here... (ndk)
      // and it looks like it's all in the first -=
      for(lit.reset(); lit.ok(); ++lit)
        {
          //if(fromBox.bigEnd(0) < ghost.smallEnd(0)) {
            //can skip rest cuz we haven't gotten to something interesting
            //continue;
          //}
          //  pout() << " box(lit) = " << a_fineBoxes[lit()] << endl;
          //pout() << "edgebox.bigEnd(0) = " << edgebox.bigEnd(0) << "," << edgebox.bigEnd(1)
          //   << " edgebox.smallEnd(0) = " << edgebox.smallEnd(0)<< "," << edgebox.smallEnd(1)
          //   << " m_fineIVS.minBox().bigEnd(0) " << m_fineIVS.minBox().bigEnd(0)<< "," << m_fineIVS.minBox().bigEnd(1)
          //   << " .smallEnd(0) " << m_fineIVS.minBox().smallEnd(0)<< "," << m_fineIVS.minBox().smallEnd(1)
          //   << endl;
          m_fineIVS -= a_fineBoxes[lit()];
          //if (m_fineIVS.numPts() == 0) break;  // runs slower!
          //pout() << "  m_fineIVS.numPts() = " << m_fineIVS.numPts() << endl;
          // if periodic, also need to subtract periodic images
          // only do this IF we're periodic _and_ both boxes
          // adjoin the domain box boundary somewhere
          if (a_fineDomain.isPeriodic() && !periodicTestBox.contains(edgebox)
              && !periodicTestBox.contains(a_fineBoxes[lit()]))
            {
              ShiftIterator shiftIt = a_fineDomain.shiftIterator();
              IntVect shiftMult(a_fineDomain.domainBox().size());
              Box shiftedBox(a_fineBoxes[lit()]);
              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                {
                  IntVect shiftVect = shiftMult*shiftIt();
                  shiftedBox.shift(shiftVect);
                  m_fineIVS -= shiftedBox;
                  shiftedBox.shift(-shiftVect);
                } // end loop over periodic shift directions
            } // end if periodic
        }
      //pout() << "  m_fineIVS.numPts() = " << m_fineIVS.numPts() << endl;
      //if (m_fineIVS.numPts() != save && edgebox.bigEnd(0) > 0) exit(0);

    }

  //exit(0);

  //ivs where all coarse slopes are defined
  //== coarsened fine ivs
  m_coarIVS.define(m_fineIVS);
  m_coarIVS.coarsen(a_refRatio);
  // this is a trick to get around the lack of a IntVectSet intersection
  // operator which works with a ProblemDomain
  Box domainIntersectBox = m_coarIVS.minBox();
  domainIntersectBox = coardom & domainIntersectBox;
  m_coarIVS &= domainIntersectBox;
}

CFStencil&
CFStencil::operator=(const CFStencil& a_ecfsIn)
{
    if(&a_ecfsIn != this)
    {
        m_direction= a_ecfsIn.m_direction;
        m_hiorlo = a_ecfsIn.m_hiorlo;
        m_dataIndex = a_ecfsIn.m_dataIndex;
        m_fineIVS = a_ecfsIn.m_fineIVS;
        m_coarIVS= a_ecfsIn.m_coarIVS;
    }
    return *this;
}
