/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DERIVSTENCIL_H_
#define _DERIVSTENCIL_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "REAL.H"
#include "IntVect.H"
#include "Vector.H"

///class to encapsulate the operations to create derivs on irreg stencils
/**
   DerivStencil is meant to be used to encapsulate the information
   necessary to take finite difference derivatives at a point in
   space.  Every point in the stencil has a weight.  You add them
   (the boxarrayindex and the weight) at the same time and you can manipulate
   the weights enmasse by real number operations.  Stencils may not
   interact with each other with the same sort of arithmetic because
   that would bring up issues as to what to do when there is
   incomplete intersection between the stencils.

 */
class DerivStencil
{
public:
  ///
  void define();

  ///
  /** return true if  any define function been called.
   */
  bool isDefined() const
  {return isdefined;}

  ///
  /**
     default constructor;  creates empty vectors
  */
  DerivStencil();

  ///
  /**
     copy constructor;  sets *this = a_dsin
  */
  DerivStencil(const DerivStencil& a_dsin);

  ///
  ~DerivStencil();

  ///
  /**
     make derivstencil empty
  */
  void clear();

  ///
  /**
     return length of vectors
  */
  int size() const;

  ///
  /**
     get iv at ivec
  */
  const IntVect& getIndex(int a_ivec) const;

  ///
  /**
     get weight at ivec
  */
  const Real& getWeight(int a_ivec) const;

  ///
  /**
     add another set if the IntVect is not in the
     stencil already.  \\
     **Add the weight to the existing weight otherwise**
     */
  void accumulate(const IntVect& a_iv,  Real a_weight);

  ///
  /**
     assignment operator
  */
  const DerivStencil& operator=(const DerivStencil& a_dsin);

  ///
  /**
     Multiply each weight by a_facin
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator*=(Real a_facin);

  ///
  /**
     divide each weight by a_denom
     does nothing if vectors are of zero length.
     Be advised ---
     This function does no checking to see if a_denom
     is close to zero.
  */
  const DerivStencil& operator/=(Real a_denom);

  ///
  /**
     add a_facin to each weight
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator+=(Real a_facin);

  ///
  /**
     subtract a_facin from each weight
     does nothing if vectors are of zero length
  */
  const DerivStencil& operator-=(Real a_facin);

protected:
  ///
  Vector<IntVect> m_vectIV;
  ///
  Vector<Real>          m_vectWgt;
  ///
  bool isdefined;
};

#endif
