/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _LEVELFLUXREGISTER_H_
#define _LEVELFLUXREGISTER_H_
#include "REAL.H"
#include "Vector.H"
#include "FArrayBox.H"
#include "IntVectSet.H"
#include "CFStencil.H"
#include "LoHiSide.H"
#include "LevelData.H"
#include "LayoutData.H"
#include "ProblemDomain.H"

///LevelFluxRegister-A class to encapsulate a levels worth of flux registers.
/**
   A LevelFluxRegister  handles all the data choreography
   necessary to create a levels worth of flux registers.
   You send a pointer to a valid LevelCFStencil to its constructor
   or define function and it parsels out all the necessary information
   to the grid flux registers which hands it to the flux registers.
   The actual functions to manipulate data in the flux registers
   live in the class FluxRegister.
*/
class LevelFluxRegister
{
public:

  ///
  /**
     Default constructor.  Creates an uninitialized levelfluxregister.
  */
  LevelFluxRegister();

  ///
  /**
     Full constructor.  Calls the define function which creates
     a levels worth of flux registers.  a_dProblem is fine-level
     domain
  */
  LevelFluxRegister(
                    const DisjointBoxLayout& a_dbl,
                    const DisjointBoxLayout& a_dblCoarse,
                    const Box& a_dProblem,
                    int a_nRefine,
                    int a_nComp);

  ///
  /**
     Full constructor.  Calls the define function which creates
     a levels worth of flux registers.  a_dProblem is fine-level
     domain
  */
  LevelFluxRegister(
                    const DisjointBoxLayout& a_dbl,
                    const DisjointBoxLayout& a_dblCoarse,
                    const ProblemDomain& a_dProblem,
                    int a_nRefine,
                    int a_nComp);

  ///
  ~LevelFluxRegister();

  ///
  /**
     Full define function.  Creates a levels worth of flux registers.
     The values in the flux registers are still undefined, however.
     To zero the fluxregisters, you must call setToZero().
  */
  void define(
              const DisjointBoxLayout& a_dbl,
              const DisjointBoxLayout& a_dblCoarse,
              const Box& a_dProblem,
              int a_nRefine,
              int a_nComp);

  ///
  /**
     Full define function.  Creates a levels worth of flux registers.
     The values in the flux registers are still undefined, however.
     To zero the fluxregisters, you must call setToZero().
  */
  void define(
              const DisjointBoxLayout& a_dbl,
              const DisjointBoxLayout& a_dblCoarse,
              const ProblemDomain& a_dProblem,
              int a_nRefine,
              int a_nComp);

  ///
  /**
     Modifies this LevelFluxRegister so that it is returned to the
     uninitialized state.  User must now call the full define() before
     using it.

  */
  void
  undefine();

  ///
  /**
      Initialize values of registers  to zero.
  */
  void setToZero();

  ///
  /**
     increments the register with data from coarseFlux, multiplied by scale.
     coarseFlux must contain the coarse fluxes in the dir direction
     for the grid   m_coarseLayout[coarseDataIndex].
     By convention, only the low side flux is used to
     avoid double-counting at coarse-fine interfaces.
     This operation is local.
     Increment coarse does register += -sign(a_sd)*incrScale*coarseFlux
  */
  void incrementCoarse(
                       FArrayBox& a_coarseFlux,
                       Real a_scale,
                       const DataIndex& a_coarseDataIndex,
                       const Interval& a_srcInterval,
                       const Interval& a_dstInterval,
                       int a_dir);

  ///
  /**
     Same as above, except you only increment the a_sd side.
  */
  void incrementCoarse(
                       FArrayBox& a_coarseFlux,
                       Real a_scale,
                       const DataIndex& a_coarseDataIndex,
                       const Interval& a_srcInterval,
                       const Interval& a_dstInterval,
                       int a_dir,
                       Side::LoHiSide a_sd);

  ///
  /**
     increments the register with data from fineFlux, multiplied by scale.
     fineFlux must contain the coarse fluxes in the dir direction
     for the grid m_fineLayout[fineDataIndex].
     This operation is local.
     increment fine   does register +=  sign(a_sd)*incrScale*Ave(fineFlux)
  */
  void incrementFine(
                     FArrayBox& a_fineFlux,
                     Real a_scale,
                     const DataIndex& a_fineDataIndex,
                     const Interval& a_srcInterval,
                     const Interval& a_dstInterval,
                     int a_dir,
                     Side::LoHiSide a_sd);

  ///
  /**
     increments uCoarse with the reflux divergence of the
     contents of the flux register.
     Note that there is no srccomp etc crap here.
     this is done for all components so uCoarse has
     to have the same number of components as input nComp.
     This operation is global and blocking.

     Reflux does cellValue -= refluxScale*registerContents
  */
  void reflux(
              LevelData<FArrayBox>& a_uCoarse,
              Real a_scale);

  ///
  /**
     increments uCoarse with the reflux divergence of the
     contents of the flux register.
     This operation is global and blocking.
     Reflux does cellValue -= refluxScale*registerContents
  */
  void reflux(
              LevelData<FArrayBox>& a_uCoarse,
              const Interval& a_coarse_interval,
              const Interval& a_flux_interval,
              Real a_scale);

  ///
  /**
     has full define function been called?  return true if so.
  */
  bool isDefined() const;

  ///
  void poutCoarseRegisters() const;

  ///
  void poutFineRegisters() const;
protected:

  // both of these are defined in the coarse index space,
  // but, m_coarFlux uses the coarse grid layout, and
  // m_fineFlux uses the fine grid layout.
  LevelData<FArrayBox> m_coarFlux;
  LevelData<FArrayBox> m_fineFlux;

  //LayoutData<IntVectSet> m_coarseLocations[CH_SPACEDIM*2];
  LayoutData< Vector<Box> > m_coarseLocations[CH_SPACEDIM*2];

  ProblemDomain m_domain;

  bool m_isDefined;

  int m_nRefine;

  Copier m_reverseCopier;

private:

  ///there is no operator= for this class
  void operator= (const LevelFluxRegister&);
  ///there is no copy constructor for this class
  LevelFluxRegister(const LevelFluxRegister&);
};
#endif
