/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "QuadCFStencil.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
#include "DataIterator.H"
#include "ProblemDomain.H"

#include "Timer.H"
extern Timer Everything;
Timer qcfs_all("qcfs all", Everything, 670, true);
Timer qcfs1("qcfs1", qcfs_all);
Timer qcfs2("qcfs2", qcfs_all);
Timer qcfs3("qcfs3", qcfs_all);
Timer qcfs4("qcfs4", qcfs_all);
Timer qcfs5("qcfs5", qcfs_all);
Timer qcfs6("qcfs6", qcfs_all);
Timer qcfs7("qcfs7", qcfs_all);
Timer qcfs8("qcfs8", qcfs_all);
Timer qcfs9("qcfs9", qcfs_all);
Timer qcfs10("qcfs10", qcfs_all);
Timer qcfs11("qcfs11", qcfs_all);

bool
QuadCFStencil::isDefined() const
{
  return m_isDefined;
}

/**
   compute second derivative in devdir direction
   at coarse point ivin
   Uses centered finite diff approximation if
   m_ivsStandard.contains(ivin ).
   Otherwise, it uses the stencil from m_second
   Asserts that derivdir != direction of face
*/
Real
QuadCFStencil::computeSecondDerivative(
                                       const BaseFab<Real> & a_phic,
                                       int a_derivdir,
                                       int a_ivar,
                                       const IntVect& a_ivin,
                                       Real a_dx) const
{
  assert(isDefined());
  assert(a_ivar >= 0);
  assert(a_ivar < a_phic.nComp());
  assert(!a_phic.box().isEmpty());
  assert(a_derivdir >= 0);
  assert(a_derivdir != m_direction);
  assert(a_derivdir < SpaceDim);
  assert(a_dx > 0);
  //use standard centered diff approx if
  //a_ivin is contained in m_ivsStandard
  Real secondd;
  if(m_ivsStandard.contains(a_ivin))
    {
      IntVect basisv = BASISV(a_derivdir);
      IntVect ivcc = a_ivin;
      IntVect ivlo = a_ivin - basisv;
      IntVect ivhi = a_ivin + basisv;
      assert(a_phic.box().contains(ivcc));
      assert(a_phic.box().contains(ivlo));
      assert(a_phic.box().contains(ivhi));
      Real phicc = a_phic(ivcc, a_ivar);
      Real philo = a_phic(ivlo, a_ivar);
      Real phihi = a_phic(ivhi, a_ivar);
      secondd = (phihi + philo - 2.0*phicc)/(a_dx*a_dx);
    }
  else
    {
      assert(m_ivsQuadd.contains(a_ivin));
      assert(m_dropOrd.box().contains(a_ivin));
      if(m_dropOrd(a_ivin))
        {
          secondd = 0.0;
        }
      else
        {
          const BaseFab<DerivStencil>& stenfab
            = m_second[a_derivdir];
          assert(stenfab.box().contains(a_ivin));
          const DerivStencil& dersten = stenfab(a_ivin);
          secondd = 0.0;
          bool keepzer = false;
          for(int isten = 0; isten < dersten.size(); isten++)
            {
              Real wgt = dersten.getWeight(isten);
              IntVect ivphi = dersten.getIndex(isten);
              //can't really assert this because blocking factor
              //could screw it up (grid one cell wide and all that)
              //assert(a_phic.box().contains(ivphi));
              if(a_phic.box().contains(ivphi))
                secondd += wgt*a_phic(ivphi, a_ivar);
              else
                keepzer = true;
            }
          if(keepzer) secondd = 0.0;
          secondd /= (a_dx*a_dx);
        }
    }
  return secondd;
}

/**
   compute first derivative in devdir direction
   at coarse point a_ivin
   Uses centered finite diff approximation if
   m_ivsStandard.contains(a_ivin ).
   Otherwise, it uses the stencil from m_firstD
   Asserts that a_derivdir != direction of face
*/

Real
QuadCFStencil::computeFirstDerivative(
                                      const BaseFab<Real> & a_phic,
                                      int a_derivdir,
                                      int a_ivar,
                                      const IntVect& a_ivin,
                                      Real a_dx) const
{
  assert(isDefined());
  assert(a_ivar >= 0);
  assert(a_ivar < a_phic.nComp());
  assert(!a_phic.box().isEmpty());
  assert(a_derivdir >= 0);
  assert(a_derivdir < SpaceDim);
  assert(a_derivdir != m_direction);
  assert(a_dx > 0);
  //use standard centered diff approx if
  //a_ivin is contained in m_ivsStandard
  Real firstd;
  if(m_ivsStandard.contains(a_ivin))
    {
      IntVect basisv = BASISV(a_derivdir);
      IntVect ivlo = a_ivin - basisv;
      IntVect ivhi = a_ivin + basisv;
      assert(a_phic.box().contains(ivlo));
      assert(a_phic.box().contains(ivhi));
      Real philo = a_phic(ivlo, a_ivar);
      Real phihi = a_phic(ivhi, a_ivar);
      firstd = (phihi - philo)/(2.0*a_dx);
    }
  else
    {
      const BaseFab<DerivStencil>& stenfab
        = m_firstD[a_derivdir];
      assert(stenfab.box().contains(a_ivin));
      const DerivStencil& dersten = stenfab(a_ivin);
      firstd = 0.0;
      bool keepzer = false;
      for(int isten = 0; isten < dersten.size(); isten++)
        {
          Real wgt = dersten.getWeight(isten);
          const IntVect& ivphi= dersten.getIndex(isten);
          //can't really assert this because blocking factor
          //could screw it up (grid one cell wide and all that)
          //assert(a_phic.box().contains(ivphi));
          if(a_phic.box().contains(ivphi))
            firstd += wgt*a_phic(ivphi, a_ivar);
          else
            keepzer = true;
        }
      if(keepzer) firstd = 0.0;
      firstd /= (a_dx);
    }
  return firstd;
}

/**
   compute mixed derivative (direction is unambiguous.
   x and y are the two directions tangential to face)
   at coarse point a_ivin. \\
   In two dimensions, always returns 0. \\
   Uses centered finite diff approximation if
   m_ivsStandard.contains(a_ivin ).
   It uses the stencil from m_mixedSten.\\
   Returns 0 if SpaceDim != 3
*/
Real
QuadCFStencil::computeMixedDerivative(
                                      const BaseFab<Real> & a_phic,
                                      int a_ivar,
                                      const IntVect& a_ivin,
                                      Real a_dx) const
{
#if (CH_SPACEDIM == 3)
  assert(isDefined());
  assert(a_ivar >= 0);
  assert(a_ivar < a_phic.nComp());
  assert(!a_phic.box().isEmpty());
  assert(a_dx > 0);
  //compute derivative directions
  //(defined as normal to m_direction)
  int itran1,itran2;
  if(m_direction==0)
    {
      itran2 =1;
      itran1 =2;
    }
  else if(m_direction==1)
    {
      itran2 =0;
      itran1 =2;
    }
  else if(m_direction==2)
    {
      itran2 =0;
      itran1 =1;
    }
  else
    {
      std::cerr << "Quadcfstencil::define-bogus m_direction" << endl;
      abort();
    }

  //use standard centered diff approx if
  //a_ivin is contained in m_ivsStandard
  Real mixedd;
  if(m_ivsStandard.contains(a_ivin))
    {
      IntVect basex = BASISV(itran2);
      IntVect basey = BASISV(itran1);
      IntVect ivur = a_ivin + basex + basey;
      IntVect ivul = a_ivin - basex + basey;
      IntVect ivlr = a_ivin + basex - basey;
      IntVect ivll = a_ivin - basex - basey;
      assert(a_phic.box().contains(ivur));
      assert(a_phic.box().contains(ivul));
      assert(a_phic.box().contains(ivlr));
      assert(a_phic.box().contains(ivll));
      Real phiur = a_phic(ivur, a_ivar);
      Real phiul = a_phic(ivul, a_ivar);
      Real philr = a_phic(ivlr, a_ivar);
      Real phill = a_phic(ivll, a_ivar);
      mixedd = (phiur+phill-philr-phiul)/(4.*a_dx*a_dx);
    }
  else
    {
      assert(m_ivsQuadd.contains(a_ivin));
      assert(m_dropOrd.box().contains(a_ivin));
      if(m_dropOrd(a_ivin))
        {
          mixedd = 0.0;
        }
      else
        {
          assert(m_mixedSten.box().contains(a_ivin));
          const DerivStencil& dersten = m_mixedSten(a_ivin);
          mixedd = 0.0;
          bool keepzer = false;
          for(int isten = 0; isten < dersten.size(); isten++)
            {
              Real wgt = dersten.getWeight(isten);
              const IntVect& ivphi =  dersten.getIndex(isten);
              //can't really assert this because blocking factor
              //could screw it up (grid one cell wide and all that)
              //assert(a_phic.box().contains(ivphi));
              if(a_phic.box().contains(ivphi))
                mixedd += wgt*a_phic(ivphi, a_ivar);
              else
                keepzer = true;
            }
          if(keepzer) mixedd = 0.0;
          mixedd /= (a_dx*a_dx);
        }
    }
  return mixedd;
#else
  return 0.0;
#endif
}

void
QuadCFStencil::setDefaultValues()
{
  m_isDefined = false;
  m_direction = -1;
}

QuadCFStencil::QuadCFStencil()
{
  setDefaultValues();
}

QuadCFStencil::~QuadCFStencil()
{

}

QuadCFStencil::QuadCFStencil(
                             const Box& a_fine_domain,
                             const Box& a_grid,
                             const DisjointBoxLayout& a_fineBoxes,
                             const DisjointBoxLayout& a_coarBoxes,
                             int a_refRatio, int a_direction,
                             Side::LoHiSide a_hiorlo)
{
  setDefaultValues();
  ProblemDomain fineProbDomain(a_fine_domain);
  define(fineProbDomain,  a_grid, a_fineBoxes, a_coarBoxes,
         a_refRatio,  a_direction, a_hiorlo);
}

QuadCFStencil::QuadCFStencil(
                             const ProblemDomain& a_fine_domain,
                             const Box& a_grid,
                             const DisjointBoxLayout& a_fineBoxes,
                             const DisjointBoxLayout& a_coarBoxes,
                             int a_refRatio, int a_direction,
                             Side::LoHiSide a_hiorlo)
{
  setDefaultValues();
  define(a_fine_domain, a_grid, a_fineBoxes, a_coarBoxes,
         a_refRatio,  a_direction, a_hiorlo);
}

void
QuadCFStencil::define(
                      const Box& a_fine_domain,
                      const Box& a_grid,
                      const DisjointBoxLayout& a_fineBoxes,
                      const DisjointBoxLayout& a_coarBoxes,
                      int a_refRatio,
                      int a_direction,
                      Side::LoHiSide a_hiorlo)
{
  ProblemDomain fineProbDomain(a_fine_domain);
  define(fineProbDomain, a_grid, a_fineBoxes, a_coarBoxes,
         a_refRatio, a_direction, a_hiorlo);

}

void
QuadCFStencil::define(
                      const ProblemDomain& a_fine_domain,
                      const Box& a_grid,
                      const DisjointBoxLayout& a_fineBoxes,
                      const DisjointBoxLayout& a_coarBoxes,
                      int a_refRatio,
                      int a_direction,
                      Side::LoHiSide a_hiorlo)
{
  qcfs_all.start();
  assert(!a_fine_domain.isEmpty());

  assert(a_refRatio >= 1);
  assert(a_direction >= 0);
  assert(a_direction < SpaceDim);
  assert(!a_grid.isEmpty());
  assert(a_fine_domain.contains(a_grid));
  assert((a_hiorlo == Side::Lo) ||
         (a_hiorlo == Side::Hi));
  m_isDefined = true;

  m_direction = a_direction;

  //qcfs1.start(); // 19%
  //define base cfstencil
  m_baseCFS.define(a_fine_domain,
                   a_grid,
                   a_fineBoxes, a_coarBoxes,
                   a_refRatio, a_direction,
                   a_hiorlo);
  //qcfs1.stop();

  //qcfs2.start();
  const IntVectSet& baseIVS = m_baseCFS.getCoarIVS();

  //these are the directions tangential to the face
  Tuple<int,SpaceDim-1> vinttran;
  int iint = 0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(idir != m_direction)
        {
          vinttran[iint] = idir;
          iint++;
        }
    }

  //compute all cells necessary for both stencils
  //and interpolation points
  Box coarseGrid = coarsen(a_grid, a_refRatio);
  ProblemDomain coarseDomain = coarsen(a_fine_domain, a_refRatio);
  //qcfs2.stop();

  // create a box which will indicate whether a cell is adjacent
  // to a periodic boundary or not (all cells contained in the box
  // are _not_ adjacent to a periodic boundary)
  //qcfs3.start();
  Box periodicTestBox(coarseDomain.domainBox());
  //qcfs3.stop();
  //qcfs4.start();
  if (coarseDomain.isPeriodic())
    for (int idir=0; idir<SpaceDim; idir++)
      if (coarseDomain.isPeriodic(idir))
        periodicTestBox.grow(idir,-1);
  //qcfs4.stop();

  //qcfs5.start();
  Box edgeBoxG1, edgeBoxG2;
  if(a_hiorlo == Side::Lo)
    {
      edgeBoxG2 = adjCellLo(coarseGrid,m_direction,1);
      edgeBoxG1 = adjCellLo(coarseGrid,m_direction,1);
    }
  else
    {
      edgeBoxG2 = adjCellHi(coarseGrid,m_direction,1);
      edgeBoxG1 = adjCellHi(coarseGrid,m_direction,1);
    }

  for(int iint = 0; iint < SpaceDim-1; iint++)
    {
      int itran = vinttran[iint];
      edgeBoxG2.grow(itran, 2);
      edgeBoxG1.grow(itran, 1);
    }
  edgeBoxG2 &= coarseDomain;
  edgeBoxG1 &= coarseDomain;
  //this is the coarsened version of the fine union
  //of rectangles
  //qcfs5.stop();
  //qcfs6.start();  // 26%
  DisjointBoxLayout coarsenedFineBoxes;
  coarsen(coarsenedFineBoxes, a_fineBoxes, a_refRatio);
  //qcfs6.stop();
  //qcfs7.start();
  IntVectSet ivsAllGood(edgeBoxG2);
  LayoutIterator litFine = coarsenedFineBoxes.layoutIterator();
  //qcfs7.stop();
  //qcfs8.start(); // 19%
  for(litFine.reset(); litFine.ok(); ++litFine)
    {
      ivsAllGood -= coarsenedFineBoxes.get(litFine());
      // need to subtract periodic images as well
      if (coarseDomain.isPeriodic()
          && !periodicTestBox.contains(coarsenedFineBoxes.get(litFine()))
          && !periodicTestBox.contains(coarseGrid))
        {
          ShiftIterator shiftIt = coarseDomain.shiftIterator();
          IntVect shiftMult = coarseDomain.domainBox().size();
          Box shiftedBox = coarsenedFineBoxes.get(litFine());
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
            {
              IntVect shiftVect = shiftMult*shiftIt();
              shiftedBox.shift(shiftVect);
              ivsAllGood -= shiftedBox;
              shiftedBox.shift(-shiftVect);
            } // end loop over periodic shift directions
        } // end if is periodic
    }
  //qcfs8.stop();
  //qcfs9.start();  // 19%
  //create temp IVS which has all points in stencils
  //which fell off coarse grid and subtract those
  IntVectSet ivsEdgeOfEarth = ivsAllGood;
  LayoutIterator litCoar =a_coarBoxes.layoutIterator();
  for(litCoar.reset(); litCoar.ok(); ++litCoar)
    {
      ivsEdgeOfEarth -= a_coarBoxes.get(litCoar());
      // if is periodic, need to subtract off periodic images here as well
      if (coarseDomain.isPeriodic()
          && !periodicTestBox.contains(a_coarBoxes.get(litCoar()))
          && !periodicTestBox.contains(coarseGrid))
        {
          ShiftIterator shiftIt = coarseDomain.shiftIterator();
          IntVect shiftMult = coarseDomain.domainBox().size();
          Box shiftedBox = a_coarBoxes.get(litCoar());
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
            {
              IntVect shiftVect = shiftMult*shiftIt();
              shiftedBox.shift(shiftVect);
              ivsEdgeOfEarth -= shiftedBox;
              shiftedBox.shift(-shiftVect);
            } // end loop over periodic shift directions
        } // end if is periodic

    }
  ivsAllGood -= ivsEdgeOfEarth;
  //qcfs9.stop();
  //qcfs10.start();

  //compute cells where pure centered diffs can be taken
  //start with the set of all coarse intvects
  //on the coarse-fine interface and compute
  //the set of all cells for which all the
  //centered-difference stencil is within
  //the coarse grid == m_ivsStandard
  m_ivsStandard = ivsAllGood;
  //ivsStandard = grow((ivsAllGood & edgeBoxG1), itran,-1)
  m_ivsStandard &= edgeBoxG1;

  for(int iint = 0; iint < SpaceDim-1; iint++)
    {
      int itran = vinttran[iint];
      m_ivsStandard.grow(itran, -1);
    }
  //the points where something special =
  //all points- standard points
  //by conservation
  m_ivsQuadd = baseIVS;
  m_ivsQuadd -= m_ivsStandard;
  //qcfs10.stop();

  //qcfs11.start();
  if(!m_ivsQuadd.isEmpty())
    {
      const Box& osbox = m_ivsQuadd.minBox();
      m_dropOrd.resize(osbox);
      m_dropOrd.setVal(false);

#if(CH_SPACEDIM == 3)
      //first calculate mixed derivative stencils
      //only in the case of 3 dimensions

      m_mixedSten.resize(osbox);
      //create four boxes of stencil
      int itran1 = vinttran[0];
      int itran2 = vinttran[1];
      Box stenbox(IntVect::Zero, BASISV(itran1)+BASISV(itran2));
      //fabs for four corners of stencil
      //0 shifts shown as comments for pedagogical reasons

      BaseFab<Real>  fabur(stenbox, 1);
      fabur.setVal(-1.0);
      fabur(BASISV(itran1)) = 1.0;
      fabur(BASISV(itran2)) = 1.0;
      //fabur.shift(itran1, 0);
      //fabur.shift(itran1, 0);

      BaseFab<Real>  fabll(stenbox, 1);
      fabll.copy(fabur);
      fabll.shift(itran1, -1);
      fabll.shift(itran2, -1);

      BaseFab<Real>  fablr(stenbox, 1);
      fablr.copy(fabur);
      //fablr.shift(itran1, 0);
      fablr.shift(itran2, -1);

      BaseFab<Real>  fabul(stenbox, 1);
      fabul.copy(fabur);
      fabur.shift(itran1, -1);
      //fabll.shift(itran2, 0);
      IVSIterator irregit(m_ivsQuadd);

      for(irregit.begin(); irregit.ok(); ++irregit)
        {
          IntVect ivirr = irregit();
          fabur.shift(ivirr);
          fablr.shift(ivirr);
          fabul.shift(ivirr);
          fabll.shift(ivirr);
          Box bur(fabur.box());
          Box bul(fabul.box());
          Box blr(fablr.box());
          Box bll(fabll.box());
          DerivStencil& dersten= m_mixedSten(ivirr);
          dersten.define();
          bool hasamixed = false;
          int imixed = 0;
          //add each stencil whose box is
          //contained in ivsAllGood and
          //average the result
          if(ivsAllGood.contains(bur))
            {
              hasamixed = true;
              addFabToSten(fabur, dersten);
              imixed++;
            }
          if(ivsAllGood.contains(bul))
            {
              hasamixed = true;
              addFabToSten(fabul, dersten);
              imixed++;
            }
          if(ivsAllGood.contains(blr))
            {
              hasamixed = true;
              addFabToSten(fablr, dersten);
              imixed++;
            }
          if(ivsAllGood.contains(bll))
            {
              hasamixed = true;
              addFabToSten(fabll, dersten);
              imixed++;
            }
          if(imixed > 0)
            {
              Real rnmixed = imixed;
              dersten /= rnmixed;
            }
          m_dropOrd(ivirr) = (!hasamixed);
          //shift fabs back so that stencils work
          fabur.shift(-ivirr);
          fablr.shift(-ivirr);
          fabul.shift(-ivirr);
          fabll.shift(-ivirr);
        }
#endif
      //now calculate all non-mixed derivative stencils
      for(int ivec = 0; ivec < SpaceDim-1; ivec++)
        {
          int idir = vinttran[ivec];
          IntVect basisv = BASISV(idir);
          BaseFab<DerivStencil>& second_ds = m_second [idir];
          BaseFab<DerivStencil>& firstd_ds = m_firstD[idir];
          second_ds.resize(osbox);
          firstd_ds.resize(osbox);
          IVSIterator irregit(m_ivsQuadd);
          for(irregit.begin(); irregit.ok(); ++irregit)
            {
              IntVect iv = irregit();
              if(!m_dropOrd(iv))
                {
                  Box bsten(iv, iv);
                  bsten.grow(idir, 1);
                  if(ivsAllGood.contains(bsten))
                    {
                      //here centered diffs in this direction
                      //apply.  this should never get hit in
                      //two dimensions.  In 2d, if iv is part of
                      //irregit, then it should need one
                      //sided stuff
                      assert(SpaceDim == 3);
                      BaseFab<Real>  fabsten(bsten,1);
                      fabsten(iv-basisv) =  1.0;
                      fabsten(iv+basisv) =  1.0;
                      fabsten(iv)        = -2.0;
                      DerivStencil& secdersten= second_ds(iv);
                      secdersten.define();
                      addFabToSten(fabsten, secdersten);
                      fabsten(iv-basisv) = -0.5;
                      fabsten(iv+basisv) =  0.5;
                      fabsten(iv)        =  0.0;
                      DerivStencil& firdersten= firstd_ds(iv);
                      firdersten.define();
                      addFabToSten(fabsten, firdersten);
                    }
                  else
                    {
                      Box bshift(bsten);
                      bshift.shift(basisv);
                      if(ivsAllGood.contains(bshift))
                        {
                          //point on the negative side is covered but
                          //the other one is not so introduce shifted
                          //stencil
                          BaseFab<Real>  fabsten(bshift,1);
                          fabsten(iv)          =  1.0;
                          fabsten(iv+  basisv) = -2.0;
                          fabsten(iv+2*basisv) =  1.0;
                          DerivStencil& secdersten= second_ds(iv);
                          secdersten.define();
                          addFabToSten(fabsten, secdersten);
                          //need three point stencil for derivative
                          //to get second-order, shifted
                          fabsten(iv)          = -3.0/2.0;
                          fabsten(iv+  basisv) =  4.0/2.0;
                          fabsten(iv+2*basisv) = -1.0/2.0;
                          DerivStencil& firdersten= firstd_ds(iv);
                          firdersten.define();
                          addFabToSten(fabsten, firdersten);
                        }
                      else
                        {
                          bshift.shift(-2*basisv);
                          if(ivsAllGood.contains(bshift))
                            {
                                //point on the positive side is covered but
                                //the other one is not so introduce shifted
                                //stencil
                              BaseFab<Real>  fabsten(bshift,1);
                              fabsten(iv)          =  1.0;
                              fabsten(iv-  basisv) = -2.0;
                              fabsten(iv-2*basisv) =  1.0;
                              DerivStencil& secdersten= second_ds(iv);
                              secdersten.define();
                              addFabToSten(fabsten, secdersten);
                                //need three point stencil for derivative
                                //to get second-order, shifted
                              fabsten(iv)          =  3.0/2.0;
                              fabsten(iv-  basisv) = -4.0/2.0;
                              fabsten(iv-2*basisv) =  1.0/2.0;
                              DerivStencil& firdersten= firstd_ds(iv);
                              firdersten.define();
                              addFabToSten(fabsten, firdersten);
                            }
                          else
                            {
                                //both sides have a covered point.
                                //need to drop order and use
                                //one-sided first derivative
                              m_dropOrd(iv) = true;
                              IntVect ivhi = iv + basisv;
                              if(ivsAllGood.contains(ivhi))
                                {
                                  //derivative taken on positive side
                                  Box bfirst(iv, ivhi);
                                  BaseFab<Real>  fabsten(bfirst, 1);
                                  fabsten(iv  ) = -1.0;
                                  fabsten(ivhi) =  1.0;
                                  DerivStencil& firdersten= firstd_ds(iv);
                                  firdersten.define();
                                  addFabToSten(fabsten, firdersten);
                                }
                              else
                                {
                                  //derivative taken on negative side
                                  IntVect ivlo = iv - basisv;
                                  if(ivsAllGood.contains(ivlo))
                                    {
                                      Box bfirst(ivlo, iv);
                                      BaseFab<Real>  fabsten(bfirst, 1);
                                      fabsten(iv  ) =  1.0;
                                      fabsten(ivlo) = -1.0;
                                      DerivStencil& firdersten= firstd_ds(iv);
                                      firdersten.define();
                                      addFabToSten(fabsten, firdersten);
                                    }
                                  else
                                    {
                                      //neither is there.  set fab sten to zero.
                                      Box bfirst(iv, iv);
                                      BaseFab<Real>  fabsten(bfirst, 1);
                                      fabsten(iv  ) =  0.0;
                                      DerivStencil& firdersten= firstd_ds(iv);
                                      firdersten.define();
                                      addFabToSten(fabsten, firdersten);
                                    } //end error message
                                }//end positive point covered
                            }//end drop order else
                        }//end point on positive side covered else
                    }//end if(there needs one side in THIS direction)
                } //end if(!m_dropOrd) (if mixed is calculable)
            } //end loop over one sided cells
        }//end loop over tangential directions
    } //end if(there are any Quadd points)
  //qcfs11.stop();
  qcfs_all.stop();
} //end constructor for Quadcfstencil

/**
   add fab to the stencil.
   the weight equal to the value of the
   of the fab data at the box
*/
void
QuadCFStencil::addFabToSten(
                            const BaseFab<Real> & a_fabin,
                            DerivStencil&  a_sten
                            )
{
  assert(isDefined());
  assert(!a_fabin.box().isEmpty());
  assert(a_fabin.nComp() == 1);
  BoxIterator bit(a_fabin.box());
  for(bit.begin(); bit.ok(); ++bit)
    {
      IntVect iv = bit();
      Real weight = a_fabin(iv);
      a_sten.accumulate(iv, weight);
    }
}
