/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DATAITERATOR_H_
#define _DATAITERATOR_H_

#include "DataIndex.H"
#include "BoxLayout.H"
#include "SPMD.H"

#ifdef MPI

///An Iterator based on a BoxLayout object.
/**
   An Iterator based on a BoxLayout object.  It does not
   support a dereferencing operation, since it is intended
   to work with all of BoxLayouts, DisjointBoxLayouts, BoxLayoutDatas
   LevelDatas, and any object that is built on top
   of a BoxLayout object.

   DataIterator accesses the data in a BoxLayout-based
   object in a <b> data-parallel </b> manner.
   This means that it skips over entries for Boxes not
   assigned to this processor.  The order of access is not defined.

   In serial execution mode, there is no difference between
   DataIterator and LayoutIterator.
*/
class DataIterator
{
  //Note: parallel version doesn't inherit LayoutIterator, serial version does
public:

  /// a null constructed DataIterator will return false on ok()
  DataIterator();

  DataIterator(const BoxLayout& a_layout){*this = a_layout.dataIterator();}

  ~DataIterator()
  {}

  /// return the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  inline const DataIndex& operator()() const ;
  /// return a copy of the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  DataIndex i() const { return this->operator()();}

  /// move the iterator to the next index in the layout
  inline void operator++();
  /// move the iterator to the next index in the layout
  void incr() { ++(*this);}

  /// return true if this iterator is still in the layout
  bool ok() const;

  /// initialize this iterator to the first index in the layout
  void begin();

  /// same as begin()
  void reset();

  /// move this iterator to after the last index in the layout
  /** The iterator will be !ok() afterwards. */
  void end();

private:

  friend class BoxLayout;
  friend class DisjointBoxLayout;

  DataIterator(const BoxLayout& boxlayout, const int* layoutID);

  BoxLayout m_layout;

  unsigned int m_index;

  DataIndex m_current;

  unsigned int m_procID;

};

inline DataIterator::DataIterator()
  : m_index(0), m_procID(0)
{}

inline const DataIndex& DataIterator::operator()() const
{
  assert(ok());
  return m_current;
}

inline void DataIterator::operator++()
{
  const Entry* box;
  while(++m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
    }

}

inline bool DataIterator::ok() const
{
  return m_index < m_layout.size();
}

inline void DataIterator::reset()
{
  begin();
}

inline void DataIterator::begin()
{
  m_index = 0;
  const Entry* box;
  while(m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
      ++m_index;
    }
}

#else

// serial version
#include "LayoutIterator.H"

class DataIterator : public LayoutIterator
{
public:
  DataIterator()
  {}

  DataIterator(const BoxLayout& a_layout){*this = a_layout.dataIterator();}

  /// return the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  const DataIndex& operator()() const
  {
    return (const DataIndex&)(LayoutIterator::operator()());
  };

  /// return a copy of the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  DataIndex i() const {return this->operator()();}

private:
  friend class BoxLayout;
  friend class DisjointBoxLayout;

  DataIterator(const BoxLayout& boxlayout, const int* layoutID)
    :
    LayoutIterator(boxlayout, layoutID)
  {}
};

#endif  /*MPI*/

#define DATAITERATOR(CLASS, BOXLAYOUT) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN1(CLASS, DataIndex, di)

#define ENDITERATOR(CLASS)  \
        MT_END1(CLASS, DataIndex, di) \
        }

#define DATAITERATOR1(CLASS, BOXLAYOUT, TYPE1, VAL1) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN2(CLASS, TYPE1, VAL1, DataIndex, di)

#define ENDITERATOR1(CLASS, TYPE1, VAL1)  \
        MT_END2(CLASS, TYPE1, VAL1, DataIndex, di) \
        }

#define DATAITERATOR2(CLASS, BOXLAYOUT, TYPE1, VAL1, TYPE2, VAL2) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN3(CLASS, TYPE1, VAL1, TYPE2, VAL2, DataIndex, di)

#define ENDITERATOR2(CLASS, TYPE1, VAL1, TYPE2, VAL2)  \
        MT_END3(CLASS, TYPE1, VAL1, TYPE2, VAL2, DataIndex, di) \
        }

#endif
