/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "ProblemDomain.H"
#include "MayDay.H"

using std::istream;
using std::ws;
ShiftIterator::ShiftIterator(const bool* a_isPeriodic)
{
  m_index = 0;
  computeShifts(a_isPeriodic);
}

ShiftIterator::~ShiftIterator()
{
  m_shift_vectors.clear();
}

void
ShiftIterator::computeShifts(const bool* a_isPeriodic)
{
  m_shift_vectors.resize(0);

  // handle no directions are periodic
  if (D_TERM(!a_isPeriodic[0],
             && !a_isPeriodic[1],
             && !a_isPeriodic[2]))
    {
      // return with no periodic directions defined
    }
  // now handle if all directions are periodic
  else if (D_TERM(a_isPeriodic[0],
                  && a_isPeriodic[1],
                  && a_isPeriodic[2]))
    {
      // loop over all possible shifts in this case (including corners)
      D_TERM(
             for (int xVect=-1; xVect<2; xVect++),
             for (int yVect=-1; yVect<2; yVect++),
             for (int zVect=-1; zVect<2; zVect++) )
        {
          IntVect shiftVect(D_DECL(xVect,yVect,zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero)
            {
              m_shift_vectors.push_back(shiftVect);
            }
        } // end loop over possible shift vectors
    }
  // case where some but not all directions are periodic
  // this could probably also swallow the completely periodic case
  else
    {
      // i think this should work
      // this is designed to filter out non-periodic directions
      D_TERM(int xMult = (a_isPeriodic[0]?1:0);,
             int yMult = (a_isPeriodic[1]?1:0);,
             int zMult = (a_isPeriodic[2]?1:0););

      // in non-periodic dirctions, shift vector component should
      // always be 0
      D_TERM(
             for (int xVect=-1*xMult; xVect<1+xMult; xVect++),
             for (int yVect=-1*yMult; yVect<1+yMult; yVect++),
             for (int zVect=-1*zMult; zVect<1+zMult; zVect++)  )
        {
          IntVect shiftVect(D_DECL(xVect, yVect, zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero)
            {
              m_shift_vectors.push_back(shiftVect);
                   }
        } // end loop over possible shift vectors
    } // end case where some directions are periodic
}

ProblemDomain::ProblemDomain(const Box& a_domBox)
  : m_domainBox(a_domBox)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const Box& a_domBox,
                             const bool* a_isPeriodic)
  : m_domainBox(a_domBox)
{
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big,
                             const bool* a_isPeriodic)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len,
                             const bool* a_isPeriodic)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  }
  m_shiftIt.computeShifts(m_isPeriodic);
}

// returns true if box intersects problem domain
bool
ProblemDomain::intersects(const Box& a_box) const
{
  if (isEmpty() || a_box.isEmpty()) return false;
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0]
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0)
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1]
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1)
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));
}

bool
ProblemDomain::intersectsNotEmpty(const Box& a_box) const
{
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0]
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0)
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1]
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1)
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));

}

bool
ProblemDomain::intersects(const Box& box1, const Box& box2) const
{

  bool boxesIntersect = false;
  // first do the obvious and easy check
  boxesIntersect = box1.intersects(box2);

  // now do the harder check whether periodic images intersect
  if (!boxesIntersect && isPeriodic())
    {
      if (m_domainBox.contains(box1) && m_domainBox.contains(box2))
        {
          // no need to do any periodic checking
        }
      else
        {
          // loop over periodic shift directions, shift box1, and
          // check for intersections
          IntVect shiftMult(m_domainBox.size());
          ShiftIterator shiftIt = m_shiftIt;
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
            {
              Box shiftBox(box1);
              IntVect shiftVect = shiftMult*shiftIt();
              shiftBox.shift(shiftVect);
              if (shiftBox.intersects(box2)) boxesIntersect = true;
            }
        }
    } // end if we need to do periodic check
  return boxesIntersect;
}

void
ProblemDomain::setPeriodic(int a_dir, bool a_isPeriodic)
{
  assert (a_dir < SpaceDim);
  m_isPeriodic[a_dir] = a_isPeriodic;
  m_shiftIt.computeShifts(m_isPeriodic);
}

Box
bdryLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryLo(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
bdryHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryHi(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
adjCellLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      adjBox = adjCellLo(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;

}

Box
adjCellHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      adjBox = adjCellHi(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;
}

Box
ProblemDomain::operator&(const Box& b) const
{

  Box intersectBox;
  // check for empty -- if either box is empty, return empty intersectBox
  if (!b.isEmpty() && !isEmpty())
    {
      // do this by creating an intersection box which will
      // take the periodic case into account
      intersectBox = m_domainBox;
      if (isPeriodic())
        {
          for (int dir=0; dir<SpaceDim; dir++)
            {
              if (m_isPeriodic[dir])
                {
                  // the idea behind this is that intersect does nothing in
                  // periodic directions
                  intersectBox.setRange(dir, b.smallEnd(dir), b.size(dir));
                }
            }
        }

      intersectBox &= b;
    } // end if neither box is empty

  return intersectBox;
}

ProblemDomain&
ProblemDomain::refine(int a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
refine(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{

  if (a_probdomain.isEmpty()) return a_probdomain;
  Box newdomain = a_probdomain.domainBox();
  newdomain.refine(a_refinement_ratio);

  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ProblemDomain&
ProblemDomain::refine(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
refine(const ProblemDomain& a_probdomain, const IntVect& a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());
  newdomain.refine(a_refinement_ratio);

  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ProblemDomain&
ProblemDomain::coarsen(int a_refinement_ratio)
{
  m_domainBox.coarsen(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());

  newdomain.coarsen(a_refinement_ratio);
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ProblemDomain&
ProblemDomain::coarsen(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain,
        const IntVect& a_refinement_ratio)
{

  Box newdomain(a_probdomain.domainBox());
  newdomain.coarsen(a_refinement_ratio);

  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ostream& operator<< (ostream& os, const ProblemDomain& a_probdomain)
{
  os << a_probdomain.domainBox()
     << ", " << D_TERM( a_probdomain.m_isPeriodic[0],
                        << " " << a_probdomain.m_isPeriodic[1],
                        << " " << a_probdomain.m_isPeriodic[2]);

  if (os.fail())
    MayDay::Error("operator<<(ostream&,ProblemDomain&) failed");

  return os;
}

//
// Copied from <Utility.H>
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream& is, ProblemDomain& a_probdomain)
{

  is >> a_probdomain.m_domainBox;
  is >> ws;
  char c;
  is >> c;
  is.putback(c);
  if (c == ',')
    {
      is.ignore(CH_IGNORE_MAX, ',');
      for (int dir=0; dir<SpaceDim; dir++)
        {
          is >> a_probdomain.m_isPeriodic[dir];
        }
    }

  else
    MayDay::Error("operator>>(istream&,ProblemDomain&): expected \',\'");

  if (is.fail())
    MayDay::Error("operator>>(istream&,ProblemDomain&) failed");

    return is;
}

void
ProblemDomain::dumpOn(ostream& strm) const
{

  strm << "domainBox: ";
  m_domainBox.dumpOn(strm);
  strm << " isPeriodic = "
       << D_TERM(m_isPeriodic[0],
                 << " " << m_isPeriodic[1],
                 << " " << m_isPeriodic[2])
       << '\n';

  if (strm.fail())
    MayDay::Error("ProblemDomain::dumpOn(ostream&) failed");

}

void operator &= (Box& a_box, const ProblemDomain& a_probdomain)
{
  a_box = a_probdomain & a_box;
}

Box operator & (const Box& a_box, const ProblemDomain& a_probdomain)
{
  Box returnBox = a_probdomain & a_box;
  return returnBox;
}
