/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cassert>
#include "MayDay.H"
#include "RealVect.H"
#include "Misc.H"
#ifdef HDF5
#include "hdf5.h"
#else
#define HOFFSET(S,M)    (offsetof(S,M))
#endif
using std::ostream;
using std::istream;
using std::ws;

size_t RealVect::io_offset = HOFFSET(RealVect, vect);

const RealVect RealVect::Unit(D_DECL(1.0,1.0,1.0));
const RealVect RealVect::Zero(D_DECL(0.0,0.0,0.0));

const Real*
RealVect::dataPtr() const
{
  return vect;
}

Real*
RealVect::dataPtr()
{
  return vect;
}

RealVect::RealVect (D_DECL(Real i, Real j, Real k))
{
  D_EXPR(vect[0] = i, vect[1] = j, vect[2] = k);
}

RealVect::RealVect (const RealVect &iv)
{
  D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
}

RealVect::RealVect ()
{
  D_EXPR(vect[0]=0.0, vect[1]=0.0, vect[2] = 0.0);
}

RealVect&
RealVect::operator= (const RealVect &iv)
{
  D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
  return *this;
}

Real&
RealVect::operator[] (int i)
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

const Real&
RealVect::operator[] (int i) const
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

bool
RealVect::operator== (const RealVect& p) const
{
  return D_TERM(vect[0] == p[0], && vect[1] == p[1], && vect[2] == p[2]);
}

bool
RealVect::operator!= (const RealVect& p) const
{
  return D_TERM(vect[0] != p[0], || vect[1] != p[1], || vect[2] != p[2]);
}

RealVect&
RealVect::operator+= (Real s)
{
  D_EXPR(vect[0] += s, vect[1] += s, vect[2] += s);
  return *this;
}

RealVect&
RealVect::operator+= (const RealVect& p)
{
  D_EXPR(vect[0] += p[0], vect[1] += p[1], vect[2] += p[2]);
  return *this;
}

RealVect&
RealVect::operator*= (Real s)
{
  D_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);
  return *this;
}

RealVect&
RealVect::operator*= (const RealVect &p)
{
  D_EXPR(vect[0] *= p[0], vect[1] *= p[1], vect[2] *= p[2]);
  return *this;
}

RealVect
RealVect::operator* (const RealVect& p) const
{
  RealVect v(D_DECL(vect[0]*p[0], vect[1]*p[1], vect[2]*p[2]));
  return v;
}

RealVect
RealVect::operator* (Real s) const
{
  RealVect v(D_DECL(vect[0]*s, vect[1]*s, vect[2]*s));
  return v;
}

RealVect&
RealVect::operator/= (Real s)
{
  D_EXPR(vect[0] /= s, vect[1] /= s, vect[2] /= s);
  return *this;
}

RealVect&
RealVect::operator/= (const RealVect& p)
{
  D_EXPR(vect[0] /= p[0], vect[1] /= p[1], vect[2] /= p[2]);
  return *this;
}

//XXXRealVect
//XXXRealVect::operator/ (const RealVect & p) const
//XXX{
//XXX  RealVect result( D_DECL( vect[0] / p[0], vect[1] / p[1], vect[2] / p[2] ) );
//XXX  return result ;
//XXX}

RealVect
RealVect::operator/ (Real s) const
{
  RealVect result( D_DECL( vect[0] / s, vect[1] / s, vect[2] / s) );
  return result ;
}

RealVect&
RealVect::operator-= (const RealVect& p)
{
  D_EXPR(vect[0] -= p[0], vect[1] -= p[1], vect[2] -= p[2]);
  return *this;
}

RealVect
BASISREALV (int dir)
{
  assert(dir >= 0 && dir < SpaceDim);
  RealVect tmp = RealVect::Zero ;
  tmp.vect[dir] = 1;
  return tmp;
}

RealVect
operator/ (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(s/p[0], s/p[1], s/p[2]));
}
RealVect
operator+ (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] + s, p[1] + s, p[2] + s));
}

RealVect
operator- (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(s - p[0], s - p[1], s - p[2]));
}

RealVect
operator* (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(s * p[0], s * p[1], s * p[2]));
}

RealVect
operator/ (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(s[0] / p[0], s[1] /p[1], s[2] / p[2]));
}

RealVect
operator+ (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] + s[0], p[1] +s[1], p[2] + s[2]));
}

RealVect
operator- (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(s[0] - p[0], s[1] - p[1], s[2] - p[2]));
}

RealVect
operator* (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] * s[0], p[1] *s[1], p[2] * s[2]));
}

std::ostream&
operator<< (std::ostream& ostr, const RealVect& p)
{
  ostr << "(" << D_TERM( p[0] ,<< "," << p[1], << "," << p[2] ) << ")" ;
  if (ostr.fail()) MayDay::Error("operator<<(ostream&,RealVect&) failed");
  return ostr;
}

/*
template < >
int linearSize(const RealVect& vindex)
{
  return sizeof(RealVect);
}

//RealVect specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf)
{
  unsigned char* bob = (unsigned char*)inBuf;
  unsigned char* to = (unsigned char*)&a_outputT;
  memcpy(to + RealVect::io_offset, bob, SpaceDim*sizeof(Real));
}

//RealVect specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT)
{
  unsigned char* bob = (unsigned char*)a_outBuf;
  const unsigned char* from = (const unsigned char*)&a_inputT;
  memcpy(bob, from + RealVect::io_offset, SpaceDim*sizeof(Real));
}

*/
