/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

//
// These are the inevitable functions that people
// can't avoid using when writing a SPMD program.
// It is a minimal set from which more elaborate
// functionality can be generated.  As always, a
// user is free to utilize the entire MPI programming
// on their own platform.  The functions are
// assured to work on all machines supported.
//

#ifndef _SPMD_H_
#define _SPMD_H_

#include "REAL.H"
#include "Vector.H"
#include "Box.H"
#include "MayDay.H"
#include "RealVect.H"

class IntVectSet;

/// local process ID
/**
    Returns the ID of the locally running process in the range 0 <=
    procID() < numProc().  This has no relation to the operating system
    pid.  There is always a procID() == 0.  */
int procID();

/// number of parallel processes
/**
    Returns the number of parallel processes running.
    Always returns at least 1. */
unsigned int numProc();

///all ranks wait here to sync-up
/**
   All MPI ranks wait here to sync-up.  Calls MPI_Barrier(comm)  */
void barrier(void);

template <class T>
int linearSize(const T& inputT);

template <class T>
void linearIn(T& a_outputT, const void* const inBuf);

template <class T>
void linearOut(void* const a_outBuf, const T& inputT);

#ifdef MPI
///gather from multiple processes
/**
  Gathers a_input of every process onto Vector<T> a_outVec on process
  number a_dest.  It is an error if a_dest < 0 or a_dest >= numProc().
  a_outVec is numProc() long with the value of a_input on every
  process in its elements.\\

T must have have the following functions:
  \\

  int linearSize<T>(const T& inputT); \\
  void linearIn<T>(T& a_outputT, const void* const inBuf); \\
  void linearOut<T>(void* a_outBuf, const T& inputT); \\

*/
template <class T>
void gather(Vector<T>& a_outVec, const T& a_input, int a_dest);

// this has to be here so that linearSize<IntVectSet>, linearIn<IntVectSet>,
//  and linearOut<IntVectSet> are defined.
// template < >
// void gather(Vector<IntVectSet>& a_outVec,
//             const IntVectSet& a_input,
//             int a_dest);
#endif

///broadcast to every process
/**
  Broadcasts a_inAndOut to every process from process number a_src.
  It is an error if a_src < 0 or a_src >= numProc().  \\

  T must have have the following functions: \\

  int linearSize<T>(const T& inputT); \\
  void linearIn<T>(T& a_outputT, const void* const inBuf); \\
  void linearOut<T>(void* a_outBuf, const T& inputT); \\ */
template <class T>
void broadcast(T& a_inAndOut,  int a_src);

//integer specialization of linearSize
template < >
int linearSize(const int& a_input);

//integer specialization of linearIn
template < >
void linearIn(int& a_outputT, const void* const inBuf);

//integer specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const int& a_inputT);

//Real specialization of linearSize
template < >
int linearSize(const float& a_input);

template < >
int linearSize(const double& a_input);

//Real specialization of linearIn
template < >
void linearIn(float& a_outputT, const void* const a_inBuf);

template < >
void linearIn(double& a_outputT, const void* const a_inBuf);

//Real specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const float& a_inputT);

template < >
void linearOut(void* const a_outBuf, const double& a_inputT);

//Box specialization of linearSize
template < >
int linearSize(const Box& a_input);

//Box specialization of linearIn
template < >
void linearIn(Box& a_outputT, const void* const a_inBuf);

//Box specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const Box& a_inputT);

//Vector<int>  specialization
template < >
int linearSize(const Vector<int>& a_input);
template < >
void linearIn(Vector<int>& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<int>& a_inputT);

//Vector<Real>  specialization
template < >
int linearSize(const Vector<float>& a_input);
template < >
void linearIn(Vector<float>& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<float>& a_inputT);

template < >
int linearSize(const Vector<double>& a_input);
template < >
void linearIn(Vector<double>& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<double>& a_inputT);

//Vector<Box>  specialization
template < >
int linearSize(const Vector<Box>& a_input);
template < >
void linearIn(Vector<Box>& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<Box>& a_inputT);

//Vector<Vector<Box> > specialization
template < >
int linearSize(const Vector<Vector<Box> >& a_input);
template < >
void linearIn(Vector<Vector<Box> >& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<Vector<Box> >& a_inputT);

// RealVect spcializations of linearization
template < >
int linearSize(const RealVect& vindex);

//VolIndex specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf);

//VolIndex specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT);

//Vector<Vector<int> > specialization
template < >
int linearSize(const Vector<Vector<int> >& a_input);
template < >
void linearIn(Vector<Vector<int> >& a_outputT, const void* const inBuf);
template < >
void linearOut(void* const a_outBuf, const Vector<Vector<int> >& a_inputT);

//Vector<T> specialization of linearSize
template <class T>
int linearListSize(const Vector<T>& a_input);

//Vector<T> specialization of linearIn
template <class T>
void linearListIn(Vector<T>& a_outputT, const void* const a_inBuf);

//Vector<T> specialization of linearOut
template <class T>
void linearListOut(void* const a_outBuf, const Vector<T>& a_inputT);

class SerialTask
{
public:
    enum task { compute=0 };
};

// return id of unique processor for special serial tasks
int
uniqueProc(const SerialTask::task& a_task);

#ifdef MPI
#include <mpi.h>
struct Chombo_MPI{ static MPI_Comm comm;};
void setChomboMPIErrorHandler();

#else
// this can be changed for debugging parallel code in serial
extern int num_procs ;
#endif

#include "SPMDI.H"

#endif
