/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

//#if defined(NDEBUG)
//int  main()
//{
//  return -1;
//}
//#else

#include "HDF5.H"
#include "BRMeshRefine.H"
#include "IntVectSet.H"
#include "LoadBalance.H"
#include "DebugDump.H"
#include "UGIO.H"
#include "AMRIO.H"
#include "FABView.H"
#include <algorithm>

#include "Timer.H"
#include "memusage.H"
#include "Timer.H"
Timer Everything        ("Everything");
int global_granularity=1;

#ifdef TIMER
// global variables
int IterationFlavor;
int BIterationFlavor;
CounterStruct NoelsSooperIterationCounter[MAX_ITERATION_COUNTERS];
#endif

/// Prototypes:
void
setCircleTags(IntVectSet& ivs, int circleR, int thickness = 1,
              const IntVect& center= IntVect::TheZeroVector());
void
buildDisjointBoxLayout(DisjointBoxLayout& plan,
                       const IntVectSet& tags, const Box& domain);

void
parseTestOptions( int argc ,char* argv[] ) ;

int
test();

/// Global variables for handling output:
static const char *pgmname = "HDF5data" ;
static const char *indent = "   ", *indent2 = "      " ;
static bool verbose = true ;

/// Code:

int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  //  InitWriteFAB();
  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  if(test() != 0)
    {
      cout << indent << pgmname <<" failed"<<endl;
    }
  else
    {
      cout << indent << pgmname <<" passed"<<endl;
    }
#ifdef MPI
  MPI_Finalize();
#endif
}

struct values
{
  static void setVal1(const Box& box, int comps, BaseFab<int>& fab)
    {
      int center = (box.smallEnd()[0] + box.bigEnd()[0])/2;
      fab.setVal(center, 0);
      fab.setVal(2*center, 1);
      fab.setVal(3*center, 2);

    }
  static void setVal2(const Box& box, int comps, BaseFab<int>& fab)
    {
      int center = (box.smallEnd()[1] + box.bigEnd()[1])/2;
      fab.setVal(center);
    }

  static void setVal3(const Box& box, int comps, FArrayBox& fab)
    {
      int center = (box.smallEnd()[1] + box.bigEnd()[1])/2;
      for(int c=0; c<comps; ++c)
                {
                  for(BoxIterator bit(box); bit.ok(); ++bit)
                        {
                          fab(bit(), c) = c + center + bit()[0];
                        }
                }
    }
};

// returns 0 on all tests passed.

int test()
{
  int error;
  HDF5Handle testFile;

  assert(!testFile.isOpen());

  error = testFile.open("data.h5", HDF5Handle::CREATE);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File creation failed "<<error<<endl;
      return error;
    }

  assert(testFile.isOpen());

  Box domain(IntVect::TheZeroVector(), 20*IntVect::TheUnitVector());

  DisjointBoxLayout plan1, plan2;
  {
    IntVectSet tags;

    IntVect center = 10*IntVect::TheUnitVector();

    setCircleTags(tags, 6, 1, center); //circle/sphere

    buildDisjointBoxLayout(plan1, tags, domain);

    tags.makeEmpty();

    setCircleTags(tags, 5, 2, center);

    buildDisjointBoxLayout(plan2, tags, domain);
  }
  if( verbose )
  {
    cout << "plan1: " << procID() << "...." << plan1 << endl;
    cout << "plan2: " << procID() << "...." << plan2 << endl;
  }

  LevelData<BaseFab<int> > level1(plan1, 3, IntVect::TheUnitVector());

  LevelData<BaseFab<int> > level2;

  level2.define(level1);

  level2.define(plan2, 1);

  for(DataIterator i(level2.dataIterator()); i.ok(); ++i)
  {
    level2[i()].setVal(2);
  }

  level1.apply(values::setVal1);

  level2.apply(values::setVal2);

  HDF5HeaderData set1;
  Real dx=0.004;
  Box b1(IntVect(D_DECL(1,2,1)), IntVect(D_DECL(4,4,4)));
  Box b2(IntVect(D_DECL(5,2,1)), IntVect(D_DECL(12,4,4)));
  int currentStep = 2332;

  set1.m_string["name"] = "set1";
  set1.m_real["dx"] = dx;
  set1.m_int["currentStep"] = currentStep;
  set1.m_intvect["some intvect or other"] = b1.smallEnd();
  set1.m_box["b1"] = b1;
  set1.m_box["b2"] = b2;

  testFile.setGroupToLevel(1);

  error = write(testFile, plan1);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "box write failed "<<error<<endl;
      return error;
    }

  error = write(testFile, level1, "level1 state vector");
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayoutData 1 write failed "<<error<<endl;
      return error;
    }

  testFile.setGroupToLevel(2);

  error = write(testFile, plan2);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "box2 write failed "<<error<<endl;
      return error;
    }

  error = write(testFile, level2, "level2 state vector");
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayoutData 2 write failed "<<error<<endl;
      return error;
    }

  LevelData<FArrayBox> state(plan2, 3);

  state.apply(values::setVal3);
  testFile.setGroupToLevel(0);
  set1.writeToFile(testFile);

  error = write(testFile, plan2);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "box2 write failed "<<error<<endl;
      return error;
    }

  testFile.setGroup("/");

  error = writeLevel(testFile, 0, state, 2, 1, 0.001, b2, 2);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayoutData 2 write failed "<<error<<endl;
      return error;
    }

  set1.writeToFile(testFile);
  set1.writeToFile(testFile);

  testFile.close();

  assert(!testFile.isOpen());

  // test the utility functions ReadUGHDF5 and WriteUGHDF5

  WriteUGHDF5("UGIO.hdf5", plan2, state, domain);

  ReadUGHDF5("UGIO.hdf5", plan2, state, domain);
  //========================================================================
  //
  //  now, read this data back in
  //
  //========================================================================

  BoxLayoutData<BaseFab<int> > readlevel1, readlevel2;

  error = testFile.open("data.h5", HDF5Handle::OPEN_RDONLY);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "File open failed "<<error<<endl;
      return error;
    }

  testFile.setGroupToLevel(2);
  Vector<Box> boxes;
  error = read(testFile, boxes);

  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "box read failed "<<error<<endl;
      return error;
    }
  boxes.sort();
  Vector<int> assign;
  error = LoadBalance(assign, boxes);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayout LoadBalance failed "<<error<<endl;
      return error;
    }
  BoxLayout readplan2(boxes, assign);
  readplan2.close();
  error = read(testFile, readlevel2, "level2 state vector", readplan2);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayoutData<BaseFab<int>> read failed "<<error<<endl;
      return error;
    }

  testFile.setGroupToLevel(1);
  error = read(testFile, boxes);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "box read failed "<<error<<endl;
      return error;
    }

  error = LoadBalance(assign, boxes);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayout LoadBalance failed "<<error<<endl;
      return error;
    }
  BoxLayout readplan1(boxes, assign);
  readplan1.close();
  if( verbose )
  {
    cout << "readplan1: " << procID() << "...." << readplan1 << endl;
    cout << "readplan2: " << procID() << "...." << readplan2 << endl;
  }
  error = read(testFile, readlevel1, "level1 state vector", readplan1);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "BoxLayoutData<BaseFab<int>> read failed "<<error<<endl;
      return error;
    }

  if( verbose )
    cout << plan1<<readplan1<<endl;

  // real test of IO, make sure the data is the same coming and going

  DataIterator l1  = level1.dataIterator();
  DataIterator rl1 = readlevel1.dataIterator();
  DataIterator l2  = level2.dataIterator();
  DataIterator rl2 = readlevel2.dataIterator();

  if(level1.boxLayout().size() != readlevel1.boxLayout().size())
    {
      if( verbose )
        cout << indent2 << "level1.size() != readl1.size() read failed "<<error<<endl;
      return 1;
    }
  if(level2.boxLayout().size() != readlevel2.boxLayout().size())
    {
      if( verbose )
        cout << indent2 << "level2.size() != readl2.size() read failed "<<error<<endl;
      return 1;
    }

  // we can assume that BoxLayout IO is tested in HDF5boxIO
  BaseFab<int>* before, *after;
  for(; l1.ok(); ++l1, ++rl1)
    {

      before = &(level1[l1()]); after = &(readlevel1[rl1()]);
      for(int c=0; c<before->nComp(); ++c)
        {
          for(BoxIterator it(level1.box(l1())); it.ok(); ++it)
            {
              if((*before)(it(), c) != (*after)(it(), c))
                {
                  if( verbose )
                    cout << indent2 << "l1 != readl1 read failed "<<error<<endl;
                  return 2;
                }
            }
        }
    }

  for(; l2.ok(); ++l2, ++rl2)
    {

      before = &(level2[l2()]); after = &(readlevel2[rl2()]);
      for(int c=0; c<before->nComp(); ++c)
        {
          for(BoxIterator it(level2.box(l2())); it.ok(); ++it)
            {
              if((*before)(it(), c) != (*after)(it(), c))
                {
                  if( verbose )
                    cout << indent2 << "level2 != readlevel2 read failed "<<error<<endl;
                  return 3;
                }
            }
        }
    }

  LevelData<FArrayBox> readState;
  Real dt, time;
  int refRatio;

  testFile.setGroup("/");

  error = readLevel(testFile, 0, readState, dx, dt, time, b2, refRatio);
  if(error != 0)
    {
      if( verbose )
        cout << indent2 << "readLevel failed "<<error<<endl;
      return error;
    }

#ifndef MPI
  // OK, now try to read one FArrayBox at a time
  // problem with DataIterator and running the out-of-core in parallel, so
  // have to think about that for now  BVS.
  FArrayBox readFAB;
  Interval  interval(1,2);
  testFile.setGroup("/");
  int index=0;
  for(DataIterator dit(state.dataIterator()); dit.ok(); ++index,++dit)
        {
          FArrayBox& fab = state[dit()];
          readFArrayBox(testFile, readFAB, 0, index, interval);
          for(BoxIterator it(state.box(dit())) ; it.ok() ; ++it)
                {
                  if(readFAB(it(), 0) != fab(it(), 1))
                        {
                          if( verbose )
                                cout << indent2 << "state != after for out-of-core "<<error<<endl;
                          return 3;
                        }
                }
        }

#endif

  testFile.close();

  assert(!testFile.isOpen());

  return 0;

}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
  {
    if( argv[i][0] == '-' ) //if it is an option
    {
      // compare 3 chars to differentiate -x from -xx
      if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
      {
        verbose = true ;
        argv[i] = "" ;
      }
      else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
      {
        verbose = false ;
        argv[i] = "" ;
      }
      else
      {
        break ;
      }
    }
  }
  return ;
}

void
setCircleTags(IntVectSet& ivs, int circleR, int thickness,
              const IntVect& center)
{
  assert(circleR > 0);
  assert(thickness < circleR && thickness > 0);
  for(int t = 0; t<thickness; ++t){
    for(int x=-circleR-t; x<=circleR+t; ++x)
    {
      // not perfect, but I'm not fussy here
      int Y = (int)sqrt((Real)abs(circleR*circleR - x*x));
#if CH_SPACEDIM == 2
      // circle
      ivs |= center+IntVect(x,Y);
      ivs |= center+IntVect(x,-Y);
#elif CH_SPACEDIM == 3
      //sphere
      for(int y=-Y; y<=Y; ++y)
      {
        int Z = (int)sqrt((Real)abs(circleR*circleR - x*x - y*y));
        ivs |= center+IntVect(x,y,Z);
        ivs |= center+IntVect(x,y,-Z);
      }
#else
#error arrayTest failed: implemented only for 2D and 3D
#endif
    }
  }
}

void
buildDisjointBoxLayout(DisjointBoxLayout& plan,
                       const IntVectSet& tags, const Box& domain)
{
  Vector<Box> vectBox;
  Vector<int> assignments;

  IntVectSet domainivs(domain);
  int maxsize = 128;
  Vector<int> fakeNRef(2,2);
  Real fillRatio = 0.5;
  int bufferSize = 0;
  int blockFactor = 1;
  BRMeshRefine mr(domain, fakeNRef, fillRatio, blockFactor,
                  bufferSize, maxsize);

  mr.makeBoxes(vectBox, tags, domainivs, domain, maxsize);

  vectBox.sort();
  int stat = LoadBalance(assignments,  vectBox);
  if( stat != 0 ) {
    cout << indent << pgmname
         << ":  loadBalance() FAILED." << endl ;
    exit(stat) ;
  }

  plan.define(vectBox, assignments);

  plan.close();
};

//#endif
