/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdio>
#include <cmath>
#include "Box.H"
#include "SPMD.H"
#include "REAL.H"
#include "MayDay.H"
#include "IntVectSet.H"
#include "BRMeshRefine.H"
#include "parstream.H"
#include <iostream>
#include <fstream>
#include <string>
#ifdef MPI
#include <mpi.h>
#endif

using std::endl;
using std::cout;
using std::cerr;

/// Global variables for test output:
static const char *pgmname = "domainSplitTest" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

const int destProc = 0;

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;
/**
   domainSplitTest returns:
   0: all tests passed
   1: boxes coming out of domain split are not in
   the original domain.
   2: boxes coming out of domain split have a side
   that is too long
   3: domainsplit returned an error code
   4: boxes do not cover  domain
  */
int domainSplitTest(void);

int main(int argc, char* argv[])
{
  using std::ostream;
  bool passed = true;
  int icode = 0;

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    pout () << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = domainSplitTest();

  if (procID() == destProc)
  {
      if(icode != 0)
      {
          cout << indent << pgmname
               << ": failed with error code " << icode
               << endl;
          passed = false;
      }
      if (passed)
      {
          cout << indent << pgmname
               << ": passed all tests"
               << endl;
      }
  }

#ifdef MPI
  MPI_Finalize();
#endif

  return 0;
}
int domainSplitTest()
{
  int retflag = 0;
  int domlen = 100;
  int maxlen = 25;
  int blockf = 2;

  Box domain(IntVect::Zero, (domlen-1)*IntVect::Unit);

  Vector<Box> vbox;

  domainSplit(domain, vbox, maxlen, blockf);

  IntVectSet ivsDom(domain);
  for(int ibox = 0; ibox < vbox.size(); ibox++)
    {
      const Box& boxloc = vbox[ibox];
      if(!domain.contains(boxloc))
        {
          if(verbose)
            cerr << "error: " << boxloc << "not contained in " << domain  << endl;
          return 1;
        }
      ivsDom -= boxloc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          if(boxloc.size(idir) > maxlen)
            {
              if(verbose)
                cerr << "error: " << boxloc << "is too big in " << idir
                     << " direction (maxlen == " << maxlen << ")"  << endl;
              return 2;
            }
        }
    }
  if(!ivsDom.isEmpty())
    {
      if(verbose)
        cerr << "error: domain not covered " << endl;
      return 4;
    }
  if(verbose)
    {
      cout << "domain = " << domain << endl;
      cout << "max box size = " << maxlen << endl;
      cout << "boxes = " << endl;
      for(int ibox = 0; ibox < vbox.size(); ibox++)
        {
          const Box& boxloc = vbox[ibox];
          cout << boxloc << endl;
        }
      cout << "box sizes = " << endl;
      for(int ibox = 0; ibox < vbox.size(); ibox++)
        {
          const Box& boxloc = vbox[ibox];
          for(int idir = 0; idir < SpaceDim; idir ++)
            cout << boxloc.size(idir)  << "   ";
          cout << endl;
        }
    }

  return retflag;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
