/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// Purpose:
//  Test the infrastucture for periodic boundary conditions

// Usage:
//  <program-name> [-q|-v] ...
//
//  where:
//    -q means run quietly (only pass/fail messages printed)
//    -v means run verbosely (all messages printed)
//    -write means write out the boxes; otherwise will read in
//           boxes from a file and compare with boxes generated
//           in tests.
//    ... all non-option arguments are ignored (Chombo convention)
//
//  Default is `-v'
//
//  Unknown options are treated as errors.

////////////////////////////////////////////////////////////////

// Include files:
#include <cstdio>

#include "parstream.H"
#include "LoadBalance.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"
#include "BRMeshRefine.H"
#include "AMRIO.H"

#if 0
#include "GhostBC.H"
#include "PoissonBC.H"
#endif

#ifdef MPI
#include <mpi.h>
#endif

//////////////////////////////////////////////////////////////
using std::cout;
using std::endl;

void parseTestOptions(int argc, char* argv[]);

void initData(LevelData<FArrayBox>& a_data, const Real a_dx);

/// Global variables for handling output

static const char *pgmname = "testPeriodic";
static const char *indent = "   ";
//static const char *indent2 = "      " ;

static bool verbose = true ;
// if true, use arrayview to display results
static bool display = false;

/// Code:

int
main(int argc ,char *argv[] )
{

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions(argc, argv);

  // establish periodic domain -- first multiply periodic in all directions
  int status = 0;
  int baseDomainSize = 8;
  int numGhost = 2;
  IntVect ghostVect(numGhost*IntVect::Unit);

  Box baseDomBox(IntVect::Zero, (baseDomainSize-1)*IntVect::Unit);
  ProblemDomain baseDomain(baseDomBox);

  // set periodic in all directions
  for (int dir=0; dir<SpaceDim; dir++) {
    baseDomain.setPeriodic(dir, true);
  }

#if 0
  // set up physical BC class for partially periodic tests
  DomainGhostBC physBC;
  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.begin(); sit.ok(); ++sit)
      {
        // it should be easy to see if Neumann bcs are doing the right thing
        NeumannBC thisBC(dir,sit());
        physBC.setBoxGhostBC(thisBC);
      }
  }
#endif

  {
    // first try a single box on this level
    Vector<Box> levelBoxes(1, baseDomBox);
    Vector<int> procAssign(1, 0);
    DisjointBoxLayout levelGrids(levelBoxes,procAssign, baseDomain);

    LevelData<FArrayBox> tester(levelGrids, 1, ghostVect);

    Real dx = 1.0/baseDomainSize;

    initData(tester, dx);

    // test exchange in this case -- should fill ghost cells with
    // periodically copied data
    Interval comps = tester.interval();
    tester.exchange(comps);
  }

  // now try refining things, this time look at 2 boxes
  baseDomain.refine(2);
  {
    const Box domainBox = baseDomain.domainBox();
    Vector<Box> levelBoxes1(1, domainBox);
    Vector<int> procAssign1(1,0);
    DisjointBoxLayout levelGrids1(levelBoxes1, procAssign1, baseDomain);

    LevelData<FArrayBox> tester0(levelGrids1, 1, IntVect::Zero);
    LevelData<FArrayBox> tester1(levelGrids1, 1, ghostVect);

    Real dx = 1.0/(2.0*baseDomainSize);
    initData(tester1, dx);
    initData(tester0, dx);

    // first test exchange for single box
    Interval comps = tester1.interval();
    tester1.exchange(comps);

    // now try multiple boxes
    Vector<Box> levelBoxes2(3);
    Vector<int> procAssign2(3);

    levelBoxes2[0] = Box(IntVect::Zero, 9*IntVect::Unit);
    levelBoxes2[1] = Box(IntVect(D_DECL(7,10,10)), 15*IntVect::Unit);
    // this box should fail the disjointness test
    //levelBoxes2[1] = Box(IntVect(D_DECL(7,10,10)), 17*IntVect::Unit);
    levelBoxes2[2] = Box(IntVect(D_DECL(11,0,0)), IntVect(D_DECL(15,7,7)));

    int loadbalancestatus = LoadBalance(procAssign2, levelBoxes2);
    assert (loadbalancestatus == 0);

    DisjointBoxLayout levelGrids2;
    levelGrids2.define(levelBoxes2, procAssign2, baseDomain);
    LevelData<FArrayBox> tester2(levelGrids2, 1, ghostVect);

    tester0.copyTo(comps, tester2, comps);
  }

  {
    // now test partially periodic case (also coarsen back to original domain
    baseDomain.coarsen(2);
    baseDomain.setPeriodic(0,false);

    // try similar tests to the second case, only coarsened
    const Box domainBox = baseDomain.domainBox();
    Vector<Box> levelBoxes1(1, domainBox);
    Vector<int> procAssign1(1,0);
    DisjointBoxLayout levelGrids1(levelBoxes1, procAssign1, baseDomain);

    LevelData<FArrayBox> tester0(levelGrids1, 1, IntVect::Zero);
    LevelData<FArrayBox> tester1(levelGrids1, 1, ghostVect);

    Real dx = 1.0/(baseDomainSize);
    initData(tester1, dx);
    initData(tester0, dx);

    // first test exchange for single box
    Interval comps = tester1.interval();
    tester1.exchange(comps);

#if 0
    // now set physcial bcs
    DataIterator dit1 = tester1.dataIterator();
    for (dit1.begin(); dit1.ok(); ++dit1)
      {
        FArrayBox& thisData = tester1[dit1()];
        physBC.applyHomogeneousBCs(thisData, baseDomain, dx);
      }
#endif

    // now try multiple boxes
    Vector<Box> levelBoxes2(3);
    Vector<int> procAssign2(3);

    levelBoxes2[0] = Box(IntVect::Zero, 4*IntVect::Unit);
    levelBoxes2[1] = Box(IntVect(D_DECL(3,5,5)), 7*IntVect::Unit);
    // this box should fail the disjointness test
    //levelBoxes2[1] = Box(IntVect(D_DECL(3,5,5)), 10*IntVect::Unit);
    levelBoxes2[2] = Box(IntVect(D_DECL(6,0,0)), IntVect(D_DECL(7,1,1)));

    int loadbalancestatus = LoadBalance(procAssign2, levelBoxes2);
    assert (loadbalancestatus == 0);

    DisjointBoxLayout levelGrids2(levelBoxes2, procAssign2, baseDomain);
    LevelData<FArrayBox> tester2(levelGrids2, 1, ghostVect);

    tester0.copyTo(comps, tester2, comps);

#if 0
    DataIterator dit2 = tester2.dataIterator();
    for (dit2.begin(); dit2.ok(); ++dit2)
      {
        FArrayBox& thisData = tester2[dit2()];
        physBC.applyInhomogeneousBCs(thisData, baseDomain, dx);
      }
#endif
  }

  // i'm thinking all pout()'s should be before MPI_Finalize... (ndk)
  pout() << indent << pgmname << ": "
     << ( (status == 0) ? "passed all tests" : "failed at least one test,")
     << " status = " << status
     << endl;

#ifdef MPI
  MPI_Finalize();
#endif

  return status ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-d" ,3 ) == 0 )
            {
              display = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-p" ,3 ) == 0 )
            {
              display = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}

void
initData(LevelData<FArrayBox>& a_data, const Real a_dx)
{

  DataIterator dit = a_data.dataIterator();
  const DisjointBoxLayout& interiorBoxes = a_data.getBoxes();

  for (dit.begin(); dit.ok(); ++dit)
    {
      // first set to a bogus value which will persist in ghost cells
      // after initialization
      a_data[dit()].setVal(1.0e9);

      // this will be slow, but who cares?
      FArrayBox& localData = a_data[dit()];
      BoxIterator boxIt(interiorBoxes[dit()]);
      Real localVal;
      for (boxIt.begin(); boxIt.ok(); ++boxIt)
        {
          const IntVect& loc = boxIt();
          localVal = a_dx*(D_TERM(loc[0]+0.5,
                                + loc[1]+0.5,
                                + loc[2]+0.5));
          localData(loc) = localVal;
        }

    }
}
