/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "REAL.H"
#include "Box.H"
#include "Vector.H"
#include "TreeIntVectSet.H"
#include "BoxIterator.H"
#ifdef MPI
#include <mpi.h>
#endif

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
testTreeIntVectSet();

/// Global variables for handling output:
static const char *pgmname = "testTreeIntVectSet" ;
static const char *indent = "   ", *indent2 = "      " ;
static bool verbose = true ;

/// Code:

int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  int ret = testTreeIntVectSet() ;
  if(ret == 0)
    cout << indent2 << pgmname << " passed." << endl ;
  else
    cout << indent2 << pgmname << " failed." << endl ;
#ifdef MPI
  MPI_Finalize();
#endif
  return ret;
}

int
testTreeIntVectSet()
{
  {
    Box b1(IntVect(D_DECL(13,5,-1)),
           31 * IntVect::TheUnitVector());
    Box b2(32 * IntVect::TheUnitVector(),
           53 * IntVect::TheUnitVector());
    Box b3(14 * IntVect::TheUnitVector(),
           53 * IntVect::TheUnitVector());

    /*
    TreeIntVectSet ivs(b);

    ivs -= b2;
    ivs -= b1;

    for(TreeIntVectSetIterator it(ivs); it.ok(); ++it)
      {
        if(b2.contains(it()) || b1.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "subtraction left in bad data: ";
                cout << pgmname << endl ;
              }
            return 1;
          }

        if(!b3.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "IntVect beyond domain: ";
                cout << pgmname << endl ;
              }
            return 1;
          }
      }
    */

    int numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();

    TreeIntVectSet ivs2(b3);

    ivs2 |= b2;

    ivs2 |= b1;

    cout <<endl;
    int count=0;

    TreeIntVectSet otherSet(ivs2);

    TreeIntVectSetIterator it(otherSet), it2(ivs2);

    for(it.begin(); it.ok(); ++it)
      {
        if(!(b2.contains(it()) || b1.contains(it()) || b3.contains(it())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region: ";
                cout << pgmname << endl ;
              }
            return 2;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "count != numpts from union ";
        cout << pgmname << endl ;
        return 2;
      }

    IntVect shifter(D_DECL(1,-2,3));
    ivs2.shift(shifter);
    b1.shift(shifter);
    b2.shift(shifter);
    b3.shift(shifter);
    for(it2.begin(); it2.ok(); ++it2)
      {
        if(!(b2.contains(it2()) || b1.contains(it2()) || b3.contains(it2())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region: shift";
                cout << pgmname << endl ;
              }
            return 2;
          }
      }

    b1.refine(4);
    b2.refine(4);
    b3.refine(4);
    ivs2.refine(4);
    numpts = b2.numPts()+b3.numPts() + b1.numPts() -(b2&b3).numPts() -
      (b2&b1).numPts() - (b1&b3).numPts();
    count = 0;
    for(it2.begin(); it2.ok(); ++it2)
      {
        if(!(b2.contains(it2()) || b1.contains(it2()) || b3.contains(it2())))
          {
            if( verbose )
              {
                cout << indent2 << "union point falls outside region:refine ";
                cout << pgmname << endl ;
              }
            return 2;
          }
        count++;
      }
    if(count != numpts)
      {
        cout << indent2 << "count != numpts from union ";
        cout << pgmname << endl ;
        return 2;
      }

  }

  {
    IntVect a(D_DECL(-2,-1,0)), b(D_DECL(0,1,1)), c(D_DECL(14,3,4)),
      d(D_DECL(16,6,5));
    Box b1(a,c);
    Box b2(b,d);
    Box b3(b,c);
    TreeIntVectSet ivs;
    ivs|= b1;
    ivs&= b2;

    long count = 0;
    for(TreeIntVectSetIterator it(ivs); it.ok(); ++it)
      {
        if(!b3.contains(it()))
          {
            if( verbose )
              {
                cout << indent2 << "intersection point falls outside region:";
                cout << pgmname << endl ;
              }
            return 2;
          }
        count++;
      }
    if(count != b3.numPts())
      {
        cout << indent2 << "count != numpts from intersection ";
        cout << pgmname << endl ;
        return 2;
      }

    BoxIterator bit(b1);
    for(bit.begin(); bit.ok(); ++bit)
      {
        if(b2.contains(bit()))
          {
            if(!ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "intersection point missing";
                    cout << pgmname << endl ;
                  }
              }
          }
        else
          {
            if(ivs.contains(bit()))
              {
                if( verbose )
                  {
                    cout << indent2 << "extraneous intersection point";
                    cout << pgmname << endl ;
                  }
              }
          }
      }
  }
  return 0;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
