/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// This program tests ChomboFortran.

// Note: the standard_io tests are not done by default because
//       they can't automatically determine if the result is
//       correct and they might fail by core-dumping.
// Note: the MayDay tests are not done by default because they
//       are _intended_ to core dump if they succeed.

// Usage: a.out [-mqsv] [args...]*
//
// Where:
//   -m selects the mayday tests
//   -q selects `quiet' execution: only pass/fail messages are printed
//   -s selects the standard_io tests
//   -v selects `verbose' execution: all messages are printed
//   args... are ignored.
//
// Defaults:  -q
//
// Returns:
//  The exit code of the program is the result code from the first
//  test routine that fails.

#include <iostream>
using std::cout ;
using std::endl ;
#include <cstring>
#include "ChFIOF_F.H"
#ifdef MPI
#include <mpi.h>
#endif
/// Prototypes:
void
parseTestOptions(bool *option_s ,bool *option_m ,int argc ,char* argv[]) ;

/// Global variables for handling output:
static const char* pgmname = "testChFIO" ;
static const char* indent = "   ";
static const char* indent2 = "      " ;
static bool verbose = true ;

int
main(int argc ,char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  bool test_stdio = false ; //dont do this test by default
  bool test_mayday = false ; //dont do this by default either

  parseTestOptions( &test_stdio ,&test_mayday ,argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int i1 = 23 ;
  Real r1 = 23.45 ;

  // Test File input/output in Chombo Fortran
  int status;
  FORT_CHF_FIO( CHF_CONST_INT( i1 ) ,CHF_CONST_REAL( r1 ), CHF_INT(status) ) ;

  if( status == 0 )
    cout << indent << pgmname << " passed the file I/O tests." << endl ;
  else
    cout << indent << pgmname << " failed the file I/O tests with return code " << status << endl ;

  // Test standard input/output in ChF
  if( test_stdio )
    {
      cout << indent << pgmname << ": running std I/O tests.  You should see some output" << endl ;
      int status2;
      FORT_CHF_STDIO( CHF_CONST_INT( i1 ) ,CHF_CONST_REAL( r1 ), CHF_INT(status2) ) ;
      if( status2 == 0 )
        {
          cout << indent << pgmname << " passed the std I/O tests." << endl ;
        }
      else
        {
          cout << indent << pgmname << " failed the std I/O tests with return code " << status << endl ;
          if( status == 0 ) status = status2 ;
        }

    }

  // Test calling MayDay routines (testing only one is enough).
  if( test_mayday )
    {
      // this shouldn't return, but if it does, something went wrong
      cout << indent << pgmname << ": running MayDay tests. An Error or Abort should happen." << endl ;
      int status2;
      FORT_CHF_MAYDAY(CHF_INT(status2) ) ;
      cout << indent << pgmname << " failed the MayDay tests with return code " << status << endl ;
      if( status == 0 ) status = status2 ;
    }
#ifdef MPI
  MPI_Finalize();
#endif
  return status ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Also parse the `-s' option to set the first argument
///
void
parseTestOptions(bool *option_s, bool *option_m,int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-s" ,3 ) == 0 )
            {
              *option_s = true ;
            }
          else if( strncmp( argv[i] ,"-m" ,3 ) == 0 )
            {
              *option_m = true ;
            }
        }
    }
  return ;
}
