/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// PhysBCUtil.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _PhysBCUtil_H_
#define _PhysBCUtil_H_

#include "GhostBC.H"
#include "Tuple.H"
#include "ParmParse.H"
#include "LoHiSide.H"
#include "basicVelBC.H"

// include support for *IBC stuff from AMRGodunov, for use with
// PatchGodunov-derived stuff
#ifdef NEW_ADVECTION
#include "OldPhysIBC.H"
#endif


/// Encapsulates all physical BC's for IAMR solves
/** This class is designed to contain all the information
needed to specify all physical boundary conditions needed
for the IAMR code. This includes BC's for velocity, pressures,
scalars.  While basic solid-wall BCs are coded in to this
class, the idea is that for a more complicated set of physical
boundary conditions (inflow, etc) a problem-dependent class
will be derived from this one which will override the generic
BC's in this class */
class PhysBCUtil
{

public:
  /*@ManMemo: this will access the ParmParse database using the
    physBC prefix and use it to define the physical BC types (hi 
    and lo) in each direction.  The BC type "custom" will imply
    that the BC is a somewhat nonstandard one, but that the 
    derived class will know what to do with it. */
  PhysBCUtil();

  ///
  virtual ~PhysBCUtil();

  /// 
  enum BCType {
    ///
    bogusBC = -1,
    ///
    SolidWall,
    ///
    Inflow,
    ///
    Outflow, 
    ///
    Symmetry,
    ///
    noShear,
    ///
    Custom,
    ///
    NUM_PHYS_BC_TYPES};



  /// "virtual" constructor
  virtual PhysBCUtil* newPhysBCUtil() const;

  /// Pressure BCs

  /// returns pressure BC for MAC projection
  virtual DomainGhostBC* MacPressureBC() const;

  /// returns BC applied to MAC pressure before computing gradient
  virtual DomainGhostBC* gradMacPressureBC() const;

  /// returns pressure BC for level projection
  virtual DomainGhostBC* LevelPressureBC() const;

  /// returns BC applied to Pi before computing gradient
  virtual DomainGhostBC* gradPiBC() const;

  /// returns pressure BC for sync projection
  virtual DomainGhostBC* SyncProjBC() const;

  /// returns BC applied to sync correction before computing gradient
  virtual DomainGhostBC* gradESyncBC() const;

  /// returns BC for freestream preservation correction
  virtual DomainGhostBC* FreestreamCorrBC() const;

  /// BC for grad eLambda
  virtual DomainGhostBC* gradELambdaBC() const;

  /// returns pressure BC for initial velocity projection
  virtual DomainGhostBC* InitialProjBC() const;

  /// BC for gradient of correction for initial projection
  virtual DomainGhostBC* gradInitialProjCorrBC() const;

  
  /// Velocity BC's
  
  /// returns velocity BC for viscous source
  virtual basicVelBC* viscousVelBC() const;

  /// returns velocity BC for tracing (inviscid)
  virtual basicVelBC* tracingVelBC() const;

  /// returns edge-centered vel BC for edge-ctrd vel (single component)
  virtual basicEdgeVelBC* advectionVelBC() const;

  /// returns edge-centered velocity BC for edge-centered vel (multicomponent)
  virtual basicEdgeVelBC* edgeVelBC() const;

  /// returns BC on cell-centered u-dot-del-u term
  virtual basicVelBC* uDelUBC(bool a_isViscous) const;

  /// returns multi-component BC for viscous solves
  virtual DomainGhostBC* viscousSolveBC() const;


  /// returns single-component BC for viscous solves
  virtual DomainGhostBC* viscousSolveBC(int a_dir) const;

  /// returns pre-projection velocity BC
  virtual basicVelBC* uStarBC(bool a_isViscous = true) const;
  
  /// returns single-component BC for viscous refluxing solves
  virtual DomainGhostBC* viscousRefluxBC(int a_dir) const;

#ifdef NEW_ADVECTION
  /// returns a BC object compatible with AMRGodunov advection infrastructure
  virtual OldPhysIBC* advectionVelIBC() const;
#endif

  /// Scalar BC's

  ///
  virtual DomainGhostBC* diffusiveBC(int a_scalarType) const;

  ///
  virtual DomainGhostBC* scalarTraceBC(int a_scalarType) const;
  
  ///
  virtual DomainGhostBC* scalarDiffusionSolveBC(int a_scalarType) const;

  ///
  virtual DomainGhostBC* scalarRefluxSolveBC(int a_scalarType) const;

#ifdef NEW_ADVECTION
  /// returns a BC object compatible with AMRGodunov advection infrastructure
  virtual OldPhysIBC* scalarTraceIBC(int a_scalarType) const;
#endif

  /// volume discrepancy BC's
  virtual DomainGhostBC* lambdaBC() const;

#ifdef NEW_ADVECTION
  ///
  virtual OldPhysIBC* lambdaTraceIBC() const;
#endif

  /// Miscellaneous BC's

  /// BC on viscous source term itself
  virtual basicVelBC* viscousSrcBC() const;

  /// BC on diffusive source term itself
  virtual basicVelBC* diffusiveSrcBC(int scalarType) const;

  // BC on stream function (used when computing stream func. from vorticity)
  virtual DomainGhostBC* streamFunctionBC(const Interval& a_comps) const;


  /// miscellaneous utility functions

  /// sets cell spacing
  void Dx(const Real a_dx);

  /// returns cell spacing
  Real Dx() const;

  /// sets time
  void Time(const Real a_time);

  /// returns current time
  Real Time() const;

  ///
  int loBC(int a_dir) const;
  
  ///
  int hiBC(int a_dir) const;

  /// 
  int getBC(int a_dir, const Side::LoHiSide& a_side) const;

  ///  this is in case the BC's have an effect on the timestep
  /** pass in currently computed dt, along with cfl and dx.  If
      effect of boundary conditions requires a decreased timestep,
      the newly reduced timestep is returned. In the default case,
      this just returns a_dt back; however, derived classes may
      actually have an effect.*/
  virtual void computeBoundaryDt(Real& a_dt, Real a_cfl, Real a_dx) const;




protected:
  /// sets basic physical BC on pressure (called by public BC functions)
  virtual DomainGhostBC* basicPressureBC(bool a_isHomogeneous) const;
  

  /// sets basic physical BC on pressure before taking gradients
  virtual DomainGhostBC* basicGradPressureBC() const;

  /// basic physical BC on cell-centered velocity
  virtual basicVelBC* basicCCVelBC(bool a_isHomogeneous,
				   bool a_isViscous) const;


  /// returns a single-component BC for cell-centered velocity
  virtual void basicCCVelBC(DomainGhostBC& a_newBC,
			    bool a_isHomogeneous,
			    bool a_isViscous,
			    int a_velComponent,
			    Interval& a_interval) const;
  
  
  ///
  virtual basicEdgeVelBC* basicECVelBC(bool a_isHomogeneous,
				       bool a_isViscous,
				       bool a_isSingleComponent) const;

  
  ///
  virtual DomainGhostBC* basicScalBC(bool a_isHomogeneous,
				     int a_scalType) const;

  ///  interact with ParmParse to set physical BC types
  virtual void setBCs();

  // contains the enumerated values for the lo-end BC
  Tuple<int, SpaceDim> m_loBC;

  // contains the enumerated values for the hi-end BC
  Tuple<int, SpaceDim> m_hiBC;


private:
  /// cell spacing
  Real m_dx;

  /// current time
  Real m_time;

  /// copy constructor
  PhysBCUtil(const PhysBCUtil& rhs);

  ///
  virtual PhysBCUtil& operator=(const PhysBCUtil& rhs);



};

#endif


