/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef  VEL_IBC_H_
#define  VEL_IBC_H_

#include  <iostream>

#include "LevelData.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RealVect.H"
#include "AMRIO.H"

#include "OldPhysIBC.H"

/// IBC for simple Velocity advection (solid wall) boundary conditions
/**
   This IBC sets normal component of velocities to the prescribed value
   (normally 0) at physical boundaries, and uses extrapolation to set 
   tangential velocities.

 */
class VelIBC:public OldPhysIBC
{
public:
  /// Null constructor
  /**
   */
  VelIBC();
  
  /// Constructor which defines parameters used by Fortran routines
  /**
   */
  VelIBC(const Real&     a_gamma,
            const Real&     a_ambientDensity,
            const Real&     a_deltaDensity,
            const int&      a_pressure,
            const RealVect& a_center,
            const Real&     a_size,
            const RealVect& a_velocity);

  /// Destructor
  /**
   */
  ~VelIBC() {};


  /// Factory method - this object is its own factory
  /**
     Return a pointer to a new OldPhysIBC object with m_isDefined = false (i.e.,
     its define() must be called before it is used) and m_isFortranCommonSet
     set to value of m_isFortranCommonset in the current (factory) object.
   */
  OldPhysIBC *new_physIBC();

  /// set default boundary values
  void setDefaultValues();

  /// Set boundary fluxes
  /**
   */
  void fluxBC(FArrayBox&            a_F,
              const FArrayBox&      a_W,
              const FArrayBox&      a_Wextrap,
              const int&            a_dir,
              const Side::LoHiSide& a_side,
              const Real&           a_time);
  
  /// Set boundary slopes
  /**
     The boundary slopes in a_dW are already set to one sided difference
     approximations.  If this function doesn't change them they will be
     used for the slopes at the boundaries.
   */
  void setBdrySlopes(FArrayBox&       a_dW,
                     const FArrayBox& a_W,
                     const int&       a_dir,
                     const Real&      a_time);
  
  /// Set up initial conditions
  /**
   */
  void initialize(LevelData<FArrayBox>& a_U);
  
  /// set velocity
  void advectionVel(const RealVect& a_advVel);

  ///
  const RealVect& advectionVel() const;

  /// set problem type
  void probType(const int a_probtype);

  /// set boundary value (default is 0)
  void setNormalWallVel(Real a_bcVal, int a_dir,
                        Side::LoHiSide a_hiLo);

  /// access boundary value
  Real getNormalWallVel(int a_dir, Side::LoHiSide a_hiLo) const;
  
  /// set slope value (default is 0)
  void setSlopeValue(Real a_slopeVal, int a_dir,
                     Side::LoHiSide a_hiLo);

  /// access slope value
  Real getSlopeValue(int a_dir, Side::LoHiSide a_hiLo) const;

  /// accessor
  int probType() const;
  
protected:
  
  /// advection velocity field
  RealVect m_velocity;

  Real m_bcVal[SpaceDim][2];
  Real m_slopeVal[SpaceDim][2];

  bool m_isBCvalSet;
  bool m_isSlopeValSet;

  /// initial condition type
  int m_probtype;

};

#endif

