/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// basicVelBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "basicVelBC.H"

// -------------------------------------------------------------
basicVelBC::basicVelBC()
{
  // explicitly set pointers to NULL
  for (int comp = 0; comp<SpaceDim; comp++) {
    m_componentBC[comp] = NULL;
  }
}

// --------------------------------------------------------------
basicVelBC::~basicVelBC() 
{
  for (int comp=0; comp<SpaceDim; comp++) {
    if (m_componentBC[comp] != NULL) {
      delete m_componentBC[comp];
      m_componentBC[comp] = NULL;
    }
  }
}


// --------------------------------------------------------------
void
basicVelBC::setBCType(const DomainGhostBC& a_ghostBC, int a_component) 
{

  if (m_componentBC[a_component] == NULL) {
    m_componentBC[a_component] = new DomainGhostBC;
    DomainGhostBC& thisBC = *m_componentBC[a_component];
    thisBC = a_ghostBC;    
  } else {
    DomainGhostBC& thisBC = *m_componentBC[a_component];
    thisBC = a_ghostBC;
  }
}



// -----------------------------------------------------------------
void
basicVelBC::applyBCs(LevelData<FArrayBox>& a_stateMF, 
		     const Box& a_domain, const Real a_dx) const
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_stateMF, physdomain, a_dx);
}


// -----------------------------------------------------------------
void
basicVelBC::applyBCs(LevelData<FArrayBox>& a_stateMF, 
		     const ProblemDomain& a_domain, 
                     const Real a_dx) const
{
  
  // just loop through boxes and call single-grid version
  DataIterator dit=a_stateMF.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) {
    applyBCs(a_stateMF[dit()], a_domain, a_dx);
  }
}


// ------------------------------------------------------------
void
basicVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
		     const Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, a_domain, a_dx);
}

// ------------------------------------------------------------
void
basicVelBC::applyBCs(FArrayBox& a_state, const ProblemDomain& a_domain,
		     const Real a_dx) const 
{
  
  // basic assumption here that all velocity BC's are inhomogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class. 

  // loop over velocity components and apply appropriate BC's
  for (int comp = 0; comp<SpaceDim; comp++) {
    DomainGhostBC& thisBC = *(m_componentBC[comp]);
    thisBC.applyInhomogeneousBCs(a_state, a_domain, a_dx);
  }

}

// ------------------------------------------------------------
void
basicVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
		   const Real a_dx, int a_comp) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, a_domain, a_dx);
}


// ------------------------------------------------------------
void
basicVelBC::applyBCs(FArrayBox& a_state, 
                     const ProblemDomain& a_domain,
                     const Real a_dx, int a_comp) const 
{
  
  // basic assumption here that all velocity BC's are inhomogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class
  

  DomainGhostBC& thisBC = *(m_componentBC[a_comp]);
  thisBC.applyInhomogeneousBCs(a_state, a_domain, a_dx);


}


// access function
const DomainGhostBC& 
basicVelBC::DomGhostBC(const int dir) const
{
  assert (dir < SpaceDim);

  return *m_componentBC[dir];
}




// -------------------------------------------------------------
basicEdgeVelBC::basicEdgeVelBC()
{
  // explicitly set pointers to NULL
  for (int comp = 0; comp<SpaceDim; comp++) {
    m_componentBC[comp] = NULL;
  }
}

// --------------------------------------------------------------
basicEdgeVelBC::~basicEdgeVelBC() 
{
  for (int comp=0; comp<SpaceDim; comp++) {
    if (m_componentBC[comp] != NULL) {
      delete m_componentBC[comp];
      m_componentBC[comp] = NULL;
    }
  }
}


// --------------------------------------------------------------
void
basicEdgeVelBC::setBCType(const DomainGhostBC& a_ghostBC, int a_component) 
{

  if (m_componentBC[a_component] == NULL) {
    m_componentBC[a_component] = new DomainGhostBC;
    DomainGhostBC& thisBC = *m_componentBC[a_component];
    thisBC = a_ghostBC;    
  } else {
    DomainGhostBC& thisBC = *m_componentBC[a_component];
    thisBC = a_ghostBC;
  }
}



// -----------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(LevelData<FluxBox>& a_stateMF, 
		     const ProblemDomain& a_domain, const Real a_dx) const
{
  ProblemDomain physdomain(a_domain);
  // just loop through boxes and call single-grid version
  DataIterator dit=a_stateMF.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) {
    applyBCs(a_stateMF[dit()], physdomain, a_dx);
  }
}


// -----------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(LevelData<FluxBox>& a_stateMF, 
		     const Box& a_domain, const Real a_dx) const
{
  
  // just loop through boxes and call single-grid version
  DataIterator dit=a_stateMF.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) {
    applyBCs(a_stateMF[dit()], a_domain, a_dx);
  }
}


// ---------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FluxBox& a_edgeState, const Box& a_domain, 
			 const Real a_dx) const
{
  ProblemDomain physdomain(a_domain);
  // loop over directions and call single-fab version
  for (int dir=0; dir<SpaceDim; dir++) {
    DomainGhostBC& thisBC = *m_componentBC[dir];
    thisBC.applyInhomogeneousBCs(a_edgeState[dir],physdomain, a_dx);
  }
}



// ---------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FluxBox& a_edgeState, const ProblemDomain& a_domain, 
			 const Real a_dx) const
{
  // loop over directions and call single-fab version
  for (int dir=0; dir<SpaceDim; dir++) {
    DomainGhostBC& thisBC = *m_componentBC[dir];
    thisBC.applyInhomogeneousBCs(a_edgeState[dir],a_domain, a_dx);
  }
}


// ------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
                         const Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, physdomain, a_dx);
}


// ------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FArrayBox& a_state, 
                         const ProblemDomain& a_domain,
                         const Real a_dx) const 
{
  
  // basic assumption here that all velocity BC's are inhomogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class

  // this needs to be rethought -- as currently implemented,
  // doesn't take edge-direction into account...
  assert(false);
  // loop over velocity components and apply appropriate BC's
  for (int comp = 0; comp<SpaceDim; comp++) {
    DomainGhostBC& thisBC = *(m_componentBC[comp]);
    thisBC.applyInhomogeneousBCs(a_state, a_domain, a_dx);
  }

}

// ------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
                         const Real a_dx, int a_comp) const 
{
  ProblemDomain physdomain(a_domain);
  // basic assumption here that all velocity BC's are inhomogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class
  DomainGhostBC& thisBC = *(m_componentBC[a_comp]);
  thisBC.applyInhomogeneousBCs(a_state, physdomain, a_dx);
}



// ------------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FArrayBox& a_state, 
                         const ProblemDomain& a_domain,
                         const Real a_dx, int a_comp) const 
{
  
  // basic assumption here that all velocity BC's are inhomogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class
  DomainGhostBC& thisBC = *(m_componentBC[a_comp]);
  thisBC.applyInhomogeneousBCs(a_state, a_domain, a_dx);

}

// ----------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FluxBox& a_state, const Box& a_domain,
                         const Real a_dx, int a_comp) const
{
  ProblemDomain physdomain(a_domain);
  for (int dir=0; dir<SpaceDim; dir++) {
    applyBCs(a_state[dir],physdomain, a_dx, a_comp);
  }
}


// ----------------------------------------------------------
void
basicEdgeVelBC::applyBCs(FluxBox& a_state, 
                         const ProblemDomain& a_domain,
                         const Real a_dx, int a_comp) const
{

  for (int dir=0; dir<SpaceDim; dir++) {
    applyBCs(a_state[dir],a_domain, a_dx, a_comp);
  }
}


// access function
const DomainGhostBC& 
basicEdgeVelBC::DomGhostBC(const int dir) const
{
  assert (dir < SpaceDim);

  return *m_componentBC[dir];
}



// will this be applied to single-component FluxBoxes?
void 
basicEdgeVelBC::isSingleComponent(bool a_isSingleComp)
{
  m_isSingleComponent = a_isSingleComp;
}

bool
basicEdgeVelBC::isSingleComponent() const
{
  return m_isSingleComponent;
}


// -----------------------------------------------------------
// functions for InflowVelBC

InflowVelBC::InflowVelBC() : BoxGhostBC()
{
  m_bcVal = 1.0e8;
}


InflowVelBC::~InflowVelBC()
{
}

InflowVelBC::InflowVelBC(int dir, Side::LoHiSide side) : BoxGhostBC(dir,side)
{
}

InflowVelBC::InflowVelBC(int dir, Side::LoHiSide side, 
			 const Interval& a_comps) 
  : BoxGhostBC(dir,side,a_comps)
{
  // set default inflow value to be a bogus number
  m_bcVal = 1.0e8;
}

BoxGhostBC*
InflowVelBC::new_boxghostbc() const
{
  InflowVelBC* newop = new InflowVelBC();
  if (newop == NULL) {
    MayDay::Error("Out of memory in InflowBC::new_boxghostbc");
  }
  newop->isHomogeneous(m_isHomogeneous);
  newop->setBCVal(m_bcVal);
  return static_cast<BoxGhostBC*> (newop);
}

void
InflowVelBC::fillBCValues(FArrayBox& a_neumfac,
			  FArrayBox& a_dircfac,
			  FArrayBox& a_inhmval,
			  Real a_dx,
			  const Box& a_domain) const
{
  ProblemDomain physdomain(a_domain);
  fillBCValues(a_neumfac, a_dircfac, a_inhmval, a_dx, physdomain);
}

void
InflowVelBC::fillBCValues(FArrayBox& a_neumfac,
			  FArrayBox& a_dircfac,
			  FArrayBox& a_inhmval,
			  Real a_dx,
			  const ProblemDomain& a_domain) const
{

  assert(!a_neumfac.box().isEmpty());
  assert(!a_dircfac.box().isEmpty());
  assert(!a_inhmval.box().isEmpty());
  a_neumfac.setVal(0.0);
  a_dircfac.setVal(1.0);

  // for now, set inflow velocity to be one.
  if (m_isHomogeneous) {
    a_inhmval.setVal(0.0);
  } else {
    a_inhmval.setVal(m_bcVal);
  }
}  


void
InflowVelBC::isHomogeneous(bool a_isHomogeneous) 
{
  m_isHomogeneous = a_isHomogeneous;
}

bool
InflowVelBC::isHomogeneous() const
{
  return m_isHomogeneous;
}

void 
InflowVelBC::setBCVal(const Real a_bcVal) 
{
  m_bcVal = a_bcVal;
}

Real
InflowVelBC::BCVal() const
{
  return m_bcVal;
}

void 
InflowVelBC::Time(const Real a_time) 
{
m_time = a_time;
}


Real
InflowVelBC::Time() const
{
  return  m_time;
}


