/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// boxVelBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#include "Box.H"
#include "boxVelBC.H"
#include "PoissonBC.H"
#include "ExtrapBC.H"
#include "PoissonBC.H"
#include "LoHiSide.H"
#include "edgeGhostBC.H"

// -----------------------------------------------------------
boxVelBC::boxVelBC(bool a_isViscous) 
{
  m_isViscous = a_isViscous;
}


// ----------------------------------------------------------
void
boxVelBC::setAllBCTypes() 
{
  // first loop over components
  for (int comp=0; comp<SpaceDim; comp++) {
    DomainGhostBC thisCompBC;
    Interval compInterval(comp,comp);
   
    // now loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {

      // loop over hi-lo
      SideIterator sit;
      for (sit.begin(); sit.ok(); ++sit) {

	//int bctype = m_physBC.getBC(dir,sit());
	int bctype = -5;
	MayDay::Error("boxVelBC::setAllBCTypes is deprecated!");
#if 0	
	// solid-wall BC's
	if (bctype == PhysBCUtil::SolidWall) {
	  // normal BC's are always no-flow
	  if (dir == comp) {
	    DircheletBC thisBC(dir, sit(),compInterval);
	    thisCompBC.setBoxGhostBC(thisBC);
	  } else {
	    // tangential BC's are no-slip if viscous, extrap if inviscid
	    if (m_isViscous) {
	      DircheletBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    } else {
	      ExtrapBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    } 
	  } // end if tangential comp
	} else {
	  // undefined BC type
	  MayDay::Error("boxVelBC::setAllBCTypes -- invalid BC type");
	}
#endif
      } // end loop over hi/lo
    } // end loop over directions
    
    setBCType(thisCompBC, comp);

  } // end loop over components

}


// ------------------------------------------------------------
bool
boxVelBC::isViscous()
{
  return m_isViscous;
}

// ------------------------------------------------------------
void
boxVelBC::isViscous(bool a_isViscous)
{
  m_isViscous = a_isViscous;
  // changing whether we're viscous or not will normally change
  // type of BC -- do this automatically.

  setAllBCTypes();

}




// -----------------------------------------------------------
boxEdgeVelBC::boxEdgeVelBC(bool a_isViscous) 
{
  m_isViscous = a_isViscous;
}


// ----------------------------------------------------------
void
boxEdgeVelBC::setAllBCTypes() 
{
  // first loop over components
  for (int comp=0; comp<SpaceDim; comp++) {
    DomainGhostBC thisCompBC;
    
    Interval compInterval(0,0);
    //Interval compInterval(comp,comp);
   
    // now loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {

      // loop over hi-lo
      SideIterator sit;
      for (sit.begin(); sit.ok(); ++sit) {

	//int bctype = m_physBC.getBC(dir,sit());
	int bctype = -5;
	MayDay::Error("boxEdgeVelBC::setAllBCTypes is deprecated!");

#if 0	
	// solid-wall BC's
	if (bctype == PhysBCUtil::SolidWall) {	
	  // normal BC's are always no-flow -- in this case,
	  // edge-centered homogeneous Dirichlet
	  if (dir == comp) {
	    edgeHomoDirichletBC thisBC(dir, sit(),compInterval);
	    IndexType edgeVelType;
	    // first set all directions to be cell-centered
	    edgeVelType.clear();
	    // then set normal direction to edge-centered
	    edgeVelType.set(dir);
	    thisBC.ixType(edgeVelType);
	    thisCompBC.setBoxGhostBC(thisBC);
	  } else {
	    // tangential BC's are no-slip if inviscid, extrap if viscous
	    if (m_isViscous) {
	      DircheletBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    } else {
	      ExtrapBC thisBC(dir,sit(),compInterval);
	      thisCompBC.setBoxGhostBC(thisBC);
	    } 
	  } // end if tangential comp
	} else {
	  	  // undefined BC type
	  MayDay::Error("boxEdgeVelBC::setAllBCTypes -- invalid BC type");
	}
#endif
      } // end loop over hi/lo
    } // end loop over directions
    
    setBCType(thisCompBC, comp);

  } // end loop over components

}


// ------------------------------------------------------------
bool
boxEdgeVelBC::isViscous()
{
  return m_isViscous;
}


// ------------------------------------------------------------
void 
boxEdgeVelBC::applyBCs(LevelData<FluxBox>& a_edgeStateMF,
		       const Box& a_domain, const Real a_dx) const
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_edgeStateMF, physdomain, a_dx);
}



// ------------------------------------------------------------
void 
boxEdgeVelBC::applyBCs(LevelData<FluxBox>& a_edgeStateMF,
		       const ProblemDomain& a_domain, 
                       const Real a_dx) const
{
  // just loop through boxes and call single-grid version
  DataIterator dit = a_edgeStateMF.dataIterator();
  
  for (dit.reset(); dit.ok(); ++dit) {
    applyBCs(a_edgeStateMF[dit()], a_domain, a_dx);
  }
}
				     


// ------------------------------------------------------------
void 
boxEdgeVelBC::applyBCs(FluxBox& a_edgeState,
		       const Box& a_domain, const Real a_dx) const
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_edgeState, physdomain, a_dx);
}




// ------------------------------------------------------------
void 
boxEdgeVelBC::applyBCs(FluxBox& a_edgeState,
		       const ProblemDomain& a_domain, 
                       const Real a_dx) const
{
  // loop over directions and call single-fab version
  for (int dir=0; dir<SpaceDim; dir++) {
    applyBCs(a_edgeState[dir], a_domain, a_dx, dir);
  }

}


// ------------------------------------------------------------
void
boxEdgeVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
		     const Real a_dx, const int a_velComp) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, physdomain, a_dx, a_velComp);
}


// ------------------------------------------------------------
void
boxEdgeVelBC::applyBCs(FArrayBox& a_state, const ProblemDomain& a_domain,
		     const Real a_dx, const int a_velComp) const 
{
  // this operates in the same way as the basicVelBC::applyBC's,
  // except that the nature of the edgeVels is that an
  // edge-centered FAB is generally one velocity component

  DomainGhostBC& thisBC = *(m_componentBC[a_velComp]);
  thisBC.applyHomogeneousBCs(a_state, a_domain, a_dx);
}


// ------------------------------------------------------------
void
boxEdgeVelBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
		     const Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, a_domain, a_dx);
}



// ------------------------------------------------------------
void
boxEdgeVelBC::applyBCs(FArrayBox& a_state, 
                       const ProblemDomain& a_domain,
                       const Real a_dx) const 
{
  // not clear that we should ever be here!
  MayDay::Warning("boxEdgeVelBC::applyBCs(FArrayBox&, const Box&, const Real) -- don't think we should be here!");

  basicVelBC::applyBCs(a_state, a_domain, a_dx);
}








