/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AmrEuler.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _AMREuler_H_
#define _AMREuler_H_

#include "AMRLevel.H"
#include "LevelFluxRegister.H"
#include "CoarseAverage.H"
#include "PiecewiseLinearFillPatch.H"
#include "CCProjector.H"

///
class AMREuler: public AMRLevel
{
public:
  
  ///
  AMREuler();
  
  ///
  AMREuler(AMRLevel* a_coarser_level_ptr, int a_level);

  ///
  AMREuler(AMRLevel* a_coarser_level_ptr, 
	   const Box& a_prob_domain,
	   int a_level, int a_ref_ratio);

  ///
  virtual ~AMREuler();

  /// workaround for the no-virtual-constructors issue
  virtual AMRLevel*   makeAMRLevel(AMRLevel* a_coarser_level_ptr,
				   int _alevel) const;

  ///
  virtual AMRLevel* makeAMRLevel(AMRLevel* a_coarser_level_ptr,
				 const Box& a_problem_domain,
				 int a_level, int a_ref_ratio) const;

  ///
  virtual void define(AMRLevel* a_coarse_level_ptr, 
		      const Box& a_problem_domain,
		      int a_level, int a_ref_ratio);

  /// 
  virtual void define(AMRLevel* a_coarser_level_ptr,
		      int a_level);

  /// advance by one timestep (returns what?)
  virtual Real advance();

  /// things to do after a timestep
  virtual void postTimeStep();

  /// create tags 
  virtual void tagCells(IntVectSet& a_tags) const;

  /// create tags at initialization
  virtual void tagCellsInit(IntVectSet& a_tags) const;

  /// regrid
  virtual void regrid(const Vector<Box>& a_new_grids);



  /// perform any post-regridding ops -- lBase is the finest unchanged level
  virtual void postRegrid(int a_lBase);


  /// initialize grids
  virtual void initialGrid(const Vector<Box>& a_new_grids);

  /// initialize data
  virtual void initialData();

  /// things do do after initialization
  virtual void postInitialize();

#ifdef HDF5
  /// write checkpoint header
  virtual void writeCheckpointHeader(HDF5Handle& a_handle) const;

  /// write checkpoint data for this level
  virtual void writeCheckpointLevel(HDF5Handle& a_handle) const;

  /// read checkpoint header
  virtual void readCheckpointHeader(HDF5Handle& a_handle);

  /// read checkpoint data for this level
  virtual void readCheckpointLevel(HDF5Handle& a_handle);

  /// write plotfile header
  virtual void writePlotHeader(HDF5Handle& a_handle) const;

  /// write plotfile data for this level
  virtual void writePlotLevel(HDF5Handle& a_handle) const;
#endif

  /// compute dt
  virtual Real computeDt();

  /// compute dt with initial data
  virtual Real computeInitialDt();

  /// set CFL number
  virtual void CFL(Real a_cfl);

  /// set refinement threshold
  virtual void refinementThreshold(Real a_refine_threshold);

  /// is this this finest level?
  bool finestLevel() const;
  
  /// is this an empty level?
  bool isEmpty() const;
  
  // compute vorticity
  void computeVorticity(LevelData<FArrayBox>& a_vorticity) const;

  /// access functions

  ///
  Real Dx() const;
  
  /// encapsulates the finer-level-pointer casting
  AMREuler* finerEulerPtr() const;

  /// encapsulates the coarser-level pointer casting
  AMREuler* crseEulerPtr() const;

  /// 
  LevelData<FArrayBox>& newVel();

  /// 
  const LevelData<FArrayBox>& newVel() const;

  /// 
  LevelData<FArrayBox>& oldVel();

  /// 
  const LevelData<FArrayBox>& oldVel() const;

  /// returns velocity at time t
  void velocity(LevelData<FArrayBox>& a_vel, Real a_time) const;

  ///
  LevelData<FArrayBox>& newLambda();

  ///
  const LevelData<FArrayBox>& newLambda() const;

  ///
  LevelData<FArrayBox>& oldLambda();

  ///
  const LevelData<FArrayBox>& oldLambda() const;

  /// returns lambda at time t
  void lambda(LevelData<FArrayBox>& a_lambda, Real a_time) const;
  


protected:
  ///
  DisjointBoxLayout loadBalance(const Vector<Box>& a_grids);

  ///
  void levelSetup (const DisjointBoxLayout& level_domain);

  /// read parameters from parmParse database
  void readParameters();

  /// set whether this is the finest level or not
  void finestLevel(bool a_finest_level);

  /// move new-time state into old-time state -- also advances time by dt
  void swapOldAndNewStates();

  /// compute advection velocities
  void computeAdvectionVelocities(LevelData<FluxBox>& a_adv_vel);


  /// do predictor -- returns approx. to (U dot del)U
  void predictVelocities(LevelData<FArrayBox>& a_uDelU,
			 LevelData<FluxBox>& a_advVel);


  /// do scalar advection -- assumes all BC's already done
  void advectScalar(LevelData<FArrayBox>& a_new_scal,
		    LevelData<FArrayBox>& a_old_scal,
		    LevelData<FluxBox>& a_adv_vel,
		    LevelFluxRegister* a_crseFluxRegPtr,
		    LevelFluxRegister& a_fineFluxReg,
		    Real a_dt);

  /// fill grown velocity field using piecewise-constant interp
  void fillVelocity(LevelData<FArrayBox>& a_vel, Real a_time);

  /// resize temporary lambda for advection tracing computation
  /// and fill interior and boundary cells
  void fillLambda(LevelData<FArrayBox>& a_lambda, Real a_time) const;

  /// set up the BC class for advected scalars
  void setScalarBCs(DomainGhostBC& a_scalBC) const;

  /// number of components in plotfiles
  virtual int numPlotComps() const;

  /// stuff plotfile data
  virtual void getPlotData(LevelData<FArrayBox>& a_plot_data) const;

protected:
  /// velocity vector at old time
  LevelData<FArrayBox>* m_vel_old_ptr;

  /// velocity vector at new time
  LevelData<FArrayBox>* m_vel_new_ptr;

  /// freestream preservation qty at old time
  LevelData<FArrayBox>* m_lambda_old_ptr;
  
  /// freestream preservation qty at new time
  LevelData<FArrayBox>* m_lambda_new_ptr;

  /// number of components of m_state (vel comps + lambda)
  static const int s_num_vel_comps = SpaceDim;

  /// number of scalars 
  static const int s_num_scal_comps = 1;

  /// names of components
  static const char* s_vel_names[s_num_vel_comps];

  /// names of scalars
  static const char* s_scal_names[s_num_scal_comps];

  /// size of physical domain
  static Vector<Real> s_domLength;
  
  /// have we initialized static variables from parmParse?
  static bool s_ppInit;

  /// factor to shrink initial timestep 
  static Real s_init_shrink;

  /// maximum allowable timestep
  static Real s_max_dt;

  /// should we initialize pressures after grid generation & regridding?
  static bool s_initialize_pressures;

  /// should we reflux momentum?
  static bool s_reflux_momentum;

  /// debugging option -- initially set arrays to s_bogus_value
  static bool s_set_bogus_values;

  /// debugging option -- value to which arrays are initially set
  static Real s_bogus_value;

  /// tag on vorticity?
  static bool s_tag_vorticity;

  /// factor for tagging -- tag if vorticity greater than factor*max_vort
  static Real s_vort_factor;


  /// grid spacing
  Real m_dx;

  /// cfl number
  Real m_cfl;

  /// dt from most recent timestep
  Real m_dt_save;

  /// averaging from fine to coarse level for velocity
  CoarseAverage m_coarse_average;

  /// averaging from fine to coarse level for scalar
  CoarseAverage m_coarse_average_scal;

  /// cell-centered projection object
  CCProjector m_projection;
  
  /// momentum flux register
  LevelFluxRegister m_flux_register;

  /// conserved scalar flux register
  LevelFluxRegister m_scal_flux_reg;

  /// refinement threshold for regridding
  Real m_refine_threshold;

  /// true if this is the finest extant level
  bool m_finest_level;

  /// true if this level is an empty level (no grids)
  bool m_is_empty;


};

#endif


