/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AmrEuler.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
#include <cmath>
#include <cstdio>

#include "parstream.H"
#include "LoadBalance.H"
#include "LayoutIterator.H"
#include "FineInterp.H"
#include "ParmParse.H"
#include "LoHiSide.H"
#include "BoxIterator.H"

#ifdef HDF5
#include "CH_HDF5.H"
#endif

#include "AMREuler.H"
#include "GodunovTrace.H"
#include "boxVelBC.H"
#include "ExtrapBC.H"
#include "Divergence.H"
#include "CellToEdge.H"
#include "EdgeToCell.H"
#include "computeNorm.H"
#include "IAMRGodunovF_F.H"
#include "PROB_F_F.H"
#include "AMREulerF_F.H"
#include "timeInterp.H"


#include "Timer.H"

extern Timer NST_Everything;
Timer TF_UDELS ("UDELS",          NST_Everything, DIAG_TABLE4);

// #define TEST_TAGGING  

#define ADVECT_GROW 2

#ifdef CH_USE_DOUBLE
#define TIME_EPS 1.0e-10
#else
#define TIME_EPS 1.0e-5
#endif

// initialize static variables here
const int AMREuler::s_num_vel_comps;
const int AMREuler::s_num_scal_comps;
Vector<Real> AMREuler::s_domLength(SpaceDim);
bool AMREuler::s_ppInit = false;
Real AMREuler::s_init_shrink = 1.0;
Real AMREuler::s_max_dt = 1.0e8;
bool AMREuler::s_initialize_pressures = false;
bool AMREuler::s_reflux_momentum = true;
bool AMREuler::s_set_bogus_values = true;
bool AMREuler::s_tag_vorticity = false;
Real AMREuler::s_vort_factor = 0.30;
Real AMREuler::s_bogus_value = 1.0e8;

// names of state variables
const char* AMREuler::s_vel_names[s_num_vel_comps] = {"x-velocity",
#if CH_SPACEDIM >= 2
						      "y-velocity",
#if CH_SPACEDIM >= 3
						      "z-velocity"
#endif
#endif
};

const char* AMREuler::s_scal_names[s_num_scal_comps] = {"lambda"};

// default constructor
AMREuler::AMREuler() : m_cfl(0.5),
		       m_refine_threshold(0.5),
		       m_finest_level(false),
		       m_is_empty(true)
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler default constructor" << endl;
  }
  m_vel_new_ptr = NULL;
  m_vel_old_ptr = NULL;
  m_lambda_new_ptr = NULL;
  m_lambda_old_ptr = NULL;

}


// constructor for restart
AMREuler::AMREuler( AMRLevel* a_coarser_level_ptr, int a_level) 
  : m_cfl(0.5),
    m_refine_threshold(0.5),
    m_finest_level(false),
    m_is_empty(true)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler restart constructor" << endl;
  } 
  m_vel_new_ptr = NULL;
  m_vel_old_ptr = NULL;
  m_lambda_new_ptr = NULL;
  m_lambda_old_ptr = NULL;

  define (a_coarser_level_ptr, a_level);
}

// full constructor
AMREuler::AMREuler(AMRLevel* a_coarser_level_ptr, 
		   const Box& a_prob_domain,
		   int a_level, int a_ref_ratio)
  : m_cfl(0.5),
    m_refine_threshold(0.5),
    m_finest_level(false),
    m_is_empty(true)

{
  if (s_verbosity >= 3) {
    pout () << "AMREuler full constructor" << endl;
  }   

  m_vel_new_ptr = NULL;
  m_vel_old_ptr = NULL;
  m_lambda_new_ptr = NULL;
  m_lambda_old_ptr = NULL;
  
  define(a_coarser_level_ptr, a_prob_domain, a_level, a_ref_ratio);
}

// destructor
AMREuler::~AMREuler()
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler destructor" << endl;
  }   
  
  if (m_vel_new_ptr != NULL) {
    delete m_vel_new_ptr;
    m_vel_new_ptr = NULL;
  }

  if (m_vel_old_ptr != NULL) {
    delete m_vel_old_ptr;
    m_vel_old_ptr = NULL;
  }

  if (m_lambda_new_ptr != NULL) {
    delete m_lambda_new_ptr;
    m_lambda_new_ptr = NULL;
  }

  if (m_lambda_old_ptr != NULL) {
    delete m_lambda_old_ptr;
    m_lambda_old_ptr = NULL;
  }


}

// define for restart
void 
AMREuler::define(AMRLevel* a_coarser_level_ptr, int a_level)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::define" << endl;
  }   
  AMRLevel::define(a_coarser_level_ptr, a_level);

  
  // read stuff from parmParse database
  if (!s_ppInit) {
    readParameters();
  }
  
  if (a_coarser_level_ptr != NULL) 
    {
      AMREuler* amre_ptr = dynamic_cast <AMREuler*> (a_coarser_level_ptr);
      if (amre_ptr != NULL) {
	m_cfl = amre_ptr->m_cfl;
	m_refine_threshold = amre_ptr->m_refine_threshold;
      } else {
	MayDay::Error ("AMREuler::define: a_coarser_level_ptr is not castable to AMREuler*"); 
      }
      // set coarser levels finer_level_ptr to point to this
      a_coarser_level_ptr->finerLevelPtr(static_cast<AMRLevel*> (this));
    }

  // projection is defined as part of restart

}

// define for starting from scratch
void
AMREuler::define(AMRLevel* a_coarser_level_ptr, 
		 const Box& a_problem_domain,
		 int a_level,
		 int a_ref_ratio)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::define " << a_level << endl;
  }

  AMRLevel::define(a_coarser_level_ptr, a_problem_domain,
		   a_level, a_ref_ratio);


  // read stuff from parmParse database
  if (!s_ppInit) {
    readParameters();
  }
  
  if (a_coarser_level_ptr != NULL) {
    AMREuler* amre_ptr = dynamic_cast <AMREuler*> (a_coarser_level_ptr);
    if (amre_ptr != NULL) {
      m_cfl = amre_ptr->m_cfl;
      m_refine_threshold = amre_ptr->m_refine_threshold;
    } else {
      MayDay::Error ("AMREuler::define: a_coarser_level_ptr is not castable to AMREuler*");
    }
    // set coarser levels finer_level_ptr to point to this
    a_coarser_level_ptr->finerLevelPtr(static_cast<AMRLevel*> (this));
  }

}

// "virtual constructor" for restart
AMRLevel* 
AMREuler::makeAMRLevel (AMRLevel* a_coarser_level_ptr, int a_level) const
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::makeAMRLevel" << endl;
  }
  AMREuler* amre_ptr = new AMREuler (a_coarser_level_ptr, a_level);
  return (static_cast <AMRLevel*> (amre_ptr));
}

// "virtual" full constructor
AMRLevel* 
AMREuler::makeAMRLevel (AMRLevel* a_coarser_level_ptr,
			const Box& a_prob_domain,
			int a_level, int a_ref_ratio) const
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::makeAMRLevel " << a_level << endl;
  }
  
  AMREuler* amre_ptr = new AMREuler(a_coarser_level_ptr, a_prob_domain,
				    a_level, a_ref_ratio);
  return (static_cast <AMRLevel*> (amre_ptr));
}

// advance solution by one timestep 
// return max safe timestep
Real
AMREuler::advance() 
{
  if (s_verbosity >= 2) {
    pout () << "AMREuler::advance " << m_level 
	    << ", starting time = " << m_time 
	    << ", dt = " << dt() << endl;
  } 
  
  Real old_time = m_time;
  Real new_time = m_time + m_dt;
  m_dt_save = m_dt;
  swapOldAndNewStates();

  const DisjointBoxLayout levelGrids = newVel().getBoxes();
  
  // initialize flux registers
  if (!m_finest_level) {
      m_flux_register.setToZero();
      m_scal_flux_reg.setToZero();
  }

  // compute advection velocities
  LevelData<FluxBox> advectionVel(levelGrids, 1);
  
  if (s_set_bogus_values) {
    DataIterator dit = advectionVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      advectionVel[dit()].setVal(s_bogus_value);
    }
  }

  computeAdvectionVelocities(advectionVel);

  // now that advection velocities have been computed,
  // update advected scalars

  // at the moment, just advecting lambda

  // do scalar boundary conditions on lambda 
  // (also grow to appropriate size)
  LevelData<FArrayBox> lambdaTemp;

  fillLambda(lambdaTemp, old_time);

  LevelFluxRegister* crseScalFluxRegPtr = NULL;
  if (m_level > 0) {
    crseScalFluxRegPtr = &(crseEulerPtr()->m_scal_flux_reg);
  }
  advectScalar(newLambda(), lambdaTemp, advectionVel,
	       crseScalFluxRegPtr, m_scal_flux_reg, 
	       m_dt);

  // now predict velocities -- this returns the advection term
  // put this in "new_vel"
  LevelData<FArrayBox>& uStar = newVel();
  predictVelocities(uStar, advectionVel);

  /// compue uStar = oldVel - dt*(u-dot-del-u)
  DataIterator dit = uStar.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) {
    Real mult = -1.0*m_dt;
    uStar[dit()] *= mult;

    uStar[dit()] += oldVel()[dit()];
  }

  // if a coarser level exists, will need coarse-level data for proj
  LevelData<FArrayBox>* crseVelPtr = NULL;
  
  if (m_level > 0) {
    const DisjointBoxLayout& crseGrids = crseEulerPtr()->newVel().getBoxes();
    crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
    // coarse velocity BC data is interpolated in time 
    crseEulerPtr()->fillVelocity(*crseVelPtr, new_time);
  }

  // need to do physical boundary conditions and exchanges
  boxVelBC velBC(false);
  velBC.setAllBCTypes();
  velBC.applyBCs(uStar, problemDomain(), m_dx);
  Interval velComps(0,SpaceDim-1);
  uStar.exchange(velComps);

  m_projection.LevelProject(uStar, crseVelPtr, new_time, m_dt);

  // as things stand now, physical BC's are re-set in LevelProjection

  // compute maximum safe timestep for next iteration
  Real newDt = computeDt();

  // clean up temp storage
  if (crseVelPtr != NULL) {
    delete crseVelPtr;
    crseVelPtr = NULL;
  }


  return newDt;

}


void 
AMREuler::computeAdvectionVelocities(LevelData<FluxBox>& a_advVel)
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::computeAdvectionVelocities: " << m_level << endl;
  }

  const DisjointBoxLayout& levelGrids = newVel().getBoxes();

  /// need to build grown grids to get be able to do all of
  /// tracing properly
  IntVect advect_grow(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));

  LevelData<FArrayBox> old_vel(levelGrids, SpaceDim, advect_grow);
  
  if (s_set_bogus_values) {
    DataIterator dit = old_vel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      old_vel[dit()].setVal(s_bogus_value);
    }
  }
  
  // m_time contains the time at which the new state is centered
  Real old_time = m_time - m_dt;
  fillVelocity(old_vel, old_time);
  
  // set physical boundary conditions here
  // set physical boundary conditions on velocity
  boxVelBC velBC(false);
  velBC.setAllBCTypes();

  velBC.applyBCs(old_vel,m_problem_domain, m_dx);

  FArrayBox srcFab;

  // trace to edges at time n+1/2
  DataIterator dit = a_advVel.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    const Box& grownBox = old_vel[dit()].box();
    srcFab.resize(grownBox, SpaceDim);
    srcFab.setVal(0.0);
    FluxBox& gridAdvVel = a_advVel[dit()];
    FArrayBox& gridCellVel = old_vel[dit()];

    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& edgeVelDir = gridAdvVel[dir];      
      
      Interval destComp(0,0);
      Interval srcComp(dir,dir);

      TraceState(edgeVelDir, 
		 gridCellVel,
		 gridCellVel,
		 srcFab,
		 problemDomain(),
		 levelGrids[dit()],
		 m_dt, m_dx, dir, 
		 srcComp, destComp);
    }
  }

  // need to set physical boundary conditions
  boxEdgeVelBC edgeVelBC(false);
  edgeVelBC.setAllBCTypes();
  
  edgeVelBC.applyBCs(a_advVel, m_problem_domain, m_dx);

  // now MAC project
  m_projection.levelMacProject(a_advVel, old_time, m_dt);

  // finally, add volume discrepancy correction
  if (m_projection.etaLambda() > 0) {
    LevelData<FluxBox>& grad_e_Lambda = m_projection.grad_eLambda();

    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisGrad_eLambda = grad_e_Lambda[dit()];
      FluxBox& thisAdvVel = a_advVel[dit()];
      for (int dir=0; dir<SpaceDim; dir++) {
	thisAdvVel[dir] += thisGrad_eLambda[dir];
      }
    }

  }

  // reset physical BC's
  edgeVelBC.applyBCs(a_advVel, m_problem_domain, m_dx);
}


void 
AMREuler::predictVelocities(LevelData<FArrayBox>& a_uDelU,
			    LevelData<FluxBox>& a_advVel)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::predictVelocities: " << m_level << endl;
  }

  const DisjointBoxLayout& levelGrids = a_advVel.getBoxes();
  
  // will need edge-centered storage for all velocity components
  LevelData<FluxBox> uHalf(levelGrids, SpaceDim);

  // will also need grad(eLambda) and grad(phi_mac)
  LevelData<FluxBox>& grad_eLambda = m_projection.grad_eLambda();
  LevelData<FluxBox> gradPhi(levelGrids, SpaceDim);
  m_projection.gradPhi(gradPhi);


  // for tracing, will need to fill in boundary values for 
  // grown copy of velocity
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  LevelData<FArrayBox> traceVel(levelGrids, SpaceDim, ghostVect);

  if (s_set_bogus_values) {
    DataIterator dit = traceVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      uHalf[dit()].setVal(s_bogus_value);
      traceVel[dit()].setVal(s_bogus_value);
    }
  }

  // now fill in velocity field
  // m_time is new time
  Real old_time = m_time - m_dt;
  fillVelocity(traceVel, old_time);

  // do physical boundary conditions on traceVel
  boxVelBC velBC(false);
  velBC.setAllBCTypes();
  velBC.applyBCs(traceVel,m_problem_domain,m_dx);


  // temp storage for source term (which will be zero)
  FArrayBox srcFab;

  // loop over velocity components
  for (int velComp = 0; velComp < SpaceDim; velComp++) {
    // need to compute predicted edge-centered values for each
    // velocity component in each coordinate direction. begin
    // loop over coordinate directions 
    for (int dir=0; dir<SpaceDim; dir++) {
      // if this is a normal edge, we already have edge-centered
      // predicted values, which are divergence-free (the advection
      // velocities).  treat this case differently
      if (dir == velComp) {
	DataIterator dit = uHalf.dataIterator();
	for (dit.reset(); dit.ok(); ++dit) {

	  // copy advVel -> uHalf
	  // srcComp is always 0, since advVel only has one component
	  int  srcComp = 0;
	  int  destComp = velComp;
	  //a_advVel.copyTo(srcComp, uHalf, destComp);
	  FluxBox& thisVel = uHalf[dit()];
	  FArrayBox& thisVelDir = thisVel[dir];

	  FluxBox& thisAdvVel = a_advVel[dit()];
	  FArrayBox& thisAdvVelDir = thisAdvVel[dir];

	  thisVelDir.copy(thisAdvVelDir,srcComp, destComp, 1);
	  
	  // now need to subtract off grad(eLambda)
	  // this is the due to the difference between _advecting_
	  // and _advected_ velocities
	
	  FluxBox& thisCorr = grad_eLambda[dit()];

	  FArrayBox& thisCorrDir = thisCorr[dir];

	  thisVelDir.minus(thisCorrDir, 0, dir, 1);
	}
      } else {
	// if not a normal direction, need to perform tracing
	Interval srcComps(velComp, velComp);
	Interval destComps(velComp, velComp);

	DataIterator dit = uHalf.dataIterator();
	for (dit.reset(); dit.ok(); ++dit) {
	  //const Box& gridBox = levelGrids[dit()];
	  const Box& grownBox = traceVel[dit()].box();
	  srcFab.resize(grownBox, SpaceDim);
	  srcFab.setVal(0.0);

	  FluxBox& thisUhalf = uHalf[dit()];
	  FArrayBox& thisUHalfDir = thisUhalf[dir];

	  FluxBox& thisAdvVel = a_advVel[dit()];
	  FArrayBox& thisOldVel = traceVel[dit()];

	  TraceState(thisUHalfDir, thisOldVel, thisOldVel, 
		     thisAdvVel, srcFab, problemDomain(),
		     levelGrids[dit()], m_dt, m_dx, dir,
		     srcComps, destComps);


	// need to add MAC correction to traced velocities
	thisUHalfDir.minus(gradPhi[dit()][dir], velComp, 
			   velComp, 1);

	}


      }
    } // end loop over tracing directions
  } // end loop over velocity components
  
  // now loop over grids to compute advection and update flux registers
  DataIterator dit = a_uDelU.dataIterator();
  
  // for nonconservative update, also need a cell-centered
  // advection velocity -- get this by averaging edge-centered
  // advection velocity to cell centers.  this one shouldn't need
  // any ghost cells.

  LevelData<FArrayBox> cellAdvVel(levelGrids, SpaceDim);
  EdgeToCell(a_advVel, cellAdvVel);

  for (dit.reset(); dit.ok(); ++dit) {
    FluxBox& thisUHalf = uHalf[dit()];
    FArrayBox& thisCellAdvVel = cellAdvVel[dit()];
    const Box& thisBox = levelGrids[dit()];
    FArrayBox& this_uDelU = a_uDelU[dit()];
    this_uDelU.setVal(0.0);
    
    // to do this in a dimensionality independent way, 
    // loop over directions.
    //timercomment//TF_UDELS.start();
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& uHalfDir = thisUHalf[dir];
     
      // this will increment this_uDelU with
      // cellVel*d(uHalfDir)/d[xyz] for all velocity components
      FORT_UDELS(CHF_FRA(this_uDelU),
		 CHF_FRA(uHalfDir),
		 CHF_FRA(thisCellAdvVel),
		 CHF_BOX(thisBox),
		 CHF_REAL(m_dx),
		 CHF_INT(dir));
    }
    //timercomment//TF_UDELS.stop();

    // now need to do flux register stuff
    // first compute fluxes (u_edge*u_adv)
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& this_uEdge = uHalf[dit()][dir];
      FArrayBox& thisAdvVel = a_advVel[dit()][dir];

      // now need to loop over traced velocity components
      for (int velComp=0; velComp<SpaceDim; ++velComp) {
	this_uEdge.mult(thisAdvVel,0,velComp,1);
      }
    }
    

    Interval velComps(0,SpaceDim-1);
    if (!m_finest_level) {
      assert (m_flux_register.isDefined());
      for (int dir=0; dir < SpaceDim; dir++) {
	// because we eventually want -D_R(U-del-U), signs
	// on fluxMult are opposite of what you might expect...
	Real fluxMult = -m_dt;
	m_flux_register.incrementCoarse(uHalf[dit()][dir],
					fluxMult, dit(), velComps,
					velComps, dir);
      }
    }
    
    if (m_level > 0) {
      LevelFluxRegister& crseFR = crseEulerPtr()->m_flux_register;
      assert(crseFR.isDefined());
      // because we will eventually want -D_R(U-del-U) signs
      // on fluxMult are opposite of what you might expect...
      Real fluxMult = -m_dt;
      for (int dir=0; dir<SpaceDim; dir++) {
	crseFR.incrementFine(uHalf[dit()][dir],
			     fluxMult, dit(), velComps,
			     velComps, dir, Side::Lo);

	crseFR.incrementFine(uHalf[dit()][dir],
			     fluxMult, dit(), velComps,
			     velComps, dir, Side::Hi);
      
      }
    }


  } // end loop over grids
      


}


void 
AMREuler::advectScalar(LevelData<FArrayBox>& a_new_scal,
		       LevelData<FArrayBox>& a_old_scal,
		       LevelData<FluxBox>& a_advVel,
		       LevelFluxRegister* a_crseFluxRegPtr,
		       LevelFluxRegister& a_fineFluxReg,
		       Real a_dt)
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::advectScalar: " << m_level << endl;
  }

  const DisjointBoxLayout& levelGrids = a_new_scal.getBoxes();

  DataIterator dit = a_new_scal.dataIterator();
  dit.reset();
  int nScal = a_new_scal[dit()].nComp();
  Interval scalComps(0,nScal-1);  
  
  LevelData<FluxBox> edgeScal(levelGrids, nScal);

  FArrayBox srcFab;
  // now trace scalars to edges at time n+1/2
  for (dit.reset(); dit.ok(); ++dit) {
    const Box& thisGrownBox = a_old_scal[dit()].box();
    FluxBox& thisEdgeScal = edgeScal[dit()];
    FluxBox& thisAdvVel = a_advVel[dit()];
    FArrayBox& thisCellVel = oldVel()[dit()];
    FArrayBox& thisOldScal = a_old_scal[dit()];

    srcFab.resize(thisGrownBox, nScal);
    srcFab.setVal(0.0);

    for (int dir=0; dir<SpaceDim; dir++) {
      TraceState(thisEdgeScal[dir], thisOldScal,
		 thisCellVel, thisAdvVel,
		 srcFab, problemDomain(),
		 levelGrids[dit()], m_dt, m_dx, dir, 
		 scalComps, scalComps);

      //multiply by edge velocity to get flux
      thisEdgeScal[dir] *= thisAdvVel[dir];
    }
  }

  
  // compute div(us)
  levelDivergenceMAC(a_new_scal, edgeScal, m_dx);

  // now multiply div(us) by -dt and add to old_scal
  for (dit.reset(); dit.ok(); ++dit) {
    Real mult = -1.0*a_dt;
    a_new_scal[dit()] *= mult;
    a_new_scal[dit()] += a_old_scal[dit()];
    
    
    // now do flux register stuff
    FluxBox& thisFlux = edgeScal[dit()];

    if (!m_finest_level) {
      assert(a_fineFluxReg.isDefined());
      for (int dir=0; dir<SpaceDim; dir++) {
	// since we will want the increment to be -U-del-U,
	// make this mult negative
	Real fluxMult = -a_dt;
	a_fineFluxReg.incrementCoarse(thisFlux[dir], fluxMult, dit(), 
				      scalComps, scalComps, dir);
      }
    } // end if a finer level exists
    

    if (m_level > 0) {
      assert(a_crseFluxRegPtr != NULL);
      assert(a_crseFluxRegPtr->isDefined());
      for (int dir=0; dir<SpaceDim; dir++) {
	Real fluxMult = -a_dt;
	a_crseFluxRegPtr->incrementFine(thisFlux[dir], fluxMult, dit(),
					scalComps, scalComps, dir,
					Side::Lo);
	
	a_crseFluxRegPtr->incrementFine(thisFlux[dir], fluxMult, dit(),
					scalComps, scalComps, dir,
					Side::Hi);
      }
    } // end if coarser level exists
  } // end loop over grids
				    
				  
  // that should be it!
}
		     


// things do do after a basic timestep
void
AMREuler::postTimeStep()
{

  if (s_verbosity >= 3) {
    pout () << "AMREuler::postTimeStep " << m_level << endl;
  } 

  // none of these things are done if this is the finest level
  
  if (!m_finest_level) {
    AMREuler* fineAmrEulerPtr = finerEulerPtr();
    // create boundary condition object
    boxVelBC velBC(false);
    velBC.setAllBCTypes();
    

    // first do refluxing and avgDown for conservation
    // do momentum refluxing first
    Real refluxScale = 1.0/m_dx;
    LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
    if (s_reflux_momentum) {
      m_flux_register.reflux(newVel, refluxScale);
      CoarseAverage& velAvgDown = fineAmrEulerPtr->m_coarse_average;
      velAvgDown.averageToCoarse(newVel,*fineAmrEulerPtr->m_vel_new_ptr);
    }
    
    // now do scalar
    LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
    m_scal_flux_reg.reflux(newLambda,refluxScale);
    CoarseAverage& scalAvgDown = fineAmrEulerPtr->m_coarse_average_scal;
    scalAvgDown.averageToCoarse(newLambda,*fineAmrEulerPtr->m_lambda_new_ptr);
  
    // now call sync projection if necessary
    Real crseTime = 0.0;
    if (m_level > 0) crseTime = m_coarser_level_ptr->time();
    
    // do sync if this is the coarsest level or if
    // coarser level is not at the same time as this level
    if (m_level == 0 || (abs(crseTime-m_time)>TIME_EPS)) {
      int finest_level = m_level;
      // index through levels to find out what finest level is
      // (there must be a better way to do this!)
      AMREuler* thisEulerPtr = fineAmrEulerPtr;
      while (!thisEulerPtr->finestLevel()) {
	thisEulerPtr = thisEulerPtr->finerEulerPtr();
      }
      assert(thisEulerPtr->finestLevel());
      finest_level = thisEulerPtr->m_level;

      // now allocate container for composite velocity and lambda
      Vector< LevelData<FArrayBox>* > compVel(finest_level+1);
      Vector< LevelData<FArrayBox>* > compLambda(finest_level+1);

      // just to be sure, set all pointers to NULL to start off
      for (int lev=0; lev<compVel.size(); lev++) {
	compVel[lev] = NULL;
	compLambda[lev]=NULL;
      }

      if (m_level > 0) { 
	// coarser level will be coarse-fine BC
	AMREuler* crseAmrEulerPtr = crseEulerPtr();
	DisjointBoxLayout crseGrids = crseAmrEulerPtr->newVel().getBoxes(); 

	LevelData<FArrayBox>* crseVelPtr =
	  new LevelData<FArrayBox>(crseGrids, SpaceDim);	

	crseAmrEulerPtr->velocity(*crseVelPtr, m_time);		
	compVel[m_level-1] = crseVelPtr;

	LevelData<FArrayBox>* crseLambdaPtr = 
	  new LevelData<FArrayBox>(crseGrids, 1);

	crseAmrEulerPtr->lambda(*crseLambdaPtr, m_time);
	compLambda[m_level-1] = crseLambdaPtr;

      }	

      thisEulerPtr = this;
      // now loop over levels and construct composite vel, lambda
      for (int lev=m_level; lev <= finest_level; lev++) {
	compVel[lev] = thisEulerPtr->m_vel_new_ptr;
	compLambda[lev] = thisEulerPtr->m_lambda_new_ptr;
	velBC.applyBCs(*compVel[lev], thisEulerPtr->problemDomain(),
		       thisEulerPtr->m_dx);
	thisEulerPtr = thisEulerPtr->finerEulerPtr();
      }

      // call projection-based sync operations
      m_projection.doSyncOperations(compVel, compLambda, m_time, m_dt);


      // re-apply physical boundary conditions
      thisEulerPtr = this;
      for (int lev = m_level; lev < finest_level; lev++) {	
	velBC.applyBCs(thisEulerPtr->newVel(), 
		       thisEulerPtr->problemDomain(), 
		       thisEulerPtr->m_dx);      
	thisEulerPtr = thisEulerPtr->finerEulerPtr();
      }
	
      // this is a diagnostic -- compute max(divergence)
      if (m_level == 0) {
	LevelData<FArrayBox>* crseVelPtr = NULL;
	LevelData<FArrayBox>* fineVelPtr = NULL;
	int nRefCrse = -1;
	int nRefFine = -1;

	AMREuler* thisLevelPtr = this;

	Real maxDiv = 0.0;

	for (int lev=0; lev <= finest_level; lev++) {
	  LevelData<FArrayBox>& levelVel = *compVel[lev];
	  const DisjointBoxLayout& levelGrids = levelVel.getBoxes();
	  LevelData<FArrayBox> levelDiv(levelGrids, 1);
	  const Box& levelDomain = thisLevelPtr->problemDomain();
	  
	  if (lev < finest_level) {
	    fineVelPtr = compVel[lev+1];
	    nRefFine = m_ref_ratio;
	  } else {
	    fineVelPtr = NULL;
	    nRefFine = -1;
	  }

	  compDivergenceCC(levelDiv, levelVel, crseVelPtr, fineVelPtr,
			   thisLevelPtr->m_dx, nRefCrse, nRefFine, 
			   levelDomain, true);

	  DataIterator dit = levelDiv.dataIterator();
	  Real levelMax = 0.0;
	  Interval divComps(0,0);

	  if (fineVelPtr != NULL) {
	    const DisjointBoxLayout* finerGridsPtr = &fineVelPtr->getBoxes();
	    levelMax = computeNorm(levelDiv, finerGridsPtr, 
				   nRefFine, thisLevelPtr->m_dx,
				   divComps, 0);
	  } else {
	    const DisjointBoxLayout* finerGridsPtr = NULL;
	    levelMax = computeNorm(levelDiv, finerGridsPtr, 
				   nRefFine, thisLevelPtr->m_dx,
				   divComps, 0);
	  }
	  if (levelMax > maxDiv) maxDiv = levelMax;
	  
	  // now index to next level	  
	  nRefCrse = nRefFine;
	  crseVelPtr = &levelVel;
	  thisLevelPtr = thisLevelPtr->finerEulerPtr();	  
	}
	
	pout() << "Max Div(u) = " << maxDiv << endl;
      }	  
      
	
      if (m_level > 0) {
	// need to clean up storage
	if (compLambda[m_level-1] != NULL) {
	  delete compLambda[m_level-1];
	  compLambda[m_level-1] = NULL;
	}

	if (compVel[m_level-1] != NULL) {
	  delete compVel[m_level-1];
	  compVel[m_level-1] = NULL;
	}
      }
    

    } // end sync projection operations


    // end if not finest level
  } else {
    // in case this is the only level
    if (m_level == 0) {

      // create boundary condition object and set physical BC's
      boxVelBC velBC(false);
      velBC.setAllBCTypes();
      velBC.applyBCs(newVel(), m_problem_domain, m_dx);

      LevelData<FArrayBox>* crseVelPtr = NULL;
      LevelData<FArrayBox>* fineVelPtr = NULL;
      int nRefCrse = -1;
      int nRefFine = -1;
      
      Real maxDiv = 0.0;
      const DisjointBoxLayout& grids = newVel().getBoxes();
      LevelData<FArrayBox> div(grids, 1);

      compDivergenceCC(div, newVel(), crseVelPtr, fineVelPtr,
		       m_dx, nRefCrse, nRefFine, m_problem_domain,
		       true);

      DataIterator dit=div.dataIterator();
      Interval divComps(0,0);

      const DisjointBoxLayout* finerGridsPtr = NULL;
      maxDiv = computeNorm(div, finerGridsPtr, nRefFine,
			   m_dx, divComps, 0);

      pout() << "Max Div(u) = " << maxDiv << endl;
    } // end if single-grid case...
  } // end if finest level      

}

// tag cells for regridding
void
AMREuler::tagCells(IntVectSet & a_tags) const
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::tagCells " << m_level << endl;
  }

  // use same function as at initial time
  tagCellsInit(a_tags);
}

// create tags at initial time
void 
AMREuler::tagCellsInit(IntVectSet& a_tags) const
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::tagCellsInit " << m_level << endl;
  }

  IntVectSet local_tags;
  
  // create tags based on something or other 
  // for now, don't do anything
  const DisjointBoxLayout& level_domain = newVel().getBoxes();


#ifdef TEST_TAGGING  
  LevelData<FArrayBox> taggingState(level_domain, 1);
  
  DataIterator dit = taggingState.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    FArrayBox& tagStateFab = taggingState[dit()];
    tagStateFab.setVal(0.0);
    
    FORT_SETTAGGINGVAR(CHF_FRA1(tagStateFab,0),
		       CHF_INT(m_dx));

    // now tag where taggingState is greater than or equal to 1
    BoxIterator bit(tagStateFab.box());
    for (bit.begin(); bit.ok(); ++bit) {
      const IntVect& iv = bit();
      if (tagStateFab(iv) >= 1.0) {
	local_tags |= iv;
      }
    }
  }

#endif

  
  if (s_tag_vorticity) {
    LevelData<FArrayBox> vorticity;
    if (SpaceDim == 2) {
      vorticity.define(level_domain,1);
    } else if (SpaceDim == 3) {
      vorticity.define(level_domain,SpaceDim);
    }
    computeVorticity(vorticity);

    Interval vortInterval(0,0);
    Real tagLevel = norm(vorticity, vortInterval, 0);
    tagLevel *= s_vort_factor;

    // now tag where vorticity is greater than or equal to tagLevel
    DataIterator dit = vorticity.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      FArrayBox& vortFab = vorticity[dit()];
      BoxIterator bit(vortFab.box());
      for (bit.begin(); bit.ok(); ++bit) {
	const IntVect& iv = bit();
	if (SpaceDim == 2) {
	  if (abs(vortFab(iv)) >= tagLevel) {
	    local_tags |= iv;
	  }
	} else if (SpaceDim == 3) {
	  for (int dir=0; dir<SpaceDim; dir++) {
	    if (abs(vortFab(iv,dir)) >= tagLevel) {
	      local_tags |= iv;
	    }
	  } // end loop over directions
	} // end if DIM=3
	    
      } // end loop over interior of box
    }

  }
    

  Vector<IntVectSet> all_tags;
  const int dest_proc = uniqueProc(SerialTask::compute);
  gather(all_tags, local_tags, dest_proc);

  if (procID() == uniqueProc(SerialTask::compute) ) {
    for (int i=0; i < all_tags.size(); ++i) {
      a_tags |= all_tags[i];
    }

    a_tags &= m_problem_domain;
  }
}


// create a load-balanced DisjointBoxLayout from a collection of boxes
DisjointBoxLayout
AMREuler::loadBalance(const Vector<Box>& a_grids)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::loadBalance " << m_level << endl;
  } 

  // this is copied pretty much directly from the AMRGodunov class
  Vector<int> proc_map;
  if (procID () == uniqueProc (SerialTask::compute) ) {
    LoadBalance(proc_map, a_grids);
  }

  broadcast (proc_map, uniqueProc (SerialTask::compute) );

  if (s_verbosity >= 4) {
    pout () << "AMREuler::loadBalance: processor map: " << endl;
    for (int igrid=0; igrid< a_grids.size(); ++igrid) {
      pout() << igrid << ": " << proc_map[igrid] << "  " << endl;
    }
    pout() << endl;
  }

  return (DisjointBoxLayout(a_grids, proc_map) );
}


 
// set up data on this level after regridding
void
AMREuler::regrid(const Vector<Box>& a_new_grids)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::regrid " << m_level << endl;
  }

  m_level_grids = a_new_grids;

  const DisjointBoxLayout level_domain = loadBalance(a_new_grids);

  if (s_verbosity >= 4) {
    pout () << "new grids: " << endl;
    for (LayoutIterator lit = level_domain.layoutIterator(); lit.ok(); ++lit)
      {
	pout () << level_domain[lit()] << endl;
      }
  }

  // save data for later copy -- only do this if level is not empty
  if (a_new_grids.size() > 0) {
    // also set flag to indicate that grids exist on this level
    m_is_empty = false;

    // since we're using pointers here, it's easy to save old
    // data -- then clean up afterwards
    LevelData<FArrayBox>* old_newVelPtr = m_vel_new_ptr;
    LevelData<FArrayBox>* old_oldVelPtr = m_vel_old_ptr;
    LevelData<FArrayBox>* old_newLambdaPtr = m_lambda_new_ptr;
    LevelData<FArrayBox>* old_oldLambdaPtr = m_lambda_old_ptr;    

    // reshape state with new grids
    IntVect ghostVect(D_DECL(1,1,1));
    m_vel_new_ptr = new LevelData<FArrayBox>(level_domain, 
					     s_num_vel_comps, ghostVect);
    m_vel_old_ptr = new LevelData<FArrayBox>(level_domain, 
					     s_num_vel_comps, ghostVect);
    
    m_lambda_new_ptr = new LevelData<FArrayBox>(level_domain, 
						s_num_scal_comps, ghostVect);
    m_lambda_old_ptr = new LevelData<FArrayBox>(level_domain, 
						s_num_scal_comps, ghostVect);
    
    if (s_set_bogus_values) {
      LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
      LevelData<FArrayBox>& oldVel = *m_vel_old_ptr;

      LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
      LevelData<FArrayBox>& oldLambda = *m_lambda_old_ptr;
      
      DataIterator dit = m_vel_new_ptr->dataIterator();
      for (dit.reset(); dit.ok(); ++dit) {
	oldVel[dit()].setVal(s_bogus_value);
	newVel[dit()].setVal(s_bogus_value);
	oldLambda[dit()].setVal(s_bogus_value);
	newLambda[dit()].setVal(s_bogus_value);
      }
    }


    // set up data structures
    levelSetup( level_domain);
    
    // interpolate from coarser level
    if (m_coarser_level_ptr != NULL) {
      AMREuler* amre_ptr = dynamic_cast<AMREuler*> (m_coarser_level_ptr);
      if (amre_ptr != NULL) {
	int nRefCrse = amre_ptr->m_ref_ratio;
	FineInterp fine_interp(level_domain,s_num_vel_comps,nRefCrse,
			       m_problem_domain);
	
	fine_interp.interpToFine (newVel(), amre_ptr->newVel());

	fine_interp.interpToFine (oldVel(), amre_ptr->oldVel());
	
	FineInterp fine_interp_lambda(level_domain, s_num_scal_comps, 
				      nRefCrse, m_problem_domain);
	fine_interp_lambda.interpToFine(newLambda(), amre_ptr->newLambda());
      } else {
	MayDay::Error("in AMREuler::regrid: m_coarser_level is not castable to AMREuler*");
      }
    }
    
    // copy from old state
    old_newVelPtr->copyTo(old_newVelPtr->interval(),
			  newVel(),
			  newVel().interval());

    old_oldVelPtr->copyTo(old_oldVelPtr->interval(),
			  oldVel(),
			  oldVel().interval());
    
    old_newLambdaPtr->copyTo(old_newLambdaPtr->interval(),
			     newLambda(), newLambda().interval() );

    // clean up before these pointers go out of scope...    
    if (old_newVelPtr != 0) {
      delete old_newVelPtr;
      old_newVelPtr = NULL;
    }

    if (old_oldVelPtr != NULL) {
      delete old_oldVelPtr;
      old_oldVelPtr = NULL;
    }

    if (old_newLambdaPtr != NULL) {
      delete old_newLambdaPtr;
      old_newLambdaPtr = NULL;
    }

    if (old_oldLambdaPtr != NULL) {
      delete old_oldLambdaPtr;
      old_oldLambdaPtr = NULL;
    }

  } else {
    // if new level is empty, then just clear everything

    // should just be able to delete pointers here    
    if (m_vel_new_ptr != NULL) {
      delete m_vel_new_ptr;
      m_vel_new_ptr = NULL;
    }
    
    if (m_vel_old_ptr != NULL) {
      delete m_vel_old_ptr;
      m_vel_old_ptr = NULL;
    }
    
    if (m_lambda_new_ptr != NULL) {
      delete m_lambda_new_ptr;
      m_lambda_new_ptr = NULL;
    }
    
    if (m_lambda_old_ptr != NULL) {
      delete m_lambda_old_ptr;
      m_lambda_old_ptr = NULL;
    }
    
    
    
    // this also implies that this is no longer the finest level
    finestLevel(false);
    // if a coarser level exists, let it know that it's the finest level
    if (m_coarser_level_ptr != NULL) {
      AMREuler* crse_amre_ptr = dynamic_cast<AMREuler*> (m_coarser_level_ptr);
      if (crse_amre_ptr != NULL) {
	if (!crse_amre_ptr->isEmpty()) crse_amre_ptr->finestLevel(true);
      }
    } else {
      MayDay::Error("in AMREuler::regrid: m_coarser_level is not castable to AMREuler*");
    }
  }

}


// initialize grid
void 
AMREuler::initialGrid(const Vector<Box>& a_new_grids)
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::initialGrid " << m_level << endl;
  }
  m_level_grids = a_new_grids;
  if (a_new_grids.size() > 0) m_is_empty = false;


  // note -- this assumes levels are defined from coarse->fine
  // (which is a reasonable assumption)
  finestLevel(true);
  if (m_coarser_level_ptr != NULL) {
    AMREuler* crse_amre_ptr = dynamic_cast<AMREuler*> (m_coarser_level_ptr);
    if (crse_amre_ptr != NULL) {
      crse_amre_ptr->finestLevel(false);
    } else {
      MayDay::Error("in AMREuler::initialGrid: m_coarser_level is not castable to AMREuler*");
    }
  }



  const DisjointBoxLayout level_domain = loadBalance(a_new_grids);

  if (s_verbosity >= 4) {
    pout () << "new grids: " << endl;
    for (LayoutIterator lit = level_domain.layoutIterator(); lit.ok(); ++lit)
      {
	pout() << level_domain[lit()] << endl;
      }
  }

  IntVect ghostVect(D_DECL(1,1,1));

  // this is necessary because of the way initialization
  // is done -- AMREuler::initialGrid is called each time
  // through level-by-level building of the initial grid
  // hierarchy
  
  if (m_vel_new_ptr != NULL) {
    delete m_vel_new_ptr;
    m_vel_new_ptr = NULL;
  } 
  m_vel_new_ptr = new LevelData<FArrayBox>(level_domain, 
					   s_num_vel_comps, 
					   ghostVect);

  if (m_vel_old_ptr != NULL) {
    delete m_vel_old_ptr;
    m_vel_old_ptr = NULL;
  }
  m_vel_old_ptr = new LevelData<FArrayBox>(level_domain, 
					   s_num_vel_comps, 
					   ghostVect);

  if (m_lambda_new_ptr != NULL) {
    delete m_lambda_new_ptr;
    m_lambda_new_ptr = NULL;
  }
  m_lambda_new_ptr = new LevelData<FArrayBox>(level_domain, 
					      s_num_scal_comps, 
					      ghostVect);

  if (m_lambda_old_ptr != NULL) {
    delete m_lambda_old_ptr;
    m_lambda_old_ptr = NULL;
  }
  m_lambda_old_ptr = new LevelData<FArrayBox>(level_domain, 
					      s_num_scal_comps, 
					      ghostVect);
  
  if (s_set_bogus_values) {
    LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
    LevelData<FArrayBox>& oldVel = *m_vel_old_ptr;
    
    LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
    LevelData<FArrayBox>& oldLambda = *m_lambda_old_ptr;
    
    DataIterator dit = m_vel_new_ptr->dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      oldVel[dit()].setVal(s_bogus_value);
      newVel[dit()].setVal(s_bogus_value);
      oldLambda[dit()].setVal(s_bogus_value);
      newLambda[dit()].setVal(s_bogus_value);
    }
  }
  


  m_dx = s_domLength[0]/problemDomain().size(0);
  

  // set up data structures -- actually, defer this until postInitialize
  //levelSetup(level_domain);
}

// set up data structures associated with this level
void
AMREuler::levelSetup(const DisjointBoxLayout& level_domain)
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::levelSetup " << m_level << endl;
  }


  CCProjector* crseProjPtr = NULL;
  CCProjector* fineProjPtr = NULL;
  int nRefCrse = -1;

  if (m_coarser_level_ptr != NULL) {
    m_coarse_average.define(level_domain, s_num_vel_comps, m_ref_ratio);
    m_coarse_average_scal.define(level_domain, s_num_scal_comps, m_ref_ratio);

    AMREuler* amre_ptr = dynamic_cast<AMREuler*> (m_coarser_level_ptr);
    if (amre_ptr != NULL) {
      const DisjointBoxLayout& coarser_level_domain 
	= amre_ptr->newVel().getBoxes();
      
      crseProjPtr = &(amre_ptr->m_projection);
      nRefCrse = amre_ptr->m_ref_ratio;

      // flux registers
      amre_ptr->m_flux_register.undefine();
      amre_ptr->m_flux_register.define(level_domain,
				       coarser_level_domain,
				       m_problem_domain,
				       nRefCrse,
				       s_num_vel_comps);

      amre_ptr->m_scal_flux_reg.undefine();
      amre_ptr->m_scal_flux_reg.define(level_domain,
				       coarser_level_domain,
				       m_problem_domain,
				       nRefCrse,
				       s_num_scal_comps);

      amre_ptr->m_flux_register.setToZero();
    } else {
      MayDay::Error("in AMREuler::levelSetup: m_coarser_level_ptr is not castable to AMREuler*");
    }

  }

  if (m_finer_level_ptr != NULL) {
    AMREuler* amre_ptr = dynamic_cast<AMREuler*> (m_finer_level_ptr);
    if (amre_ptr != NULL) {
      fineProjPtr = &(amre_ptr->m_projection);
    } else { 
      MayDay::Error("in AMREuler::levelSetup: m_finer_level_ptr is not castable to AMREuler*");
    }
  } else {
    m_flux_register.undefine();
  }

  m_projection.define(level_domain, m_problem_domain, m_dx, fineProjPtr,
		      crseProjPtr, nRefCrse, m_level);
  
}

			    
  
void  
AMREuler::readParameters() 
{

  if (s_verbosity >= 3) {
    pout () << "AMREuler::readParameters " << endl;
  } 

  ParmParse ppEuler("euler");

  Real defaultDomainLength = 1.0;
  for (int dir=0; dir<SpaceDim; dir++) {
    s_domLength[dir] = defaultDomainLength;
  }
  ppEuler.queryarr("domainLength", s_domLength, 0, SpaceDim);
  
  ppEuler.query("init_shrink", s_init_shrink);

  ppEuler.query("max_dt", s_max_dt);

  int temp = s_initialize_pressures;
  ppEuler.query("init_pressures", temp);
  s_initialize_pressures = (temp == 1);

  temp = s_reflux_momentum;
  ppEuler.query("reflux_momentum", temp);
  s_reflux_momentum = (temp == 1);

  temp = s_tag_vorticity;
  ppEuler.query("tag_vorticity", temp);
  s_tag_vorticity = (temp == 1);

  ppEuler.query("vorticity_tagging_factor", s_vort_factor);

  // set flag to indicate that we've done this
  s_ppInit = true;

}

void
AMREuler::initialData()
{

  if (s_verbosity >= 3) {
    pout () << "AMREuler::initialData " << m_level << endl;
  } 
  // for now, don't do anything -- this will (obviously) change!
  DataIterator dit = newVel().dataIterator();


  for (dit.reset(); dit.ok(); ++dit) {
    
    //newVel()[dit()].setVal(0.0,newVel()[dit()].box(),0,SpaceDim);
    
    FORT_INITVEL(CHF_FRA(newVel()[dit()]),
		 CHF_REAL(m_dx));

    // now initialize Lambda
    newLambda()[dit()].setVal(1.0, 0);
  }
}

void 
AMREuler::postInitialize()
{

  if (s_verbosity >= 3) {
    pout () << "AMREuler::postInitialize " << m_level << endl;
  } 

  // initialize data structures which haven't yet been initialized
  const DisjointBoxLayout& level_domain = newVel().getBoxes();
  levelSetup(level_domain);

  // set up for initial velocity projection -- do this from the finest level
  if (finestLevel()) {
    int numLevels = m_level +1;
    Vector<LevelData<FArrayBox>* > amrVel(numLevels);
    AMREuler* thisLevelData = this;
    for (int lev=m_level; lev >=0; lev--) {
      amrVel[lev] = &(thisLevelData->newVel());

      if (thisLevelData->m_coarser_level_ptr != NULL) {
	thisLevelData = 
	  dynamic_cast <AMREuler*> (thisLevelData->m_coarser_level_ptr);
      }
    }
    
    assert(thisLevelData->m_level == 0);
    AMREuler* level0AMREuler = thisLevelData;

    CCProjector& level0Proj = thisLevelData->m_projection;

    // set physical boundary conditions on velocity
    boxVelBC velBC(false);
    velBC.setAllBCTypes();

    for (int lev=0; lev<=m_level; lev++) {
      Box levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();
      
      velBC.applyBCs(*amrVel[lev], levelDomain, levelDx);

      if (thisLevelData->m_finer_level_ptr != NULL) {
	thisLevelData = dynamic_cast<AMREuler*> (thisLevelData->m_finer_level_ptr);
      }
    }


    bool homoBC = false;
    level0Proj.initialVelocityProject(amrVel, homoBC);
    thisLevelData = level0AMREuler;

    // need to reset boundary conditions here
    for (int lev=0; lev<=m_level; lev++) {
      Box levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();
      
      velBC.applyBCs(*amrVel[lev], levelDomain, levelDx);
      
      if (thisLevelData->m_finer_level_ptr != NULL) {
	thisLevelData = dynamic_cast<AMREuler*> (thisLevelData->m_finer_level_ptr);
      }
    }
    

    if (s_initialize_pressures) {
      MayDay::Error ("AMREuler::postInitialize -- pressure initialization not uet defined");

    } 

  }

}

Real
AMREuler::computeInitialDt() 
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::computeInitialDt " << m_level << endl;
  }

  return s_init_shrink*computeDt();
}

Real
AMREuler::computeDt()
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::computeDt " << m_level << endl;
  }

  Real MaxVel = 0;
  
  Interval velComps(0,SpaceDim-1);
  // compute maximum velocity  (MaxNorm)
  MaxVel = norm(newVel(), velComps, 0);

  
  Real dt; 
  if (MaxVel != 0) {
    dt = m_cfl*m_dx/MaxVel;
  } else {
    dt = s_max_dt;
  }

  return dt;
}




#ifdef HDF5
void 
AMREuler::writeCheckpointHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::writeCheckpointHeader" << endl;
  }

  HDF5HeaderData header;
  // since number of velocity components is obviously SpaceDim,
  // only need to do number of scalar components here
  assert (s_num_vel_comps == SpaceDim);
  header.m_int["num_components"] = s_num_scal_comps;
  char comp_str[30];
  for (int comp=0; comp<s_num_scal_comps; ++comp) {
    sprintf (comp_str, "component_%d", comp);
    header.m_string[comp_str] = s_scal_names[comp];
  }

  // now write velocity names..
  for (int comp=0; comp<s_num_vel_comps; ++comp) {
    sprintf (comp_str, "vel_component_%d", comp);
    header.m_string[comp_str] = s_vel_names[comp];
  }
  header.writeToFile(a_handle);

  if (s_verbosity >= 3) {
    pout () << header << endl;
  }

  // now write projection's header
  //m_projection.writeCheckpointHeader(a_handle);

}

void 
AMREuler::writeCheckpointLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3) {
    pout () << "AMRGodunov::writeCheckpointLevel" << endl;
  }

  char level_str[20];
  sprintf (level_str, "%d", m_level);
  const std::string label = std::string("level_") + level_str;

  a_handle.setGroup(label);

  HDF5HeaderData header;

  header.m_int ["ref_ratio"] = m_ref_ratio;
  header.m_real["dx"]        = m_dx;
  header.m_real["dt"]        = m_dt;
  header.m_real["time"]      = m_time;
  header.m_real["cfl"]       = m_cfl;

  header.m_int["finest_level"] = m_finest_level;
  header.m_int["is_empty"] = m_is_empty;

  // don't write out static variables -- get them from inputs files?
  // or should we checkpoint those?
  

  header.m_box["prob_domain"] = m_problem_domain;

  
  header.writeToFile(a_handle);

  if (s_verbosity >= 3) {
    pout () << header << endl;
  }

  // now dump out this level's data
  write (a_handle, m_vel_new_ptr->boxLayout());
  write (a_handle, *m_vel_new_ptr, "new_velocity");
  write (a_handle, *m_lambda_new_ptr, "new_lambda");

  // now do level checkpoint for projection
  m_projection.writeCheckpointLevel(a_handle);

}

void 
AMREuler::readCheckpointHeader(HDF5Handle& a_handle) 
{ 
  if (s_verbosity >= 3) {
    pout () << "AMRGodunov::readCheckpointHeader" << endl;
  }

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3) {
    pout() << "hdf5 header data:" << endl;
    pout () << header << endl;
  }

  // read number of components
  if (header.m_int.find("num_components") == header.m_int.end()) {
    MayDay::Error("AMREuler::readCheckpointHeader: checkfile does not have num_components");
  }
  int num_comps = header.m_int["num_components"];
  if (num_comps != s_num_scal_comps) {
    MayDay::Error("AMREuler::readCheckpointHeader: num_components in input file does not match solver");
  }

  // read component names
  std::string state_name;
  char comp_str[60];
  for (int comp=0; comp < s_num_scal_comps; ++comp) {
    sprintf (comp_str, "component_%d", comp);
    if (header.m_string.find(comp_str) == header.m_string.end()) {
      MayDay::Error("AMREuler::readCheckpointHeader: checkfile does not have enough component names");
    }
    state_name = header.m_string[comp_str];
    if (state_name != s_scal_names[comp]) {
      MayDay::Error("AMREuler::readCheckpointHeader: state_name in checkpoint does not match solver");
    }

    sprintf(comp_str, "vel_component_%d", comp);
    if (header.m_string.find(comp_str) == header.m_string.end()) {
      MayDay::Error("AMREuler::readCheckpointHeader: checkfile does not have enough velocity names");
    }
    state_name = header.m_string[comp_str];
    if (state_name != s_vel_names[comp]) {
      MayDay::Error("AMREuler::readCheckpointHEader: vel_name in checkfile does not match solver");
    }
  } 

}


// read level data from checkpoint file
void 
AMREuler::readCheckpointLevel (HDF5Handle& a_handle) 
{
  if (s_verbosity >= 3) {
    pout () << "AMREuler::readCheckpointLevel " << m_level << endl;
  }

  char level_str[20];
  sprintf(level_str, "%d", m_level);
  const std::string label = std::string("level_") + level_str;
  
  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3) {
    pout () << "hdf5 header data:" << endl;
    pout () << header << endl;
  }

  // read refinement ratio
  if (header.m_int.find("ref_ratio") == header.m_int.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain ref_ratio");
  }
  m_ref_ratio = header.m_int["ref_ratio"];

  // read dx
  if (header.m_real.find("dx") == header.m_real.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain dx");
  }
  m_dx = header.m_real["dx"];

  // read dt
  if (header.m_real.find("dt") == header.m_real.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain dt");
  }
  m_dt = header.m_real["dt"];

  // read time
  if (header.m_real.find("time") == header.m_real.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain time");
  }
  m_time = header.m_real["time"];

  // read cfl
  if (header.m_real.find("cfl") == header.m_real.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain cfl");
  }
  m_cfl = header.m_real["cfl"];

  // read finest_level
  if (header.m_int.find("finest_level") == header.m_int.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain finest_level");
  }
  m_finest_level = header.m_int["finest_level"];

  // read is_empty
  if (header.m_int.find("is_empty") == header.m_int.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain is_empty");
  }
  m_is_empty = header.m_int["is_empty"];

  // read problem domain
  if (header.m_box.find("prob_domain") == header.m_box.end()) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain prob_domain");
  }
  m_problem_domain = header.m_box["prob_domain"];

  // read grids
  Vector<Box> grids;
  const int grid_status = read(a_handle, grids);
  if (grid_status != 0) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain a Vector<Box>");
  }
  // create level domain
  const DisjointBoxLayout level_domain = loadBalance(grids);
  
  LayoutIterator lit = level_domain.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit) {
    const Box& b = level_domain[lit()];
    m_level_grids.push_back(b);
  }
  
  if (s_verbosity >= 4) {
    pout () << "read level domain: " << endl;
    LayoutIterator lit = level_domain.layoutIterator();
    for (lit.begin(); lit.ok(); ++lit) {
      const Box& b = level_domain[lit()];
      pout() << lit().intCode() << ": " << b << endl;
    } 
    pout () << endl;
  }
  
  // reshape data holders with new grids
  // reshape state with new grids
  IntVect ghostVect(D_DECL(1,1,1));
  m_vel_new_ptr = new LevelData<FArrayBox>(level_domain, 
					   s_num_vel_comps, ghostVect);
  m_vel_old_ptr = new LevelData<FArrayBox>(level_domain, 
					   s_num_vel_comps, ghostVect);
  
  m_lambda_new_ptr = new LevelData<FArrayBox>(level_domain, 
					      s_num_scal_comps, ghostVect);
  m_lambda_old_ptr = new LevelData<FArrayBox>(level_domain, 
					      s_num_scal_comps, ghostVect);
  
  if (s_set_bogus_values) {
    LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
    LevelData<FArrayBox>& oldVel = *m_vel_old_ptr;
    
    LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
    LevelData<FArrayBox>& oldLambda = *m_lambda_old_ptr;
    
    DataIterator dit = m_vel_new_ptr->dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      oldVel[dit()].setVal(s_bogus_value);
      newVel[dit()].setVal(s_bogus_value);
      oldLambda[dit()].setVal(s_bogus_value);
      newLambda[dit()].setVal(s_bogus_value);
    }
  }
  
  // read data 
  LevelData<FArrayBox>& new_vel = *m_vel_new_ptr;

  const int velData_status = read<FArrayBox> (a_handle,
					      new_vel,
					      "new_velocity",
					      level_domain);
  
  if (velData_status != 0) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain new_velocity data");
  }

  LevelData<FArrayBox>& new_lambda = *m_lambda_new_ptr;
  
  const int lambdaData_status = read<FArrayBox> (a_handle,
						new_lambda,
						"new_lambda",
						level_domain);

  if (lambdaData_status != 0) {
    MayDay::Error("AMREuler::readCheckpointLevel: file does not contain new_lambda data");
  }

  levelSetup(level_domain);

  m_projection.readCheckpointLevel(a_handle);


}

void 
AMREuler::writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::writePlotHeader " << m_level << endl;
  }

  HDF5HeaderData header;

  int numcomp = numPlotComps();
  
  header.m_int ["num_components"] = numcomp;
  char comp_str[30];
  int comp = 0;
  
  sprintf(comp_str, "component_%d", comp);
  header.m_string[comp_str] = "xVel";
  comp++;

  if (SpaceDim >1) {
    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "yVel";
    comp++;
   
    if (SpaceDim >2) {
      sprintf(comp_str, "component_%d", comp);
      header.m_string[comp_str] = "zVel";
      comp++;
    }
  }

  sprintf(comp_str, "component_%d", comp);
  header.m_string[comp_str] = "divergence";
  comp++;

  sprintf(comp_str, "component_%d", comp);
  header.m_string[comp_str] = "lambda";
  comp++;

  sprintf(comp_str, "component_%d", comp);
  header.m_string[comp_str] = "duDt";
  comp++;

  if (SpaceDim >1) {
    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "dvDt";
    comp++;
   
    if (SpaceDim >2) {
      sprintf(comp_str, "component_%d", comp);
      header.m_string[comp_str] = "dwDt";
      comp++;
    }
  }

  if (SpaceDim == 2) {
    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "vorticity";
    comp++;
  } else if (SpaceDim == 3) {
    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "x_vort";
    comp++;

    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "y_vort";
    comp++;

    sprintf(comp_str, "component_%d", comp);
    header.m_string[comp_str] = "z_vort";
    comp++;
  }

  header.writeToFile(a_handle);

  if (s_verbosity >= 3) {
    pout () << header << endl;
  }
}
    
   

void 
AMREuler::writePlotLevel(HDF5Handle& a_handle) const
{ 
  if (s_verbosity >= 3) {
    pout() << "AMREuler::writePlotLevel " << m_level << endl;
  }
  char level_str[20];
  sprintf (level_str, "%d", m_level);
  const std::string label = std::string("level_") + level_str;
  
  a_handle.setGroup(label);

  HDF5HeaderData header;

  header.m_int ["ref_ratio"]    = m_ref_ratio;
  header.m_real ["dx"]          = m_dx;
  header.m_real ["dt"]          = m_dt;
  header.m_real ["time"]        = m_time;
  header.m_box  ["prob_domain"] = m_problem_domain;
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout () << header << endl;
  }
  
  const DisjointBoxLayout& levelGrids = m_vel_new_ptr->getBoxes();
  int numcomp = numPlotComps();  

  LevelData<FArrayBox> plotData(levelGrids, numcomp);
  getPlotData(plotData);

  write (a_handle, levelGrids);
  write (a_handle, plotData, "data");
  
}
#endif

void 
AMREuler::CFL(Real a_cfl)
{ 
  m_cfl = a_cfl;
}

void
AMREuler::refinementThreshold(Real a_thresh) 
{ 
  m_refine_threshold = a_thresh;
}
    
bool
AMREuler::finestLevel() const
{
  return m_finest_level;
}

bool
AMREuler::isEmpty() const
{
  return m_is_empty;
}

void
AMREuler::finestLevel(bool a_finest_level) 
{
  m_finest_level = a_finest_level;
  m_projection.isFinestLevel(a_finest_level);
}

Real
AMREuler::Dx() const 
{
  return m_dx;
}

void 
AMREuler::postRegrid(int a_lBase)
{
  if (s_verbosity >= 3) {
    pout() << "AMREuler::postRegrid " << m_level << endl;
  }

  // this looks remarkably similar to postInitialize
  // need to re-project velocity and recompute e_lambda;
  // do this from the finest level

  if (finestLevel()) {
    int numLevels = m_level+1;
    Vector<LevelData<FArrayBox>* > amrVel(numLevels);
    Vector<LevelData<FArrayBox>* > amrLambda(numLevels);

    AMREuler* thisLevelData = this;
    for (int lev=m_level; lev >=a_lBase; --lev) {
      amrVel[lev] = &(thisLevelData->newVel());
      amrLambda[lev] = &(thisLevelData->newLambda());

      if (thisLevelData->m_coarser_level_ptr != NULL) {
	thisLevelData = dynamic_cast<AMREuler*> (thisLevelData->m_coarser_level_ptr);
      }
    }
  
    assert(thisLevelData->m_level == a_lBase);

    AMREuler* lBaseAMREuler = thisLevelData;

    CCProjector& lBaseProj = thisLevelData->m_projection;
    Real dt_lBase = thisLevelData->m_dt;

    // set physical boundary conditions on velocity
    boxVelBC velBC(false);
    velBC.setAllBCTypes();

    for (int lev=a_lBase; lev <= m_level; lev++) {
      Box levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();

      velBC.applyBCs(*amrVel[lev], levelDomain, levelDx);

      if (thisLevelData->m_finer_level_ptr != NULL) {
	thisLevelData = dynamic_cast<AMREuler*> (thisLevelData->m_finer_level_ptr);
      }
    } // end loop over levels
#if 0
    // shouldn't need this, since we shouldn't be changing lbase vel
    // may need coarser CF-BC
    LevelData<FArrayBox> crseVel;
    if (a_lBase > 0) {
      const DisjointBoxLayout& crseGrids = lBaseAMREuler->newVel().getBoxes();
      // this shouldn't need any ghost cells (proper nesting)
      crseVel.define(crseGrids, SpaceDim);
      AMREuler* crseLevelPtr = dynamic_cast<AMREuler*> (lBaseAMREuler->m_coarser_level_ptr); 
      crseLevelPtr->fillVelocity(crseVel, m_time);
      amrVel[a_lBase-1] = &crseVel;
    }
#endif

    // note that initial velocity projection is done holding
    // lBase velocity constant (so we call the finer-level projection)
    bool homogeneousCFBC = true;
    lBaseProj.fineProjPtr()->initialVelocityProject(amrVel,homogeneousCFBC); 

    lBaseProj.doPostRegridOps(amrVel, amrLambda, dt_lBase, m_time);
    thisLevelData = lBaseAMREuler;
    
    // need to reset boundary conditions here
    for (int lev=a_lBase; lev<=m_level; lev++) {
      Box levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();
      
      velBC.applyBCs(*amrVel[lev], levelDomain, levelDx);
      
      if (thisLevelData->m_finer_level_ptr != NULL) {
	thisLevelData = dynamic_cast<AMREuler*> (thisLevelData->m_finer_level_ptr);
      }
    } // end loop over levels to reset physBC's
    
  } // end if finest level

}


AMREuler* 
AMREuler::finerEulerPtr() const
{
  
  AMREuler* fine_e_ptr;
  if (m_finer_level_ptr != NULL) {
    fine_e_ptr = dynamic_cast <AMREuler*> (m_finer_level_ptr);
    if (fine_e_ptr == NULL) {
      MayDay::Error ("AMREuler::fineEulerPtr -- fineptr not castable to AMREuler*");
    }
  } else {
    fine_e_ptr = NULL;
  }
  

  return fine_e_ptr;
}

AMREuler* 
AMREuler::crseEulerPtr() const
{
  
  AMREuler* crse_e_ptr;
  if (m_coarser_level_ptr != NULL) {
    crse_e_ptr = dynamic_cast <AMREuler*> (m_coarser_level_ptr);
    if (crse_e_ptr == NULL) {
      MayDay::Error ("AMREuler::crseEulerPtr -- crseptr not castable to AMREuler*");
    }
  } else {
    crse_e_ptr = NULL;
  }
  

  return crse_e_ptr;
}


LevelData<FArrayBox>& 
AMREuler::newVel()
{
  assert(m_vel_new_ptr != NULL);

  return *(m_vel_new_ptr);
}

LevelData<FArrayBox>& 
AMREuler::oldVel()
{
  assert(m_vel_old_ptr != NULL);

  return *(m_vel_old_ptr);
}


LevelData<FArrayBox>& 
AMREuler::newLambda()
{
  assert(m_lambda_new_ptr != NULL);

  return *(m_lambda_new_ptr);
}


LevelData<FArrayBox>& 
AMREuler::oldLambda()
{
  assert(m_lambda_old_ptr != NULL);

  return *(m_lambda_old_ptr);
}


const LevelData<FArrayBox>& 
AMREuler::newVel() const
{
  assert(m_vel_new_ptr != NULL);

  return *(m_vel_new_ptr);
}

const LevelData<FArrayBox>& 
AMREuler::oldVel() const
{
  assert(m_vel_old_ptr != NULL);

  return *(m_vel_old_ptr);
}

const LevelData<FArrayBox>& 
AMREuler::newLambda() const 
{
  assert(m_lambda_new_ptr != NULL);

  return *(m_lambda_new_ptr);
}


const LevelData<FArrayBox>& 
AMREuler::oldLambda() const 
{
  assert(m_lambda_old_ptr != NULL);

  return *(m_lambda_old_ptr);
}


// this also changes time levels
void
AMREuler::swapOldAndNewStates() 
{
  LevelData<FArrayBox>* tempPtr;
  
  // swap both velocity and scalar data
  tempPtr = m_vel_new_ptr;
  m_vel_new_ptr = m_vel_old_ptr;
  m_vel_old_ptr = tempPtr;

  tempPtr = m_lambda_new_ptr;
  m_lambda_new_ptr = m_lambda_old_ptr;
  m_lambda_old_ptr = tempPtr;

  time(m_time + m_dt);
}

void 
AMREuler::fillLambda(LevelData<FArrayBox>& a_lambda, Real a_time) const
{
  assert (!a_lambda.isDefined());
  
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  const DisjointBoxLayout& levelGrids = m_vel_new_ptr->getBoxes();
  int ncomp = 1;

  a_lambda.define(levelGrids, ncomp, ghostVect);

  if (s_set_bogus_values) {
    DataIterator dit = a_lambda.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      a_lambda[dit()].setVal(s_bogus_value);
    }
  }

  Interval lambdaComps(0, ncomp-1);

  Real old_time = m_time - m_dt;
  // first copy lambda to a_lambda
  if (abs(a_time - old_time) < TIME_EPS) {
    // copy from old values
    m_lambda_old_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    // copy from new values
    m_lambda_new_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else {
    // interp in time
    timeInterp(a_lambda, a_time, oldLambda(), old_time,
	       newLambda(), m_time, lambdaComps);
  }

  if (m_level > 0) {
    // interpolate boundary values from coarse grids
    AMREuler* crseLevelPtr = crseEulerPtr();
    assert (crseLevelPtr != NULL);
    LevelData<FArrayBox>& crse_old_lambda = crseLevelPtr->oldLambda();
    LevelData<FArrayBox>& crse_new_lambda = crseLevelPtr->newLambda();
    const DisjointBoxLayout& crseGrids = crse_new_lambda.getBoxes();
    const Box& crseDomain = crseLevelPtr->problemDomain();
    int nRefCrse = crseLevelPtr->refRatio();

    Real crse_dt = crseLevelPtr->dt();
    // crse time should be the new time    
    Real crse_old_time = crseLevelPtr->time() - crse_dt;
    Real time_interp_coeff;

    // check for "essentially 0 or 1"
    if (abs(a_time - crse_old_time) < TIME_EPS) {
      time_interp_coeff = 0.0;
    } else if (abs(a_time - crseLevelPtr->time()) < TIME_EPS) {
      time_interp_coeff = 1.0;
    } else {
      time_interp_coeff = (a_time - crse_old_time)/crse_dt;
    }




    PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids,
					ncomp, crseDomain, nRefCrse,
					ADVECT_GROW);

    
    filpatcher.fillInterp(a_lambda, crse_old_lambda, crse_new_lambda, 
			  time_interp_coeff, 0, 0, ncomp);

  }

  // now do physical boundary conditions
  DomainGhostBC lambdaBC;
  setScalarBCs(lambdaBC);

  // loop over boxes
  const Box& physDomain = problemDomain();
  DataIterator dit = a_lambda.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    lambdaBC.applyInhomogeneousBCs(a_lambda[dit()], physDomain, m_dx);
  }


}
  
void
AMREuler::setScalarBCs(DomainGhostBC& a_scalBC) const
{
  // for now just implement solid-wall (extrap) BC's
  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      ExtrapBC thisBC(dir,sit());
      a_scalBC.setBoxGhostBC(thisBC);
    }
  }
}



void
AMREuler::fillVelocity(LevelData<FArrayBox>& a_vel, Real a_time)
{
  Interval velComps(0,SpaceDim-1);


  Real old_time = m_time - m_dt;
  // first copy on this level
  if (abs(a_time - old_time) < TIME_EPS) {
    m_vel_old_ptr->copyTo(velComps, a_vel, velComps);
  } else if (abs(a_time - m_time) < TIME_EPS) {
    m_vel_new_ptr->copyTo(velComps, a_vel, velComps);
  } else {
    // do linear interpolation in time
    timeInterp(a_vel, a_time, oldVel(), old_time,
	       newVel(), m_time, velComps);
  }
  
  if (m_level > 0) {
    const DisjointBoxLayout& levelGrids = a_vel.getBoxes();
    
    // fill in coarse-fine BC data by conservative linear interp
    AMREuler& crseLevel = *crseEulerPtr();
    
    LevelData<FArrayBox>& oldCrseVel = crseLevel.oldVel();
    LevelData<FArrayBox>& newCrseVel = crseLevel.newVel();
    const DisjointBoxLayout& crseGrids = oldCrseVel.getBoxes();
    const Box& crseDomain = crseLevel.problemDomain();
    int nRefCrse = crseLevel.refRatio();
    
    Real crse_new_time = crseLevel.m_time;
    Real crse_dt = crseLevel.dt();
    Real crse_old_time = crse_new_time - crse_dt;
    Real crse_time_interp_coeff;
    
    // check for "essentially 0 or 1"
    if (abs(a_time - crse_old_time) < TIME_EPS) {
      crse_time_interp_coeff = 0.0;
    } else if (abs(a_time - crse_new_time) < TIME_EPS) {
      crse_time_interp_coeff = 1.0;
    } else {
      crse_time_interp_coeff = (a_time - crse_old_time)/crse_dt;
    }
    
    
    
    const IntVect& velGrowVect = a_vel.ghostVect();
    int velGrow = velGrowVect[0];
    
    PiecewiseLinearFillPatch filpatcher(levelGrids, crseGrids, 
					SpaceDim, crseDomain,
					nRefCrse, velGrow);
    
    
    filpatcher.fillInterp(a_vel, oldCrseVel, newCrseVel,
			  crse_time_interp_coeff, 0,0,SpaceDim);
  }
  
  
}


// ------------------------------------------------------------------
void
AMREuler::velocity(LevelData<FArrayBox>& a_vel, Real a_time) const
{
  
  Interval velComps(0,SpaceDim-1);
  if (abs(a_time-m_time) < TIME_EPS) {
    m_vel_new_ptr->copyTo(velComps, a_vel, velComps);
  } else if (abs(a_time-(m_time-m_dt)) < TIME_EPS) {
    m_vel_old_ptr->copyTo(velComps, a_vel, velComps);
  } else {
    Real old_time = m_time - m_dt;
    timeInterp(a_vel, a_time, *m_vel_old_ptr, old_time, 
	       *m_vel_new_ptr, m_time, velComps);
  }

}


// ------------------------------------------------------------------
void
AMREuler::lambda(LevelData<FArrayBox>& a_lambda, Real a_time) const
{
  
  Interval lambdaComps(0,0);
  if (abs(a_time-m_time) < TIME_EPS) {
    m_lambda_new_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else if (abs(a_time-(m_time-m_dt)) < TIME_EPS) {
    m_lambda_old_ptr->copyTo(lambdaComps, a_lambda, lambdaComps);
  } else {
    Real old_time = m_time - m_dt;
    timeInterp(a_lambda, a_time, *m_lambda_old_ptr, old_time, 
	       *m_lambda_new_ptr, m_time, lambdaComps);
  }

}

int 
AMREuler::numPlotComps() const
{
  // for now, hardwire number of components:
  // velocity
  int numcomp = SpaceDim;
  // divergence
  ++numcomp;
  // lambda
  ++numcomp;
  // du/dt
  numcomp += SpaceDim;
  
  // vorticity
  if (SpaceDim == 2) {
    ++numcomp;
  } else if (SpaceDim == 3) {
    numcomp += SpaceDim;
  }
 
  return numcomp;
}

void
AMREuler::getPlotData(LevelData<FArrayBox>& a_plot_data) const
{
  
  assert (a_plot_data.nComp() == numPlotComps());
  const DisjointBoxLayout& levelGrids = a_plot_data.getBoxes();

  int plot_data_counter = 0;

  LevelData<FArrayBox> temp(levelGrids, 1);
  Interval srcComp(0,0);

  // first do velocities
  Interval velComps(0, SpaceDim-1);
  Interval destVelComps(plot_data_counter, plot_data_counter +SpaceDim-1);
  
  m_vel_new_ptr->copyTo(velComps, a_plot_data, destVelComps);
  plot_data_counter += SpaceDim;
  
  // now compute divergence
  LevelData<FArrayBox>* fineVelPtr = NULL;
  LevelData<FArrayBox>* crseVelPtr = NULL;
  int nRefCrse = -1;

  if (m_level > 0) {
    crseVelPtr = crseEulerPtr()->m_vel_new_ptr;
    nRefCrse = m_coarser_level_ptr->refRatio();
  }
  
  if (!finestLevel()) {
    fineVelPtr = finerEulerPtr()->m_vel_new_ptr;
  }
  
  compDivergenceCC(temp, *m_vel_new_ptr, crseVelPtr, fineVelPtr, 
		   m_dx, nRefCrse, m_ref_ratio, m_problem_domain,
		   true);


  Interval divDestComp(plot_data_counter, plot_data_counter);
  temp.copyTo(srcComp, a_plot_data, divDestComp);
  plot_data_counter += 1;

  // now copy lambda
  Interval lambdaDestComp(plot_data_counter, plot_data_counter);
  m_lambda_new_ptr->copyTo(srcComp, a_plot_data, lambdaDestComp);
  plot_data_counter += 1;

  // now do du/dt
  LevelData<FArrayBox> velTemp(levelGrids, SpaceDim);
  m_vel_new_ptr->copyTo(velComps, velTemp, velComps);
  
  const LevelData<FArrayBox>& old_vel = oldVel();
  DataIterator dit = velTemp.dataIterator();
  
  
  for (dit.reset(); dit.ok(); ++dit) {    
    if (m_dt_save != 0.0) {
      velTemp[dit()].minus(old_vel[dit()], levelGrids[dit()],
			   0,0,SpaceDim);
      
      velTemp[dit()].divide(m_dt_save);
    } else {
      velTemp[dit()].setVal(0.0);
    }
  }



  Interval duDtDestComps(plot_data_counter, plot_data_counter +SpaceDim-1);
  velTemp.copyTo(velComps, a_plot_data, duDtDestComps);
  plot_data_counter += SpaceDim;

  // finally do vorticity  
  if (SpaceDim == 2) {
    computeVorticity(temp);
    Interval vortComp(plot_data_counter, plot_data_counter);
    temp.copyTo(srcComp, a_plot_data, vortComp);
    ++plot_data_counter;
  } else if (SpaceDim == 3) {
    computeVorticity(velTemp);
    Interval vortComps(plot_data_counter, plot_data_counter +SpaceDim-1);
    velTemp.copyTo(velComps, a_plot_data, vortComps);
    plot_data_counter += SpaceDim;
  }



  

}


// ---------------------------------------------------------
void
AMREuler::computeVorticity(LevelData<FArrayBox>& a_vorticity) const
{
  // this will obviously have to be rethought in 3d

  // this breaks the "const"-ness of the function,
  // but is necessary to ensure that boundary
  // conditions are properly set.
  LevelData<FArrayBox>& vel = *(m_vel_new_ptr);
  
  if (m_level > 0) {
    // do quadratic C/F BC's
    const DisjointBoxLayout& grids = vel.getBoxes();
    // for now, assume that BC's are with new velocity
    // (may need to be changed for fine-level regridding)
    LevelData<FArrayBox>& crseVel = crseEulerPtr()->newVel();
    const DisjointBoxLayout& crseGrids = crseVel.getBoxes();
    int nRefCrse = crseEulerPtr()->refRatio();

    QuadCFInterp interpolator(grids, &crseGrids, m_dx, nRefCrse, 
			      SpaceDim, m_problem_domain);

    interpolator.coarseFineInterp(vel, crseVel);

  }

  Interval velComps(0,SpaceDim-1);  
  vel.exchange(velComps);

  // at the moment, do extrap at physical boundaries
  // (same effect as the one-sided differencing we have
  // been doing, right?
  DomainGhostBC vortVelBC;
  for (int dir=0; dir<SpaceDim; ++dir) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      ExtrapBC thisBC(dir,sit());
      vortVelBC.setBoxGhostBC(thisBC);
    }
  }

  DataIterator dit = vel.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    // apply physical BC's
    vortVelBC.applyHomogeneousBCs(vel[dit()], 
				  m_problem_domain, m_dx);

    if (SpaceDim == 3) {
      for (int dir=0; dir<SpaceDim; dir++) {
	// and then compute vorticity
	FORT_COMPUTEVORT(CHF_FRA1(a_vorticity[dit()],dir),
			 CHF_CONST_FRA(vel[dit()]),
			 CHF_CONST_REAL(m_dx), 
			 CHF_CONST_INT(dir));
      }
    } else if (SpaceDim == 2) {
      int dir = 2;
      FORT_COMPUTEVORT(CHF_FRA1(a_vorticity[dit()],0),
		       CHF_CONST_FRA(vel[dit()]),
		       CHF_CONST_REAL(m_dx),
		       CHF_CONST_INT(dir));
    } // end dimensionality choice
  } // end loop over grids

}

