/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// Main.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
#include "MeshRefineObject.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "ArrayView.H"
#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"
#include "AMREuler.H"

// global variables for handling output
static const char* pgmname = "amreuler";
static const char* indent = "   ";
static const char* indent2 = "      " ;

// define main loop here because that's the way amrgodunov does it!
void mainLoop();

int main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &arvg);
#endif
  
  // read inputs
  char* in_file = argv[1];
  ParmParse pp(argc-2, argv+2,NULL, in_file);

  // run actual algorithm
  mainLoop();

#ifdef MPI
  MPI_Finalize();
#endif
}

void
mainLoop()
{
  // read inputs
  ParmParse ppMain("main");
  
  int verbosity = 0;
  ppMain.query("verbosity", verbosity);
  assert (verbosity >= 0);
  AMR::verbosity(verbosity);
  AMRLevel::verbosity(verbosity);

  int nstop = 0;
  ppMain.query("max_step", nstop);

  Real stopTime = 0.;
  ppMain.query("max_time", stopTime);

  Vector<int> num_cells(SpaceDim,0);
  ppMain.queryarr("num_cells", num_cells, 0, SpaceDim);

  assert ( D_TERM ( ( num_cells[0] > 0 ), && 
                    ( num_cells[1] > 0 ), && 
                    ( num_cells[2] > 0 ) ) );
  assert ( D_TERM ( ( num_cells[0] % 2 == 0 ), &&
                    ( num_cells[1] % 2 == 0 ), &&
                    ( num_cells[2] % 2 == 0 ) ) );

  int max_level = 0;
  ppMain.query("max_level", max_level);
  
  int num_read_levels = Max(max_level,1);
  Vector<int> ref_ratios (num_read_levels,1);
  ppMain.queryarr("ref_ratio", ref_ratios,0,num_read_levels);
  
  Vector<int> regrid_intervals(num_read_levels,1);
  ppMain.queryarr("regrid_interval", regrid_intervals,0,num_read_levels);
  
  Real refine_threshold = 0.3;
  ppMain.query("refine_threshold", refine_threshold);
  
  int block_factor = 2;
  ppMain.query("block_factor", block_factor);

  int max_grid_size = 32;
  ppMain.query("max_grid_size", max_grid_size);

  Real fill_ratio = 0.75;
  ppMain.query("fill_ratio", fill_ratio);

  int buffer_size = 2;
  ppMain.query("grid_buffer_size", buffer_size);

  int checkpoint_interval = -1;
  ppMain.query("checkpoint_interval", checkpoint_interval);

  int plot_interval = -1;
  ppMain.query("plot_interval", plot_interval);

  Real cfl = 0.5;
  ppMain.query("cfl", cfl);

  if (verbosity >= 2) {
    // display inputs
    pout() << "nstop = " << nstop << endl;
    pout() << "stopTime = " << stopTime << endl;

    pout() << "num_cells = " << D_TERM ( num_cells[0] << "  " <<,
                                       num_cells[1] << "  " <<,
                                       num_cells[2] << ) endl;
  
    pout() << "max level = " << max_level << endl;
    pout() << "refinement ratios = ";
    for (int i = 0; i < ref_ratios.size(); ++i) pout() << ref_ratios[i] << " ";
    pout() << endl;
    pout() << "regrid intervals = ";
    for (int i = 0; i < regrid_intervals.size(); ++i) pout() << regrid_intervals[i] << " ";
    pout() << endl;
    pout() << "refinement threshold = " << refine_threshold << endl;

    pout() << "blocking factor = " << block_factor << endl;
    pout() << "max grid size = " << max_grid_size << endl;
    pout() << "fill ratio = " << fill_ratio << endl;
    pout() << "grid buffer size = " << buffer_size << endl;

    pout() << "checkpoint interval = " << checkpoint_interval << endl;
    pout() << "plot interval = " << plot_interval << endl;
    pout() << "CFL = " << cfl << endl;
  }


  // create objects
  Box prob_domain(IntVect::Zero,
		  IntVect(D_DECL(num_cells[0]-1,
				 num_cells[1]-1,
				 num_cells[2]-1)) );

  MeshRefineObject mr;
  mr.blockFactor(block_factor);
  mr.maxSize(max_grid_size);
  mr.fillRatio(fill_ratio);
  mr.bufferSize(buffer_size);

  if (!ppMain.contains("restart_file") ) {
    // initialize from scratch
    AMRLevel* bogus_amrlevel = NULL;

    AMREuler amreuler (bogus_amrlevel,
		       prob_domain, 0, ref_ratios[0]);

    amreuler.CFL(cfl);
    amreuler.refinementThreshold(refine_threshold);
    
    AMR thisAmr;

    thisAmr.define(prob_domain, ref_ratios,
		   regrid_intervals, max_level, 
		   mr, &amreuler);

    AMR::verbosity(verbosity);
    AMRLevel::verbosity(verbosity);

    thisAmr.checkpointInterval(checkpoint_interval);
    thisAmr.plotInterval(plot_interval);
    thisAmr.maxGridSize(max_grid_size);

    // initialize hierarchy of levels
    thisAmr.initialize();
    
    // run the specified number of timesteps (or until stopTime)
    thisAmr.coarseTimeStep(0., stopTime, 0, nstop);
    
    thisAmr.conclude();
  } else {
    // initialize from restart file
    AMREuler amreuler;

    amreuler.CFL(cfl);
    amreuler.refinementThreshold(refine_threshold);

    AMR restartAMR;
    restartAMR.define(regrid_intervals, mr, &amreuler);
    
    AMR::verbosity(verbosity);
    AMRLevel::verbosity(verbosity);

    restartAMR.checkpointInterval(checkpoint_interval);
    restartAMR.plotInterval(plot_interval);
    restartAMR.maxGridSize(max_grid_size);

    std::string restart_file;
    ppMain.query("restart_file", restart_file);
    
#ifdef HDF5
    HDF5Handle handle(restart_file, HDF5Handle::OPEN_RDONLY);
    HDF5HeaderData header;
    header.readFromFile(handle);

    // read current iteration number
    if (header.m_int.find("iteration") == header.m_int.end())
    {
      MayDay::Error("main: checkpoint file does not contain iteration");
    }
    int step = header.m_int ["iteration"];

// read time
    if (header.m_real.find("time") == header.m_real.end())
    {
      MayDay::Error("main: checkpoint file does not contain time");
    }
    Real time = header.m_real ["time"];

    if ( verbosity >= 2 )
    {
      pout () << "restart file = " << restart_file << endl;
      pout () << "read current timestep = " << step << endl;
      pout () << "read time = " << time << endl;
    }

// read from checkpoint file
    restartAMR.restart(handle);
    handle.close();

    restartAMR.coarseTimeStep(time, stopTime, step, nstop);

    restartAMR.conclude();
#endif
  }    

}
