/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"
#include "Timer.H"

// advance solution by one timestep 
// return max safe timestep
Real
AMRNavierStokes::advance() 
{
  TimeNSadvance.start();
  Timer TimeAdvanceStep;
  TimeAdvanceStep.start();
  GlobalLevel = m_level;

  TimeNSadvanceIntro.start();
  if (s_verbosity >= 2) {

    pout () << "AMRNavierStokes::advance " << m_level 
	    << ", starting time = " 
            << setiosflags(ios::fixed) << setprecision(6) << setw(12) << m_time 
	    << ", dt = " 
            << setiosflags(ios::fixed) << setprecision(6) << setw(12) << dt() 
            << endl;
  } 
  
#ifdef DEBUG
  // if we're at the beginning of a composite timestep,
  // save current multilevel state
  if (m_level == 0) {
    AMRNavierStokes* levelPtr = this;
    // figure out finest level
    while (!levelPtr->finestLevel()) {
      levelPtr = levelPtr->finerNSPtr();
    }
    int finest_level = levelPtr->m_level;

    levelPtr = this;
    for (int lev=0; lev<= finest_level; lev++) {
      levelPtr->m_saved_time = m_time;
      
      levelPtr->newVel().copyTo(levelPtr->newVel().interval(),
				*levelPtr->m_vel_save_ptr,
				levelPtr->m_vel_save_ptr->interval());

      levelPtr->newLambda().copyTo(levelPtr->newLambda().interval(),
                                   *levelPtr->m_lambda_save_ptr,
                                   levelPtr->m_lambda_save_ptr->interval());
      
      for (int comp=0; comp<s_num_scal_comps; comp++) {
	levelPtr->newScal(comp).copyTo(levelPtr->newScal(comp).interval(),
					*levelPtr->m_scal_save[comp], 
					levelPtr->m_scal_save[comp]->interval());
      }
      
      levelPtr = levelPtr->finerNSPtr();
    }
  }
#endif
      

  Real old_time = m_time;
  Real new_time = m_time + m_dt;
  m_dt_save = m_dt;
  swapOldAndNewStates();

  const DisjointBoxLayout levelGrids = newVel().getBoxes();
  
  // initialize flux registers
  if (!finestLevel()) {
      m_flux_register.setToZero();
      for (int comp=0; comp<s_num_scal_comps; comp++) {
	m_scal_fluxreg_ptrs[comp]->setToZero();
      }
      m_lambda_flux_reg.setToZero();
  }

  // compute advection velocities -- if we're using the
  // patchGodunov approach, these will have ghost cells.
#ifdef NEW_ADVECTION
  IntVect advVelGhost(IntVect::Unit);
  LevelData<FluxBox> advectionVel(levelGrids, 1, advVelGhost);
#else
  // old approach, no ghost cells needed  
  LevelData<FluxBox> advectionVel(levelGrids, 1);
#endif
  
  if (s_set_bogus_values) {
    DataIterator dit = advectionVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      advectionVel[dit()].setVal(s_bogus_value);
    }
  }
  TimeNSadvanceIntro.stop();

  TimeNSadvanceCompAV.start(); 
  computeAdvectionVelocities(advectionVel);
  TimeNSadvanceCompAV.stop();

  // now that advection velocities have been computed,
  // update advected scalars

  // do scalar boundary conditions on lambda 
  // (also grow to appropriate size)
  TimeNSadvanceAdvectLambda.start();

  LevelData<FArrayBox> lambdaTemp;

  fillLambda(lambdaTemp, old_time);

  LevelFluxRegister* crseLambdaFluxRegPtr = NULL;
  if (m_level > 0) {
    crseLambdaFluxRegPtr = &(crseNSPtr()->m_lambda_flux_reg);
  }
  
#ifdef NEW_ADVECTION
  //#if 0
  OldPhysIBC* lambdaIBCPtr = m_physBCPtr->lambdaTraceIBC();
#endif

  advectScalar(newLambda(), lambdaTemp, advectionVel,
	       crseLambdaFluxRegPtr, m_lambda_flux_reg, 
#ifdef NEW_ADVECTION
               //#if 0
               lambdaIBCPtr,
#endif
	       m_dt);
  TimeNSadvanceAdvectLambda.stop();

  TimeNSadvanceAdvectDiffScalar.start();
  // now do advected-diffused scalars
  if (s_num_scal_comps > 0) {
    Vector<DomainGhostBC*> scalPhysBCs(s_num_scal_comps,NULL);
    
    // loop over scalar components
    for (int comp=0; comp<s_num_scal_comps; comp++) {
      scalPhysBCs[comp] = m_physBCPtr->scalarTraceBC(comp);
      
#ifdef NEW_ADVECTION
      //#if 0
      OldPhysIBC* scalIBC = m_physBCPtr->scalarTraceIBC(comp);
#endif

      LevelData<FArrayBox> scalarTemp;
      fillScalars(scalarTemp, old_time,comp);
      
      // now get coarse-level scalars for BC's if necessary
      LevelData<FArrayBox>* newCrseScalPtr = NULL;
      LevelData<FArrayBox>* oldCrseScalPtr = NULL;
      Real oldCrseTime = -s_bogus_value;
      Real newCrseTime = s_bogus_value;
      LevelFluxRegister* crseScalFluxRegPtr = NULL;
    
      if (m_level > 0) {
	newCrseScalPtr = &(crseNSPtr()->newScal(comp));
	oldCrseScalPtr = &(crseNSPtr()->oldScal(comp));
	newCrseTime = crseNSPtr()->time();
	oldCrseTime = newCrseTime - crseNSPtr()->dt();
	crseScalFluxRegPtr = (crseNSPtr()->m_scal_fluxreg_ptrs[comp]);
      }

      
      advectDiffuseScalar(newScal(comp), scalarTemp,
			  advectionVel,
			  s_scal_coeffs[comp], 
			  oldCrseScalPtr, newCrseScalPtr,
			  oldCrseTime, newCrseTime,
			  crseScalFluxRegPtr, 
			  *(m_scal_fluxreg_ptrs[comp]),
			  *(scalPhysBCs[comp]),
#ifdef NEW_ADVECTION
                          //#if 0
                          scalIBC,
#endif
			  m_dt);
      

      delete scalPhysBCs[comp];
    } // end loop over components

  } // end advected-diffused scalars
  TimeNSadvanceAdvectDiffScalar.stop();

  // now predict velocities -- this returns the advection term
  // put this in "new_vel"
  TimeNSadvancePredictVelocities.start();
  LevelData<FArrayBox>& uStar = newVel();
  predictVelocities(uStar, advectionVel);
  TimeNSadvancePredictVelocities.stop();

  TimeNSadvanceCompUStar.start();   
  computeUStar(uStar); 
  TimeNSadvanceCompUStar.stop(); 

  TimeNSadvanceLevelProject.start();
  // if a coarser level exists, will need coarse-level data for proj
  LevelData<FArrayBox>* crseVelPtr = NULL;
  
  if (m_level > 0) {
    const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
    crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
    // coarse velocity BC data is interpolated in time 
    crseNSPtr()->fillVelocity(*crseVelPtr, new_time);
  }

  // need to do physical boundary conditions and exchanges
  bool isViscous = (s_nu > 0);
  basicVelBC* velBCPtr = m_physBCPtr->uStarBC(isViscous);

  velBCPtr->applyBCs(uStar, problemDomain(), m_dx);
  //Interval velComps(0,SpaceDim-1);
  // exchange is done in level projection
  //uStar.exchange(velComps);
  delete velBCPtr;

  // noel -- all time in level project 
  m_projection.LevelProject(uStar, crseVelPtr, new_time, m_dt);
  TimeNSadvanceLevelProject.stop(); 

  // as things stand now, physical BC's are re-set in LevelProjection

  // compute maximum safe timestep for next iteration
  Real newDt = computeDt();

  // clean up temp storage
  if (crseVelPtr != NULL) {
    delete crseVelPtr;
    crseVelPtr = NULL;
  }


  ++m_level_steps;
  TimeAdvanceStep.stop();

  if (s_verbosity >= 2) {

    pout () << "AMRNavierStokes::advance " << m_level 
	    << ",      end time = " 
            << setiosflags(ios::fixed) << setprecision(6) << setw(12) << m_time 
	    << ", dt = " 
            << setiosflags(ios::fixed) << setprecision(6) << setw(12) << dt() 
#ifdef TIMER
            << "  wc="
            << setiosflags(ios::fixed) << setprecision(4) << setw(12) 
            << TimeAdvanceStep.wc_time() 
#endif
            << endl;

  } 

  TimeNSadvance.stop();
  return newDt;

}



void 
AMRNavierStokes::computeAdvectionVelocities(LevelData<FluxBox>& a_advVel)
{
  TimecomputeAdvectionVelocities.start();
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::computeAdvectionVelocities: " 
           << m_level << endl;
  }

  bool isViscous = (s_nu > 0.0);

  const DisjointBoxLayout& levelGrids = newVel().getBoxes();

  /// need to build grown grids to get be able to do all of
  /// tracing properly
  IntVect advect_grow(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));

  LevelData<FArrayBox> old_vel(levelGrids, SpaceDim, advect_grow);
  LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim, advect_grow);
  LevelData<FArrayBox>* crseVelPtr = NULL;
  
  if (s_set_bogus_values) {
    DataIterator dit = old_vel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      old_vel[dit()].setVal(s_bogus_value);
      viscousSource[dit()].setVal(s_bogus_value);
    }
  }

  // m_time contains the time at which the new state is centered
  Real old_time = m_time - m_dt;
  TimeCAVfillVelocity.start();
  fillVelocity(old_vel, old_time);
  TimeCAVfillVelocity.stop();
    
  // set physical boundary conditions here
  // set physical boundary conditions on velocity
  basicVelBC* velBCPtr = NULL;

  Time_computeLapVel.start();
  if (isViscous) {
    velBCPtr = m_physBCPtr->viscousVelBC();
    // rather than resetting BC's on old_vel, and then setting them 
    // back, just copy old_vel to a temporary holder, and set 
    // BCs on that one.
    LevelData<FArrayBox> viscousVel(levelGrids, SpaceDim, advect_grow);
    DataIterator dit = viscousVel.dataIterator();
    for (dit.begin(); dit.ok(); ++dit)
      {
        viscousVel[dit()].copy(old_vel[dit()]);
      }
    velBCPtr->applyBCs(viscousVel, m_problem_domain, m_dx);
    delete velBCPtr;
    
    // if crse level exists, fill coarse velocity BC
    if (m_level > 0) {
      const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
      crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
      crseNSPtr()->fillVelocity(*crseVelPtr, old_time);
    }

    computeLapVel(viscousSource, viscousVel, crseVelPtr);

    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].mult(s_nu);
    }

    
  } 
  else 
    {
      DataIterator dit = viscousSource.dataIterator();
      for (dit.reset(); dit.ok(); ++dit) {
        viscousSource[dit()].setVal(0.0);
      }
    }
  Time_computeLapVel.stop();

  TimeTraceState.start();
  // tracing will use inviscid BC's
  velBCPtr = m_physBCPtr->tracingVelBC();
  velBCPtr->applyBCs(old_vel,m_problem_domain, m_dx);
  velBCPtr->applyBCs(old_vel,m_problem_domain, m_dx);
  delete velBCPtr;
  

#ifdef NEW_ADVECTION
  // set up patchGodunov for this problem
  // first get boundary condition object 
  TTS_setupPatchGodunov.start();
  OldPhysIBC* velIBCPtr = m_physBCPtr->advectionVelIBC();
  m_patchGod.setPhysIBC(velIBCPtr);
  
  // do all component directions at the same time
  m_patchGod.setNumVar(SpaceDim);
  m_patchGod.setCurrentTime(old_time);

  // this should now be unnecessary
  // also need to fill grown advection velocity
  // by cell-to-edge averaging old-time velocity
  CellToEdge(old_vel, a_advVel);

  // loop over grids and predict face-centered velocities at the half
  // time using the patchGodunov infrastructure
  DataIterator dit = a_advVel.dataIterator();
  FArrayBox tempFlux[SpaceDim];
  TTS_setupPatchGodunov.stop();
  
  for (dit.begin(); dit.ok(); ++dit)
    {
      Box gridBox = levelGrids[dit()];
      FArrayBox& thisOldVel = old_vel[dit()];
      FluxBox& thisAdvVel = a_advVel[dit()];
      FArrayBox& srcFab = viscousSource[dit()];
      
      // set up OldPatchGodunov for this patch
      m_patchGod.setCurrentBox(gridBox);
      m_patchGod.setCellVelPtr(&thisOldVel);
      m_patchGod.setAdvVelPtr(&thisAdvVel);

      // we will need some temp storage here
      TTS_tempStorage.start();
      for (int dir=0; dir<SpaceDim; dir++)
        {
          Box fluxBox = gridBox;
          fluxBox.grow(1);
          fluxBox.grow(dir,-1);
          fluxBox &= m_problem_domain;
          fluxBox.surroundingNodes(dir);
          tempFlux[dir].resize(fluxBox, SpaceDim);
        }
      TTS_tempStorage.stop();
      
      // now we should be set to predict face-centered velocities
      TTS_computeFluxes.start();
      m_patchGod.computeFluxes(thisOldVel, tempFlux, 
                               srcFab, m_dt, gridBox);
      TTS_computeFluxes.stop();
      
      // copy this to advection velocity FluxBox
      TTS_copyflux.start();
      for (int dir=0; dir<SpaceDim; dir++)
        {          
          // only copy valid faces
          Box faceBox = gridBox;
          faceBox.surroundingNodes(dir);
          thisAdvVel[dir].copy(tempFlux[dir], faceBox, dir,
                               faceBox, 0, 1);
          
        } // end loop over directions
      TTS_copyflux.stop();
      
    } // end loop over grids
           
  
#else    
  // trace to edges at time n+1/2 using levelTrace-based approach
  DataIterator dit = a_advVel.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    //    const Box& grownBox = old_vel[dit()].box();
    FArrayBox& srcFab = viscousSource[dit()];

    FluxBox& gridAdvVel = a_advVel[dit()];
    FArrayBox& gridCellVel = old_vel[dit()];

    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& edgeVelDir = gridAdvVel[dir];      
      
      Interval destComp(0,0);
      Interval srcComp(dir,dir);
      
      TraceState(edgeVelDir, 
		 gridCellVel,
		 gridCellVel,
		 srcFab,
		 problemDomain(),
		 levelGrids[dit()],
		 m_dt, m_dx, dir, 
		 srcComp, destComp);
    }
  } // end loop over grids
#endif // end leveltrace-based approach
  TimeTraceState.stop();  

  // need to set physical boundary conditions
  TimelevelMacProject.start();
  basicEdgeVelBC* edgeVelBCPtr = m_physBCPtr->advectionVelBC();
  
  edgeVelBCPtr->applyBCs(a_advVel, m_problem_domain, m_dx);

  // noel  levelMacProject is big guy
  // now MAC project
  m_projection.levelMacProject(a_advVel, old_time, m_dt);
  TimelevelMacProject.stop();  

  // finally, add volume discrepancy correction
  if (m_projection.etaLambda() > 0 && s_applyFreestreamCorrection) {
    LevelData<FluxBox>& grad_e_Lambda = m_projection.grad_eLambda();
    
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisGrad_eLambda = grad_e_Lambda[dit()];
      FluxBox& thisAdvVel = a_advVel[dit()];
      for (int dir=0; dir<SpaceDim; dir++) {
	thisAdvVel[dir] += thisGrad_eLambda[dir];
      }
    }

  }

  // reset physical BC's
  edgeVelBCPtr->applyBCs(a_advVel, m_problem_domain, m_dx);

  delete edgeVelBCPtr;

  // clean up storage
  if (crseVelPtr != NULL) {
    delete crseVelPtr;
    crseVelPtr = NULL;
  }

  TimecomputeAdvectionVelocities.stop();
}


