/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"


#include "Timer.H"
extern Timer Everything;
Timer TMPI_computeVorticity_exchange("compVorti ex", Everything, DTIMECOMM, true);
Timer TMPI_computeLapVel_exchange   ("LapVel ex",    Everything, DTIMECOMM, true);
Timer TMPI_computeLapScal_exchange  ("LapScal ex",   Everything, DTIMECOMM, true);

Timer TF_COMPUTEVORT2 ("COMPUTEVORT2",  Everything, DTIMEFORT, true);
Timer TF_KINETICENERGY("KINETICENERGY", Everything, DTIMEFORT, true);

// ---------------------------------------------------------
void
AMRNavierStokes::computeVorticity(LevelData<FArrayBox>& a_vorticity) const
{
  // this will obviously have to be rethought in 3d

  // this breaks the "const"-ness of the function,
  // but is necessary to ensure that boundary
  // conditions are properly set.
  LevelData<FArrayBox>& vel = *(m_vel_new_ptr);
  
  if (m_level > 0) {
    // do quadratic C/F BC's
    const DisjointBoxLayout& grids = vel.getBoxes();
    // for now, assume that BC's are with new velocity
    // (may need to be changed for fine-level regridding)
    LevelData<FArrayBox>& crseVel = crseNSPtr()->newVel();
    const DisjointBoxLayout& crseGrids = crseVel.getBoxes();
    int nRefCrse = crseNSPtr()->refRatio();

    QuadCFInterp interpolator(grids, &crseGrids, m_dx, nRefCrse, 
			      SpaceDim, m_problem_domain);

    interpolator.coarseFineInterp(vel, crseVel);

  }

  Interval velComps(0,SpaceDim-1);  
  TMPI_computeVorticity_exchange.start();
  vel.exchange(velComps);
  TMPI_computeVorticity_exchange.stop();

  // at the moment, do extrap at physical boundaries
  // (same effect as the one-sided differencing we have
  // been doing, right?
  DomainGhostBC vortVelBC;
  for (int dir=0; dir<SpaceDim; ++dir) {
    SideIterator sit;
    for (sit.reset(); sit.ok(); ++sit) {
      Interval velComps(0,SpaceDim-1);
      ExtrapBC thisBC(dir,sit(),velComps);
      vortVelBC.setBoxGhostBC(thisBC);
    }
  }

  DataIterator dit = vel.dataIterator();

  const DisjointBoxLayout& grids = a_vorticity.getBoxes();
  for (dit.reset(); dit.ok(); ++dit) {
    // apply physical BC's
    vortVelBC.applyHomogeneousBCs(vel[dit()], 
				  m_problem_domain, m_dx);

    TF_COMPUTEVORT2.start();
    if (SpaceDim == 3) {
      for (int dir=0; dir<SpaceDim; dir++) {
	// and then compute vorticity
	FORT_COMPUTEVORT(CHF_FRA1(a_vorticity[dit()],dir),
			 CHF_CONST_FRA(vel[dit()]),
			 CHF_BOX(grids[dit()]),
			 CHF_CONST_REAL(m_dx), 
			 CHF_CONST_INT(dir));
      }
    } else if (SpaceDim == 2) {
      int dir = 2;
      FORT_COMPUTEVORT(CHF_FRA1(a_vorticity[dit()],0),
		       CHF_CONST_FRA(vel[dit()]),
		       CHF_BOX(grids[dit()]),
		       CHF_CONST_REAL(m_dx),
		       CHF_CONST_INT(dir));
    } // end dimensionality choice
    TF_COMPUTEVORT2.stop();
  } // end loop over grids

}

void 
AMRNavierStokes::computeLapVel(LevelData<FArrayBox>& a_lapVel,
			       LevelData<FArrayBox>& a_vel,
			       const LevelData<FArrayBox>* a_crseVelPtr)
{

  // set BC's
  basicVelBC* BCPtr = m_physBCPtr->viscousVelBC();
  Vector<DomainGhostBC> bcvect(SpaceDim);
  for (int dir=0; dir<SpaceDim; dir++) {
    bcvect[dir] = BCPtr->DomGhostBC(dir);
  }
  

  m_velocityPoissonOp.applyOpI(a_vel, a_crseVelPtr, a_lapVel,
			       bcvect);

  // may need to extend lapVel to cover ghost cells as well
  basicVelBC* viscBCPtr = m_physBCPtr->viscousSrcBC();

  viscBCPtr->applyBCs(a_lapVel, m_problem_domain, m_dx);

  // finally, do exchange
  TMPI_computeLapVel_exchange.start();
  a_lapVel.exchange(a_lapVel.interval());
  TMPI_computeLapVel_exchange.stop();

  delete viscBCPtr;
  delete BCPtr;

}

void 
AMRNavierStokes::computeLapScal(LevelData<FArrayBox>& a_lapScal,
				LevelData<FArrayBox>& a_scal,
				DomainGhostBC& a_physBC,
				const LevelData<FArrayBox>* a_crseScalPtr) 
{

  m_scalarsPoissonOp.setDomainGhostBC(a_physBC);
  m_scalarsPoissonOp.applyOpI(a_scal, a_crseScalPtr, a_lapScal);

  // may need to extend Laplacian to cover ghost cells as well
  viscousBC viscBC;
  viscBC.applyBCs(a_lapScal, m_problem_domain, m_dx);

  // finally, do exchange
  TMPI_computeLapScal_exchange.start();
  a_lapScal.exchange(a_lapScal.interval());
  TMPI_computeLapScal_exchange.stop();

}


// ---------------------------------------------------------
void
AMRNavierStokes::computeKineticEnergy(LevelData<FArrayBox>& a_energy) const
{
  
  // this is simple since no BC's need to be set.
  const LevelData<FArrayBox>& levelVel = *m_vel_new_ptr;
  const DisjointBoxLayout& levelGrids = levelVel.getBoxes();

  DataIterator dit = a_energy.dataIterator();
  TF_KINETICENERGY.start();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FORT_KINETICENERGY(CHF_FRA1(a_energy[dit()],0),
                         CHF_CONST_FRA(levelVel[dit()]),
                         CHF_BOX(levelGrids[dit()]));
    }
  TF_KINETICENERGY.stop();


}
