/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include "AMRNavierStokesHeaders.H"

#include "Timer.H"
extern Timer Everything;
Timer TMPI_INS_initialData_exchange("initData ex",  Everything, DTIMECOMM, true);
Timer TMPI_initialData_cp1         ("initData cp1", Everything, DTIMECOMM, true);
Timer TMPI_initialData_cp2         ("initData cp2", Everything, DTIMECOMM, true);
Timer TMPI_initialData_cp3         ("initData cp3", Everything, DTIMECOMM, true);

// initialize grid
void 
AMRNavierStokes::initialGrid(const Vector<Box>& a_new_grids)
{
  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::initialGrid " << m_level << endl;
  }
  m_level_grids = a_new_grids;
  if (a_new_grids.size() > 0) m_is_empty = false;


  // note -- this assumes levels are defined from coarse->fine
  // (which is a reasonable assumption)
  if (!m_is_empty) 
    {
      finestLevel(true);
      if (m_coarser_level_ptr != NULL) {
        AMRNavierStokes* crse_amrns_ptr = dynamic_cast<AMRNavierStokes*> (m_coarser_level_ptr);
        if (crse_amrns_ptr != NULL) {
          crse_amrns_ptr->finestLevel(false);
        } else {
          MayDay::Error("in AMRNavierStokes::initialGrid: m_coarser_level is not castable to AMRNavierStokes*");
        }
      }
    }


  const DisjointBoxLayout level_domain = loadBalance(a_new_grids);

  if (s_verbosity >= 4) {
    pout () << "new grids: " << endl;
    for (LayoutIterator lit = level_domain.layoutIterator(); lit.ok(); ++lit)
      {
        pout() << level_domain[lit()] << endl;
      }
  }

  IntVect ghostVect(D_DECL(1,1,1));

  // this is necessary because of the way initialization
  // is done -- AMRNavierStokes::initialGrid is called each time
  // through level-by-level building of the initial grid
  // hierarchy
  
  if (m_vel_new_ptr != NULL) {
    delete m_vel_new_ptr;
    m_vel_new_ptr = NULL;
  } 
  m_vel_new_ptr = new LevelData<FArrayBox>(level_domain, 
                                           s_num_vel_comps, 
                                           ghostVect);

  if (m_vel_old_ptr != NULL) {
    delete m_vel_old_ptr;
    m_vel_old_ptr = NULL;
  }
  m_vel_old_ptr = new LevelData<FArrayBox>(level_domain, 
                                           s_num_vel_comps, 
                                           ghostVect);

#ifdef DEBUG
  if (m_vel_save_ptr != NULL) {
    delete m_vel_save_ptr;
    m_vel_save_ptr = NULL;
  }
  m_vel_save_ptr = new LevelData<FArrayBox>(level_domain, 
                                            s_num_vel_comps, 
                                            ghostVect);  
#endif

  if (m_lambda_new_ptr != NULL) {
    delete m_lambda_new_ptr;
    m_lambda_new_ptr = NULL;
  }
  m_lambda_new_ptr = new LevelData<FArrayBox>(level_domain, 
                                              1, ghostVect);

  if (m_lambda_old_ptr != NULL) {
    delete m_lambda_old_ptr;
    m_lambda_old_ptr = NULL;
  }
  m_lambda_old_ptr = new LevelData<FArrayBox>(level_domain, 
                                              1, ghostVect);


#ifdef DEBUG
  if (m_lambda_save_ptr != NULL) {
    delete m_lambda_save_ptr;
    m_lambda_save_ptr = NULL;
  }
  m_lambda_save_ptr = new LevelData<FArrayBox>(level_domain, 
                                               1,
                                               ghostVect);  
#endif


  // now do scalars
  if (m_scal_new.size() != s_num_scal_comps) {
    // this seems a good way to initialize pointers
    // to null
    Vector<LevelData<FArrayBox>* > tempVect(s_num_scal_comps,NULL);
    m_scal_new = tempVect;
  }
  if (m_scal_old.size() != s_num_scal_comps) {
    Vector<LevelData<FArrayBox>* > tempVect(s_num_scal_comps,NULL);
    m_scal_old = tempVect;
  }

#ifdef DEBUG
  if (m_scal_save.size() != s_num_scal_comps) {
    Vector<LevelData<FArrayBox>* > tempVect(s_num_scal_comps,NULL);
    m_scal_save = tempVect;
  }
#endif

  // do flux registers as well
  if (m_scal_fluxreg_ptrs.size() < s_num_scal_comps) {
    m_scal_fluxreg_ptrs.resize(s_num_scal_comps,NULL);
  }

  for (int comp=0; comp<s_num_scal_comps; comp++) {
    if (m_scal_new[comp] != NULL) {
      delete m_scal_new[comp];
      m_scal_new[comp] = NULL;
    }
    m_scal_new[comp] = new LevelData<FArrayBox>(level_domain, 
                                                1,ghostVect);

    if (m_scal_old[comp] != NULL) {
      delete m_scal_old[comp];
      m_scal_old[comp] = NULL;
    }
    m_scal_old[comp] = new LevelData<FArrayBox>(level_domain, 
                                                1, ghostVect);

#ifdef DEBUG
    if (m_scal_save[comp] != NULL) {
      delete m_scal_save[comp];
      m_scal_save[comp] = NULL;
    }
    m_scal_save[comp] = new LevelData<FArrayBox>(level_domain,
                                                 1, ghostVect);
#endif  

    if (m_scal_fluxreg_ptrs[comp] != NULL) {
      delete m_scal_fluxreg_ptrs[comp];
      m_scal_fluxreg_ptrs[comp] = NULL;
    }
    m_scal_fluxreg_ptrs[comp] = new LevelFluxRegister;

  } // end loop over scalar components

  if (s_set_bogus_values) {
    LevelData<FArrayBox>& newVel = *m_vel_new_ptr;
    LevelData<FArrayBox>& oldVel = *m_vel_old_ptr;
    
    LevelData<FArrayBox>& newLambda = *m_lambda_new_ptr;
    LevelData<FArrayBox>& oldLambda = *m_lambda_old_ptr;
    
    DataIterator dit = m_vel_new_ptr->dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      oldVel[dit()].setVal(s_bogus_value);
      newVel[dit()].setVal(s_bogus_value);
      oldLambda[dit()].setVal(s_bogus_value);
      newLambda[dit()].setVal(s_bogus_value);
    }

    for (int comp=0; comp<s_num_scal_comps; comp++) {
      LevelData<FArrayBox>& newScal = *m_scal_new[comp];
      LevelData<FArrayBox>& oldScal = *m_scal_old[comp];
      
      
      for (dit.reset(); dit.ok(); ++dit) {
        oldScal[dit()].setVal(s_bogus_value);
        newScal[dit()].setVal(s_bogus_value);
      }
    } // end loop over scalar components
  }


  m_dx = s_domLength[1]/problemDomain().domainBox().size(1);
  

  // set up data structures -- actually, defer this until postInitialize
  //levelSetup(level_domain);
}

// set up data structures associated with this level
void
AMRNavierStokes::levelSetup(const DisjointBoxLayout& level_domain)
{
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::levelSetup " << m_level << endl;
  }


  CCProjector* crseProjPtr = NULL;
  CCProjector* fineProjPtr = NULL;
  int nRefCrse = -1;

  const DisjointBoxLayout* crseGridsPtr = NULL;

  if (m_coarser_level_ptr != NULL) {

    AMRNavierStokes* amrns_ptr = dynamic_cast<AMRNavierStokes*> (m_coarser_level_ptr);
    if (amrns_ptr != NULL) {
      crseGridsPtr = &(amrns_ptr->newVel().getBoxes());
      
      crseProjPtr = &(amrns_ptr->m_projection);
      nRefCrse = amrns_ptr->m_ref_ratio;

      m_coarse_average.define(level_domain, s_num_vel_comps, nRefCrse);
      m_coarse_average_lambda.define(level_domain, 1, nRefCrse);
      //m_coarse_average_scal.define(level_domain, 1, nRefCrse);
      

      // flux registers
      amrns_ptr->m_flux_register.undefine();
      amrns_ptr->m_flux_register.define(level_domain,
                                        *crseGridsPtr,
                                        m_problem_domain,
                                        nRefCrse,
                                        s_num_vel_comps);
      
      amrns_ptr->m_lambda_flux_reg.undefine();
      amrns_ptr->m_lambda_flux_reg.define(level_domain,
                                          *crseGridsPtr,
                                          m_problem_domain,
                                          nRefCrse,
                                          1);


      for (int comp=0; comp<s_num_scal_comps; comp++) {
        amrns_ptr->m_scal_fluxreg_ptrs[comp]->undefine();
        amrns_ptr->m_scal_fluxreg_ptrs[comp]->define(level_domain,
                                                     *crseGridsPtr,
                                                     m_problem_domain,
                                                     nRefCrse,
                                                     1);
      }


      amrns_ptr->m_flux_register.setToZero();
      amrns_ptr->m_lambda_flux_reg.setToZero();
      for (int comp=0; comp<s_num_scal_comps; comp++) {
        amrns_ptr->m_scal_fluxreg_ptrs[comp]->setToZero();
      }

      // define coarse-fine interpolation operator
      m_velCFInterp.define(level_domain, crseGridsPtr,
                           m_dx, nRefCrse, SpaceDim, m_problem_domain);

      // define viscous solver here as well
      if (s_nu > 0.0) {

        m_velocityPoissonOp.define(level_domain, crseGridsPtr,
                                   m_dx, nRefCrse, m_problem_domain,
                                   false, SpaceDim);
        
        // set levelOp physical BCs -- use fact that viscous
        // BCs are the same for each component (this may break at
        // some point)
        //const DomainGhostBC* thisBCPtr = velBCPtr.DomGhostBC(0);

        // don't do this here -- need to do differently for each comp.
        //m_velocityPoissonOp.setDomainGhostBC(thisBC);
        
        HelmholtzOp viscousOp;
        viscousOp.define(level_domain, crseGridsPtr,
                         m_dx, nRefCrse, m_problem_domain, false);
        // set viscousOp physical BCs -- use fact that viscous
        // BCs are the same for each component (this may break at
        // some point)

        // do this in a componentwise manner when we need it
        // this is just a placeholder so that things can be defined
        DomainGhostBC* thisBCPtr = m_physBCPtr->viscousSolveBC(0);
        viscousOp.setDomainGhostBC(*thisBCPtr);
        delete thisBCPtr;
        // for now, set helmholtz coefficient to be one
        viscousOp.setHelmCoeff(1.0);

        // for now, set nComp to be one -- eventually probably 
        // will want this to be SpaceDim
        int nComp = 1;
        m_viscous_solver.define(level_domain, crseGridsPtr,
                                m_problem_domain,m_dx,nRefCrse,
                                &viscousOp, nComp, m_limitSolverCoarsening);
        m_viscous_solver.setTolerance(s_viscous_solver_tol);
        m_viscous_solver.setNumSmoothUp(s_viscous_num_smooth_up);
        m_viscous_solver.setNumSmoothDown(s_viscous_num_smooth_down);

        if (s_verbosity >= 3) {
          m_viscous_solver.setVerbose(true);
        } else {
          m_viscous_solver.setVerbose(false);
        }
      }
    
      bool isDiffusive = false;
      for (int comp=0; comp<s_num_scal_comps; comp++) {
        if (s_scal_coeffs[comp] > 0) isDiffusive = true;
      }

      if (isDiffusive) {
        m_scalarsPoissonOp.define(level_domain, crseGridsPtr,
                                  m_dx, nRefCrse, m_problem_domain,
                                  false, 1);
      }
      
    } else {
      MayDay::Error("in AMRNavierStokes::levelSetup: m_coarser_level_ptr is not castable to AMRNavierStokes*");
    }

  } else {
    // if no coarser level, still need to define viscous solver
    if (s_nu > 0.0) {

      m_velocityPoissonOp.define(level_domain, crseGridsPtr,
                                 m_dx, nRefCrse, m_problem_domain,
                                 false, SpaceDim);

      // set levelOp physical BCs -- use fact that viscous
      // BCs are the same for each component (this may break at
      // some point)
      //boxVelBC velBC(true);
      //velBC.setAllBCTypes();
      //const DomainGhostBC& thisBC = velBC.DomGhostBC(0);
      //m_velocityPoissonOp.setDomainGhostBC(thisBC);
      
      bool isDiffusive = false;
      for (int comp=0; comp<s_num_scal_comps; comp++) {
        if (s_scal_coeffs[comp] > 0) isDiffusive = true;
      }

      if (isDiffusive) {
        m_scalarsPoissonOp.define(level_domain, crseGridsPtr,
                                   m_dx, nRefCrse, m_problem_domain,
                                   false, 1);
      }

      HelmholtzOp viscousOp;
      viscousOp.define(level_domain, crseGridsPtr,
                       m_dx, nRefCrse, m_problem_domain, false);

      // for now, set helmholtz coefficient to be one
      viscousOp.setHelmCoeff(1.0);
      DomainGhostBC* thisBCPtr = m_physBCPtr->viscousSolveBC(0);      
      viscousOp.setDomainGhostBC(*thisBCPtr);    
      delete thisBCPtr;

      // for now, set viscous solver's ncomp to be 1
      int ncomp = 1;
      m_viscous_solver.define(level_domain, crseGridsPtr, m_problem_domain,
                              m_dx, 0, &viscousOp, ncomp,
                              m_limitSolverCoarsening);
      m_viscous_solver.setTolerance(s_viscous_solver_tol);
      m_viscous_solver.setNumSmoothUp(s_viscous_num_smooth_up);
      m_viscous_solver.setNumSmoothDown(s_viscous_num_smooth_down);

      if (s_verbosity >= 3) {
        m_viscous_solver.setVerbose(true);
      } else {
        m_viscous_solver.setVerbose(false);
      }
      
    }
  }

  if (m_finer_level_ptr != NULL) {
    AMRNavierStokes* amrns_ptr = dynamic_cast<AMRNavierStokes*> (m_finer_level_ptr);
    if (amrns_ptr != NULL) {
      fineProjPtr = &(amrns_ptr->m_projection);
    } else { 
      MayDay::Error("in AMRNavierStokes::levelSetup: m_finer_level_ptr is not castable to AMRNavierStokes*");
    }
  } else {
    m_flux_register.undefine();
  }

  // pass on verbosity
  m_projection.verbosity(s_verbosity);
  // pass on whether to limit coarsenings
  m_projection.limitSolverCoarsening(m_limitSolverCoarsening);
  
  m_projection.define(level_domain, crseGridsPtr, m_problem_domain, 
                      m_dx, fineProjPtr,
                      crseProjPtr, nRefCrse, m_level, *m_physBCPtr);
  

#ifdef NEW_ADVECTION
  // set up a default scalar IBC so we can at least define the OldPatchGodunov object
  OldPhysIBC* scalIBC = m_physBCPtr->scalarTraceIBC(0);  
  m_patchGod.setPhysIBC(scalIBC);

  // set up OldPatchGodunov object
  m_patchGod.define(m_problem_domain, m_dx);
  
  // don't use artificial viscosity
  bool useArtVisc = false;
  Real artVisc = -1.0;
  m_patchGod.setArtificialViscosity(useArtVisc, artVisc);

  
  // slope computation parameters -- may want to make these user-controlled
  bool fourthOrderSlopes = true;
  //bool fourthOrderSlopes = false;
  bool useFlattening = false;
  bool limitSlopes = false;
  m_patchGod.setSlopeParameters(fourthOrderSlopes, 
                                useFlattening,
                                limitSlopes);
#endif

  if (s_boussinesq) {
    // set up data points to be recorded
    m_dataPoints.resize(1);
    m_dataPoints[0][0] = 0.1810;
    m_dataPoints[0][1] = 7.3700;
    assert (SpaceDim != 3);
  }

}

                            
  
void  
AMRNavierStokes::readParameters() 
{

  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::readParameters " << endl;
  } 

  ParmParse ppNS("ns");

  Real defaultDomainLength = 1.0;
  vector<Real> tempVect(SpaceDim);
  for (int dir=0; dir<SpaceDim; dir++) {
    tempVect[dir] = defaultDomainLength;
  }
  //  ppNS.queryarr("domainLength", s_domLength, 0, SpaceDim);
  ppNS.queryarr("domainLength", tempVect, 0, SpaceDim);
  s_domLength = tempVect;
  
  ppNS.query("init_shrink", s_init_shrink);

  ppNS.query("max_dt", s_max_dt);

  ppNS.query("setDt", s_prescribedDt);

  int temp = s_project_initial_vel;
  ppNS.query("project_initial_vel", temp);
  s_project_initial_vel = (temp == 1);

  temp = s_initialize_pressures;
  ppNS.query("init_pressures", temp);
  s_initialize_pressures = (temp == 1);

  temp = s_smooth_after_regrid;
  ppNS.query("smooth_after_regrid", temp);
  s_smooth_after_regrid = (temp == 1);

  ppNS.query("postRegrid_smoothing_coeff", s_regrid_smoothing_coeff);

  temp = s_dump_smoothing_plots;
  ppNS.query("dump_smoothing_plots", temp);
  s_dump_smoothing_plots = (temp == 1);

  if (s_initialize_pressures) {
    ppNS.query("num_init_passes", s_num_init_passes);
  }

  temp = s_reflux_momentum;
  ppNS.query("reflux_momentum", temp);
  s_reflux_momentum = (temp == 1);

  temp = s_reflux_normal_momentum;
  ppNS.query("reflux_normal_momentum", temp);
  s_reflux_normal_momentum = (temp == 1);

  temp = s_tag_vorticity;
  ppNS.query("tag_vorticity", temp);
  s_tag_vorticity = (temp == 1);

  temp = s_tag_theta;
  ppNS.query("tag_theta", temp);
  s_tag_theta = (temp == 1);

  ppNS.query("vorticity_tagging_factor", s_vort_factor);
  
  ppNS.query("tags_grow", s_tags_grow);

  ppNS.query("viscosity", s_nu);

  temp = s_implicit_reflux;
  ppNS.query("implicit_reflux", temp);
  s_implicit_reflux = (temp == 1);

  temp = s_applyFreestreamCorrection;
  ppNS.query("applyFreestreamCorrection",temp);
  s_applyFreestreamCorrection = (temp == 1);

  temp = s_reflux_scal;
  ppNS.query("reflux_scalars", temp);
  s_reflux_scal = (temp == 1);

  temp = s_implicit_scal_reflux;
  ppNS.query("implicit_scal_reflux", temp);
  s_implicit_scal_reflux = (temp == 1);

  ppNS.query("viscous_solver_type", s_viscous_solver_type);

  ppNS.query("viscous_solver_tolerance", s_viscous_solver_tol);

  ppNS.query("viscous_num_smooth_up", s_viscous_num_smooth_up);
  ppNS.query("viscous_num_smooth_down", s_viscous_num_smooth_down);

  ppNS.query("num_scalars", s_num_scal_comps);

  if (s_num_scal_comps > 0) {
    s_scal_coeffs.resize(s_num_scal_comps);
    ppNS.getarr("scal_diffusion_coeffs", tempVect,
                0,s_num_scal_comps);
    s_scal_coeffs = tempVect;
  }

#ifdef BOUSSINESQ
  temp = s_boussinesq;
  ppNS.query("boussinesq", temp);
  s_boussinesq = (temp == 1);

  if (s_boussinesq) {
    // this will overwrite viscosity and theta's diffusion coeff
    ppNS.query("RayleighNo", s_rayleighNo);
    ppNS.query("PrandtlNo", s_prandtlNo);

    s_nu = sqrt(s_prandtlNo/s_rayleighNo);
    s_scal_coeffs[THETA_COMP] = 1.0/(sqrt(s_rayleighNo*s_prandtlNo));

    //ppNS.query("wallTemp", s_wallTemp);
  }
#endif

  temp = s_specifyInitialGrids;
  ppNS.query("specifyInitialGrids", temp);
  s_specifyInitialGrids = (temp == 1);

  if (s_specifyInitialGrids) {
    ppNS.get("initialGridFile", s_initialGridFile);
  }

  
  temp = s_init_vel_from_vorticity;
  ppNS.query("initVelFromVorticity", temp);
  s_init_vel_from_vorticity = (temp == 1);
  
  ppNS.query("backgroundVelocity", s_backgroundVel);


  temp = s_write_divergence;
  ppNS.query("writeDivergence", temp);
  s_write_divergence = (temp == 1);

  temp =  s_write_lambda;
  ppNS.query("writeLambda", temp);
  s_write_lambda = (temp == 1);

  temp = s_write_time_derivatives;
  ppNS.query("writeTimeDerivatives", temp);
  s_write_time_derivatives = (temp == 1);

  temp = s_write_vorticity;
  ppNS.query("writeVorticity", temp);
  s_write_vorticity = (temp == 1);

  temp = s_write_scalars;
  ppNS.query("writeScalars", temp);
  s_write_scalars = (temp == 1);

  /// include d(scalars)/dt in plotfile?
  temp = s_write_dScalar_dt;
  ppNS.query("writeDscalDt", temp);
  s_write_dScalar_dt = (temp == 1);

  /// include strain-rates in plotfile? (2D only)
  temp = s_write_strains;
  ppNS.query("writeStrains", temp);
  s_write_strains = (temp == 1);

  /// include freestream preservation correcion in plotfile?
  temp = s_write_grad_eLambda;  
  ppNS.query("writeGradELambda", temp);
  s_write_grad_eLambda = (temp == 1);

  /// include solution error (if there is an exact solution)
  temp = s_write_error;
  ppNS.query("writeError", temp);
  s_write_error = (temp == 1);

#if 0
  /// include solution error (if there is an exact solution)
  temp = s_write_curl_error;
  ppNS.query("writeCurlError", temp);
  s_write_curl_error = (temp == 1);
#endif

  // include processor id information in plotfile?
  temp = s_write_proc_ids;
  ppNS.query("writeProcIDs", temp);
  s_write_proc_ids = (temp == 1);
 
  // compute and report scalar error after composite timestep?
  temp = s_compute_scal_err;
  ppNS.query("computeScalErr", temp);
  s_compute_scal_err = (temp ==1);

  // write out RHS for initial vortex-stream function solves?
  temp = s_write_initial_solve_rhs;
  ppNS.query("writeInitialSolveRHS", temp);
  s_write_initial_solve_rhs = (temp == 1);

  temp = s_write_grids;
  ppNS.query("writeGrids", temp);
  s_write_grids = (temp == 1);

  // now dump this all back out...
  if (s_verbosity > 2)
    {
      pout() << "NavierStokes inputs:" << endl;

      pout() << "  domainLength = " << s_domLength << endl;
      
      pout() << "  init_shrink = " <<  s_init_shrink << endl;
      
      pout() << "  max_dt = " << s_max_dt << endl;
      
      pout() << "  setDt = " << s_prescribedDt << endl;
      
      pout() << "  project_initial_vel"  << " = " << s_project_initial_vel
             << endl;

      pout() << "  init_pressures = " << s_initialize_pressures << endl;
      
      pout() << "  smooth_after_regrid = " << s_smooth_after_regrid
             << endl;

      pout() << "  postRegrid_smoothing_coeff = " 
             << s_regrid_smoothing_coeff  << endl;
      
      pout() << "  dump_smoothing_plots = " << s_dump_smoothing_plots
             << endl;
      
      if (s_initialize_pressures) {
        pout() << "  num_init_passes = " << s_num_init_passes << endl;
      }
      
      pout() << "  reflux_momentum = " << s_reflux_momentum << endl;

      pout() << "  reflux_normal_momentum = " 
             << s_reflux_normal_momentum << endl;
      
      pout() << "  tag_vorticity = " << s_tag_vorticity << endl;
      
      pout() << "  tag_theta = " << s_tag_theta << endl;
      
      pout() << "  vorticity_tagging_factor = " << s_vort_factor << endl;
      
      pout() << "  tags_grow = " << s_tags_grow << endl;
      
      pout() << "  viscosity = " << s_nu << endl;
      
      pout() << "  implicit_reflux = " << s_implicit_reflux << endl;
      
      pout() << "  applyFreestreamCorrection = " 
             << s_applyFreestreamCorrection << endl;
      
      pout() << "  reflux_scalars = " << s_reflux_scal << endl;
      
      pout() << "  implicit_scal_reflux = " << s_implicit_scal_reflux
             << endl;
      
      pout() << "  viscous_solver_type = " << s_viscous_solver_type
             << endl;
      
      pout() << "  viscous_solver_tolerance = " << s_viscous_solver_tol
             << endl;

      pout() << "  viscous_solver_num_smooth_up = " 
             << s_viscous_num_smooth_up << endl;

      pout() << "  viscous_solver_num_smooth_down = " 
             << s_viscous_num_smooth_down << endl;
      
      pout() << "  num_scalars = " << s_num_scal_comps << endl;
      
      if (s_num_scal_comps > 0) {
        pout () << "  scal_diffusion_coeffs = " << s_scal_coeffs << endl;
      }
      
#ifdef BOUSSINESQ
      pout() << "  boussinesq = " << s_boussinesq << endl;
      
      if (s_boussinesq) {
        // this will overwrite viscosity and theta's diffusion coeff
        pout() << "RayleighNo = " << s_rayleighNo << endl;
        pout() << "  PrandtlNo = " << s_prandtlNo << endl;
        
      }
#endif
      
      pout() << "  specifyInitialGrids = " << s_specifyInitialGrids
             << endl;
      
      if (s_specifyInitialGrids)
        {
          pout() << "  initialGridFile = " << s_initialGridFile << endl;
        }
      
      
      pout() << "  initVelFromVorticity = " << s_init_vel_from_vorticity
             << endl;
      
      pout() << "  backgroundVelocity = " << s_backgroundVel << endl;
      
      pout() << "  writeDivergence = " << s_write_divergence << endl;
      
      pout() << "  writeLambda = " << s_write_lambda << endl;
      
      pout() << "  writeTimeDerivatives = " << s_write_time_derivatives
             << endl;
      
      pout() << "  writeVorticity = " << s_write_vorticity << endl;
      
      pout() << "  writeScalars = " << s_write_scalars << endl;
      
      /// include d(scalars)/dt in plotfile?
      pout() << "  writeDscalDt = " << s_write_dScalar_dt << endl;
      
      /// include strain-rates in plotfile? (2D only)
      pout() << "  writeStrains = " << s_write_strains << endl;
      
      /// include freestream preservation correcion in plotfile?
      pout() << "  writeGradELambda = " << s_write_grad_eLambda << endl;
      
      /// include solution error (if there is an exact solution)
      pout() << "  writeError = " << s_write_error  << endl;
      
#if 0
      /// include solution error (if there is an exact solution)
      pout() << "  writeCurlError = " << s_write_curl_error << endl;
#endif
      
      // include processor id information in plotfile?
      pout() << "  writeProcIDs = " << s_write_proc_ids << endl;

      pout() << "  compute scalar error = " << s_compute_scal_err << endl;
    } 
  
  // set flag to indicate that we've done this
  s_ppInit  =true;
  
}


void
AMRNavierStokes::initialData()
{

  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::initialData " << m_level << endl;
  } 

  DataIterator dit = newVel().dataIterator();
  
  if (s_init_vel_from_vorticity) {
    // first initialize vorticity field -- do this in an AMR way
    if (finestLevel()) {
      
      Vector<LevelData<FArrayBox>*> initialVort(m_level+1, NULL);
      Vector<LevelData<FArrayBox>*> streamFunc(m_level+1, NULL);
      Vector<DisjointBoxLayout> amrGrids(m_level+1);
      Vector<ProblemDomain> amrDomains(m_level+1);
      Vector<Real> amrDx(m_level+1);
      Vector<int> refRatios(m_level+1);

      int nVortComp;
      if (SpaceDim == 2) {
        nVortComp = 1;
      } else {
        nVortComp = SpaceDim;
      }
      
      // start at finest (this) level and progress to coarser levels
      AMRNavierStokes* thisLevelPtr = this;
      for (int lev=m_level; lev>=0; lev--) {
        const DisjointBoxLayout& levelGrids 
          = thisLevelPtr->newVel().getBoxes();
        
        amrGrids[lev] = levelGrids;
        initialVort[lev] = new LevelData<FArrayBox>(levelGrids, nVortComp,
                                                    IntVect::TheZeroVector());
        
        streamFunc[lev] = new LevelData<FArrayBox>(levelGrids, nVortComp,
                                                   IntVect::TheUnitVector());
        
        amrDomains[lev] = thisLevelPtr->problemDomain();
        Real levelDx = thisLevelPtr->Dx();
        amrDx[lev] = levelDx;
        refRatios[lev] = thisLevelPtr->refRatio();
        
        LevelData<FArrayBox>& thisLevStreamFunc = *streamFunc[lev];
        
        DataIterator levelDit = levelGrids.dataIterator();
        for (levelDit.reset(); levelDit.ok(); ++levelDit) {
          // initialize data
          thisLevStreamFunc[levelDit()].setVal(0.0);          
          
          // since there is only one component of vorticity
          // in this case, put it in the 0th component
          TF_INITVORT.start();
          for (int comp=0; comp<nVortComp; comp++) {
            FORT_INITVORT(CHF_FRA1((*initialVort[lev])[levelDit()],comp),
                          CHF_INT(comp),
                          CHF_REAL(levelDx));
          } 
          TF_INITVORT.stop();
        }
        
        // index to next coarser level
        if (lev > 0) {
          thisLevelPtr = thisLevelPtr->crseNSPtr();
        }
      }     
      
      PoissonOp bogusPoissonOp;
      Interval solverComps(0,0);
      DomainGhostBC* solverBCPtr = m_physBCPtr->streamFunctionBC(solverComps);
      bogusPoissonOp.setDomainGhostBC(*solverBCPtr);
      delete solverBCPtr;
      
      int numComp = 1;
      AMRSolver solver(amrGrids,amrDomains,amrDx,refRatios,
                       amrGrids.size(),0, &bogusPoissonOp,
                       numComp, m_limitSolverCoarsening);

      if (s_verbosity >= 3) {
        solver.setVerbose(true);
      } else {
        solver.setVerbose(false);
      }

      // if desired, write out this RHS
      if (s_write_initial_solve_rhs)
        {
          string fname;
          if (SpaceDim == 2) 
            {
              fname = "InitialRHS.2d.hdf5";
            }
          else
            { 
              fname = "InitialRHS.3d.hdf5";
            }

          Vector<string> dataNames(nVortComp);
          if (SpaceDim == 2)
            {
              dataNames[0] = "rhs";
            }
          else 
            {
              dataNames[0] = "xRHS"; 
              dataNames[1] = "yRHS"; 
              dataNames[2] = "zRHS"; 
            }

#ifdef HDF5
          WriteAMRHierarchyHDF5(fname, 
                                amrGrids,
                                initialVort,
                                dataNames,
                                amrDomains[0].domainBox(),
                                amrDx[0],
                                m_dt,
                                m_time, 
                                refRatios,
                                amrGrids.size());
#endif
        }

      if (SpaceDim == 2) {
        TimeSA_initialData.start();
        IterationFlavor = COUNTER_solveAMR_INSinit + m_level;
        solver.solveAMR(streamFunc, initialVort);
        TimeSA_initialData.stop();

        // reapply BCs and compute velocity field
        int nRefCrse=-1;
        int nRefFine=-1;

        assert (thisLevelPtr->m_level == 0);    

        Interval psiComps(0,0);
        DomainGhostBC* psiBCPtr = m_physBCPtr->streamFunctionBC(psiComps);

        for (int lev=0; lev<=m_level; lev++) {
          LevelData<FArrayBox> gradPsi(amrGrids[lev],SpaceDim,
                                       IntVect::TheZeroVector());

          DataIterator levelDit = streamFunc[lev]->dataIterator();        
          for (levelDit.reset(); levelDit.ok(); ++levelDit) {
            psiBCPtr->applyInhomogeneousBCs((*streamFunc[lev])[levelDit()],
                                             amrDomains[lev], amrDx[lev]);
          }

          // now compute derivatives of psi
          LevelData<FArrayBox>* crsePsiPtr = NULL;
          LevelData<FArrayBox>* finePsiPtr = NULL;

          if (lev > 0) {
            crsePsiPtr = streamFunc[lev-1];
            nRefCrse = refRatios[lev-1];
          }

          if (lev < m_level) {
            finePsiPtr = streamFunc[lev+1];
            nRefFine = refRatios[lev];
          }
          
          compGradientCC(gradPsi, *streamFunc[lev],
                         crsePsiPtr, finePsiPtr,
                         amrDx[lev], nRefCrse, nRefFine,
                         amrDomains[lev]);

          // now copy derivatives of psi into the proper velocity components
          Interval xVelComp(0,0);
          Interval yVelComp(1,1);
          Interval xDerivativeComp(0,0);
          Interval yDerivativeComp(1,1);

          LevelData<FArrayBox>& levelVel = thisLevelPtr->newVel();

          TMPI_initialData_cp1.start();
          gradPsi.copyTo(xDerivativeComp,levelVel,yVelComp);
          gradPsi.copyTo(yDerivativeComp,levelVel,xVelComp);
          TMPI_initialData_cp1.stop();

          // multiply x-component of vel by -1
          for (levelDit.reset(); levelDit.ok(); ++levelDit) {
            levelVel[levelDit()].mult(-1.0,0,1);
          }
          
          thisLevelPtr = thisLevelPtr->finerNSPtr();
        } // end loop over levels

        delete psiBCPtr;

        // end if SpaceDim == 2
      } else if (SpaceDim == 3) {
        // need to loop over directions in this case and do each one
        // independently.
        
        // first define temporary storage
        Vector<LevelData<FArrayBox>*> compVort(m_level+1,NULL);
        Vector<LevelData<FArrayBox>*> compPsi(m_level+1,NULL);

        // start at finest (this) level and progress to coarse levels
        for (int lev=0; lev<compVort.size(); lev++) {
          compVort[lev] = new LevelData<FArrayBox>(amrGrids[lev],1,
                                                   IntVect::TheZeroVector());

          compPsi[lev] = new LevelData<FArrayBox>(amrGrids[lev],1,
                                                  IntVect::TheUnitVector());
        }

        // now compute stream function one direction at a time
        for (int dim=0; dim<SpaceDim; ++dim) {
          // copy vorticity into temp storage
          Interval dirComps(dim,dim);
          Interval destComps(0,0);

          TMPI_initialData_cp2.start();
          for (int lev=0; lev < initialVort.size(); lev++) {
            initialVort[lev]->copyTo(dirComps, *compVort[lev], destComps);
          }  
          TMPI_initialData_cp2.stop();
          
          TimeSA_initialData.start();
          IterationFlavor = COUNTER_solveAMR_INSinit + m_level;
          solver.solveAMR(compPsi, compVort);
          TimeSA_initialData.stop();

          // now copy to composite psi
          TMPI_initialData_cp3.start();
          for (int lev=0; lev<compPsi.size(); lev++) {
            compPsi[lev]->copyTo(destComps, *streamFunc[lev], dirComps);
          }
          TMPI_initialData_cp3.stop();

        } // end loop over dimensions

        // now compute vel = curl(streamFunc) 
        thisLevelPtr = this;

        Interval psiComps(0,SpaceDim-1);
        DomainGhostBC* psiBCPtr = m_physBCPtr->streamFunctionBC(psiComps);

        for (int lev=m_level; lev >= 0; lev--) {
          //LevelData<FArrayBox>* crsePsiPtr = NULL; //not used?  (ndk)
          //LevelData<FArrayBox>* finePsiPtr = NULL; //not used?  (ndk)
          const DisjointBoxLayout& levelGrids = streamFunc[lev]->getBoxes();

          DataIterator levelDit = streamFunc[lev]->dataIterator();
          
          LevelData<FArrayBox>& levelVel = thisLevelPtr->newVel();

          // this isn't tested yet
          // do averaging down if necessary
          if (lev < m_level)
            {
              LevelData<FArrayBox>& finePsi = *streamFunc[lev+1];
              const DisjointBoxLayout& fineGrids = finePsi.getBoxes();
              CoarseAverage PsiAvgDown(fineGrids, finePsi.nComp(),
                                       refRatios[lev]);
              PsiAvgDown.averageToCoarse(*streamFunc[lev],
                                         finePsi);
            }
              
          
          // first apply physical BC's to stream function
          for (levelDit.reset(); levelDit.ok(); ++levelDit) {
            psiBCPtr->applyInhomogeneousBCs((*streamFunc[lev])[levelDit()],
                                             amrDomains[lev], amrDx[lev]);
          }

          // do coarse-fine BC's
          if (lev > 0) {
            QuadCFInterp interpolator(amrGrids[lev],
                                      &amrGrids[lev-1],
                                      amrDx[lev],
                                      refRatios[lev-1],
                                      SpaceDim, amrDomains[lev]);

            interpolator.coarseFineInterp(*streamFunc[lev], 
                                          *streamFunc[lev-1]);

          }
          
          TMPI_INS_initialData_exchange.start();
          streamFunc[lev]->exchange(psiComps);
          TMPI_INS_initialData_exchange.stop();
        
          for (levelDit.reset(); levelDit.ok(); ++levelDit) {

            TF_COMPUTEVORT.start();
            for (int dir=0; dir<SpaceDim; dir++) {
              FORT_COMPUTEVORT(CHF_FRA1(levelVel[levelDit()],dir),
                               CHF_CONST_FRA((*streamFunc[lev])[levelDit()]),
                               CHF_BOX(levelGrids[levelDit()]),
                               CHF_CONST_REAL(amrDx[lev]),
                               CHF_CONST_INT(dir));
            } // end loop over directions
            TF_COMPUTEVORT.stop();

            // get signs right!
            levelVel[levelDit()].mult(-1.0);
            // add in background velocity
            levelVel[levelDit()].plus(s_backgroundVel, SpaceDim-1,1);
          } // end loop over grids
#if 0
          // this is just for debugging
          Interval velComps(0,SpaceDim-1);
          streamFunc[lev]->copyTo(velComps, levelVel, velComps);
#endif

          // average down if necessary
                    // this isn't tested yet
          // do averaging down if necessary
          if (lev < m_level)
            {
              LevelData<FArrayBox>& fineVel = thisLevelPtr->finerNSPtr()->newVel();
              

              const DisjointBoxLayout& fineGrids = fineVel.getBoxes();
              CoarseAverage VelAvgDown(fineGrids, fineVel.nComp(),
                                       refRatios[lev]);
              VelAvgDown.averageToCoarse(levelVel,
                                         fineVel);
            }

          thisLevelPtr = thisLevelPtr->crseNSPtr();
        } // end loop over levels
          
        delete psiBCPtr;
          
        // clean up temporary storage!
        for (int lev=0; lev<compVort.size(); lev++) {
          if (compVort[lev] != NULL) {
            delete compVort[lev];
            compVort[lev] = NULL;
          }
          
          if (compPsi[lev] != NULL) {
            delete compPsi[lev];
            compPsi[lev] = NULL;
          }
        }
        
      } else {
        MayDay::Error("HEY! What's with the oddly dimensional space???");
      }
      
      // clean up memory
      for (int lev=0; lev<amrGrids.size(); lev++) {
        if (initialVort[lev] != NULL) {
          delete initialVort[lev];
          initialVort[lev] = NULL;
        } 
        if (streamFunc[lev] != NULL) {
          delete streamFunc[lev];
          streamFunc[lev] = NULL;
        }
      }
      
      
    } // end if finest level      
  } else {
    TF_INITVEL.start();
    for (dit.reset(); dit.ok(); ++dit) {
      
      //newVel()[dit()].setVal(0.0,newVel()[dit()].box(),0,SpaceDim);
      
      FORT_INITVEL(CHF_FRA(newVel()[dit()]),
                   CHF_REAL(m_dx));
    }
    TF_INITVEL.stop();
  }  
  
  for (dit.reset(); dit.ok(); ++dit) {
    // now initialize Lambda
    newLambda()[dit()].setVal(1.0, 0);
    
    // now init (cell-centered) scalars
    int isEdgeCentered = 0;
    TF_INITSCAL.start();
    for (int comp=0; comp<s_num_scal_comps; comp++) {
      FORT_INITSCAL(CHF_FRA(newScal(comp)[dit()]),
                    CHF_REAL(m_dx), 
                    CHF_REAL(m_time),
                    CHF_INT(comp),
                    CHF_BOX(m_problem_domain.domainBox()),
                    CHF_INT(isEdgeCentered));
      
      // if we're doing boussinesq, also want old scalars initialized
      FORT_INITSCAL(CHF_FRA(oldScal(comp)[dit()]),
                    CHF_REAL(m_dx), 
                    CHF_REAL(m_time),
                    CHF_INT(comp),
                    CHF_BOX(m_problem_domain.domainBox()),
                    CHF_INT(isEdgeCentered));
    }
    TF_INITSCAL.stop();
  }
}
  

void 
AMRNavierStokes::postInitialize()
{

  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::postInitialize " << m_level << endl;
  } 

  // initialize data structures which haven't yet been initialized
  const DisjointBoxLayout& level_domain = newVel().getBoxes();
  levelSetup(level_domain);

  // set up for initial velocity projection -- do this from level 0,
  // since post_initialize is called from fine->coarse
  if (m_level == 0) {
    AMRNavierStokes* thisLevelData = this;
    while (!thisLevelData->finestLevel()) 
      {
        thisLevelData = thisLevelData->finerNSPtr();
      }
    assert (thisLevelData->finestLevel());
    int numLevels = thisLevelData->m_level + 1;

    Vector<LevelData<FArrayBox>* > amrVel(numLevels);

    thisLevelData = this;
    for (int lev=m_level; lev <numLevels; lev++) {
      amrVel[lev] = &(thisLevelData->newVel());

      if (!thisLevelData->finestLevel())
        thisLevelData = 
          dynamic_cast <AMRNavierStokes*> (thisLevelData->m_finer_level_ptr);
    }
  
    
    
    CCProjector& level0Proj = m_projection;
    
    // set physical boundary conditions on velocity
    bool isViscous = (s_nu > 0.0);
    basicVelBC* velBCPtr = m_physBCPtr->uStarBC(isViscous);
    
    thisLevelData = this;
    
    for (int lev=0; lev< numLevels; lev++) {
      const ProblemDomain& levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();
      
      velBCPtr->applyBCs(*amrVel[lev], levelDomain, levelDx);

      if (thisLevelData->m_finer_level_ptr != NULL) {
        thisLevelData = dynamic_cast<AMRNavierStokes*> (thisLevelData->m_finer_level_ptr);
      }
    }


    bool homoBC = false;
    if (s_project_initial_vel) {
      level0Proj.initialVelocityProject(amrVel, homoBC);
    }
    
    thisLevelData = this;

    // need to reset boundary conditions here
    for (int lev=0; lev< numLevels; lev++) {
      const ProblemDomain& levelDomain = thisLevelData->problemDomain();
      Real levelDx = thisLevelData->Dx();
      
      velBCPtr->applyBCs(*amrVel[lev], levelDomain, levelDx);
      
      if (thisLevelData->m_finer_level_ptr != NULL) {
        thisLevelData = dynamic_cast<AMRNavierStokes*> (thisLevelData->m_finer_level_ptr);
      }
    }

    delete velBCPtr;
    

    if (s_initialize_pressures) {
      initializeGlobalPressure();

    }

    // finally, if desired, dump out grids and processor mappings
    if (s_write_grids)
      {
#ifdef MPI         
        if (procID() == 0) {
#endif
          ofstream os("grids.out", ios::out);
          if (os.fail())
            {
              pout() << "cannot open grid output file " << os << endl;
              MayDay::Error();
            }
          
          thisLevelData = this;
          os << "NumLevels = " << numLevels << endl;
          for (int lev=0; lev<numLevels; lev++) 
            {
              const DisjointBoxLayout& thisLevelGrids 
                = thisLevelData->m_vel_new_ptr->getBoxes();
              os << "Number of Grids on Level " << lev << ": " 
                 << thisLevelGrids.size() << endl;
              // note that we don't need an endl here because
              // BoxLayout::operator<< provides a '\n' at the end
              os << thisLevelGrids;
              thisLevelData = thisLevelData->finerNSPtr();
            }
          os.close();
#ifdef MPI
        } // end if procID == 0
#endif
      } // end if writing grids to file
  

  } // end if level 0

}
