/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/


#include "AMRNavierStokesHeaders.H"

#include "Timer.H"
extern Timer Everything;
Timer TMPI_advectDiffScalar_exchange("advDiffScal ex", Everything, DTIMECOMM, true);
Timer TMPI_advectDiffScal_cp        ("advDiffScal cp", Everything, DTIMECOMM, true);
Timer TMPI_advectDiffScal_NA_cp     ("advDiffScalNA cp", Everything, DTIMECOMM, true);

void
AMRNavierStokes::advectScalar(LevelData<FArrayBox>& a_new_scal,
			      LevelData<FArrayBox>& a_old_scal,
			      LevelData<FluxBox>& a_advVel,
			      LevelFluxRegister* a_crseFluxRegPtr,
			      LevelFluxRegister& a_fineFluxReg,
#ifdef NEW_ADVECTION
                              OldPhysIBC* a_scalIBC,
#endif
			      Real a_dt)
{
  
  if (s_verbosity >= 3) 
    {
      pout() << "AMRNavierStokes::advectScalar: " << m_level << endl;
    }
  
  const DisjointBoxLayout& levelGrids = a_new_scal.getBoxes();


  /// need to build grown grids to get be able to do all of
  /// tracing properly -- this is necessary to sync with old
  /// code
  IntVect advect_grow(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));

  LevelData<FArrayBox> local_old_vel(levelGrids, SpaceDim, advect_grow);
  // m_time contains the time at which the new state is centered
  Real old_time = m_time - m_dt;
  fillVelocity(local_old_vel, old_time);
  //local_old_vel.exchange(local_old_vel.interval());

  DataIterator dit = a_new_scal.dataIterator();
  dit.reset();
  int nScal = a_new_scal.nComp();
  Interval scalComps(0,nScal-1);  
  
#ifdef NEW_ADVECTION
  // set up patchGodunov for this problem
  m_patchGod.setPhysIBC(a_scalIBC);
  m_patchGod.setNumVar(nScal);
  m_patchGod.setCurrentTime(old_time);

#endif

  
#ifdef NEW_ADVECTION
  IntVect ghostVect = IntVect::Unit;
  LevelData<FluxBox> edgeScal(levelGrids, nScal, ghostVect);
#else
  LevelData<FluxBox> edgeScal(levelGrids, nScal);
#endif

#ifdef MPI
  if (s_verbosity >= 3) {
    //pout() << "procID: " << procID() << "tracing scalars" << endl;
  }
#endif

  // haven't implemented this yet
#if 0
  // if we're using freestream preservation correction, need
  // to add in the additional source term for advection.
  // the source term looks like -s*div(advVel).  div(advVel)
  // is the same as div(grad(eLambda)).  however, since we've
  // got the advection velocity here, it's easiest just to compute
  // it directly from the advection velocity
  LevelData<FArrayBox> divAdvVel;
  if (m_projection.etaLambda() > 0 && s_applyFreestreamCorrection)
    {
      divAdvVel.define(levelGrids);
      
    }
#endif

  FArrayBox srcFab;
  // now trace scalars to edges at time n+1/2
  for (dit.reset(); dit.ok(); ++dit) {
    const Box& thisGrownBox = a_old_scal[dit()].box();
    FluxBox& thisEdgeScal = edgeScal[dit()];
#ifdef NEW_ADVECTION
    // note that this is now redundant
    //FluxBox& thisAdvVel = grownAdvVel[dit()];
    FluxBox& thisAdvVel = a_advVel[dit()];
#else
    FluxBox& thisAdvVel = a_advVel[dit()];
#endif

#ifdef MPI
  if (s_verbosity >= 3) {
    //pout() << "procID: " << procID() << "tracing grid " << thisGrownBox << endl;
  }
#endif


#ifdef MATCH_OLDCODE
    // this is necessary to synchronize with old-version code
    FArrayBox&  thisCellVel = local_old_vel[dit()];
#else 
    FArrayBox& thisCellVel = oldVel()[dit()];
#endif

    FArrayBox& thisOldScal = a_old_scal[dit()];

    srcFab.resize(thisGrownBox, nScal);
    srcFab.setVal(0.0);


#ifdef NEW_ADVECTION
    // new OldPatchGodunov-based approach to advection
    const Box& curBox = levelGrids[dit()];
    m_patchGod.setCurrentBox(curBox);
    // set cell-centered velocity field
    m_patchGod.setCellVelPtr(&thisCellVel);
    // set advection velocity field
    m_patchGod.setAdvVelPtr(&thisAdvVel);
    
    // set up fluxBox->FArrayBox[SpaceDim] conversion
    FArrayBox faceScalArray[SpaceDim];
    for (int dir=0; dir<SpaceDim; dir++)
      {
        // use aliasing to handle this 
        Interval scalInterval = thisEdgeScal[dir].interval();
        faceScalArray[dir].define(scalInterval, thisEdgeScal[dir]);
      }

    // now compute fluxes
    m_patchGod.computeFluxes(thisOldScal, 
                             faceScalArray, srcFab,
                             m_dt, curBox);
  
    for (int dir=0; dir<SpaceDim; dir++)
      {
        //multiply by edge velocity to get flux
        // do componentwise
        for (int comp=0; comp<nScal; comp++) {	
          thisEdgeScal[dir].mult(thisAdvVel[dir],0,comp,1);
        }
      }    
#else
    // old LevelTrace advection scheme
    for (int dir=0; dir<SpaceDim; dir++) {
      TraceState(thisEdgeScal[dir], thisOldScal,
		 thisCellVel, thisAdvVel,
		 srcFab, problemDomain(),
		 levelGrids[dit()], m_dt, m_dx, dir, 
		 scalComps, scalComps);
      
      //multiply by edge velocity to get flux
      // do componentwise
      for (int comp=0; comp<nScal; comp++) {	
	thisEdgeScal[dir].mult(thisAdvVel[dir],0,comp,1);
      }
    }
#endif
    
  } // end loop over grids

  
  // compute div(us)
#ifdef MPI
  if (s_verbosity >= 3) {
    //pout() << "procID: " << procID() << "taking divergence" << endl;
  }
#endif

  levelDivergenceMAC(a_new_scal, edgeScal, m_dx);

  // now multiply div(us) by -dt and add to old_scal
  for (dit.reset(); dit.ok(); ++dit) {
    Real mult = -1.0*a_dt;
    a_new_scal[dit()] *= mult;
    a_new_scal[dit()] += a_old_scal[dit()];
    
    
    // now do flux register stuff
    FluxBox& thisFlux = edgeScal[dit()];

#ifdef MPI
  if (s_verbosity >= 3) {
    //pout() << "procID: " << procID() << "beginning flux register stuff" << endl;
  }
#endif

    if (!finestLevel()) {
      assert(a_fineFluxReg.isDefined());
      for (int dir=0; dir<SpaceDim; dir++) {
	// since we will want the increment to be -U-del-U,
	// make this mult negative
	Real fluxMult = -a_dt;
	a_fineFluxReg.incrementCoarse(thisFlux[dir], fluxMult, dit(), 
				      scalComps, scalComps, dir);
      }
    } // end if a finer level exists
    

    if (m_level > 0) {
      assert(a_crseFluxRegPtr != NULL);
      assert(a_crseFluxRegPtr->isDefined());
      for (int dir=0; dir<SpaceDim; dir++) {
	Real fluxMult = -a_dt;
	a_crseFluxRegPtr->incrementFine(thisFlux[dir], fluxMult, dit(),
					scalComps, scalComps, dir,
					Side::Lo);
	
	a_crseFluxRegPtr->incrementFine(thisFlux[dir], fluxMult, dit(),
					scalComps, scalComps, dir,
					Side::Hi);
      }
    } // end if coarser level exists
  } // end loop over grids
				    
				  
  // that should be it!
}
		     

void
AMRNavierStokes::advectDiffuseScalar(LevelData<FArrayBox>& a_new_scal,
				     LevelData<FArrayBox>& a_old_scal,
				     LevelData<FluxBox>& a_adv_vel,
				     const Real& a_diffusive_coeff,
				     const LevelData<FArrayBox>* a_crse_scal_old,
				     const LevelData<FArrayBox>* a_crse_scal_new,
				     const Real a_old_crse_time,
				     const Real a_new_crse_time,
				     LevelFluxRegister* a_crseFluxRegPtr,
				     LevelFluxRegister& a_fineFluxReg,
				     DomainGhostBC& a_scalPhysBC,
#ifdef NEW_ADVECTION
                                     OldPhysIBC* a_scalIBC,
#endif
				     Real a_dt)
{
  
  if (s_verbosity >= 3) {
    pout() << "AMRNavierStokes::advectDiffuseScalar: " << m_level << endl;
  }
  
  int numScal = a_new_scal.nComp();
  Interval scalComps(0,numScal-1);
  Real old_time = m_time - m_dt;
  //check to see if we need diffusive stuff or not...
  bool isDiffusive = false;
  if (a_diffusive_coeff > 0) isDiffusive = true;
  

#if 1
  if (!isDiffusive) {
    // if not diffusive,  just turn around and call nondiffusive scalar advection
    advectScalar(a_new_scal, a_old_scal, a_adv_vel, a_crseFluxRegPtr,
		 a_fineFluxReg, 
#ifdef NEW_ADVECTION
                 a_scalIBC,
#endif
                 a_dt);
  } else {
#endif
    
    // otherwise, do diffusive algorithm
    
    // at the moment, can only do one component at a time for diffusion
    assert (numScal == 1);
    
    // these are parameters for TGA diffusive solves
    Real a = 2.0 - sqrt(2.0) - TGA_EPS;
    Real discr = sqrt(a*a - 4.0*a + 2.0);
    Real r1 = (2.0*a - 1.0)/(a + discr);
    Real r2 = (2.0*a - 1.0)/(a - discr);  
       
    const DisjointBoxLayout& levelGrids = a_new_scal.getBoxes();
    
    // define temporary storage
    LevelData<FArrayBox> diffusiveSrc(levelGrids,numScal,
				      IntVect::TheUnitVector());
    LevelData<FArrayBox> tempStorage(levelGrids,numScal, 
				     IntVect::TheUnitVector());
    LevelData<FArrayBox>* crseDataPtr = NULL;
    
    if (m_level > 0) {
      // allocate crseBC info
      const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
      crseDataPtr = new LevelData<FArrayBox>(crseGrids,1);
    }

    LevelData<FluxBox> diffusiveFlux(levelGrids, numScal);
    DataIterator dit = a_new_scal.dataIterator();
    
    // these are necessary because solver can only handle one
    // component at a time
    //LevelData<FArrayBox> solverRHS(levelGrids, 1, IntVect::TheUnitVector());
    LevelData<FArrayBox> intermediateScal(levelGrids, 1,
					  IntVect::TheUnitVector());
    LevelData<FluxBox> tempFlux(levelGrids,1);
    //LevelData<FArrayBox>* crseSolverDataPtr = NULL;
    
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
	diffusiveSrc[dit()].setVal(s_bogus_value);
	tempStorage[dit()].setVal(s_bogus_value);
	diffusiveFlux[dit()].setVal(s_bogus_value);
	//solverRHS[dit()].setVal(s_bogus_value);
	intermediateScal[dit()].setVal(s_bogus_value);
	tempFlux[dit()].setVal(s_bogus_value);
      }
    } // end bogus-value initialization
    
    
    // will need two QuadCFInterps -- one for all components, 
    // and one for single component
    QuadCFInterp scalarCFInterp;
    QuadCFInterp singleCompCFInterp;

    if (m_level > 0 ) {
      // only define these if necessary
      const DisjointBoxLayout& crseBoxes = crseNSPtr()->newVel().getBoxes();
      //crseSolverDataPtr = new LevelData<FArrayBox>(crseBoxes, 1);
      int nRefCrse = crseNSPtr()->refRatio();
      scalarCFInterp.define(levelGrids, &crseBoxes, m_dx, 
			    nRefCrse, 1, m_problem_domain);
    }
    

    // now compute viscous source term
    // copy scalar into temp storage so we don't overwrite
    // C/F BC's 

    // define our own copier here to prevent ghost cells 
    // being filled (avoiding a potentially expensive exchange)
    // since they will be filled a bit later when we compute the Laplacian
    TMPI_advectDiffScal_cp.start();
    Copier thisCopier(a_old_scal.getBoxes(), tempStorage.getBoxes(),
                      IntVect::TheZeroVector());
    a_old_scal.copyTo(a_old_scal.interval(), tempStorage, 
		      tempStorage.interval(), thisCopier);
    TMPI_advectDiffScal_cp.stop();
    // now set up crse level BC
    if (crseDataPtr != NULL) {
      assert (a_crse_scal_old != NULL);
      assert (a_crse_scal_new != NULL);
      // need to interpolate in time 
      timeInterp(*crseDataPtr, old_time,
		 *a_crse_scal_old, a_old_crse_time,
		 *a_crse_scal_new, a_new_crse_time,
		 scalComps);
    }
    // exchange for tempStorage (scal) not necessary since it's 
    // done in PoissonOp::ApplyOpI
    computeLapScal(diffusiveSrc,tempStorage,a_scalPhysBC,crseDataPtr);

    // exchange for diffusiveSrc not necessary because it's done
    // in computeLapScal

    // also, since quadInterp BC's were applied in laplacian,
    // can also compute viscous fluxes due to old-time data
    // and can do it w/o recomputing C/F interp
    levelGradientMAC(diffusiveFlux, tempStorage, m_dx);
  
    // loop through and apply proper scaling
    //Real fluxMult = (0.5-a)*m_dt*a_diffusive_coeff;
    Real fluxMult = (1.0-a)*a_diffusive_coeff;
    // this is an experiment
    //fluxMult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) { 
      FArrayBox& thisSrc = diffusiveSrc[dit()];
      FluxBox& thisFlux = diffusiveFlux[dit()];
      thisSrc *= a_diffusive_coeff;
      for (int dir=0; dir<SpaceDim; dir++) {
	thisFlux[dir] *= fluxMult;
      }
      
    }
    
    
    // this is pretty much identical to advectScalar()
    /// need to build grown grids to get be able to do all of
    /// tracing properly -- this is necessary to sync with old
    /// code
    IntVect advect_grow(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
    
    LevelData<FArrayBox> local_old_vel(levelGrids, SpaceDim, advect_grow);
    // m_time contains the time at which the new state is centered
    fillVelocity(local_old_vel, old_time);

    
    if (s_verbosity >= 3) {
      //pout() << "procID " << procID() << " return from fillVelocity" << endl;
    }  
    
#ifdef NEW_ADVECTION
    // set up patchGodunov for this problem
    m_patchGod.setPhysIBC(a_scalIBC);
    m_patchGod.setNumVar(numScal);
    m_patchGod.setCurrentTime(old_time);
    
    // also need to build a grown advection velocity 
    IntVect advectVelGrow(IntVect::Unit);
    LevelData<FluxBox> grownAdvVel(levelGrids, 1, advectVelGrow);
    // this is horribly inefficient, but we'll see how bad it really is
    // for now, fill extra cells with cell-to-face averaged velocities
    CellToEdge(local_old_vel, grownAdvVel);
    // now overwrite with advection velocities wherever possible
    TMPI_advectDiffScal_NA_cp.start();
    a_adv_vel.copyTo(a_adv_vel.interval(), grownAdvVel, 
                     grownAdvVel.interval());
    TMPI_advectDiffScal_NA_cp.stop();
#endif
    
#ifdef NEW_ADVECTION
    IntVect ghostVect = IntVect::Unit;
    LevelData<FluxBox> edgeScal(levelGrids, numScal, ghostVect);
#else
    LevelData<FluxBox> edgeScal(levelGrids, numScal);
#endif
    
    
    // now trace scalars to edges at time n+1/2
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisEdgeScal = edgeScal[dit()];
#ifdef NEW_ADVECTION
      FluxBox& thisAdvVel = grownAdvVel[dit()];
#else
      FluxBox& thisAdvVel = a_adv_vel[dit()];
#endif
      
#ifdef MATCH_OLDCODE
      FArrayBox& thisCellVel = local_old_vel[dit()];
#else
      FArrayBox& thisCellVel = oldVel()[dit()];
#endif
      FArrayBox& thisOldScal = a_old_scal[dit()];
      FArrayBox& thisSrc = diffusiveSrc[dit()];


#ifdef NEW_ADVECTION
      // OldPatchGodunov-based approach to advection
      const Box& curBox = levelGrids[dit()];
      m_patchGod.setCurrentBox(curBox);
      // set cell-centered velocity field
      m_patchGod.setCellVelPtr(&thisCellVel);
      // set advection velocity field
      m_patchGod.setAdvVelPtr(&thisAdvVel);

      // set up fluxBox->FArrayBox[SpaceDim] conversion
      FArrayBox faceScalArray[SpaceDim];
      for (int dir=0; dir<SpaceDim; dir++)
        {
          // use aliasing to handle this
          Interval scalInterval = thisEdgeScal[dir].interval();
          faceScalArray[dir].define(scalInterval, thisEdgeScal[dir]);
        }

      // now compute fluxes
      m_patchGod.computeFluxes(thisOldScal, faceScalArray, 
                               thisSrc, m_dt, curBox);
      
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          // multiply by edge velocity to get flux
          // do componentwise
          for (int comp=0; comp<numScal; comp++) {
            thisEdgeScal[dir].mult(thisAdvVel[dir],0,comp,1);
          }	
        } // end loop over tracing directions
#else
      // old LevelTrace advection scheme
      for (int dir=0; dir<SpaceDim; dir++) {
	TraceState(thisEdgeScal[dir], thisOldScal,
		   thisCellVel, thisAdvVel, 
		   thisSrc,problemDomain(),
		   levelGrids[dit()], m_dt, m_dx, dir,
		   scalComps, scalComps);

	// multiply by edge velocity to get flux
	// do componentwise
	for (int comp=0; comp<numScal; comp++) {
	  thisEdgeScal[dir].mult(thisAdvVel[dir],0,comp,1);
	}	
      } // end loop over tracing directions
#endif // end LevelTrace version


    } // end loop over grids for tracing

    // compute div(us)
    levelDivergenceMAC(a_new_scal, edgeScal, m_dx);

    // now add advective flux to diffusive flux 
    // (edgeSCal already contains fluxes)
    // advective flux has opposite sign of diffusive flux
    //fluxMult = m_dt;
    fluxMult = -1.0;
    // this is a debugging thing
    //fluxMult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = diffusiveFlux[dit()];
      FluxBox& thisAdvFlux = edgeScal[dit()];
      for (int dir=0; dir<SpaceDim; ++dir) {
	thisAdvFlux[dir] *= fluxMult;
	thisFlux[dir].plus(thisAdvFlux[dir]);
      }
    }
  
    
    // now begin to set up viscous solves...
    //"f" term is -u-del-s, for scalars
    for (dit.reset(); dit.ok(); ++dit) {
      a_new_scal[dit()] *= -1.0;
    }
    // now compute diffused version of this

    // want to set up extrap C/F BC's for this
    //if (m_level > 0) {
      // this BC object will contain only extrapBC's
      // and will be used at C/F interfaces
      DomainGhostBC bogusBC;
      for (int dir=0; dir<SpaceDim; dir++) {
	SideIterator sit;
	for (sit.begin(); sit.ok(); ++sit) {
	  HOExtrapBC thisBC(dir,sit(), scalComps);
	  bogusBC.setBoxGhostBC(thisBC);
	}
      }
      
      // now loop over grids
      for (dit.reset(); dit.ok(); ++dit) {
	Box thisGridBox = levelGrids[dit()];
	FArrayBox& thisData = a_new_scal[dit()];
	bogusBC.applyHomogeneousBCs(thisData,thisGridBox,m_dx);
	
	// at this point, all ghost cells are set by extrap
	// now set physical BCs on (-uDelS)
	// not sure if this is the right thing to do for this --
	// leave it at extrap for now (safe)
	//  for (int comp=0; comp<SpaceDim; comp++) {
	// a_scalPhysBCs[comp].applyHomogeneousBC(thisData,
	// m_problem_domain,
	// m_dx);
	// }
      }
      //}

      // finally, do exchange
      TMPI_advectDiffScalar_exchange.start();
      a_new_scal.exchange(scalComps);
      TMPI_advectDiffScalar_exchange.stop();
    
    // this computes L(-uDelS) and puts it in tempStorage
    TF_OPERATORLAP_advectDiffuseScalar.start();
    for (dit.reset(); dit.ok(); ++dit) {
      FORT_OLDOPERATORLAP(CHF_FRA(tempStorage[dit()]),
		                      CHF_FRA(a_new_scal[dit()]),
		                      CHF_BOX(levelGrids.get(dit())),
		                      CHF_CONST_REAL(m_dx));
    }
    TF_OPERATORLAP_advectDiffuseScalar.stop();
    
    // also need to compute flux due to antidiffusive piece
    // in this case, we've already set C/F and physical BC's
    levelGradientMAC(tempFlux,a_new_scal,m_dx);

#if 0
    // debugging check
    for (dit.reset(); dit.ok(); ++dit) {
      tempStorage[dit()].setVal(0.0);
      tempFlux[dit()].setVal(0.0);
    }
#endif

    
    // now add this to total flux
    fluxMult = (0.5-a)*m_dt*a_diffusive_coeff;
    // this is for debugging
    //fluxMult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisLocalFlux = tempFlux[dit()];
      FluxBox& thisTotalFlux = diffusiveFlux[dit()];
      // now loop over components and multiply 
      for (int dir=0; dir<SpaceDim; dir++) {
	thisLocalFlux[dir] *= fluxMult;
	
	
	// now add in flux
	thisTotalFlux[dir] += thisLocalFlux[dir];
      }
    } // end loop over grids for antidiffusive flux addition
    
    
    // src term for first solve will be (I + (half-a)*dt*nu*L)(-uDelS)
    // this will be stored for now in a_new_scal
    Real mult = (0.5-a)*m_dt;
    for (dit.reset(); dit.ok(); ++dit) {
      tempStorage[dit()] *= mult*a_diffusive_coeff;
      
      a_new_scal[dit()] += tempStorage[dit()];
    }
    
    // now loop over boxes, multiply diffusiveSrc by mult,
    // then add old-time scalar and diffusiveSrc to the RHS for
    // the first solve, which will all be in a_new_scal, and which is
    // RHS = (I + (1-a)*nu*dt*L)u_n + dt*f
    // where f = (I + (half-a)*dt*nu*L)(-uDelS)
    mult = (1.0-a)*m_dt;
    for (dit.reset(); dit.ok(); ++dit) {
      diffusiveSrc[dit()] *= mult;
      a_new_scal[dit()] *= m_dt;
      a_new_scal[dit()] += diffusiveSrc[dit()];
      a_new_scal[dit()] += a_old_scal[dit()];
    }
    
    // now set up crse level BC
    Real intermediateTime = old_time + (1-r1)*m_dt;
    //Real intermediateTime = old_time + m_dt;
    if (crseDataPtr != NULL) {
      assert (a_crse_scal_old != NULL);
      assert (a_crse_scal_new != NULL);
      // need to interpolate in time 
      timeInterp(*crseDataPtr, intermediateTime,
		 *a_crse_scal_old, a_old_crse_time,
		 *a_crse_scal_new, a_new_crse_time,
		 scalComps);
    }
    
    
    // now that we've got everything, perform diffusion solves
    
    // copy RHS and coarse BC into single-component holder
    LevelData<FArrayBox>& solverRHS = a_new_scal;
    for (dit.reset(); dit.ok(); ++dit) {
      //solverRHS[dit()].copy(a_new_scal[dit()],comp,0,1);
      // initalize soln to be RHS
      intermediateScal[dit()].copy(solverRHS[dit()]);
    }
    
    // can use velocity viscous solver here as well
    mult = -r2*m_dt*a_diffusive_coeff;
    m_viscous_solver.setBeta(mult);
    
    m_viscous_solver.setBC(a_scalPhysBC);
    
    TimeLS_advectScalar1.start();
    IterationFlavor = COUNTER_levelSolve_viscous1 + m_level;
    //pout() << " advectScalar1 levelSolve Level" << m_level << endl;
    m_viscous_solver.levelSolve(intermediateScal,crseDataPtr,
				solverRHS, false);
    TimeLS_advectScalar1.stop();

    // increment fluxes with flux due to intermediate soln.
    // set physical BC's 
    for (dit.reset(); dit.ok(); ++dit) {
      a_scalPhysBC.applyInhomogeneousBCs(intermediateScal[dit()],
					 m_problem_domain, m_dx);
    }
     
    levelGradientMAC(tempFlux, intermediateScal, crseDataPtr,
		     m_dx, scalarCFInterp);

    mult = -r2*a_diffusive_coeff;
    // this is for debugging
    //mult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = tempFlux[dit()];
      FluxBox& thisTotalFlux = diffusiveFlux[dit()];
      for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
	thisFlux[fluxDir].mult(mult);
	thisTotalFlux[fluxDir].minus(thisFlux[fluxDir],0,0,1);
      }
    }
    
    // now ready to do second solve.
    mult = -r1*m_dt*a_diffusive_coeff;
    m_viscous_solver.setBeta(mult);

    
    // now set up crse level BC
    Real new_time = m_time;
    if (crseDataPtr != NULL) {
      assert (a_crse_scal_old != NULL);
      assert (a_crse_scal_new != NULL);
      // need to interpolate in time 
      timeInterp(*crseDataPtr, new_time,
		 *a_crse_scal_old, a_old_crse_time,
		 *a_crse_scal_new, a_new_crse_time,
		 scalComps);
    }
    // initialize initial guess
    for (dit.reset(); dit.ok(); ++dit) {
      a_new_scal[dit()].copy(intermediateScal[dit()]);
    }

    // and solve away...
    // (can just use new-time boundary condition)
    TimeLS_advectScalar2.start();
    IterationFlavor = COUNTER_levelSolve_viscous2 + m_level;
    //pout() << " advectScalar2 levelSolve Level" << m_level << endl;
    m_viscous_solver.levelSolve(a_new_scal, crseDataPtr,
				intermediateScal, false);
    TimeLS_advectScalar2.stop();

    // reset physical BC's
    for (dit.reset(); dit.ok(); ++dit) {
      a_scalPhysBC.applyInhomogeneousBCs(a_new_scal[dit()],
					m_problem_domain, m_dx);
    }

    // compute flux due to final piece
    levelGradientMAC(tempFlux, a_new_scal, crseDataPtr,
		     m_dx, scalarCFInterp);

    mult = -r1*a_diffusive_coeff;
    // this is for debugging
    //mult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = tempFlux[dit()];
      FluxBox& thisTotalFlux = diffusiveFlux[dit()];
      for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
	thisFlux[fluxDir].mult(mult);
	thisTotalFlux[fluxDir].minus(thisFlux[fluxDir],0,0,1);
      }
    }
				
    // now actually increment flux registers with entire flux
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& localFlux = diffusiveFlux[dit()];
      
      if (m_level > 0) {
	LevelFluxRegister& crseFR = *a_crseFluxRegPtr;
	assert (crseFR.isDefined());
	// use same multipliers and sign convention as in computeUStar
	Real fluxMult = m_dt;
	for (int dir=0; dir<SpaceDim; dir++) {
	  crseFR.incrementFine(localFlux[dir],
			       fluxMult, dit(), scalComps,
			       scalComps, dir, Side::Lo);
	  
	  crseFR.incrementFine(localFlux[dir],
			       fluxMult, dit(), scalComps,
			       scalComps, dir, Side::Hi);
	} // end loop over directions
      } // end if coarser level exists

      if (!finestLevel()) {
	assert(a_fineFluxReg.isDefined());
	Real fluxMult = m_dt;
	for (int dir=0; dir<SpaceDim; ++dir) {
	  a_fineFluxReg.incrementCoarse(localFlux[dir],
					fluxMult, dit(),
					scalComps, scalComps,
					dir);
	} // end loop over directions
      } // end if finer level exists
    } // end loop over grids for flux register updates

    // clean up storage, if necessary
    if (m_level > 0) {
      if (crseDataPtr != NULL) {
	delete crseDataPtr;
      }
    }

#if 1
  } // end if is diffusive
#endif

}
