/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include "AMRNavierStokesHeaders.H"

#include "Timer.H"
extern Timer Everything;
Timer TMPI_computeUStar_cp          ("compUstar cp",  Everything, DTIMECOMM, true);
Timer TMPI_computeUStar_exchange    ("compUstar ex",  Everything, DTIMECOMM, true);

//#include "memusage.H"
//char crap[40];

void 
AMRNavierStokes::predictVelocities(LevelData<FArrayBox>& a_uDelU,
                                   LevelData<FluxBox>& a_advVel)
{
  if (s_verbosity >= 3) {
    pout () << "AMRNavierStokes::predictVelocities: " << m_level << endl;
  }

  bool isViscous = (s_nu > 0.0);

  const DisjointBoxLayout& levelGrids = a_advVel.getBoxes();
  
  ////sprintf(crap, "predictVel0 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;

  // for tracing, will need to fill in boundary values for 
  // grown copy of velocity
  IntVect ghostVect(D_DECL(ADVECT_GROW, ADVECT_GROW, ADVECT_GROW));
  LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim, ghostVect);

  if (s_set_bogus_values) {
    DataIterator dit = viscousSource.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].setVal(s_bogus_value);
    }
  }
  
  LevelData<FArrayBox> traceVel     (levelGrids, SpaceDim, ghostVect);
  if (s_set_bogus_values) {
    DataIterator dit = traceVel.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      traceVel[dit()].setVal(s_bogus_value);
    }
  }

  // now fill in velocity field
  // m_time is new time
  Real old_time = m_time - m_dt;
  ////sprintf(crap, "predictVel2 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
  fillVelocity(traceVel, old_time);
  ////sprintf(crap, "predictVel3 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;

  // do physical boundary conditions on traceVel
  basicVelBC* velBCPtr = NULL;

  if (isViscous) {    
    // rather than resetting BC's on old_vel, and then setting them 
    // back, just copy old_vel to a temporary holder, and set 
    // BCs on that one.
    LevelData<FArrayBox> viscousVel(levelGrids, SpaceDim, ghostVect);
    ////sprintf(crap, "predictVel4 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
    DataIterator dit = viscousVel.dataIterator();
    for (dit.begin(); dit.ok(); ++dit)
      {
        viscousVel[dit()].copy(traceVel[dit()]);
      }

    ////sprintf(crap, "predictVel5 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
    
    
    velBCPtr = m_physBCPtr->viscousVelBC();
    velBCPtr->applyBCs(viscousVel,m_problem_domain,m_dx);
    delete velBCPtr;
    
    // if crse level exists, fill coarse velocity BC for laplacian
    LevelData<FArrayBox>* crseVelPtr = NULL;
    if (m_level > 0) {
      const DisjointBoxLayout& crseGrids = crseNSPtr()->newVel().getBoxes();
      crseVelPtr = new LevelData<FArrayBox>(crseGrids, SpaceDim);
      crseNSPtr()->fillVelocity(*crseVelPtr, old_time);
      ////sprintf(crap, "predictVel8 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
    }
    
    computeLapVel(viscousSource, viscousVel, crseVelPtr);
    ////sprintf(crap, "predictVel9 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
    
    if (crseVelPtr != NULL) {
      delete crseVelPtr;
      crseVelPtr = NULL;
    }

    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].mult(s_nu);
    }
    
    ////sprintf(crap, "predictVel10 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
  } else {
    // inviscid case
    DataIterator dit = viscousSource.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()].setVal(0.0);
    }
  }

  ////sprintf(crap, "predictVel11 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
  // tracing uses inviscid BC's
  velBCPtr = m_physBCPtr->tracingVelBC();
  velBCPtr->applyBCs(traceVel, m_problem_domain, m_dx);
  ////sprintf(crap, "predictVel12 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
  delete velBCPtr;
  
  //---------------------------------------------------------------------------  
  ////sprintf(crap, "predictVel13 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
  
#ifdef NEW_ADVECTION
  // set up patchGodunov for this problem
  // first get boundary condition object 
  OldPhysIBC* velIBCPtr = m_physBCPtr->advectionVelIBC();
  m_patchGod.setPhysIBC(velIBCPtr);
  
  // do all component directions at the same time
  m_patchGod.setNumVar(SpaceDim);
  m_patchGod.setCurrentTime(old_time);
#endif // end new advection part


  // will need edge-centered storage for all velocity components
  LevelData<FluxBox> uHalf(levelGrids, SpaceDim);

  { //allow this fluxbox to go out of scope
    // will also need grad(eLambda) and grad(phi_mac)
    LevelData<FluxBox>& grad_eLambda = m_projection.grad_eLambda();
    LevelData<FluxBox> gradPhi(levelGrids, SpaceDim);
    m_projection.gradPhi(gradPhi);
    
    if (s_set_bogus_values) {
      DataIterator dit = uHalf.dataIterator();
      for (dit.reset(); dit.ok(); ++dit) {
        uHalf[dit()].setVal(s_bogus_value);
      }
    }

#ifdef NEW_ADVECTION
    // temporary storage
    FArrayBox tempFlux[SpaceDim];

    // loop over grids and do prediction
    DataIterator dit = levelGrids.dataIterator();
    for (dit.begin(); dit.ok(); ++dit)
      {
        Box gridBox = levelGrids[dit()];
        FArrayBox& thisOldVel = traceVel[dit()];
        FluxBox& thisAdvVel = a_advVel[dit()];
        FArrayBox& srcFab = viscousSource[dit()];
        
        // set up patchGodunov for this patch
        m_patchGod.setCurrentBox(gridBox);
        m_patchGod.setCellVelPtr(&thisOldVel);
        m_patchGod.setAdvVelPtr(&thisAdvVel);
        
        // allocate temp storage
        for (int dir=0; dir<SpaceDim; dir++)
          {
            Box fluxBox = gridBox;
            fluxBox.grow(1);
            fluxBox.grow(dir,-1);
            fluxBox &= m_problem_domain;
            fluxBox.surroundingNodes(dir);
            tempFlux[dir].resize(fluxBox, SpaceDim);
          }

        // compute face-centered variables using "Godunov Box"
        m_patchGod.computeFluxes(thisOldVel, tempFlux, 
                                 srcFab, m_dt, gridBox);
        
        // now loop over directions -- normal direction velocities 
        // are copied from advection velocities; for tangential direction 
        // we use face-centered predicted velocities (flux divided by 
        // advection velocity)
        FluxBox& thisPredictedVel = uHalf[dit()];
        for (int velComp = 0; velComp<SpaceDim; velComp++) 
          {
            for (int dir=0; dir<SpaceDim; dir++)
              {
                FArrayBox& thisVelDir = thisPredictedVel[dir];
                FArrayBox& thisAdvVelDir = thisAdvVel[dir];
                    
                if (dir == velComp)
                  {
                    // normal direction -- copy from advVel->uHalf
                    // srcComp is always 0, since advVel only has 
                    // one component
                    int srcComp=0; 
                    int destComp = velComp;
                    
                    thisVelDir.copy(thisAdvVelDir, srcComp, destComp, 1);

                    if (s_applyFreestreamCorrection) 
                      {
                        // now need to subtract off grad(eLambda)
                        // this is the due to the difference between 
                        // _advecting_ and _advected_ velocities              
                        FluxBox& thisCorr = grad_eLambda[dit()];              
                        FArrayBox& thisCorrDir = thisCorr[dir];              
                        thisVelDir.minus(thisCorrDir, 0, dir, 1);
                      }                  
                  } else {
                    // if not a normal direction, then copy predicted
                    // velocity

                    int srcComp = velComp; 
                    int destComp = velComp;
                    
                    FArrayBox& thisFlux = tempFlux[dir];
                    
                    thisVelDir.copy(thisFlux, srcComp, destComp, 1);
                    
                    // need to add MAC correction to traced velocities
                    thisVelDir.minus(gradPhi[dit()][dir], velComp, 
                                     velComp, 1);
                    
                  } // end if tangential direction
              } // end loop over face directions
          } // end loop over velocity components            
      } // end loop over grids
    
#else
    // loop over velocity components -- old LevelTrace-based approach
    for (int velComp = 0; velComp < SpaceDim; velComp++) {
      // need to compute predicted edge-centered values for each
      // velocity component in each coordinate direction. begin
      // loop over coordinate directions 
      for (int dir=0; dir<SpaceDim; dir++) {
        // if this is a normal edge, we already have edge-centered
        // predicted values, which are divergence-free (the advection
        // velocities).  treat this case differently
        if (dir == velComp) {
          DataIterator dit = uHalf.dataIterator();
          for (dit.reset(); dit.ok(); ++dit) {
            
            // copy advVel -> uHalf
            // srcComp is always 0, since advVel only has one component
            int  srcComp = 0;
            int  destComp = velComp;
            //a_advVel.copyTo(srcComp, uHalf, destComp);
            FluxBox& thisVel = uHalf[dit()];
            FArrayBox& thisVelDir = thisVel[dir];
            
            FluxBox& thisAdvVel = a_advVel[dit()];
            FArrayBox& thisAdvVelDir = thisAdvVel[dir];
            
            thisVelDir.copy(thisAdvVelDir,srcComp, destComp, 1);
            
            if (s_applyFreestreamCorrection) 
              {
                // now need to subtract off grad(eLambda)
                // this is the due to the difference between _advecting_
                // and _advected_ velocities              
                FluxBox& thisCorr = grad_eLambda[dit()];              
                FArrayBox& thisCorrDir = thisCorr[dir];              
                thisVelDir.minus(thisCorrDir, 0, dir, 1);
              }
          }
        } else {
          // if not a normal direction, need to perform tracing
          Interval srcComps(velComp, velComp);
          Interval destComps(velComp, velComp);
          
          DataIterator dit = uHalf.dataIterator();
          for (dit.reset(); dit.ok(); ++dit) {
            // const Box& gridBox = levelGrids[dit()];
            //const Box& grownBox = traceVel[dit()].box();
            FArrayBox& srcFab = viscousSource[dit()];
            
            FluxBox& thisUhalf = uHalf[dit()];
            FArrayBox& thisUHalfDir = thisUhalf[dir];
            
            FluxBox& thisAdvVel = a_advVel[dit()];
            FArrayBox& thisOldVel = traceVel[dit()];
            
            TraceState(thisUHalfDir, thisOldVel, thisOldVel, 
                       thisAdvVel, srcFab, problemDomain(),
                       levelGrids[dit()], m_dt, m_dx, dir,
                       srcComps, destComps);
            
            
            // need to add MAC correction to traced velocities
            thisUHalfDir.minus(gradPhi[dit()][dir], velComp, 
                               velComp, 1);
            
          }
          
          
        }
      } // end loop over tracing directions
    } // end loop over velocity components
#endif // end levelTrace-based approach
  
  } // FluxBox goes out of scope, memory reclaimed.
  
  // now loop over grids to compute advection and update flux registers
  DataIterator dit = a_uDelU.dataIterator();
  
  // for nonconservative update, also need a cell-centered
  // advection velocity -- get this by averaging edge-centered
  // advection velocity to cell centers.  this one shouldn't need
  // any ghost cells.

  ////sprintf(crap, "predictVel14 level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;

  // reduce, reuse, recycle...
  //LevelData<FArrayBox> cellAdvVel(levelGrids, SpaceDim);
  LevelData<FArrayBox>&  cellAdvVel = viscousSource;

  EdgeToCell(a_advVel, cellAdvVel);

  for (dit.reset(); dit.ok(); ++dit) {
    FluxBox& thisUHalf = uHalf[dit()];
    FArrayBox& thisCellAdvVel = cellAdvVel[dit()];
    const Box& thisBox = levelGrids[dit()];
    FArrayBox& this_uDelU = a_uDelU[dit()];
    this_uDelU.setVal(0.0);
    
    // to do this in a dimensionality independent way, 
    // loop over directions.
    TF_UDELS_predictVelocities.start();
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& uHalfDir = thisUHalf[dir];
     
      // this will increment this_uDelU with
      // cellVel*d(uHalfDir)/d[xyz] for all velocity components
      FORT_UDELS(CHF_FRA(this_uDelU),
                 CHF_FRA(uHalfDir),
                 CHF_FRA(thisCellAdvVel),
                 CHF_BOX(thisBox),
                 CHF_REAL(m_dx),
                 CHF_INT(dir));
    }
    TF_UDELS_predictVelocities.stop();

    // now need to do flux register stuff
    // first compute fluxes (u_edge*u_adv)
    for (int dir=0; dir<SpaceDim; dir++) {
      FArrayBox& this_uEdge = uHalf[dit()][dir];
      FArrayBox& thisAdvVel = a_advVel[dit()][dir];

      // now need to loop over traced velocity components
      for (int velComp=0; velComp<SpaceDim; ++velComp) {
        this_uEdge.mult(thisAdvVel,0,velComp,1);
      }
    }

    Interval velComps(0,SpaceDim-1);
    if (!finestLevel()) {
      assert (m_flux_register.isDefined());
      for (int dir=0; dir < SpaceDim; dir++) {
        // because we eventually want -D_R(U-del-U), signs
        // on fluxMult are opposite of what you might expect...
        Real fluxMult = -m_dt;
        // debugging change
        //fluxMult = 0.0;
        if (s_reflux_normal_momentum)
          {
            m_flux_register.incrementCoarse(uHalf[dit()][dir],
                                            fluxMult, dit(), velComps,
                                            velComps, dir);
          }
        else
          {
            // if we're not refluxing normal momentum component, 
            // then do this component by component
            for (int velComp=0; velComp<SpaceDim; velComp++) 
              {
                if (velComp != dir)
                  {
                    Interval velInt(velComp, velComp);
                    m_flux_register.incrementCoarse(uHalf[dit()][dir],
                                                    fluxMult, dit(), 
                                                    velInt,
                                                    velInt, dir);
                  } // end if a tangential velocity component
              } // end loop over velocity components
          } // end if we're not refluxing normal momentum
                      
      } // end loop over directions
    } // end if a finer level exists
    
    if (m_level > 0) {
      LevelFluxRegister& crseFR = crseNSPtr()->m_flux_register;
      assert(crseFR.isDefined());
      // because we will eventually want -D_R(U-del-U) signs
      // on fluxMult are opposite of what you might expect...      
      Real fluxMult = -m_dt;
      // debugging change
      //fluxMult = 0.0;
      for (int dir=0; dir<SpaceDim; dir++) {
        if (s_reflux_normal_momentum)
          {
            crseFR.incrementFine(uHalf[dit()][dir],
                                 fluxMult, dit(), velComps,
                                 velComps, dir, Side::Lo);
            
            crseFR.incrementFine(uHalf[dit()][dir],
                                 fluxMult, dit(), velComps,
                                 velComps, dir, Side::Hi);
          }
        else
          {
            // for case where we only want to reflux tangential
            // momentum components, do this component by component
            // so that we can isolate the normal component.
            for (int velComp = 0; velComp<SpaceDim; velComp++)
              {
                if (velComp != dir)
                  {
                    Interval velInt(velComp, velComp);
                    
                    crseFR.incrementFine(uHalf[dit()][dir],
                                         fluxMult, dit(), velInt,
                                         velInt, dir, Side::Lo);
                    
                    crseFR.incrementFine(uHalf[dit()][dir],
                                         fluxMult, dit(), velInt,
                                         velInt, dir, Side::Hi);
                  } // end if this is a tangential component
              } // end loop over velocity components
          } // end if we're not refluxing normal momentum
      
      }
    }


  } // end loop over grids 


  // at this point, we've computed the advective term, and updated
  // the flux registers with the momentum flux 

  ////sprintf(crap, "predictVelEND level%d mem=%-10.3f\n", m_level, get_memory_usage()); pout() << crap;
}

// the approximation to u-dot-del-u is passed into this
// function in a_uStar -- at end of function, a_uStar
// will contain approximation to new velocity:
// uStar + dt*gradPi
// noel, don't forget about computeUStarInit!
void
AMRNavierStokes::computeUStar(LevelData<FArrayBox>& a_uStar)
{

  ////sprintf(crap, "compUstar0 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
  bool isViscous = (s_nu > 0.0);
  Interval velComps(0,SpaceDim-1);

  if (!isViscous) {
    /// compute uStar = oldVel - dt*(u-dot-del-u)
    TimeUStar_notViscous.start();
    DataIterator dit = a_uStar.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      Real mult = -1.0*m_dt;
      a_uStar[dit()] *= mult;
      a_uStar[dit()] += oldVel()[dit()];
    }
    TimeUStar_notViscous.stop();
  } else { 
    // viscous case is more elaborate -- need to do viscous solves.

    const DisjointBoxLayout& levelGrids = a_uStar.getBoxes();

    // these are parameters for TGA viscous solves
    Real a = 2.0 - sqrt(2.0) - TGA_EPS;
    Real discr = sqrt(a*a - 4.0*a + 2.0);
    Real r1 = (2.0*a - 1.0)/(a + discr);
    Real r2 = (2.0*a - 1.0)/(a - discr);  

    // set convergence metric to be norm of velocity field
    TimeUStarMetricNorm.start();
    Interval allVelComps(0,SpaceDim-1);
    int normType = 0;
    Real velNorm = norm(oldVel(), allVelComps, normType);
    m_viscous_solver.setConvergenceMetric(velNorm, 0);
    TimeUStarMetricNorm.stop();
    TimeUStarDefines.start();
    
    // m_time is new time
    Real old_time = m_time - m_dt;

    // define temporary storage
    ////sprintf(crap, "compUstar1 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    LevelData<FArrayBox> gradPi(levelGrids, SpaceDim);
    LevelData<FArrayBox>* crseDataPtr = NULL;

    //LevelData<FluxBox> viscousFlux(levelGrids, SpaceDim);
    DataIterator dit = a_uStar.dataIterator();

    // these are necessary because solver can only handle one component
    // at a time.
    LevelData<FArrayBox> solverRHS(levelGrids, 1, 
                                   IntVect::TheUnitVector());
    LevelData<FArrayBox> intermediateVel(levelGrids,1,
                                         IntVect::TheUnitVector());

    LevelData<FArrayBox>* crseSolverDataPtr = NULL;
    ////sprintf(crap, "compUstar5 mem=%-10.3f\n", get_memory_usage()); pout() << crap;

    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        //viscousSource[dit()].setVal(s_bogus_value);
        //tempStorage[dit()].setVal(s_bogus_value);
        solverRHS[dit()].setVal(s_bogus_value);
        intermediateVel[dit()].setVal(s_bogus_value);
        for (int dir=0; dir<SpaceDim; ++dir) {
          //viscousFlux[dit()][dir].setVal(s_bogus_value);
          //tempFlux[dit()][dir].setVal(s_bogus_value);
        }
      }
    }   

    // will need single-component QuadCFInterp to do fluxes,
    // and Laplacians
    QuadCFInterp singleCompCFInterp;

    if (m_level > 0) {
      const DisjointBoxLayout& crseBoxes = crseNSPtr()->newVel().getBoxes();
      crseDataPtr = new LevelData<FArrayBox>(crseBoxes, SpaceDim);
      crseSolverDataPtr = new LevelData<FArrayBox>(crseBoxes, 1);
      int nRefCrse = crseNSPtr()->refRatio();
      singleCompCFInterp.define(levelGrids, &crseBoxes, m_dx, 
                                nRefCrse, 1, m_problem_domain);
    }

    //Real viscousBeta; //not used?  (ndk)
    if (s_viscous_solver_type == TGA) {
      //viscousBeta = -r2*s_nu*m_dt; //not used? (ndk)
    } else {
      MayDay::Error("undefined viscous solver type");
    }

    // first get gradPi
    m_projection.gradPi(gradPi);
    TimeUStarDefines.stop();

    TimeUStarComputeLap.start();
    // "f" is -u-del-u -gradPi +source
    for (dit.reset(); dit.ok(); ++dit) {
      a_uStar[dit()] += gradPi[dit()];
      a_uStar[dit()] *= -1.0;
    }

    LevelData<FArrayBox> tempStorage(levelGrids, SpaceDim,
                                     IntVect::TheUnitVector());

    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        tempStorage[dit()].setVal(s_bogus_value); 
      }
    }   

    // now compute diffused version of this
    if (crseDataPtr != NULL) {
      DataIterator crseDit = crseDataPtr->dataIterator();
      for (crseDit.reset(); crseDit.ok(); ++crseDit) {
        ((*crseDataPtr)[crseDit()]).setVal(0.0);
      }
    }

#ifdef MATCH_OLDCODE
    // this is a kluge to make new code agree with old code
    // C/F BC's are different -- here we can force them
    // to be extrap by being clever
    if (m_level > 0) {
      // this BC object will contain only extrapBC's 
      // and will be used at C/F interfaces
      DomainGhostBC bogusBC;
      // loop over directions
      for (int dir=0; dir<SpaceDim; dir++) {
        // loop over hi-lo
        SideIterator sit;
        for (sit.begin(); sit.ok(); ++sit) {
          HOExtrapBC thisBC(dir,sit(), velComps);
          bogusBC.setBoxGhostBC(thisBC);
        }
      }
    
      // now loop over grids
      for (dit.reset(); dit.ok(); ++dit) {
        Box thisGridBox = levelGrids[dit()];
        FArrayBox& thisData = a_uStar[dit()];
        bogusBC.applyHomogeneousBCs(thisData,thisGridBox,m_dx);
      }
      
    } // at this point, ALL ghost cells are set by extrap       
    ////sprintf(crap, "compUstar6 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    
    // now set physical BC's on (-uDelU-GradPi)
    bool isViscous = (s_nu > 0.0);
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC(isViscous);

#ifdef MATCH_OLDCODE
    //velBC.isViscous(false);
#endif

    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);    
    TMPI_computeUStar_exchange.start();
    a_uStar.exchange(velComps);
    TMPI_computeUStar_exchange.stop();
    delete velBCPtr;

    // this computes L(-uDelU-GradPi) and puts it in tempStorage

    TF_OPERATORLAP_ustar.start();
    for (dit.reset(); dit.ok(); ++dit) {
      FORT_OLDOPERATORLAP(CHF_FRA(tempStorage[dit()]),
                          CHF_FRA(a_uStar[dit()]),
                          CHF_BOX(levelGrids.get(dit())),
                          CHF_CONST_REAL(m_dx));
    }
    TF_OPERATORLAP_ustar.stop();

#if 0
    {
      // put these in their own localized scope
      viscousBC viscBC;
      viscBC.applyBCs(tempStorage,m_problem_domain, m_dx);
      TMPI_computeUStar_exchange.start();
      tempStorage.exchange(tempStorage.interval());
      TMPI_computeUStar_exchange.stop();
    }
#endif

    LevelData<FluxBox> viscousFlux(levelGrids, SpaceDim);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        for (int dir=0; dir<SpaceDim; ++dir) {
          viscousFlux[dit()][dir].setVal(s_bogus_value); 
        }
      }
    }   

    ////sprintf(crap, "compUstar7 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    // before we add in stuff, compute flux due to antidiffusive part
    // in this case, use the extrap C/F BC we've already set.
    // (this version of the levelGradientMac fn doesn't do C/F BC's)
    levelGradientMAC(viscousFlux, a_uStar, m_dx);
    ////sprintf(crap, "compUstar8 mem=%-10.3f\n", get_memory_usage()); pout() << crap;

#else    

    // set physical BC's on (-uDelU-GradPi)
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC(isViscous);
    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);
    a_uStar.exchange(velComps);
    delete velBCPtr;

    ////sprintf(crap, "compUstar9 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    // this computes L(-uDelU-GradPi) and puts it in tempStorage
    computeLapVel(tempStorage, a_uStar, crseDataPtr);

    // before we add in stuff, compute flux due to antidiffusive part
    levelGradientMAC(viscousFlux, a_uStar, crseDataPtr, m_dx, 
                     m_velCFInterp);
    ////sprintf(crap, "compUstar10 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
#endif
    TimeUStarComputeLap.stop();

    TimeUStarCompSource.start();
    // multiply grad(forcing term) by mult, then initialize
    // and increment viscousFlux (will increment flux registers
    // all at once)
    // sign convention will be so that flux will have the opposite
    // sign when incrementing flux registers as the fluxes
    // computed in earlier case.  (same sign convention as in 
    // old code).

    Real mult = (0.5-a)*m_dt*s_nu;
    // this is experimental, for debugging purposes
    //mult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = viscousFlux[dit()];
      for (int dir=0; dir<SpaceDim; dir++) {
        thisFlux[dir] *= mult;
      }
    }
        
    // source term we will use is (I + (half-a)*dt*nu*L)(-uDelU-gradPi)
    // this will be stored in a_uStar
    for (dit.reset(); dit.ok(); ++dit) {
      tempStorage[dit()] *= mult;
      a_uStar[dit()] += tempStorage[dit()];
    }
    
    // now need to get old-time velocity
    ////sprintf(crap, "compUstar11 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    fillVelocity(tempStorage, old_time);
    if (crseDataPtr != NULL) {
      crseNSPtr()->fillVelocity(*crseDataPtr, old_time);
    }
    ////sprintf(crap, "compUstar12 mem=%-10.3f\n", get_memory_usage()); pout() << crap;

    velBCPtr = m_physBCPtr->viscousVelBC();
    velBCPtr->applyBCs(tempStorage, m_problem_domain, m_dx);
    delete velBCPtr;

    mult = -(1.0-a)*s_nu;
    // this is experimental, for debugging
    //mult = 0.0;

    // compute viscous flux due to old-time velocity
    {
      // need a second multicomponent LevelData<FluxBox>
      // to hold these fluxes -- goes out of scope and 
      // is deleted when we're done
      LevelData<FluxBox> oldViscFlux(levelGrids, SpaceDim);
      levelGradientMAC(oldViscFlux, tempStorage, crseDataPtr, 
                       m_dx, m_velCFInterp);
            
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = oldViscFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int dir = 0; dir<SpaceDim; dir++) {
          thisFlux[dir].mult(mult);
          thisViscFlux[dir] -= thisFlux[dir];
        }
      }
    } // temporary flux container goes out of scope here.
    
    LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        viscousSource[dit()].setVal(s_bogus_value);
      }
    }   

    // compute L(old_velocity)
    computeLapVel(viscousSource, tempStorage, crseDataPtr);   
    ////sprintf(crap, "compUstar13 mem=%-10.3f\n", get_memory_usage()); pout() << crap;

    // now loop over boxes, multiply viscousSource by mult,
    // then add old-time velocity and viscousSource to 
    // the RHS for the first solve, which will all be in
    // a_uStar, and which is 
    // RHS = (I + (1-a)*nu*dt*L)u_n +dt*f,
    // where f = (I +(half-a)*dt*nu*L)(-uDelU - gradPi)

    mult = (1.0-a)*m_dt*s_nu;
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()] *= mult;
      a_uStar[dit()] *= m_dt;
      
      a_uStar[dit()] += viscousSource[dit()];
      a_uStar[dit()] += tempStorage[dit()];
    }

    // set up coarse-level BC
    Real intermediateTime = old_time + (1-r1)*m_dt;
    if (crseDataPtr != NULL) {
      ////sprintf(crap, "compUstar14 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
      crseNSPtr()->fillVelocity(*crseDataPtr, intermediateTime);
      ////sprintf(crap, "compUstar15 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    }


    // finally ready to do intermediate solve!
    // actually, don't need this one
    //a_uStar.exchange(velComps);

    LevelData<FluxBox> tempFlux(levelGrids, 1);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        for (int dir=0; dir<SpaceDim; ++dir) {
          tempFlux[dit()][dir].setVal(s_bogus_value);
        }
      }
    }
    TimeUStarCompSource.stop();

    // loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {

      TimeUStarSetupForSolves.start();
      // copy RHS and coarse BC into single-component holder
      LevelData<FArrayBox>& crseBC = *crseSolverDataPtr;
      LevelData<FArrayBox>& crseVel = *crseDataPtr;
      for (dit.reset(); dit.ok(); ++dit) {
        solverRHS[dit()].copy(a_uStar[dit()],dir,0,1);
        // initialize soln to be RHS
        intermediateVel[dit()].copy(a_uStar[dit()],dir,0,1);
      }
      if (crseDataPtr != NULL) {
        DataIterator crseDit = crseVel.dataIterator();
        for (crseDit.reset(); crseDit.ok(); ++crseDit) {
          // copy coarse vel BC into single-component holder
          crseBC[crseDit()].copy(crseVel[crseDit()],dir,0,1);
        }
      }
      
      mult = -r2*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);            

      DomainGhostBC* thisVelBCPtr = m_physBCPtr->viscousSolveBC(dir);
      m_viscous_solver.setBC(*thisVelBCPtr);

      TimeUStarSetupForSolves.stop();

      TimeLS_compUstar1.start();
      TimeUStarlevelSolve1.start();    
      IterationFlavor = COUNTER_levelSolve_ustar1 + m_level;
      //pout() << " UStarlevelSolve1 levelSolve Level" << m_level << endl;
      m_viscous_solver.levelSolve(intermediateVel, crseSolverDataPtr,
                                  solverRHS, false);    
      TimeUStarlevelSolve1.stop();
      TimeLS_compUstar1.stop();

      TimeUStarSetupForSolves.start();

      // increment fluxes with flux due to intermediate velocity
      // set physical BC's -- this also relies on the fact that
      // we use the same velocity components for all directions
      // (however, need to use the x-direction DomainGhostBC
      // because of the single-component nature of intermediateVel
      
      //const DomainGhostBC& thisVelBC = velBC.DomGhostBC(dir);
      for (dit.reset(); dit.ok(); ++dit) {
        thisVelBCPtr->applyHomogeneousBCs(intermediateVel[dit()], 
                                          m_problem_domain, m_dx);
      }
    
      levelGradientMAC(tempFlux, intermediateVel, crseSolverDataPtr, 
                       m_dx, singleCompCFInterp);
      
      mult = -r2*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = tempFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
          thisFlux[fluxDir].mult(mult);
          thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
        }
      }
      
      // now ready to do second solve
      mult = -r1*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);
      
      
      // fill coarse-level boundary condition...
      if (crseDataPtr != NULL) {
        ////sprintf(crap, "compUstar16 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
        crseNSPtr()->fillVelocity(*crseSolverDataPtr, m_time,
                                  dir, 0, 1);
        ////sprintf(crap, "compUstar17 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
      }
      
      // initialize initial guess 
      for (dit.reset(); dit.ok(); ++dit) {
        solverRHS[dit()].copy(intermediateVel[dit()]);
      }
      TimeUStarSetupForSolves.stop();

      // and solve away
      TimeLS_compUstar2.start();
      TimeUStarlevelSolve2.start();
      IterationFlavor = COUNTER_levelSolve_ustar2 + m_level;
      //pout() << " UStarlevelSolve2 levelSolve Level" << m_level << endl;
      m_viscous_solver.levelSolve(solverRHS, crseSolverDataPtr, intermediateVel,
                                  false);      
      TimeUStarlevelSolve2.stop();
      TimeLS_compUstar2.stop();

      TimeUStarSetupForSolves.start();

      // set physical BC's
      for (dit.reset(); dit.ok(); ++dit) {
        thisVelBCPtr->applyHomogeneousBCs(solverRHS[dit()], 
                                          m_problem_domain, m_dx);
      }
      
      // compute flux due to uStar
      levelGradientMAC(tempFlux, solverRHS, crseSolverDataPtr, 
                       m_dx, singleCompCFInterp);
      
      mult = -r1*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = tempFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
          thisFlux[fluxDir].mult(mult);
          thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
        }

        // need to copy computed uStar into a_uStar
        a_uStar[dit()].copy(solverRHS[dit()],0,dir,1);
        
      } // end loop over grids

      delete thisVelBCPtr;
      TimeUStarSetupForSolves.stop();
    } // end loop over directions
    ////sprintf(crap, "compUstar18 mem=%-10.3f\n", get_memory_usage()); pout() << crap;

    TimeUStarConcludeSolve.start();
    // now increment flux registers
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisViscFlux = viscousFlux[dit()];
      if (!finestLevel()) {
        assert(m_flux_register.isDefined());
        for (int dir=0; dir<SpaceDim; dir++) {
          // this is to make things look the same as old code
          //thisViscFlux[dir] *= m_dx;
          // use same multipliers as momentum fluxes, but
          // opposite sign convention
          Real fluxMult = m_dt;
          m_flux_register.incrementCoarse(thisViscFlux[dir],
                                          fluxMult, dit(), velComps,
                                          velComps, dir);
        }
      }
      
      if (m_level > 0) {
        LevelFluxRegister& crseFR = crseNSPtr()->m_flux_register;
        assert(crseFR.isDefined());
        // use same multipliers as momemtum fluxes, but opposite
        // sign convention
        Real fluxMult = m_dt;
        for (int dir=0; dir<SpaceDim; dir++) {
          // this is to make things look the same as old code
          //thisViscFlux[dir] *= m_dx;

          crseFR.incrementFine(thisViscFlux[dir],
                               fluxMult, dit(), velComps, 
                               velComps, dir, Side::Lo);
          
          crseFR.incrementFine(thisViscFlux[dir],
                               fluxMult, dit(), velComps, 
                               velComps, dir, Side::Hi);
        }
      }
    
    } // end loop over grids for flux register updates

    // final thing to do -- subtract off dt*gradPi term
    for (dit.reset(); dit.ok(); ++dit) {
      gradPi[dit()] *= m_dt;
      a_uStar[dit()] += gradPi[dit()];
    }
  
    
    // clean up temp storage
    if (crseDataPtr != NULL) {
      delete crseDataPtr;
      crseDataPtr = NULL;
    }
    
    if (crseSolverDataPtr != NULL) {
      delete crseSolverDataPtr;
      crseSolverDataPtr = NULL;
    }
    TimeUStarConcludeSolve.stop();

  } // end if viscous

 ////sprintf(crap, "compUstarEND mem=%-10.3f\n", get_memory_usage()); pout() << crap;
}
