/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// Main.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
using std::endl;
#include "BRMeshRefine.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "AMRIO.H"
#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"
#include "AMRNavierStokes.H"
#include "AMRNavierStokesFactory.H"
#include "PhysBCUtil.H"
#include "channelBC.H"
#include "ProblemDomain.H"

#include "NoelsPrettyPrint.H"
#include "whatisdefined.H"
#include "memusage.H"
#include "Timer.H"
Timer Everything        ("ns Everything", 0);
Timer NST_ReadInput     ("ns Read Input",   Everything);
Timer NST_SetupAMR      ("ns Setup AMR",    Everything);
Timer NST_Run           ("ns Run",          Everything);
Timer NST_Conclude      ("ns Conclude",     Everything);

Timer TMPI_Barrier2  ("ns MPI_Barrier2",  Everything, DTIMECOMM, true);
Timer TMPI_Barrier3  ("ns MPI_Barrier3",  Everything, DTIMECOMM, true);
Timer TMPI_readInput ("ns MPI readInput", Everything, DTIMECOMM, true);

#ifdef TIMER
extern void writeOutNoelsSooperIterationCounterStuff(void);
#endif

#ifdef COUNT_PEX
extern unsigned int pex_Nsends[NPEX];
extern double pex_delta_time[NPEX];
extern unsigned int pex_Nlocal[NPEX];
extern unsigned int pex_Nto[NPEX];
extern unsigned int pex_Nfrom[NPEX];
#endif

int global_granularity=1; //temp hack for now (noel)
int write_bottomsmoother_file=0; //temp hack for now (noel)
int num_levelGSRBs_inPC=2; //temp hack for now (noel)
int global_relaxation_method=1;  // default GSRB    GS 0   GSRB 1
int global_mgRelax_numBottomGSRB=4;
int global_mgRelax_numSmoothUp=4;
int global_mgRelax_numSmoothDown=4;
int global_BS_algorithm=0; // default bicgstab
double global_BS_tolerance = 1.0e-10;
double global_GSRB_relaxFactor = 1.0; // only matters if turned on in FORT_GSRB

// define main loop here because that's the way amrgodunov does it!
void mainLoop();

extern void dumpmemoryatexit(void);

//#define TRAP_FPE  //(should be off by default)
#ifdef TRAP_FPE
// funproto
static void __attribute__ ((constructor)) trapfpe(void);
#endif

//int get_memory_usage_from_OS(void) {return 0;}

int main(int argc, char* argv[])
{

#ifdef MPI
  MPI_Init(&argc, &argv);
#ifdef CH_AIX  
  H5dont_atexit();
#endif
  setChomboMPIErrorHandler();
  MPI_Barrier(Chombo_MPI::comm);  // Barrier #1 
#endif

  int rank, number_procs;
#ifdef MPI
  MPI_Comm_rank(Chombo_MPI::comm, &rank);
  MPI_Comm_size(Chombo_MPI::comm, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif

  // init silly global variables
  IterationFlavor = 0;
  for (int i=0; i<MAX_ITERATION_COUNTERS; ++i) {
    NoelsSooperIterationCounter[i].Count = 0;
    NoelsSooperIterationCounter[i].Value = 0;
    NoelsSooperIterationCounter[i].NmaxIteration = 0;
    NoelsSooperIterationCounter[i].wc = 0.0;
  }

  if (rank==0) {
    pout() << " number_procs= " << number_procs << endl;
    dumpAllThoseCrazyDefines();
  }

  Timer::TimerInit(rank);

#ifdef TRAP_FPE
  trapfpe();
#endif
  

  Everything.start();

  // read inputs
  char* in_file = argv[1];
  NoelsDumpInputFile(in_file);
  ParmParse pp(argc-2, argv+2,NULL, in_file);

#ifdef MPI
  TMPI_Barrier2.start();
  MPI_Barrier(Chombo_MPI::comm);
  TMPI_Barrier2.stop();
#endif
  // run actual algorithm
  mainLoop();

  Everything.stop();

  Real end_memory = get_memory_usage_from_OS();
  
  pout() << "Everything done.  mem= " << end_memory 
         << "MB wall-clock time: "    << setiosflags(ios::fixed)
         << Everything.wc_time()      << " seconds" << endl;

#if defined(TIMER) && defined(MPI)
  // Gather peak memory from procs and write a single line to screen.  that's all.
  Real avg_memory, min_memory, max_memory;
  gather_memory_from_procs(end_memory, avg_memory, min_memory, max_memory);
#endif

  Timer::TimerSummary();

#ifdef MPI
  dumpmemoryatexit();
#endif  

#ifdef COUNT_PEX  
  FILE *PEXOUT;
  char filename[80];
  sprintf(filename, "pex%03d.dat", rank);
  PEXOUT = fopen(filename, "w");
  if(PEXOUT == NULL) {
    printf("problem opening pex out\n");  
    exit(0); 
  }
  if (rank==0) printf("rank%d NPEX = %d\n", rank, NPEX);
  for(int i=1; i<NPEX; ++i) { 
    fprintf(PEXOUT, "%6d %5d %10d %10d %10d %10.2e\n", i, pex_Nsends[i], pex_Nlocal[i], 
            pex_Nto[i], pex_Nfrom[i], 1000*pex_delta_time[i]);
    if (pex_Nlocal[i] == 0 && pex_delta_time[i] < 1.0e-16) break;
  }
  fclose(PEXOUT);
#endif

#ifdef MPI
  MPI_Finalize();
#endif
}

void
mainLoop()
{
  NST_ReadInput.start();
  // read inputs
  ParmParse ppMain("main");
  
  num_levelGSRBs_inPC=2;
  ppMain.query("num_levelGSRBs_inPC", num_levelGSRBs_inPC);

  global_relaxation_method=1;
  ppMain.query("global_relaxation_method", global_relaxation_method);

  global_mgRelax_numBottomGSRB=4;
  ppMain.query("global_mgRelax_numBottomGSRB", global_mgRelax_numBottomGSRB);
  global_mgRelax_numSmoothUp=4;
  ppMain.query("global_mgRelax_numSmoothUp", global_mgRelax_numSmoothUp);
  global_mgRelax_numSmoothDown=4;
  ppMain.query("global_mgRelax_numSmoothDown", global_mgRelax_numSmoothDown);

  global_BS_tolerance=1.0e-10;
  ppMain.query("global_BS_tolerance", global_BS_tolerance);

  global_GSRB_relaxFactor= 1.0;
  ppMain.query("global_GSRB_relaxFactor", global_GSRB_relaxFactor);

  global_BS_algorithm=0;
  ppMain.query("global_BS_algorithm", global_BS_algorithm);

  int verbosity = 0;
  ppMain.query("verbosity", verbosity);
  assert (verbosity >= 0);

  int nstop = 0;
  ppMain.query("max_step", nstop);

  Real stopTime = 0.;
  ppMain.query("max_time", stopTime);

  Real fixed_dt = -1;
  ppMain.query("fixed_dt", fixed_dt);

  vector<int> tempVectInt;
  Vector<int> num_cells(SpaceDim,0);
  ppMain.queryarr("num_cells", tempVectInt, 0, SpaceDim);
  num_cells = tempVectInt;

  assert ( D_TERM ( ( num_cells[0] > 0 ), && 
                    ( num_cells[1] > 0 ), && 
                    ( num_cells[2] > 0 ) ) );
  assert ( D_TERM ( ( num_cells[0] % 2 == 0 ), &&
                    ( num_cells[1] % 2 == 0 ), &&
                    ( num_cells[2] % 2 == 0 ) ) );

  vector<int> is_periodica(SpaceDim,0);
  bool is_periodic[SpaceDim];

  ppMain.queryarr("is_periodic", is_periodica,0,SpaceDim);
  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++)
    {
      is_periodic[dim] = (is_periodica[dim] == 1);
      if (is_periodic[dim] && verbosity >= 2 && procID() == 0)
        pout() << "Using periodic BCs in direction:" << dim << endl;
    }

  int max_level = 0;
  ppMain.query("max_level", max_level);
  
  int num_read_levels = Max(max_level,1);
  Vector<int> ref_ratios (num_read_levels+1,1);
  ppMain.queryarr("ref_ratio", tempVectInt,0,num_read_levels+1);
  ref_ratios = tempVectInt;

  Vector<int> regrid_intervals(num_read_levels,1);
  ppMain.queryarr("regrid_interval", tempVectInt,0,num_read_levels);
  regrid_intervals = tempVectInt;
  
  Real refine_threshold = 0.3;
  ppMain.query("refine_threshold", refine_threshold);
  
  int block_factor = 2;
  ppMain.query("block_factor", block_factor);

  int max_grid_size = 32;
  ppMain.query("max_grid_size", max_grid_size);

  int max_base_grid_size = max_grid_size;
  ppMain.query("max_base_grid_size", max_base_grid_size);

  Real fill_ratio = 0.75;
  ppMain.query("fill_ratio", fill_ratio);

  int buffer_size = 2;
  ppMain.query("grid_buffer_size", buffer_size);

  int checkpoint_interval = -1;
  ppMain.query("checkpoint_interval", checkpoint_interval);

  int plot_interval = -1;
  ppMain.query("plot_interval", plot_interval);
  
  Real max_dt_grow = 1.5;
  ppMain.query("max_dt_grow", max_dt_grow);

  bool predefinedGrids = ppMain.contains("gridfile");
  string gridfile;
  if (predefinedGrids) {
    ppMain.get("gridfile", gridfile);
  }
  
  string checkPrefix = "chk";
  ppMain.query("checkPrefix", checkPrefix);
  
  string plotPrefix = "plt";
  ppMain.query("plotPrefix", plotPrefix);


  Real cfl = 0.5;
  ppMain.query("cfl", cfl);

  // solver parameter which will be passed all the way throughout the code
  int tempBool = 0;
  ppMain.query("limitSolverCoarsening", tempBool);
  bool limitSolverCoarsening = (tempBool == 1);

  if (verbosity >= 1) {
    // display inputs
    pout() << "nstop = " << nstop << endl;
    pout() << "stopTime = " << stopTime << endl;
    if (fixed_dt > 0) 
      pout() << "fixed dt = " << fixed_dt << endl;

    pout() << "num_cells = " << D_TERM ( num_cells[0] << "  " <<,
                                       num_cells[1] << "  " <<,
                                       num_cells[2] << ) endl;
  
    pout() << "max level = " << max_level << endl;
    pout() << "refinement ratios = ";
    for (int i = 0; i < ref_ratios.size(); ++i) pout() << ref_ratios[i] << " ";
    pout() << endl;
    pout() << "regrid intervals = ";
    for (int i = 0; i < regrid_intervals.size(); ++i) pout() << regrid_intervals[i] << " ";
    pout() << endl;
    pout() << "refinement threshold = " << refine_threshold << endl;

    pout() << "blocking factor = " << block_factor << endl;
    pout() << "max grid size = " << max_grid_size << endl;
    pout() << "max base grid size = " << max_base_grid_size << endl;
    pout() << "fill ratio = " << fill_ratio << endl;
    pout() << "grid buffer size = " << buffer_size << endl;

    pout() << "checkpoint interval = " << checkpoint_interval << endl;
    pout() << "plot interval = " << plot_interval << endl;
    pout() << "CFL = " << cfl << endl;
    pout() << "maxDtGrow = " << max_dt_grow << endl;
    pout() << "plotPrefix = " << plotPrefix << endl;
    pout() << "checkPrefix = " << checkPrefix << endl;
    pout() << "verbosity = " << verbosity << endl;
    pout() << "limit solver coarsening = " << limitSolverCoarsening << endl;
    pout() << "num_levelGSRBs_inPC = " << num_levelGSRBs_inPC << endl;
    pout() << "global_relaxation_method = " << global_relaxation_method << endl;
    pout() << "global_mgRelax_numBottomGSRB = " << global_mgRelax_numBottomGSRB << endl;
    pout() << "global_mgRelax_numSmoothUp = " << global_mgRelax_numSmoothUp << endl;
    pout() << "global_mgRelax_numSmoothDown = " << global_mgRelax_numSmoothDown << endl;
    pout() << "global_BS_algorithm = " << global_BS_algorithm << endl;
    pout() << "global_BS_tolerance = " << global_BS_tolerance << endl;
    pout() << "global_GSRB_relaxFactor = " << global_GSRB_relaxFactor << endl;

    if (predefinedGrids) {
      pout() << "gridFile = " << gridfile << endl;
    }
  }

  // create objects
  ProblemDomain prob_domain(IntVect::Zero,
                            IntVect(D_DECL(num_cells[0]-1,
                                           num_cells[1]-1,
                                           num_cells[2]-1)),
                            is_periodic);
  
  
  Vector<Vector<Box> > amrGrids;
  if (predefinedGrids) {
#ifdef MPI
    TMPI_Barrier3.start();
    //if (procID() ==  uniqueProc(SerialTask::compute)) {
    MPI_Barrier(Chombo_MPI::comm);
    TMPI_Barrier3.stop();
    if (procID() ==  0) {
#endif
      // read in predefined grids
      ifstream is(gridfile.c_str(), ios::in);
      if (is.fail()) {
	pout() << "cannot open grids file " << gridfile << endl;
	MayDay::Error();
      }
      
#ifdef MPI 
      if (verbosity >= 3) {
	pout() << "procID: " << procID() << "  opening gridfile \n" << endl;
      }
#endif      
      
      // format of file -- number of levels, then for
      // each level starting with level 1, number of
      // grids on level, list of boxes
      int in_numLevels;
      is >> in_numLevels;
      assert (in_numLevels <= max_level+1);
      if (verbosity >= 3) {
	pout() << "numLevels = " << in_numLevels << endl;
      }
      while (is.get() != '\n');
      amrGrids.resize(in_numLevels);
      // check to see if coarsest level needs to be broken up
      domainSplit(prob_domain,amrGrids[0], max_base_grid_size,4);
      
      if (verbosity >= 3) {
	pout() << "level 0: ";
	for (int n=0; n<amrGrids[0].size(); n++) {
	  pout () << amrGrids[0][0] << endl;
	}
      }
      // now loop over levels, starting with level 1
      int ngrid;
      for (int lev=1; lev<in_numLevels; lev++) {
	is >> ngrid;
	if (verbosity >= 3) {
	  pout() << "level " << lev << " numGrids = " << ngrid << endl;
	  pout() << "Grids: ";
	}
	while (is.get() != '\n');
	amrGrids[lev].resize(ngrid);
	for (int i=0; i<ngrid; i++) {
	  Box bx;
	  is >> bx;
	  while (is.get() != '\n');
	  // quick check on box size
	  Box bxRef(bx);
	  // not really sure why i was doing this (holdover from 
	  // legacy code)
	  //bxRef.refine(ref_ratios[lev-1]);
	  if (bxRef.longside() > max_grid_size) {
	    pout() << "Grid " << bx << " too large" << endl;
	    MayDay::Error();
	  }
	  if (verbosity >= 3) {
	    pout() << bx << endl;
	  }
	  amrGrids[lev][i] = bx;
	} // end loop over boxes on this level
      } // end loop over levels
      
#ifdef MPI
    } // end if procID = 0
    TMPI_readInput.start();
    //broadcast (amrGrids, uniqueProc(SerialTask::compute));
    broadcast (amrGrids, 0);
    TMPI_readInput.stop();
#endif
    
    
  } // end if grids are predefined
  NST_ReadInput.stop();


  pout() << " Beginning AMR Setup.  mem=" << get_memory_usage_from_OS() << "MB" << endl;
  NST_SetupAMR.start();
  
  AMRNavierStokesFactory amrns_fact;
  amrns_fact.CFL(cfl);
  amrns_fact.refinementThreshold(refine_threshold);

  channelBC physBC;
  amrns_fact.setPhysBC(physBC);

  amrns_fact.setLimitSolverCoarsening(limitSolverCoarsening);

  AMR thisAMR;
  
  thisAMR.define(max_level, ref_ratios, prob_domain,
		 &amrns_fact);

  thisAMR.verbosity(verbosity);
  AMRLevel::verbosity(verbosity);
  
  thisAMR.checkpointInterval(checkpoint_interval);
  thisAMR.plotInterval(plot_interval);
  thisAMR.maxGridSize(max_grid_size);
  thisAMR.maxBaseGridSize(max_base_grid_size);
  thisAMR.fillRatio(fill_ratio);  
  thisAMR.blockFactor(block_factor);
  // this is commented out because, although it is correct,
  // gridBufferSize sets the proper nesting radius, not the
  // buffer on the tags, which is what i thought it did.
  //  thisAMR.gridBufferSize(buffer_size);
  thisAMR.regridIntervals(regrid_intervals);
  thisAMR.maxDtGrow(max_dt_grow);
  if (fixed_dt > 0) thisAMR.fixedDt(fixed_dt);
  thisAMR.checkpointPrefix(checkPrefix);
  thisAMR.plotPrefix(plotPrefix);


  if (!ppMain.contains("restart_file") ) {
    if (!predefinedGrids) {
      // initialize from scratch for an AMR run
      // initialize hierarchy of levels
      thisAMR.setupForNewAMRRun();
      
    } else {
      thisAMR.setupForFixedHierarchyRun(amrGrids, 1);
    }
  } else {
    // initialize from restart file
    std::string restart_file;
    ppMain.query("restart_file", restart_file);    
#ifdef HDF5
    HDF5Handle handle(restart_file, HDF5Handle::OPEN_RDONLY);
    thisAMR.setupForRestart(handle);
    handle.close();
#else
    MayDay::Error("amrNavierStokes restart only defined with HDF5");
#endif
  if (verbosity >= 2) 
    {
      pout() << "restart_file = " << restart_file << endl;
    }

  } // end if restart
  NST_SetupAMR.stop();

#ifdef TIMER
   pout() << "AMR Setup done.  mem= " << get_memory_usage_from_OS() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << NST_SetupAMR.wc_time() 
          << " seconds" << endl;
#endif

  // run until conclusion
  NST_Run.start();
  thisAMR.run(stopTime, nstop);
  NST_Run.stop();

#ifdef TIMER
   pout() << "AMR Run completed.  mem= " << get_memory_usage_from_OS() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << NST_Run.wc_time() 
          << " seconds" << endl;
#endif

  // output last plotfile and statistics
  NST_Conclude.start();
  thisAMR.conclude();
  NST_Conclude.stop();

#ifdef TIMER
   pout() << "AMR Conclude completed.  mem= " << get_memory_usage_from_OS() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << NST_Conclude.wc_time() 
          << " seconds" << endl;
#endif

   writeOutNoelsSooperIterationCounterStuff();
}


#ifdef TRAP_FPE
// Previous versions of glibc require the following code:
extern "C" {
#include <fpu_control.h>
}
/* IM: Invalid operation mask
 * DM: Denormalized operand mask 
 * ZM: Zero-divide mask
 * OM: Overflow mask 
 * UM: Underflow mask 
 * PM: Precision (inexact result) mask */
static void __attribute__ ((constructor)) trapfpe(void)
{
  //pout() << " Turning on floating-point traps! " << endl;
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_UM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_ZM);
  fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_OM | _FPU_MASK_UM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_UM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_DM | _FPU_MASK_UM);
  //fpu_control_t cw = _FPU_DEFAULT;
   _FPU_SETCW(cw);
   /* On x86, this expands to: */
   /* unsigned int cw = 0x037f & ~(0x01 | 0x04 | 0x08); */
   /* __asm__ ("fldcw %0" : : "m" (*&cw));              */ 
}
#endif
