/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// from PoissonOp by DTGraves, Weds, July 21, 1999
// petermc, 20 Feb 2003

#include "HyprePoissonOp.H"
#include "HypreSmoother.H"
#include "RealVect.H"

/***********************/
/***********************/
LevelOp* 
HyprePoissonOp::new_levelop() const
{
  //only check to see if the boundary conditions are
  //defined.  not necessary to have whole thing defined.
  //the boundary conditions need to be defined because
  //the interface does not know about the boundary
  //condtions.  the solvers call the other define functions
  //in our inimitable two-stage construction
  assert(m_isBCDefined);
  HyprePoissonOp* newop = new HyprePoissonOp();
  if(newop == NULL)
    {
      MayDay::Error("Out of Memory in HyprePoissonOp::new_levelop");
    }
  newop->setDomainGhostBC(m_domghostbc);
  // do you need this?
  if (m_bottomSmootherPtr != NULL) 
    {
      newop->setBottomSmoother(*m_bottomSmootherPtr);
    }
  // added 20 feb 2003
  assert(m_bottomSmootherPtr != NULL);
  return static_cast<LevelOp*>(newop);
}

/***********************/
/***********************/
HyprePoissonOp::HyprePoissonOp():PoissonOp()
{
  // first calls PoissonOp(); you can't avoid this.
  setHypreValues();
}

/***********************/
/***********************/
HyprePoissonOp::~HyprePoissonOp()
{
  clearMemory();
  if (m_bottomSmootherPtr != NULL) 
    {
      delete m_bottomSmootherPtr;
      m_bottomSmootherPtr = NULL;
    }
  //if(m_bottomOp_ptr != NULL)
  //{
  //  delete m_bottomOp_ptr;
  //  m_bottomOp_ptr = NULL;
  //}
}

/***********************/
// return to undefined state
/***********************/
void
HyprePoissonOp::setHypreValues()
{
  // replace the bottom smoother
  if (m_bottomSmootherPtr != NULL) delete m_bottomSmootherPtr;
  m_bottomSmootherPtr = new HypreSmoother;
}


/***********************/
// return to undefined state
/***********************/

void
HyprePoissonOp::clearMemory()
{
  m_dxLevel = -1.0;
  m_refRatio = -1;
  m_isDefined = false;
  m_ihcfiEnabled = false;
  // don't touch bottom smoother here, since we want it to be able
  // to pass through the define statement unaltered
}


/***********************/
/***********************/

void
HyprePoissonOp::define(
                       const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real  a_dxLevel, 
                       int a_refRatio,
                       const Box& a_domain,
                       bool a_homogeneousOnly,
                       int a_ncomp)
{
  ProblemDomain probdomain(a_domain);
  define(a_grids, a_baseBAPtr, a_dxLevel, a_refRatio, probdomain,
         a_homogeneousOnly, a_ncomp);
}

/***********************/
/***********************/

void
HyprePoissonOp::define(const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real  a_dxLevel, 
                       int   a_refRatio,
                       const ProblemDomain& a_domain,
                       bool  a_homogeneousOnly,
                       int   a_ncomp)
{
  PoissonOp::define(a_grids, a_baseBAPtr, a_dxLevel, a_refRatio, a_domain,
                    a_homogeneousOnly, a_ncomp);
  RealVect dxLevelVect(D_DECL(m_dxLevel, m_dxLevel, m_dxLevel));
  int solver_flag = 0; // 0 for SMG solver, 1 for PFMG solver
  HypreSmoother* hypreBottom = (HypreSmoother*) m_bottomSmootherPtr;
  // PoissonOp has field DisjointBoxLayout m_baseBA
  hypreBottom->setup_solver(m_grids, a_baseBAPtr,
                            m_domain, dxLevelVect,
                            m_refRatio, solver_flag);

//    HypreSmoother* hypreBottomPtr = new HypreSmoother();
//    // PoissonOp has field DisjointBoxLayout m_baseBA
//    hypreBottomPtr->setup_solver(m_grids, a_baseBAPtr,
//                                 m_domain, dxLevelVect,
//                                 m_refRatio, solver_flag);
//    delete m_bottomSmootherPtr;
//    m_bottomSmootherPtr = (BaseBottomSmoother*) hypreBottomPtr;
}

/***********************/
// a_reftofine is the refinement between this operator and a_opfine
/***********************/
void
HyprePoissonOp::define(
                       const LevelOp* a_opfine,
                       int a_reftoFine)
{
  PoissonOp::define(a_opfine, a_reftoFine);
  HypreSmoother* hypreBottom = (HypreSmoother*) m_bottomSmootherPtr;
  RealVect dxLevelVect(D_DECL(m_dxLevel, m_dxLevel, m_dxLevel));
  int solver_flag = 0; // 0 for SMG solver, 1 for PFMG solver
  hypreBottom->setup_solver(m_grids, NULL,
                            m_domain, dxLevelVect,
                            m_refRatio, solver_flag);
}
