/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// from NoOpSmoother by DFMartin, Sun, May 5, 2002
// petermc, 10 Jan 2003

#include "HypreSmoother.H"
#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"

#include "HYPRE_struct_ls.h"
#include "HABECF_F.H"

#define STENCILSIZE (2*CH_SPACEDIM + 1)

#include "Timer.H"

using std::cout;
using std::cerr;
using std::endl;


// --------------------------------------------------------------------------

struct HypreSmoother::HYPRE_impl
{
  HYPRE_StructGrid    grid;
  //HYPRE_StructStencil stencil;

  HYPRE_StructMatrix  A; // , A0;  don't need A0
  HYPRE_StructVector  b;
  HYPRE_StructVector  x;

  HYPRE_StructSolver  solver;
};

// --------------------------------------------------------------------------

#include "struct_mv.h"

// --------------------------------------------------------------------------

namespace
{
  MPI_Comm mpi_comm = MPI_COMM_WORLD;
  
  inline int ispow2(int i)
  {
    return (i == 1) ? 1 : (((i <= 0) || (i & 1)) ? 0 : ispow2(i / 2));
  }
  
  inline int*
  loV(const Box& b)
  {
    return const_cast<int*>(b.loVect());
  }
  
  inline int*
  hiV(const Box& b)
  {
    return const_cast<int*>(b.hiVect());
  }
}

// --------------------------------------------------------------------------

HypreSmoother::HypreSmoother()
{
  m_initializedHypre = false;
}

// --------------------------------------------------------------------------

HypreSmoother::~HypreSmoother()
{
  clearMemory();
}

// --------------------------------------------------------------------------

BaseBottomSmoother*
HypreSmoother::new_bottomSmoother() const
{
  HypreSmoother* newsmoother = new HypreSmoother();
  if (newsmoother == NULL) 
    {
      MayDay::Error("Out of Memory in HypreSmoother::new_bottomSmoother");
    }
  return static_cast<BaseBottomSmoother*> (newsmoother);
}

// --------------------------------------------------------------------------
void
HypreSmoother::setup_solver(const DisjointBoxLayout& a_grids, 
                            const DisjointBoxLayout* a_baseGrids, 
                            const ProblemDomain& a_domain, 
                            const RealVect& a_dxLevel,
                            int a_nRefCrse,
                            int a_solver_flag)
{
  // m_initializedHypre = false;
  clearMemory();

  m_solver_flag = a_solver_flag;
  m_numPre = 4; // was 1
  m_numPost = 4; // was 1
  m_impl = new HYPRE_impl();
  m_relTolerance = 1.0e-9;
  m_maxIter = 33;

  m_grids = a_grids;
  m_domain = a_domain;
  m_dxLevel = a_dxLevel;
  m_nRefCrse = a_nRefCrse;
  // Correction by petermc, 16 Jan 2003
  if (a_baseGrids == NULL) m_nRefCrse = 0;

  /*
    Set up structured-grid object, m_impl->grid.
  */

  // first create empty structured-grid object, m_impl->grid,
  // living on mpi_comm.
  HYPRE_StructGridCreate(mpi_comm, CH_SPACEDIM, &m_impl->grid);

  // Remove the periodic stuff because it causes problems.
  // (Must be power of 2, and boxes need to wrap around.)
  /*
  if ( m_domain.isPeriodic() )  // periodic in any direction
    {
      const Box& bx = m_domain.domainBox();
      int periodicity[SpaceDim];
      for ( int idim = 0; idim < SpaceDim; idim++ )
        {
          periodicity[idim] = 0;
          // if ( geom.isPeriodic(i) )
          if ( m_domain.isPeriodic(idim) )
            {
              periodicity[idim] = bx.size(idim);
              // verify that periodicity[idim] is a power of 2, because
              // hypre requires this.

              assert(ispow2(periodicity[idim]));

              assert(bx.smallEnd(0) == 0);
            }
        }
      // Set periodicities of box.
      HYPRE_StructGridSetPeriodic(m_impl->grid, periodicity);
    }
  */

  // DataIterator gets data in this proc only.
  DataIterator dit = m_grids.dataIterator();

  // (In the original, if num_procs != 1, then use distribution
  // mapping to set only on this proc.)
  for (dit.begin(); dit.ok(); ++dit)
    {
      // Add a new box to the structured-grid object.
      const Box& bx = a_grids.get(dit());
      HYPRE_StructGridSetExtents(m_impl->grid, loV(bx), hiV(bx));
    }

  // Finalize assembly of structured-grid object.
  HYPRE_StructGridAssemble(m_impl->grid);

  /*
    Set up stencil.
  */

#if (CH_SPACEDIM == 2)
  // 5-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1
  //   4
  int offsets[5][2] = {{ 0,  0},
                       {+1,  0},
                       {-1,  0},
                       { 0, +1},
                       { 0, -1}};
#elif (CH_SPACEDIM == 3)
  // 7-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1   (and:  5 above 0, 6 below 0)
  //   4
  int offsets[7][3] = {{ 0,  0,  0},
                       {+1,  0,  0},
                       {-1,  0,  0},
                       { 0, +1,  0},
                       { 0, -1,  0},
                       { 0,  0, +1},
                       { 0,  0, -1}};
#endif

  // number of ghost layers in each direction.  always array of length 6.
#if (CH_SPACEDIM == 2)
  //int A_num_ghost[4] = { 1, 1, 1, 1 };
  int A_num_ghost[6] = { 1, 1, 1, 1, 0, 0 };
#elif (CH_SPACEDIM == 3)
  int A_num_ghost[6] = { 1, 1, 1, 1, 1, 1 };
#endif

  HYPRE_StructStencil stencil;

  // first create empty CH_SPACEDIM-dimensional, STENCILSIZE-point
  // stencil object, stencil.
  HYPRE_StructStencilCreate(CH_SPACEDIM, STENCILSIZE, &stencil);

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      // assign array numbers for stencil entry icomp.
      HYPRE_StructStencilSetElement(stencil, icomp, offsets[icomp]);
    }

  /*
    Set up matrix.
  */

  // first create empty matrix object, m_impl->A,
  // with m_impl->grid and stencil.
  HYPRE_StructMatrixCreate(mpi_comm, m_impl->grid, stencil, &m_impl->A);
  HYPRE_StructMatrixSetNumGhost(m_impl->A, A_num_ghost);
  // tell hypre that matrix coefficients are now ready to be set.
  HYPRE_StructMatrixInitialize(m_impl->A);

  // Matrix coefficients are set in setup_solver(), not here.

  /*
    Set up right-hand side vector.
  */

  // create empty vector objects, m_impl->b and m_impl->x, with m_impl->grid.
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->b);
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->x);

  HYPRE_StructStencilDestroy(stencil); // no longer needed

  // tell hypre that vector values are now ready to be set.
  HYPRE_StructVectorInitialize(m_impl->b);
  HYPRE_StructVectorInitialize(m_impl->x);

  /*
    Other stuff:  define m_acoefs and m_bcoefs.
  */

  const int ncomp=1;
  const int ngrow=0;
  IntVect ghostVect = ngrow * IntVect::Unit;
  m_acoefs = new LevelData<FArrayBox>(a_grids, ncomp, ghostVect);
  for (dit.begin(); dit.ok(); ++dit)
    m_acoefs->operator[](dit()).setVal(0.);
 
  m_bcoefs = new LevelData<FluxBox>(a_grids, ncomp, ghostVect);
  // m_bcoefs not initialized

  /*
    Set up hypre solver.
  */

  m_alpha = 0.; // weight of identity
  m_beta = 1.; // weight of div grad

  for (dit.begin(); dit.ok(); ++dit)
    {
      m_acoefs->operator[](dit()).setVal(0.);
      m_bcoefs->operator[](dit()).setVal(1.);
    }
  setup_matrix();
}

// --------------------------------------------------------------------------

void 
HypreSmoother::setup_matrix()
{

  //pout() << " 0 HypreSmoother::setup_matrix()" << endl;
  int stencil_indices[STENCILSIZE];

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      stencil_indices[icomp] = icomp;
    }

  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  int refToCoarse = m_nRefCrse;
  // grrr...
  if (refToCoarse < 0) refToCoarse = 1;

  //int N=0;
  for (DataIterator dit = m_acoefs->dataIterator(); dit.ok(); ++dit)
    {
      const Box& bx = m_grids.get(dit());
      //pout() << "  N" << N << " " << bx << endl;
      //N++;

      // petermc:  must we allocate with hypre_CTAlloc?

      // const int volume = bx.numPts();
      // Real* mat = hypre_CTAlloc(double, STENCILSIZE*volume);
      // original has mat(0:3, regl0:regh0, regl1:regh1, regl2:regh2)
      // note that the component indices are FIRST,
      // but Chombo Fortran assumes they are LAST.
      // Will need to do some complicated choreography to transpose.
      FArrayBox matFab(bx, STENCILSIZE);

      /*
        build matrix interior
      */

      // Operator on phi(:) is:
      // L(phi)(:) = alpha * a(:) * phi(:) + beta * div (b(:) * grad phi(:))

      // Set coefficients of first term, alpha * a(:) * phi(:).
      if (m_alpha == 0.)
        {
          matFab.setVal(0., 0); // set component 0
        }
      else
        {
          FArrayBox& acoefsFab = m_acoefs->operator[](dit());
          matFab.copy(acoefsFab, 0, 0, 1); // srccomp, destcomp, numcomp
          matFab.mult(m_alpha, 0, 1); // startcomp, numcomp
        }
//        FORT_HACOEF(CHF_FRA(mat),
//                    CHF_CONST_FRA1(acoefsFab, 0),
//                    CHF_CONST_REAL(m_alpha));

      // Set coefficients of second term, beta * div (b(:) * grad phi(:)).
      for ( int idim = 0; idim < SpaceDim; idim++ )
	{
          const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];
          FORT_HBCOEF(CHF_FRA(matFab),
                      CHF_CONST_FRA1(bfab, 0),
                      CHF_CONST_REAL(m_beta),
                      CHF_CONST_REALVECT(m_dxLevel),
                      CHF_CONST_INT(idim));
	}

      /*
        add boundary conditions to matrix diagonal, and
        zero out offdiag values at domain boundaries.
      */

      SideIterator sit;
      for (int idim = 0; idim < SpaceDim; idim++)
        // if ( ! m_domain.isPeriodic(idim) ) // hypre has probs with periodic
          for (sit.begin(); sit.ok(); sit.next())
          {
            Box region(bx);

            // const BaseFab<int>& msk = m_bd.bndryMasks(oitr())[dit()];
            // const Box&      bbox = (*m_bcoefs[idim])[dit()].box();
            // const Box&      msb  = msk.box();

            // const Array<BoundCond>& bct = m_bd.bndryConds(oitr())[dit()];
            // int bct = LO_DIRICHLET;
            // BL_ASSERT( bct.size() == 1);
            int iside;
            bool intersectsBoundary = false;
            switch (sit())
              {
              case Side::Lo:
                {
                  int smallCoordinate = region.smallEnd(idim);
                  if (smallCoordinate == loDomain[idim]) 
                    {
                      intersectsBoundary = true;
                      region.setBig(idim, smallCoordinate);
                      iside = -1;
                    }
                  break;
                }
              case Side::Hi:
                {
                  int bigCoordinate = region.bigEnd(idim);
                  if (bigCoordinate == hiDomain[idim])
                    {
                      intersectsBoundary = true;
                      region.setSmall(idim, bigCoordinate);
                      iside = +1;
                    }
                  break;
                }
              default:
                cerr << "SideIterator error" << endl;
                abort();
              }

            if (intersectsBoundary)
              {
                const FArrayBox& bfab = m_bcoefs->operator[](dit())[idim];
                FORT_HBMAT(CHF_FRA(matFab),
                           CHF_CONST_FRA1(bfab, 0),
                           CHF_BOX(region),
                           CHF_CONST_INT(idim),
                           CHF_CONST_INT(iside),
                           CHF_CONST_INT(refToCoarse),
                           CHF_CONST_REAL(m_beta),
                           CHF_CONST_REALVECT(m_dxLevel));

                // dummy statement in order to be able to get around gdb bug
                if (false) assert(0 == 0);

//  	      FORT_HBMAT(mat, ARLIM(bx.loVect()), ARLIM(bx.hiVect()),
//  			 cdir, bct[0], bcl,
//  			 msk.dataPtr(), ARLIM(msb.loVect()), ARLIM(msb.hiVect()),
//  			 (*m_bcoefs[idim])[i].dataPtr(), ARLIM(bbox.loVect()), ARLIM(bbox.hiVect()),
//  			 m_beta, m_dx);
              }
          } // end iteration over faces

      const int volume = bx.numPts();
      Real* mat = hypre_CTAlloc(Real, STENCILSIZE*volume);

      transpose(mat, matFab);

      // initialize matrix

      // fill in the part of the matrix m_impl->A on box bx, with mat.
      HYPRE_StructMatrixSetBoxValues(m_impl->A, loV(bx), hiV(bx),
				     STENCILSIZE, stencil_indices, mat);

      hypre_TFree(mat);
    }
  //pout() << " 1 HypreSmoother::setup_matrix()" << endl;

  // Finalize assembly of matrix object.
  HYPRE_StructMatrixAssemble(m_impl->A);

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op
  //pout() << " 2 HypreSmoother::setup_matrix()" << endl;

  if ( m_solver_flag == 0 )
    {
      HYPRE_StructSMGCreate(mpi_comm, &m_impl->solver);
      HYPRE_StructSMGSetMemoryUse(m_impl->solver, 0);
      HYPRE_StructSMGSetMaxIter(m_impl->solver, m_maxIter);
      HYPRE_StructSMGSetRelChange(m_impl->solver, 0);
      HYPRE_StructSMGSetTol(m_impl->solver, m_relTolerance);
      HYPRE_StructSMGSetNumPreRelax(m_impl->solver, m_numPre);
      HYPRE_StructSMGSetNumPostRelax(m_impl->solver, m_numPost);
      HYPRE_StructSMGSetLogging(m_impl->solver, 1);
      //pout() << " 26 HypreSmoother::setup_matrix()" << endl;
      // getting hung right here: noel
      HYPRE_StructSMGSetup(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
      //pout() << " 27 HypreSmoother::setup_matrix()" << endl;
    }
  else if ( m_solver_flag == 1 )
    {
      HYPRE_StructPFMGCreate(mpi_comm, &m_impl->solver);
      //HYPRE_StructPFMGSetMemoryUse(solver, 0);
      // for isotropic problems, can set to 1.  otherwise, 0.
      HYPRE_StructPFMGSetSkipRelax(m_impl->solver, 1);
      HYPRE_StructPFMGSetMaxIter(m_impl->solver, m_maxIter);
      HYPRE_StructPFMGSetRelChange(m_impl->solver, 0);
      HYPRE_StructPFMGSetTol(m_impl->solver, m_relTolerance);
      HYPRE_StructPFMGSetNumPreRelax(m_impl->solver, m_numPre);
      HYPRE_StructPFMGSetNumPostRelax(m_impl->solver, m_numPost);
      HYPRE_StructPFMGSetLogging(m_impl->solver, 1);
      HYPRE_StructPFMGSetup(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else
    {
      MayDay::Error("LevelHypreSolver: no such solver");
    }
  m_initializedHypre = true;
  //pout() << " LAST HypreSmoother::setup_matrix()" << endl;
}

// --------------------------------------------------------------------------

void 
HypreSmoother::clearMemory()
{
  if (m_initializedHypre)
    {
      delete m_acoefs;
      delete m_bcoefs;
      // destroy the solver before destroying m_impl
      clear_hypre();
      HYPRE_StructVectorDestroy(m_impl->b);
      HYPRE_StructVectorDestroy(m_impl->x);
      HYPRE_StructMatrixDestroy(m_impl->A);
      HYPRE_StructGridDestroy(m_impl->grid);
      delete m_impl;
    }
  m_initializedHypre = false;
}


// --------------------------------------------------------------------------

void
HypreSmoother::clear_hypre()
{
  if ( m_solver_flag == 0 )
    {
      HYPRE_StructSMGDestroy(m_impl->solver);
    }
  else
    {
      HYPRE_StructPFMGDestroy(m_impl->solver);
    }
}

// --------------------------------------------------------------------------

void 
HypreSmoother::doBottomSmooth(LevelData<FArrayBox>& a_phi, 
                             const LevelData<FArrayBox>& a_rhs,
                             LevelOp* a_levelop_ptr)
{
  pout() << " HypreSmoother::doBottomSmooth" << endl;
  // is there a HypreSmoother object for each level?  or only one?
  // when does it (do they) get created?
  // answer:  hard to say!
  // For each level, I should set up a HypreSmoother and then
  // call m_levelop.setBottomSmoother(ph)
  // which will set m_levelop.m_bottom_smoother_ptr to ph.new_bottomSmoother().
  // Now can I get m_levelop for every level?
  // Idea:  Use HyprePoissonOp, an extension of PoissonOp, which
  // sets the smoother to HypreSmoother.

  Timer HBS;
  HBS.start();
  assert(m_initializedHypre);

  const DisjointBoxLayout& grids = a_phi.getBoxes();
  // assert( grids == m_grids );

  //phi.setVal(0.0);
  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  Real *vec;
  bool has_ghosts = ( a_phi.ghostVect() != IntVect::Zero );
  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      // initialize phi, since we will reuse the space to set up rhs below:

      FArrayBox& phiFab = a_phi[dit()];
      FArrayBox* phiFabPtr = &phiFab;

      if (has_ghosts)
	{
	  phiFabPtr = new FArrayBox(bx, 1);
          phiFabPtr->copy(phiFab);
	}

      // sharing space, phi will be overwritten below
      vec = phiFabPtr->dataPtr(); 

      // set vector coefficients for this box
      HYPRE_StructVectorSetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      const FArrayBox& rhsFab = a_rhs[dit()];
      phiFabPtr->copy(rhsFab); // why do i do this?

      // incorporate boundary conditions into rhs:
      // do nothing, because this is the homogeneous case.

      // initialize rhs

      HYPRE_StructVectorSetBoxValues(m_impl->b, loV(bx), hiV(bx), vec);

      if (has_ghosts) delete phiFabPtr; // contains vec
    }

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op

  if ( m_solver_flag == 0 )
    {
      HYPRE_StructSMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }
  else
    {
      HYPRE_StructPFMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
    }

  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      FArrayBox& phiFab = a_phi[dit()];
      FArrayBox* phiFabPtr = &phiFab;

      if (has_ghosts)
	{
	  phiFabPtr = new FArrayBox(bx, 1);
          phiFabPtr->copy(phiFab);
	}

      // sharing space, phi will be overwritten below
      vec = phiFabPtr->dataPtr(); 

      HYPRE_StructVectorGetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      if (has_ghosts) 
	{
	  a_phi[dit()].copy(*phiFabPtr);
	  delete phiFabPtr;
	}
    }


//   if (m_verbose && procID() == uniqueProc(SerialTask::compute))
//     {
//       int num_iterations=0;
//       Real res;
//       if ( m_solver_flag == 0 )
// 	{
// 	  HYPRE_StructSMGGetNumIterations(m_impl->solver, &num_iterations);
// 	  HYPRE_StructSMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
// 	}
//       else
// 	{
// 	  HYPRE_StructPFMGGetNumIterations(m_impl->solver, &num_iterations);
// 	  HYPRE_StructPFMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
// 	}
//       const int oldprec = pout().precision(20);
//       pout() << num_iterations << " Hypre Multigrid Iterations, Relative Residual "
//              << res << endl;
//       pout().precision(oldprec);
//     }

  HBS.stop();
  //#if 0
  int num_iterations=0;
  Real res;
  if ( m_solver_flag == 0 ) {
    HYPRE_StructSMGGetNumIterations(m_impl->solver, &num_iterations);
    HYPRE_StructSMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
  } else {
    HYPRE_StructPFMGGetNumIterations(m_impl->solver, &num_iterations);
    HYPRE_StructPFMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
  }
  // debug print statements
  int Npoints=0, Nboxes=0;  
  LayoutIterator lit = grids.layoutIterator();  
  for (lit.begin(); lit.ok(); ++lit) {
    Nboxes++;  
    Npoints += grids[lit()].numPts();
  }
  // Nboxes should now be same as grids.size()
  char stuff[180];
  sprintf(stuff, " end HYPREdoBottomSmooth Npoints=%7d Nboxes=%7d iterations=%3d res=%10.3e wc=%10.3f\n",
          Npoints, Nboxes, num_iterations, res,
          HBS.wc_time());
  pout() << stuff;
  //#endif
}


// --------------------------------------------------------------------------

void
HypreSmoother::transpose(Real* a_array,
                         const FArrayBox& a_fab) const
{
  Box bx = a_fab.box();
  int ncomps = a_fab.nComp();
  BoxIterator bit(bx);
  IntVect bxLo = bx.smallEnd();
  IntVect bxHi = bx.bigEnd();
  IntVect lens = bxHi - bxLo + IntVect::Unit;
  for (bit.begin(); bit.ok(); ++bit)
    {
      IntVect iv = bit();
      IntVect ivOff = iv - bxLo;
      // ptIndex = (ivOff[0] * lens[0] + ivOff[1]) * lens[1] + ivOff[2] ?
      // ptIndex = ivOff[2] + lens[2] * (ivOff[1] + lens[1] * ivOff[0]) ?
      // ptIndex = ivOff[0] + lens[0] * (ivOff[1] + lens[1] * ivOff[2])
      int ptOffIndex = ivOff[SpaceDim-1];
      for (int idim = SpaceDim-2; idim >= 0; idim--)
        {
          ptOffIndex *= lens[idim];
          ptOffIndex += ivOff[idim];
        }
      for (int nv = 0; nv < ncomps; nv++)
        {
          // int ind = nv * volume + ptOffIndex;
          int ind = ptOffIndex * ncomps + nv;
          a_array[ind] = a_fab(iv, nv);
        }
    }
}
