/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, July 23, 1999

#ifndef _LevelHypreSolver_H_
#define _LevelHypreSolver_H_

#include <cstdlib>
#include <iostream>
#include <assert.h>
#include <cmath>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "LevelMG.H"
#include "ProblemDomain.H"

/// Solves elliptic equations on a level
/**
   LevelHypreSolver Solves elliptic equations on a level
   using multigrid.
*/
class LevelHypreSolver
{

#ifdef HYPRE_AS_LEVELSOLVER
  struct HYPRE_impl;
#endif

public:

  /// 
  /**
     Default constructor leaves LevelHypreSolver undefined.
  */
  LevelHypreSolver();

  /// 
  virtual ~LevelHypreSolver();

  ///
  /**
     Full constructor.  Calls full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  LevelHypreSolver(const DisjointBoxLayout& a_grids, 
              const DisjointBoxLayout* a_baseGrids, 
              const Box& a_domain, 
              Real a_dxLevel,  
              int a_nRefCrse,
              const LevelOp* const a_opin,
              int a_ncomp = 1);


  ///
  /**
     Full constructor.  Calls full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  LevelHypreSolver(const DisjointBoxLayout& a_grids, 
              const DisjointBoxLayout* a_baseGrids, 
              const ProblemDomain& a_domain, 
              Real a_dxLevel,  
              int a_nRefCrse,
              const LevelOp* const a_opin,
              int a_ncomp);


  ///   Full define function.
  /**
     Full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  virtual void define(const DisjointBoxLayout& a_grids, 
                      const DisjointBoxLayout* a_baseGrids,
                      const Box& a_domain, 
                      Real a_dxLevel,
                      int a_nRefCrse,
                      const LevelOp* const a_opin,
                      int a_ncomp=1);


  ///   Full define function.
  /**
     Full define function.\\
     {\bf Inputs:} \\
     a_grids: Grid on which solution will be computed.
     \\
     a_baseGrids: pointer to next coarser grids.  
     Should be == NULL if there is no coarser level.
     \\
     a_domain: domain of solution level
     \\
     a_dxLevel: grid spacing at solution level
     \\
     a_nRefCrse: refinement ratio between two levels.
     Not used if there is no finer level
     \\
     a_opin: operator L used in solution of Lphi=rho
  */
  virtual void define(const DisjointBoxLayout& a_grids, 
                      const DisjointBoxLayout* a_baseGrids,
                      const ProblemDomain& a_domain, 
                      Real a_dxLevel,
                      int a_nRefCrse,
                      const LevelOp* const a_opin,
                      int a_ncomp = 1);


  /// 
  bool isDefined() const;

  ///
  void clearMemory();

  ///
  void setDefaultValues();

  ///
  /**
     do a level solve on this level using multigrid
     uses homogeneous bcs at coarse/fine interfaces \\
     {\bf Inputs:} \\
     phi: initial guess at solution. 
     rhs: right-hand side. \\
     {\bf Outputs:} \\
     phi: solution.
  */
  void levelSolveH(LevelData<FArrayBox>& phi, 
                   const LevelData<FArrayBox>& rhs,
                   bool initializePhiToZero=true);

  ///
  /**
     do a level solve on this level using multigrid
     uses inhomogeneous bcs at interfaces\\
     {\bf Inputs:} \\
     phi: initial guess at solution.
     phic: solution at next coarser level.  should be == NULL
     if there is no coarser level.
     rhs: right-hand side. \\
     {\bf Outputs:} \\
     phi: solution.
  */
  void levelSolve(LevelData<FArrayBox>& phi, 
                  const LevelData<FArrayBox>* phic, 
                  const LevelData<FArrayBox>& rhs,
                  bool initializePhiToZero=true);

  /// set solution tolerance
  void setTolerance(Real a_tolerance);

  ///
  /**
     Set "operator tolerance" of iterative solution.  Iteration will 
     stop if (new_residual/old_residual) > 1-operatorTolerance (and 
     at least minIter iterations have been performed).
     Default is 1.0e-5.
  **/
  void setOperatorTolerance(Real a_operatorTolerance);


  /// 
  void setVerbose(bool a_verbose);

  /// set max number multigrid vcycles to use in solution
  void setMaxIter(int a_maxIter);

  ///
  /**
     Set min number of iterations.  Only relevant when residual is 
     not decreasing fast enough to satisfy the "operator tolerance".  
     Default is 4.
  */
  // added by Dan M, (9/20/02)
  void setMinIter(int a_minIter);

#ifdef HYPRE_AS_LEVELSOLVER

  void defineHypre(int a_hypreSolverFlag);

  void setupHypreMatrix();

  void setHypre(bool a_hypreOn);

  void clearHypre();

  void levelSolveHypre(LevelData<FArrayBox>& a_dest,
                       const LevelData<FArrayBox>& a_rhs);
#endif

protected:

  // LevelMG level solver object to relax on this level
  LevelMG m_levelMG;

  // LevelMG level solver object to relax on this level
  LevelOp* m_levelOpPtr;

  // has this LevelHypreSolver been defined yet?
  bool m_isDefined;

  // grids at this level
  DisjointBoxLayout  m_grids;

  // problem domain at this level
  ProblemDomain m_domain;

  // residual
  LevelData<FArrayBox> m_resid;

  // useful scratch space.
  LevelData<FArrayBox> m_scratch;

  // correction
  LevelData<FArrayBox> m_corr;

  //
  int m_nRefCrse;

  //
  Real m_dxLevel;

  //
  int m_maxIter;

  //
  int m_minIter;

  //
  Real m_tolerance;

  // 
  Real m_operatorTolerance;
  

  //
  bool m_verbose;

#ifdef HYPRE_AS_LEVELSOLVER

  int m_hypreSolverFlag;

  int m_hypreNumPre;

  int m_hypreNumPost;

  bool m_initializedHypre;

  // true if going to use hypre solver, false otherwise
  bool m_hypreOn;

  HYPRE_impl* m_impl;

  void transpose(Real* a_array,
                 const FArrayBox& a_fab) const;
#endif

private:
  void operator = (const LevelHypreSolver&){;}
  LevelHypreSolver(const LevelHypreSolver&){;}

};

#endif


