/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, July 23, 1999

#include "LevelHypreSolver.H"
#include "IntVectSet.H"
#include "parstream.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
#include "CFIVS.H"

#include "Timer.H"

using std::cout;
using std::cerr;
using std::endl;

///////////////

// stuff for hypre

#ifdef HYPRE_AS_LEVELSOLVER

#include "HYPRE_struct_ls.h"
// #include "struct_mv.h"
#include "struct_ls.h" // for hypre_SMGPrintLogging

#define STENCILSIZE (2*CH_SPACEDIM + 1)

namespace
{
  MPI_Comm mpi_comm = MPI_COMM_WORLD;

  inline int ispow2(int i)
  {
    return (i == 1) ? 1 : (((i <= 0) || (i & 1)) ? 0 : ispow2(i / 2));
  }

  inline int*
  loV(const Box& b)
  {
    return const_cast<int*>(b.loVect());
  }

  inline int*
  hiV(const Box& b)
  {
    return const_cast<int*>(b.hiVect());
  }
}


struct LevelHypreSolver::HYPRE_impl
{
  HYPRE_StructGrid    grid;
  //HYPRE_StructStencil stencil;

  HYPRE_StructMatrix  A;
  HYPRE_StructVector  b;
  HYPRE_StructVector  x;

  HYPRE_StructSolver  solver;
};

#endif // end #ifdef HYPRE

///////////////
bool
LevelHypreSolver::isDefined() const
{
  return m_isDefined;
}
///////////////
void 
LevelHypreSolver::setDefaultValues()
{
#ifdef HYPRE_AS_LEVELSOLVER
  m_initializedHypre = false;
#endif
  m_levelOpPtr = NULL;
  m_isDefined = false;
  // set this to an invalid value
  m_nRefCrse = -1;
  m_dxLevel = -1.0;
  m_maxIter = 33;
  m_minIter = 4;
  m_tolerance = 1.0e-10;
  m_operatorTolerance = 1.0e-4;
#ifdef CH_USE_FLOAT
  m_tolerance = sqrt(m_tolerance);
#endif
  m_verbose = false;
}

void 
LevelHypreSolver::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}
///////////////
void 
LevelHypreSolver::setTolerance(Real a_tolerance)
{
  m_tolerance = a_tolerance;
}
//////////////
// set tolerance for stopping due to hung convergence
void 
LevelHypreSolver::setOperatorTolerance(Real a_tolerance)
{
  assert (a_tolerance >= 0.0);
  m_operatorTolerance = a_tolerance;
}

///////////////
void 
LevelHypreSolver::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
#ifdef HYPRE_AS_LEVELSOLVER
  if (m_initializedHypre)
    cerr << "max number of iterations remains " << m_maxIter 
         << " and not changed to " << a_maxIter << endl;
  // You need max iterations to be set before calling setup on hypre solver.
  // A new call HYPRE_StructPFMGSetMaxIter(m_impl->solver, m_maxIter)
  // won't help, because the setup call allocates space depending on m_maxIter.
#endif
}

///////////////
void
LevelHypreSolver::setMinIter(int a_minIter)
{
  m_minIter = a_minIter;
}

///////////////
// default constructor
LevelHypreSolver::LevelHypreSolver()
{
  setDefaultValues();
}

// destructor
LevelHypreSolver::~LevelHypreSolver()
{
  clearMemory();
}

void 
LevelHypreSolver::define(const DisjointBoxLayout& a_grids, 
                    const DisjointBoxLayout* a_baseGrids, 
                    const Box& a_domain, 
                    Real a_dxLevel, int a_nRefCrse,
                    const LevelOp* const a_opin,
                    int a_ncomp)
{
  ProblemDomain physdomain(a_domain);
  define(a_grids, a_baseGrids, a_domain, a_dxLevel, a_nRefCrse,
         a_opin, a_ncomp);
}


void 
LevelHypreSolver::define(const DisjointBoxLayout& a_grids, 
                    const DisjointBoxLayout* a_baseGrids, 
                    const ProblemDomain& a_domain, 
                    Real a_dxLevel, int a_nRefCrse,
                    const LevelOp* const a_opin,
                    int a_ncomp)
{
  clearMemory();
  m_grids = a_grids;
  m_domain = a_domain;
  m_nRefCrse = a_nRefCrse;
  m_dxLevel = a_dxLevel;
  m_resid.define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_scratch.define(m_grids, a_ncomp, IntVect::TheZeroVector());
  m_corr.define(m_grids, a_ncomp, IntVect::TheUnitVector());
  m_isDefined = true;

  m_levelOpPtr = a_opin->new_levelop();
  m_levelOpPtr->define(m_grids, a_baseGrids, m_dxLevel, 
                       m_nRefCrse, m_domain, false, a_ncomp);

  // coarsen boxes once to start, because coarsest level boxes
  // must also be coarsenable (for proper coarse-fine interpolation)
  // however, if grid merely covers entire physical domain, then 
  // want to coarsen all the way down

  int nCoarserLevels = -1;

  // (DFM 10/21/02) don't bother doing _any_ of this if m_grids
  // is an empty set

  if (m_grids.size() > 0) 
    {
      // why do it this way rather than as a dense IVS?
      IntVectSet ivsCheck(m_domain.domainBox());
      LayoutIterator lit = m_grids.layoutIterator();
      for(lit.reset(); lit.ok(); ++lit)
        ivsCheck -= m_grids.get(lit());
      if(ivsCheck.isEmpty()) 
        nCoarserLevels += 1;
      
      int refrattot = 2;
      bool getOut = false;
      while(!getOut)
        {
          bool addone = true;
          for(lit.reset(); lit.ok(); ++lit)
            {
              Box blocal = m_grids.get(lit());
              if(refine(coarsen(blocal,refrattot),refrattot) != blocal)
                addone = false;
              if(!addone) break;
            }
          if(addone)
            {
              nCoarserLevels += 1;
              refrattot *= 2;
            }
          else
            {
              getOut = true;
            }
        }
    }
  nCoarserLevels = Max(nCoarserLevels, 0);
    
  
  m_levelMG.define(m_grids, a_baseGrids, m_dxLevel,
                   m_nRefCrse, m_domain, nCoarserLevels, a_opin, a_ncomp);

#ifdef HYPRE_AS_LEVELSOLVER
  // if you are going to use hypre then you'll need to turn it on
  // with setHypre(true)
  m_hypreOn = true;
  // check this!
  if (a_baseGrids == NULL) m_nRefCrse = 0;
#endif
}

#ifdef HYPRE_AS_LEVELSOLVER

void
LevelHypreSolver::defineHypre(int a_hypreSolverFlag)
{
  clearHypre();
  m_hypreSolverFlag = a_hypreSolverFlag;
  m_hypreNumPre = 4; // number of pre-relaxation sweeps; can tweak this
  m_hypreNumPost = 4; // number of post-relaxation sweeps; can tweak this
  Timer tm;
  tm.start();
  m_impl = new HYPRE_impl();

  /*
    Set up structured-grid object, m_impl->grid.
  */

  // first create empty structured-grid object, m_impl->grid,
  // living on mpi_comm.
  HYPRE_StructGridCreate(mpi_comm, CH_SPACEDIM, &m_impl->grid);

  // FIX this part so that it treats periodic correctly when needed
  /*
  if ( m_domain.isPeriodic() )  // periodic in any direction
    {
      const Box& bx = m_domain.domainBox();
      int periodicity[SpaceDim];
      for ( int idim = 0; idim < SpaceDim; idim++ )
        {
          periodicity[idim] = 0;
          if ( m_domain.isPeriodic(idim) )
            {
              periodicity[idim] = bx.size(idim);
              // verify that periodicity[idim] is a power of 2, because
              // hypre requires this.

              assert(ispow2(periodicity[idim]));

              assert(bx.smallEnd(0) == 0);
            }
        }
      // Set periodicities of box.
      HYPRE_StructGridSetPeriodic(m_impl->grid, periodicity);
    }
  */

  // DataIterator gets data in this proc only.
  DataIterator dit = m_grids.dataIterator();

  // (In the original, if num_procs != 1, then use distribution
  // mapping to set only on this proc.)
  for (dit.begin(); dit.ok(); ++dit)
    {
      // Add a new box to the structured-grid object.
      const Box& bx = m_grids.get(dit());
      HYPRE_StructGridSetExtents(m_impl->grid, loV(bx), hiV(bx));
    }

  // Finalize assembly of structured-grid object.
  HYPRE_StructGridAssemble(m_impl->grid);

  /*
    Set up stencil.
  */

#if (CH_SPACEDIM == 2)
  // 5-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1
  //   4
  int offsets[5][2] = {{ 0,  0},
                       {+1,  0},
                       {-1,  0},
                       { 0, +1},
                       { 0, -1}};
#elif (CH_SPACEDIM == 3)
  // 7-point stencil:  offsets[stencil size][dimensions]
  //
  //   3
  // 2 0 1   (and:  5 above 0, 6 below 0)
  //   4
  int offsets[7][3] = {{ 0,  0,  0},
                       {+1,  0,  0},
                       {-1,  0,  0},
                       { 0, +1,  0},
                       { 0, -1,  0},
                       { 0,  0, +1},
                       { 0,  0, -1}};
#endif

  // number of ghost layers in each direction.  always array of length 6.
#if (CH_SPACEDIM == 2)
  //int A_num_ghost[4] = { 1, 1, 1, 1 };
  int A_num_ghost[6] = { 1, 1, 1, 1, 0, 0 };
#elif (CH_SPACEDIM == 3)
  int A_num_ghost[6] = { 1, 1, 1, 1, 1, 1 };
#endif

  HYPRE_StructStencil stencil;

  // first create empty CH_SPACEDIM-dimensional, STENCILSIZE-point
  // stencil object, stencil.
  HYPRE_StructStencilCreate(CH_SPACEDIM, STENCILSIZE, &stencil);

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      // assign array numbers for stencil entry icomp.
      HYPRE_StructStencilSetElement(stencil, icomp, offsets[icomp]);
    }

  /*
    Set up matrix.
  */

  // first create empty matrix object, m_impl->A,
  // with m_impl->grid and stencil.
  HYPRE_StructMatrixCreate(mpi_comm, m_impl->grid, stencil, &m_impl->A);
  HYPRE_StructMatrixSetNumGhost(m_impl->A, A_num_ghost);
  // tell hypre that matrix coefficients are now ready to be set.
  HYPRE_StructMatrixInitialize(m_impl->A);

  // Matrix coefficients are set in setup_solver(), not here.

  /*
    Set up right-hand side vector.
  */

  // create empty vector objects, m_impl->b and m_impl->x, with m_impl->grid.
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->b);
  HYPRE_StructVectorCreate(mpi_comm, m_impl->grid, &m_impl->x);

  HYPRE_StructStencilDestroy(stencil); // no longer needed

  // tell hypre that vector values are now ready to be set.
  HYPRE_StructVectorInitialize(m_impl->b);
  HYPRE_StructVectorInitialize(m_impl->x);

  tm.stop();
  pout() << "hypre time to set up grids, stencils, matrix, vectors:  " << tm.wc_time() << endl;
  setupHypreMatrix();
}

void
LevelHypreSolver::setupHypreMatrix()
{
  Timer tm;
  tm.start();

  int stencil_indices[STENCILSIZE];

  for ( int icomp = 0; icomp < STENCILSIZE; icomp++ )
    {
      stencil_indices[icomp] = icomp;
    }

  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();
  DataIterator dit = m_grids.dataIterator();

  int refToCoarse = m_nRefCrse;
  // check this!
  if (refToCoarse < 0) refToCoarse = 1;

  LevelData<FArrayBox> matAll(m_grids, STENCILSIZE);
  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = m_grids.get(dit());

      // petermc:  must we allocate with hypre_CTAlloc?

      // const int volume = bx.numPts();
      // Real* mat = hypre_CTAlloc(double, STENCILSIZE*volume);
      // original has mat(0:3, regl0:regh0, regl1:regh1, regl2:regh2)
      // note that the component indices are FIRST,
      // but Chombo Fortran assumes they are LAST.
      // Will need to do some complicated choreography to transpose.
      
      // FArrayBox matFab(bx, STENCILSIZE);
      FArrayBox& matFab = matAll[dit()];

      /*
        build matrix interior
      */

      // Operator on phi(:) is:
      // L(phi)(:) = alpha * a(:) * phi(:) + beta * div (b(:) * grad phi(:))
      // alpha = 0, a == 0, beta = 1, b == 1

      Real dx2inv = 1.0 / (m_dxLevel * m_dxLevel);
      matFab.setVal(dx2inv);
      matFab.setVal((-2.0 * SpaceDim) * dx2inv, 0); // set component 0

      /*
        add boundary conditions to matrix diagonal, and
        zero out offdiag values at domain boundaries.

        need to check how this relates to c/f boundaries.
      */

      Real distToCoarse = (refToCoarse + 1) * 0.5;
      Real wt0 = (2. - 2. / distToCoarse) * dx2inv;
      Real wtoff = (2. / (1. + distToCoarse) - 1.) * dx2inv;

      SideIterator sit;
      for (int idim = 0; idim < SpaceDim; idim++)
        // if ( ! m_domain.isPeriodic(idim) ) // not really periodic!  FIX this
          for (sit.begin(); sit.ok(); sit.next())
            {
              Side::LoHiSide hiorlo = sit();
              CFIVS mycfivs(m_domain, bx, m_grids, idim, hiorlo);
              
              // interp returns the ghost cells; therefore need to shift
              IntVectSet interp(mycfivs.getFineIVS());
              if (!interp.isEmpty())
                {
                  int iside = 0;
                  int compNextInside = 0;
                  int compNextOutside = 0;
                  if (hiorlo == Side::Lo)
                    {
                      iside = -1;
                      compNextInside = 2*idim + 1;
                      compNextOutside = 2*idim + 2;
                    }
                  else if (hiorlo == Side::Hi)
                    {
                      iside = +1;
                      compNextInside = 2*idim + 2;
                      compNextOutside = 2*idim + 1;
                    }

                  // shift by shiftVector to get the fine cells
                  // that are in the box, instead of ghost cells
                  IntVect shiftVector = -iside * BASISV(idim);
                  if (mycfivs.isPacked())
                    {
                      Box interpBox(mycfivs.packedBox());
                      interpBox.shift(shiftVector);

                      matFab.setVal(0., interpBox, compNextOutside);
                      matFab.plus(wtoff, interpBox, compNextInside, 1);
                      matFab.plus(wt0, interpBox, 0, 1);
                    }
                  else
                    {
                      interp.shift(shiftVector);

                      IVSIterator ivsit(interp);
                      for (ivsit.begin(); ivsit.ok(); ++ivsit)
                        {
                          IntVect ivf = ivsit();
                          matFab(ivf, compNextOutside) = 0.;
                          matFab(ivf, compNextInside) = wtoff;
                          matFab(ivf, 0) = wt0;
                        }
                    }
                }
            } // end iteration over faces

      const int volume = bx.numPts();
      Real* mat = hypre_CTAlloc(Real, STENCILSIZE*volume);

      transpose(mat, matFab);

      // initialize matrix

      // fill in the part of the matrix m_impl->A on box bx, with mat.
      HYPRE_StructMatrixSetBoxValues(m_impl->A, loV(bx), hiV(bx),
				     STENCILSIZE, stencil_indices, mat);

      hypre_TFree(mat);
    }

  tm.stop();
  pout() << "hypre time to fill in matrix:  " << tm.wc_time() << endl;

  // Finalize assembly of matrix object.
  tm.clear(); tm.start();
  HYPRE_StructMatrixAssemble(m_impl->A);

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op
  tm.stop();
  pout() << "hypre time to assemble:  " << tm.wc_time() << endl;

  if ( m_hypreSolverFlag == 0 )
    {
      tm.clear(); tm.start();
      HYPRE_StructSMGCreate(mpi_comm, &m_impl->solver);
      // HYPRE_StructSMGSetMemoryUse(m_impl->solver, 0);
      HYPRE_StructSMGSetMaxIter(m_impl->solver, m_maxIter);
      HYPRE_StructSMGSetRelChange(m_impl->solver, 0);
      HYPRE_StructSMGSetTol(m_impl->solver, m_tolerance);
      HYPRE_StructSMGSetNumPreRelax(m_impl->solver, m_hypreNumPre);
      HYPRE_StructSMGSetNumPostRelax(m_impl->solver, m_hypreNumPost);
      // This gives us updates of residual.
      HYPRE_StructSMGSetLogging(m_impl->solver, 1);
      tm.stop();
      pout() << "time for hypre SMG setup calls:  " << tm.wc_time() << endl;
      tm.clear(); tm.start();
      HYPRE_StructSMGSetup(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
      tm.stop();
      pout() << "time for HYPRE_StructSMGSetup:  " << tm.wc_time() << endl;
      // HYPRE_StructMatrixPrint("hyprematrix", m_impl->A, 0);
    }
  else if ( m_hypreSolverFlag == 1 )
    {
      tm.clear(); tm.start();
      HYPRE_StructPFMGCreate(mpi_comm, &m_impl->solver);
      //HYPRE_StructPFMGSetMemoryUse(solver, 0);
      // HYPRE_StructPFMGSetSkipRelax(m_impl->solver, 0);
      // Skip relaxation on certain grids for isotropic problems.
      HYPRE_StructPFMGSetSkipRelax(m_impl->solver, 1);
      HYPRE_StructPFMGSetMaxIter(m_impl->solver, m_maxIter);
      HYPRE_StructPFMGSetRelChange(m_impl->solver, 0);
      HYPRE_StructPFMGSetTol(m_impl->solver, m_tolerance);
      HYPRE_StructPFMGSetNumPreRelax(m_impl->solver, m_hypreNumPre);
      HYPRE_StructPFMGSetNumPostRelax(m_impl->solver, m_hypreNumPost);
      // This gives us updates of residual.
      HYPRE_StructPFMGSetLogging(m_impl->solver, 1);
      tm.stop();
      pout() << "time for hypre PFMG setup calls:  " << tm.wc_time() << endl;
      tm.clear(); tm.start();
      HYPRE_StructPFMGSetup(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
      tm.stop();
      pout() << "time for HYPRE_StructPFMGSetup:  " << tm.wc_time() << endl;
      // HYPRE_StructMatrixPrint("hyprematrix", m_impl->A, 0);
    }
  else
    {
      MayDay::Error("LevelHypreSolver: no such solver in hypre");
    }
  m_initializedHypre = true;
}

void
LevelHypreSolver::setHypre(bool a_hypreOn)
{
  m_hypreOn = a_hypreOn;
}

void 
LevelHypreSolver::clearHypre()
{
  if (m_initializedHypre)
    {
      Timer tm;
      tm.start();
      // delete m_acoefs;
      // delete m_bcoefs;
      // clear_hypre();
      if ( m_hypreSolverFlag == 0 )
        {
          HYPRE_StructSMGDestroy(m_impl->solver);
        }
      else
        {
          HYPRE_StructPFMGDestroy(m_impl->solver);
        }
      HYPRE_StructVectorDestroy(m_impl->b);
      HYPRE_StructVectorDestroy(m_impl->x);
      HYPRE_StructMatrixDestroy(m_impl->A);
      HYPRE_StructGridDestroy(m_impl->grid);
      delete m_impl;
      tm.stop();
      pout() << "hypre time for destroy:  " << tm.wc_time() << endl;
    }
  m_initializedHypre = false;
}


void
LevelHypreSolver::levelSolveHypre(LevelData<FArrayBox>& a_dest,
                             const LevelData<FArrayBox>& a_rhs)
{
  const DisjointBoxLayout& grids = a_dest.getBoxes();

  //dest.setVal(0.0);
  const IntVect loDomain = m_domain.domainBox().smallEnd();
  const IntVect hiDomain = m_domain.domainBox().bigEnd();

  Real *vec;
  bool has_ghosts = ( a_dest.ghostVect() != IntVect::Zero );
  DataIterator dit = a_dest.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      // initialize dest, since we will reuse the space to set up rhs below:

      FArrayBox& destFab = a_dest[dit()];
      FArrayBox* destFabPtr = &destFab;

      if (has_ghosts)
	{
	  destFabPtr = new FArrayBox(bx, 1);
          destFabPtr->copy(destFab);
	}

      // sharing space, dest will be overwritten below
      vec = destFabPtr->dataPtr(); 

      // set vector coefficients for this box
      HYPRE_StructVectorSetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      const FArrayBox& rhsFab = a_rhs[dit()];
      destFabPtr->copy(rhsFab); // why do i do this?

      // incorporate boundary conditions into rhs:
      // do nothing, because this is the homogeneous case.

      // initialize rhs

      HYPRE_StructVectorSetBoxValues(m_impl->b, loV(bx), hiV(bx), vec);

      if (has_ghosts) delete destFabPtr; // contains vec
    }

  HYPRE_StructVectorAssemble(m_impl->b); // currently a no-op
  HYPRE_StructVectorAssemble(m_impl->x); // currently a no-op

  // HYPRE_StructMatrixPrint("hyprematrix", m_impl->A, 0);
  // HYPRE_StructVectorPrint("hyprevector", m_impl->b, 0);
  if ( m_hypreSolverFlag == 0 )
    {
      HYPRE_StructSMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
      if (m_verbose) hypre_SMGPrintLogging(m_impl->solver, procID());
    }
  else
    {
      HYPRE_StructPFMGSolve(m_impl->solver, m_impl->A, m_impl->b, m_impl->x);
      if (m_verbose) hypre_PFMGPrintLogging(m_impl->solver, procID());
    }

  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& bx = grids.get(dit());

      FArrayBox& destFab = a_dest[dit()];
      FArrayBox* destFabPtr = &destFab;

      if (has_ghosts)
	{
	  destFabPtr = new FArrayBox(bx, 1);
          destFabPtr->copy(destFab);
	}

      // sharing space, dest will be overwritten below
      vec = destFabPtr->dataPtr(); 

      HYPRE_StructVectorGetBoxValues(m_impl->x, loV(bx), hiV(bx), vec);

      if (has_ghosts) 
	{
	  a_dest[dit()].copy(*destFabPtr);
	  delete destFabPtr;
	}
    }
  if (m_verbose && procID() == uniqueProc(SerialTask::compute))
    {
      int num_iterations;
      Real res;
      if ( m_hypreSolverFlag == 0 )
	{
	  HYPRE_StructSMGGetNumIterations(m_impl->solver, &num_iterations);
	  HYPRE_StructSMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
	}
      else
	{
	  HYPRE_StructPFMGGetNumIterations(m_impl->solver, &num_iterations);
	  HYPRE_StructPFMGGetFinalRelativeResidualNorm(m_impl->solver, &res);
	}
      pout() << num_iterations
             << " Hypre Multigrid Iterations, Relative Residual "
             << res << std::endl;
    }
}


void
LevelHypreSolver::transpose(Real* a_array,
                       const FArrayBox& a_fab) const
{
  Box bx = a_fab.box();
  int ncomps = a_fab.nComp();
  BoxIterator bit(bx);
  IntVect bxLo = bx.smallEnd();
  IntVect bxHi = bx.bigEnd();
  IntVect lens = bxHi - bxLo + IntVect::Unit;
  for (bit.begin(); bit.ok(); ++bit)
    {
      IntVect iv = bit();
      IntVect ivOff = iv - bxLo;
      // ptIndex = (ivOff[0] * lens[0] + ivOff[1]) * lens[1] + ivOff[2] ?
      // ptIndex = ivOff[2] + lens[2] * (ivOff[1] + lens[1] * ivOff[0]) ?
      // ptIndex = ivOff[0] + lens[0] * (ivOff[1] + lens[1] * ivOff[2])
      int ptOffIndex = ivOff[SpaceDim-1];
      for (int idim = SpaceDim-2; idim >= 0; idim--)
        {
          ptOffIndex *= lens[idim];
          ptOffIndex += ivOff[idim];
        }
      for (int nv = 0; nv < ncomps; nv++)
        {
          // int ind = nv * volume + ptOffIndex;
          int ind = ptOffIndex * ncomps + nv;
          a_array[ind] = a_fab(iv, nv);
        }
    }
}

#endif // end #ifdef HYPRE_AS_LEVELSOLVER


// complete constructor
LevelHypreSolver::LevelHypreSolver(const DisjointBoxLayout& a_grids, 
                         const DisjointBoxLayout* a_baseGrids,
                         const Box& a_domain, 
                         Real a_dxLevel, int a_nRefCrse,
                         const LevelOp* const a_opin, int a_ncomp)
{
  setDefaultValues();
  ProblemDomain physdomain(a_domain);
  define(a_grids,  a_baseGrids, physdomain, a_dxLevel, 
         a_nRefCrse, a_opin, a_ncomp);
}


// complete constructor
LevelHypreSolver::LevelHypreSolver(const DisjointBoxLayout& a_grids, 
                         const DisjointBoxLayout* a_baseGrids,
                         const ProblemDomain& a_domain, 
                         Real a_dxLevel, int a_nRefCrse,
                         const LevelOp* const a_opin,
                         int a_ncomp)
{
  setDefaultValues();
  define(a_grids,  a_baseGrids, a_domain, a_dxLevel, 
         a_nRefCrse, a_opin, a_ncomp);
}


// -----------------------------------------------------------------
// returns LevelHypreSolver to a basically undefined state
void 
LevelHypreSolver::clearMemory()
{
  if(m_levelOpPtr != NULL)
    delete m_levelOpPtr;
  m_levelOpPtr = NULL;
#ifdef HYPRE_AS_LEVELSOLVER
  clearHypre();
#endif
}

// solve on just this level, inhomogeneous bcs
void
LevelHypreSolver::levelSolve(LevelData<FArrayBox>& a_phi, 
                        const LevelData<FArrayBox>* a_phic, 
                        const LevelData<FArrayBox>& a_rhs,
                        bool a_initializePhiToZero)
{
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /* compute initial residual 
     -- can't use residual() fn because 
     don't want to zero out higher levels
     -- also note that we use LNf, 
     because this is a LEVEL solve
  */

  assert(m_levelOpPtr->isDefined());
  DataIterator dit = a_phi.dataIterator();
  if (a_initializePhiToZero) {
    //initial guess of phi is zero
    for(dit.reset(); dit.ok(); ++dit)
      a_phi[dit()].setVal(0.0);
  }

  m_levelOpPtr->applyOpI(a_phi, a_phic, m_resid);

  for(dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();
      m_corr[dit()].setVal(0.0);
    }

#ifdef HYPRE_AS_LEVELSOLVER
  if (m_hypreOn)
    {
      levelSolveHypre(m_corr, m_resid);
    }
  else
    {
#endif

  Vector<Real> initRes(m_resid.nComp());
  Vector<Real> oldRes(m_resid.nComp());

  bool done = true;
  
  for (int comp=0; comp<m_resid.nComp(); comp++) 
    {
      Interval thisInterval(comp,comp);
      initRes[comp] = norm(m_resid, thisInterval, 0);
      // initialze oldRes to be initRes as well
      oldRes[comp] = initRes[comp];
      
      // if initRes = 0, don't bother solving
      if (initRes[comp] >  0.01*m_tolerance) done=false;
    }

  if (done) return;
      
  int iter = 0;
  while (!done) 
    {
      m_levelMG.mgRelax(m_corr, m_resid, true);

      m_levelOpPtr->applyOpH(m_corr, m_scratch);

      for(dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }

      iter++;

      done = true;
      for (int comp=0; comp<m_resid.nComp(); comp++) 
        {
          Real currentRes = 0.0;
          Interval thisInterval(comp,comp);

          currentRes =norm(m_scratch, thisInterval, 0);

          if(m_verbose)
            pout() << "iteration #= " << iter 
                   << ": Max(res) = " << currentRes 
                   << endl;
          
          // DFM (9/20/02) third test is to catch case where
          // convergence hangs due to machine precision (or
          // solvability) issues.  This follows the approach
          // used in AMRSolver
          done = done && ((iter > m_maxIter)
                          || (currentRes <= m_tolerance*initRes[comp] )
                          || ((iter > m_minIter) 
                              && ((currentRes/oldRes[comp]) 
                                  > 1.0-m_operatorTolerance)));
          // save currentRes into oldRes
          oldRes[comp] = currentRes;
        } // end loop over components

    }  // end solver iterations

#ifdef HYPRE_AS_LEVELSOLVER
    }
#endif

  // update phi
  for(dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }
}

// solve on just this level, homogeneous bcs
void
LevelHypreSolver::levelSolveH(LevelData<FArrayBox>& a_phi, 
                         const LevelData<FArrayBox>& a_rhs,
                         bool a_initializePhiToZero)
{
  assert(isDefined());
  assert(a_phi.nComp() == a_rhs.nComp());
  assert(a_phi.nComp() == m_resid.nComp());

  /** compute initial residual 
      -- can't use residual() fn because 
      don't want to zero out higher levels
      -- also note that we use LNf, 
      because this is a LEVEL solve
  **/

  assert(m_levelOpPtr->isDefined());
  DataIterator dit = a_phi.dataIterator();
  if (a_initializePhiToZero) {
    //initial guess of phi is zero
    for(dit.reset(); dit.ok(); ++dit)
      a_phi[dit()].setVal(0.0);
  }

  m_levelOpPtr->applyOpH(a_phi, m_resid);

  for (dit.reset(); dit.ok(); ++dit)
    {
      m_resid[dit()] -= a_rhs[dit()];
      m_resid[dit()].negate();
      m_corr[dit()].setVal(0.0);
    }

#ifdef HYPRE_AS_LEVELSOLVER
  if (m_hypreOn)
    {
      levelSolveHypre(m_corr, m_resid);
    }
  else
    {
#endif

  Vector<Real> initRes(m_resid.nComp());
  Vector<Real> oldRes(m_resid.nComp());
  

  bool done = true;

  for (int comp=0; comp<m_resid.nComp(); comp++)
    {
      Interval thisInterval(comp,comp);
      initRes[comp] = norm(m_resid, thisInterval, 0);
      // also initialze oldRes to be initRes
      oldRes[comp] = initRes[comp];

  // if initRes = 0, don't bother solving
      if (initRes[comp] > 0.01*m_tolerance) done = false;
    }
  
  if (done) return;

  int iter = 0;
  while (!done) 
    {
      iter++;

      // compute modified residual LofPhi = res - LNf(corr)
      // use homogeneous form of ApplyOp because want to use 
      // homogeneous CF BC's for correction
      m_levelMG.mgRelax(m_corr, m_resid, true);

      m_levelOpPtr->applyOpH(m_corr, m_scratch);

      for (dit.reset(); dit.ok(); ++dit)
        {
          m_scratch[dit()] -= m_resid[dit()];
          m_scratch[dit()].negate();
        }

      done = true;
      for (int comp=0; comp<m_scratch.nComp(); comp++)
        {
          Interval thisInterval(comp,comp);
          Real currentRes = norm(m_scratch, thisInterval, 0);

          if(m_verbose)    
            pout() << "iteration #= " << iter 
                   << ": Max(res) = " << currentRes 
                   << endl;
          

          // DFM (9/20/02) third test is to catch case where
          // convergence hangs due to machine precision (or
          // solvability) issues.  This follows the approach
          // used in AMRSolver
          done = done && ((iter > m_maxIter)
                          || (currentRes <= m_tolerance*initRes[comp] )
                              || ((iter < m_minIter)
                                  && ((currentRes/oldRes[comp]) 
                                      > 1.0-m_operatorTolerance)));
            oldRes[comp] = currentRes;
        } // end loop over components
      
    }  // end solver iterations
  
#ifdef HYPRE_AS_LEVELSOLVER
    }
#endif

  // update phi
  for (dit.reset(); dit.ok(); ++dit)
    {
      a_phi[dit()] += m_corr[dit()];
    }
}
