/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// LevelProjTest.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S. Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#include <iostream>
#include "LevelHypreSolver.H"
#include "PoissonOp.H"
#include "ParmParse.H"
#include "DisjointBoxLayout.H"

#include "LoadBalance.H"
#include "LayoutIterator.H"
#include "PoissonBC.H"

#include "Timer.H"
Timer Everything("Everything");

#ifdef TIMER
// global variables
int IterationFlavor;
int BIterationFlavor;
CounterStruct NoelsSooperIterationCounter[MAX_ITERATION_COUNTERS];
#endif

#ifdef HDF5
#include "HDF5.H"
#include "AMRIO.H"
#include "FABView.H"
#endif


void readData(LevelData<FArrayBox> & a_rhs,
              Vector<LevelData<FArrayBox>* >& a_solnVect,
              ProblemDomain& a_domain,
              Real& a_dx, 
              int& a_nRefCrse,
              const string& a_filename);
              



int main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  Everything.start();
  //scoping trick
  {
#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
 
    char* filename = NULL;
    
    for (int i=1; i<argc; i++) {
      if (strlen(argv[i]) > 0 && argv[i][0] != '-') {
        filename = argv[i];
        argv[i] = "";
        break;
      }
    }
    

    // storage for RHS and solution.  Note that soln is two levels, 
    // due to the need for coarse-fine boundary conditions
    LevelData<FArrayBox> rhs;
    Vector<LevelData<FArrayBox>* > solnVect(2);
    ProblemDomain domain;
    Real dx;
    int nRefCrse;
    
    string fname(filename);
    
    readData(rhs, solnVect, domain, 
             dx, nRefCrse, fname);
    
    const DisjointBoxLayout& grids = rhs.getBoxes();
    const DisjointBoxLayout* crseGridsPtr = &(solnVect[0]->getBoxes());
    
    PoissonOp solverOp;
    
    DomainGhostBC physBC;
    SideIterator sit;
    Interval BCcomps = solnVect[0]->interval();
    for (int dir=0; dir<SpaceDim; dir++)
      {
        for (sit.reset(); sit.ok(); ++sit)
          {
            // example BC's are neumann bc's
            NeumannBC thisBC(dir, sit(), BCcomps);
            physBC.setBoxGhostBC(thisBC);
          }
      }
    
    solverOp.setDomainGhostBC(physBC);
    
    LevelHypreSolver projectionSolver(grids, crseGridsPtr, domain,
                                      dx, nRefCrse, &solverOp, 
                                      solnVect[0]->nComp());
    
    projectionSolver.setVerbose(true);

#ifdef HYPRE_AS_LEVELSOLVER
    int hypreSolverFlag = 1; // 0 for SMG solver, 1 for PFMG solver
    
    Timer timerHypreSetup;
    timerHypreSetup.start();
    if (hypreSolverFlag == 1) // if using PFMG, need more steps
      projectionSolver.setMaxIter(200);
    projectionSolver.defineHypre(hypreSolverFlag);
    timerHypreSetup.stop();
    double wcSetup = timerHypreSetup.wc_time();
    pout() << "hypre setup time:  " << wcSetup << endl;
    
    // if true, then use hypre solver.  if false, then use Chombo solver.
    bool useHypreSolver = true;
    projectionSolver.setHypre(useHypreSolver);
#endif
    
    

    // we'll do this explicitly ourselves, just to be sure
    bool initializeSolnToZero = false;
    //bool initializeSolnToZero = true;

    // temp storage, since we're doing this in residual-correction form
    // resid won't need ghost cells
    LevelData<FArrayBox> resid(grids, rhs.nComp());
    // corr will need ghost cells
    IntVect ghostVect = IntVect::TheUnitVector();
    LevelData<FArrayBox> corr(grids, solnVect[1]->nComp(), 
                              ghostVect);

    // define solverOp for computing residual
    bool homogeneousOnly = false;
    solverOp.define(grids, crseGridsPtr, dx, nRefCrse, 
                    domain, homogeneousOnly, corr.nComp());
    

    LevelData<FArrayBox>& soln = *solnVect[1];
    DataIterator dit = soln.dataIterator();

    // to get a better idea of timing, we can run this more than once
    int numIter = 2;
    for (int iter = 0; iter< numIter; iter++)
      {
        for (dit.begin(); dit.ok(); ++dit)
          {
            soln[dit()].setVal(0.0);
            corr[dit()].setVal(0.0);
          }
        
        pout() << endl;
        pout() << "solve number " << iter << endl;

        // put solution into residual-correction form, 
        // so that we can use homogeneous C/F BC's
        solverOp.applyOpI(soln, solnVect[0], resid);
        for (dit.reset(); dit.ok(); ++dit)
          {
            resid[dit()] -= rhs[dit()];
            resid[dit()].negate();
          }
        
        Timer timerSolve;
        timerSolve.start();
        projectionSolver.levelSolveH(corr, rhs, initializeSolnToZero);      
        timerSolve.stop();
        double wcSolve = timerSolve.wc_time();
        pout() << "levelSolve time:  " << wcSolve << endl;

        // add correction to solution
        for (dit.begin(); dit.ok(); ++dit)
          {
            soln[dit()] += corr[dit()];
          }
      }

    bool write_output = false;
    if (write_output)
      {
        // write out solution
        string filename("outfile.hdf5");
        Vector<DisjointBoxLayout> vectGrids(2);
        vectGrids[0] = rhs.getBoxes();
        vectGrids[1] = solnVect[0]->getBoxes();
        Vector<string> vectName(1);
        string solnname("soln");
        vectName[0] = solnname;
        Box coarseDomain = coarsen(domain.domainBox(), nRefCrse);
        Vector<int> vectRatio(2);
        vectRatio[0] = nRefCrse;
        vectRatio[1] = nRefCrse;

        WriteAMRHierarchyHDF5(filename,
                              vectGrids,
                              solnVect,
                              vectName,
                              coarseDomain,
                              dx, // should be at coarse level
                              0., 0., // dt and time, not used
                              vectRatio,
                              2); // 2 levels
      }

    // clean up memory
    delete solnVect[0];
    delete solnVect[1];
    
  }//end scoping trick
  Everything.stop();
#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
  
}


void 
readData(LevelData<FArrayBox> & a_rhs,
         Vector<LevelData<FArrayBox>* >& a_solnVect,
         ProblemDomain& a_domain,
         Real& a_dx, 
         int& a_nRefCrse,
         const string& a_filename)
{

  Vector<DisjointBoxLayout> vectGrids;
  Vector<LevelData<FArrayBox>*> vectData;
  Vector<string> vectNames;
  Real dt, time;
  Vector<int> nRefVect;
  int numLevels;
  IntVect ghostVect(IntVect::TheZeroVector());
  Box domainBox;


  ReadAMRHierarchyHDF5(a_filename,
                       vectGrids, 
                       vectData,
                       vectNames,
                       domainBox,
                       a_dx,
                       dt,
                       time,
                       nRefVect,
                       numLevels,
                       ghostVect);


  assert (numLevels = 2);

  a_nRefCrse = nRefVect[0];
  // a_dx is now on level0 -- want it to be on level 1
  a_dx = a_dx/a_nRefCrse;

  // construct problem domains here
  ProblemDomain crseDomain(domainBox);
  // to mirror ring problem, only set periodic in the z-direction.
  crseDomain.setPeriodic(SpaceDim-1, true);

  // now refine up to fine-level domain
  a_domain = crseDomain;
  a_domain.refine(a_nRefCrse);
  
  // since we read in DisjointBoxLayouts which will not have
  // ProblemDomain (periodicity info), need to generate
  // real disjointBoxLayouts which include periodicity info

  // first do coarse-level:
  int numGridsCrse = vectGrids[0].size();
  Vector<Box> crseBoxes(numGridsCrse);
  Vector<int> crseProcAssign(numGridsCrse,0);

  LayoutIterator crseSrcLit = vectGrids[0].layoutIterator();
  int crseBox = 0;

  for (crseSrcLit.begin(); crseSrcLit.ok(); crseBox++, ++crseSrcLit)
    {
      crseBoxes[crseBox] = vectGrids[0].get(crseSrcLit());
      crseProcAssign[crseBox] = vectGrids[0].procID(crseSrcLit());
    }
  // define periodic version of crse grids
  DisjointBoxLayout crseGrids(crseBoxes, crseProcAssign, crseDomain);

  // now define fine-level
  int numGridsFine = vectGrids[1].size();
  Vector<Box> fineBoxes(numGridsFine);
  Vector<int> fineProcAssign(numGridsFine,0);

  LayoutIterator fineSrcLit = vectGrids[1].layoutIterator();
  int fineBox = 0;

  for (fineSrcLit.begin(); fineSrcLit.ok(); fineBox++, ++fineSrcLit)
    {
      fineBoxes[fineBox] = vectGrids[1].get(fineSrcLit());
      fineProcAssign[fineBox] = vectGrids[1].procID(fineSrcLit());
    }
  // define periodic version of fine grids
  DisjointBoxLayout fineGrids(fineBoxes, fineProcAssign, a_domain);


  // loop over variables in data
  for (int comp=0; comp<vectNames.size(); comp++)
    {
      if (vectNames[comp] == "RHS")
        {      
          a_rhs.define(fineGrids, 1, IntVect::TheZeroVector());
          Interval rhsDataInterval(comp,comp);
          Interval rhsInterval(0,0);
          vectData[1]->copyTo(rhsDataInterval, a_rhs, rhsInterval);
        }

      // soln is called "pi" because it's from a level projection in 
      // the navier-stokes code
      else if (vectNames[comp] == "pi")
        {
          // this is defined on the fine and the coarse levels.
          a_solnVect[0] = new LevelData<FArrayBox>(crseGrids,
                                                 1, 
                                                 IntVect::TheUnitVector());

          a_solnVect[1] = new LevelData<FArrayBox>(fineGrids,
                                                 1, 
                                                 IntVect::TheUnitVector());

          Interval solnDataInterval(comp, comp);
          Interval solnInterval(0,0);
          
          vectData[0]->copyTo(solnDataInterval, *a_solnVect[0],
                              solnInterval);

          vectData[1]->copyTo(solnDataInterval, *a_solnVect[1],
                              solnInterval);
          
        }
    }

  // now clean up temp storage
  for (int lev=0; lev<vectData.size(); lev++)
    {
      if (vectData[lev] != NULL)
        {
          delete vectData[lev];
          vectData[lev] = NULL;
        }
    }
}
          

          
  
