/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// MacProjTest.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#include <iostream.h>
#include "CCProjector.H"
#include "ParmParse.H"
#include "DisjointBoxLayout.H"
#include "LoadBalance.H"
#include "Divergence.H"
#include "FluxBox.H"
#include "computeNorm.H"
#include "LayoutIterator.H"
#include "PhysBCUtil.H"
#include "MacProjTestF_F.H"


void buildGrids(Vector<DisjointBoxLayout>& a_amrGrids,
                Vector<int>& a_vectRefRatio,
                Vector<Real>& a_vectDx,
		ProblemDomain& a_baseDomain);


void readGrids(Vector<Vector<Box> >& a_vectGrids,
               Vector<int>& a_vectRefRatio,
               const string& a_gridfile);

void buildDefaultGrids(Vector<Vector<Box> >& a_vectGrids,
                       Vector<int> a_vectRefRatio);

void initData(Vector<LevelData<FArrayBox>* >& vectFaceVel,
              Vector<int>& a_vectRefRatio, 
              Vector<Real>& a_vectDx, 
              ProblemDomain& a_baseDomain);


int main(int argc, char* argv[])
{
 
  char* in_file = NULL;
  char* pp_prefix = NULL;

  for (int i=1; i<argc; i++) {
    if (strlen(argv[i]) > 0 && argv[i][0] != '-') {
      in_file = argv[i];
      argv[i] = "";
      break;
    }
  }

  ParmParse pp(argc-1, argv+1, pp_prefix, in_file);

  // want to test both level 0 and level 1 projections to ensure 
  // that C/F BC's work too.
  
  bool testResult = true;
  
  pout() << "initializing data..." << endl;
  Vector<DisjointBoxLayout> vectGrids;
  Vector<int> vectRefRatio;
  Vector<Real> vectDx;
  ProblemDomain baseDomain;

  buildGrids(vectGrids, vectRefRatio, vectDx, baseDomain);
  
  // now allocate what we'll need for tests.
  int numLevels = vectGrids.size();
  Vector<LevelData<FluxBox>* > vectFaceVel(numLevels, NULL);
  Vector<LevelData<FArrayBox>* > vectDiv(numLevels, NULL);
  Vector<CCProjector*> vectProj(numLevels, NULL);
  
  // first allocate projectors
  for (int lev=0; lev<numLevels; lev++)
    {
      vectProj[lev] = new CCProjector();
    }
  

  // use basic physical BC's for this test
  PhysBCUtil physBC;

  CCProjector* crseProjPtr = NULL;
  CCProjector* fineProjPtr = NULL;
  ProblemDomain levelDoman = baseDomain;
  int nRefCrse = 0;

  for (int lev=0; lev< numLevels; lev++) 
    {
      const DisjointBoxLayout& levelGrids = vectGrids[lev];
      // only need a single component, since we're doing MAC projection
      vectFaceVel[lev] = new LevelData<FluxBox>(levelGrids, 1);
      vectDiv[lev] = new LevelData<FArrayBox>(levelGrids,1);
      
      if (lev<numLevels-1) fineProjPtr = vectProj[lev+1];


      vectProj[lev]->define(levelGrids, levelDomain,
                            vectDx[lev], fineProjPtr,
                            crseProjPtr, nRefCrse,
                            lev, physBC);

      // update stuff for next level
      levelDomain.refine(vectRefRatio[lev]);
      crseProjPtr = vectProj[lev];
      nRefCrse = vectRefRatio[lev];
    }
     
  // initialize face-centered velocity
  initData(vectFaceVel, vectRefRatio, vectDx, baseDomain);

  // now do MAC projections
  Real time = 1.0;
  Real dt = 0.5;
  int passedTests = 0;
  // this is somewhat arbitrary 
  // make this something around solver tolerance
  Real maxAllowableDiv = 1.0e-14;

  pout() << "testing projector" << endl;
  for (int lev=0; lev<numLevels; lev++) 
    {
      LevelData<FluxBox>& levelFaceVel = *vectFaceVel[lev];
      LevelData<FArrayBox>& levelDiv = *vectDiv[lev];
      // first compute div(vel)
      levelDivergenceMAC(levelDiv, levelFaceVel, vectDx[lev]);

      // compute max norm of divergence
      Real normDivBefore = norm(levelDiv, vectDx[lev], 0);
      
      vectProj[lev]->levelMacProject(levelFaceVel, time, dt);
      
      // now compute divergence of projected velocities
      levelDivergenceMAC(levelDiv, levelFaceVel, vectDx[lev]);

      Real normDivAfter = norm(levelDiv, vectDx[lev], 0);

      pout () << "Level " << lev << " Max(div) = " << normDivBefore 
              << " before MAC projection, " normDivAfter
              << " after MAC projection"    << endl;

      if (normDivAfter > maxAllowableDiv) passedTests += (lev+1);

    }

  if (passedTests == 0)
    {
      pout () "Mac Projection Test: passed all tests" << endl;
    }
  else
    {
      pout() "Mac Projection Test: FAILED with code " 
        << passedTests << endl;
    }
  
}

// -------------------------------------------------------------
void buildGrids(Vector<DisjointBoxLayout>& a_vectGrids,
                Vector<int>& a_vectRefRatio,
                Vector<Real>& a_vectDx,
		ProblemDomain& dProblem0)
{

  Vector<Vector<Box> > vectBoxes;

  // read grids from file
  if (pp.contains("gridFile"))
    {
      string gridfile;
      ppMain.get("gridfile", gridfile);
      
      readGrids(vectBoxes, a_vectRefRatio, gridfile);

    }
  else
    {
      buildDefaultGrids(vectBoxes, a_vectRefRatio);
    }
  
  // now set up domain, compute dx, do load balancing, etc.

  // build base domain
  // if level 0 has more than one box, need to construct domain box
  Box domainBox;
  if (vectBoxes[0].size() == 1)
    {
      domainBox = vectBoxes[0][0];
    }
  else
    {
      assert(vectBoxes[0].size() > 1);
      
      IntVectSet baseDomain(vectBoxes[0][0]);
      for (int boxno = 1; boxno < vectBoxes[0].size(); ++boxno)
        {
          baseDomain |= vectBoxes[0][boxno];
        }
      
      domainBox = baseDomain.minBox();
    }
  
  // for now, assume non-periodic domains.
  dProblem0.define(domainBox);
  
  // also, assume that domain size = 1
  int nx = domainBox.size(0);
  a_vectDx[0] = 1.0/(nx);
  

  // load balance base level
  Vector<int> procAssign0(vectBoxes[0].size());
  loadBalance(procAssign0, vectBoxes[0]);

  a_vectGrids[0].define(vectBoxes[0], procAssign0, dProblem0);

  ProblemDomain thisDomain = dProblem0;

  // now do finer levels
  for (int lev=1; lev<vectBoxes.size(); lev++)
    {
      // build stuff from coarser levels
      thisDomain.refine(a_vectRefRatio[lev-1]);
      a_vectDx[lev] = a_vectDx[lev-1]/a_vectRefRatio[lev-1];
      
      // now load balance this level
      Vector<int> procAssign(vectBoxes[lev].size());
      loadBalance(procAssign, vectBoxes[lev]);
      a_vectGrids[lev].define(vectBoxes[lev], procAssign, thisDomain);
    }

}



void initData(Vector<LevelData<FluxBox>* >& vectFaceVel,
              Vector<int>& a_vectRefRatio, 
              Vector<Real>& a_vectDx, 
              ProblemDomain& a_baseDomain)
{
  // initialize data
  for (int lev = 0; lev<vectFaceVel.size(); lev++)
    {
      LevelData<FluxBox>& levelVel = vectFaceVel[lev];
      Real levelDx = a_vectDx[lev];

      DataIterator levelDit = levelVel.dataIterator();
      for (levelDit.begin(); levelDit.ok(); ++levelDit)
        {
          FluxBox& thisVel = levelVel[dit()];
          
          for (int dir=0; dir<SpaceDim; dir++)
            {
              FORT_INITFACEVEL(CHF_FRA(thisVel[dir]),
                               CHF_REAL(levelDx),
                               CHF_INT(dir));
            } // end loop over directions
        } // end loop over grids
    } // end loop over levels
}

void
readGrids(Vector<Vector<Box> >& a_vectBoxes, 
          Vector<int>& a_vectRefRatio,
          string& a_gridfile)
{
  if (procID() == 0 ) 
    {
      istream is(a_gridfile.c_str(), ios::in);
      if (is.fail())
        {
          pout() << "cannot open grids file " << a_gridfile << endl;
          MayDay::Error();
        }
      
      
      
#ifdef MPI 
      if (verbosity >= 3) {
        pout() << "procID: " << procID() << "  opening gridfile \n" << endl;
      }
#endif      
      
      // format of file -- number of levels, then for
      // each level starting with level 0, refinement ratio, number of
      // grids on level, list of boxes
      is >> a_vectRefRatio[lev];
      int in_numLevels;
      is >> in_numLevels;
      assert (in_numLevels <= max_level+1);
      if (verbosity >= 3) 
        {
          pout() << "numLevels = " << in_numLevels << endl;
        }

      while (is.get() != '\n');
      amrGrids.resize(in_numLevels);
      
      // now loop over levels, starting with level 1
      int ngrid;
      for (int lev=0; lev<in_numLevels; lev++) 
        {
          is >> ngrid;
          if (verbosity >= 3) 
            {
              pout() << "level " << lev << " numGrids = " << ngrid << endl;
              pout() << "Grids: ";
            }

          while (is.get() != '\n');
          amrGrids[lev].resize(ngrid);
          for (int i=0; i<ngrid; i++) 
            {
              Box bx;
              is >> bx;
              while (is.get() != '\n');
              // quick check on box size
              Box bxRef(bx);
              
              amrGrids[lev][i] = bx;
            } // end loop over boxes on this level
        } // end loop over levels
      
#ifdef MPI
    } // end if procID = 0
  NST_MPI_readInput.start();
  //broadcast (amrGrids, uniqueProc(SerialTask::compute));
  MPI_Barrier(Chombo_MPI::comm);
  broadcast (amrGrids, 0);
  MPI_Barrier(Chombo_MPI::comm);
  NST_MPI_readInput.stop();
#endif
  
}


void buildDefaultGrids(Vector<Vector<Box> >& a_vectGrids,
                       Vector<int>& a_vectRefRatio)
{
  // build simple grids for default case
  int baseDomainSize = 32;
  Box baseDomBox(IntVect::TheZeroVector(), 
                 (baseDomainSize-1)*IntVect::TheUnitVector());
  
  int baseRefRatio = 2;
  int numLevels = 3;

  a_vectGrids.resize(numLevels);
  a_vectRefRatio.resize(numLevels, baseRefRatio);

  a_vectGrids[0].push_back(baseDomBox);

  // to make it simpler to define grids, define all levels 
  // on the coarsest level, then refine.  this also makes it simpler
  // to change refinement ratio later.

  a_vectGrids[1].resize(3);
  
  int halfWayPoint = (baseDomainSize/2);
  int quarterPoint = (baseDomainSize/4);
  int threeQuarterPoint = (3*baseDomainSize/4);

  // do this in its own scope -- level 1
  { 

    Box box0(IntVect::TheZeroVector(), 
             (halfWayPoint-1)*IntVect::TheUnitVector());

    a_vectGrids[1].push_back(box0);
    
    Box box1(halfWayPoint*IntVect::TheUnitVector(),
             (threeQuarterPoint-1)*IntVect::TheUnitVector());

    a_vectGrids[1].push_back(box1);
    
    box1.shift(-quarterPoint*IntVect::BASISV(0));

    a_vectGrids[1].push_bac(box1);

  }

  // do level 2 in its own scope as well


}

  
