/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// CoarseAverageEdge.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "CoarseAverageEdge.H"
#include "AverageEdgeF_F.H"


#include "Timer.H"
extern Timer Everything;
Timer TF_AVERAGEEDGE  ("AVERAGEEDGE",     Everything, DTIMEFORT, true);

Timer TMPI_averageToCoarse_cp ("avgToCoar cp", Everything, DTIMECOMM, true);

// ----------------------------------------------------------
CoarseAverageEdge::CoarseAverageEdge() :m_isDefined(false)
{
}


// ----------------------------------------------------------
CoarseAverageEdge::~CoarseAverageEdge()
{
}

// ----------------------------------------------------------
CoarseAverageEdge::CoarseAverageEdge(const DisjointBoxLayout& a_fineGrids,
				     int a_nComp, int a_nRef) 
  : m_isDefined(false)
{
  define(a_fineGrids, a_nComp, a_nRef);
}

// ----------------------------------------------------------
void
CoarseAverageEdge::define(const DisjointBoxLayout& a_fineGrids,
			  int a_nComp, int a_nRef)
{
  m_nRef = a_nRef;
  
  DisjointBoxLayout coarsened_fine_domain;
  coarsen(coarsened_fine_domain, a_fineGrids, m_nRef);
  m_coarsenedFineData.define(coarsened_fine_domain, a_nComp);

  m_isDefined = true;
}

// ----------------------------------------------------------
bool 
CoarseAverageEdge::isDefined() const
{
  return m_isDefined;
}

// ----------------------------------------------------------
// this function is shamelessly based on the ANAG CoarseAverage 
// (cell-centered) version
void
CoarseAverageEdge::averageToCoarse(LevelData<FluxBox>& a_coarseData,
				   const LevelData<FluxBox>& a_fineData)
{
  assert(isDefined());

  DataIterator dit = a_fineData.dataIterator();
  for (dit.reset(); dit.ok(); ++dit) {
    FluxBox& coarsenedFine = m_coarsenedFineData[dit()];
    const FluxBox& fine = a_fineData[dit()];
    
    // coarsen from the entire fine grid onto the entire coarse grid
    averageGridData(coarsenedFine, fine);
  }

  TMPI_averageToCoarse_cp.start();
  m_coarsenedFineData.copyTo(m_coarsenedFineData.interval(),
			     a_coarseData, a_coarseData.interval());
  TMPI_averageToCoarse_cp.stop();

}


// ----------------------------------------------------------
void
CoarseAverageEdge::averageGridData(FluxBox& a_coarsenedFine, 
				   const FluxBox& a_fine) const 
{
  
  for (int dir=0; dir<SpaceDim; dir++) {
    FArrayBox& coarseFab = a_coarsenedFine[dir];
    const FArrayBox& fineFab = a_fine[dir];

    const Box& coarseBox = coarseFab.box();

    // set up refinement box
    int boxHi = m_nRef-1;
    IntVect hiVect(D_DECL(boxHi,boxHi,boxHi));
    // don't want to index at all in dir direction --
    // instead, want to just march along edge.
    hiVect.setVal(dir,0);
    IntVect loVect(D_DECL(0,0,0));
    Box refBox(loVect, hiVect);    


    TF_AVERAGEEDGE.start();
    FORT_AVERAGEEDGE( CHF_FRA(coarseFab),
		      CHF_CONST_FRA(fineFab),
		      CHF_BOX(coarseBox),
		      CHF_CONST_INT(dir),
		      CHF_CONST_INT(m_nRef),
		      CHF_BOX(refBox));
    TF_AVERAGEEDGE.stop();
  }

}
  
