/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AmrSolver.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _Divergence_H_
#define _Divergence_H_

#include "LevelData.H"
#include "FArrayBox.H"
#include "FluxBox.H"
#include "LevelFluxRegister.H"
#include "QuadCFInterp.H"

/*@ManMemo: computes cell-centered level-operator divergence of 
  cell-centered vector field u; if uCrse != NULL, does coarse-fine
  boundary conditions for u -- if  quadInterp == true, uses quadratic
  coarse-fine boundary conditions, otherwise, use extrap BC's */
void levelDivergenceCC(
		       ///
		       LevelData<FArrayBox>& a_div, 
		       ///
		       LevelData<FArrayBox>& a_u, 
		       ///
		       LevelData<FArrayBox>* a_uCrsePtr, 
		       ///
		       const Real a_dx,
		       ///
		       const int a_nRefCrse,
		       ///
		       const ProblemDomain& a_dProblem,
		       ///
		       const bool a_quadInterp);

/*@ManMemo: computes cell-centered level-operator divergence of 
  cell-centered vector field u; if uCrse != NULL, does coarse-fine
  boundary conditions for u -- if  quadInterp == true, uses quadratic
  coarse-fine boundary conditions, otherwise, use extrap BC's.
  This (deprecated) interface uses a Box instead of a ProblemDomain. */
void levelDivergenceCC(
		       ///
		       LevelData<FArrayBox>& a_div, 
		       ///
		       LevelData<FArrayBox>& a_u, 
		       ///
		       LevelData<FArrayBox>* a_uCrsePtr, 
		       ///
		       const Real a_dx,
		       ///
		       const int a_nRefCrse,
		       ///
		       const Box& a_dProblem,
		       ///
		       const bool a_quadInterp);

/*@ManMemo: computes cell-centered level-operator divergence of 
  cell-centered vector field u; if uCrse != NULL, does coarse-fine
  boundary conditions for u -- if  quadInterp == true, uses quadratic
  coarse-fine boundary conditions, otherwise, use extrap BC's.  This
  one takes a pre-allocated QuadCFInterp object as an argument */
void levelDivergenceCC(
		       ///
		       LevelData<FArrayBox>& a_div, 
		       ///
		       LevelData<FArrayBox>& a_u, 
		       ///
		       LevelData<FArrayBox>* a_uCrsePtr, 
		       ///
		       const Real a_dx,
		       ///
		       const bool a_quadInterp, 
		       ///
		       QuadCFInterp& a_cfInterp
		       );


/*@ManMemo: computes cell-centered composite-operator divergence
  of cell-centered vector field u; uses same coarse-fine BC's as
  levelDivergenceCC -- if uFine != NULL, also does flux-matching
  BC with finer level uFine, using same coarse-fine BC's as used
  for coarse level BC. */
void compDivergenceCC(
		      ///
		      LevelData<FArrayBox>& a_div,
		      ///
		      LevelData<FArrayBox>& a_u,
		      ///
		      LevelData<FArrayBox>* a_uCrsePtr,
		      ///
		      LevelData<FArrayBox>* a_uFinePtr,
		      ///
		      const Real a_dx, 
		      ///
		      const int a_nRefCrse,
		      ///
		      const int a_nRefFine,
		      ///
		      const ProblemDomain& a_dProblem,
		      ///
		      const bool a_quadInterp);

/*@ManMemo: computes cell-centered composite-operator divergence
  of cell-centered vector field u; uses same coarse-fine BC's as
  levelDivergenceCC -- if uFine != NULL, also does flux-matching
  BC with finer level uFine, using same coarse-fine BC's as used
  for coarse level BC. This (deprecated) interface uses a Box instead 
  of a ProblemDomain */
void compDivergenceCC(
		      ///
		      LevelData<FArrayBox>& a_div,
		      ///
		      LevelData<FArrayBox>& a_u,
		      ///
		      LevelData<FArrayBox>* a_uCrsePtr,
		      ///
		      LevelData<FArrayBox>* a_uFinePtr,
		      ///
		      const Real a_dx, 
		      ///
		      const int a_nRefCrse,
		      ///
		      const int a_nRefFine,
		      ///
		      const Box& a_dProblem,
		      ///
		      const bool a_quadInterp);


/*@ManMemo: computes cell-centered composite-operator divergence
  of cell-centered vector field u; uses same coarse-fine BC's as
  levelDivergenceCC -- if uFine != NULL, also does flux-matching
  BC with finer level uFine, using same coarse-fine BC's as used
  for coarse level BC. In this one, pre-allocated LevelFluxRegister
  and QuadCFInterp objects are passed in as an argument (saves 
  allocation inside fn.)*/
void compDivergenceCC(
		      ///
		      LevelData<FArrayBox>& a_div,
		      ///
		      LevelData<FArrayBox>& a_u,
		      ///
		      LevelData<FArrayBox>* a_uCrsePtr,
		      ///
		      LevelData<FArrayBox>* a_uFinePtr,
		      ///
		      const Real a_dx, 
		      ///
		      const int a_nRefFine,
		      ///
		      const int a_nRefCrse,
		      ///
		      const ProblemDomain& a_dProblem,
		      ///
		      const bool a_quadInterp,
		      ///
		      LevelFluxRegister* a_fluxRegFinePtr,
		      ///
		      QuadCFInterp& a_cfInterpCrse,
		      ///
		      QuadCFInterp& a_cfInterpFine);


/*@ManMemo: computes cell-centered composite-operator divergence
  of cell-centered vector field u; uses same coarse-fine BC's as
  levelDivergenceCC -- if uFine != NULL, also does flux-matching
  BC with finer level uFine, using same coarse-fine BC's as used
  for coarse level BC. In this one, pre-allocated LevelFluxRegister
  and QuadCFInterp objects are passed in as an argument (saves 
  allocation inside fn.)  This (deprecated) interface uses a 
  Box instead of a ProblemDomain. */
void compDivergenceCC(
		      ///
		      LevelData<FArrayBox>& a_div,
		      ///
		      LevelData<FArrayBox>& a_u,
		      ///
		      LevelData<FArrayBox>* a_uCrsePtr,
		      ///
		      LevelData<FArrayBox>* a_uFinePtr,
		      ///
		      const Real a_dx, 
		      ///
		      const int a_nRefFine,
		      ///
		      const int a_nRefCrse,
		      ///
		      const Box& a_dProblem,
		      ///
		      const bool a_quadInterp,
		      ///
		      LevelFluxRegister* a_fluxRegFinePtr,
		      ///
		      QuadCFInterp& a_cfInterpCrse,
		      ///
		      QuadCFInterp& a_cfInterpFine);


/*@ManMemo: computes cell-centered level-operator divergence of 
  edge-centered vector field uEdge; assumes all coarse-fine BC's
  have already been set. */
void levelDivergenceMAC(
			///
			LevelData<FArrayBox>& a_div,
			///
			const LevelData<FluxBox>& a_uEdge, 
			///
			const Real a_dx);


/// really basic single-fab MAC divergence
void simpleDivergenceMAC(///
			 FArrayBox& a_div,
			 ///
			 const FluxBox& a_uEdge,
			 /// 
			 const Real a_dx);

/*@ManMemo: computes composite cell-centered divergence of edge-centered
  vector field uEdge; if finer level data uEdgeFine exists, use 
  flux-matching condition to compute divergence along coarse-fine
  interface. */
void compDivergenceMAC(
		       ///
		       LevelData<FArrayBox>& a_div,
		       ///
		       LevelData<FluxBox>& a_uEdge,
		       ///
		       LevelData<FluxBox>* a_uEdgeFinePtr, 
		       ///
		       const Real a_dx,
		       ///
		       const Real* a_dxFine,
		       ///
		       const int a_nRefFine,
		       ///
		       const ProblemDomain& a_dProblem);

/*@ManMemo: computes composite cell-centered divergence of edge-centered
  vector field uEdge; if finer level data uEdgeFine exists, use 
  flux-matching condition to compute divergence along coarse-fine
  interface. This (deprecated) interface uses a Box instead of a
  ProblemDomain */
void compDivergenceMAC(
		       ///
		       LevelData<FArrayBox>& a_div,
		       ///
		       LevelData<FluxBox>& a_uEdge,
		       ///
		       LevelData<FluxBox>* a_uEdgeFinePtr, 
		       ///
		       const Real a_dx,
		       ///
		       const Real* a_dxFine,
		       ///
		       const int a_nRefFine,
		       ///
		       const Box& a_dProblem);

/*@ManMemo: just like normal compDivergenceMAC, but pass in a predefined 
  flux register (more efficient) */
void compDivergenceMAC(
		       /// 
		       LevelData<FArrayBox>& a_div,
		       ///
		       LevelData<FluxBox>& a_uEdge,
		       ///
		       LevelData<FluxBox>* a_uEdgeFinePtr,
		       ///
		       LevelFluxRegister* a_fluxRegPtr,
		       ///
		       const Real a_dx,
		       ///
		       const Real* a_dxFine,
		       ///
		       const int a_nRefine,
		       ///
		       const ProblemDomain& a_dProblem);



/*@ManMemo: just like normal compDivergenceMAC, but pass in a predefined 
  flux register (more efficient). This (deprecated) interface uses a
  Box instead of a ProblemDomain. */
void compDivergenceMAC(
		       /// 
		       LevelData<FArrayBox>& a_div,
		       ///
		       LevelData<FluxBox>& a_uEdge,
		       ///
		       LevelData<FluxBox>* a_uEdgeFinePtr,
		       ///
		       LevelFluxRegister* a_fluxRegPtr,
		       ///
		       const Real a_dx,
		       ///
		       const Real* a_dxFine,
		       ///
		       const int a_nRefine,
		       ///
		       const Box& a_dProblem);
		       


#endif


